/* Generated by Cython 0.29.13 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "celiagg/font.h",
            "celiagg/graphics_state.h",
            "celiagg/image.h",
            "celiagg/ndarray_canvas.h",
            "celiagg/paint.h",
            "celiagg/vertex_source.h"
        ],
        "include_dirs": [
            "./celiagg"
        ],
        "language": "c++",
        "name": "celiagg._celiagg",
        "sources": [
            "/home/runner/work/celiagg/celiagg/celiagg/_celiagg.pyx"
        ]
    },
    "module_name": "celiagg._celiagg"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_13"
#define CYTHON_HEX_VERSION 0x001D0DF0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__celiagg___celiagg
#define __PYX_HAVE_API__celiagg___celiagg
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include "font.h"
#include "image.h"
#include "graphics_state.h"
#include "paint.h"
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <vector>
#include "agg_trans_affine.h"
#include "vertex_source.h"
#include "ndarray_canvas.h"
#include "pythread.h"
#include <stdlib.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "celiagg/_celiagg.pyx",
  "celiagg/graphics_state.pxi",
  "celiagg/vertex_source.pxi",
  "celiagg/font.pxi",
  "stringsource",
  "celiagg/image.pxi",
  "celiagg/ndarray_canvas.pxi",
  "celiagg/paint.pxi",
  "celiagg/transform.pxi",
  "celiagg/conversion.pxi",
  "__init__.pxd",
  "type.pxd",
};
/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && __GNUC__ >= 4 && (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL >= 2)) &&\
                    !defined(__i386__)
    #define __pyx_atomic_incr_aligned(value, lock) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value, lock) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && 0
    #include <Windows.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type LONG
    #define __pyx_atomic_incr_aligned(value, lock) InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#elif CYTHON_ATOMICS && (defined(__ICC) || defined(__INTEL_COMPILER)) && 0
    #define __pyx_atomic_incr_aligned(value, lock) _InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) _InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using Intel atomics"
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":779
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":786
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":791
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":806
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":809
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "celiagg/ndarray_canvas.pxi":27
 * cimport numpy
 * 
 * ctypedef unsigned char _bytes_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef _ndarray_canvas.ndarray_canvas_base canvas_base_t
 */
typedef unsigned char __pyx_t_7celiagg_8_celiagg__bytes_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_7celiagg_8_celiagg_Font;
struct __pyx_obj_7celiagg_8_celiagg_Rect;
struct __pyx_obj_7celiagg_8_celiagg_GraphicsState;
struct __pyx_obj_7celiagg_8_celiagg_Image;
struct __pyx_obj_7celiagg_8_celiagg_CanvasBase;
struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128;
struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32;
struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32;
struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24;
struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16;
struct __pyx_obj_7celiagg_8_celiagg_CanvasG8;
struct __pyx_obj_7celiagg_8_celiagg_Paint;
struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint;
struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint;
struct __pyx_obj_7celiagg_8_celiagg_PatternPaint;
struct __pyx_obj_7celiagg_8_celiagg_SolidPaint;
struct __pyx_obj_7celiagg_8_celiagg_Transform;
struct __pyx_obj_7celiagg_8_celiagg_VertexSource;
struct __pyx_obj_7celiagg_8_celiagg_BSpline;
struct __pyx_obj_7celiagg_8_celiagg_Path;
struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints;
struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__;
struct __pyx_obj___Pyx_EnumMeta;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":817
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":819
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_7celiagg_8_celiagg_gray2rgb;
struct __pyx_opt_args_7celiagg_8_celiagg_rgb2rgb;
struct __pyx_opt_args_7celiagg_8_celiagg_gray2gray;
struct __pyx_opt_args_7celiagg_8_celiagg_rgb2gray;

/* "celiagg/enums.pxi":25
 * # Authors: John Wiggins
 * 
 * cpdef enum PixelFormat:             # <<<<<<<<<<<<<<
 *     Gray8 = _enums.k_PixelFormatGray8
 *     Gray16 = _enums.k_PixelFormatGray16
 */
enum __pyx_t_7celiagg_8_celiagg_PixelFormat {

  /* "celiagg/enums.pxi":46
 *     RGBA128 = _enums.k_PixelFormatRGBA128
 *     ARGB128 = _enums.k_PixelFormatARGB128
 *     ABGR128 = _enums.k_PixelFormatABGR128             # <<<<<<<<<<<<<<
 * 
 * cpdef enum InnerJoin:
 */
  __pyx_e_7celiagg_8_celiagg_Gray8 = k_PixelFormatGray8,
  __pyx_e_7celiagg_8_celiagg_Gray16 = k_PixelFormatGray16,
  __pyx_e_7celiagg_8_celiagg_Gray32 = k_PixelFormatGray32,
  __pyx_e_7celiagg_8_celiagg_BGR24 = k_PixelFormatBGR24,
  __pyx_e_7celiagg_8_celiagg_RGB24 = k_PixelFormatRGB24,
  __pyx_e_7celiagg_8_celiagg_BGR48 = k_PixelFormatBGR48,
  __pyx_e_7celiagg_8_celiagg_RGB48 = k_PixelFormatRGB48,
  __pyx_e_7celiagg_8_celiagg_BGR96 = k_PixelFormatBGR96,
  __pyx_e_7celiagg_8_celiagg_RGB96 = k_PixelFormatRGB96,
  __pyx_e_7celiagg_8_celiagg_BGRA32 = k_PixelFormatBGRA32,
  __pyx_e_7celiagg_8_celiagg_RGBA32 = k_PixelFormatRGBA32,
  __pyx_e_7celiagg_8_celiagg_ARGB32 = k_PixelFormatARGB32,
  __pyx_e_7celiagg_8_celiagg_ABGR32 = k_PixelFormatABGR32,
  __pyx_e_7celiagg_8_celiagg_BGRA64 = k_PixelFormatBGRA64,
  __pyx_e_7celiagg_8_celiagg_RGBA64 = k_PixelFormatRGBA64,
  __pyx_e_7celiagg_8_celiagg_ARGB64 = k_PixelFormatARGB64,
  __pyx_e_7celiagg_8_celiagg_ABGR64 = k_PixelFormatABGR64,
  __pyx_e_7celiagg_8_celiagg_BGRA128 = k_PixelFormatBGRA128,
  __pyx_e_7celiagg_8_celiagg_RGBA128 = k_PixelFormatRGBA128,
  __pyx_e_7celiagg_8_celiagg_ARGB128 = k_PixelFormatARGB128,
  __pyx_e_7celiagg_8_celiagg_ABGR128 = k_PixelFormatABGR128
};

/* "celiagg/enums.pxi":48
 *     ABGR128 = _enums.k_PixelFormatABGR128
 * 
 * cpdef enum InnerJoin:             # <<<<<<<<<<<<<<
 *     InnerBevel = _enums.InnerBevel
 *     InnerMiter = _enums.InnerMiter
 */
enum __pyx_t_7celiagg_8_celiagg_InnerJoin {

  /* "celiagg/enums.pxi":52
 *     InnerMiter = _enums.InnerMiter
 *     InnerJag = _enums.InnerJag
 *     InnerRound = _enums.InnerRound             # <<<<<<<<<<<<<<
 * 
 * cpdef enum LineJoin:
 */
  __pyx_e_7celiagg_8_celiagg_InnerBevel = GraphicsState::InnerBevel,
  __pyx_e_7celiagg_8_celiagg_InnerMiter = GraphicsState::InnerMiter,
  __pyx_e_7celiagg_8_celiagg_InnerJag = GraphicsState::InnerJag,
  __pyx_e_7celiagg_8_celiagg_InnerRound = GraphicsState::InnerRound
};

/* "celiagg/enums.pxi":54
 *     InnerRound = _enums.InnerRound
 * 
 * cpdef enum LineJoin:             # <<<<<<<<<<<<<<
 *     JoinMiter = _enums.JoinMiter
 *     JoinRound = _enums.JoinRound
 */
enum __pyx_t_7celiagg_8_celiagg_LineJoin {

  /* "celiagg/enums.pxi":57
 *     JoinMiter = _enums.JoinMiter
 *     JoinRound = _enums.JoinRound
 *     JoinBevel = _enums.JoinBevel             # <<<<<<<<<<<<<<
 * 
 * cpdef enum LineCap:
 */
  __pyx_e_7celiagg_8_celiagg_JoinMiter = GraphicsState::JoinMiter,
  __pyx_e_7celiagg_8_celiagg_JoinRound = GraphicsState::JoinRound,
  __pyx_e_7celiagg_8_celiagg_JoinBevel = GraphicsState::JoinBevel
};

/* "celiagg/enums.pxi":59
 *     JoinBevel = _enums.JoinBevel
 * 
 * cpdef enum LineCap:             # <<<<<<<<<<<<<<
 *     CapButt = _enums.CapButt
 *     CapSquare = _enums.CapSquare
 */
enum __pyx_t_7celiagg_8_celiagg_LineCap {

  /* "celiagg/enums.pxi":62
 *     CapButt = _enums.CapButt
 *     CapSquare = _enums.CapSquare
 *     CapRound = _enums.CapRound             # <<<<<<<<<<<<<<
 * 
 * cpdef enum DrawingMode:
 */
  __pyx_e_7celiagg_8_celiagg_CapButt = GraphicsState::CapButt,
  __pyx_e_7celiagg_8_celiagg_CapSquare = GraphicsState::CapSquare,
  __pyx_e_7celiagg_8_celiagg_CapRound = GraphicsState::CapRound
};

/* "celiagg/enums.pxi":64
 *     CapRound = _enums.CapRound
 * 
 * cpdef enum DrawingMode:             # <<<<<<<<<<<<<<
 *     DrawFill = _enums.DrawFill
 *     DrawEofFill = _enums.DrawEofFill
 */
enum __pyx_t_7celiagg_8_celiagg_DrawingMode {

  /* "celiagg/enums.pxi":69
 *     DrawStroke = _enums.DrawStroke
 *     DrawFillStroke = _enums.DrawFillStroke
 *     DrawEofFillStroke = _enums.DrawEofFillStroke             # <<<<<<<<<<<<<<
 * 
 * cpdef enum TextDrawingMode:
 */
  __pyx_e_7celiagg_8_celiagg_DrawFill = GraphicsState::DrawFill,
  __pyx_e_7celiagg_8_celiagg_DrawEofFill = GraphicsState::DrawEofFill,
  __pyx_e_7celiagg_8_celiagg_DrawStroke = GraphicsState::DrawStroke,
  __pyx_e_7celiagg_8_celiagg_DrawFillStroke = GraphicsState::DrawFillStroke,
  __pyx_e_7celiagg_8_celiagg_DrawEofFillStroke = GraphicsState::DrawEofFillStroke
};

/* "celiagg/enums.pxi":71
 *     DrawEofFillStroke = _enums.DrawEofFillStroke
 * 
 * cpdef enum TextDrawingMode:             # <<<<<<<<<<<<<<
 *     TextDrawInvisible = _enums.TextDrawInvisible
 *     TextDrawFill = _enums.TextDrawFill
 */
enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode {

  /* "celiagg/enums.pxi":79
 *     TextDrawFillClip = _enums.TextDrawFillClip
 *     TextDrawStrokeClip = _enums.TextDrawStrokeClip
 *     TextDrawFillStrokeClip = _enums.TextDrawFillStrokeClip             # <<<<<<<<<<<<<<
 * 
 * cpdef enum BlendMode:
 */
  __pyx_e_7celiagg_8_celiagg_TextDrawInvisible = GraphicsState::TextDrawInvisible,
  __pyx_e_7celiagg_8_celiagg_TextDrawFill = GraphicsState::TextDrawFill,
  __pyx_e_7celiagg_8_celiagg_TextDrawStroke = GraphicsState::TextDrawStroke,
  __pyx_e_7celiagg_8_celiagg_TextDrawClip = GraphicsState::TextDrawClip,
  __pyx_e_7celiagg_8_celiagg_TextDrawFillStroke = GraphicsState::TextDrawFillStroke,
  __pyx_e_7celiagg_8_celiagg_TextDrawFillClip = GraphicsState::TextDrawFillClip,
  __pyx_e_7celiagg_8_celiagg_TextDrawStrokeClip = GraphicsState::TextDrawStrokeClip,
  __pyx_e_7celiagg_8_celiagg_TextDrawFillStrokeClip = GraphicsState::TextDrawFillStrokeClip
};

/* "celiagg/enums.pxi":81
 *     TextDrawFillStrokeClip = _enums.TextDrawFillStrokeClip
 * 
 * cpdef enum BlendMode:             # <<<<<<<<<<<<<<
 *     BlendAlpha = _enums.BlendAlpha
 *     BlendClear = _enums.BlendClear
 */
enum __pyx_t_7celiagg_8_celiagg_BlendMode {

  /* "celiagg/enums.pxi":106
 *     BlendSoftLight = _enums.BlendSoftLight
 *     BlendDifference = _enums.BlendDifference
 *     BlendExclusion = _enums.BlendExclusion             # <<<<<<<<<<<<<<
 */
  __pyx_e_7celiagg_8_celiagg_BlendAlpha = GraphicsState::BlendAlpha,
  __pyx_e_7celiagg_8_celiagg_BlendClear = GraphicsState::BlendClear,
  __pyx_e_7celiagg_8_celiagg_BlendSrc = GraphicsState::BlendSrc,
  __pyx_e_7celiagg_8_celiagg_BlendDst = GraphicsState::BlendDst,
  __pyx_e_7celiagg_8_celiagg_BlendSrcOver = GraphicsState::BlendSrcOver,
  __pyx_e_7celiagg_8_celiagg_BlendDstOver = GraphicsState::BlendDstOver,
  __pyx_e_7celiagg_8_celiagg_BlendSrcIn = GraphicsState::BlendSrcIn,
  __pyx_e_7celiagg_8_celiagg_BlendDstIn = GraphicsState::BlendDstIn,
  __pyx_e_7celiagg_8_celiagg_BlendSrcOut = GraphicsState::BlendSrcOut,
  __pyx_e_7celiagg_8_celiagg_BlendDstOut = GraphicsState::BlendDstOut,
  __pyx_e_7celiagg_8_celiagg_BlendSrcAtop = GraphicsState::BlendSrcAtop,
  __pyx_e_7celiagg_8_celiagg_BlendDstAtop = GraphicsState::BlendDstAtop,
  __pyx_e_7celiagg_8_celiagg_BlendXor = GraphicsState::BlendXor,
  __pyx_e_7celiagg_8_celiagg_BlendAdd = GraphicsState::BlendAdd,
  __pyx_e_7celiagg_8_celiagg_BlendMultiply = GraphicsState::BlendMultiply,
  __pyx_e_7celiagg_8_celiagg_BlendScreen = GraphicsState::BlendScreen,
  __pyx_e_7celiagg_8_celiagg_BlendOverlay = GraphicsState::BlendOverlay,
  __pyx_e_7celiagg_8_celiagg_BlendDarken = GraphicsState::BlendDarken,
  __pyx_e_7celiagg_8_celiagg_BlendLighten = GraphicsState::BlendLighten,
  __pyx_e_7celiagg_8_celiagg_BlendColorDodge = GraphicsState::BlendColorDodge,
  __pyx_e_7celiagg_8_celiagg_BlendColorBurn = GraphicsState::BlendColorBurn,
  __pyx_e_7celiagg_8_celiagg_BlendHardLight = GraphicsState::BlendHardLight,
  __pyx_e_7celiagg_8_celiagg_BlendSoftLight = GraphicsState::BlendSoftLight,
  __pyx_e_7celiagg_8_celiagg_BlendDifference = GraphicsState::BlendDifference,
  __pyx_e_7celiagg_8_celiagg_BlendExclusion = GraphicsState::BlendExclusion
};

/* "celiagg/font.pxi":26
 * 
 * 
 * cpdef enum FontCacheType:             # <<<<<<<<<<<<<<
 *     RasterFontCache = _enums.RasterFontCache
 *     VectorFontCache = _enums.VectorFontCache
 */
enum __pyx_t_7celiagg_8_celiagg_FontCacheType {

  /* "celiagg/font.pxi":28
 * cpdef enum FontCacheType:
 *     RasterFontCache = _enums.RasterFontCache
 *     VectorFontCache = _enums.VectorFontCache             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_e_7celiagg_8_celiagg_RasterFontCache = Font::RasterFontCache,
  __pyx_e_7celiagg_8_celiagg_VectorFontCache = Font::VectorFontCache
};

/* "celiagg/paint.pxi":25
 * # Authors: John Wiggins
 * 
 * cpdef enum GradientSpread:             # <<<<<<<<<<<<<<
 *     SpreadPad = _enums.k_GradientSpreadPad
 *     SpreadReflect = _enums.k_GradientSpreadReflect
 */
enum __pyx_t_7celiagg_8_celiagg_GradientSpread {

  /* "celiagg/paint.pxi":28
 *     SpreadPad = _enums.k_GradientSpreadPad
 *     SpreadReflect = _enums.k_GradientSpreadReflect
 *     SpreadRepeat = _enums.k_GradientSpreadRepeat             # <<<<<<<<<<<<<<
 * 
 * cpdef enum GradientUnits:
 */
  __pyx_e_7celiagg_8_celiagg_SpreadPad = Paint::k_GradientSpreadPad,
  __pyx_e_7celiagg_8_celiagg_SpreadReflect = Paint::k_GradientSpreadReflect,
  __pyx_e_7celiagg_8_celiagg_SpreadRepeat = Paint::k_GradientSpreadRepeat
};

/* "celiagg/paint.pxi":30
 *     SpreadRepeat = _enums.k_GradientSpreadRepeat
 * 
 * cpdef enum GradientUnits:             # <<<<<<<<<<<<<<
 *     UserSpace = _enums.k_GradientUnitsUserSpace
 *     ObjectBoundingBox = _enums.k_GradientUnitsObjectBoundingBox
 */
enum __pyx_t_7celiagg_8_celiagg_GradientUnits {

  /* "celiagg/paint.pxi":32
 * cpdef enum GradientUnits:
 *     UserSpace = _enums.k_GradientUnitsUserSpace
 *     ObjectBoundingBox = _enums.k_GradientUnitsObjectBoundingBox             # <<<<<<<<<<<<<<
 * 
 * cpdef enum PatternStyle:
 */
  __pyx_e_7celiagg_8_celiagg_UserSpace = Paint::k_GradientUnitsUserSpace,
  __pyx_e_7celiagg_8_celiagg_ObjectBoundingBox = Paint::k_GradientUnitsObjectBoundingBox
};

/* "celiagg/paint.pxi":34
 *     ObjectBoundingBox = _enums.k_GradientUnitsObjectBoundingBox
 * 
 * cpdef enum PatternStyle:             # <<<<<<<<<<<<<<
 *     StyleRepeat = _enums.k_PatternStyleRepeat
 *     StyleReflect = _enums.k_PatternStyleReflect
 */
enum __pyx_t_7celiagg_8_celiagg_PatternStyle {

  /* "celiagg/paint.pxi":36
 * cpdef enum PatternStyle:
 *     StyleRepeat = _enums.k_PatternStyleRepeat
 *     StyleReflect = _enums.k_PatternStyleReflect             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_e_7celiagg_8_celiagg_StyleRepeat = Paint::k_PatternStyleRepeat,
  __pyx_e_7celiagg_8_celiagg_StyleReflect = Paint::k_PatternStyleReflect
};

/* "celiagg/image.pxi":27
 * cimport numpy as np
 * 
 * ctypedef _image.Image* img_ptr_t             # <<<<<<<<<<<<<<
 * 
 * cdef _get_format_dtype(PixelFormat pixel_format):
 */
typedef Image *__pyx_t_7celiagg_8_celiagg_img_ptr_t;

/* "celiagg/ndarray_canvas.pxi":29
 * ctypedef unsigned char _bytes_t
 * 
 * ctypedef _ndarray_canvas.ndarray_canvas_base canvas_base_t             # <<<<<<<<<<<<<<
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgba128] canvas_rgba128_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_bgra32] canvas_brga32_t
 */
typedef ndarray_canvas_base __pyx_t_7celiagg_8_celiagg_canvas_base_t;

/* "celiagg/ndarray_canvas.pxi":30
 * 
 * ctypedef _ndarray_canvas.ndarray_canvas_base canvas_base_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgba128] canvas_rgba128_t             # <<<<<<<<<<<<<<
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_bgra32] canvas_brga32_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgba32] canvas_rgba32_t
 */
typedef ndarray_canvas<agg::pixfmt_rgba128>  __pyx_t_7celiagg_8_celiagg_canvas_rgba128_t;

/* "celiagg/ndarray_canvas.pxi":31
 * ctypedef _ndarray_canvas.ndarray_canvas_base canvas_base_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgba128] canvas_rgba128_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_bgra32] canvas_brga32_t             # <<<<<<<<<<<<<<
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgba32] canvas_rgba32_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgb24] canvas_rgb24_t
 */
typedef ndarray_canvas<agg::pixfmt_bgra32>  __pyx_t_7celiagg_8_celiagg_canvas_brga32_t;

/* "celiagg/ndarray_canvas.pxi":32
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgba128] canvas_rgba128_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_bgra32] canvas_brga32_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgba32] canvas_rgba32_t             # <<<<<<<<<<<<<<
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgb24] canvas_rgb24_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_gray8] canvas_ga16_t
 */
typedef ndarray_canvas<agg::pixfmt_rgba32>  __pyx_t_7celiagg_8_celiagg_canvas_rgba32_t;

/* "celiagg/ndarray_canvas.pxi":33
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_bgra32] canvas_brga32_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgba32] canvas_rgba32_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgb24] canvas_rgb24_t             # <<<<<<<<<<<<<<
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_gray8] canvas_ga16_t
 * 
 */
typedef ndarray_canvas<agg::pixfmt_rgb24>  __pyx_t_7celiagg_8_celiagg_canvas_rgb24_t;

/* "celiagg/ndarray_canvas.pxi":34
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgba32] canvas_rgba32_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_rgb24] canvas_rgb24_t
 * ctypedef _ndarray_canvas.ndarray_canvas[_ndarray_canvas.pixfmt_gray8] canvas_ga16_t             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef ndarray_canvas<agg::pixfmt_gray8>  __pyx_t_7celiagg_8_celiagg_canvas_ga16_t;

/* "celiagg/conversion.pxi":59
 * 
 * 
 * cdef gray2rgb(src, dst_order, dtype=None, alpha=False):             # <<<<<<<<<<<<<<
 *     """Convert gray pixels to RGB[A].
 *     """
 */
struct __pyx_opt_args_7celiagg_8_celiagg_gray2rgb {
  int __pyx_n;
  PyObject *dtype;
  PyObject *alpha;
};

/* "celiagg/conversion.pxi":78
 * 
 * 
 * cdef rgb2rgb(src, src_order, dst_order, dtype=None, alpha=False):             # <<<<<<<<<<<<<<
 *     """Convert RGB[A] pixels to RGB[A] pixels.
 *     """
 */
struct __pyx_opt_args_7celiagg_8_celiagg_rgb2rgb {
  int __pyx_n;
  PyObject *dtype;
  PyObject *alpha;
};

/* "celiagg/conversion.pxi":100
 * 
 * 
 * cdef gray2gray(src, dtype=None):             # <<<<<<<<<<<<<<
 *     """Convert gray pixels to gray pixels.
 *     """
 */
struct __pyx_opt_args_7celiagg_8_celiagg_gray2gray {
  int __pyx_n;
  PyObject *dtype;
};

/* "celiagg/conversion.pxi":110
 * 
 * 
 * cdef rgb2gray(src, src_order, dtype=None):             # <<<<<<<<<<<<<<
 *     """Convert RGB[A] pixels to gray.
 *     """
 */
struct __pyx_opt_args_7celiagg_8_celiagg_rgb2gray {
  int __pyx_n;
  PyObject *dtype;
};

/* "celiagg/font.pxi":32
 * 
 * IF _ENABLE_TEXT_RENDERING:
 *     cdef class Font:             # <<<<<<<<<<<<<<
 *         """Font(path, size, cache_type)
 * 
 */
struct __pyx_obj_7celiagg_8_celiagg_Font {
  PyObject_HEAD
  Font *_this;
};


/* "celiagg/graphics_state.pxi":29
 * 
 * 
 * cdef class Rect:             # <<<<<<<<<<<<<<
 *     """Rect(x, y, w, h)
 *     A rectangle.
 */
struct __pyx_obj_7celiagg_8_celiagg_Rect {
  PyObject_HEAD
  GraphicsState::Rect *_this;
};


/* "celiagg/graphics_state.pxi":92
 * 
 * 
 * cdef class GraphicsState:             # <<<<<<<<<<<<<<
 *     """GraphicsState(**properties)
 *     A container for drawing state. Can be initialized using keyword
 */
struct __pyx_obj_7celiagg_8_celiagg_GraphicsState {
  PyObject_HEAD
  struct __pyx_vtabstruct_7celiagg_8_celiagg_GraphicsState *__pyx_vtab;
  GraphicsState *_this;
  struct __pyx_obj_7celiagg_8_celiagg_Image *_stencil_img;
};


/* "celiagg/image.pxi":112
 * 
 * 
 * cdef class Image:             # <<<<<<<<<<<<<<
 *     """Image(array, pixel_format, bottom_up=False)
 * 
 */
struct __pyx_obj_7celiagg_8_celiagg_Image {
  PyObject_HEAD
  __pyx_t_7celiagg_8_celiagg_img_ptr_t _this;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat pixel_format;
  PyObject *pixel_array;
  bool bottom_up;
};


/* "celiagg/ndarray_canvas.pxi":38
 * 
 * @cython.internal
 * cdef class CanvasBase:             # <<<<<<<<<<<<<<
 *     cdef canvas_base_t* _this
 *     cdef PixelFormat pixel_format
 */
struct __pyx_obj_7celiagg_8_celiagg_CanvasBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *__pyx_vtab;
  __pyx_t_7celiagg_8_celiagg_canvas_base_t *_this;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat pixel_format;
  PyObject *py_array;
  bool bottom_up;
};


/* "celiagg/ndarray_canvas.pxi":289
 * 
 * 
 * cdef class CanvasRGBA128(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasRGBA128(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */
struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase __pyx_base;
};


/* "celiagg/ndarray_canvas.pxi":310
 * 
 * 
 * cdef class CanvasBGRA32(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasBGRA32(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */
struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase __pyx_base;
};


/* "celiagg/ndarray_canvas.pxi":331
 * 
 * 
 * cdef class CanvasRGBA32(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasRGBA32(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */
struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase __pyx_base;
};


/* "celiagg/ndarray_canvas.pxi":352
 * 
 * 
 * cdef class CanvasRGB24(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasRGB24(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */
struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase __pyx_base;
};


/* "celiagg/ndarray_canvas.pxi":373
 * 
 * 
 * cdef class CanvasGA16(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasGA16(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */
struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase __pyx_base;
};


/* "celiagg/ndarray_canvas.pxi":394
 * 
 * 
 * cdef class CanvasG8(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasG8(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */
struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase __pyx_base;
};


/* "celiagg/paint.pxi":63
 * 
 * @cython.internal
 * cdef class Paint:             # <<<<<<<<<<<<<<
 *     cdef _paint.Paint* _this
 * 
 */
struct __pyx_obj_7celiagg_8_celiagg_Paint {
  PyObject_HEAD
  Paint *_this;
};


/* "celiagg/paint.pxi":81
 * 
 * 
 * cdef class LinearGradientPaint(Paint):             # <<<<<<<<<<<<<<
 *     """LinearGradientPaint(x1, y1, x2, y2, stops, spread, units)
 *     A ``Paint`` for drawing linear gradients
 */
struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint {
  struct __pyx_obj_7celiagg_8_celiagg_Paint __pyx_base;
  PyObject *_points;
  PyObject *_stops;
};


/* "celiagg/paint.pxi":120
 * 
 * 
 * cdef class RadialGradientPaint(Paint):             # <<<<<<<<<<<<<<
 *     """RadialGradientPaint(cx, cy, r, fx, fy, stops, spread, units)
 *     A ``Paint`` for drawing radial gradients
 */
struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint {
  struct __pyx_obj_7celiagg_8_celiagg_Paint __pyx_base;
  PyObject *_points;
  PyObject *_stops;
};


/* "celiagg/paint.pxi":160
 * 
 * 
 * cdef class PatternPaint(Paint):             # <<<<<<<<<<<<<<
 *     """PatternPaint(style, image)
 *     A ``Paint`` for repeating patterns.
 */
struct __pyx_obj_7celiagg_8_celiagg_PatternPaint {
  struct __pyx_obj_7celiagg_8_celiagg_Paint __pyx_base;
  enum __pyx_t_7celiagg_8_celiagg_PatternStyle style;
  struct __pyx_obj_7celiagg_8_celiagg_Image *img_obj;
};


/* "celiagg/paint.pxi":198
 * 
 * 
 * cdef class SolidPaint(Paint):             # <<<<<<<<<<<<<<
 *     """SolidPaint(r, g, b, a)
 *     A ``Paint`` for solid colors.
 */
struct __pyx_obj_7celiagg_8_celiagg_SolidPaint {
  struct __pyx_obj_7celiagg_8_celiagg_Paint __pyx_base;
};


/* "celiagg/transform.pxi":26
 * 
 * 
 * cdef class Transform:             # <<<<<<<<<<<<<<
 *     """Transform(sx=1.0, shy=0.0, shx=0.0, sy=1.0, tx=0.0, ty=0.0)
 *     A 2D affine transform.
 */
struct __pyx_obj_7celiagg_8_celiagg_Transform {
  PyObject_HEAD
  struct __pyx_vtabstruct_7celiagg_8_celiagg_Transform *__pyx_vtab;
  agg::trans_affine *_this;
};


/* "celiagg/vertex_source.pxi":25
 * # Authors: John Wiggins
 * 
 * cdef class VertexSource:             # <<<<<<<<<<<<<<
 *     """An object which supplies vertex pairs and other information to
 *     low-level drawing routines.
 */
struct __pyx_obj_7celiagg_8_celiagg_VertexSource {
  PyObject_HEAD
  VertexSource *_this;
};


/* "celiagg/vertex_source.pxi":83
 * 
 * 
 * cdef class BSpline(VertexSource):             # <<<<<<<<<<<<<<
 *     """BSpline(points)
 *     A B-Spline curve.
 */
struct __pyx_obj_7celiagg_8_celiagg_BSpline {
  struct __pyx_obj_7celiagg_8_celiagg_VertexSource __pyx_base;
  PyObject *_points;
};


/* "celiagg/vertex_source.pxi":117
 * 
 * 
 * cdef class Path(VertexSource):             # <<<<<<<<<<<<<<
 *     """A path object.
 *     """
 */
struct __pyx_obj_7celiagg_8_celiagg_Path {
  struct __pyx_obj_7celiagg_8_celiagg_VertexSource __pyx_base;
};


/* "celiagg/vertex_source.pxi":363
 * 
 * 
 * cdef class ShapeAtPoints(VertexSource):             # <<<<<<<<<<<<<<
 *     """ShapeAtPoints(source, points)
 *     Replicates a shape at mutiple points.
 */
struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints {
  struct __pyx_obj_7celiagg_8_celiagg_VertexSource __pyx_base;
  struct __pyx_obj_7celiagg_8_celiagg_VertexSource *_sub_source;
  PyObject *_points;
};


/* "celiagg/vertex_source.pxi":36
 *         del self._this
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         for pt in self.vertices():
 *             yield pt
 */
struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ {
  PyObject_HEAD
  PyObject *__pyx_v_pt;
  struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "EnumBase":15
 * 
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):             # <<<<<<<<<<<<<<
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 */
struct __pyx_obj___Pyx_EnumMeta {
  PyHeapTypeObject __pyx_base;
};


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":279
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "celiagg/graphics_state.pxi":92
 * 
 * 
 * cdef class GraphicsState:             # <<<<<<<<<<<<<<
 *     """GraphicsState(**properties)
 *     A container for drawing state. Can be initialized using keyword
 */

struct __pyx_vtabstruct_7celiagg_8_celiagg_GraphicsState {
  PyObject *(*_propnames)(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *);
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_GraphicsState *__pyx_vtabptr_7celiagg_8_celiagg_GraphicsState;


/* "celiagg/ndarray_canvas.pxi":38
 * 
 * @cython.internal
 * cdef class CanvasBase:             # <<<<<<<<<<<<<<
 *     cdef canvas_base_t* _this
 *     cdef PixelFormat pixel_format
 */

struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase {
  int (*base_init)(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *, PyObject *, int, bool);
  PyObject *(*_check_stencil)(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *, struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *);
  struct __pyx_obj_7celiagg_8_celiagg_Image *(*_get_native_image)(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *, struct __pyx_obj_7celiagg_8_celiagg_Image *, enum __pyx_t_7celiagg_8_celiagg_PixelFormat);
  struct __pyx_obj_7celiagg_8_celiagg_Paint *(*_get_native_paint)(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *, PyObject *, enum __pyx_t_7celiagg_8_celiagg_PixelFormat);
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *__pyx_vtabptr_7celiagg_8_celiagg_CanvasBase;


/* "celiagg/ndarray_canvas.pxi":289
 * 
 * 
 * cdef class CanvasRGBA128(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasRGBA128(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */

struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGBA128 {
  struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase __pyx_base;
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGBA128 *__pyx_vtabptr_7celiagg_8_celiagg_CanvasRGBA128;


/* "celiagg/ndarray_canvas.pxi":310
 * 
 * 
 * cdef class CanvasBGRA32(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasBGRA32(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */

struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBGRA32 {
  struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase __pyx_base;
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBGRA32 *__pyx_vtabptr_7celiagg_8_celiagg_CanvasBGRA32;


/* "celiagg/ndarray_canvas.pxi":331
 * 
 * 
 * cdef class CanvasRGBA32(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasRGBA32(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */

struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGBA32 {
  struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase __pyx_base;
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGBA32 *__pyx_vtabptr_7celiagg_8_celiagg_CanvasRGBA32;


/* "celiagg/ndarray_canvas.pxi":352
 * 
 * 
 * cdef class CanvasRGB24(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasRGB24(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */

struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGB24 {
  struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase __pyx_base;
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGB24 *__pyx_vtabptr_7celiagg_8_celiagg_CanvasRGB24;


/* "celiagg/ndarray_canvas.pxi":373
 * 
 * 
 * cdef class CanvasGA16(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasGA16(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */

struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasGA16 {
  struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase __pyx_base;
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasGA16 *__pyx_vtabptr_7celiagg_8_celiagg_CanvasGA16;


/* "celiagg/ndarray_canvas.pxi":394
 * 
 * 
 * cdef class CanvasG8(CanvasBase):             # <<<<<<<<<<<<<<
 *     """CanvasG8(array, bottom_up=False)
 *     Provides AGG (Anti-Grain Geometry) drawing routines that render to the
 */

struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasG8 {
  struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase __pyx_base;
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasG8 *__pyx_vtabptr_7celiagg_8_celiagg_CanvasG8;


/* "celiagg/transform.pxi":26
 * 
 * 
 * cdef class Transform:             # <<<<<<<<<<<<<<
 *     """Transform(sx=1.0, shy=0.0, shx=0.0, sy=1.0, tx=0.0, ty=0.0)
 *     A 2D affine transform.
 */

struct __pyx_vtabstruct_7celiagg_8_celiagg_Transform {
  PyObject *(*_assign_obj)(struct __pyx_obj_7celiagg_8_celiagg_Transform *, agg::trans_affine);
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_Transform *__pyx_vtabptr_7celiagg_8_celiagg_Transform;


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* py_dict_items.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* None.proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* None.proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_float(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn___pyx_t_7celiagg_8_celiagg__bytes_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_7celiagg_8_celiagg__bytes_t(PyObject *, int writable_flag);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__Font_3a__3a_FontCacheType(enum Font::FontCacheType value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_DrawingMode(enum GraphicsState::DrawingMode value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_TextDrawingMode(enum GraphicsState::TextDrawingMode value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_BlendMode(enum GraphicsState::BlendMode value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_LineCap(enum GraphicsState::LineCap value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_LineJoin(enum GraphicsState::LineJoin value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_InnerJoin(enum GraphicsState::InnerJoin value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(enum __pyx_t_7celiagg_8_celiagg_PixelFormat value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_float(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_float(const char *itemp, PyObject *obj);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_char(unsigned char value);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_7celiagg_8_celiagg__bytes_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_7celiagg_8_celiagg__bytes_t(const char *itemp, PyObject *obj);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__Paint_3a__3a_GradientSpread(enum Paint::GradientSpread value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__Paint_3a__3a_GradientUnits(enum Paint::GradientUnits value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(enum __pyx_t_7celiagg_8_celiagg_PatternStyle value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_FontCacheType __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_DrawingMode __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_BlendMode __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_BlendMode(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_LineCap __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_LineCap(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_LineJoin __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_LineJoin(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_InnerJoin __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_PixelFormat __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_GradientSpread __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_GradientUnits __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_PatternStyle __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned char __Pyx_PyInt_As_unsigned_char(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(enum __pyx_t_7celiagg_8_celiagg_BlendMode value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(enum __pyx_t_7celiagg_8_celiagg_DrawingMode value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(enum __pyx_t_7celiagg_8_celiagg_FontCacheType value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(enum __pyx_t_7celiagg_8_celiagg_GradientSpread value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(enum __pyx_t_7celiagg_8_celiagg_GradientUnits value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(enum __pyx_t_7celiagg_8_celiagg_InnerJoin value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineCap(enum __pyx_t_7celiagg_8_celiagg_LineCap value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineJoin(enum __pyx_t_7celiagg_8_celiagg_LineJoin value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode value);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_npy_uint8(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_npy_uint16(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_float(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn_npy_uint8(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn_npy_uint16(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_npy_int32(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_double(PyObject *, int writable_flag);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_7celiagg_8_celiagg_13GraphicsState__propnames(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto*/
static int __pyx_f_7celiagg_8_celiagg_10CanvasBase_base_init(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_image, int __pyx_v_channel_count, bool __pyx_v_has_alpha); /* proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg_10CanvasBase__check_stencil(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_state); /* proto*/
static struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_f_7celiagg_8_celiagg_10CanvasBase__get_native_image(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_image, enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_fmt); /* proto*/
static struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_f_7celiagg_8_celiagg_10CanvasBase__get_native_paint(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_paint, enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_fmt); /* proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg_9Transform__assign_obj(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, agg::trans_affine __pyx_v_trans); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'libcpp' */

/* Module declarations from 'cython.view' */

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'celiagg._enums' */

/* Module declarations from 'celiagg._font' */

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'celiagg._image' */

/* Module declarations from 'celiagg._graphics_state' */

/* Module declarations from 'celiagg._transform' */

/* Module declarations from 'celiagg._paint' */

/* Module declarations from 'celiagg._vertex_source' */

/* Module declarations from 'celiagg._ndarray_canvas' */

/* Module declarations from 'celiagg._celiagg' */
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_Font = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_Rect = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_GraphicsState = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_Image = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_CanvasBase = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_CanvasRGBA128 = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_CanvasBGRA32 = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_CanvasRGBA32 = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_CanvasRGB24 = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_CanvasGA16 = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_CanvasG8 = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_Paint = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_LinearGradientPaint = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_RadialGradientPaint = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_PatternPaint = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_SolidPaint = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_Transform = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_VertexSource = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_BSpline = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_Path = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg_ShapeAtPoints = 0;
static PyTypeObject *__pyx_ptype_7celiagg_8_celiagg___pyx_scope_struct____iter__ = 0;
static PyTypeObject *__pyx_ptype___Pyx_EnumMeta = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static PyObject *__Pyx_OrderedDict = 0;
static PyObject *__Pyx_EnumBase = 0;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static PyObject *__Pyx_globals = 0;
static PyObject *__pyx_f_7celiagg_8_celiagg__get_utf8_text(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg__get_format_dtype(enum __pyx_t_7celiagg_8_celiagg_PixelFormat); /*proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg__get_format_last_dim(enum __pyx_t_7celiagg_8_celiagg_PixelFormat); /*proto*/
static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_2d_u8_img(__Pyx_memviewslice, bool); /*proto*/
static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_2d_u16_img(__Pyx_memviewslice, bool); /*proto*/
static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_2d_f32_img(__Pyx_memviewslice, bool); /*proto*/
static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_3d_u8_img(__Pyx_memviewslice, bool); /*proto*/
static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_3d_u16_img(__Pyx_memviewslice, bool); /*proto*/
static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_3d_f32_img(__Pyx_memviewslice, bool); /*proto*/
static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_image(PyObject *, PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg__get_gradient_points(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg__get_gradient_stops(PyObject *); /*proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg__get_component_indices(PyObject *); /*proto*/
static bool __pyx_f_7celiagg_8_celiagg__has_alpha(PyObject *); /*proto*/
static bool __pyx_f_7celiagg_8_celiagg__is_gray(PyObject *); /*proto*/
static bool __pyx_f_7celiagg_8_celiagg__is_rgb(PyObject *); /*proto*/
static float __pyx_f_7celiagg_8_celiagg__get_max_value(PyObject *); /*proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg_gray2rgb(PyObject *, PyObject *, struct __pyx_opt_args_7celiagg_8_celiagg_gray2rgb *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg_rgb2rgb(PyObject *, PyObject *, PyObject *, struct __pyx_opt_args_7celiagg_8_celiagg_rgb2rgb *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg_gray2gray(PyObject *, struct __pyx_opt_args_7celiagg_8_celiagg_gray2gray *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_7celiagg_8_celiagg_rgb2gray(PyObject *, PyObject *, struct __pyx_opt_args_7celiagg_8_celiagg_rgb2gray *__pyx_optional_args); /*proto*/
static PyObject *__pyx_convert_vector_to_py_double(const std::vector<double>  &); /*proto*/
static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *, PyObject *); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_float = { "float", NULL, sizeof(float), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_7celiagg_8_celiagg__bytes_t = { "_bytes_t", NULL, sizeof(__pyx_t_7celiagg_8_celiagg__bytes_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_7celiagg_8_celiagg__bytes_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_7celiagg_8_celiagg__bytes_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn_npy_uint8 = { "npy_uint8", NULL, sizeof(npy_uint8), { 0 }, 0, IS_UNSIGNED(npy_uint8) ? 'U' : 'I', IS_UNSIGNED(npy_uint8), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn_npy_uint16 = { "npy_uint16", NULL, sizeof(npy_uint16), { 0 }, 0, IS_UNSIGNED(npy_uint16) ? 'U' : 'I', IS_UNSIGNED(npy_uint16), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn_npy_int32 = { "npy_int32", NULL, sizeof(npy_int32), { 0 }, 0, IS_UNSIGNED(npy_int32) ? 'U' : 'I', IS_UNSIGNED(npy_int32), 0 };
#define __Pyx_MODULE_NAME "celiagg._celiagg"
extern int __pyx_module_is_main_celiagg___celiagg;
int __pyx_module_is_main_celiagg___celiagg = 0;

/* Implementation of 'celiagg._celiagg' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static PyObject *__pyx_builtin_IndexError;
static const char __pyx_k_O[] = "O";
static const char __pyx_k_a[] = "a";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_g[] = "g";
static const char __pyx_k_h[] = "h";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k__3[] = "{}({}, {}, {}, {})";
static const char __pyx_k_ch[] = "ch";
static const char __pyx_k_cx[] = "cx";
static const char __pyx_k_cy[] = "cy";
static const char __pyx_k_fx[] = "fx";
static const char __pyx_k_fy[] = "fy";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_rx[] = "rx";
static const char __pyx_k_ry[] = "ry";
static const char __pyx_k_sx[] = "sx";
static const char __pyx_k_sy[] = "sy";
static const char __pyx_k_tx[] = "tx";
static const char __pyx_k_ty[] = "ty";
static const char __pyx_k_x1[] = "x1";
static const char __pyx_k_x2[] = "x2";
static const char __pyx_k_y1[] = "y1";
static const char __pyx_k_y2[] = "y2";
static const char __pyx_k_MxN[] = "MxN";
static const char __pyx_k__52[] = "{}({}, {}, {}, {}, {}, {})";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_dct[] = "dct";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_dst[] = "dst";
static const char __pyx_k_fmt[] = "fmt";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_res[] = "res";
static const char __pyx_k_s_s[] = "%s.%s";
static const char __pyx_k_shx[] = "shx";
static const char __pyx_k_shy[] = "shy";
static const char __pyx_k_src[] = "src";
static const char __pyx_k_str[] = "__str__";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_Font[] = "Font";
static const char __pyx_k_MxNx[] = "MxNx{}";
static const char __pyx_k_Path[] = "Path";
static const char __pyx_k_Rect[] = "Rect";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_ends[] = "ends";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_fill[] = "fill";
static const char __pyx_k_font[] = "font";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_repr[] = "__repr__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_text[] = "text";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_utf8[] = "utf8";
static const char __pyx_k_x_to[] = "x_to";
static const char __pyx_k_y_to[] = "y_to";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_BGR24[] = "BGR24";
static const char __pyx_k_BGR48[] = "BGR48";
static const char __pyx_k_BGR96[] = "BGR96";
static const char __pyx_k_Gray8[] = "Gray8";
static const char __pyx_k_Image[] = "Image";
static const char __pyx_k_RGB24[] = "RGB24";
static const char __pyx_k_RGB48[] = "RGB48";
static const char __pyx_k_RGB96[] = "RGB96";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_iinfo[] = "iinfo";
static const char __pyx_k_image[] = "image";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_order[] = "order";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_s_s_d[] = "<%s.%s: %d>";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_stops[] = "stops";
static const char __pyx_k_style[] = "style";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_types[] = "types";
static const char __pyx_k_uint8[] = "uint8";
static const char __pyx_k_units[] = "units";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_width[] = "width";
static const char __pyx_k_ABGR32[] = "ABGR32";
static const char __pyx_k_ABGR64[] = "ABGR64";
static const char __pyx_k_ARGB32[] = "ARGB32";
static const char __pyx_k_ARGB64[] = "ARGB64";
static const char __pyx_k_BGRA32[] = "BGRA32";
static const char __pyx_k_BGRA64[] = "BGRA64";
static const char __pyx_k_Gray16[] = "Gray16";
static const char __pyx_k_Gray32[] = "Gray32";
static const char __pyx_k_RGBA32[] = "RGBA32";
static const char __pyx_k_RGBA64[] = "RGBA64";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_height[] = "height";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_pixels[] = "pixels";
static const char __pyx_k_points[] = "points";
static const char __pyx_k_radius[] = "radius";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_spread[] = "spread";
static const char __pyx_k_starts[] = "starts";
static const char __pyx_k_stroke[] = "stroke";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_uint16[] = "uint16";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_x_ctrl[] = "x_ctrl";
static const char __pyx_k_y_ctrl[] = "y_ctrl";
static const char __pyx_k_ABGR128[] = "ABGR128";
static const char __pyx_k_ARGB128[] = "ARGB128";
static const char __pyx_k_BGRA128[] = "BGRA128";
static const char __pyx_k_BSpline[] = "BSpline";
static const char __pyx_k_CapButt[] = "CapButt";
static const char __pyx_k_IntEnum[] = "IntEnum";
static const char __pyx_k_LineCap[] = "LineCap";
static const char __pyx_k_RGBA128[] = "RGBA128";
static const char __pyx_k_asarray[] = "asarray";
static const char __pyx_k_dst_rgb[] = "dst_rgb";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_members[] = "__members__";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_parents[] = "parents";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_src_rgb[] = "src_rgb";
static const char __pyx_k_stencil[] = "stencil";
static const char __pyx_k_x_ctrl1[] = "x_ctrl1";
static const char __pyx_k_x_ctrl2[] = "x_ctrl2";
static const char __pyx_k_y_ctrl1[] = "y_ctrl1";
static const char __pyx_k_y_ctrl2[] = "y_ctrl2";
static const char __pyx_k_AggError[] = "AggError";
static const char __pyx_k_BlendAdd[] = "BlendAdd";
static const char __pyx_k_BlendDst[] = "BlendDst";
static const char __pyx_k_BlendSrc[] = "BlendSrc";
static const char __pyx_k_BlendXor[] = "BlendXor";
static const char __pyx_k_CanvasG8[] = "CanvasG8";
static const char __pyx_k_CapRound[] = "CapRound";
static const char __pyx_k_DrawFill[] = "DrawFill";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_EnumBase[] = "EnumBase";
static const char __pyx_k_EnumType[] = "EnumType";
static const char __pyx_k_HAS_TEXT[] = "HAS_TEXT";
static const char __pyx_k_InnerJag[] = "InnerJag";
static const char __pyx_k_LineJoin[] = "LineJoin";
static const char __pyx_k_dst_gray[] = "dst_gray";
static const char __pyx_k_fontName[] = "fontName";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_src_gray[] = "src_gray";
static const char __pyx_k_vertices[] = "vertices";
static const char __pyx_k_BlendMode[] = "BlendMode";
static const char __pyx_k_CapSquare[] = "CapSquare";
static const char __pyx_k_InnerJoin[] = "InnerJoin";
static const char __pyx_k_JoinBevel[] = "JoinBevel";
static const char __pyx_k_JoinMiter[] = "JoinMiter";
static const char __pyx_k_JoinRound[] = "JoinRound";
static const char __pyx_k_SpreadPad[] = "SpreadPad";
static const char __pyx_k_Transform[] = "Transform";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_UserSpace[] = "UserSpace";
static const char __pyx_k_bottom_up[] = "bottom_up";
static const char __pyx_k_clockwise[] = "clockwise";
static const char __pyx_k_dst_alpha[] = "dst_alpha";
static const char __pyx_k_dst_dtype[] = "dst_dtype";
static const char __pyx_k_dst_order[] = "dst_order";
static const char __pyx_k_end_angle[] = "end_angle";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_src_order[] = "src_order";
static const char __pyx_k_to_format[] = "to_format";
static const char __pyx_k_transform[] = "transform";
static const char __pyx_k_BlendAlpha[] = "BlendAlpha";
static const char __pyx_k_BlendClear[] = "BlendClear";
static const char __pyx_k_BlendDstIn[] = "BlendDstIn";
static const char __pyx_k_BlendSrcIn[] = "BlendSrcIn";
static const char __pyx_k_CanvasGA16[] = "CanvasGA16";
static const char __pyx_k_DrawStroke[] = "DrawStroke";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_InnerBevel[] = "InnerBevel";
static const char __pyx_k_InnerMiter[] = "InnerMiter";
static const char __pyx_k_InnerRound[] = "InnerRound";
static const char __pyx_k_SolidPaint[] = "SolidPaint";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_face_index[] = "face_index";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_BlendDarken[] = "BlendDarken";
static const char __pyx_k_BlendDstOut[] = "BlendDstOut";
static const char __pyx_k_BlendScreen[] = "BlendScreen";
static const char __pyx_k_BlendSrcOut[] = "BlendSrcOut";
static const char __pyx_k_CanvasRGB24[] = "CanvasRGB24";
static const char __pyx_k_DrawEofFill[] = "DrawEofFill";
static const char __pyx_k_DrawingMode[] = "DrawingMode";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_OrderedDict[] = "OrderedDict";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_PixelFormat[] = "PixelFormat";
static const char __pyx_k_StyleRepeat[] = "StyleRepeat";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_from_format[] = "from_format";
static const char __pyx_k_last_vertex[] = "last_vertex";
static const char __pyx_k_start_angle[] = "start_angle";
static const char __pyx_k_with_format[] = "_with_format";
static const char __pyx_k_BlendDstAtop[] = "BlendDstAtop";
static const char __pyx_k_BlendDstOver[] = "BlendDstOver";
static const char __pyx_k_BlendLighten[] = "BlendLighten";
static const char __pyx_k_BlendOverlay[] = "BlendOverlay";
static const char __pyx_k_BlendSrcAtop[] = "BlendSrcAtop";
static const char __pyx_k_BlendSrcOver[] = "BlendSrcOver";
static const char __pyx_k_CanvasBGRA32[] = "CanvasBGRA32";
static const char __pyx_k_CanvasRGBA32[] = "CanvasRGBA32";
static const char __pyx_k_PatternPaint[] = "PatternPaint";
static const char __pyx_k_PatternStyle[] = "PatternStyle";
static const char __pyx_k_Pyx_EnumBase[] = "__Pyx_EnumBase";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_SpreadRepeat[] = "SpreadRepeat";
static const char __pyx_k_StyleReflect[] = "StyleReflect";
static const char __pyx_k_TextDrawClip[] = "TextDrawClip";
static const char __pyx_k_TextDrawFill[] = "TextDrawFill";
static const char __pyx_k_VertexSource[] = "VertexSource";
static const char __pyx_k_pixel_format[] = "pixel_format";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_BlendMultiply[] = "BlendMultiply";
static const char __pyx_k_CanvasRGBA128[] = "CanvasRGBA128";
static const char __pyx_k_FontCacheType[] = "FontCacheType";
static const char __pyx_k_GradientUnits[] = "GradientUnits";
static const char __pyx_k_GraphicsState[] = "GraphicsState";
static const char __pyx_k_ShapeAtPoints[] = "ShapeAtPoints";
static const char __pyx_k_SpreadReflect[] = "SpreadReflect";
static const char __pyx_k_convert_image[] = "convert_image";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_BlendColorBurn[] = "BlendColorBurn";
static const char __pyx_k_BlendExclusion[] = "BlendExclusion";
static const char __pyx_k_BlendHardLight[] = "BlendHardLight";
static const char __pyx_k_BlendSoftLight[] = "BlendSoftLight";
static const char __pyx_k_DrawFillStroke[] = "DrawFillStroke";
static const char __pyx_k_GradientSpread[] = "GradientSpread";
static const char __pyx_k_TextDrawStroke[] = "TextDrawStroke";
static const char __pyx_k_BlendColorDodge[] = "BlendColorDodge";
static const char __pyx_k_BlendDifference[] = "BlendDifference";
static const char __pyx_k_RasterFontCache[] = "RasterFontCache";
static const char __pyx_k_TextDrawingMode[] = "TextDrawingMode";
static const char __pyx_k_VectorFontCache[] = "VectorFontCache";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_TextDrawFillClip[] = "TextDrawFillClip";
static const char __pyx_k_celiagg__celiagg[] = "celiagg._celiagg";
static const char __pyx_k_DrawEofFillStroke[] = "DrawEofFillStroke";
static const char __pyx_k_ObjectBoundingBox[] = "ObjectBoundingBox";
static const char __pyx_k_TextDrawInvisible[] = "TextDrawInvisible";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_Pyx_EnumBase___new[] = "__Pyx_EnumBase.__new__";
static const char __pyx_k_Pyx_EnumBase___str[] = "__Pyx_EnumBase.__str__";
static const char __pyx_k_TextDrawFillStroke[] = "TextDrawFillStroke";
static const char __pyx_k_TextDrawStrokeClip[] = "TextDrawStrokeClip";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_LinearGradientPaint[] = "LinearGradientPaint";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_Pyx_EnumBase___repr[] = "__Pyx_EnumBase.__repr__";
static const char __pyx_k_RadialGradientPaint[] = "RadialGradientPaint";
static const char __pyx_k_VertexSource___iter[] = "VertexSource.__iter__";
static const char __pyx_k_GetSetDescriptorType[] = "GetSetDescriptorType";
static const char __pyx_k_Unknown_enum_value_s[] = "Unknown enum value: '%s'";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_TextDrawFillStrokeClip[] = "TextDrawFillStrokeClip";
static const char __pyx_k_celiagg_conversion_pxi[] = "celiagg/conversion.pxi";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_src_must_an_Image_instance[] = "src must an Image instance";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_pyx_unpickle___Pyx_EnumMeta[] = "__pyx_unpickle___Pyx_EnumMeta";
static const char __pyx_k_font_must_be_a_Font_instance[] = "font must be a Font instance";
static const char __pyx_k_fill_must_be_a_Paint_instance[] = "fill must be a Paint instance";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_Stencil_Images_must_have_format[] = "Stencil Images must have format Gray8!";
static const char __pyx_k_The_clip_box_property_must_be_a[] = "The clip_box property must be a Rect";
static const char __pyx_k_The_format_requires_an_array_of[] = "The '{}' format requires an array of type {}, but {} was passed";
static const char __pyx_k_The_stencil_property_must_be_an[] = "The stencil property must be an Image";
static const char __pyx_k_image_argument_must_not_be_None[] = "image argument must not be None.";
static const char __pyx_k_image_must_be_an_Image_instance[] = "image must be an Image instance";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_stroke_must_be_a_Paint_instance[] = "stroke must be a Paint instance";
static const char __pyx_k_to_format_must_be_a_PixelFormat[] = "to_format must be a PixelFormat value";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_An_image_array_must_be_2_or_3_di[] = "An image array must be 2 or 3 dimensions";
static const char __pyx_k_An_image_must_be_provided_in_a_n[] = "An image must be provided in a numpy array";
static const char __pyx_k_Argument_must_be_a_unicode_strin[] = "Argument must be a unicode string";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Dashes_must_be_an_iterable_of_da[] = "Dashes must be an iterable of (dash len, gap len) pairs.";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Font_path_must_be_a_unicode_stri[] = "Font path must be a unicode string";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Incompatible_checksums_s_vs_0xb0[] = "Incompatible checksums (%s vs 0xb068931 = (name))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xd4[] = "Incompatible checksums (%s vs 0xd41d8cd = ())";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Points_argument_must_be_an_itera[] = "Points argument must be an iterable of (x, y) pairs.";
static const char __pyx_k_That_type_of_comparison_is_not_i[] = "That type of comparison is not implemented for Rect";
static const char __pyx_k_The_stencil_set_on_the_GraphicsS[] = "The stencil set on the GraphicsState object does not have the same dimensions as the target canvas! (({}, {}) vs. ({}, {}))";
static const char __pyx_k_The_text_argument_must_be_unicod[] = "The text argument must be unicode.";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_ends_argument_must_be_an_iterabl[] = "ends argument must be an iterable of (x, y) tuples.";
static const char __pyx_k_format_must_be_a_PixelFormat_val[] = "format must be a PixelFormat value";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_image_argument_must_be_channel_i[] = "image argument must be {} channel (ie, must be {}).";
static const char __pyx_k_image_must_be_an_ndarray_or_Imag[] = "image must be an ndarray or Image instance";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_points_argument_must_be_an_itera[] = "points argument must be an iterable of length {}";
static const char __pyx_k_rects_argument_must_be_an_iterab[] = "rects argument must be an iterable of (x, y, w, h) tuples.";
static const char __pyx_k_self__this_cannot_be_converted_t[] = "self._this cannot be converted to a Python object for pickling";
static const char __pyx_k_shape_must_be_a_VertexSource_Pat[] = "shape must be a VertexSource (Path, BSpline, etc)";
static const char __pyx_k_source_must_be_a_VertexSource_in[] = "source must be a VertexSource instance";
static const char __pyx_k_starts_argument_must_be_an_itera[] = "starts argument must be an iterable of (x, y) tuples.";
static const char __pyx_k_starts_argument_must_match_lengt[] = "starts argument must match length of ends argument.";
static const char __pyx_k_state_must_be_a_GraphicsState_in[] = "state must be a GraphicsState instance";
static const char __pyx_k_stops_argument_must_be_an_iterab[] = "stops argument must be an iterable of (off, r, g, b, a) tuples.";
static const char __pyx_k_transform_must_be_a_Transform_in[] = "transform must be a Transform instance";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static const char __pyx_k_The_format_requires_an_array_of_2[] = "The '{}' format requires an array of dimension MxNx{}";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static const char __pyx_k_That_type_of_comparison_is_not_i_2[] = "That type of comparison is not implemented for SolidPaint";
static const char __pyx_k_points_argument_must_be_an_itera_2[] = "points argument must be an iterable of (x, y) tuples.";
static PyObject *__pyx_n_s_ABGR128;
static PyObject *__pyx_n_s_ABGR32;
static PyObject *__pyx_n_s_ABGR64;
static PyObject *__pyx_n_s_ARGB128;
static PyObject *__pyx_n_s_ARGB32;
static PyObject *__pyx_n_s_ARGB64;
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_n_s_AggError;
static PyObject *__pyx_kp_s_An_image_array_must_be_2_or_3_di;
static PyObject *__pyx_kp_s_An_image_must_be_provided_in_a_n;
static PyObject *__pyx_kp_s_Argument_must_be_a_unicode_strin;
static PyObject *__pyx_n_s_BGR24;
static PyObject *__pyx_n_s_BGR48;
static PyObject *__pyx_n_s_BGR96;
static PyObject *__pyx_n_s_BGRA128;
static PyObject *__pyx_n_s_BGRA32;
static PyObject *__pyx_n_s_BGRA64;
static PyObject *__pyx_n_s_BSpline;
static PyObject *__pyx_n_s_BlendAdd;
static PyObject *__pyx_n_s_BlendAlpha;
static PyObject *__pyx_n_s_BlendClear;
static PyObject *__pyx_n_s_BlendColorBurn;
static PyObject *__pyx_n_s_BlendColorDodge;
static PyObject *__pyx_n_s_BlendDarken;
static PyObject *__pyx_n_s_BlendDifference;
static PyObject *__pyx_n_s_BlendDst;
static PyObject *__pyx_n_s_BlendDstAtop;
static PyObject *__pyx_n_s_BlendDstIn;
static PyObject *__pyx_n_s_BlendDstOut;
static PyObject *__pyx_n_s_BlendDstOver;
static PyObject *__pyx_n_s_BlendExclusion;
static PyObject *__pyx_n_s_BlendHardLight;
static PyObject *__pyx_n_s_BlendLighten;
static PyObject *__pyx_n_s_BlendMode;
static PyObject *__pyx_n_s_BlendMultiply;
static PyObject *__pyx_n_s_BlendOverlay;
static PyObject *__pyx_n_s_BlendScreen;
static PyObject *__pyx_n_s_BlendSoftLight;
static PyObject *__pyx_n_s_BlendSrc;
static PyObject *__pyx_n_s_BlendSrcAtop;
static PyObject *__pyx_n_s_BlendSrcIn;
static PyObject *__pyx_n_s_BlendSrcOut;
static PyObject *__pyx_n_s_BlendSrcOver;
static PyObject *__pyx_n_s_BlendXor;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_n_s_CanvasBGRA32;
static PyObject *__pyx_n_s_CanvasG8;
static PyObject *__pyx_n_s_CanvasGA16;
static PyObject *__pyx_n_s_CanvasRGB24;
static PyObject *__pyx_n_s_CanvasRGBA128;
static PyObject *__pyx_n_s_CanvasRGBA32;
static PyObject *__pyx_n_s_CapButt;
static PyObject *__pyx_n_s_CapRound;
static PyObject *__pyx_n_s_CapSquare;
static PyObject *__pyx_kp_s_Dashes_must_be_an_iterable_of_da;
static PyObject *__pyx_n_s_DrawEofFill;
static PyObject *__pyx_n_s_DrawEofFillStroke;
static PyObject *__pyx_n_s_DrawFill;
static PyObject *__pyx_n_s_DrawFillStroke;
static PyObject *__pyx_n_s_DrawStroke;
static PyObject *__pyx_n_s_DrawingMode;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_n_s_EnumBase;
static PyObject *__pyx_n_s_EnumType;
static PyObject *__pyx_n_s_Font;
static PyObject *__pyx_n_s_FontCacheType;
static PyObject *__pyx_kp_s_Font_path_must_be_a_unicode_stri;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_GetSetDescriptorType;
static PyObject *__pyx_n_s_GradientSpread;
static PyObject *__pyx_n_s_GradientUnits;
static PyObject *__pyx_n_s_GraphicsState;
static PyObject *__pyx_n_s_Gray16;
static PyObject *__pyx_n_s_Gray32;
static PyObject *__pyx_n_s_Gray8;
static PyObject *__pyx_n_s_HAS_TEXT;
static PyObject *__pyx_n_s_Image;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xb0;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd4;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_n_s_InnerBevel;
static PyObject *__pyx_n_s_InnerJag;
static PyObject *__pyx_n_s_InnerJoin;
static PyObject *__pyx_n_s_InnerMiter;
static PyObject *__pyx_n_s_InnerRound;
static PyObject *__pyx_n_s_IntEnum;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_n_s_JoinBevel;
static PyObject *__pyx_n_s_JoinMiter;
static PyObject *__pyx_n_s_JoinRound;
static PyObject *__pyx_n_s_LineCap;
static PyObject *__pyx_n_s_LineJoin;
static PyObject *__pyx_n_s_LinearGradientPaint;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_n_s_MxN;
static PyObject *__pyx_kp_s_MxNx;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_n_s_ObjectBoundingBox;
static PyObject *__pyx_n_s_OrderedDict;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_Path;
static PyObject *__pyx_n_s_PatternPaint;
static PyObject *__pyx_n_s_PatternStyle;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_PixelFormat;
static PyObject *__pyx_kp_s_Points_argument_must_be_an_itera;
static PyObject *__pyx_n_s_Pyx_EnumBase;
static PyObject *__pyx_n_s_Pyx_EnumBase___new;
static PyObject *__pyx_n_s_Pyx_EnumBase___repr;
static PyObject *__pyx_n_s_Pyx_EnumBase___str;
static PyObject *__pyx_n_s_RGB24;
static PyObject *__pyx_n_s_RGB48;
static PyObject *__pyx_n_s_RGB96;
static PyObject *__pyx_n_s_RGBA128;
static PyObject *__pyx_n_s_RGBA32;
static PyObject *__pyx_n_s_RGBA64;
static PyObject *__pyx_n_s_RadialGradientPaint;
static PyObject *__pyx_n_s_RasterFontCache;
static PyObject *__pyx_n_s_Rect;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_ShapeAtPoints;
static PyObject *__pyx_n_s_SolidPaint;
static PyObject *__pyx_n_s_SpreadPad;
static PyObject *__pyx_n_s_SpreadReflect;
static PyObject *__pyx_n_s_SpreadRepeat;
static PyObject *__pyx_kp_s_Stencil_Images_must_have_format;
static PyObject *__pyx_n_s_StyleReflect;
static PyObject *__pyx_n_s_StyleRepeat;
static PyObject *__pyx_n_s_TextDrawClip;
static PyObject *__pyx_n_s_TextDrawFill;
static PyObject *__pyx_n_s_TextDrawFillClip;
static PyObject *__pyx_n_s_TextDrawFillStroke;
static PyObject *__pyx_n_s_TextDrawFillStrokeClip;
static PyObject *__pyx_n_s_TextDrawInvisible;
static PyObject *__pyx_n_s_TextDrawStroke;
static PyObject *__pyx_n_s_TextDrawStrokeClip;
static PyObject *__pyx_n_s_TextDrawingMode;
static PyObject *__pyx_kp_s_That_type_of_comparison_is_not_i;
static PyObject *__pyx_kp_s_That_type_of_comparison_is_not_i_2;
static PyObject *__pyx_kp_s_The_clip_box_property_must_be_a;
static PyObject *__pyx_kp_s_The_format_requires_an_array_of;
static PyObject *__pyx_kp_s_The_format_requires_an_array_of_2;
static PyObject *__pyx_kp_s_The_stencil_property_must_be_an;
static PyObject *__pyx_kp_s_The_stencil_set_on_the_GraphicsS;
static PyObject *__pyx_kp_s_The_text_argument_must_be_unicod;
static PyObject *__pyx_n_s_Transform;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_kp_s_Unknown_enum_value_s;
static PyObject *__pyx_n_s_UserSpace;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_VectorFontCache;
static PyObject *__pyx_n_s_VertexSource;
static PyObject *__pyx_n_s_VertexSource___iter;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_kp_s__52;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_asarray;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_bottom_up;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_celiagg__celiagg;
static PyObject *__pyx_kp_s_celiagg_conversion_pxi;
static PyObject *__pyx_n_s_ch;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clockwise;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cls;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_convert_image;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_cx;
static PyObject *__pyx_n_s_cy;
static PyObject *__pyx_n_s_dct;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dst;
static PyObject *__pyx_n_s_dst_alpha;
static PyObject *__pyx_n_s_dst_dtype;
static PyObject *__pyx_n_s_dst_gray;
static PyObject *__pyx_n_s_dst_order;
static PyObject *__pyx_n_s_dst_rgb;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_end_angle;
static PyObject *__pyx_n_s_ends;
static PyObject *__pyx_kp_s_ends_argument_must_be_an_iterabl;
static PyObject *__pyx_n_s_enum;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_face_index;
static PyObject *__pyx_n_s_fill;
static PyObject *__pyx_kp_s_fill_must_be_a_Paint_instance;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_fmt;
static PyObject *__pyx_n_s_font;
static PyObject *__pyx_n_s_fontName;
static PyObject *__pyx_kp_s_font_must_be_a_Font_instance;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_kp_s_format_must_be_a_PixelFormat_val;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_from_format;
static PyObject *__pyx_n_s_fx;
static PyObject *__pyx_n_s_fy;
static PyObject *__pyx_n_s_g;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_h;
static PyObject *__pyx_n_s_height;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_iinfo;
static PyObject *__pyx_n_s_image;
static PyObject *__pyx_kp_s_image_argument_must_be_channel_i;
static PyObject *__pyx_kp_s_image_argument_must_not_be_None;
static PyObject *__pyx_kp_s_image_must_be_an_Image_instance;
static PyObject *__pyx_kp_s_image_must_be_an_ndarray_or_Imag;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_last_vertex;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_members;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_order;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_parents;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pixel_format;
static PyObject *__pyx_n_s_pixels;
static PyObject *__pyx_n_s_points;
static PyObject *__pyx_kp_s_points_argument_must_be_an_itera;
static PyObject *__pyx_kp_s_points_argument_must_be_an_itera_2;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_unpickle___Pyx_EnumMeta;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_r;
static PyObject *__pyx_n_s_radius;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_kp_s_rects_argument_must_be_an_iterab;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_remove;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_res;
static PyObject *__pyx_n_s_rx;
static PyObject *__pyx_n_s_ry;
static PyObject *__pyx_kp_s_s_s;
static PyObject *__pyx_kp_s_s_s_d;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self__this_cannot_be_converted_t;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_kp_s_shape_must_be_a_VertexSource_Pat;
static PyObject *__pyx_n_s_shx;
static PyObject *__pyx_n_s_shy;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_source;
static PyObject *__pyx_kp_s_source_must_be_a_VertexSource_in;
static PyObject *__pyx_n_s_spread;
static PyObject *__pyx_n_s_src;
static PyObject *__pyx_n_s_src_gray;
static PyObject *__pyx_kp_s_src_must_an_Image_instance;
static PyObject *__pyx_n_s_src_order;
static PyObject *__pyx_n_s_src_rgb;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_start_angle;
static PyObject *__pyx_n_s_starts;
static PyObject *__pyx_kp_s_starts_argument_must_be_an_itera;
static PyObject *__pyx_kp_s_starts_argument_must_match_lengt;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_kp_s_state_must_be_a_GraphicsState_in;
static PyObject *__pyx_n_s_stencil;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_n_s_stops;
static PyObject *__pyx_kp_s_stops_argument_must_be_an_iterab;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_stroke;
static PyObject *__pyx_kp_s_stroke_must_be_a_Paint_instance;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_style;
static PyObject *__pyx_n_s_sx;
static PyObject *__pyx_n_s_sy;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_text;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_to_format;
static PyObject *__pyx_kp_s_to_format_must_be_a_PixelFormat;
static PyObject *__pyx_n_s_transform;
static PyObject *__pyx_kp_s_transform_must_be_a_Transform_in;
static PyObject *__pyx_n_s_tx;
static PyObject *__pyx_n_s_ty;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_types;
static PyObject *__pyx_n_s_uint16;
static PyObject *__pyx_n_s_uint8;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_units;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_utf8;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_vertices;
static PyObject *__pyx_n_s_w;
static PyObject *__pyx_n_s_width;
static PyObject *__pyx_n_s_with_format;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_x1;
static PyObject *__pyx_n_s_x2;
static PyObject *__pyx_n_s_x_ctrl;
static PyObject *__pyx_n_s_x_ctrl1;
static PyObject *__pyx_n_s_x_ctrl2;
static PyObject *__pyx_n_s_x_to;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_y1;
static PyObject *__pyx_n_s_y2;
static PyObject *__pyx_n_s_y_ctrl;
static PyObject *__pyx_n_s_y_ctrl1;
static PyObject *__pyx_n_s_y_ctrl2;
static PyObject *__pyx_n_s_y_to;
static PyObject *__pyx_n_s_zip;
static int __pyx_pf_7celiagg_8_celiagg_4Font___cinit__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self, PyObject *__pyx_v_fontName, double __pyx_v_size, enum __pyx_t_7celiagg_8_celiagg_FontCacheType __pyx_v_ch, unsigned int __pyx_v_face_index); /* proto */
static void __pyx_pf_7celiagg_8_celiagg_4Font_2__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_4copy(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_10cache_type___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_10face_index___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_8filepath___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_6height___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_7hinting___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_4Font_7hinting_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self, bool __pyx_v_value); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_6width(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self, PyObject *__pyx_v_text); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_4Rect___cinit__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, double __pyx_v_x, double __pyx_v_y, double __pyx_v_w, double __pyx_v_h); /* proto */
static void __pyx_pf_7celiagg_8_celiagg_4Rect_2__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_4__repr__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_6__richcmp__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_1x___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_4Rect_1x_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_1w___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_4Rect_1w_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_1y___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_4Rect_1y_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_1h___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_4Rect_1h_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_5valid___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState___cinit__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static void __pyx_pf_7celiagg_8_celiagg_13GraphicsState_2__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_4__init__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_6copy(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_12anti_aliased___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12anti_aliased_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_aa); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_8clip_box___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_8clip_box_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_box); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_12drawing_mode___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12drawing_mode_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_DrawingMode __pyx_v_m); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode __pyx_v_m); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_10blend_mode___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10blend_mode_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_BlendMode __pyx_v_m); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_BlendMode __pyx_v_m); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_12master_alpha___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12master_alpha_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_a); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_10line_width___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10line_width_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_w); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_8line_cap___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_8line_cap_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_LineCap __pyx_v_cap); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_9line_join___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_9line_join_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_LineJoin __pyx_v_join); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_10inner_join___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10inner_join_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_InnerJoin __pyx_v_join); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_11miter_limit___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_11miter_limit_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, double __pyx_v_limit); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, double __pyx_v_limit); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_dashes); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, double __pyx_v_phase); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_7stencil___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_7stencil_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_img); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_5Image___cinit__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self, PyObject *__pyx_v_array, enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_pixel_format, bool __pyx_v_bottom_up); /* proto */
static void __pyx_pf_7celiagg_8_celiagg_5Image_2__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_4copy(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_6format___get__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_6pixels___get__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_6height___get__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_5width___get__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_pf_7celiagg_8_celiagg_10CanvasBase___dealloc__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_13channel_count___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_5width___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_6height___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_5array___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_5image___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_2clear(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, double __pyx_v_r, double __pyx_v_g, double __pyx_v_b, double __pyx_v_a); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_4draw_image(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_image, PyObject *__pyx_v_fmt, PyObject *__pyx_v_transform, PyObject *__pyx_v_state, PyObject *__pyx_v_bottom_up); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_6draw_shape(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_shape, PyObject *__pyx_v_transform, PyObject *__pyx_v_state, PyObject *__pyx_v_stroke, PyObject *__pyx_v_fill); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_8draw_text(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_font, PyObject *__pyx_v_transform, PyObject *__pyx_v_state, PyObject *__pyx_v_stroke, PyObject *__pyx_v_fill); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13CanvasRGBA128___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13CanvasRGBA128_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13CanvasRGBA128_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_12CanvasBGRA32___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12CanvasBGRA32_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12CanvasBGRA32_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_12CanvasRGBA32___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12CanvasRGBA32_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12CanvasRGBA32_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_11CanvasRGB24___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_11CanvasRGB24_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_11CanvasRGB24_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_10CanvasGA16___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasGA16_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasGA16_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_8CanvasG8___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_8CanvasG8_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_8CanvasG8_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_pf_7celiagg_8_celiagg_5Paint___dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Paint_9transform___get__(struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_5Paint_9transform_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self, PyObject *__pyx_v_trans); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Paint_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_5Paint_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint___cinit__(struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *__pyx_v_self, double __pyx_v_x1, double __pyx_v_y1, double __pyx_v_x2, double __pyx_v_y2, PyObject *__pyx_v_stops, enum __pyx_t_7celiagg_8_celiagg_GradientSpread __pyx_v_spread, enum __pyx_t_7celiagg_8_celiagg_GradientUnits __pyx_v_units); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint_2copy(struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint___cinit__(struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *__pyx_v_self, double __pyx_v_cx, double __pyx_v_cy, double __pyx_v_r, double __pyx_v_fx, double __pyx_v_fy, PyObject *__pyx_v_stops, enum __pyx_t_7celiagg_8_celiagg_GradientSpread __pyx_v_spread, enum __pyx_t_7celiagg_8_celiagg_GradientUnits __pyx_v_units); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint_2copy(struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_12PatternPaint___cinit__(struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_PatternStyle __pyx_v_style, PyObject *__pyx_v_image); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12PatternPaint_2_with_format(struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_fmt); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12PatternPaint_4copy(struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12PatternPaint_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12PatternPaint_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint___cinit__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_r, double __pyx_v_g, double __pyx_v_b, double __pyx_v_a); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_2__repr__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_4__richcmp__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_6copy(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_1a___get__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1a_2__set__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_1r___get__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1r_2__set__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_1g___get__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1g_2__set__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_1b___get__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1b_2__set__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_9Transform___init__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_sx, double __pyx_v_shy, double __pyx_v_shx, double __pyx_v_sy, double __pyx_v_tx, double __pyx_v_ty); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_9Transform_2__cinit__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static void __pyx_pf_7celiagg_8_celiagg_9Transform_4__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_6__repr__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_2sx___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_2sy___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_3shx___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_3shy___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_2tx___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_2ty___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_8copy(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_10reset(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_12invert(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_14scale(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_sx, double __pyx_v_sy); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_16rotate(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_angle); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_18skew(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_sx, double __pyx_v_sy); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_20translate(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_x, double __pyx_v_y); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_22premultiply(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_24multiply(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_26worldToScreen(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_x, double __pyx_v_y); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_28screenToWorld(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_x, double __pyx_v_y); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_30__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_32__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_pf_7celiagg_8_celiagg_12VertexSource___dealloc__(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_2__iter__(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_5bounding_rect(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_7copy(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_9final_point(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_11length(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_13vertices(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_15__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_17__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_7BSpline___cinit__(struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self, PyObject *__pyx_v_points); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_7BSpline_2copy(struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_7BSpline_4final_point(struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_7BSpline_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_7BSpline_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_4Path___cinit__(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_2copy(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_4final_point(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_6begin(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_8close(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_10reset(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_12add_path(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_14move_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x, double __pyx_v_y); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_16line_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x, double __pyx_v_y); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_18arc_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x1, double __pyx_v_y1, double __pyx_v_x2, double __pyx_v_y2, double __pyx_v_radius); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_20quadric_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x_ctrl, double __pyx_v_y_ctrl, double __pyx_v_x_to, double __pyx_v_y_to); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_22cubic_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x_ctrl1, double __pyx_v_y_ctrl1, double __pyx_v_x_ctrl2, double __pyx_v_y_ctrl2, double __pyx_v_x_to, double __pyx_v_y_to); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_24arc(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x, double __pyx_v_y, double __pyx_v_radius, double __pyx_v_start_angle, double __pyx_v_end_angle, bool __pyx_v_clockwise); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_26ellipse(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_cx, double __pyx_v_cy, double __pyx_v_rx, double __pyx_v_ry); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_28rect(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x, double __pyx_v_y, double __pyx_v_width, double __pyx_v_height); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_30rects(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, PyObject *__pyx_v_rects); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_32lines(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, PyObject *__pyx_v_points); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_34lines_set(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, PyObject *__pyx_v_starts, PyObject *__pyx_v_ends); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints___cinit__(struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_points); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_2copy(struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_4final_point(struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7celiagg_8_celiagg_convert_image(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src, PyObject *__pyx_v_to_format, PyObject *__pyx_v_bottom_up); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Font(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Rect(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_GraphicsState(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Image(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasRGBA128(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasBGRA32(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasRGBA32(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasRGB24(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasGA16(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasG8(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Paint(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_LinearGradientPaint(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_RadialGradientPaint(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_PatternPaint(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_SolidPaint(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Transform(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_VertexSource(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_BSpline(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Path(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg_ShapeAtPoints(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7celiagg_8_celiagg___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___Pyx_EnumMeta(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items = {0, &__pyx_n_s_items, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_remove = {0, &__pyx_n_s_remove, 0, 0, 0};
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_float_3_0;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_222419149;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__7;
static PyObject *__pyx_slice__8;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__56;
static PyObject *__pyx_slice__76;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_tuple__78;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__80;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__92;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__94;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__97;
static PyObject *__pyx_tuple__98;
static PyObject *__pyx_tuple__99;
static PyObject *__pyx_tuple__100;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__102;
static PyObject *__pyx_tuple__103;
static PyObject *__pyx_tuple__105;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__108;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__112;
static PyObject *__pyx_tuple__114;
static PyObject *__pyx_tuple__115;
static PyObject *__pyx_tuple__116;
static PyObject *__pyx_tuple__117;
static PyObject *__pyx_tuple__118;
static PyObject *__pyx_tuple__119;
static PyObject *__pyx_codeobj__104;
static PyObject *__pyx_codeobj__106;
static PyObject *__pyx_codeobj__109;
static PyObject *__pyx_codeobj__111;
static PyObject *__pyx_codeobj__113;
static PyObject *__pyx_codeobj__120;
/* Late includes */

/* "celiagg/_celiagg.pyx":52
 * 
 * 
 * cdef _get_utf8_text(text, exp_msg):             # <<<<<<<<<<<<<<
 *     # Ensure UTF-8 encoded text is passed to C++ code.
 *     if isinstance(text, unicode):
 */

static PyObject *__pyx_f_7celiagg_8_celiagg__get_utf8_text(PyObject *__pyx_v_text, PyObject *__pyx_v_exp_msg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_get_utf8_text", 0);

  /* "celiagg/_celiagg.pyx":54
 * cdef _get_utf8_text(text, exp_msg):
 *     # Ensure UTF-8 encoded text is passed to C++ code.
 *     if isinstance(text, unicode):             # <<<<<<<<<<<<<<
 *         return text.encode('utf8')
 *     elif isinstance(text, bytes):
 */
  __pyx_t_1 = PyUnicode_Check(__pyx_v_text); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "celiagg/_celiagg.pyx":55
 *     # Ensure UTF-8 encoded text is passed to C++ code.
 *     if isinstance(text, unicode):
 *         return text.encode('utf8')             # <<<<<<<<<<<<<<
 *     elif isinstance(text, bytes):
 *         return text
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_text, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_n_s_utf8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_utf8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "celiagg/_celiagg.pyx":54
 * cdef _get_utf8_text(text, exp_msg):
 *     # Ensure UTF-8 encoded text is passed to C++ code.
 *     if isinstance(text, unicode):             # <<<<<<<<<<<<<<
 *         return text.encode('utf8')
 *     elif isinstance(text, bytes):
 */
  }

  /* "celiagg/_celiagg.pyx":56
 *     if isinstance(text, unicode):
 *         return text.encode('utf8')
 *     elif isinstance(text, bytes):             # <<<<<<<<<<<<<<
 *         return text
 *     else:
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_text); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (likely(__pyx_t_1)) {

    /* "celiagg/_celiagg.pyx":57
 *         return text.encode('utf8')
 *     elif isinstance(text, bytes):
 *         return text             # <<<<<<<<<<<<<<
 *     else:
 *         raise TypeError(exp_msg)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_text);
    __pyx_r = __pyx_v_text;
    goto __pyx_L0;

    /* "celiagg/_celiagg.pyx":56
 *     if isinstance(text, unicode):
 *         return text.encode('utf8')
 *     elif isinstance(text, bytes):             # <<<<<<<<<<<<<<
 *         return text
 *     else:
 */
  }

  /* "celiagg/_celiagg.pyx":59
 *         return text
 *     else:
 *         raise TypeError(exp_msg)             # <<<<<<<<<<<<<<
 * 
 * include "enums.pxi"
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_v_exp_msg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 59, __pyx_L1_error)
  }

  /* "celiagg/_celiagg.pyx":52
 * 
 * 
 * cdef _get_utf8_text(text, exp_msg):             # <<<<<<<<<<<<<<
 *     # Ensure UTF-8 encoded text is passed to C++ code.
 *     if isinstance(text, unicode):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg._get_utf8_text", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/font.pxi":43
 *         cdef _font.Font* _this
 * 
 *         def __cinit__(self, fontName, double size, FontCacheType ch,             # <<<<<<<<<<<<<<
 *                       unsigned face_index=0):
 *             fontName = _get_utf8_text(fontName,
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_4Font_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_4Font_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_fontName = 0;
  double __pyx_v_size;
  enum __pyx_t_7celiagg_8_celiagg_FontCacheType __pyx_v_ch;
  unsigned int __pyx_v_face_index;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fontName,&__pyx_n_s_size,&__pyx_n_s_ch,&__pyx_n_s_face_index,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fontName)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 4, 1); __PYX_ERR(3, 43, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ch)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 4, 2); __PYX_ERR(3, 43, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_face_index);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(3, 43, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_fontName = values[0];
    __pyx_v_size = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_size == (double)-1) && PyErr_Occurred())) __PYX_ERR(3, 43, __pyx_L3_error)
    __pyx_v_ch = ((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(values[2])); if (unlikely(PyErr_Occurred())) __PYX_ERR(3, 43, __pyx_L3_error)
    if (values[3]) {
      __pyx_v_face_index = __Pyx_PyInt_As_unsigned_int(values[3]); if (unlikely((__pyx_v_face_index == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(3, 44, __pyx_L3_error)
    } else {
      __pyx_v_face_index = ((unsigned int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(3, 43, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Font.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self), __pyx_v_fontName, __pyx_v_size, __pyx_v_ch, __pyx_v_face_index);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_4Font___cinit__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self, PyObject *__pyx_v_fontName, double __pyx_v_size, enum __pyx_t_7celiagg_8_celiagg_FontCacheType __pyx_v_ch, unsigned int __pyx_v_face_index) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_fontName);

  /* "celiagg/font.pxi":45
 *         def __cinit__(self, fontName, double size, FontCacheType ch,
 *                       unsigned face_index=0):
 *             fontName = _get_utf8_text(fontName,             # <<<<<<<<<<<<<<
 *                                     "Font path must be a unicode string")
 *             self._this = new _font.Font(fontName, size, ch, face_index)
 */
  __pyx_t_1 = __pyx_f_7celiagg_8_celiagg__get_utf8_text(__pyx_v_fontName, __pyx_kp_s_Font_path_must_be_a_unicode_stri); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_fontName, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/font.pxi":47
 *             fontName = _get_utf8_text(fontName,
 *                                     "Font path must be a unicode string")
 *             self._this = new _font.Font(fontName, size, ch, face_index)             # <<<<<<<<<<<<<<
 * 
 *         def __dealloc__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_v_fontName); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(3, 47, __pyx_L1_error)
  __pyx_v_self->_this = new Font(__pyx_t_2, __pyx_v_size, ((enum Font::FontCacheType)__pyx_v_ch), __pyx_v_face_index);

  /* "celiagg/font.pxi":43
 *         cdef _font.Font* _this
 * 
 *         def __cinit__(self, fontName, double size, FontCacheType ch,             # <<<<<<<<<<<<<<
 *                       unsigned face_index=0):
 *             fontName = _get_utf8_text(fontName,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Font.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fontName);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/font.pxi":49
 *             self._this = new _font.Font(fontName, size, ch, face_index)
 * 
 *         def __dealloc__(self):             # <<<<<<<<<<<<<<
 *             del self._this
 * 
 */

/* Python wrapper */
static void __pyx_pw_7celiagg_8_celiagg_4Font_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7celiagg_8_celiagg_4Font_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7celiagg_8_celiagg_4Font_2__dealloc__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7celiagg_8_celiagg_4Font_2__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "celiagg/font.pxi":50
 * 
 *         def __dealloc__(self):
 *             del self._this             # <<<<<<<<<<<<<<
 * 
 *         def copy(self):
 */
  delete __pyx_v_self->_this;

  /* "celiagg/font.pxi":49
 *             self._this = new _font.Font(fontName, size, ch, face_index)
 * 
 *         def __dealloc__(self):             # <<<<<<<<<<<<<<
 *             del self._this
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "celiagg/font.pxi":52
 *             del self._this
 * 
 *         def copy(self):             # <<<<<<<<<<<<<<
 *             return Font(self._this.filepath(), self._this.height(),
 *                         self._this.cache_type(), self._this.face_index())
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_5copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_5copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_4copy(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_4copy(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/font.pxi":53
 * 
 *         def copy(self):
 *             return Font(self._this.filepath(), self._this.height(),             # <<<<<<<<<<<<<<
 *                         self._this.cache_type(), self._this.face_index())
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->_this->filepath()); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->_this->height()); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "celiagg/font.pxi":54
 *         def copy(self):
 *             return Font(self._this.filepath(), self._this.height(),
 *                         self._this.cache_type(), self._this.face_index())             # <<<<<<<<<<<<<<
 * 
 *         property cache_type:
 */
  __pyx_t_3 = __Pyx_PyInt_From_enum__Font_3a__3a_FontCacheType(__pyx_v_self->_this->cache_type()); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->_this->face_index()); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "celiagg/font.pxi":53
 * 
 *         def copy(self):
 *             return Font(self._this.filepath(), self._this.height(),             # <<<<<<<<<<<<<<
 *                         self._this.cache_type(), self._this.face_index())
 * 
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(3, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_Font), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "celiagg/font.pxi":52
 *             del self._this
 * 
 *         def copy(self):             # <<<<<<<<<<<<<<
 *             return Font(self._this.filepath(), self._this.height(),
 *                         self._this.cache_type(), self._this.face_index())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg.Font.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/font.pxi":57
 * 
 *         property cache_type:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return FontCacheType(self._this.cache_type())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_10cache_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_10cache_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_10cache_type___get__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_10cache_type___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/font.pxi":58
 *         property cache_type:
 *             def __get__(self):
 *                 return FontCacheType(self._this.cache_type())             # <<<<<<<<<<<<<<
 * 
 *         property face_index:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_FontCacheType); if (unlikely(!__pyx_t_2)) __PYX_ERR(3, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum__Font_3a__3a_FontCacheType(__pyx_v_self->_this->cache_type()); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/font.pxi":57
 * 
 *         property cache_type:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return FontCacheType(self._this.cache_type())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.Font.cache_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/font.pxi":61
 * 
 *         property face_index:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return self._this.face_index()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_10face_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_10face_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_10face_index___get__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_10face_index___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/font.pxi":62
 *         property face_index:
 *             def __get__(self):
 *                 return self._this.face_index()             # <<<<<<<<<<<<<<
 * 
 *         property filepath:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->_this->face_index()); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/font.pxi":61
 * 
 *         property face_index:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return self._this.face_index()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Font.face_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/font.pxi":65
 * 
 *         property filepath:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return self._this.filepath()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_8filepath_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_8filepath_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_8filepath___get__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_8filepath___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/font.pxi":66
 *         property filepath:
 *             def __get__(self):
 *                 return self._this.filepath()             # <<<<<<<<<<<<<<
 * 
 *         property height:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->_this->filepath()); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/font.pxi":65
 * 
 *         property filepath:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return self._this.filepath()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Font.filepath.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/font.pxi":69
 * 
 *         property height:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return self._this.height()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_6height_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_6height_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_6height___get__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_6height___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/font.pxi":70
 *         property height:
 *             def __get__(self):
 *                 return self._this.height()             # <<<<<<<<<<<<<<
 * 
 *         property hinting:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->height()); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/font.pxi":69
 * 
 *         property height:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return self._this.height()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Font.height.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/font.pxi":73
 * 
 *         property hinting:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return self._this.hinting()
 *             def __set__(self, bool value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_7hinting_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_7hinting_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_7hinting___get__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_7hinting___get__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/font.pxi":74
 *         property hinting:
 *             def __get__(self):
 *                 return self._this.hinting()             # <<<<<<<<<<<<<<
 *             def __set__(self, bool value):
 *                 self._this.hinting(value)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_this->hinting()); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/font.pxi":73
 * 
 *         property hinting:
 *             def __get__(self):             # <<<<<<<<<<<<<<
 *                 return self._this.hinting()
 *             def __set__(self, bool value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Font.hinting.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/font.pxi":75
 *             def __get__(self):
 *                 return self._this.hinting()
 *             def __set__(self, bool value):             # <<<<<<<<<<<<<<
 *                 self._this.hinting(value)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_4Font_7hinting_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_4Font_7hinting_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  bool __pyx_v_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyObject_IsTrue(__pyx_arg_value); if (unlikely((__pyx_v_value == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(3, 75, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Font.hinting.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_7hinting_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self), ((bool)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_4Font_7hinting_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self, bool __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/font.pxi":76
 *                 return self._this.hinting()
 *             def __set__(self, bool value):
 *                 self._this.hinting(value)             # <<<<<<<<<<<<<<
 * 
 *         def width(self, text):
 */
  __pyx_v_self->_this->hinting(__pyx_v_value);

  /* "celiagg/font.pxi":75
 *             def __get__(self):
 *                 return self._this.hinting()
 *             def __set__(self, bool value):             # <<<<<<<<<<<<<<
 *                 self._this.hinting(value)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/font.pxi":78
 *                 self._this.hinting(value)
 * 
 *         def width(self, text):             # <<<<<<<<<<<<<<
 *             """width(text)
 *             Measures the width of a string rendered with the font.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_7width(PyObject *__pyx_v_self, PyObject *__pyx_v_text); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Font_6width[] = "width(text)\n            Measures the width of a string rendered with the font.\n\n            :param text: a unicode string\n            ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_7width(PyObject *__pyx_v_self, PyObject *__pyx_v_text) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("width (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_6width(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self), ((PyObject *)__pyx_v_text));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_6width(struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self, PyObject *__pyx_v_text) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  __Pyx_RefNannySetupContext("width", 0);
  __Pyx_INCREF(__pyx_v_text);

  /* "celiagg/font.pxi":84
 *             :param text: a unicode string
 *             """
 *             text = _get_utf8_text(text, "Argument must be a unicode string")             # <<<<<<<<<<<<<<
 *             return self._this.string_width(text)
 * ELSE:
 */
  __pyx_t_1 = __pyx_f_7celiagg_8_celiagg__get_utf8_text(__pyx_v_text, __pyx_kp_s_Argument_must_be_a_unicode_strin); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_text, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/font.pxi":85
 *             """
 *             text = _get_utf8_text(text, "Argument must be a unicode string")
 *             return self._this.string_width(text)             # <<<<<<<<<<<<<<
 * ELSE:
 *     cdef class Font:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_AsWritableString(__pyx_v_text); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(3, 85, __pyx_L1_error)
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->string_width(__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/font.pxi":78
 *                 self._this.hinting(value)
 * 
 *         def width(self, text):             # <<<<<<<<<<<<<<
 *             """width(text)
 *             Measures the width of a string rendered with the font.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Font.width", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_text);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_8__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Font.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Font_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Font_10__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Font_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Font.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":40
 *     cdef _graphics_state.Rect* _this
 * 
 *     def __cinit__(self, double x=0., double y=0., double w=0., double h=0.):             # <<<<<<<<<<<<<<
 *         self._this = new _graphics_state.Rect(x, y, x + w, y + h)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x;
  double __pyx_v_y;
  double __pyx_v_w;
  double __pyx_v_h;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_w,&__pyx_n_s_h,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_w);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 40, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_x = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 40, __pyx_L3_error)
    } else {
      __pyx_v_x = ((double)0.);
    }
    if (values[1]) {
      __pyx_v_y = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 40, __pyx_L3_error)
    } else {
      __pyx_v_y = ((double)0.);
    }
    if (values[2]) {
      __pyx_v_w = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_w == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 40, __pyx_L3_error)
    } else {
      __pyx_v_w = ((double)0.);
    }
    if (values[3]) {
      __pyx_v_h = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_h == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 40, __pyx_L3_error)
    } else {
      __pyx_v_h = ((double)0.);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 40, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Rect.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self), __pyx_v_x, __pyx_v_y, __pyx_v_w, __pyx_v_h);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_4Rect___cinit__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, double __pyx_v_x, double __pyx_v_y, double __pyx_v_w, double __pyx_v_h) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/graphics_state.pxi":41
 * 
 *     def __cinit__(self, double x=0., double y=0., double w=0., double h=0.):
 *         self._this = new _graphics_state.Rect(x, y, x + w, y + h)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->_this = new GraphicsState::Rect(__pyx_v_x, __pyx_v_y, (__pyx_v_x + __pyx_v_w), (__pyx_v_y + __pyx_v_h));

  /* "celiagg/graphics_state.pxi":40
 *     cdef _graphics_state.Rect* _this
 * 
 *     def __cinit__(self, double x=0., double y=0., double w=0., double h=0.):             # <<<<<<<<<<<<<<
 *         self._this = new _graphics_state.Rect(x, y, x + w, y + h)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":43
 *         self._this = new _graphics_state.Rect(x, y, x + w, y + h)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

/* Python wrapper */
static void __pyx_pw_7celiagg_8_celiagg_4Rect_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7celiagg_8_celiagg_4Rect_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7celiagg_8_celiagg_4Rect_2__dealloc__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7celiagg_8_celiagg_4Rect_2__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "celiagg/graphics_state.pxi":44
 * 
 *     def __dealloc__(self):
 *         del self._this             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  delete __pyx_v_self->_this;

  /* "celiagg/graphics_state.pxi":43
 *         self._this = new _graphics_state.Rect(x, y, x + w, y + h)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "celiagg/graphics_state.pxi":46
 *         del self._this
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         name = type(self).__name__
 *         width = self._this.x2 - self._this.x1
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_4__repr__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_4__repr__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self) {
  PyObject *__pyx_v_name = NULL;
  double __pyx_v_width;
  double __pyx_v_height;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "celiagg/graphics_state.pxi":47
 * 
 *     def __repr__(self):
 *         name = type(self).__name__             # <<<<<<<<<<<<<<
 *         width = self._this.x2 - self._this.x1
 *         height = self._this.y2 - self._this.y1
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/graphics_state.pxi":48
 *     def __repr__(self):
 *         name = type(self).__name__
 *         width = self._this.x2 - self._this.x1             # <<<<<<<<<<<<<<
 *         height = self._this.y2 - self._this.y1
 *         return "{}({}, {}, {}, {})".format(name, self._this.x1, self._this.y1,
 */
  __pyx_v_width = (__pyx_v_self->_this->x2 - __pyx_v_self->_this->x1);

  /* "celiagg/graphics_state.pxi":49
 *         name = type(self).__name__
 *         width = self._this.x2 - self._this.x1
 *         height = self._this.y2 - self._this.y1             # <<<<<<<<<<<<<<
 *         return "{}({}, {}, {}, {})".format(name, self._this.x1, self._this.y1,
 *                                            width, height)
 */
  __pyx_v_height = (__pyx_v_self->_this->y2 - __pyx_v_self->_this->y1);

  /* "celiagg/graphics_state.pxi":50
 *         width = self._this.x2 - self._this.x1
 *         height = self._this.y2 - self._this.y1
 *         return "{}({}, {}, {}, {})".format(name, self._this.x1, self._this.y1,             # <<<<<<<<<<<<<<
 *                                            width, height)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s__3, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->_this->x1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->_this->y1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "celiagg/graphics_state.pxi":51
 *         height = self._this.y2 - self._this.y1
 *         return "{}({}, {}, {}, {})".format(name, self._this.x1, self._this.y1,
 *                                            width, height)             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(Rect self, Rect other, int op):
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_width); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_height); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[6] = {__pyx_t_7, __pyx_v_name, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[6] = {__pyx_t_7, __pyx_v_name, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(5+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_name);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 4+__pyx_t_8, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":46
 *         del self._this
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         name = type(self).__name__
 *         width = self._this.x2 - self._this.x1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":53
 *                                            width, height)
 * 
 *     def __richcmp__(Rect self, Rect other, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return (self._this.x1 == other._this.x1 and
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_7__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_7__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7celiagg_8_celiagg_Rect, 1, "other", 0))) __PYX_ERR(1, 53, __pyx_L1_error)
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_6__richcmp__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self), ((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_6__richcmp__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "celiagg/graphics_state.pxi":54
 * 
 *     def __richcmp__(Rect self, Rect other, int op):
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return (self._this.x1 == other._this.x1 and
 *                     self._this.x2 == other._this.x2 and
 */
  __pyx_t_1 = ((__pyx_v_op == 2) != 0);
  if (likely(__pyx_t_1)) {

    /* "celiagg/graphics_state.pxi":55
 *     def __richcmp__(Rect self, Rect other, int op):
 *         if op == 2:  # ==
 *             return (self._this.x1 == other._this.x1 and             # <<<<<<<<<<<<<<
 *                     self._this.x2 == other._this.x2 and
 *                     self._this.y1 == other._this.y1 and
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = (__pyx_v_self->_this->x1 == __pyx_v_other->_this->x1);
    if (__pyx_t_1) {
    } else {
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 55, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L4_bool_binop_done;
    }

    /* "celiagg/graphics_state.pxi":56
 *         if op == 2:  # ==
 *             return (self._this.x1 == other._this.x1 and
 *                     self._this.x2 == other._this.x2 and             # <<<<<<<<<<<<<<
 *                     self._this.y1 == other._this.y1 and
 *                     self._this.y2 == other._this.y2)
 */
    __pyx_t_1 = (__pyx_v_self->_this->x2 == __pyx_v_other->_this->x2);
    if (__pyx_t_1) {
    } else {
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 56, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L4_bool_binop_done;
    }

    /* "celiagg/graphics_state.pxi":57
 *             return (self._this.x1 == other._this.x1 and
 *                     self._this.x2 == other._this.x2 and
 *                     self._this.y1 == other._this.y1 and             # <<<<<<<<<<<<<<
 *                     self._this.y2 == other._this.y2)
 *         else:
 */
    __pyx_t_1 = (__pyx_v_self->_this->y1 == __pyx_v_other->_this->y1);
    if (__pyx_t_1) {
    } else {
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 57, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L4_bool_binop_done;
    }

    /* "celiagg/graphics_state.pxi":58
 *                     self._this.x2 == other._this.x2 and
 *                     self._this.y1 == other._this.y1 and
 *                     self._this.y2 == other._this.y2)             # <<<<<<<<<<<<<<
 *         else:
 *             msg = "That type of comparison is not implemented for Rect"
 */
    __pyx_t_1 = (__pyx_v_self->_this->y2 == __pyx_v_other->_this->y2);
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_L4_bool_binop_done:;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "celiagg/graphics_state.pxi":54
 * 
 *     def __richcmp__(Rect self, Rect other, int op):
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return (self._this.x1 == other._this.x1 and
 *                     self._this.x2 == other._this.x2 and
 */
  }

  /* "celiagg/graphics_state.pxi":60
 *                     self._this.y2 == other._this.y2)
 *         else:
 *             msg = "That type of comparison is not implemented for Rect"             # <<<<<<<<<<<<<<
 *             raise NotImplementedError(msg)
 * 
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_kp_s_That_type_of_comparison_is_not_i);
    __pyx_v_msg = __pyx_kp_s_That_type_of_comparison_is_not_i;

    /* "celiagg/graphics_state.pxi":61
 *         else:
 *             msg = "That type of comparison is not implemented for Rect"
 *             raise NotImplementedError(msg)             # <<<<<<<<<<<<<<
 * 
 *     property x:
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_v_msg); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 61, __pyx_L1_error)
  }

  /* "celiagg/graphics_state.pxi":53
 *                                            width, height)
 * 
 *     def __richcmp__(Rect self, Rect other, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return (self._this.x1 == other._this.x1 and
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":64
 * 
 *     property x:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.x1
 *         def __set__(self, value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_1x_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_1x_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_1x___get__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_1x___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":65
 *     property x:
 *         def __get__(self):
 *             return self._this.x1             # <<<<<<<<<<<<<<
 *         def __set__(self, value):
 *             self._this.x1 = value
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->x1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":64
 * 
 *     property x:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.x1
 *         def __set__(self, value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.x.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":66
 *         def __get__(self):
 *             return self._this.x1
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self._this.x1 = value
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1x_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1x_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_1x_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_4Rect_1x_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":67
 *             return self._this.x1
 *         def __set__(self, value):
 *             self._this.x1 = value             # <<<<<<<<<<<<<<
 * 
 *     property w:
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 67, __pyx_L1_error)
  __pyx_v_self->_this->x1 = __pyx_t_1;

  /* "celiagg/graphics_state.pxi":66
 *         def __get__(self):
 *             return self._this.x1
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self._this.x1 = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Rect.x.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":70
 * 
 *     property w:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.x2 - self._this.x1
 *         def __set__(self, value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_1w_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_1w_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_1w___get__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_1w___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":71
 *     property w:
 *         def __get__(self):
 *             return self._this.x2 - self._this.x1             # <<<<<<<<<<<<<<
 *         def __set__(self, value):
 *             self._this.x2 = self._this.x1 + value
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_this->x2 - __pyx_v_self->_this->x1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":70
 * 
 *     property w:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.x2 - self._this.x1
 *         def __set__(self, value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.w.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":72
 *         def __get__(self):
 *             return self._this.x2 - self._this.x1
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self._this.x2 = self._this.x1 + value
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1w_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1w_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_1w_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_4Rect_1w_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":73
 *             return self._this.x2 - self._this.x1
 *         def __set__(self, value):
 *             self._this.x2 = self._this.x1 + value             # <<<<<<<<<<<<<<
 * 
 *     property y:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->x1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->_this->x2 = __pyx_t_3;

  /* "celiagg/graphics_state.pxi":72
 *         def __get__(self):
 *             return self._this.x2 - self._this.x1
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self._this.x2 = self._this.x1 + value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.w.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":76
 * 
 *     property y:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.y1
 *         def __set__(self, value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_1y_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_1y_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_1y___get__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_1y___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":77
 *     property y:
 *         def __get__(self):
 *             return self._this.y1             # <<<<<<<<<<<<<<
 *         def __set__(self, value):
 *             self._this.y1 = value
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->y1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":76
 * 
 *     property y:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.y1
 *         def __set__(self, value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.y.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":78
 *         def __get__(self):
 *             return self._this.y1
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self._this.y1 = value
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1y_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1y_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_1y_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_4Rect_1y_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":79
 *             return self._this.y1
 *         def __set__(self, value):
 *             self._this.y1 = value             # <<<<<<<<<<<<<<
 * 
 *     property h:
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 79, __pyx_L1_error)
  __pyx_v_self->_this->y1 = __pyx_t_1;

  /* "celiagg/graphics_state.pxi":78
 *         def __get__(self):
 *             return self._this.y1
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self._this.y1 = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Rect.y.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":82
 * 
 *     property h:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.y2 - self._this.y1
 *         def __set__(self, value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_1h_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_1h_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_1h___get__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_1h___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":83
 *     property h:
 *         def __get__(self):
 *             return self._this.y2 - self._this.y1             # <<<<<<<<<<<<<<
 *         def __set__(self, value):
 *             self._this.y2 = self._this.y1 + value
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_this->y2 - __pyx_v_self->_this->y1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":82
 * 
 *     property h:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.y2 - self._this.y1
 *         def __set__(self, value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.h.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":84
 *         def __get__(self):
 *             return self._this.y2 - self._this.y1
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self._this.y2 = self._this.y1 + value
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1h_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_4Rect_1h_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_1h_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_4Rect_1h_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  double __pyx_t_3;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":85
 *             return self._this.y2 - self._this.y1
 *         def __set__(self, value):
 *             self._this.y2 = self._this.y1 + value             # <<<<<<<<<<<<<<
 * 
 *     property valid:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->y1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->_this->y2 = __pyx_t_3;

  /* "celiagg/graphics_state.pxi":84
 *         def __get__(self):
 *             return self._this.y2 - self._this.y1
 *         def __set__(self, value):             # <<<<<<<<<<<<<<
 *             self._this.y2 = self._this.y1 + value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.h.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":88
 * 
 *     property valid:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.is_valid()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_5valid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_5valid_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_5valid___get__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_5valid___get__(struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":89
 *     property valid:
 *         def __get__(self):
 *             return self._this.is_valid()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_this->is_valid()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":88
 * 
 *     property valid:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.is_valid()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.valid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_8__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Rect_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Rect_10__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Rect_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Rect.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":120
 *     cdef Image _stencil_img
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._this = new _graphics_state.GraphicsState()
 *         self._stencil_img = None
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState___cinit__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/graphics_state.pxi":121
 * 
 *     def __cinit__(self):
 *         self._this = new _graphics_state.GraphicsState()             # <<<<<<<<<<<<<<
 *         self._stencil_img = None
 * 
 */
  __pyx_v_self->_this = new GraphicsState();

  /* "celiagg/graphics_state.pxi":122
 *     def __cinit__(self):
 *         self._this = new _graphics_state.GraphicsState()
 *         self._stencil_img = None             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_stencil_img);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_stencil_img));
  __pyx_v_self->_stencil_img = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)Py_None);

  /* "celiagg/graphics_state.pxi":120
 *     cdef Image _stencil_img
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._this = new _graphics_state.GraphicsState()
 *         self._stencil_img = None
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":124
 *         self._stencil_img = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

/* Python wrapper */
static void __pyx_pw_7celiagg_8_celiagg_13GraphicsState_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7celiagg_8_celiagg_13GraphicsState_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7celiagg_8_celiagg_13GraphicsState_2__dealloc__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7celiagg_8_celiagg_13GraphicsState_2__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "celiagg/graphics_state.pxi":125
 * 
 *     def __dealloc__(self):
 *         del self._this             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, **kwargs):
 */
  delete __pyx_v_self->_this;

  /* "celiagg/graphics_state.pxi":124
 *         self._stencil_img = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "celiagg/graphics_state.pxi":127
 *         del self._this
 * 
 *     def __init__(self, **kwargs):             # <<<<<<<<<<<<<<
 *         properties = self._propnames()
 *         for key, value in kwargs.items():
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (__pyx_kwds && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 1))) return -1;
  __pyx_v_kwargs = (__pyx_kwds) ? PyDict_Copy(__pyx_kwds) : PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_4__init__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_4__init__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_properties = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "celiagg/graphics_state.pxi":128
 * 
 *     def __init__(self, **kwargs):
 *         properties = self._propnames()             # <<<<<<<<<<<<<<
 *         for key, value in kwargs.items():
 *             if key in properties:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self->__pyx_vtab)->_propnames(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_properties = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/graphics_state.pxi":129
 *     def __init__(self, **kwargs):
 *         properties = self._propnames()
 *         for key, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             if key in properties:
 *                 setattr(self, key, value)
 */
  __pyx_t_1 = __Pyx_PyDict_Items(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 129, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 129, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 129, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 129, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 129, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 129, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 129, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(1, 129, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 129, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "celiagg/graphics_state.pxi":130
 *         properties = self._propnames()
 *         for key, value in kwargs.items():
 *             if key in properties:             # <<<<<<<<<<<<<<
 *                 setattr(self, key, value)
 * 
 */
    __pyx_t_9 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_v_properties, Py_EQ)); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(1, 130, __pyx_L1_error)
    __pyx_t_10 = (__pyx_t_9 != 0);
    if (__pyx_t_10) {

      /* "celiagg/graphics_state.pxi":131
 *         for key, value in kwargs.items():
 *             if key in properties:
 *                 setattr(self, key, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef list _propnames(self):
 */
      __pyx_t_11 = PyObject_SetAttr(((PyObject *)__pyx_v_self), __pyx_v_key, __pyx_v_value); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(1, 131, __pyx_L1_error)

      /* "celiagg/graphics_state.pxi":130
 *         properties = self._propnames()
 *         for key, value in kwargs.items():
 *             if key in properties:             # <<<<<<<<<<<<<<
 *                 setattr(self, key, value)
 * 
 */
    }

    /* "celiagg/graphics_state.pxi":129
 *     def __init__(self, **kwargs):
 *         properties = self._propnames()
 *         for key, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             if key in properties:
 *                 setattr(self, key, value)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "celiagg/graphics_state.pxi":127
 *         del self._this
 * 
 *     def __init__(self, **kwargs):             # <<<<<<<<<<<<<<
 *         properties = self._propnames()
 *         for key, value in kwargs.items():
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_properties);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":133
 *                 setattr(self, key, value)
 * 
 *     cdef list _propnames(self):             # <<<<<<<<<<<<<<
 *         return [name for name in GraphicsState.__dict__.keys()
 *                 if isinstance(getattr(GraphicsState, name),
 */

static PyObject *__pyx_f_7celiagg_8_celiagg_13GraphicsState__propnames(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("_propnames", 0);

  /* "celiagg/graphics_state.pxi":134
 * 
 *     cdef list _propnames(self):
 *         return [name for name in GraphicsState.__dict__.keys()             # <<<<<<<<<<<<<<
 *                 if isinstance(getattr(GraphicsState, name),
 *                               GetSetDescriptorType)]
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_GraphicsState), __pyx_n_s_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_keys); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 134, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 134, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 134, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 134, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 134, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "celiagg/graphics_state.pxi":135
 *     cdef list _propnames(self):
 *         return [name for name in GraphicsState.__dict__.keys()
 *                 if isinstance(getattr(GraphicsState, name),             # <<<<<<<<<<<<<<
 *                               GetSetDescriptorType)]
 * 
 */
    __pyx_t_2 = __Pyx_GetAttr(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_GraphicsState), __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);

    /* "celiagg/graphics_state.pxi":136
 *         return [name for name in GraphicsState.__dict__.keys()
 *                 if isinstance(getattr(GraphicsState, name),
 *                               GetSetDescriptorType)]             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_GetSetDescriptorType); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "celiagg/graphics_state.pxi":135
 *     cdef list _propnames(self):
 *         return [name for name in GraphicsState.__dict__.keys()
 *                 if isinstance(getattr(GraphicsState, name),             # <<<<<<<<<<<<<<
 *                               GetSetDescriptorType)]
 * 
 */
    __pyx_t_7 = PyObject_IsInstance(__pyx_t_2, __pyx_t_3); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 135, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "celiagg/graphics_state.pxi":134
 * 
 *     cdef list _propnames(self):
 *         return [name for name in GraphicsState.__dict__.keys()             # <<<<<<<<<<<<<<
 *                 if isinstance(getattr(GraphicsState, name),
 *                               GetSetDescriptorType)]
 */
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_v_name))) __PYX_ERR(1, 134, __pyx_L1_error)

      /* "celiagg/graphics_state.pxi":135
 *     cdef list _propnames(self):
 *         return [name for name in GraphicsState.__dict__.keys()
 *                 if isinstance(getattr(GraphicsState, name),             # <<<<<<<<<<<<<<
 *                               GetSetDescriptorType)]
 * 
 */
    }

    /* "celiagg/graphics_state.pxi":134
 * 
 *     cdef list _propnames(self):
 *         return [name for name in GraphicsState.__dict__.keys()             # <<<<<<<<<<<<<<
 *                 if isinstance(getattr(GraphicsState, name),
 *                               GetSetDescriptorType)]
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":133
 *                 setattr(self, key, value)
 * 
 *     cdef list _propnames(self):             # <<<<<<<<<<<<<<
 *         return [name for name in GraphicsState.__dict__.keys()
 *                 if isinstance(getattr(GraphicsState, name),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState._propnames", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":138
 *                               GetSetDescriptorType)]
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Return a deep copy of the object
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_7copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_13GraphicsState_6copy[] = "Return a deep copy of the object\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_7copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_6copy(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_6copy(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_cpy = NULL;
  PyObject *__pyx_v_properties = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/graphics_state.pxi":141
 *         """Return a deep copy of the object
 *         """
 *         cpy = GraphicsState()             # <<<<<<<<<<<<<<
 *         properties = self._propnames()
 *         properties.remove("stencil")
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_GraphicsState)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cpy = ((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/graphics_state.pxi":142
 *         """
 *         cpy = GraphicsState()
 *         properties = self._propnames()             # <<<<<<<<<<<<<<
 *         properties.remove("stencil")
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self->__pyx_vtab)->_propnames(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_properties = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/graphics_state.pxi":143
 *         cpy = GraphicsState()
 *         properties = self._propnames()
 *         properties.remove("stencil")             # <<<<<<<<<<<<<<
 * 
 *         if self._stencil_img is not None:
 */
  __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyList_Type_remove, __pyx_v_properties, __pyx_n_s_stencil); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/graphics_state.pxi":145
 *         properties.remove("stencil")
 * 
 *         if self._stencil_img is not None:             # <<<<<<<<<<<<<<
 *             cpy.stencil = self._stencil_img.copy()
 * 
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_self->_stencil_img) != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "celiagg/graphics_state.pxi":146
 * 
 *         if self._stencil_img is not None:
 *             cpy.stencil = self._stencil_img.copy()             # <<<<<<<<<<<<<<
 * 
 *         for name in properties:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_stencil_img), __pyx_n_s_copy); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cpy), __pyx_n_s_stencil, __pyx_t_1) < 0) __PYX_ERR(1, 146, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "celiagg/graphics_state.pxi":145
 *         properties.remove("stencil")
 * 
 *         if self._stencil_img is not None:             # <<<<<<<<<<<<<<
 *             cpy.stencil = self._stencil_img.copy()
 * 
 */
  }

  /* "celiagg/graphics_state.pxi":148
 *             cpy.stencil = self._stencil_img.copy()
 * 
 *         for name in properties:             # <<<<<<<<<<<<<<
 *             value = getattr(self, name)
 *             setattr(cpy, name, value)
 */
  if (unlikely(__pyx_v_properties == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 148, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_properties; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
  for (;;) {
    if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(1, 148, __pyx_L1_error)
    #else
    __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "celiagg/graphics_state.pxi":149
 * 
 *         for name in properties:
 *             value = getattr(self, name)             # <<<<<<<<<<<<<<
 *             setattr(cpy, name, value)
 * 
 */
    __pyx_t_4 = __Pyx_GetAttr(((PyObject *)__pyx_v_self), __pyx_v_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "celiagg/graphics_state.pxi":150
 *         for name in properties:
 *             value = getattr(self, name)
 *             setattr(cpy, name, value)             # <<<<<<<<<<<<<<
 * 
 *         return cpy
 */
    __pyx_t_7 = PyObject_SetAttr(((PyObject *)__pyx_v_cpy), __pyx_v_name, __pyx_v_value); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 150, __pyx_L1_error)

    /* "celiagg/graphics_state.pxi":148
 *             cpy.stencil = self._stencil_img.copy()
 * 
 *         for name in properties:             # <<<<<<<<<<<<<<
 *             value = getattr(self, name)
 *             setattr(cpy, name, value)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/graphics_state.pxi":152
 *             setattr(cpy, name, value)
 * 
 *         return cpy             # <<<<<<<<<<<<<<
 * 
 *     property anti_aliased:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cpy));
  __pyx_r = ((PyObject *)__pyx_v_cpy);
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":138
 *                               GetSetDescriptorType)]
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Return a deep copy of the object
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cpy);
  __Pyx_XDECREF(__pyx_v_properties);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":155
 * 
 *     property anti_aliased:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.anti_aliased()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_12anti_aliased_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_12anti_aliased_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12anti_aliased___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_12anti_aliased___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":156
 *     property anti_aliased:
 *         def __get__(self):
 *             return self._this.anti_aliased()             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, aa):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_this->anti_aliased()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":155
 * 
 *     property anti_aliased:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.anti_aliased()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.anti_aliased.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":158
 *             return self._this.anti_aliased()
 * 
 *         def __set__(self, aa):             # <<<<<<<<<<<<<<
 *             self._this.anti_aliased(aa)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12anti_aliased_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_aa); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12anti_aliased_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_aa) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12anti_aliased_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((PyObject *)__pyx_v_aa));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12anti_aliased_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_aa) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":159
 * 
 *         def __set__(self, aa):
 *             self._this.anti_aliased(aa)             # <<<<<<<<<<<<<<
 * 
 *     property clip_box:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_aa); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(1, 159, __pyx_L1_error)
  __pyx_v_self->_this->anti_aliased(__pyx_t_1);

  /* "celiagg/graphics_state.pxi":158
 *             return self._this.anti_aliased()
 * 
 *         def __set__(self, aa):             # <<<<<<<<<<<<<<
 *             self._this.anti_aliased(aa)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.anti_aliased.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":162
 * 
 *     property clip_box:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef _graphics_state.Rect ret = self._this.clip_box()
 *             return Rect(ret.x1, ret.y1, ret.x2 - ret.x1, ret.y2 - ret.y1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_8clip_box_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_8clip_box_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_8clip_box___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_8clip_box___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  GraphicsState::Rect __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":163
 *     property clip_box:
 *         def __get__(self):
 *             cdef _graphics_state.Rect ret = self._this.clip_box()             # <<<<<<<<<<<<<<
 *             return Rect(ret.x1, ret.y1, ret.x2 - ret.x1, ret.y2 - ret.y1)
 * 
 */
  __pyx_v_ret = __pyx_v_self->_this->clip_box();

  /* "celiagg/graphics_state.pxi":164
 *         def __get__(self):
 *             cdef _graphics_state.Rect ret = self._this.clip_box()
 *             return Rect(ret.x1, ret.y1, ret.x2 - ret.x1, ret.y2 - ret.y1)             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, Rect box):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_ret.x1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_ret.y1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble((__pyx_v_ret.x2 - __pyx_v_ret.x1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_ret.y2 - __pyx_v_ret.y1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_Rect), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":162
 * 
 *     property clip_box:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef _graphics_state.Rect ret = self._this.clip_box()
 *             return Rect(ret.x1, ret.y1, ret.x2 - ret.x1, ret.y2 - ret.y1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.clip_box.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":166
 *             return Rect(ret.x1, ret.y1, ret.x2 - ret.x1, ret.y2 - ret.y1)
 * 
 *         def __set__(self, Rect box):             # <<<<<<<<<<<<<<
 *             if not isinstance(box, Rect):
 *                 raise TypeError("The clip_box property must be a Rect")
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_8clip_box_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_box); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_8clip_box_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_box) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_box), __pyx_ptype_7celiagg_8_celiagg_Rect, 1, "box", 0))) __PYX_ERR(1, 166, __pyx_L1_error)
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_8clip_box_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_v_box));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_8clip_box_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_box) {
  struct __pyx_obj_7celiagg_8_celiagg_Rect *__pyx_v_rect = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":167
 * 
 *         def __set__(self, Rect box):
 *             if not isinstance(box, Rect):             # <<<<<<<<<<<<<<
 *                 raise TypeError("The clip_box property must be a Rect")
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_box), __pyx_ptype_7celiagg_8_celiagg_Rect); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/graphics_state.pxi":168
 *         def __set__(self, Rect box):
 *             if not isinstance(box, Rect):
 *                 raise TypeError("The clip_box property must be a Rect")             # <<<<<<<<<<<<<<
 * 
 *             cdef Rect rect = <Rect>box
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 168, __pyx_L1_error)

    /* "celiagg/graphics_state.pxi":167
 * 
 *         def __set__(self, Rect box):
 *             if not isinstance(box, Rect):             # <<<<<<<<<<<<<<
 *                 raise TypeError("The clip_box property must be a Rect")
 * 
 */
  }

  /* "celiagg/graphics_state.pxi":170
 *                 raise TypeError("The clip_box property must be a Rect")
 * 
 *             cdef Rect rect = <Rect>box             # <<<<<<<<<<<<<<
 *             self._this.clip_box(dereference(rect._this))
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_box);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_rect = ((struct __pyx_obj_7celiagg_8_celiagg_Rect *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/graphics_state.pxi":171
 * 
 *             cdef Rect rect = <Rect>box
 *             self._this.clip_box(dereference(rect._this))             # <<<<<<<<<<<<<<
 * 
 *     property drawing_mode:
 */
  __pyx_v_self->_this->clip_box((*__pyx_v_rect->_this));

  /* "celiagg/graphics_state.pxi":166
 *             return Rect(ret.x1, ret.y1, ret.x2 - ret.x1, ret.y2 - ret.y1)
 * 
 *         def __set__(self, Rect box):             # <<<<<<<<<<<<<<
 *             if not isinstance(box, Rect):
 *                 raise TypeError("The clip_box property must be a Rect")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.clip_box.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_rect);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":174
 * 
 *     property drawing_mode:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return DrawingMode(self._this.drawing_mode())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_12drawing_mode_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_12drawing_mode_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12drawing_mode___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_12drawing_mode___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":175
 *     property drawing_mode:
 *         def __get__(self):
 *             return DrawingMode(self._this.drawing_mode())             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, DrawingMode m):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum__GraphicsState_3a__3a_DrawingMode(__pyx_v_self->_this->drawing_mode()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":174
 * 
 *     property drawing_mode:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return DrawingMode(self._this.drawing_mode())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.drawing_mode.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":177
 *             return DrawingMode(self._this.drawing_mode())
 * 
 *         def __set__(self, DrawingMode m):             # <<<<<<<<<<<<<<
 *             self._this.drawing_mode(m)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12drawing_mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_m); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12drawing_mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_m) {
  enum __pyx_t_7celiagg_8_celiagg_DrawingMode __pyx_v_m;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_m); {
    __pyx_v_m = ((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_arg_m)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 177, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.drawing_mode.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12drawing_mode_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)__pyx_v_m));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12drawing_mode_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_DrawingMode __pyx_v_m) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":178
 * 
 *         def __set__(self, DrawingMode m):
 *             self._this.drawing_mode(m)             # <<<<<<<<<<<<<<
 * 
 *     property text_drawing_mode:
 */
  __pyx_v_self->_this->drawing_mode(((enum GraphicsState::DrawingMode)__pyx_v_m));

  /* "celiagg/graphics_state.pxi":177
 *             return DrawingMode(self._this.drawing_mode())
 * 
 *         def __set__(self, DrawingMode m):             # <<<<<<<<<<<<<<
 *             self._this.drawing_mode(m)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":181
 * 
 *     property text_drawing_mode:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return TextDrawingMode(self._this.text_drawing_mode())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":182
 *     property text_drawing_mode:
 *         def __get__(self):
 *             return TextDrawingMode(self._this.text_drawing_mode())             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, TextDrawingMode m):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum__GraphicsState_3a__3a_TextDrawingMode(__pyx_v_self->_this->text_drawing_mode()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":181
 * 
 *     property text_drawing_mode:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return TextDrawingMode(self._this.text_drawing_mode())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.text_drawing_mode.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":184
 *             return TextDrawingMode(self._this.text_drawing_mode())
 * 
 *         def __set__(self, TextDrawingMode m):             # <<<<<<<<<<<<<<
 *             self._this.text_drawing_mode(m)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_m); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_m) {
  enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode __pyx_v_m;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_m); {
    __pyx_v_m = ((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_arg_m)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 184, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.text_drawing_mode.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)__pyx_v_m));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode __pyx_v_m) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":185
 * 
 *         def __set__(self, TextDrawingMode m):
 *             self._this.text_drawing_mode(m)             # <<<<<<<<<<<<<<
 * 
 *     property blend_mode:
 */
  __pyx_v_self->_this->text_drawing_mode(((enum GraphicsState::TextDrawingMode)__pyx_v_m));

  /* "celiagg/graphics_state.pxi":184
 *             return TextDrawingMode(self._this.text_drawing_mode())
 * 
 *         def __set__(self, TextDrawingMode m):             # <<<<<<<<<<<<<<
 *             self._this.text_drawing_mode(m)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":188
 * 
 *     property blend_mode:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return BlendMode(self._this.blend_mode())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_10blend_mode_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_10blend_mode_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10blend_mode___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_10blend_mode___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":189
 *     property blend_mode:
 *         def __get__(self):
 *             return BlendMode(self._this.blend_mode())             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, BlendMode m):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum__GraphicsState_3a__3a_BlendMode(__pyx_v_self->_this->blend_mode()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":188
 * 
 *     property blend_mode:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return BlendMode(self._this.blend_mode())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.blend_mode.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":191
 *             return BlendMode(self._this.blend_mode())
 * 
 *         def __set__(self, BlendMode m):             # <<<<<<<<<<<<<<
 *             self._this.blend_mode(m)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10blend_mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_m); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10blend_mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_m) {
  enum __pyx_t_7celiagg_8_celiagg_BlendMode __pyx_v_m;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_m); {
    __pyx_v_m = ((enum __pyx_t_7celiagg_8_celiagg_BlendMode)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_arg_m)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 191, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.blend_mode.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10blend_mode_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((enum __pyx_t_7celiagg_8_celiagg_BlendMode)__pyx_v_m));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10blend_mode_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_BlendMode __pyx_v_m) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":192
 * 
 *         def __set__(self, BlendMode m):
 *             self._this.blend_mode(m)             # <<<<<<<<<<<<<<
 * 
 *     property image_blend_mode:
 */
  __pyx_v_self->_this->blend_mode(((enum GraphicsState::BlendMode)__pyx_v_m));

  /* "celiagg/graphics_state.pxi":191
 *             return BlendMode(self._this.blend_mode())
 * 
 *         def __set__(self, BlendMode m):             # <<<<<<<<<<<<<<
 *             self._this.blend_mode(m)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":195
 * 
 *     property image_blend_mode:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return BlendMode(self._this.image_blend_mode())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":196
 *     property image_blend_mode:
 *         def __get__(self):
 *             return BlendMode(self._this.image_blend_mode())             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, BlendMode m):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum__GraphicsState_3a__3a_BlendMode(__pyx_v_self->_this->image_blend_mode()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":195
 * 
 *     property image_blend_mode:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return BlendMode(self._this.image_blend_mode())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.image_blend_mode.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":198
 *             return BlendMode(self._this.image_blend_mode())
 * 
 *         def __set__(self, BlendMode m):             # <<<<<<<<<<<<<<
 *             self._this.image_blend_mode(m)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_m); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_m) {
  enum __pyx_t_7celiagg_8_celiagg_BlendMode __pyx_v_m;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_m); {
    __pyx_v_m = ((enum __pyx_t_7celiagg_8_celiagg_BlendMode)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_arg_m)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 198, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.image_blend_mode.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((enum __pyx_t_7celiagg_8_celiagg_BlendMode)__pyx_v_m));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_BlendMode __pyx_v_m) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":199
 * 
 *         def __set__(self, BlendMode m):
 *             self._this.image_blend_mode(m)             # <<<<<<<<<<<<<<
 * 
 *     property master_alpha:
 */
  __pyx_v_self->_this->image_blend_mode(((enum GraphicsState::BlendMode)__pyx_v_m));

  /* "celiagg/graphics_state.pxi":198
 *             return BlendMode(self._this.image_blend_mode())
 * 
 *         def __set__(self, BlendMode m):             # <<<<<<<<<<<<<<
 *             self._this.image_blend_mode(m)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":202
 * 
 *     property master_alpha:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.master_alpha()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_12master_alpha_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_12master_alpha_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12master_alpha___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_12master_alpha___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":203
 *     property master_alpha:
 *         def __get__(self):
 *             return self._this.master_alpha()             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, a):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->master_alpha()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":202
 * 
 *     property master_alpha:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.master_alpha()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.master_alpha.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":205
 *             return self._this.master_alpha()
 * 
 *         def __set__(self, a):             # <<<<<<<<<<<<<<
 *             self._this.master_alpha(a)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12master_alpha_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_a); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12master_alpha_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_a) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12master_alpha_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((PyObject *)__pyx_v_a));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_12master_alpha_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_a) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":206
 * 
 *         def __set__(self, a):
 *             self._this.master_alpha(a)             # <<<<<<<<<<<<<<
 * 
 *     property line_width:
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_a); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 206, __pyx_L1_error)
  __pyx_v_self->_this->master_alpha(__pyx_t_1);

  /* "celiagg/graphics_state.pxi":205
 *             return self._this.master_alpha()
 * 
 *         def __set__(self, a):             # <<<<<<<<<<<<<<
 *             self._this.master_alpha(a)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.master_alpha.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":209
 * 
 *     property line_width:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.line_width()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_10line_width_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_10line_width_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10line_width___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_10line_width___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":210
 *     property line_width:
 *         def __get__(self):
 *             return self._this.line_width()             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, w):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->line_width()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":209
 * 
 *     property line_width:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.line_width()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_width.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":212
 *             return self._this.line_width()
 * 
 *         def __set__(self, w):             # <<<<<<<<<<<<<<
 *             self._this.line_width(w)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10line_width_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_w); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10line_width_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_w) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10line_width_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((PyObject *)__pyx_v_w));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10line_width_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_w) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":213
 * 
 *         def __set__(self, w):
 *             self._this.line_width(w)             # <<<<<<<<<<<<<<
 * 
 *     property line_cap:
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_w); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 213, __pyx_L1_error)
  __pyx_v_self->_this->line_width(__pyx_t_1);

  /* "celiagg/graphics_state.pxi":212
 *             return self._this.line_width()
 * 
 *         def __set__(self, w):             # <<<<<<<<<<<<<<
 *             self._this.line_width(w)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_width.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":216
 * 
 *     property line_cap:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return LineCap(self._this.line_cap())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_8line_cap_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_8line_cap_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_8line_cap___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_8line_cap___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":217
 *     property line_cap:
 *         def __get__(self):
 *             return LineCap(self._this.line_cap())             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, LineCap cap):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LineCap); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum__GraphicsState_3a__3a_LineCap(__pyx_v_self->_this->line_cap()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":216
 * 
 *     property line_cap:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return LineCap(self._this.line_cap())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_cap.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":219
 *             return LineCap(self._this.line_cap())
 * 
 *         def __set__(self, LineCap cap):             # <<<<<<<<<<<<<<
 *             self._this.line_cap(cap)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_8line_cap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_cap); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_8line_cap_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_cap) {
  enum __pyx_t_7celiagg_8_celiagg_LineCap __pyx_v_cap;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_cap); {
    __pyx_v_cap = ((enum __pyx_t_7celiagg_8_celiagg_LineCap)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_LineCap(__pyx_arg_cap)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 219, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_cap.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_8line_cap_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((enum __pyx_t_7celiagg_8_celiagg_LineCap)__pyx_v_cap));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_8line_cap_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_LineCap __pyx_v_cap) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":220
 * 
 *         def __set__(self, LineCap cap):
 *             self._this.line_cap(cap)             # <<<<<<<<<<<<<<
 * 
 *     property line_join:
 */
  __pyx_v_self->_this->line_cap(((enum GraphicsState::LineCap)__pyx_v_cap));

  /* "celiagg/graphics_state.pxi":219
 *             return LineCap(self._this.line_cap())
 * 
 *         def __set__(self, LineCap cap):             # <<<<<<<<<<<<<<
 *             self._this.line_cap(cap)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":223
 * 
 *     property line_join:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return LineJoin(self._this.line_join())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_9line_join_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_9line_join_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_9line_join___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_9line_join___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":224
 *     property line_join:
 *         def __get__(self):
 *             return LineJoin(self._this.line_join())             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, LineJoin join):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LineJoin); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum__GraphicsState_3a__3a_LineJoin(__pyx_v_self->_this->line_join()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":223
 * 
 *     property line_join:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return LineJoin(self._this.line_join())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_join.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":226
 *             return LineJoin(self._this.line_join())
 * 
 *         def __set__(self, LineJoin join):             # <<<<<<<<<<<<<<
 *             self._this.line_join(join)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_9line_join_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_join); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_9line_join_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_join) {
  enum __pyx_t_7celiagg_8_celiagg_LineJoin __pyx_v_join;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_join); {
    __pyx_v_join = ((enum __pyx_t_7celiagg_8_celiagg_LineJoin)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_LineJoin(__pyx_arg_join)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 226, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_join.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_9line_join_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((enum __pyx_t_7celiagg_8_celiagg_LineJoin)__pyx_v_join));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_9line_join_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_LineJoin __pyx_v_join) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":227
 * 
 *         def __set__(self, LineJoin join):
 *             self._this.line_join(join)             # <<<<<<<<<<<<<<
 * 
 *     property inner_join:
 */
  __pyx_v_self->_this->line_join(((enum GraphicsState::LineJoin)__pyx_v_join));

  /* "celiagg/graphics_state.pxi":226
 *             return LineJoin(self._this.line_join())
 * 
 *         def __set__(self, LineJoin join):             # <<<<<<<<<<<<<<
 *             self._this.line_join(join)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":230
 * 
 *     property inner_join:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return InnerJoin(self._this.inner_join())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_10inner_join_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_10inner_join_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10inner_join___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_10inner_join___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":231
 *     property inner_join:
 *         def __get__(self):
 *             return InnerJoin(self._this.inner_join())             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, InnerJoin join):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_InnerJoin); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum__GraphicsState_3a__3a_InnerJoin(__pyx_v_self->_this->inner_join()); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":230
 * 
 *     property inner_join:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return InnerJoin(self._this.inner_join())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.inner_join.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":233
 *             return InnerJoin(self._this.inner_join())
 * 
 *         def __set__(self, InnerJoin join):             # <<<<<<<<<<<<<<
 *             self._this.inner_join(join)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10inner_join_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_join); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10inner_join_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_join) {
  enum __pyx_t_7celiagg_8_celiagg_InnerJoin __pyx_v_join;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_join); {
    __pyx_v_join = ((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(__pyx_arg_join)); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 233, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.inner_join.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10inner_join_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)__pyx_v_join));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10inner_join_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_InnerJoin __pyx_v_join) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":234
 * 
 *         def __set__(self, InnerJoin join):
 *             self._this.inner_join(join)             # <<<<<<<<<<<<<<
 * 
 *     property miter_limit:
 */
  __pyx_v_self->_this->inner_join(((enum GraphicsState::InnerJoin)__pyx_v_join));

  /* "celiagg/graphics_state.pxi":233
 *             return InnerJoin(self._this.inner_join())
 * 
 *         def __set__(self, InnerJoin join):             # <<<<<<<<<<<<<<
 *             self._this.inner_join(join)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":237
 * 
 *     property miter_limit:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.miter_limit()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_11miter_limit_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_11miter_limit_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_11miter_limit___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_11miter_limit___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":238
 *     property miter_limit:
 *         def __get__(self):
 *             return self._this.miter_limit()             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, double limit):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->miter_limit()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":237
 * 
 *     property miter_limit:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.miter_limit()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.miter_limit.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":240
 *             return self._this.miter_limit()
 * 
 *         def __set__(self, double limit):             # <<<<<<<<<<<<<<
 *             self._this.miter_limit(limit)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_11miter_limit_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_limit); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_11miter_limit_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_limit) {
  double __pyx_v_limit;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_limit); {
    __pyx_v_limit = __pyx_PyFloat_AsDouble(__pyx_arg_limit); if (unlikely((__pyx_v_limit == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 240, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.miter_limit.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_11miter_limit_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((double)__pyx_v_limit));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_11miter_limit_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, double __pyx_v_limit) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":241
 * 
 *         def __set__(self, double limit):
 *             self._this.miter_limit(limit)             # <<<<<<<<<<<<<<
 * 
 *     property inner_miter_limit:
 */
  __pyx_v_self->_this->miter_limit(__pyx_v_limit);

  /* "celiagg/graphics_state.pxi":240
 *             return self._this.miter_limit()
 * 
 *         def __set__(self, double limit):             # <<<<<<<<<<<<<<
 *             self._this.miter_limit(limit)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":244
 * 
 *     property inner_miter_limit:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.inner_miter_limit()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":245
 *     property inner_miter_limit:
 *         def __get__(self):
 *             return self._this.inner_miter_limit()             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, double limit):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->inner_miter_limit()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":244
 * 
 *     property inner_miter_limit:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.inner_miter_limit()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.inner_miter_limit.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":247
 *             return self._this.inner_miter_limit()
 * 
 *         def __set__(self, double limit):             # <<<<<<<<<<<<<<
 *             self._this.inner_miter_limit(limit)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_limit); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_limit) {
  double __pyx_v_limit;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_limit); {
    __pyx_v_limit = __pyx_PyFloat_AsDouble(__pyx_arg_limit); if (unlikely((__pyx_v_limit == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 247, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.inner_miter_limit.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((double)__pyx_v_limit));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, double __pyx_v_limit) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":248
 * 
 *         def __set__(self, double limit):
 *             self._this.inner_miter_limit(limit)             # <<<<<<<<<<<<<<
 * 
 *     property line_dash_pattern:
 */
  __pyx_v_self->_this->inner_miter_limit(__pyx_v_limit);

  /* "celiagg/graphics_state.pxi":247
 *             return self._this.inner_miter_limit()
 * 
 *         def __set__(self, double limit):             # <<<<<<<<<<<<<<
 *             self._this.inner_miter_limit(limit)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":251
 * 
 *     property line_dash_pattern:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef vector[double] dashes = self._this.line_dash_pattern()
 *             return list(zip(dashes[:-1:2], dashes[1::2]))
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  std::vector<double>  __pyx_v_dashes;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":252
 *     property line_dash_pattern:
 *         def __get__(self):
 *             cdef vector[double] dashes = self._this.line_dash_pattern()             # <<<<<<<<<<<<<<
 *             return list(zip(dashes[:-1:2], dashes[1::2]))
 * 
 */
  __pyx_v_dashes = __pyx_v_self->_this->line_dash_pattern();

  /* "celiagg/graphics_state.pxi":253
 *         def __get__(self):
 *             cdef vector[double] dashes = self._this.line_dash_pattern()
 *             return list(zip(dashes[:-1:2], dashes[1::2]))             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, dashes):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert_vector_to_py_double(__pyx_v_dashes); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_slice__7); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_convert_vector_to_py_double(__pyx_v_dashes); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_slice__8); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PySequence_List(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":251
 * 
 *     property line_dash_pattern:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef vector[double] dashes = self._this.line_dash_pattern()
 *             return list(zip(dashes[:-1:2], dashes[1::2]))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_dash_pattern.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":255
 *             return list(zip(dashes[:-1:2], dashes[1::2]))
 * 
 *         def __set__(self, dashes):             # <<<<<<<<<<<<<<
 *             if len(dashes) == 0:
 *                 self._this.line_dash_pattern(<const double *>0, 0)
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_dashes); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_dashes) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((PyObject *)__pyx_v_dashes));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_dashes) {
  __Pyx_memviewslice __pyx_v__dashes = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_msg = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":256
 * 
 *         def __set__(self, dashes):
 *             if len(dashes) == 0:             # <<<<<<<<<<<<<<
 *                 self._this.line_dash_pattern(<const double *>0, 0)
 *                 return
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_dashes); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 256, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_1 == 0) != 0);
  if (__pyx_t_2) {

    /* "celiagg/graphics_state.pxi":257
 *         def __set__(self, dashes):
 *             if len(dashes) == 0:
 *                 self._this.line_dash_pattern(<const double *>0, 0)             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
    __pyx_v_self->_this->line_dash_pattern(((double const *)0), 0);

    /* "celiagg/graphics_state.pxi":258
 *             if len(dashes) == 0:
 *                 self._this.line_dash_pattern(<const double *>0, 0)
 *                 return             # <<<<<<<<<<<<<<
 * 
 *             cdef double[:,::1] _dashes = numpy.asarray(dashes,
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "celiagg/graphics_state.pxi":256
 * 
 *         def __set__(self, dashes):
 *             if len(dashes) == 0:             # <<<<<<<<<<<<<<
 *                 self._this.line_dash_pattern(<const double *>0, 0)
 *                 return
 */
  }

  /* "celiagg/graphics_state.pxi":260
 *                 return
 * 
 *             cdef double[:,::1] _dashes = numpy.asarray(dashes,             # <<<<<<<<<<<<<<
 *                                                        dtype=numpy.float64,
 *                                                        order='c')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_dashes);
  __Pyx_GIVEREF(__pyx_v_dashes);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_dashes);

  /* "celiagg/graphics_state.pxi":261
 * 
 *             cdef double[:,::1] _dashes = numpy.asarray(dashes,
 *                                                        dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                                        order='c')
 *             if _dashes.shape[1] != 2:
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_float64); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(1, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(1, 261, __pyx_L1_error)

  /* "celiagg/graphics_state.pxi":260
 *                 return
 * 
 *             cdef double[:,::1] _dashes = numpy.asarray(dashes,             # <<<<<<<<<<<<<<
 *                                                        dtype=numpy.float64,
 *                                                        order='c')
 */
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v__dashes = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "celiagg/graphics_state.pxi":263
 *                                                        dtype=numpy.float64,
 *                                                        order='c')
 *             if _dashes.shape[1] != 2:             # <<<<<<<<<<<<<<
 *                 msg = 'Dashes must be an iterable of (dash len, gap len) pairs.'
 *                 raise ValueError(msg)
 */
  __pyx_t_2 = (((__pyx_v__dashes.shape[1]) != 2) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/graphics_state.pxi":264
 *                                                        order='c')
 *             if _dashes.shape[1] != 2:
 *                 msg = 'Dashes must be an iterable of (dash len, gap len) pairs.'             # <<<<<<<<<<<<<<
 *                 raise ValueError(msg)
 *             self._this.line_dash_pattern(&_dashes[0][0], _dashes.shape[0])
 */
    __Pyx_INCREF(__pyx_kp_s_Dashes_must_be_an_iterable_of_da);
    __pyx_v_msg = __pyx_kp_s_Dashes_must_be_an_iterable_of_da;

    /* "celiagg/graphics_state.pxi":265
 *             if _dashes.shape[1] != 2:
 *                 msg = 'Dashes must be an iterable of (dash len, gap len) pairs.'
 *                 raise ValueError(msg)             # <<<<<<<<<<<<<<
 *             self._this.line_dash_pattern(&_dashes[0][0], _dashes.shape[0])
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(1, 265, __pyx_L1_error)

    /* "celiagg/graphics_state.pxi":263
 *                                                        dtype=numpy.float64,
 *                                                        order='c')
 *             if _dashes.shape[1] != 2:             # <<<<<<<<<<<<<<
 *                 msg = 'Dashes must be an iterable of (dash len, gap len) pairs.'
 *                 raise ValueError(msg)
 */
  }

  /* "celiagg/graphics_state.pxi":266
 *                 msg = 'Dashes must be an iterable of (dash len, gap len) pairs.'
 *                 raise ValueError(msg)
 *             self._this.line_dash_pattern(&_dashes[0][0], _dashes.shape[0])             # <<<<<<<<<<<<<<
 * 
 *     property line_dash_phase:
 */
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = -1;
  if (__pyx_t_9 < 0) {
    __pyx_t_9 += __pyx_v__dashes.shape[0];
    if (unlikely(__pyx_t_9 < 0)) __pyx_t_11 = 0;
  } else if (unlikely(__pyx_t_9 >= __pyx_v__dashes.shape[0])) __pyx_t_11 = 0;
  if (__pyx_t_10 < 0) {
    __pyx_t_10 += __pyx_v__dashes.shape[1];
    if (unlikely(__pyx_t_10 < 0)) __pyx_t_11 = 1;
  } else if (unlikely(__pyx_t_10 >= __pyx_v__dashes.shape[1])) __pyx_t_11 = 1;
  if (unlikely(__pyx_t_11 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_11);
    __PYX_ERR(1, 266, __pyx_L1_error)
  }
  __pyx_v_self->_this->line_dash_pattern((&(*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__dashes.data + __pyx_t_9 * __pyx_v__dashes.strides[0]) )) + __pyx_t_10)) )))), (__pyx_v__dashes.shape[0]));

  /* "celiagg/graphics_state.pxi":255
 *             return list(zip(dashes[:-1:2], dashes[1::2]))
 * 
 *         def __set__(self, dashes):             # <<<<<<<<<<<<<<
 *             if len(dashes) == 0:
 *                 self._this.line_dash_pattern(<const double *>0, 0)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_dash_pattern.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v__dashes, 1);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":269
 * 
 *     property line_dash_phase:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.line_dash_phase()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":270
 *     property line_dash_phase:
 *         def __get__(self):
 *             return self._this.line_dash_phase()             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, double phase):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->line_dash_phase()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":269
 * 
 *     property line_dash_phase:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.line_dash_phase()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_dash_phase.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":272
 *             return self._this.line_dash_phase()
 * 
 *         def __set__(self, double phase):             # <<<<<<<<<<<<<<
 *             self._this.line_dash_phase(phase)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_phase); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_phase) {
  double __pyx_v_phase;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_phase); {
    __pyx_v_phase = __pyx_PyFloat_AsDouble(__pyx_arg_phase); if (unlikely((__pyx_v_phase == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 272, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.line_dash_phase.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((double)__pyx_v_phase));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, double __pyx_v_phase) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":273
 * 
 *         def __set__(self, double phase):
 *             self._this.line_dash_phase(phase)             # <<<<<<<<<<<<<<
 * 
 *     property stencil:
 */
  __pyx_v_self->_this->line_dash_phase(__pyx_v_phase);

  /* "celiagg/graphics_state.pxi":272
 *             return self._this.line_dash_phase()
 * 
 *         def __set__(self, double phase):             # <<<<<<<<<<<<<<
 *             self._this.line_dash_phase(phase)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":276
 * 
 *     property stencil:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._stencil_img
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_7stencil_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_7stencil_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_7stencil___get__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_7stencil___get__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/graphics_state.pxi":277
 *     property stencil:
 *         def __get__(self):
 *             return self._stencil_img             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, img):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_stencil_img));
  __pyx_r = ((PyObject *)__pyx_v_self->_stencil_img);
  goto __pyx_L0;

  /* "celiagg/graphics_state.pxi":276
 * 
 *     property stencil:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._stencil_img
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/graphics_state.pxi":279
 *             return self._stencil_img
 * 
 *         def __set__(self, img):             # <<<<<<<<<<<<<<
 *             if img is not None and not isinstance(img, Image):
 *                 raise TypeError("The stencil property must be an Image")
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_7stencil_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_img); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13GraphicsState_7stencil_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_img) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_7stencil_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((PyObject *)__pyx_v_img));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13GraphicsState_7stencil_2__set__(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, PyObject *__pyx_v_img) {
  struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_image = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/graphics_state.pxi":280
 * 
 *         def __set__(self, img):
 *             if img is not None and not isinstance(img, Image):             # <<<<<<<<<<<<<<
 *                 raise TypeError("The stencil property must be an Image")
 * 
 */
  __pyx_t_2 = (__pyx_v_img != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_img, __pyx_ptype_7celiagg_8_celiagg_Image); 
  __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "celiagg/graphics_state.pxi":281
 *         def __set__(self, img):
 *             if img is not None and not isinstance(img, Image):
 *                 raise TypeError("The stencil property must be an Image")             # <<<<<<<<<<<<<<
 * 
 *             cdef Image image = <Image>img
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 281, __pyx_L1_error)

    /* "celiagg/graphics_state.pxi":280
 * 
 *         def __set__(self, img):
 *             if img is not None and not isinstance(img, Image):             # <<<<<<<<<<<<<<
 *                 raise TypeError("The stencil property must be an Image")
 * 
 */
  }

  /* "celiagg/graphics_state.pxi":283
 *                 raise TypeError("The stencil property must be an Image")
 * 
 *             cdef Image image = <Image>img             # <<<<<<<<<<<<<<
 *             self._stencil_img = image
 *             if self._stencil_img is not None:
 */
  __pyx_t_4 = __pyx_v_img;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_image = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "celiagg/graphics_state.pxi":284
 * 
 *             cdef Image image = <Image>img
 *             self._stencil_img = image             # <<<<<<<<<<<<<<
 *             if self._stencil_img is not None:
 *                 if image.format != PixelFormat.Gray8:
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_image));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_image));
  __Pyx_GOTREF(__pyx_v_self->_stencil_img);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_stencil_img));
  __pyx_v_self->_stencil_img = __pyx_v_image;

  /* "celiagg/graphics_state.pxi":285
 *             cdef Image image = <Image>img
 *             self._stencil_img = image
 *             if self._stencil_img is not None:             # <<<<<<<<<<<<<<
 *                 if image.format != PixelFormat.Gray8:
 *                     raise ValueError("Stencil Images must have format Gray8!")
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->_stencil_img) != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "celiagg/graphics_state.pxi":286
 *             self._stencil_img = image
 *             if self._stencil_img is not None:
 *                 if image.format != PixelFormat.Gray8:             # <<<<<<<<<<<<<<
 *                     raise ValueError("Stencil Images must have format Gray8!")
 *                 self._this.stencil(image._this)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_image), __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray8); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_4, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_t_2)) {

      /* "celiagg/graphics_state.pxi":287
 *             if self._stencil_img is not None:
 *                 if image.format != PixelFormat.Gray8:
 *                     raise ValueError("Stencil Images must have format Gray8!")             # <<<<<<<<<<<<<<
 *                 self._this.stencil(image._this)
 *             else:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 287, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 287, __pyx_L1_error)

      /* "celiagg/graphics_state.pxi":286
 *             self._stencil_img = image
 *             if self._stencil_img is not None:
 *                 if image.format != PixelFormat.Gray8:             # <<<<<<<<<<<<<<
 *                     raise ValueError("Stencil Images must have format Gray8!")
 *                 self._this.stencil(image._this)
 */
    }

    /* "celiagg/graphics_state.pxi":288
 *                 if image.format != PixelFormat.Gray8:
 *                     raise ValueError("Stencil Images must have format Gray8!")
 *                 self._this.stencil(image._this)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._this.stencil(<_image.Image*>0)
 */
    __pyx_v_self->_this->stencil(__pyx_v_image->_this);

    /* "celiagg/graphics_state.pxi":285
 *             cdef Image image = <Image>img
 *             self._stencil_img = image
 *             if self._stencil_img is not None:             # <<<<<<<<<<<<<<
 *                 if image.format != PixelFormat.Gray8:
 *                     raise ValueError("Stencil Images must have format Gray8!")
 */
    goto __pyx_L6;
  }

  /* "celiagg/graphics_state.pxi":290
 *                 self._this.stencil(image._this)
 *             else:
 *                 self._this.stencil(<_image.Image*>0)             # <<<<<<<<<<<<<<
 */
  /*else*/ {
    __pyx_v_self->_this->stencil(((Image *)0));
  }
  __pyx_L6:;

  /* "celiagg/graphics_state.pxi":279
 *             return self._stencil_img
 * 
 *         def __set__(self, img):             # <<<<<<<<<<<<<<
 *             if img is not None and not isinstance(img, Image):
 *                 raise TypeError("The stencil property must be an Image")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.stencil.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_image);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_8__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13GraphicsState_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13GraphicsState_10__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13GraphicsState_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.GraphicsState.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":29
 * ctypedef _image.Image* img_ptr_t
 * 
 * cdef _get_format_dtype(PixelFormat pixel_format):             # <<<<<<<<<<<<<<
 *     dtypes = {
 *         numpy.uint8: (PixelFormat.Gray8, PixelFormat.BGR24,
 */

static PyObject *__pyx_f_7celiagg_8_celiagg__get_format_dtype(enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_pixel_format) {
  PyObject *__pyx_v_dtypes = NULL;
  PyObject *__pyx_v_format_dtypes = NULL;
  PyObject *__pyx_7genexpr__pyx_v_dt = NULL;
  PyObject *__pyx_7genexpr__pyx_v_fmts = NULL;
  PyObject *__pyx_7genexpr__pyx_v_fmt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *(*__pyx_t_13)(PyObject *);
  Py_ssize_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  __Pyx_RefNannySetupContext("_get_format_dtype", 0);

  /* "celiagg/image.pxi":31
 * cdef _get_format_dtype(PixelFormat pixel_format):
 *     dtypes = {
 *         numpy.uint8: (PixelFormat.Gray8, PixelFormat.BGR24,             # <<<<<<<<<<<<<<
 *                       PixelFormat.RGB24, PixelFormat.BGRA32,
 *                       PixelFormat.RGBA32, PixelFormat.ARGB32,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_uint8); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray8); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR24); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "celiagg/image.pxi":32
 *     dtypes = {
 *         numpy.uint8: (PixelFormat.Gray8, PixelFormat.BGR24,
 *                       PixelFormat.RGB24, PixelFormat.BGRA32,             # <<<<<<<<<<<<<<
 *                       PixelFormat.RGBA32, PixelFormat.ARGB32,
 *                       PixelFormat.ABGR32),
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB24); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA32); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "celiagg/image.pxi":33
 *         numpy.uint8: (PixelFormat.Gray8, PixelFormat.BGR24,
 *                       PixelFormat.RGB24, PixelFormat.BGRA32,
 *                       PixelFormat.RGBA32, PixelFormat.ARGB32,             # <<<<<<<<<<<<<<
 *                       PixelFormat.ABGR32),
 *         numpy.uint16: (PixelFormat.Gray16, PixelFormat.BGR48,
 */
  __pyx_t_7 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA32); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB32); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "celiagg/image.pxi":34
 *                       PixelFormat.RGB24, PixelFormat.BGRA32,
 *                       PixelFormat.RGBA32, PixelFormat.ARGB32,
 *                       PixelFormat.ABGR32),             # <<<<<<<<<<<<<<
 *         numpy.uint16: (PixelFormat.Gray16, PixelFormat.BGR48,
 *                        PixelFormat.RGB48, PixelFormat.BGRA64,
 */
  __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR32); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "celiagg/image.pxi":31
 * cdef _get_format_dtype(PixelFormat pixel_format):
 *     dtypes = {
 *         numpy.uint8: (PixelFormat.Gray8, PixelFormat.BGR24,             # <<<<<<<<<<<<<<
 *                       PixelFormat.RGB24, PixelFormat.BGRA32,
 *                       PixelFormat.RGBA32, PixelFormat.ARGB32,
 */
  __pyx_t_10 = PyTuple_New(7); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_t_9);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_10) < 0) __PYX_ERR(5, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "celiagg/image.pxi":35
 *                       PixelFormat.RGBA32, PixelFormat.ARGB32,
 *                       PixelFormat.ABGR32),
 *         numpy.uint16: (PixelFormat.Gray16, PixelFormat.BGR48,             # <<<<<<<<<<<<<<
 *                        PixelFormat.RGB48, PixelFormat.BGRA64,
 *                        PixelFormat.RGBA64, PixelFormat.ARGB64,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_numpy); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_uint16); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray16); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR48); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "celiagg/image.pxi":36
 *                       PixelFormat.ABGR32),
 *         numpy.uint16: (PixelFormat.Gray16, PixelFormat.BGR48,
 *                        PixelFormat.RGB48, PixelFormat.BGRA64,             # <<<<<<<<<<<<<<
 *                        PixelFormat.RGBA64, PixelFormat.ARGB64,
 *                        PixelFormat.ABGR64),
 */
  __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB48); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA64); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "celiagg/image.pxi":37
 *         numpy.uint16: (PixelFormat.Gray16, PixelFormat.BGR48,
 *                        PixelFormat.RGB48, PixelFormat.BGRA64,
 *                        PixelFormat.RGBA64, PixelFormat.ARGB64,             # <<<<<<<<<<<<<<
 *                        PixelFormat.ABGR64),
 *         numpy.float32: (PixelFormat.Gray32, PixelFormat.BGR96,
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA64); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB64); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "celiagg/image.pxi":38
 *                        PixelFormat.RGB48, PixelFormat.BGRA64,
 *                        PixelFormat.RGBA64, PixelFormat.ARGB64,
 *                        PixelFormat.ABGR64),             # <<<<<<<<<<<<<<
 *         numpy.float32: (PixelFormat.Gray32, PixelFormat.BGR96,
 *                         PixelFormat.RGB96, PixelFormat.BGRA128,
 */
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR64); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "celiagg/image.pxi":35
 *                       PixelFormat.RGBA32, PixelFormat.ARGB32,
 *                       PixelFormat.ABGR32),
 *         numpy.uint16: (PixelFormat.Gray16, PixelFormat.BGR48,             # <<<<<<<<<<<<<<
 *                        PixelFormat.RGB48, PixelFormat.BGRA64,
 *                        PixelFormat.RGBA64, PixelFormat.ARGB64,
 */
  __pyx_t_2 = PyTuple_New(7); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_t_4);
  __pyx_t_10 = 0;
  __pyx_t_9 = 0;
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_2) < 0) __PYX_ERR(5, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "celiagg/image.pxi":39
 *                        PixelFormat.RGBA64, PixelFormat.ARGB64,
 *                        PixelFormat.ABGR64),
 *         numpy.float32: (PixelFormat.Gray32, PixelFormat.BGR96,             # <<<<<<<<<<<<<<
 *                         PixelFormat.RGB96, PixelFormat.BGRA128,
 *                         PixelFormat.RGBA128, PixelFormat.ARGB128,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float32); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray32); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR96); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "celiagg/image.pxi":40
 *                        PixelFormat.ABGR64),
 *         numpy.float32: (PixelFormat.Gray32, PixelFormat.BGR96,
 *                         PixelFormat.RGB96, PixelFormat.BGRA128,             # <<<<<<<<<<<<<<
 *                         PixelFormat.RGBA128, PixelFormat.ARGB128,
 *                         PixelFormat.ABGR128),
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB96); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA128); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "celiagg/image.pxi":41
 *         numpy.float32: (PixelFormat.Gray32, PixelFormat.BGR96,
 *                         PixelFormat.RGB96, PixelFormat.BGRA128,
 *                         PixelFormat.RGBA128, PixelFormat.ARGB128,             # <<<<<<<<<<<<<<
 *                         PixelFormat.ABGR128),
 *     }
 */
  __pyx_t_7 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA128); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB128); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "celiagg/image.pxi":42
 *                         PixelFormat.RGB96, PixelFormat.BGRA128,
 *                         PixelFormat.RGBA128, PixelFormat.ARGB128,
 *                         PixelFormat.ABGR128),             # <<<<<<<<<<<<<<
 *     }
 *     format_dtypes = {fmt: dt for dt, fmts in dtypes.items() for fmt in fmts}
 */
  __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR128); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "celiagg/image.pxi":39
 *                        PixelFormat.RGBA64, PixelFormat.ARGB64,
 *                        PixelFormat.ABGR64),
 *         numpy.float32: (PixelFormat.Gray32, PixelFormat.BGR96,             # <<<<<<<<<<<<<<
 *                         PixelFormat.RGB96, PixelFormat.BGRA128,
 *                         PixelFormat.RGBA128, PixelFormat.ARGB128,
 */
  __pyx_t_10 = PyTuple_New(7); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 5, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 6, __pyx_t_9);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_t_3, __pyx_t_10) < 0) __PYX_ERR(5, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_dtypes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/image.pxi":44
 *                         PixelFormat.ABGR128),
 *     }
 *     format_dtypes = {fmt: dt for dt, fmts in dtypes.items() for fmt in fmts}             # <<<<<<<<<<<<<<
 *     return format_dtypes[pixel_format]
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 44, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyDict_Items(__pyx_v_dtypes); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 44, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (likely(PyList_CheckExact(__pyx_t_10)) || PyTuple_CheckExact(__pyx_t_10)) {
      __pyx_t_3 = __pyx_t_10; __Pyx_INCREF(__pyx_t_3); __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
    } else {
      __pyx_t_11 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 44, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_12 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(5, 44, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    for (;;) {
      if (likely(!__pyx_t_12)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_11); __Pyx_INCREF(__pyx_t_10); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(5, 44, __pyx_L5_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 44, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        } else {
          if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_11); __Pyx_INCREF(__pyx_t_10); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(5, 44, __pyx_L5_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 44, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        }
      } else {
        __pyx_t_10 = __pyx_t_12(__pyx_t_3);
        if (unlikely(!__pyx_t_10)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(5, 44, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_10);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
        PyObject* sequence = __pyx_t_10;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(5, 44, __pyx_L5_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 44, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 44, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_7 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 44, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_13 = Py_TYPE(__pyx_t_7)->tp_iternext;
        index = 0; __pyx_t_9 = __pyx_t_13(__pyx_t_7); if (unlikely(!__pyx_t_9)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_9);
        index = 1; __pyx_t_8 = __pyx_t_13(__pyx_t_7); if (unlikely(!__pyx_t_8)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_7), 2) < 0) __PYX_ERR(5, 44, __pyx_L5_error)
        __pyx_t_13 = NULL;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_13 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(5, 44, __pyx_L5_error)
        __pyx_L9_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_dt, __pyx_t_9);
      __pyx_t_9 = 0;
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_fmts, __pyx_t_8);
      __pyx_t_8 = 0;
      if (likely(PyList_CheckExact(__pyx_7genexpr__pyx_v_fmts)) || PyTuple_CheckExact(__pyx_7genexpr__pyx_v_fmts)) {
        __pyx_t_10 = __pyx_7genexpr__pyx_v_fmts; __Pyx_INCREF(__pyx_t_10); __pyx_t_14 = 0;
        __pyx_t_15 = NULL;
      } else {
        __pyx_t_14 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_7genexpr__pyx_v_fmts); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 44, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_15 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(5, 44, __pyx_L5_error)
      }
      for (;;) {
        if (likely(!__pyx_t_15)) {
          if (likely(PyList_CheckExact(__pyx_t_10))) {
            if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_10)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_14); __Pyx_INCREF(__pyx_t_8); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(5, 44, __pyx_L5_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_10, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 44, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          } else {
            if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_14); __Pyx_INCREF(__pyx_t_8); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(5, 44, __pyx_L5_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_10, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 44, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          }
        } else {
          __pyx_t_8 = __pyx_t_15(__pyx_t_10);
          if (unlikely(!__pyx_t_8)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(5, 44, __pyx_L5_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_fmt, __pyx_t_8);
        __pyx_t_8 = 0;
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_7genexpr__pyx_v_fmt, (PyObject*)__pyx_7genexpr__pyx_v_dt))) __PYX_ERR(5, 44, __pyx_L5_error)
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_dt); __pyx_7genexpr__pyx_v_dt = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_fmt); __pyx_7genexpr__pyx_v_fmt = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_fmts); __pyx_7genexpr__pyx_v_fmts = 0;
    goto __pyx_L12_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_dt); __pyx_7genexpr__pyx_v_dt = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_fmt); __pyx_7genexpr__pyx_v_fmt = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_fmts); __pyx_7genexpr__pyx_v_fmts = 0;
    goto __pyx_L1_error;
    __pyx_L12_exit_scope:;
  } /* exit inner scope */
  __pyx_v_format_dtypes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/image.pxi":45
 *     }
 *     format_dtypes = {fmt: dt for dt, fmts in dtypes.items() for fmt in fmts}
 *     return format_dtypes[pixel_format]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_pixel_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_format_dtypes, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/image.pxi":29
 * ctypedef _image.Image* img_ptr_t
 * 
 * cdef _get_format_dtype(PixelFormat pixel_format):             # <<<<<<<<<<<<<<
 *     dtypes = {
 *         numpy.uint8: (PixelFormat.Gray8, PixelFormat.BGR24,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("celiagg._celiagg._get_format_dtype", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dtypes);
  __Pyx_XDECREF(__pyx_v_format_dtypes);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_dt);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_fmts);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_fmt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":48
 * 
 * 
 * cdef _get_format_last_dim(PixelFormat pixel_format):             # <<<<<<<<<<<<<<
 *     dims = {
 *         1: (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 */

static PyObject *__pyx_f_7celiagg_8_celiagg__get_format_last_dim(enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_pixel_format) {
  PyObject *__pyx_v_dims = NULL;
  PyObject *__pyx_v_format_dims = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_dim = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_fmts = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_fmt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  PyObject *(*__pyx_t_17)(PyObject *);
  Py_ssize_t __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  __Pyx_RefNannySetupContext("_get_format_last_dim", 0);

  /* "celiagg/image.pxi":50
 * cdef _get_format_last_dim(PixelFormat pixel_format):
 *     dims = {
 *         1: (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),             # <<<<<<<<<<<<<<
 *         3: (PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *             PixelFormat.RGB48, PixelFormat.BGR96, PixelFormat.RGB96),
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray8); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray16); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray32); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_1, __pyx_t_5) < 0) __PYX_ERR(5, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "celiagg/image.pxi":51
 *     dims = {
 *         1: (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 *         3: (PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,             # <<<<<<<<<<<<<<
 *             PixelFormat.RGB48, PixelFormat.BGR96, PixelFormat.RGB96),
 *         4: (PixelFormat.BGRA32, PixelFormat.RGBA32, PixelFormat.ARGB32,
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR24); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB24); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR48); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "celiagg/image.pxi":52
 *         1: (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 *         3: (PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *             PixelFormat.RGB48, PixelFormat.BGR96, PixelFormat.RGB96),             # <<<<<<<<<<<<<<
 *         4: (PixelFormat.BGRA32, PixelFormat.RGBA32, PixelFormat.ARGB32,
 *             PixelFormat.ABGR32, PixelFormat.BGRA64, PixelFormat.RGBA64,
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB48); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR96); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB96); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "celiagg/image.pxi":51
 *     dims = {
 *         1: (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 *         3: (PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,             # <<<<<<<<<<<<<<
 *             PixelFormat.RGB48, PixelFormat.BGR96, PixelFormat.RGB96),
 *         4: (PixelFormat.BGRA32, PixelFormat.RGBA32, PixelFormat.ARGB32,
 */
  __pyx_t_8 = PyTuple_New(6); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 5, __pyx_t_7);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_3, __pyx_t_8) < 0) __PYX_ERR(5, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "celiagg/image.pxi":53
 *         3: (PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *             PixelFormat.RGB48, PixelFormat.BGR96, PixelFormat.RGB96),
 *         4: (PixelFormat.BGRA32, PixelFormat.RGBA32, PixelFormat.ARGB32,             # <<<<<<<<<<<<<<
 *             PixelFormat.ABGR32, PixelFormat.BGRA64, PixelFormat.RGBA64,
 *             PixelFormat.ARGB64, PixelFormat.ABGR64, PixelFormat.BGRA128,
 */
  __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA32); if (unlikely(!__pyx_t_8)) __PYX_ERR(5, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_7 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA32); if (unlikely(!__pyx_t_7)) __PYX_ERR(5, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB32); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "celiagg/image.pxi":54
 *             PixelFormat.RGB48, PixelFormat.BGR96, PixelFormat.RGB96),
 *         4: (PixelFormat.BGRA32, PixelFormat.RGBA32, PixelFormat.ARGB32,
 *             PixelFormat.ABGR32, PixelFormat.BGRA64, PixelFormat.RGBA64,             # <<<<<<<<<<<<<<
 *             PixelFormat.ARGB64, PixelFormat.ABGR64, PixelFormat.BGRA128,
 *             PixelFormat.RGBA128, PixelFormat.ARGB128, PixelFormat.ABGR128)
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR32); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA64); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA64); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "celiagg/image.pxi":55
 *         4: (PixelFormat.BGRA32, PixelFormat.RGBA32, PixelFormat.ARGB32,
 *             PixelFormat.ABGR32, PixelFormat.BGRA64, PixelFormat.RGBA64,
 *             PixelFormat.ARGB64, PixelFormat.ABGR64, PixelFormat.BGRA128,             # <<<<<<<<<<<<<<
 *             PixelFormat.RGBA128, PixelFormat.ARGB128, PixelFormat.ABGR128)
 *     }
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB64); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR64); if (unlikely(!__pyx_t_9)) __PYX_ERR(5, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA128); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "celiagg/image.pxi":56
 *             PixelFormat.ABGR32, PixelFormat.BGRA64, PixelFormat.RGBA64,
 *             PixelFormat.ARGB64, PixelFormat.ABGR64, PixelFormat.BGRA128,
 *             PixelFormat.RGBA128, PixelFormat.ARGB128, PixelFormat.ABGR128)             # <<<<<<<<<<<<<<
 *     }
 *     format_dims = {fmt: dim for dim, fmts in dims.items() for fmt in fmts}
 */
  __pyx_t_11 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA128); if (unlikely(!__pyx_t_11)) __PYX_ERR(5, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB128); if (unlikely(!__pyx_t_12)) __PYX_ERR(5, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR128); if (unlikely(!__pyx_t_13)) __PYX_ERR(5, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "celiagg/image.pxi":53
 *         3: (PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *             PixelFormat.RGB48, PixelFormat.BGR96, PixelFormat.RGB96),
 *         4: (PixelFormat.BGRA32, PixelFormat.RGBA32, PixelFormat.ARGB32,             # <<<<<<<<<<<<<<
 *             PixelFormat.ABGR32, PixelFormat.BGRA64, PixelFormat.RGBA64,
 *             PixelFormat.ARGB64, PixelFormat.ABGR64, PixelFormat.BGRA128,
 */
  __pyx_t_14 = PyTuple_New(12); if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_14, 2, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_14, 3, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_14, 4, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_14, 5, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_14, 6, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_14, 7, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_14, 8, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_14, 9, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_14, 10, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_14, 11, __pyx_t_13);
  __pyx_t_8 = 0;
  __pyx_t_7 = 0;
  __pyx_t_6 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_4, __pyx_t_14) < 0) __PYX_ERR(5, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_v_dims = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/image.pxi":58
 *             PixelFormat.RGBA128, PixelFormat.ARGB128, PixelFormat.ABGR128)
 *     }
 *     format_dims = {fmt: dim for dim, fmts in dims.items() for fmt in fmts}             # <<<<<<<<<<<<<<
 *     return format_dims[pixel_format]
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 58, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = __Pyx_PyDict_Items(__pyx_v_dims); if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 58, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
      __pyx_t_13 = __pyx_t_14; __Pyx_INCREF(__pyx_t_13); __pyx_t_15 = 0;
      __pyx_t_16 = NULL;
    } else {
      __pyx_t_15 = -1; __pyx_t_13 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_13)) __PYX_ERR(5, 58, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_16 = Py_TYPE(__pyx_t_13)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(5, 58, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    for (;;) {
      if (likely(!__pyx_t_16)) {
        if (likely(PyList_CheckExact(__pyx_t_13))) {
          if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_13)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_14 = PyList_GET_ITEM(__pyx_t_13, __pyx_t_15); __Pyx_INCREF(__pyx_t_14); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(5, 58, __pyx_L5_error)
          #else
          __pyx_t_14 = PySequence_ITEM(__pyx_t_13, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 58, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_14);
          #endif
        } else {
          if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_13)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_13, __pyx_t_15); __Pyx_INCREF(__pyx_t_14); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(5, 58, __pyx_L5_error)
          #else
          __pyx_t_14 = PySequence_ITEM(__pyx_t_13, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 58, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_14);
          #endif
        }
      } else {
        __pyx_t_14 = __pyx_t_16(__pyx_t_13);
        if (unlikely(!__pyx_t_14)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(5, 58, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_14);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_14))) || (PyList_CheckExact(__pyx_t_14))) {
        PyObject* sequence = __pyx_t_14;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(5, 58, __pyx_L5_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_12 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_11 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_12 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_11 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_11);
        #else
        __pyx_t_12 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(5, 58, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_11 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(5, 58, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_11);
        #endif
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_10 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_10)) __PYX_ERR(5, 58, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_17 = Py_TYPE(__pyx_t_10)->tp_iternext;
        index = 0; __pyx_t_12 = __pyx_t_17(__pyx_t_10); if (unlikely(!__pyx_t_12)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_12);
        index = 1; __pyx_t_11 = __pyx_t_17(__pyx_t_10); if (unlikely(!__pyx_t_11)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_11);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_17(__pyx_t_10), 2) < 0) __PYX_ERR(5, 58, __pyx_L5_error)
        __pyx_t_17 = NULL;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_17 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(5, 58, __pyx_L5_error)
        __pyx_L9_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_dim, __pyx_t_12);
      __pyx_t_12 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_fmts, __pyx_t_11);
      __pyx_t_11 = 0;
      if (likely(PyList_CheckExact(__pyx_8genexpr1__pyx_v_fmts)) || PyTuple_CheckExact(__pyx_8genexpr1__pyx_v_fmts)) {
        __pyx_t_14 = __pyx_8genexpr1__pyx_v_fmts; __Pyx_INCREF(__pyx_t_14); __pyx_t_18 = 0;
        __pyx_t_19 = NULL;
      } else {
        __pyx_t_18 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_8genexpr1__pyx_v_fmts); if (unlikely(!__pyx_t_14)) __PYX_ERR(5, 58, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_19 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_19)) __PYX_ERR(5, 58, __pyx_L5_error)
      }
      for (;;) {
        if (likely(!__pyx_t_19)) {
          if (likely(PyList_CheckExact(__pyx_t_14))) {
            if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_14)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_11 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_18); __Pyx_INCREF(__pyx_t_11); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(5, 58, __pyx_L5_error)
            #else
            __pyx_t_11 = PySequence_ITEM(__pyx_t_14, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_11)) __PYX_ERR(5, 58, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_11);
            #endif
          } else {
            if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_18); __Pyx_INCREF(__pyx_t_11); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(5, 58, __pyx_L5_error)
            #else
            __pyx_t_11 = PySequence_ITEM(__pyx_t_14, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_11)) __PYX_ERR(5, 58, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_11);
            #endif
          }
        } else {
          __pyx_t_11 = __pyx_t_19(__pyx_t_14);
          if (unlikely(!__pyx_t_11)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(5, 58, __pyx_L5_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_fmt, __pyx_t_11);
        __pyx_t_11 = 0;
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr1__pyx_v_fmt, (PyObject*)__pyx_8genexpr1__pyx_v_dim))) __PYX_ERR(5, 58, __pyx_L5_error)
      }
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    }
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_dim); __pyx_8genexpr1__pyx_v_dim = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_fmt); __pyx_8genexpr1__pyx_v_fmt = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_fmts); __pyx_8genexpr1__pyx_v_fmts = 0;
    goto __pyx_L12_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_dim); __pyx_8genexpr1__pyx_v_dim = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_fmt); __pyx_8genexpr1__pyx_v_fmt = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_fmts); __pyx_8genexpr1__pyx_v_fmts = 0;
    goto __pyx_L1_error;
    __pyx_L12_exit_scope:;
  } /* exit inner scope */
  __pyx_v_format_dims = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/image.pxi":59
 *     }
 *     format_dims = {fmt: dim for dim, fmts in dims.items() for fmt in fmts}
 *     return format_dims[pixel_format]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_pixel_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = __Pyx_PyDict_GetItem(__pyx_v_format_dims, __pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(5, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_13;
  __pyx_t_13 = 0;
  goto __pyx_L0;

  /* "celiagg/image.pxi":48
 * 
 * 
 * cdef _get_format_last_dim(PixelFormat pixel_format):             # <<<<<<<<<<<<<<
 *     dims = {
 *         1: (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("celiagg._celiagg._get_format_last_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dims);
  __Pyx_XDECREF(__pyx_v_format_dims);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_dim);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_fmts);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_fmt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":62
 * 
 * 
 * cdef img_ptr_t _get_2d_u8_img(np.npy_uint8[:, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_2d_u8_img(__Pyx_memviewslice __pyx_v_arr, bool __pyx_v_bottom_up) {
  __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannySetupContext("_get_2d_u8_img", 0);

  /* "celiagg/image.pxi":63
 * 
 * cdef img_ptr_t _get_2d_u8_img(np.npy_uint8[:, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = -1;
  if (__pyx_t_1 < 0) {
    __pyx_t_1 += __pyx_v_arr.shape[0];
    if (unlikely(__pyx_t_1 < 0)) __pyx_t_3 = 0;
  } else if (unlikely(__pyx_t_1 >= __pyx_v_arr.shape[0])) __pyx_t_3 = 0;
  if (__pyx_t_2 < 0) {
    __pyx_t_2 += __pyx_v_arr.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __pyx_t_3 = 1;
  } else if (unlikely(__pyx_t_2 >= __pyx_v_arr.shape[1])) __pyx_t_3 = 1;
  if (unlikely(__pyx_t_3 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_3);
    __PYX_ERR(5, 63, __pyx_L1_error)
  }

  /* "celiagg/image.pxi":64
 * cdef img_ptr_t _get_2d_u8_img(np.npy_uint8[:, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if ((__pyx_v_bottom_up != 0)) {
    __pyx_t_4 = (-(__pyx_v_arr.strides[0]));
  } else {
    __pyx_t_4 = (__pyx_v_arr.strides[0]);
  }

  /* "celiagg/image.pxi":63
 * 
 * cdef img_ptr_t _get_2d_u8_img(np.npy_uint8[:, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_r = new Image(((unsigned char *)(&(*((npy_uint8 *) ( /* dim=1 */ ((char *) (((npy_uint8 *) ( /* dim=0 */ (__pyx_v_arr.data + __pyx_t_1 * __pyx_v_arr.strides[0]) )) + __pyx_t_2)) ))))), (__pyx_v_arr.shape[1]), (__pyx_v_arr.shape[0]), __pyx_t_4);
  goto __pyx_L0;

  /* "celiagg/image.pxi":62
 * 
 * 
 * cdef img_ptr_t _get_2d_u8_img(np.npy_uint8[:, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("celiagg._celiagg._get_2d_u8_img", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":67
 * 
 * 
 * cdef img_ptr_t _get_2d_u16_img(np.npy_uint16[:, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_2d_u16_img(__Pyx_memviewslice __pyx_v_arr, bool __pyx_v_bottom_up) {
  __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannySetupContext("_get_2d_u16_img", 0);

  /* "celiagg/image.pxi":68
 * 
 * cdef img_ptr_t _get_2d_u16_img(np.npy_uint16[:, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = -1;
  if (__pyx_t_1 < 0) {
    __pyx_t_1 += __pyx_v_arr.shape[0];
    if (unlikely(__pyx_t_1 < 0)) __pyx_t_3 = 0;
  } else if (unlikely(__pyx_t_1 >= __pyx_v_arr.shape[0])) __pyx_t_3 = 0;
  if (__pyx_t_2 < 0) {
    __pyx_t_2 += __pyx_v_arr.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __pyx_t_3 = 1;
  } else if (unlikely(__pyx_t_2 >= __pyx_v_arr.shape[1])) __pyx_t_3 = 1;
  if (unlikely(__pyx_t_3 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_3);
    __PYX_ERR(5, 68, __pyx_L1_error)
  }

  /* "celiagg/image.pxi":69
 * cdef img_ptr_t _get_2d_u16_img(np.npy_uint16[:, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if ((__pyx_v_bottom_up != 0)) {
    __pyx_t_4 = (-(__pyx_v_arr.strides[0]));
  } else {
    __pyx_t_4 = (__pyx_v_arr.strides[0]);
  }

  /* "celiagg/image.pxi":68
 * 
 * cdef img_ptr_t _get_2d_u16_img(np.npy_uint16[:, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_r = new Image(((unsigned char *)(&(*((npy_uint16 *) ( /* dim=1 */ ((char *) (((npy_uint16 *) ( /* dim=0 */ (__pyx_v_arr.data + __pyx_t_1 * __pyx_v_arr.strides[0]) )) + __pyx_t_2)) ))))), (__pyx_v_arr.shape[1]), (__pyx_v_arr.shape[0]), __pyx_t_4);
  goto __pyx_L0;

  /* "celiagg/image.pxi":67
 * 
 * 
 * cdef img_ptr_t _get_2d_u16_img(np.npy_uint16[:, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("celiagg._celiagg._get_2d_u16_img", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":72
 * 
 * 
 * cdef img_ptr_t _get_2d_f32_img(float[:, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_2d_f32_img(__Pyx_memviewslice __pyx_v_arr, bool __pyx_v_bottom_up) {
  __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  __Pyx_RefNannySetupContext("_get_2d_f32_img", 0);

  /* "celiagg/image.pxi":73
 * 
 * cdef img_ptr_t _get_2d_f32_img(float[:, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = -1;
  if (__pyx_t_1 < 0) {
    __pyx_t_1 += __pyx_v_arr.shape[0];
    if (unlikely(__pyx_t_1 < 0)) __pyx_t_3 = 0;
  } else if (unlikely(__pyx_t_1 >= __pyx_v_arr.shape[0])) __pyx_t_3 = 0;
  if (__pyx_t_2 < 0) {
    __pyx_t_2 += __pyx_v_arr.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __pyx_t_3 = 1;
  } else if (unlikely(__pyx_t_2 >= __pyx_v_arr.shape[1])) __pyx_t_3 = 1;
  if (unlikely(__pyx_t_3 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_3);
    __PYX_ERR(5, 73, __pyx_L1_error)
  }

  /* "celiagg/image.pxi":74
 * cdef img_ptr_t _get_2d_f32_img(float[:, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if ((__pyx_v_bottom_up != 0)) {
    __pyx_t_4 = (-(__pyx_v_arr.strides[0]));
  } else {
    __pyx_t_4 = (__pyx_v_arr.strides[0]);
  }

  /* "celiagg/image.pxi":73
 * 
 * cdef img_ptr_t _get_2d_f32_img(float[:, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_r = new Image(((unsigned char *)(&(*((float *) ( /* dim=1 */ ((char *) (((float *) ( /* dim=0 */ (__pyx_v_arr.data + __pyx_t_1 * __pyx_v_arr.strides[0]) )) + __pyx_t_2)) ))))), (__pyx_v_arr.shape[1]), (__pyx_v_arr.shape[0]), __pyx_t_4);
  goto __pyx_L0;

  /* "celiagg/image.pxi":72
 * 
 * 
 * cdef img_ptr_t _get_2d_f32_img(float[:, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("celiagg._celiagg._get_2d_f32_img", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":77
 * 
 * 
 * cdef img_ptr_t _get_3d_u8_img(np.npy_uint8[:, :, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_3d_u8_img(__Pyx_memviewslice __pyx_v_arr, bool __pyx_v_bottom_up) {
  __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("_get_3d_u8_img", 0);

  /* "celiagg/image.pxi":78
 * 
 * cdef img_ptr_t _get_3d_u8_img(np.npy_uint8[:, :, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = -1;
  if (__pyx_t_1 < 0) {
    __pyx_t_1 += __pyx_v_arr.shape[0];
    if (unlikely(__pyx_t_1 < 0)) __pyx_t_4 = 0;
  } else if (unlikely(__pyx_t_1 >= __pyx_v_arr.shape[0])) __pyx_t_4 = 0;
  if (__pyx_t_2 < 0) {
    __pyx_t_2 += __pyx_v_arr.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __pyx_t_4 = 1;
  } else if (unlikely(__pyx_t_2 >= __pyx_v_arr.shape[1])) __pyx_t_4 = 1;
  if (__pyx_t_3 < 0) {
    __pyx_t_3 += __pyx_v_arr.shape[2];
    if (unlikely(__pyx_t_3 < 0)) __pyx_t_4 = 2;
  } else if (unlikely(__pyx_t_3 >= __pyx_v_arr.shape[2])) __pyx_t_4 = 2;
  if (unlikely(__pyx_t_4 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_4);
    __PYX_ERR(5, 78, __pyx_L1_error)
  }

  /* "celiagg/image.pxi":79
 * cdef img_ptr_t _get_3d_u8_img(np.npy_uint8[:, :, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if ((__pyx_v_bottom_up != 0)) {
    __pyx_t_5 = (-(__pyx_v_arr.strides[0]));
  } else {
    __pyx_t_5 = (__pyx_v_arr.strides[0]);
  }

  /* "celiagg/image.pxi":78
 * 
 * cdef img_ptr_t _get_3d_u8_img(np.npy_uint8[:, :, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_r = new Image(((unsigned char *)(&(*((npy_uint8 *) ( /* dim=2 */ ((char *) (((npy_uint8 *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_arr.data + __pyx_t_1 * __pyx_v_arr.strides[0]) ) + __pyx_t_2 * __pyx_v_arr.strides[1]) )) + __pyx_t_3)) ))))), (__pyx_v_arr.shape[1]), (__pyx_v_arr.shape[0]), __pyx_t_5);
  goto __pyx_L0;

  /* "celiagg/image.pxi":77
 * 
 * 
 * cdef img_ptr_t _get_3d_u8_img(np.npy_uint8[:, :, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("celiagg._celiagg._get_3d_u8_img", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":82
 * 
 * 
 * cdef img_ptr_t _get_3d_u16_img(np.npy_uint16[:, :, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_3d_u16_img(__Pyx_memviewslice __pyx_v_arr, bool __pyx_v_bottom_up) {
  __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("_get_3d_u16_img", 0);

  /* "celiagg/image.pxi":83
 * 
 * cdef img_ptr_t _get_3d_u16_img(np.npy_uint16[:, :, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = -1;
  if (__pyx_t_1 < 0) {
    __pyx_t_1 += __pyx_v_arr.shape[0];
    if (unlikely(__pyx_t_1 < 0)) __pyx_t_4 = 0;
  } else if (unlikely(__pyx_t_1 >= __pyx_v_arr.shape[0])) __pyx_t_4 = 0;
  if (__pyx_t_2 < 0) {
    __pyx_t_2 += __pyx_v_arr.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __pyx_t_4 = 1;
  } else if (unlikely(__pyx_t_2 >= __pyx_v_arr.shape[1])) __pyx_t_4 = 1;
  if (__pyx_t_3 < 0) {
    __pyx_t_3 += __pyx_v_arr.shape[2];
    if (unlikely(__pyx_t_3 < 0)) __pyx_t_4 = 2;
  } else if (unlikely(__pyx_t_3 >= __pyx_v_arr.shape[2])) __pyx_t_4 = 2;
  if (unlikely(__pyx_t_4 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_4);
    __PYX_ERR(5, 83, __pyx_L1_error)
  }

  /* "celiagg/image.pxi":84
 * cdef img_ptr_t _get_3d_u16_img(np.npy_uint16[:, :, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if ((__pyx_v_bottom_up != 0)) {
    __pyx_t_5 = (-(__pyx_v_arr.strides[0]));
  } else {
    __pyx_t_5 = (__pyx_v_arr.strides[0]);
  }

  /* "celiagg/image.pxi":83
 * 
 * cdef img_ptr_t _get_3d_u16_img(np.npy_uint16[:, :, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_r = new Image(((unsigned char *)(&(*((npy_uint16 *) ( /* dim=2 */ ((char *) (((npy_uint16 *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_arr.data + __pyx_t_1 * __pyx_v_arr.strides[0]) ) + __pyx_t_2 * __pyx_v_arr.strides[1]) )) + __pyx_t_3)) ))))), (__pyx_v_arr.shape[1]), (__pyx_v_arr.shape[0]), __pyx_t_5);
  goto __pyx_L0;

  /* "celiagg/image.pxi":82
 * 
 * 
 * cdef img_ptr_t _get_3d_u16_img(np.npy_uint16[:, :, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("celiagg._celiagg._get_3d_u16_img", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":87
 * 
 * 
 * cdef img_ptr_t _get_3d_f32_img(float[:, :, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_3d_f32_img(__Pyx_memviewslice __pyx_v_arr, bool __pyx_v_bottom_up) {
  __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("_get_3d_f32_img", 0);

  /* "celiagg/image.pxi":88
 * 
 * cdef img_ptr_t _get_3d_f32_img(float[:, :, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = -1;
  if (__pyx_t_1 < 0) {
    __pyx_t_1 += __pyx_v_arr.shape[0];
    if (unlikely(__pyx_t_1 < 0)) __pyx_t_4 = 0;
  } else if (unlikely(__pyx_t_1 >= __pyx_v_arr.shape[0])) __pyx_t_4 = 0;
  if (__pyx_t_2 < 0) {
    __pyx_t_2 += __pyx_v_arr.shape[1];
    if (unlikely(__pyx_t_2 < 0)) __pyx_t_4 = 1;
  } else if (unlikely(__pyx_t_2 >= __pyx_v_arr.shape[1])) __pyx_t_4 = 1;
  if (__pyx_t_3 < 0) {
    __pyx_t_3 += __pyx_v_arr.shape[2];
    if (unlikely(__pyx_t_3 < 0)) __pyx_t_4 = 2;
  } else if (unlikely(__pyx_t_3 >= __pyx_v_arr.shape[2])) __pyx_t_4 = 2;
  if (unlikely(__pyx_t_4 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_4);
    __PYX_ERR(5, 88, __pyx_L1_error)
  }

  /* "celiagg/image.pxi":89
 * cdef img_ptr_t _get_3d_f32_img(float[:, :, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if ((__pyx_v_bottom_up != 0)) {
    __pyx_t_5 = (-(__pyx_v_arr.strides[0]));
  } else {
    __pyx_t_5 = (__pyx_v_arr.strides[0]);
  }

  /* "celiagg/image.pxi":88
 * 
 * cdef img_ptr_t _get_3d_f32_img(float[:, :, ::1] arr, bool bottom_up):
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],             # <<<<<<<<<<<<<<
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 * 
 */
  __pyx_r = new Image(((unsigned char *)(&(*((float *) ( /* dim=2 */ ((char *) (((float *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_arr.data + __pyx_t_1 * __pyx_v_arr.strides[0]) ) + __pyx_t_2 * __pyx_v_arr.strides[1]) )) + __pyx_t_3)) ))))), (__pyx_v_arr.shape[1]), (__pyx_v_arr.shape[0]), __pyx_t_5);
  goto __pyx_L0;

  /* "celiagg/image.pxi":87
 * 
 * 
 * cdef img_ptr_t _get_3d_f32_img(float[:, :, ::1] arr, bool bottom_up):             # <<<<<<<<<<<<<<
 *     return new _image.Image(<unsigned char*>&arr[0][0][0], arr.shape[1], arr.shape[0],
 *                             -arr.strides[0] if bottom_up else arr.strides[0])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("celiagg._celiagg._get_3d_f32_img", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":92
 * 
 * 
 * cdef img_ptr_t _get_image(array, pixel_format, bottom_up):             # <<<<<<<<<<<<<<
 *     dtype = _get_format_dtype(pixel_format)
 * 
 */

static __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_f_7celiagg_8_celiagg__get_image(PyObject *__pyx_v_array, PyObject *__pyx_v_pixel_format, PyObject *__pyx_v_bottom_up) {
  PyObject *__pyx_v_dtype = NULL;
  __pyx_t_7celiagg_8_celiagg_img_ptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  bool __pyx_t_7;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_9 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannySetupContext("_get_image", 0);

  /* "celiagg/image.pxi":93
 * 
 * cdef img_ptr_t _get_image(array, pixel_format, bottom_up):
 *     dtype = _get_format_dtype(pixel_format)             # <<<<<<<<<<<<<<
 * 
 *     # Finally build the image
 */
  __pyx_t_1 = ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_pixel_format)); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 93, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_7celiagg_8_celiagg__get_format_dtype(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_dtype = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "celiagg/image.pxi":96
 * 
 *     # Finally build the image
 *     if array.ndim == 2:             # <<<<<<<<<<<<<<
 *         if dtype is numpy.uint8:
 *             return _get_2d_u8_img(array, bottom_up)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_array, __pyx_n_s_ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_t_2, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(5, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "celiagg/image.pxi":97
 *     # Finally build the image
 *     if array.ndim == 2:
 *         if dtype is numpy.uint8:             # <<<<<<<<<<<<<<
 *             return _get_2d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_uint8); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_v_dtype == __pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "celiagg/image.pxi":98
 *     if array.ndim == 2:
 *         if dtype is numpy.uint8:
 *             return _get_2d_u8_img(array, bottom_up)             # <<<<<<<<<<<<<<
 *         elif dtype is numpy.uint16:
 *             return _get_2d_u16_img(array, bottom_up)
 */
      __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_npy_uint8(__pyx_v_array, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(5, 98, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_7 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(5, 98, __pyx_L1_error)
      __pyx_r = __pyx_f_7celiagg_8_celiagg__get_2d_u8_img(__pyx_t_6, __pyx_t_7);
      __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
      __pyx_t_6.memview = NULL;
      __pyx_t_6.data = NULL;
      goto __pyx_L0;

      /* "celiagg/image.pxi":97
 *     # Finally build the image
 *     if array.ndim == 2:
 *         if dtype is numpy.uint8:             # <<<<<<<<<<<<<<
 *             return _get_2d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:
 */
    }

    /* "celiagg/image.pxi":99
 *         if dtype is numpy.uint8:
 *             return _get_2d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:             # <<<<<<<<<<<<<<
 *             return _get_2d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_uint16); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__pyx_v_dtype == __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {

      /* "celiagg/image.pxi":100
 *             return _get_2d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:
 *             return _get_2d_u16_img(array, bottom_up)             # <<<<<<<<<<<<<<
 *         elif dtype is numpy.float32:
 *             return _get_2d_f32_img(array, bottom_up)
 */
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_npy_uint16(__pyx_v_array, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(5, 100, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_7 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(5, 100, __pyx_L1_error)
      __pyx_r = __pyx_f_7celiagg_8_celiagg__get_2d_u16_img(__pyx_t_8, __pyx_t_7);
      __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
      __pyx_t_8.memview = NULL;
      __pyx_t_8.data = NULL;
      goto __pyx_L0;

      /* "celiagg/image.pxi":99
 *         if dtype is numpy.uint8:
 *             return _get_2d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:             # <<<<<<<<<<<<<<
 *             return _get_2d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:
 */
    }

    /* "celiagg/image.pxi":101
 *         elif dtype is numpy.uint16:
 *             return _get_2d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:             # <<<<<<<<<<<<<<
 *             return _get_2d_f32_img(array, bottom_up)
 *     else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float32); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_v_dtype == __pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "celiagg/image.pxi":102
 *             return _get_2d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:
 *             return _get_2d_f32_img(array, bottom_up)             # <<<<<<<<<<<<<<
 *     else:
 *         if dtype is numpy.uint8:
 */
      __pyx_t_9 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_float(__pyx_v_array, PyBUF_WRITABLE); if (unlikely(!__pyx_t_9.memview)) __PYX_ERR(5, 102, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_7 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(5, 102, __pyx_L1_error)
      __pyx_r = __pyx_f_7celiagg_8_celiagg__get_2d_f32_img(__pyx_t_9, __pyx_t_7);
      __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
      __pyx_t_9.memview = NULL;
      __pyx_t_9.data = NULL;
      goto __pyx_L0;

      /* "celiagg/image.pxi":101
 *         elif dtype is numpy.uint16:
 *             return _get_2d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:             # <<<<<<<<<<<<<<
 *             return _get_2d_f32_img(array, bottom_up)
 *     else:
 */
    }

    /* "celiagg/image.pxi":96
 * 
 *     # Finally build the image
 *     if array.ndim == 2:             # <<<<<<<<<<<<<<
 *         if dtype is numpy.uint8:
 *             return _get_2d_u8_img(array, bottom_up)
 */
    goto __pyx_L3;
  }

  /* "celiagg/image.pxi":104
 *             return _get_2d_f32_img(array, bottom_up)
 *     else:
 *         if dtype is numpy.uint8:             # <<<<<<<<<<<<<<
 *             return _get_3d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_uint8); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__pyx_v_dtype == __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {

      /* "celiagg/image.pxi":105
 *     else:
 *         if dtype is numpy.uint8:
 *             return _get_3d_u8_img(array, bottom_up)             # <<<<<<<<<<<<<<
 *         elif dtype is numpy.uint16:
 *             return _get_3d_u16_img(array, bottom_up)
 */
      __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn_npy_uint8(__pyx_v_array, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(5, 105, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_7 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(5, 105, __pyx_L1_error)
      __pyx_r = __pyx_f_7celiagg_8_celiagg__get_3d_u8_img(__pyx_t_10, __pyx_t_7);
      __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
      __pyx_t_10.memview = NULL;
      __pyx_t_10.data = NULL;
      goto __pyx_L0;

      /* "celiagg/image.pxi":104
 *             return _get_2d_f32_img(array, bottom_up)
 *     else:
 *         if dtype is numpy.uint8:             # <<<<<<<<<<<<<<
 *             return _get_3d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:
 */
    }

    /* "celiagg/image.pxi":106
 *         if dtype is numpy.uint8:
 *             return _get_3d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:             # <<<<<<<<<<<<<<
 *             return _get_3d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_uint16); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_v_dtype == __pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "celiagg/image.pxi":107
 *             return _get_3d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:
 *             return _get_3d_u16_img(array, bottom_up)             # <<<<<<<<<<<<<<
 *         elif dtype is numpy.float32:
 *             return _get_3d_f32_img(array, bottom_up)
 */
      __pyx_t_11 = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn_npy_uint16(__pyx_v_array, PyBUF_WRITABLE); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(5, 107, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_7 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(5, 107, __pyx_L1_error)
      __pyx_r = __pyx_f_7celiagg_8_celiagg__get_3d_u16_img(__pyx_t_11, __pyx_t_7);
      __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
      __pyx_t_11.memview = NULL;
      __pyx_t_11.data = NULL;
      goto __pyx_L0;

      /* "celiagg/image.pxi":106
 *         if dtype is numpy.uint8:
 *             return _get_3d_u8_img(array, bottom_up)
 *         elif dtype is numpy.uint16:             # <<<<<<<<<<<<<<
 *             return _get_3d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:
 */
    }

    /* "celiagg/image.pxi":108
 *         elif dtype is numpy.uint16:
 *             return _get_3d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:             # <<<<<<<<<<<<<<
 *             return _get_3d_f32_img(array, bottom_up)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float32); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__pyx_v_dtype == __pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {

      /* "celiagg/image.pxi":109
 *             return _get_3d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:
 *             return _get_3d_f32_img(array, bottom_up)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_float(__pyx_v_array, PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(5, 109, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_7 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(5, 109, __pyx_L1_error)
      __pyx_r = __pyx_f_7celiagg_8_celiagg__get_3d_f32_img(__pyx_t_12, __pyx_t_7);
      __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
      __pyx_t_12.memview = NULL;
      __pyx_t_12.data = NULL;
      goto __pyx_L0;

      /* "celiagg/image.pxi":108
 *         elif dtype is numpy.uint16:
 *             return _get_3d_u16_img(array, bottom_up)
 *         elif dtype is numpy.float32:             # <<<<<<<<<<<<<<
 *             return _get_3d_f32_img(array, bottom_up)
 * 
 */
    }
  }
  __pyx_L3:;

  /* "celiagg/image.pxi":92
 * 
 * 
 * cdef img_ptr_t _get_image(array, pixel_format, bottom_up):             # <<<<<<<<<<<<<<
 *     dtype = _get_format_dtype(pixel_format)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_9, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __Pyx_WriteUnraisable("celiagg._celiagg._get_image", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":124
 *     cdef bool bottom_up
 * 
 *     def __cinit__(self, array, PixelFormat pixel_format, bool bottom_up=False):             # <<<<<<<<<<<<<<
 *         expected_dtype = _get_format_dtype(pixel_format)
 *         expected_dim = _get_format_last_dim(pixel_format)
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_5Image_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_5Image_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_array = 0;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_pixel_format;
  bool __pyx_v_bottom_up;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_array,&__pyx_n_s_pixel_format,&__pyx_n_s_bottom_up,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_array)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pixel_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(5, 124, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bottom_up);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(5, 124, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_array = values[0];
    __pyx_v_pixel_format = ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(values[1])); if (unlikely(PyErr_Occurred())) __PYX_ERR(5, 124, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_bottom_up = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_bottom_up == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(5, 124, __pyx_L3_error)
    } else {
      __pyx_v_bottom_up = ((bool)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(5, 124, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Image.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Image___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_v_self), __pyx_v_array, __pyx_v_pixel_format, __pyx_v_bottom_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_5Image___cinit__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self, PyObject *__pyx_v_array, enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_pixel_format, bool __pyx_v_bottom_up) {
  PyObject *__pyx_v_expected_dtype = NULL;
  PyObject *__pyx_v_expected_dim = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_v_pix_fmt_name = NULL;
  PyObject *__pyx_v_ex_dtype_name = NULL;
  PyObject *__pyx_v_dtype_name = NULL;
  PyObject *__pyx_v_last_dim = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/image.pxi":125
 * 
 *     def __cinit__(self, array, PixelFormat pixel_format, bool bottom_up=False):
 *         expected_dtype = _get_format_dtype(pixel_format)             # <<<<<<<<<<<<<<
 *         expected_dim = _get_format_last_dim(pixel_format)
 * 
 */
  __pyx_t_1 = __pyx_f_7celiagg_8_celiagg__get_format_dtype(__pyx_v_pixel_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expected_dtype = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/image.pxi":126
 *     def __cinit__(self, array, PixelFormat pixel_format, bool bottom_up=False):
 *         expected_dtype = _get_format_dtype(pixel_format)
 *         expected_dim = _get_format_last_dim(pixel_format)             # <<<<<<<<<<<<<<
 * 
 *         # Do several sanity checks...
 */
  __pyx_t_1 = __pyx_f_7celiagg_8_celiagg__get_format_last_dim(__pyx_v_pixel_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_expected_dim = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/image.pxi":129
 * 
 *         # Do several sanity checks...
 *         if not isinstance(array, numpy.ndarray):             # <<<<<<<<<<<<<<
 *             raise TypeError("An image must be provided in a numpy array")
 * 
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_array, __pyx_ptype_5numpy_ndarray); 
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "celiagg/image.pxi":130
 *         # Do several sanity checks...
 *         if not isinstance(array, numpy.ndarray):
 *             raise TypeError("An image must be provided in a numpy array")             # <<<<<<<<<<<<<<
 * 
 *         if array.dtype.type is not expected_dtype:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(5, 130, __pyx_L1_error)

    /* "celiagg/image.pxi":129
 * 
 *         # Do several sanity checks...
 *         if not isinstance(array, numpy.ndarray):             # <<<<<<<<<<<<<<
 *             raise TypeError("An image must be provided in a numpy array")
 * 
 */
  }

  /* "celiagg/image.pxi":132
 *             raise TypeError("An image must be provided in a numpy array")
 * 
 *         if array.dtype.type is not expected_dtype:             # <<<<<<<<<<<<<<
 *             msg = "The '{}' format requires an array of type {}, but {} was passed"
 *             pix_fmt_name = PixelFormat(pixel_format).name
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_array, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_type); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_4 != __pyx_v_expected_dtype);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/image.pxi":133
 * 
 *         if array.dtype.type is not expected_dtype:
 *             msg = "The '{}' format requires an array of type {}, but {} was passed"             # <<<<<<<<<<<<<<
 *             pix_fmt_name = PixelFormat(pixel_format).name
 *             ex_dtype_name = numpy.dtype(expected_dtype).name
 */
    __Pyx_INCREF(__pyx_kp_s_The_format_requires_an_array_of);
    __pyx_v_msg = __pyx_kp_s_The_format_requires_an_array_of;

    /* "celiagg/image.pxi":134
 *         if array.dtype.type is not expected_dtype:
 *             msg = "The '{}' format requires an array of type {}, but {} was passed"
 *             pix_fmt_name = PixelFormat(pixel_format).name             # <<<<<<<<<<<<<<
 *             ex_dtype_name = numpy.dtype(expected_dtype).name
 *             dtype_name = array.dtype.name
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_pixel_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_pix_fmt_name = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "celiagg/image.pxi":135
 *             msg = "The '{}' format requires an array of type {}, but {} was passed"
 *             pix_fmt_name = PixelFormat(pixel_format).name
 *             ex_dtype_name = numpy.dtype(expected_dtype).name             # <<<<<<<<<<<<<<
 *             dtype_name = array.dtype.name
 *             raise TypeError(msg.format(pix_fmt_name, ex_dtype_name, dtype_name))
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_v_expected_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_ex_dtype_name = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "celiagg/image.pxi":136
 *             pix_fmt_name = PixelFormat(pixel_format).name
 *             ex_dtype_name = numpy.dtype(expected_dtype).name
 *             dtype_name = array.dtype.name             # <<<<<<<<<<<<<<
 *             raise TypeError(msg.format(pix_fmt_name, ex_dtype_name, dtype_name))
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_array, __pyx_n_s_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_dtype_name = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "celiagg/image.pxi":137
 *             ex_dtype_name = numpy.dtype(expected_dtype).name
 *             dtype_name = array.dtype.name
 *             raise TypeError(msg.format(pix_fmt_name, ex_dtype_name, dtype_name))             # <<<<<<<<<<<<<<
 * 
 *         if array.ndim not in (2, 3):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_msg, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_pix_fmt_name, __pyx_v_ex_dtype_name, __pyx_v_dtype_name};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 137, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_v_pix_fmt_name, __pyx_v_ex_dtype_name, __pyx_v_dtype_name};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 137, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_v_pix_fmt_name);
      __Pyx_GIVEREF(__pyx_v_pix_fmt_name);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_v_pix_fmt_name);
      __Pyx_INCREF(__pyx_v_ex_dtype_name);
      __Pyx_GIVEREF(__pyx_v_ex_dtype_name);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_v_ex_dtype_name);
      __Pyx_INCREF(__pyx_v_dtype_name);
      __Pyx_GIVEREF(__pyx_v_dtype_name);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_7, __pyx_v_dtype_name);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 137, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(5, 137, __pyx_L1_error)

    /* "celiagg/image.pxi":132
 *             raise TypeError("An image must be provided in a numpy array")
 * 
 *         if array.dtype.type is not expected_dtype:             # <<<<<<<<<<<<<<
 *             msg = "The '{}' format requires an array of type {}, but {} was passed"
 *             pix_fmt_name = PixelFormat(pixel_format).name
 */
  }

  /* "celiagg/image.pxi":139
 *             raise TypeError(msg.format(pix_fmt_name, ex_dtype_name, dtype_name))
 * 
 *         if array.ndim not in (2, 3):             # <<<<<<<<<<<<<<
 *             raise ValueError("An image array must be 2 or 3 dimensions")
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_array, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyInt_NeObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(5, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyInt_NeObjC(__pyx_t_4, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(5, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_3)) {

    /* "celiagg/image.pxi":140
 * 
 *         if array.ndim not in (2, 3):
 *             raise ValueError("An image array must be 2 or 3 dimensions")             # <<<<<<<<<<<<<<
 * 
 *         last_dim = array.shape[-1] if array.ndim == 3 else 1
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(5, 140, __pyx_L1_error)

    /* "celiagg/image.pxi":139
 *             raise TypeError(msg.format(pix_fmt_name, ex_dtype_name, dtype_name))
 * 
 *         if array.ndim not in (2, 3):             # <<<<<<<<<<<<<<
 *             raise ValueError("An image array must be 2 or 3 dimensions")
 * 
 */
  }

  /* "celiagg/image.pxi":142
 *             raise ValueError("An image array must be 2 or 3 dimensions")
 * 
 *         last_dim = array.shape[-1] if array.ndim == 3 else 1             # <<<<<<<<<<<<<<
 *         if last_dim != expected_dim:
 *             msg = "The '{}' format requires an array of dimension MxNx{}"
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_array, __pyx_n_s_ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_3, 3, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(5, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_array, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_6, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = __pyx_t_1;
    __pyx_t_1 = 0;
  } else {
    __Pyx_INCREF(__pyx_int_1);
    __pyx_t_4 = __pyx_int_1;
  }
  __pyx_v_last_dim = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "celiagg/image.pxi":143
 * 
 *         last_dim = array.shape[-1] if array.ndim == 3 else 1
 *         if last_dim != expected_dim:             # <<<<<<<<<<<<<<
 *             msg = "The '{}' format requires an array of dimension MxNx{}"
 *             pix_fmt_name = PixelFormat(pixel_format).name
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_last_dim, __pyx_v_expected_dim, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 143, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(5, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "celiagg/image.pxi":144
 *         last_dim = array.shape[-1] if array.ndim == 3 else 1
 *         if last_dim != expected_dim:
 *             msg = "The '{}' format requires an array of dimension MxNx{}"             # <<<<<<<<<<<<<<
 *             pix_fmt_name = PixelFormat(pixel_format).name
 *             raise ValueError(msg.format(pix_fmt_name, expected_dim))
 */
    __Pyx_INCREF(__pyx_kp_s_The_format_requires_an_array_of_2);
    __pyx_v_msg = __pyx_kp_s_The_format_requires_an_array_of_2;

    /* "celiagg/image.pxi":145
 *         if last_dim != expected_dim:
 *             msg = "The '{}' format requires an array of dimension MxNx{}"
 *             pix_fmt_name = PixelFormat(pixel_format).name             # <<<<<<<<<<<<<<
 *             raise ValueError(msg.format(pix_fmt_name, expected_dim))
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_pixel_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(5, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_pix_fmt_name = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "celiagg/image.pxi":146
 *             msg = "The '{}' format requires an array of dimension MxNx{}"
 *             pix_fmt_name = PixelFormat(pixel_format).name
 *             raise ValueError(msg.format(pix_fmt_name, expected_dim))             # <<<<<<<<<<<<<<
 * 
 *         self._this = _get_image(array, pixel_format, bottom_up)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_msg, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_pix_fmt_name, __pyx_v_expected_dim};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 146, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_v_pix_fmt_name, __pyx_v_expected_dim};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 146, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(5, 146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_v_pix_fmt_name);
      __Pyx_GIVEREF(__pyx_v_pix_fmt_name);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_7, __pyx_v_pix_fmt_name);
      __Pyx_INCREF(__pyx_v_expected_dim);
      __Pyx_GIVEREF(__pyx_v_expected_dim);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_7, __pyx_v_expected_dim);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(5, 146, __pyx_L1_error)

    /* "celiagg/image.pxi":143
 * 
 *         last_dim = array.shape[-1] if array.ndim == 3 else 1
 *         if last_dim != expected_dim:             # <<<<<<<<<<<<<<
 *             msg = "The '{}' format requires an array of dimension MxNx{}"
 *             pix_fmt_name = PixelFormat(pixel_format).name
 */
  }

  /* "celiagg/image.pxi":148
 *             raise ValueError(msg.format(pix_fmt_name, expected_dim))
 * 
 *         self._this = _get_image(array, pixel_format, bottom_up)             # <<<<<<<<<<<<<<
 *         self.pixel_format = pixel_format
 *         self.pixel_array = array
 */
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_pixel_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(5, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_bottom_up); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_self->_this = __pyx_f_7celiagg_8_celiagg__get_image(__pyx_v_array, __pyx_t_4, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/image.pxi":149
 * 
 *         self._this = _get_image(array, pixel_format, bottom_up)
 *         self.pixel_format = pixel_format             # <<<<<<<<<<<<<<
 *         self.pixel_array = array
 *         self.bottom_up = bottom_up
 */
  __pyx_v_self->pixel_format = __pyx_v_pixel_format;

  /* "celiagg/image.pxi":150
 *         self._this = _get_image(array, pixel_format, bottom_up)
 *         self.pixel_format = pixel_format
 *         self.pixel_array = array             # <<<<<<<<<<<<<<
 *         self.bottom_up = bottom_up
 * 
 */
  __Pyx_INCREF(__pyx_v_array);
  __Pyx_GIVEREF(__pyx_v_array);
  __Pyx_GOTREF(__pyx_v_self->pixel_array);
  __Pyx_DECREF(__pyx_v_self->pixel_array);
  __pyx_v_self->pixel_array = __pyx_v_array;

  /* "celiagg/image.pxi":151
 *         self.pixel_format = pixel_format
 *         self.pixel_array = array
 *         self.bottom_up = bottom_up             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->bottom_up = __pyx_v_bottom_up;

  /* "celiagg/image.pxi":124
 *     cdef bool bottom_up
 * 
 *     def __cinit__(self, array, PixelFormat pixel_format, bool bottom_up=False):             # <<<<<<<<<<<<<<
 *         expected_dtype = _get_format_dtype(pixel_format)
 *         expected_dim = _get_format_last_dim(pixel_format)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("celiagg._celiagg.Image.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_expected_dtype);
  __Pyx_XDECREF(__pyx_v_expected_dim);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_pix_fmt_name);
  __Pyx_XDECREF(__pyx_v_ex_dtype_name);
  __Pyx_XDECREF(__pyx_v_dtype_name);
  __Pyx_XDECREF(__pyx_v_last_dim);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":153
 *         self.bottom_up = bottom_up
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

/* Python wrapper */
static void __pyx_pw_7celiagg_8_celiagg_5Image_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7celiagg_8_celiagg_5Image_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7celiagg_8_celiagg_5Image_2__dealloc__(((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7celiagg_8_celiagg_5Image_2__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "celiagg/image.pxi":154
 * 
 *     def __dealloc__(self):
 *         del self._this             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  delete __pyx_v_self->_this;

  /* "celiagg/image.pxi":153
 *         self.bottom_up = bottom_up
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "celiagg/image.pxi":156
 *         del self._this
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the image.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_5copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_5Image_4copy[] = "Returns a deep copy of the image.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_5copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Image_4copy(((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_4copy(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self) {
  PyObject *__pyx_v_array = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/image.pxi":159
 *         """Returns a deep copy of the image.
 *         """
 *         array = self.pixel_array.copy()             # <<<<<<<<<<<<<<
 *         return Image(array, self.pixel_format, bottom_up=self.bottom_up)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pixel_array, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_array = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/image.pxi":160
 *         """
 *         array = self.pixel_array.copy()
 *         return Image(array, self.pixel_format, bottom_up=self.bottom_up)             # <<<<<<<<<<<<<<
 * 
 *     property format:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_self->pixel_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(5, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_array);
  __Pyx_GIVEREF(__pyx_v_array);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_array);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->bottom_up); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_bottom_up, __pyx_t_3) < 0) __PYX_ERR(5, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_Image), __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(5, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/image.pxi":156
 *         del self._this
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the image.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.Image.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_array);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":163
 * 
 *     property format:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.pixel_format
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_6format_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_6format_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Image_6format___get__(((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_6format___get__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/image.pxi":164
 *     property format:
 *         def __get__(self):
 *             return self.pixel_format             # <<<<<<<<<<<<<<
 * 
 *     property pixels:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_self->pixel_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/image.pxi":163
 * 
 *     property format:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.pixel_format
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Image.format.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":167
 * 
 *     property pixels:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.pixel_array.base is not None:
 *                 self.pixel_array.base
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_6pixels_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_6pixels_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Image_6pixels___get__(((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_6pixels___get__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/image.pxi":168
 *     property pixels:
 *         def __get__(self):
 *             if self.pixel_array.base is not None:             # <<<<<<<<<<<<<<
 *                 self.pixel_array.base
 *             return self.pixel_array
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pixel_array, __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "celiagg/image.pxi":169
 *         def __get__(self):
 *             if self.pixel_array.base is not None:
 *                 self.pixel_array.base             # <<<<<<<<<<<<<<
 *             return self.pixel_array
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->pixel_array, __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "celiagg/image.pxi":168
 *     property pixels:
 *         def __get__(self):
 *             if self.pixel_array.base is not None:             # <<<<<<<<<<<<<<
 *                 self.pixel_array.base
 *             return self.pixel_array
 */
  }

  /* "celiagg/image.pxi":170
 *             if self.pixel_array.base is not None:
 *                 self.pixel_array.base
 *             return self.pixel_array             # <<<<<<<<<<<<<<
 * 
 *     property height:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->pixel_array);
  __pyx_r = __pyx_v_self->pixel_array;
  goto __pyx_L0;

  /* "celiagg/image.pxi":167
 * 
 *     property pixels:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.pixel_array.base is not None:
 *                 self.pixel_array.base
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Image.pixels.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":173
 * 
 *     property height:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.height()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_6height_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_6height_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Image_6height___get__(((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_6height___get__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/image.pxi":174
 *     property height:
 *         def __get__(self):
 *             return self._this.height()             # <<<<<<<<<<<<<<
 * 
 *     property width:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->_this->height()); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/image.pxi":173
 * 
 *     property height:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.height()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Image.height.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/image.pxi":177
 * 
 *     property width:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.width()
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_5width_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_5width_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Image_5width___get__(((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_5width___get__(struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/image.pxi":178
 *     property width:
 *         def __get__(self):
 *             return self._this.width()             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->_this->width()); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/image.pxi":177
 * 
 *     property width:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.width()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Image.width.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Image_6__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Image.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Image_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Image_8__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Image_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Image.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":44
 *     cdef bool bottom_up
 * 
 *     cdef int base_init(self, image, int channel_count, bool has_alpha) except -1:             # <<<<<<<<<<<<<<
 *         if image is None:
 *             raise ValueError('image argument must not be None.')
 */

static int __pyx_f_7celiagg_8_celiagg_10CanvasBase_base_init(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_image, int __pyx_v_channel_count, bool __pyx_v_has_alpha) {
  __Pyx_memviewslice __pyx_v_image_shape = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_image_ndim;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_v_msg_tmp = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  __Pyx_RefNannySetupContext("base_init", 0);

  /* "celiagg/ndarray_canvas.pxi":45
 * 
 *     cdef int base_init(self, image, int channel_count, bool has_alpha) except -1:
 *         if image is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('image argument must not be None.')
 * 
 */
  __pyx_t_1 = (__pyx_v_image == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":46
 *     cdef int base_init(self, image, int channel_count, bool has_alpha) except -1:
 *         if image is None:
 *             raise ValueError('image argument must not be None.')             # <<<<<<<<<<<<<<
 * 
 *         cdef numpy.npy_int32[:] image_shape = numpy.asarray(image.shape,
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 46, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":45
 * 
 *     cdef int base_init(self, image, int channel_count, bool has_alpha) except -1:
 *         if image is None:             # <<<<<<<<<<<<<<
 *             raise ValueError('image argument must not be None.')
 * 
 */
  }

  /* "celiagg/ndarray_canvas.pxi":48
 *             raise ValueError('image argument must not be None.')
 * 
 *         cdef numpy.npy_int32[:] image_shape = numpy.asarray(image.shape,             # <<<<<<<<<<<<<<
 *                                                             dtype=numpy.int32,
 *                                                             order='c')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":49
 * 
 *         cdef numpy.npy_int32[:] image_shape = numpy.asarray(image.shape,
 *                                                             dtype=numpy.int32,             # <<<<<<<<<<<<<<
 *                                                             order='c')
 *         cdef int image_ndim = <int> image.ndim
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_int32); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(6, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(6, 49, __pyx_L1_error)

  /* "celiagg/ndarray_canvas.pxi":48
 *             raise ValueError('image argument must not be None.')
 * 
 *         cdef numpy.npy_int32[:] image_shape = numpy.asarray(image.shape,             # <<<<<<<<<<<<<<
 *                                                             dtype=numpy.int32,
 *                                                             order='c')
 */
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_npy_int32(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(6, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_image_shape = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "celiagg/ndarray_canvas.pxi":51
 *                                                             dtype=numpy.int32,
 *                                                             order='c')
 *         cdef int image_ndim = <int> image.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not has_alpha:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ndim); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_7); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(6, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_image_ndim = ((int)__pyx_t_9);

  /* "celiagg/ndarray_canvas.pxi":53
 *         cdef int image_ndim = <int> image.ndim
 * 
 *         if not has_alpha:             # <<<<<<<<<<<<<<
 *             # Draw colors for images without alpha channels are specified with
 *             # alpha channels and the drawing operations are blended as if all
 */
  __pyx_t_2 = ((!(__pyx_v_has_alpha != 0)) != 0);
  if (__pyx_t_2) {

    /* "celiagg/ndarray_canvas.pxi":59
 *             # channel count is one less than default color channel count if
 *             # image does not have an alpha channel.
 *             channel_count -= 1             # <<<<<<<<<<<<<<
 *         if (channel_count > 1 and (image_ndim != 3
 *                 or image_shape[2] != channel_count) or
 */
    __pyx_v_channel_count = (__pyx_v_channel_count - 1);

    /* "celiagg/ndarray_canvas.pxi":53
 *         cdef int image_ndim = <int> image.ndim
 * 
 *         if not has_alpha:             # <<<<<<<<<<<<<<
 *             # Draw colors for images without alpha channels are specified with
 *             # alpha channels and the drawing operations are blended as if all
 */
  }

  /* "celiagg/ndarray_canvas.pxi":60
 *             # image does not have an alpha channel.
 *             channel_count -= 1
 *         if (channel_count > 1 and (image_ndim != 3             # <<<<<<<<<<<<<<
 *                 or image_shape[2] != channel_count) or
 *                 channel_count == 1 and image_ndim != 2):
 */
  __pyx_t_1 = ((__pyx_v_channel_count > 1) != 0);
  if (!__pyx_t_1) {
    goto __pyx_L7_next_or;
  } else {
  }

  /* "celiagg/ndarray_canvas.pxi":61
 *             channel_count -= 1
 *         if (channel_count > 1 and (image_ndim != 3
 *                 or image_shape[2] != channel_count) or             # <<<<<<<<<<<<<<
 *                 channel_count == 1 and image_ndim != 2):
 *             if channel_count > 1:
 */
  __pyx_t_1 = ((__pyx_v_image_ndim != 3) != 0);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_10 = 2;
  __pyx_t_9 = -1;
  if (__pyx_t_10 < 0) {
    __pyx_t_10 += __pyx_v_image_shape.shape[0];
    if (unlikely(__pyx_t_10 < 0)) __pyx_t_9 = 0;
  } else if (unlikely(__pyx_t_10 >= __pyx_v_image_shape.shape[0])) __pyx_t_9 = 0;
  if (unlikely(__pyx_t_9 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_9);
    __PYX_ERR(6, 61, __pyx_L1_error)
  }
  __pyx_t_1 = (((*((npy_int32 *) ( /* dim=0 */ (__pyx_v_image_shape.data + __pyx_t_10 * __pyx_v_image_shape.strides[0]) ))) != __pyx_v_channel_count) != 0);
  if (!__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_L7_next_or:;

  /* "celiagg/ndarray_canvas.pxi":62
 *         if (channel_count > 1 and (image_ndim != 3
 *                 or image_shape[2] != channel_count) or
 *                 channel_count == 1 and image_ndim != 2):             # <<<<<<<<<<<<<<
 *             if channel_count > 1:
 *                 e = 'MxNx{}'.format(channel_count)
 */
  __pyx_t_1 = ((__pyx_v_channel_count == 1) != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = ((__pyx_v_image_ndim != 2) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L6_bool_binop_done:;

  /* "celiagg/ndarray_canvas.pxi":60
 *             # image does not have an alpha channel.
 *             channel_count -= 1
 *         if (channel_count > 1 and (image_ndim != 3             # <<<<<<<<<<<<<<
 *                 or image_shape[2] != channel_count) or
 *                 channel_count == 1 and image_ndim != 2):
 */
  if (__pyx_t_2) {

    /* "celiagg/ndarray_canvas.pxi":63
 *                 or image_shape[2] != channel_count) or
 *                 channel_count == 1 and image_ndim != 2):
 *             if channel_count > 1:             # <<<<<<<<<<<<<<
 *                 e = 'MxNx{}'.format(channel_count)
 *             else:
 */
    __pyx_t_2 = ((__pyx_v_channel_count > 1) != 0);
    if (__pyx_t_2) {

      /* "celiagg/ndarray_canvas.pxi":64
 *                 channel_count == 1 and image_ndim != 2):
 *             if channel_count > 1:
 *                 e = 'MxNx{}'.format(channel_count)             # <<<<<<<<<<<<<<
 *             else:
 *                 e = 'MxN'
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_MxNx, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_channel_count); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_7 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_e = __pyx_t_7;
      __pyx_t_7 = 0;

      /* "celiagg/ndarray_canvas.pxi":63
 *                 or image_shape[2] != channel_count) or
 *                 channel_count == 1 and image_ndim != 2):
 *             if channel_count > 1:             # <<<<<<<<<<<<<<
 *                 e = 'MxNx{}'.format(channel_count)
 *             else:
 */
      goto __pyx_L11;
    }

    /* "celiagg/ndarray_canvas.pxi":66
 *                 e = 'MxNx{}'.format(channel_count)
 *             else:
 *                 e = 'MxN'             # <<<<<<<<<<<<<<
 *             msg_tmp = 'image argument must be {} channel (ie, must be {}).'
 *             raise ValueError(msg_tmp.format(channel_count, e))
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_n_s_MxN);
      __pyx_v_e = __pyx_n_s_MxN;
    }
    __pyx_L11:;

    /* "celiagg/ndarray_canvas.pxi":67
 *             else:
 *                 e = 'MxN'
 *             msg_tmp = 'image argument must be {} channel (ie, must be {}).'             # <<<<<<<<<<<<<<
 *             raise ValueError(msg_tmp.format(channel_count, e))
 *         # In order to ensure that our backing memory is not deallocated from
 */
    __Pyx_INCREF(__pyx_kp_s_image_argument_must_be_channel_i);
    __pyx_v_msg_tmp = __pyx_kp_s_image_argument_must_be_channel_i;

    /* "celiagg/ndarray_canvas.pxi":68
 *                 e = 'MxN'
 *             msg_tmp = 'image argument must be {} channel (ie, must be {}).'
 *             raise ValueError(msg_tmp.format(channel_count, e))             # <<<<<<<<<<<<<<
 *         # In order to ensure that our backing memory is not deallocated from
 *         # underneath us, we retain a reference to the memory view supplied to
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_msg_tmp, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_channel_count); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_5, __pyx_v_e};
      __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 68, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_5, __pyx_v_e};
      __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 68, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(6, 68, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_e);
      __Pyx_GIVEREF(__pyx_v_e);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_v_e);
      __pyx_t_5 = 0;
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 68, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 68, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":60
 *             # image does not have an alpha channel.
 *             channel_count -= 1
 *         if (channel_count > 1 and (image_ndim != 3             # <<<<<<<<<<<<<<
 *                 or image_shape[2] != channel_count) or
 *                 channel_count == 1 and image_ndim != 2):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":74
 *         # with ndarray_canvas<>'s reference to that memory and destroyed when
 *         # that reference is lost.
 *         self.py_array = image             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(__pyx_v_image);
  __Pyx_GIVEREF(__pyx_v_image);
  __Pyx_GOTREF(__pyx_v_self->py_array);
  __Pyx_DECREF(__pyx_v_self->py_array);
  __pyx_v_self->py_array = __pyx_v_image;

  /* "celiagg/ndarray_canvas.pxi":44
 *     cdef bool bottom_up
 * 
 *     cdef int base_init(self, image, int channel_count, bool has_alpha) except -1:             # <<<<<<<<<<<<<<
 *         if image is None:
 *             raise ValueError('image argument must not be None.')
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.base_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_image_shape, 1);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_msg_tmp);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":76
 *         self.py_array = image
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

/* Python wrapper */
static void __pyx_pw_7celiagg_8_celiagg_10CanvasBase_1__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7celiagg_8_celiagg_10CanvasBase_1__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7celiagg_8_celiagg_10CanvasBase___dealloc__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7celiagg_8_celiagg_10CanvasBase___dealloc__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "celiagg/ndarray_canvas.pxi":77
 * 
 *     def __dealloc__(self):
 *         del self._this             # <<<<<<<<<<<<<<
 * 
 *     property channel_count:
 */
  delete __pyx_v_self->_this;

  /* "celiagg/ndarray_canvas.pxi":76
 *         self.py_array = image
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "celiagg/ndarray_canvas.pxi":80
 * 
 *     property channel_count:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.channel_count()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_13channel_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_13channel_count_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_13channel_count___get__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_13channel_count___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/ndarray_canvas.pxi":81
 *     property channel_count:
 *         def __get__(self):
 *             return self._this.channel_count()             # <<<<<<<<<<<<<<
 * 
 *     property width:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->_this->channel_count()); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/ndarray_canvas.pxi":80
 * 
 *     property channel_count:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.channel_count()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.channel_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":84
 * 
 *     property width:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.width()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_5width_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_5width_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_5width___get__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_5width___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/ndarray_canvas.pxi":85
 *     property width:
 *         def __get__(self):
 *             return self._this.width()             # <<<<<<<<<<<<<<
 * 
 *     property height:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->_this->width()); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/ndarray_canvas.pxi":84
 * 
 *     property width:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.width()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.width.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":88
 * 
 *     property height:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.height()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_6height_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_6height_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_6height___get__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_6height___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/ndarray_canvas.pxi":89
 *     property height:
 *         def __get__(self):
 *             return self._this.height()             # <<<<<<<<<<<<<<
 * 
 *     property array:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->_this->height()); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/ndarray_canvas.pxi":88
 * 
 *     property height:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.height()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.height.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":92
 * 
 *     property array:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             # User is not likely to be concerned with the details of the cython
 *             # memory view around array or buffer supplied to the constructor or
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_5array_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_5array_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_5array___get__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_5array___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/ndarray_canvas.pxi":97
 *             # attach, but instead expects the original array or buffer (which
 *             # is why self.py_array.base is returned rather than self.py_array).
 *             return self.py_array.base             # <<<<<<<<<<<<<<
 * 
 *     property image:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->py_array, __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/ndarray_canvas.pxi":92
 * 
 *     property array:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             # User is not likely to be concerned with the details of the cython
 *             # memory view around array or buffer supplied to the constructor or
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.array.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":100
 * 
 *     property image:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return Image(self.py_array.base, self.pixel_format,
 *                          bottom_up=self.bottom_up)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_5image_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_5image_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_5image___get__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_5image___get__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/ndarray_canvas.pxi":101
 *     property image:
 *         def __get__(self):
 *             return Image(self.py_array.base, self.pixel_format,             # <<<<<<<<<<<<<<
 *                          bottom_up=self.bottom_up)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->py_array, __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_self->pixel_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;

  /* "celiagg/ndarray_canvas.pxi":102
 *         def __get__(self):
 *             return Image(self.py_array.base, self.pixel_format,
 *                          bottom_up=self.bottom_up)             # <<<<<<<<<<<<<<
 * 
 *     def clear(self, double r, double g, double b, double a=1.0):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->bottom_up); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bottom_up, __pyx_t_1) < 0) __PYX_ERR(6, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/ndarray_canvas.pxi":101
 *     property image:
 *         def __get__(self):
 *             return Image(self.py_array.base, self.pixel_format,             # <<<<<<<<<<<<<<
 *                          bottom_up=self.bottom_up)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_Image), __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/ndarray_canvas.pxi":100
 * 
 *     property image:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return Image(self.py_array.base, self.pixel_format,
 *                          bottom_up=self.bottom_up)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.image.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":104
 *                          bottom_up=self.bottom_up)
 * 
 *     def clear(self, double r, double g, double b, double a=1.0):             # <<<<<<<<<<<<<<
 *         """clear(r, g, b, a)
 *         Fill the canvas with a single RGBA value
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_3clear(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_10CanvasBase_2clear[] = "clear(r, g, b, a)\n        Fill the canvas with a single RGBA value\n\n        :param r: Red value in [0, 1]\n        :param g: Green value in [0, 1]\n        :param b: Blue value in [0, 1]\n        :param a: Alpha value in [0, 1] (defaults to 1.0)\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_3clear(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_r;
  double __pyx_v_g;
  double __pyx_v_b;
  double __pyx_v_a;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_r,&__pyx_n_s_g,&__pyx_n_s_b,&__pyx_n_s_a,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_r)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("clear", 0, 3, 4, 1); __PYX_ERR(6, 104, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("clear", 0, 3, 4, 2); __PYX_ERR(6, 104, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "clear") < 0)) __PYX_ERR(6, 104, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_r = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_r == (double)-1) && PyErr_Occurred())) __PYX_ERR(6, 104, __pyx_L3_error)
    __pyx_v_g = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_g == (double)-1) && PyErr_Occurred())) __PYX_ERR(6, 104, __pyx_L3_error)
    __pyx_v_b = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_b == (double)-1) && PyErr_Occurred())) __PYX_ERR(6, 104, __pyx_L3_error)
    if (values[3]) {
      __pyx_v_a = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_a == (double)-1) && PyErr_Occurred())) __PYX_ERR(6, 104, __pyx_L3_error)
    } else {
      __pyx_v_a = ((double)1.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("clear", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 104, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_2clear(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_v_r, __pyx_v_g, __pyx_v_b, __pyx_v_a);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_2clear(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, double __pyx_v_r, double __pyx_v_g, double __pyx_v_b, double __pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear", 0);

  /* "celiagg/ndarray_canvas.pxi":113
 *         :param a: Alpha value in [0, 1] (defaults to 1.0)
 *         """
 *         self._this.clear(r, g, b, a)             # <<<<<<<<<<<<<<
 * 
 *     def draw_image(self, image, fmt, transform, state, bottom_up=False):
 */
  __pyx_v_self->_this->clear(__pyx_v_r, __pyx_v_g, __pyx_v_b, __pyx_v_a);

  /* "celiagg/ndarray_canvas.pxi":104
 *                          bottom_up=self.bottom_up)
 * 
 *     def clear(self, double r, double g, double b, double a=1.0):             # <<<<<<<<<<<<<<
 *         """clear(r, g, b, a)
 *         Fill the canvas with a single RGBA value
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":115
 *         self._this.clear(r, g, b, a)
 * 
 *     def draw_image(self, image, fmt, transform, state, bottom_up=False):             # <<<<<<<<<<<<<<
 *         """draw_image(image, format, transform, state, bottom_up=False)
 *         Draw an image on the canvas.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_5draw_image(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_10CanvasBase_4draw_image[] = "draw_image(image, format, transform, state, bottom_up=False)\n        Draw an image on the canvas.\n\n        :param image: A 2D or 3D numpy array containing image data\n        :param format: A ``PixelFormat`` describing the array's data\n        :param transform: A ``Transform`` object\n        :param state: A ``GraphicsState`` object\n        :param bottom_up: If True, the image data is flipped in the y axis\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_5draw_image(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_fmt = 0;
  PyObject *__pyx_v_transform = 0;
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v_bottom_up = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("draw_image (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_fmt,&__pyx_n_s_transform,&__pyx_n_s_state,&__pyx_n_s_bottom_up,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[4] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fmt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("draw_image", 0, 4, 5, 1); __PYX_ERR(6, 115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_transform)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("draw_image", 0, 4, 5, 2); __PYX_ERR(6, 115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("draw_image", 0, 4, 5, 3); __PYX_ERR(6, 115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bottom_up);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "draw_image") < 0)) __PYX_ERR(6, 115, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = values[0];
    __pyx_v_fmt = values[1];
    __pyx_v_transform = values[2];
    __pyx_v_state = values[3];
    __pyx_v_bottom_up = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("draw_image", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 115, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.draw_image", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_4draw_image(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_v_image, __pyx_v_fmt, __pyx_v_transform, __pyx_v_state, __pyx_v_bottom_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_4draw_image(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_image, PyObject *__pyx_v_fmt, PyObject *__pyx_v_transform, PyObject *__pyx_v_state, PyObject *__pyx_v_bottom_up) {
  struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_gs = 0;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_pix_fmt;
  struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_trans = 0;
  struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_img = 0;
  struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_input_img = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_t_5;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("draw_image", 0);

  /* "celiagg/ndarray_canvas.pxi":125
 *         :param bottom_up: If True, the image data is flipped in the y axis
 *         """
 *         if not isinstance(image, (numpy.ndarray, Image)):             # <<<<<<<<<<<<<<
 *             raise TypeError("image must be an ndarray or Image instance")
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_image, __pyx_ptype_5numpy_ndarray); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_image, __pyx_ptype_7celiagg_8_celiagg_Image); 
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":126
 *         """
 *         if not isinstance(image, (numpy.ndarray, Image)):
 *             raise TypeError("image must be an ndarray or Image instance")             # <<<<<<<<<<<<<<
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):
 *             raise TypeError("format must be a PixelFormat value")
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(6, 126, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":125
 *         :param bottom_up: If True, the image data is flipped in the y axis
 *         """
 *         if not isinstance(image, (numpy.ndarray, Image)):             # <<<<<<<<<<<<<<
 *             raise TypeError("image must be an ndarray or Image instance")
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":127
 *         if not isinstance(image, (numpy.ndarray, Image)):
 *             raise TypeError("image must be an ndarray or Image instance")
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):             # <<<<<<<<<<<<<<
 *             raise TypeError("format must be a PixelFormat value")
 *         if not isinstance(transform, Transform):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyObject_IsInstance(__pyx_v_fmt, __pyx_t_4); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(6, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_image, __pyx_ptype_5numpy_ndarray); 
  __pyx_t_1 = (__pyx_t_3 != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":128
 *             raise TypeError("image must be an ndarray or Image instance")
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):
 *             raise TypeError("format must be a PixelFormat value")             # <<<<<<<<<<<<<<
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(6, 128, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":127
 *         if not isinstance(image, (numpy.ndarray, Image)):
 *             raise TypeError("image must be an ndarray or Image instance")
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):             # <<<<<<<<<<<<<<
 *             raise TypeError("format must be a PixelFormat value")
 *         if not isinstance(transform, Transform):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":129
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):
 *             raise TypeError("format must be a PixelFormat value")
 *         if not isinstance(transform, Transform):             # <<<<<<<<<<<<<<
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_transform, __pyx_ptype_7celiagg_8_celiagg_Transform); 
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "celiagg/ndarray_canvas.pxi":130
 *             raise TypeError("format must be a PixelFormat value")
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")             # <<<<<<<<<<<<<<
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(6, 130, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":129
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):
 *             raise TypeError("format must be a PixelFormat value")
 *         if not isinstance(transform, Transform):             # <<<<<<<<<<<<<<
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":131
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):             # <<<<<<<<<<<<<<
 *             raise TypeError("state must be a GraphicsState instance")
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_state, __pyx_ptype_7celiagg_8_celiagg_GraphicsState); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":132
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")             # <<<<<<<<<<<<<<
 * 
 *         cdef GraphicsState gs = <GraphicsState>state
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(6, 132, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":131
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):             # <<<<<<<<<<<<<<
 *             raise TypeError("state must be a GraphicsState instance")
 * 
 */
  }

  /* "celiagg/ndarray_canvas.pxi":134
 *             raise TypeError("state must be a GraphicsState instance")
 * 
 *         cdef GraphicsState gs = <GraphicsState>state             # <<<<<<<<<<<<<<
 *         cdef PixelFormat pix_fmt = <PixelFormat>image.format if fmt is None else fmt
 *         cdef Transform trans = <Transform>transform
 */
  __pyx_t_4 = __pyx_v_state;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_gs = ((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "celiagg/ndarray_canvas.pxi":135
 * 
 *         cdef GraphicsState gs = <GraphicsState>state
 *         cdef PixelFormat pix_fmt = <PixelFormat>image.format if fmt is None else fmt             # <<<<<<<<<<<<<<
 *         cdef Transform trans = <Transform>transform
 *         cdef Image img
 */
  __pyx_t_2 = (__pyx_v_fmt == Py_None);
  if ((__pyx_t_2 != 0)) {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_t_4)); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 135, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)__pyx_t_6);
  } else {
    __pyx_t_6 = ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_fmt)); if (unlikely(PyErr_Occurred())) __PYX_ERR(6, 135, __pyx_L1_error)
    __pyx_t_5 = __pyx_t_6;
  }
  __pyx_v_pix_fmt = __pyx_t_5;

  /* "celiagg/ndarray_canvas.pxi":136
 *         cdef GraphicsState gs = <GraphicsState>state
 *         cdef PixelFormat pix_fmt = <PixelFormat>image.format if fmt is None else fmt
 *         cdef Transform trans = <Transform>transform             # <<<<<<<<<<<<<<
 *         cdef Image img
 *         cdef Image input_img
 */
  __pyx_t_4 = __pyx_v_transform;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_trans = ((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "celiagg/ndarray_canvas.pxi":140
 *         cdef Image input_img
 * 
 *         self._check_stencil(gs)             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(image, Image):
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self->__pyx_vtab)->_check_stencil(__pyx_v_self, __pyx_v_gs); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "celiagg/ndarray_canvas.pxi":142
 *         self._check_stencil(gs)
 * 
 *         if isinstance(image, Image):             # <<<<<<<<<<<<<<
 *             input_img = image
 *         else:
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_image, __pyx_ptype_7celiagg_8_celiagg_Image); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "celiagg/ndarray_canvas.pxi":143
 * 
 *         if isinstance(image, Image):
 *             input_img = image             # <<<<<<<<<<<<<<
 *         else:
 *             input_img = Image(image, pix_fmt, bottom_up=bottom_up)
 */
    if (!(likely(((__pyx_v_image) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_image, __pyx_ptype_7celiagg_8_celiagg_Image))))) __PYX_ERR(6, 143, __pyx_L1_error)
    __pyx_t_4 = __pyx_v_image;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_v_input_img = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "celiagg/ndarray_canvas.pxi":142
 *         self._check_stencil(gs)
 * 
 *         if isinstance(image, Image):             # <<<<<<<<<<<<<<
 *             input_img = image
 *         else:
 */
    goto __pyx_L11;
  }

  /* "celiagg/ndarray_canvas.pxi":145
 *             input_img = image
 *         else:
 *             input_img = Image(image, pix_fmt, bottom_up=bottom_up)             # <<<<<<<<<<<<<<
 * 
 *         img = self._get_native_image(input_img, self.pixel_format)
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_pix_fmt); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_v_image);
    __Pyx_GIVEREF(__pyx_v_image);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_image);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_bottom_up, __pyx_v_bottom_up) < 0) __PYX_ERR(6, 145, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_Image), __pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_input_img = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_t_8);
    __pyx_t_8 = 0;
  }
  __pyx_L11:;

  /* "celiagg/ndarray_canvas.pxi":147
 *             input_img = Image(image, pix_fmt, bottom_up=bottom_up)
 * 
 *         img = self._get_native_image(input_img, self.pixel_format)             # <<<<<<<<<<<<<<
 *         self._this.draw_image(dereference(img._this),
 *                               dereference(trans._this),
 */
  __pyx_t_8 = ((PyObject *)((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self->__pyx_vtab)->_get_native_image(__pyx_v_self, __pyx_v_input_img, __pyx_v_self->pixel_format)); if (unlikely(!__pyx_t_8)) __PYX_ERR(6, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_v_img = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "celiagg/ndarray_canvas.pxi":148
 * 
 *         img = self._get_native_image(input_img, self.pixel_format)
 *         self._this.draw_image(dereference(img._this),             # <<<<<<<<<<<<<<
 *                               dereference(trans._this),
 *                               dereference(gs._this))
 */
  __pyx_v_self->_this->draw_image((*__pyx_v_img->_this), (*__pyx_v_trans->_this), (*__pyx_v_gs->_this));

  /* "celiagg/ndarray_canvas.pxi":115
 *         self._this.clear(r, g, b, a)
 * 
 *     def draw_image(self, image, fmt, transform, state, bottom_up=False):             # <<<<<<<<<<<<<<
 *         """draw_image(image, format, transform, state, bottom_up=False)
 *         Draw an image on the canvas.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.draw_image", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_gs);
  __Pyx_XDECREF((PyObject *)__pyx_v_trans);
  __Pyx_XDECREF((PyObject *)__pyx_v_img);
  __Pyx_XDECREF((PyObject *)__pyx_v_input_img);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":152
 *                               dereference(gs._this))
 * 
 *     def draw_shape(self, shape, transform, state, stroke=None, fill=None):             # <<<<<<<<<<<<<<
 *         """draw_shape(shape, transform, state, stroke=SolidColor(0, 0, 0), fill=SolidColor(0, 0, 0))
 *         Draw a shape on the canvas.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_7draw_shape(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_10CanvasBase_6draw_shape[] = "draw_shape(shape, transform, state, stroke=SolidColor(0, 0, 0), fill=SolidColor(0, 0, 0))\n        Draw a shape on the canvas.\n\n        .. note::\n           Use ``GraphicsState.drawing_mode`` to enable/disable stroke or fill\n           drawing.\n\n        :param shape: A ``VertexSource`` object\n        :param transform: A ``Transform`` object\n        :param state: A ``GraphicsState`` object\n        :param stroke: The ``Paint`` to use for outlines. Defaults to black.\n        :param fill: The ``Paint`` to use for fills. Defaults to black.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_7draw_shape(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  PyObject *__pyx_v_transform = 0;
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v_stroke = 0;
  PyObject *__pyx_v_fill = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("draw_shape (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_transform,&__pyx_n_s_state,&__pyx_n_s_stroke,&__pyx_n_s_fill,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_transform)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("draw_shape", 0, 3, 5, 1); __PYX_ERR(6, 152, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("draw_shape", 0, 3, 5, 2); __PYX_ERR(6, 152, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stroke);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "draw_shape") < 0)) __PYX_ERR(6, 152, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = values[0];
    __pyx_v_transform = values[1];
    __pyx_v_state = values[2];
    __pyx_v_stroke = values[3];
    __pyx_v_fill = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("draw_shape", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 152, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.draw_shape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_6draw_shape(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_v_shape, __pyx_v_transform, __pyx_v_state, __pyx_v_stroke, __pyx_v_fill);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_6draw_shape(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_shape, PyObject *__pyx_v_transform, PyObject *__pyx_v_state, PyObject *__pyx_v_stroke, PyObject *__pyx_v_fill) {
  struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_shp = 0;
  struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_gs = 0;
  struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_trans = 0;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_fmt;
  struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_stroke_paint = 0;
  struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_fill_paint = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_t_5;
  __Pyx_RefNannySetupContext("draw_shape", 0);

  /* "celiagg/ndarray_canvas.pxi":166
 *         :param fill: The ``Paint`` to use for fills. Defaults to black.
 *         """
 *         if not isinstance(shape, VertexSource):             # <<<<<<<<<<<<<<
 *             raise TypeError("shape must be a VertexSource (Path, BSpline, etc)")
 *         if not isinstance(transform, Transform):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_shape, __pyx_ptype_7celiagg_8_celiagg_VertexSource); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":167
 *         """
 *         if not isinstance(shape, VertexSource):
 *             raise TypeError("shape must be a VertexSource (Path, BSpline, etc)")             # <<<<<<<<<<<<<<
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 167, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":166
 *         :param fill: The ``Paint`` to use for fills. Defaults to black.
 *         """
 *         if not isinstance(shape, VertexSource):             # <<<<<<<<<<<<<<
 *             raise TypeError("shape must be a VertexSource (Path, BSpline, etc)")
 *         if not isinstance(transform, Transform):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":168
 *         if not isinstance(shape, VertexSource):
 *             raise TypeError("shape must be a VertexSource (Path, BSpline, etc)")
 *         if not isinstance(transform, Transform):             # <<<<<<<<<<<<<<
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_transform, __pyx_ptype_7celiagg_8_celiagg_Transform); 
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "celiagg/ndarray_canvas.pxi":169
 *             raise TypeError("shape must be a VertexSource (Path, BSpline, etc)")
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")             # <<<<<<<<<<<<<<
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 169, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":168
 *         if not isinstance(shape, VertexSource):
 *             raise TypeError("shape must be a VertexSource (Path, BSpline, etc)")
 *         if not isinstance(transform, Transform):             # <<<<<<<<<<<<<<
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":170
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):             # <<<<<<<<<<<<<<
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_state, __pyx_ptype_7celiagg_8_celiagg_GraphicsState); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":171
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")             # <<<<<<<<<<<<<<
 *         if stroke is not None and not isinstance(stroke, Paint):
 *             raise TypeError("stroke must be a Paint instance")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 171, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":170
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):             # <<<<<<<<<<<<<<
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":172
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):             # <<<<<<<<<<<<<<
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):
 */
  __pyx_t_1 = (__pyx_v_stroke != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_stroke, __pyx_ptype_7celiagg_8_celiagg_Paint); 
  __pyx_t_1 = ((!(__pyx_t_4 != 0)) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":173
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):
 *             raise TypeError("stroke must be a Paint instance")             # <<<<<<<<<<<<<<
 *         if fill is not None and not isinstance(fill, Paint):
 *             raise TypeError("fill must be a Paint instance")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 173, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":172
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):             # <<<<<<<<<<<<<<
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":174
 *         if stroke is not None and not isinstance(stroke, Paint):
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):             # <<<<<<<<<<<<<<
 *             raise TypeError("fill must be a Paint instance")
 * 
 */
  __pyx_t_1 = (__pyx_v_fill != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_fill, __pyx_ptype_7celiagg_8_celiagg_Paint); 
  __pyx_t_1 = ((!(__pyx_t_4 != 0)) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L10_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":175
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):
 *             raise TypeError("fill must be a Paint instance")             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 175, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":174
 *         if stroke is not None and not isinstance(stroke, Paint):
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):             # <<<<<<<<<<<<<<
 *             raise TypeError("fill must be a Paint instance")
 * 
 */
  }

  /* "celiagg/ndarray_canvas.pxi":178
 * 
 *         cdef:
 *             VertexSource shp = <VertexSource>shape             # <<<<<<<<<<<<<<
 *             GraphicsState gs = <GraphicsState>state
 *             Transform trans = <Transform>transform
 */
  __pyx_t_3 = __pyx_v_shape;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_shp = ((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":179
 *         cdef:
 *             VertexSource shp = <VertexSource>shape
 *             GraphicsState gs = <GraphicsState>state             # <<<<<<<<<<<<<<
 *             Transform trans = <Transform>transform
 *             PixelFormat fmt = self.pixel_format
 */
  __pyx_t_3 = __pyx_v_state;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_gs = ((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":180
 *             VertexSource shp = <VertexSource>shape
 *             GraphicsState gs = <GraphicsState>state
 *             Transform trans = <Transform>transform             # <<<<<<<<<<<<<<
 *             PixelFormat fmt = self.pixel_format
 *             Paint stroke_paint
 */
  __pyx_t_3 = __pyx_v_transform;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_trans = ((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":181
 *             GraphicsState gs = <GraphicsState>state
 *             Transform trans = <Transform>transform
 *             PixelFormat fmt = self.pixel_format             # <<<<<<<<<<<<<<
 *             Paint stroke_paint
 *             Paint fill_paint
 */
  __pyx_t_5 = __pyx_v_self->pixel_format;
  __pyx_v_fmt = __pyx_t_5;

  /* "celiagg/ndarray_canvas.pxi":185
 *             Paint fill_paint
 * 
 *         self._check_stencil(gs)             # <<<<<<<<<<<<<<
 *         stroke_paint = self._get_native_paint(stroke, fmt)
 *         fill_paint = self._get_native_paint(fill, fmt)
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self->__pyx_vtab)->_check_stencil(__pyx_v_self, __pyx_v_gs); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":186
 * 
 *         self._check_stencil(gs)
 *         stroke_paint = self._get_native_paint(stroke, fmt)             # <<<<<<<<<<<<<<
 *         fill_paint = self._get_native_paint(fill, fmt)
 * 
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self->__pyx_vtab)->_get_native_paint(__pyx_v_self, __pyx_v_stroke, __pyx_v_fmt)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_stroke_paint = ((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":187
 *         self._check_stencil(gs)
 *         stroke_paint = self._get_native_paint(stroke, fmt)
 *         fill_paint = self._get_native_paint(fill, fmt)             # <<<<<<<<<<<<<<
 * 
 *         self._this.draw_shape(dereference(shp._this),
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self->__pyx_vtab)->_get_native_paint(__pyx_v_self, __pyx_v_fill, __pyx_v_fmt)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_fill_paint = ((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":189
 *         fill_paint = self._get_native_paint(fill, fmt)
 * 
 *         self._this.draw_shape(dereference(shp._this),             # <<<<<<<<<<<<<<
 *                               dereference(trans._this),
 *                               dereference(stroke_paint._this),
 */
  __pyx_v_self->_this->draw_shape((*__pyx_v_shp->_this), (*__pyx_v_trans->_this), (*__pyx_v_stroke_paint->_this), (*__pyx_v_fill_paint->_this), (*__pyx_v_gs->_this));

  /* "celiagg/ndarray_canvas.pxi":152
 *                               dereference(gs._this))
 * 
 *     def draw_shape(self, shape, transform, state, stroke=None, fill=None):             # <<<<<<<<<<<<<<
 *         """draw_shape(shape, transform, state, stroke=SolidColor(0, 0, 0), fill=SolidColor(0, 0, 0))
 *         Draw a shape on the canvas.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.draw_shape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_shp);
  __Pyx_XDECREF((PyObject *)__pyx_v_gs);
  __Pyx_XDECREF((PyObject *)__pyx_v_trans);
  __Pyx_XDECREF((PyObject *)__pyx_v_stroke_paint);
  __Pyx_XDECREF((PyObject *)__pyx_v_fill_paint);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":195
 *                               dereference(gs._this))
 * 
 *     def draw_text(self, text, font, transform, state, stroke=None, fill=None):             # <<<<<<<<<<<<<<
 *         """draw_text(text, font, transform, state, stroke=SolidColor(0, 0, 0), fill=SolidColor(0, 0, 0))
 *         Draw a line of text on the canvas.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_9draw_text(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_10CanvasBase_8draw_text[] = "draw_text(text, font, transform, state, stroke=SolidColor(0, 0, 0), fill=SolidColor(0, 0, 0))\n        Draw a line of text on the canvas.\n\n        .. note::\n           Use ``GraphicsState.drawing_mode`` to enable/disable stroke or fill\n           drawing.\n\n        :param text: A Unicode string of text to be renderered. Newlines will\n                     be ignored.\n        :param font: A ``Font`` object\n        :param transform: A ``Transform`` object\n        :param state: A ``GraphicsState`` object\n        :param stroke: The ``Paint`` to use for outlines. Defaults to black.\n        :param fill: The ``Paint`` to use for fills. Defaults to black.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_9draw_text(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_text = 0;
  PyObject *__pyx_v_font = 0;
  PyObject *__pyx_v_transform = 0;
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v_stroke = 0;
  PyObject *__pyx_v_fill = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("draw_text (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_text,&__pyx_n_s_font,&__pyx_n_s_transform,&__pyx_n_s_state,&__pyx_n_s_stroke,&__pyx_n_s_fill,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_text)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_font)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("draw_text", 0, 4, 6, 1); __PYX_ERR(6, 195, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_transform)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("draw_text", 0, 4, 6, 2); __PYX_ERR(6, 195, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("draw_text", 0, 4, 6, 3); __PYX_ERR(6, 195, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stroke);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fill);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "draw_text") < 0)) __PYX_ERR(6, 195, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_text = values[0];
    __pyx_v_font = values[1];
    __pyx_v_transform = values[2];
    __pyx_v_state = values[3];
    __pyx_v_stroke = values[4];
    __pyx_v_fill = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("draw_text", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 195, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.draw_text", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_8draw_text(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_v_text, __pyx_v_font, __pyx_v_transform, __pyx_v_state, __pyx_v_stroke, __pyx_v_fill);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_8draw_text(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_font, PyObject *__pyx_v_transform, PyObject *__pyx_v_state, PyObject *__pyx_v_stroke, PyObject *__pyx_v_fill) {
  struct __pyx_obj_7celiagg_8_celiagg_Font *__pyx_v_fnt = 0;
  struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_gs = 0;
  struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_trans = 0;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_fmt;
  struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_stroke_paint = 0;
  struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_fill_paint = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_t_5;
  char const *__pyx_t_6;
  __Pyx_RefNannySetupContext("draw_text", 0);
  __Pyx_INCREF(__pyx_v_text);

  /* "celiagg/ndarray_canvas.pxi":217
 *             raise RuntimeError(msg)
 * 
 *         if not isinstance(font, Font):             # <<<<<<<<<<<<<<
 *             raise TypeError("font must be a Font instance")
 *         if not isinstance(transform, Transform):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_font, __pyx_ptype_7celiagg_8_celiagg_Font); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":218
 * 
 *         if not isinstance(font, Font):
 *             raise TypeError("font must be a Font instance")             # <<<<<<<<<<<<<<
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 218, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":217
 *             raise RuntimeError(msg)
 * 
 *         if not isinstance(font, Font):             # <<<<<<<<<<<<<<
 *             raise TypeError("font must be a Font instance")
 *         if not isinstance(transform, Transform):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":219
 *         if not isinstance(font, Font):
 *             raise TypeError("font must be a Font instance")
 *         if not isinstance(transform, Transform):             # <<<<<<<<<<<<<<
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_transform, __pyx_ptype_7celiagg_8_celiagg_Transform); 
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "celiagg/ndarray_canvas.pxi":220
 *             raise TypeError("font must be a Font instance")
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")             # <<<<<<<<<<<<<<
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 220, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":219
 *         if not isinstance(font, Font):
 *             raise TypeError("font must be a Font instance")
 *         if not isinstance(transform, Transform):             # <<<<<<<<<<<<<<
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":221
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):             # <<<<<<<<<<<<<<
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_state, __pyx_ptype_7celiagg_8_celiagg_GraphicsState); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":222
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")             # <<<<<<<<<<<<<<
 *         if stroke is not None and not isinstance(stroke, Paint):
 *             raise TypeError("stroke must be a Paint instance")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 222, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":221
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):             # <<<<<<<<<<<<<<
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":223
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):             # <<<<<<<<<<<<<<
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):
 */
  __pyx_t_1 = (__pyx_v_stroke != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_stroke, __pyx_ptype_7celiagg_8_celiagg_Paint); 
  __pyx_t_1 = ((!(__pyx_t_4 != 0)) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":224
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):
 *             raise TypeError("stroke must be a Paint instance")             # <<<<<<<<<<<<<<
 *         if fill is not None and not isinstance(fill, Paint):
 *             raise TypeError("fill must be a Paint instance")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 224, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":223
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):             # <<<<<<<<<<<<<<
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):
 */
  }

  /* "celiagg/ndarray_canvas.pxi":225
 *         if stroke is not None and not isinstance(stroke, Paint):
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):             # <<<<<<<<<<<<<<
 *             raise TypeError("fill must be a Paint instance")
 * 
 */
  __pyx_t_1 = (__pyx_v_fill != Py_None);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_TypeCheck(__pyx_v_fill, __pyx_ptype_7celiagg_8_celiagg_Paint); 
  __pyx_t_1 = ((!(__pyx_t_4 != 0)) != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L10_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/ndarray_canvas.pxi":226
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):
 *             raise TypeError("fill must be a Paint instance")             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(6, 226, __pyx_L1_error)

    /* "celiagg/ndarray_canvas.pxi":225
 *         if stroke is not None and not isinstance(stroke, Paint):
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):             # <<<<<<<<<<<<<<
 *             raise TypeError("fill must be a Paint instance")
 * 
 */
  }

  /* "celiagg/ndarray_canvas.pxi":229
 * 
 *         cdef:
 *             Font fnt = <Font>font             # <<<<<<<<<<<<<<
 *             GraphicsState gs = <GraphicsState>state
 *             Transform trans = <Transform>transform
 */
  __pyx_t_3 = __pyx_v_font;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_fnt = ((struct __pyx_obj_7celiagg_8_celiagg_Font *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":230
 *         cdef:
 *             Font fnt = <Font>font
 *             GraphicsState gs = <GraphicsState>state             # <<<<<<<<<<<<<<
 *             Transform trans = <Transform>transform
 *             PixelFormat fmt = self.pixel_format
 */
  __pyx_t_3 = __pyx_v_state;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_gs = ((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":231
 *             Font fnt = <Font>font
 *             GraphicsState gs = <GraphicsState>state
 *             Transform trans = <Transform>transform             # <<<<<<<<<<<<<<
 *             PixelFormat fmt = self.pixel_format
 *             Paint stroke_paint
 */
  __pyx_t_3 = __pyx_v_transform;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_trans = ((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":232
 *             GraphicsState gs = <GraphicsState>state
 *             Transform trans = <Transform>transform
 *             PixelFormat fmt = self.pixel_format             # <<<<<<<<<<<<<<
 *             Paint stroke_paint
 *             Paint fill_paint
 */
  __pyx_t_5 = __pyx_v_self->pixel_format;
  __pyx_v_fmt = __pyx_t_5;

  /* "celiagg/ndarray_canvas.pxi":236
 *             Paint fill_paint
 * 
 *         self._check_stencil(gs)             # <<<<<<<<<<<<<<
 *         stroke_paint = self._get_native_paint(stroke, fmt)
 *         fill_paint = self._get_native_paint(fill, fmt)
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self->__pyx_vtab)->_check_stencil(__pyx_v_self, __pyx_v_gs); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":237
 * 
 *         self._check_stencil(gs)
 *         stroke_paint = self._get_native_paint(stroke, fmt)             # <<<<<<<<<<<<<<
 *         fill_paint = self._get_native_paint(fill, fmt)
 * 
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self->__pyx_vtab)->_get_native_paint(__pyx_v_self, __pyx_v_stroke, __pyx_v_fmt)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_stroke_paint = ((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":238
 *         self._check_stencil(gs)
 *         stroke_paint = self._get_native_paint(stroke, fmt)
 *         fill_paint = self._get_native_paint(fill, fmt)             # <<<<<<<<<<<<<<
 * 
 *         text = _get_utf8_text(text, "The text argument must be unicode.")
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self->__pyx_vtab)->_get_native_paint(__pyx_v_self, __pyx_v_fill, __pyx_v_fmt)); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_fill_paint = ((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":240
 *         fill_paint = self._get_native_paint(fill, fmt)
 * 
 *         text = _get_utf8_text(text, "The text argument must be unicode.")             # <<<<<<<<<<<<<<
 *         self._this.draw_text(text, dereference(fnt._this),
 *                              dereference(trans._this),
 */
  __pyx_t_3 = __pyx_f_7celiagg_8_celiagg__get_utf8_text(__pyx_v_text, __pyx_kp_s_The_text_argument_must_be_unicod); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF_SET(__pyx_v_text, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/ndarray_canvas.pxi":241
 * 
 *         text = _get_utf8_text(text, "The text argument must be unicode.")
 *         self._this.draw_text(text, dereference(fnt._this),             # <<<<<<<<<<<<<<
 *                              dereference(trans._this),
 *                              dereference(stroke_paint._this),
 */
  __pyx_t_6 = __Pyx_PyObject_AsString(__pyx_v_text); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(6, 241, __pyx_L1_error)

  /* "celiagg/ndarray_canvas.pxi":245
 *                              dereference(stroke_paint._this),
 *                              dereference(fill_paint._this),
 *                              dereference(gs._this))             # <<<<<<<<<<<<<<
 * 
 *     cdef _check_stencil(self, GraphicsState state):
 */
  __pyx_v_self->_this->draw_text(__pyx_t_6, (*__pyx_v_fnt->_this), (*__pyx_v_trans->_this), (*__pyx_v_stroke_paint->_this), (*__pyx_v_fill_paint->_this), (*__pyx_v_gs->_this));

  /* "celiagg/ndarray_canvas.pxi":195
 *                               dereference(gs._this))
 * 
 *     def draw_text(self, text, font, transform, state, stroke=None, fill=None):             # <<<<<<<<<<<<<<
 *         """draw_text(text, font, transform, state, stroke=SolidColor(0, 0, 0), fill=SolidColor(0, 0, 0))
 *         Draw a line of text on the canvas.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.draw_text", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_fnt);
  __Pyx_XDECREF((PyObject *)__pyx_v_gs);
  __Pyx_XDECREF((PyObject *)__pyx_v_trans);
  __Pyx_XDECREF((PyObject *)__pyx_v_stroke_paint);
  __Pyx_XDECREF((PyObject *)__pyx_v_fill_paint);
  __Pyx_XDECREF(__pyx_v_text);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":247
 *                              dereference(gs._this))
 * 
 *     cdef _check_stencil(self, GraphicsState state):             # <<<<<<<<<<<<<<
 *         """Internal. Checks if a stencil's dimensions match those of the
 *         canvas.
 */

static PyObject *__pyx_f_7celiagg_8_celiagg_10CanvasBase__check_stencil(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *__pyx_v_state) {
  struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_stencil = 0;
  PyObject *__pyx_v_w = NULL;
  PyObject *__pyx_v_h = NULL;
  PyObject *__pyx_v_sw = NULL;
  PyObject *__pyx_v_sh = NULL;
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_check_stencil", 0);

  /* "celiagg/ndarray_canvas.pxi":251
 *         canvas.
 *         """
 *         cdef Image stencil = state.stencil             # <<<<<<<<<<<<<<
 *         if stencil is not None:
 *             w, h = self.width, self.height
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_state), __pyx_n_s_stencil); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7celiagg_8_celiagg_Image))))) __PYX_ERR(6, 251, __pyx_L1_error)
  __pyx_v_stencil = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/ndarray_canvas.pxi":252
 *         """
 *         cdef Image stencil = state.stencil
 *         if stencil is not None:             # <<<<<<<<<<<<<<
 *             w, h = self.width, self.height
 *             sw, sh = stencil.width, stencil.height
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_stencil) != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "celiagg/ndarray_canvas.pxi":253
 *         cdef Image stencil = state.stencil
 *         if stencil is not None:
 *             w, h = self.width, self.height             # <<<<<<<<<<<<<<
 *             sw, sh = stencil.width, stencil.height
 *             if sw != w or sh != h:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_width); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_height); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_w = __pyx_t_1;
    __pyx_t_1 = 0;
    __pyx_v_h = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "celiagg/ndarray_canvas.pxi":254
 *         if stencil is not None:
 *             w, h = self.width, self.height
 *             sw, sh = stencil.width, stencil.height             # <<<<<<<<<<<<<<
 *             if sw != w or sh != h:
 *                 msg = ("The stencil set on the GraphicsState object does not "
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_stencil), __pyx_n_s_width); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_stencil), __pyx_n_s_height); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_sw = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_v_sh = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "celiagg/ndarray_canvas.pxi":255
 *             w, h = self.width, self.height
 *             sw, sh = stencil.width, stencil.height
 *             if sw != w or sh != h:             # <<<<<<<<<<<<<<
 *                 msg = ("The stencil set on the GraphicsState object does not "
 *                        "have the same dimensions as the target canvas! "
 */
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_sw, __pyx_v_w, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 255, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(6, 255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_sh, __pyx_v_h, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 255, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(6, 255, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;
    if (unlikely(__pyx_t_3)) {

      /* "celiagg/ndarray_canvas.pxi":258
 *                 msg = ("The stencil set on the GraphicsState object does not "
 *                        "have the same dimensions as the target canvas! "
 *                        "(({}, {}) vs. ({}, {}))").format(sw, sh, w, h)             # <<<<<<<<<<<<<<
 *                 raise AggError(msg)
 * 
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_The_stencil_set_on_the_GraphicsS, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_sw, __pyx_v_sh, __pyx_v_w, __pyx_v_h};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 258, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[5] = {__pyx_t_5, __pyx_v_sw, __pyx_v_sh, __pyx_v_w, __pyx_v_h};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 4+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 258, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(4+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(6, 258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_INCREF(__pyx_v_sw);
        __Pyx_GIVEREF(__pyx_v_sw);
        PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_v_sw);
        __Pyx_INCREF(__pyx_v_sh);
        __Pyx_GIVEREF(__pyx_v_sh);
        PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_sh);
        __Pyx_INCREF(__pyx_v_w);
        __Pyx_GIVEREF(__pyx_v_w);
        PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_v_w);
        __Pyx_INCREF(__pyx_v_h);
        __Pyx_GIVEREF(__pyx_v_h);
        PyTuple_SET_ITEM(__pyx_t_7, 3+__pyx_t_6, __pyx_v_h);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 258, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_msg = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "celiagg/ndarray_canvas.pxi":259
 *                        "have the same dimensions as the target canvas! "
 *                        "(({}, {}) vs. ({}, {}))").format(sw, sh, w, h)
 *                 raise AggError(msg)             # <<<<<<<<<<<<<<
 * 
 *     cdef Image _get_native_image(self, Image image, PixelFormat fmt):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_AggError); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_7, __pyx_v_msg) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_msg);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(6, 259, __pyx_L1_error)

      /* "celiagg/ndarray_canvas.pxi":255
 *             w, h = self.width, self.height
 *             sw, sh = stencil.width, stencil.height
 *             if sw != w or sh != h:             # <<<<<<<<<<<<<<
 *                 msg = ("The stencil set on the GraphicsState object does not "
 *                        "have the same dimensions as the target canvas! "
 */
    }

    /* "celiagg/ndarray_canvas.pxi":252
 *         """
 *         cdef Image stencil = state.stencil
 *         if stencil is not None:             # <<<<<<<<<<<<<<
 *             w, h = self.width, self.height
 *             sw, sh = stencil.width, stencil.height
 */
  }

  /* "celiagg/ndarray_canvas.pxi":247
 *                              dereference(gs._this))
 * 
 *     cdef _check_stencil(self, GraphicsState state):             # <<<<<<<<<<<<<<
 *         """Internal. Checks if a stencil's dimensions match those of the
 *         canvas.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase._check_stencil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_stencil);
  __Pyx_XDECREF(__pyx_v_w);
  __Pyx_XDECREF(__pyx_v_h);
  __Pyx_XDECREF(__pyx_v_sw);
  __Pyx_XDECREF(__pyx_v_sh);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":261
 *                 raise AggError(msg)
 * 
 *     cdef Image _get_native_image(self, Image image, PixelFormat fmt):             # <<<<<<<<<<<<<<
 *         """_get_native_image(image, format)
 * 
 */

static struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_f_7celiagg_8_celiagg_10CanvasBase__get_native_image(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_image, enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_fmt) {
  struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_get_native_image", 0);

  /* "celiagg/ndarray_canvas.pxi":268
 *         :param format: The desired output pixel format
 *         """
 *         if image.pixel_format == fmt:             # <<<<<<<<<<<<<<
 *             return image
 * 
 */
  __pyx_t_1 = ((__pyx_v_image->pixel_format == __pyx_v_fmt) != 0);
  if (__pyx_t_1) {

    /* "celiagg/ndarray_canvas.pxi":269
 *         """
 *         if image.pixel_format == fmt:
 *             return image             # <<<<<<<<<<<<<<
 * 
 *         return convert_image(image, fmt, bottom_up=image.bottom_up)
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_image));
    __pyx_r = __pyx_v_image;
    goto __pyx_L0;

    /* "celiagg/ndarray_canvas.pxi":268
 *         :param format: The desired output pixel format
 *         """
 *         if image.pixel_format == fmt:             # <<<<<<<<<<<<<<
 *             return image
 * 
 */
  }

  /* "celiagg/ndarray_canvas.pxi":271
 *             return image
 * 
 *         return convert_image(image, fmt, bottom_up=image.bottom_up)             # <<<<<<<<<<<<<<
 * 
 *     cdef Paint _get_native_paint(self, paint, PixelFormat fmt):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_convert_image); if (unlikely(!__pyx_t_2)) __PYX_ERR(6, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_fmt); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(((PyObject *)__pyx_v_image));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_image));
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_image));
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_image->bottom_up); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_bottom_up, __pyx_t_5) < 0) __PYX_ERR(6, 271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_7celiagg_8_celiagg_Image))))) __PYX_ERR(6, 271, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "celiagg/ndarray_canvas.pxi":261
 *                 raise AggError(msg)
 * 
 *     cdef Image _get_native_image(self, Image image, PixelFormat fmt):             # <<<<<<<<<<<<<<
 *         """_get_native_image(image, format)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase._get_native_image", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":273
 *         return convert_image(image, fmt, bottom_up=image.bottom_up)
 * 
 *     cdef Paint _get_native_paint(self, paint, PixelFormat fmt):             # <<<<<<<<<<<<<<
 *         """_get_native_paint(paint, format)
 * 
 */

static struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_f_7celiagg_8_celiagg_10CanvasBase__get_native_paint(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, PyObject *__pyx_v_paint, enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_fmt) {
  struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("_get_native_paint", 0);

  /* "celiagg/ndarray_canvas.pxi":280
 *         :param format: The desired output pixel format
 *         """
 *         if paint is None:             # <<<<<<<<<<<<<<
 *             return SolidPaint(0.0, 0.0, 0.0)
 * 
 */
  __pyx_t_1 = (__pyx_v_paint == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "celiagg/ndarray_canvas.pxi":281
 *         """
 *         if paint is None:
 *             return SolidPaint(0.0, 0.0, 0.0)             # <<<<<<<<<<<<<<
 * 
 *         if not hasattr(paint, '_with_format'):
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_SolidPaint), __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = ((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "celiagg/ndarray_canvas.pxi":280
 *         :param format: The desired output pixel format
 *         """
 *         if paint is None:             # <<<<<<<<<<<<<<
 *             return SolidPaint(0.0, 0.0, 0.0)
 * 
 */
  }

  /* "celiagg/ndarray_canvas.pxi":283
 *             return SolidPaint(0.0, 0.0, 0.0)
 * 
 *         if not hasattr(paint, '_with_format'):             # <<<<<<<<<<<<<<
 *             return paint
 * 
 */
  __pyx_t_2 = __Pyx_HasAttr(__pyx_v_paint, __pyx_n_s_with_format); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(6, 283, __pyx_L1_error)
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "celiagg/ndarray_canvas.pxi":284
 * 
 *         if not hasattr(paint, '_with_format'):
 *             return paint             # <<<<<<<<<<<<<<
 * 
 *         return paint._with_format(fmt)
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    if (!(likely(((__pyx_v_paint) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_paint, __pyx_ptype_7celiagg_8_celiagg_Paint))))) __PYX_ERR(6, 284, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_paint);
    __pyx_r = ((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_v_paint);
    goto __pyx_L0;

    /* "celiagg/ndarray_canvas.pxi":283
 *             return SolidPaint(0.0, 0.0, 0.0)
 * 
 *         if not hasattr(paint, '_with_format'):             # <<<<<<<<<<<<<<
 *             return paint
 * 
 */
  }

  /* "celiagg/ndarray_canvas.pxi":286
 *             return paint
 * 
 *         return paint._with_format(fmt)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_paint, __pyx_n_s_with_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(6, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_fmt); if (unlikely(!__pyx_t_5)) __PYX_ERR(6, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(6, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_7celiagg_8_celiagg_Paint))))) __PYX_ERR(6, 286, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/ndarray_canvas.pxi":273
 *         return convert_image(image, fmt, bottom_up=image.bottom_up)
 * 
 *     cdef Paint _get_native_paint(self, paint, PixelFormat fmt):             # <<<<<<<<<<<<<<
 *         """_get_native_paint(paint, format)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase._get_native_paint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_10__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasBase_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasBase_12__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasBase_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":299
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, float[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.RGBA128
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13CanvasRGBA128_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13CanvasRGBA128_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_image = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_bottom_up = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_bottom_up,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bottom_up);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(6, 299, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_float(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_image.memview)) __PYX_ERR(6, 299, __pyx_L3_error)
    __pyx_v_bottom_up = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 299, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGBA128.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13CanvasRGBA128___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *)__pyx_v_self), __pyx_v_image, __pyx_v_bottom_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13CanvasRGBA128___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  bool __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/ndarray_canvas.pxi":300
 *     """
 *     def __cinit__(self, float[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 4, True)             # <<<<<<<<<<<<<<
 *         self.pixel_format = PixelFormat.RGBA128
 *         self.bottom_up = bottom_up
 */
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_image, 3, (PyObject *(*)(char *)) __pyx_memview_get_float, (int (*)(char *, PyObject *)) __pyx_memview_set_float, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGBA128 *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.base_init(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_t_1, 4, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(6, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/ndarray_canvas.pxi":301
 *     def __cinit__(self, float[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.RGBA128             # <<<<<<<<<<<<<<
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_rgba128_t(<_bytes_t*>&image[0][0][0],
 */
  __pyx_v_self->__pyx_base.pixel_format = __pyx_e_7celiagg_8_celiagg_RGBA128;

  /* "celiagg/ndarray_canvas.pxi":302
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.RGBA128
 *         self.bottom_up = bottom_up             # <<<<<<<<<<<<<<
 *         self._this = <canvas_base_t*> new canvas_rgba128_t(<_bytes_t*>&image[0][0][0],
 *                                                            image.shape[1],
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 302, __pyx_L1_error)
  __pyx_v_self->__pyx_base.bottom_up = __pyx_t_3;

  /* "celiagg/ndarray_canvas.pxi":303
 *         self.pixel_format = PixelFormat.RGBA128
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_rgba128_t(<_bytes_t*>&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                            image.shape[1],
 *                                                            image.shape[0],
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_2 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_v_image.shape[0];
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_2 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_v_image.shape[0])) __pyx_t_2 = 0;
  if (__pyx_t_5 < 0) {
    __pyx_t_5 += __pyx_v_image.shape[1];
    if (unlikely(__pyx_t_5 < 0)) __pyx_t_2 = 1;
  } else if (unlikely(__pyx_t_5 >= __pyx_v_image.shape[1])) __pyx_t_2 = 1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_v_image.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_2 = 2;
  } else if (unlikely(__pyx_t_6 >= __pyx_v_image.shape[2])) __pyx_t_2 = 2;
  if (unlikely(__pyx_t_2 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_2);
    __PYX_ERR(6, 303, __pyx_L1_error)
  }

  /* "celiagg/ndarray_canvas.pxi":307
 *                                                            image.shape[0],
 *                                                            image.strides[0], 4,
 *                                                            bottom_up)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 307, __pyx_L1_error)

  /* "celiagg/ndarray_canvas.pxi":303
 *         self.pixel_format = PixelFormat.RGBA128
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_rgba128_t(<_bytes_t*>&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                            image.shape[1],
 *                                                            image.shape[0],
 */
  __pyx_v_self->__pyx_base._this = ((__pyx_t_7celiagg_8_celiagg_canvas_base_t *)new __pyx_t_7celiagg_8_celiagg_canvas_rgba128_t(((__pyx_t_7celiagg_8_celiagg__bytes_t *)(&(*((float *) ( /* dim=2 */ ((char *) (((float *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_image.data + __pyx_t_4 * __pyx_v_image.strides[0]) ) + __pyx_t_5 * __pyx_v_image.strides[1]) )) + __pyx_t_6)) ))))), (__pyx_v_image.shape[1]), (__pyx_v_image.shape[0]), (__pyx_v_image.strides[0]), 4, __pyx_t_3));

  /* "celiagg/ndarray_canvas.pxi":299
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, float[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.RGBA128
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGBA128.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_image, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13CanvasRGBA128_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13CanvasRGBA128_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13CanvasRGBA128_2__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13CanvasRGBA128_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGBA128.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13CanvasRGBA128_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13CanvasRGBA128_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13CanvasRGBA128_4__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13CanvasRGBA128_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGBA128.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":320
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.BGRA32
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_12CanvasBGRA32_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_12CanvasBGRA32_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_image = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_bottom_up = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_bottom_up,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bottom_up);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(6, 320, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn___pyx_t_7celiagg_8_celiagg__bytes_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_image.memview)) __PYX_ERR(6, 320, __pyx_L3_error)
    __pyx_v_bottom_up = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 320, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBGRA32.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12CanvasBGRA32___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *)__pyx_v_self), __pyx_v_image, __pyx_v_bottom_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_12CanvasBGRA32___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  bool __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/ndarray_canvas.pxi":321
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 4, True)             # <<<<<<<<<<<<<<
 *         self.pixel_format = PixelFormat.BGRA32
 *         self.bottom_up = bottom_up
 */
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_image, 3, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_7celiagg_8_celiagg__bytes_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_7celiagg_8_celiagg__bytes_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBGRA32 *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.base_init(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_t_1, 4, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(6, 321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/ndarray_canvas.pxi":322
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.BGRA32             # <<<<<<<<<<<<<<
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_brga32_t(&image[0][0][0],
 */
  __pyx_v_self->__pyx_base.pixel_format = __pyx_e_7celiagg_8_celiagg_BGRA32;

  /* "celiagg/ndarray_canvas.pxi":323
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.BGRA32
 *         self.bottom_up = bottom_up             # <<<<<<<<<<<<<<
 *         self._this = <canvas_base_t*> new canvas_brga32_t(&image[0][0][0],
 *                                                           image.shape[1],
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 323, __pyx_L1_error)
  __pyx_v_self->__pyx_base.bottom_up = __pyx_t_3;

  /* "celiagg/ndarray_canvas.pxi":324
 *         self.pixel_format = PixelFormat.BGRA32
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_brga32_t(&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                           image.shape[1],
 *                                                           image.shape[0],
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_2 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_v_image.shape[0];
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_2 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_v_image.shape[0])) __pyx_t_2 = 0;
  if (__pyx_t_5 < 0) {
    __pyx_t_5 += __pyx_v_image.shape[1];
    if (unlikely(__pyx_t_5 < 0)) __pyx_t_2 = 1;
  } else if (unlikely(__pyx_t_5 >= __pyx_v_image.shape[1])) __pyx_t_2 = 1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_v_image.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_2 = 2;
  } else if (unlikely(__pyx_t_6 >= __pyx_v_image.shape[2])) __pyx_t_2 = 2;
  if (unlikely(__pyx_t_2 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_2);
    __PYX_ERR(6, 324, __pyx_L1_error)
  }

  /* "celiagg/ndarray_canvas.pxi":328
 *                                                           image.shape[0],
 *                                                           image.strides[0], 4,
 *                                                           bottom_up)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 328, __pyx_L1_error)

  /* "celiagg/ndarray_canvas.pxi":324
 *         self.pixel_format = PixelFormat.BGRA32
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_brga32_t(&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                           image.shape[1],
 *                                                           image.shape[0],
 */
  __pyx_v_self->__pyx_base._this = ((__pyx_t_7celiagg_8_celiagg_canvas_base_t *)new __pyx_t_7celiagg_8_celiagg_canvas_brga32_t((&(*((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=2 */ ((char *) (((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_image.data + __pyx_t_4 * __pyx_v_image.strides[0]) ) + __pyx_t_5 * __pyx_v_image.strides[1]) )) + __pyx_t_6)) )))), (__pyx_v_image.shape[1]), (__pyx_v_image.shape[0]), (__pyx_v_image.strides[0]), 4, __pyx_t_3));

  /* "celiagg/ndarray_canvas.pxi":320
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.BGRA32
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBGRA32.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_image, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12CanvasBGRA32_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12CanvasBGRA32_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12CanvasBGRA32_2__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12CanvasBGRA32_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBGRA32.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12CanvasBGRA32_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12CanvasBGRA32_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12CanvasBGRA32_4__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12CanvasBGRA32_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasBGRA32.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":341
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.RGBA32
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_12CanvasRGBA32_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_12CanvasRGBA32_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_image = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_bottom_up = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_bottom_up,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bottom_up);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(6, 341, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn___pyx_t_7celiagg_8_celiagg__bytes_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_image.memview)) __PYX_ERR(6, 341, __pyx_L3_error)
    __pyx_v_bottom_up = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 341, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGBA32.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12CanvasRGBA32___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *)__pyx_v_self), __pyx_v_image, __pyx_v_bottom_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_12CanvasRGBA32___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  bool __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/ndarray_canvas.pxi":342
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 4, True)             # <<<<<<<<<<<<<<
 *         self.pixel_format = PixelFormat.RGBA32
 *         self.bottom_up = bottom_up
 */
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_image, 3, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_7celiagg_8_celiagg__bytes_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_7celiagg_8_celiagg__bytes_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGBA32 *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.base_init(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_t_1, 4, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(6, 342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/ndarray_canvas.pxi":343
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.RGBA32             # <<<<<<<<<<<<<<
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_rgba32_t(&image[0][0][0],
 */
  __pyx_v_self->__pyx_base.pixel_format = __pyx_e_7celiagg_8_celiagg_RGBA32;

  /* "celiagg/ndarray_canvas.pxi":344
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.RGBA32
 *         self.bottom_up = bottom_up             # <<<<<<<<<<<<<<
 *         self._this = <canvas_base_t*> new canvas_rgba32_t(&image[0][0][0],
 *                                                           image.shape[1],
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 344, __pyx_L1_error)
  __pyx_v_self->__pyx_base.bottom_up = __pyx_t_3;

  /* "celiagg/ndarray_canvas.pxi":345
 *         self.pixel_format = PixelFormat.RGBA32
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_rgba32_t(&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                           image.shape[1],
 *                                                           image.shape[0],
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_2 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_v_image.shape[0];
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_2 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_v_image.shape[0])) __pyx_t_2 = 0;
  if (__pyx_t_5 < 0) {
    __pyx_t_5 += __pyx_v_image.shape[1];
    if (unlikely(__pyx_t_5 < 0)) __pyx_t_2 = 1;
  } else if (unlikely(__pyx_t_5 >= __pyx_v_image.shape[1])) __pyx_t_2 = 1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_v_image.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_2 = 2;
  } else if (unlikely(__pyx_t_6 >= __pyx_v_image.shape[2])) __pyx_t_2 = 2;
  if (unlikely(__pyx_t_2 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_2);
    __PYX_ERR(6, 345, __pyx_L1_error)
  }

  /* "celiagg/ndarray_canvas.pxi":349
 *                                                           image.shape[0],
 *                                                           image.strides[0], 4,
 *                                                           bottom_up)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 349, __pyx_L1_error)

  /* "celiagg/ndarray_canvas.pxi":345
 *         self.pixel_format = PixelFormat.RGBA32
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_rgba32_t(&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                           image.shape[1],
 *                                                           image.shape[0],
 */
  __pyx_v_self->__pyx_base._this = ((__pyx_t_7celiagg_8_celiagg_canvas_base_t *)new __pyx_t_7celiagg_8_celiagg_canvas_rgba32_t((&(*((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=2 */ ((char *) (((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_image.data + __pyx_t_4 * __pyx_v_image.strides[0]) ) + __pyx_t_5 * __pyx_v_image.strides[1]) )) + __pyx_t_6)) )))), (__pyx_v_image.shape[1]), (__pyx_v_image.shape[0]), (__pyx_v_image.strides[0]), 4, __pyx_t_3));

  /* "celiagg/ndarray_canvas.pxi":341
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 4, True)
 *         self.pixel_format = PixelFormat.RGBA32
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGBA32.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_image, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12CanvasRGBA32_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12CanvasRGBA32_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12CanvasRGBA32_2__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12CanvasRGBA32_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGBA32.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12CanvasRGBA32_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12CanvasRGBA32_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12CanvasRGBA32_4__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12CanvasRGBA32_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGBA32.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":362
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 4, False)
 *         self.pixel_format = PixelFormat.RGB24
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_11CanvasRGB24_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_11CanvasRGB24_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_image = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_bottom_up = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_bottom_up,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bottom_up);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(6, 362, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn___pyx_t_7celiagg_8_celiagg__bytes_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_image.memview)) __PYX_ERR(6, 362, __pyx_L3_error)
    __pyx_v_bottom_up = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 362, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGB24.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_11CanvasRGB24___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *)__pyx_v_self), __pyx_v_image, __pyx_v_bottom_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_11CanvasRGB24___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  bool __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/ndarray_canvas.pxi":363
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 4, False)             # <<<<<<<<<<<<<<
 *         self.pixel_format = PixelFormat.RGB24
 *         self.bottom_up = bottom_up
 */
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_image, 3, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_7celiagg_8_celiagg__bytes_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_7celiagg_8_celiagg__bytes_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGB24 *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.base_init(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_t_1, 4, 0); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(6, 363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/ndarray_canvas.pxi":364
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 4, False)
 *         self.pixel_format = PixelFormat.RGB24             # <<<<<<<<<<<<<<
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_rgb24_t(&image[0][0][0],
 */
  __pyx_v_self->__pyx_base.pixel_format = __pyx_e_7celiagg_8_celiagg_RGB24;

  /* "celiagg/ndarray_canvas.pxi":365
 *         self.base_init(image, 4, False)
 *         self.pixel_format = PixelFormat.RGB24
 *         self.bottom_up = bottom_up             # <<<<<<<<<<<<<<
 *         self._this = <canvas_base_t*> new canvas_rgb24_t(&image[0][0][0],
 *                                                          image.shape[1],
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 365, __pyx_L1_error)
  __pyx_v_self->__pyx_base.bottom_up = __pyx_t_3;

  /* "celiagg/ndarray_canvas.pxi":366
 *         self.pixel_format = PixelFormat.RGB24
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_rgb24_t(&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                          image.shape[1],
 *                                                          image.shape[0],
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_2 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_v_image.shape[0];
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_2 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_v_image.shape[0])) __pyx_t_2 = 0;
  if (__pyx_t_5 < 0) {
    __pyx_t_5 += __pyx_v_image.shape[1];
    if (unlikely(__pyx_t_5 < 0)) __pyx_t_2 = 1;
  } else if (unlikely(__pyx_t_5 >= __pyx_v_image.shape[1])) __pyx_t_2 = 1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_v_image.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_2 = 2;
  } else if (unlikely(__pyx_t_6 >= __pyx_v_image.shape[2])) __pyx_t_2 = 2;
  if (unlikely(__pyx_t_2 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_2);
    __PYX_ERR(6, 366, __pyx_L1_error)
  }

  /* "celiagg/ndarray_canvas.pxi":370
 *                                                          image.shape[0],
 *                                                          image.strides[0], 3,
 *                                                          bottom_up)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 370, __pyx_L1_error)

  /* "celiagg/ndarray_canvas.pxi":366
 *         self.pixel_format = PixelFormat.RGB24
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_rgb24_t(&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                          image.shape[1],
 *                                                          image.shape[0],
 */
  __pyx_v_self->__pyx_base._this = ((__pyx_t_7celiagg_8_celiagg_canvas_base_t *)new __pyx_t_7celiagg_8_celiagg_canvas_rgb24_t((&(*((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=2 */ ((char *) (((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_image.data + __pyx_t_4 * __pyx_v_image.strides[0]) ) + __pyx_t_5 * __pyx_v_image.strides[1]) )) + __pyx_t_6)) )))), (__pyx_v_image.shape[1]), (__pyx_v_image.shape[0]), (__pyx_v_image.strides[0]), 3, __pyx_t_3));

  /* "celiagg/ndarray_canvas.pxi":362
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 4, False)
 *         self.pixel_format = PixelFormat.RGB24
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGB24.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_image, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_11CanvasRGB24_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_11CanvasRGB24_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_11CanvasRGB24_2__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_11CanvasRGB24_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGB24.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_11CanvasRGB24_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_11CanvasRGB24_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_11CanvasRGB24_4__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_11CanvasRGB24_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasRGB24.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":383
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 2, True)
 *         self.pixel_format = PixelFormat.Gray8
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_10CanvasGA16_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_10CanvasGA16_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_image = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_bottom_up = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_bottom_up,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bottom_up);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(6, 383, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn___pyx_t_7celiagg_8_celiagg__bytes_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_image.memview)) __PYX_ERR(6, 383, __pyx_L3_error)
    __pyx_v_bottom_up = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 383, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasGA16.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasGA16___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *)__pyx_v_self), __pyx_v_image, __pyx_v_bottom_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_10CanvasGA16___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  bool __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/ndarray_canvas.pxi":384
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 2, True)             # <<<<<<<<<<<<<<
 *         self.pixel_format = PixelFormat.Gray8
 *         self.bottom_up = bottom_up
 */
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_image, 3, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_7celiagg_8_celiagg__bytes_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_7celiagg_8_celiagg__bytes_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasGA16 *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.base_init(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_t_1, 2, 1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(6, 384, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/ndarray_canvas.pxi":385
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):
 *         self.base_init(image, 2, True)
 *         self.pixel_format = PixelFormat.Gray8             # <<<<<<<<<<<<<<
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_ga16_t(&image[0][0][0],
 */
  __pyx_v_self->__pyx_base.pixel_format = __pyx_e_7celiagg_8_celiagg_Gray8;

  /* "celiagg/ndarray_canvas.pxi":386
 *         self.base_init(image, 2, True)
 *         self.pixel_format = PixelFormat.Gray8
 *         self.bottom_up = bottom_up             # <<<<<<<<<<<<<<
 *         self._this = <canvas_base_t*> new canvas_ga16_t(&image[0][0][0],
 *                                                         image.shape[1],
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 386, __pyx_L1_error)
  __pyx_v_self->__pyx_base.bottom_up = __pyx_t_3;

  /* "celiagg/ndarray_canvas.pxi":387
 *         self.pixel_format = PixelFormat.Gray8
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_ga16_t(&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                         image.shape[1],
 *                                                         image.shape[0],
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_2 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_v_image.shape[0];
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_2 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_v_image.shape[0])) __pyx_t_2 = 0;
  if (__pyx_t_5 < 0) {
    __pyx_t_5 += __pyx_v_image.shape[1];
    if (unlikely(__pyx_t_5 < 0)) __pyx_t_2 = 1;
  } else if (unlikely(__pyx_t_5 >= __pyx_v_image.shape[1])) __pyx_t_2 = 1;
  if (__pyx_t_6 < 0) {
    __pyx_t_6 += __pyx_v_image.shape[2];
    if (unlikely(__pyx_t_6 < 0)) __pyx_t_2 = 2;
  } else if (unlikely(__pyx_t_6 >= __pyx_v_image.shape[2])) __pyx_t_2 = 2;
  if (unlikely(__pyx_t_2 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_2);
    __PYX_ERR(6, 387, __pyx_L1_error)
  }

  /* "celiagg/ndarray_canvas.pxi":391
 *                                                         image.shape[0],
 *                                                         image.strides[0], 2,
 *                                                         bottom_up)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 391, __pyx_L1_error)

  /* "celiagg/ndarray_canvas.pxi":387
 *         self.pixel_format = PixelFormat.Gray8
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_ga16_t(&image[0][0][0],             # <<<<<<<<<<<<<<
 *                                                         image.shape[1],
 *                                                         image.shape[0],
 */
  __pyx_v_self->__pyx_base._this = ((__pyx_t_7celiagg_8_celiagg_canvas_base_t *)new __pyx_t_7celiagg_8_celiagg_canvas_ga16_t((&(*((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=2 */ ((char *) (((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=1 */ (( /* dim=0 */ (__pyx_v_image.data + __pyx_t_4 * __pyx_v_image.strides[0]) ) + __pyx_t_5 * __pyx_v_image.strides[1]) )) + __pyx_t_6)) )))), (__pyx_v_image.shape[1]), (__pyx_v_image.shape[0]), (__pyx_v_image.strides[0]), 2, __pyx_t_3));

  /* "celiagg/ndarray_canvas.pxi":383
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 2, True)
 *         self.pixel_format = PixelFormat.Gray8
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasGA16.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_image, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasGA16_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasGA16_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasGA16_2__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasGA16_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasGA16.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasGA16_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10CanvasGA16_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10CanvasGA16_4__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10CanvasGA16_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasGA16.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/ndarray_canvas.pxi":404
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 2, False)
 *         self.pixel_format = PixelFormat.Gray8
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_8CanvasG8_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_8CanvasG8_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_image = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_bottom_up = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_bottom_up,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bottom_up);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(6, 404, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_image = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_7celiagg_8_celiagg__bytes_t(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_image.memview)) __PYX_ERR(6, 404, __pyx_L3_error)
    __pyx_v_bottom_up = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(6, 404, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.CanvasG8.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_8CanvasG8___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *)__pyx_v_self), __pyx_v_image, __pyx_v_bottom_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_8CanvasG8___cinit__(struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *__pyx_v_self, __Pyx_memviewslice __pyx_v_image, PyObject *__pyx_v_bottom_up) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  bool __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/ndarray_canvas.pxi":405
 *     """
 *     def __cinit__(self, _bytes_t[:,::1] image, bottom_up=False):
 *         self.base_init(image, 2, False)             # <<<<<<<<<<<<<<
 *         self.pixel_format = PixelFormat.Gray8
 *         self.bottom_up = bottom_up
 */
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_image, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_7celiagg_8_celiagg__bytes_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_7celiagg_8_celiagg__bytes_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasG8 *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.base_init(((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)__pyx_v_self), __pyx_t_1, 2, 0); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(6, 405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/ndarray_canvas.pxi":406
 *     def __cinit__(self, _bytes_t[:,::1] image, bottom_up=False):
 *         self.base_init(image, 2, False)
 *         self.pixel_format = PixelFormat.Gray8             # <<<<<<<<<<<<<<
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_ga16_t(&image[0][0],
 */
  __pyx_v_self->__pyx_base.pixel_format = __pyx_e_7celiagg_8_celiagg_Gray8;

  /* "celiagg/ndarray_canvas.pxi":407
 *         self.base_init(image, 2, False)
 *         self.pixel_format = PixelFormat.Gray8
 *         self.bottom_up = bottom_up             # <<<<<<<<<<<<<<
 *         self._this = <canvas_base_t*> new canvas_ga16_t(&image[0][0],
 *                                                         image.shape[1],
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 407, __pyx_L1_error)
  __pyx_v_self->__pyx_base.bottom_up = __pyx_t_3;

  /* "celiagg/ndarray_canvas.pxi":408
 *         self.pixel_format = PixelFormat.Gray8
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_ga16_t(&image[0][0],             # <<<<<<<<<<<<<<
 *                                                         image.shape[1],
 *                                                         image.shape[0],
 */
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_2 = -1;
  if (__pyx_t_4 < 0) {
    __pyx_t_4 += __pyx_v_image.shape[0];
    if (unlikely(__pyx_t_4 < 0)) __pyx_t_2 = 0;
  } else if (unlikely(__pyx_t_4 >= __pyx_v_image.shape[0])) __pyx_t_2 = 0;
  if (__pyx_t_5 < 0) {
    __pyx_t_5 += __pyx_v_image.shape[1];
    if (unlikely(__pyx_t_5 < 0)) __pyx_t_2 = 1;
  } else if (unlikely(__pyx_t_5 >= __pyx_v_image.shape[1])) __pyx_t_2 = 1;
  if (unlikely(__pyx_t_2 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_2);
    __PYX_ERR(6, 408, __pyx_L1_error)
  }

  /* "celiagg/ndarray_canvas.pxi":412
 *                                                         image.shape[0],
 *                                                         image.strides[0], 1,
 *                                                         bottom_up)             # <<<<<<<<<<<<<<
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bottom_up); if (unlikely((__pyx_t_3 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(6, 412, __pyx_L1_error)

  /* "celiagg/ndarray_canvas.pxi":408
 *         self.pixel_format = PixelFormat.Gray8
 *         self.bottom_up = bottom_up
 *         self._this = <canvas_base_t*> new canvas_ga16_t(&image[0][0],             # <<<<<<<<<<<<<<
 *                                                         image.shape[1],
 *                                                         image.shape[0],
 */
  __pyx_v_self->__pyx_base._this = ((__pyx_t_7celiagg_8_celiagg_canvas_base_t *)new __pyx_t_7celiagg_8_celiagg_canvas_ga16_t((&(*((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=1 */ ((char *) (((__pyx_t_7celiagg_8_celiagg__bytes_t *) ( /* dim=0 */ (__pyx_v_image.data + __pyx_t_4 * __pyx_v_image.strides[0]) )) + __pyx_t_5)) )))), (__pyx_v_image.shape[1]), (__pyx_v_image.shape[0]), (__pyx_v_image.strides[0]), 1, __pyx_t_3));

  /* "celiagg/ndarray_canvas.pxi":404
 *     :param bottom_up: If True, the origin is the bottom left, instead of top-left
 *     """
 *     def __cinit__(self, _bytes_t[:,::1] image, bottom_up=False):             # <<<<<<<<<<<<<<
 *         self.base_init(image, 2, False)
 *         self.pixel_format = PixelFormat.Gray8
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasG8.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_image, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_8CanvasG8_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_8CanvasG8_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_8CanvasG8_2__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_8CanvasG8_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasG8.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_8CanvasG8_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_8CanvasG8_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_8CanvasG8_4__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_8CanvasG8_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.CanvasG8.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":39
 * 
 * 
 * cdef _get_gradient_points(points, length):             # <<<<<<<<<<<<<<
 *     cdef double[::1] arr = numpy.asarray(points,
 *                                          dtype=numpy.float64,
 */

static PyObject *__pyx_f_7celiagg_8_celiagg__get_gradient_points(PyObject *__pyx_v_points, PyObject *__pyx_v_length) {
  __Pyx_memviewslice __pyx_v_arr = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_get_gradient_points", 0);

  /* "celiagg/paint.pxi":40
 * 
 * cdef _get_gradient_points(points, length):
 *     cdef double[::1] arr = numpy.asarray(points,             # <<<<<<<<<<<<<<
 *                                          dtype=numpy.float64,
 *                                          order='c')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_points);
  __Pyx_GIVEREF(__pyx_v_points);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_points);

  /* "celiagg/paint.pxi":41
 * cdef _get_gradient_points(points, length):
 *     cdef double[::1] arr = numpy.asarray(points,
 *                                          dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                          order='c')
 *     if arr.shape[0] != length:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(7, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(7, 41, __pyx_L1_error)

  /* "celiagg/paint.pxi":40
 * 
 * cdef _get_gradient_points(points, length):
 *     cdef double[::1] arr = numpy.asarray(points,             # <<<<<<<<<<<<<<
 *                                          dtype=numpy.float64,
 *                                          order='c')
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(7, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_arr = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "celiagg/paint.pxi":43
 *                                          dtype=numpy.float64,
 *                                          order='c')
 *     if arr.shape[0] != length:             # <<<<<<<<<<<<<<
 *         msg = 'points argument must be an iterable of length {}'.format(length)
 *         raise ValueError(msg)
 */
  __pyx_t_5 = PyInt_FromSsize_t((__pyx_v_arr.shape[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_v_length, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(7, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_7)) {

    /* "celiagg/paint.pxi":44
 *                                          order='c')
 *     if arr.shape[0] != length:
 *         msg = 'points argument must be an iterable of length {}'.format(length)             # <<<<<<<<<<<<<<
 *         raise ValueError(msg)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_points_argument_must_be_an_itera, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_1, __pyx_v_length) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_length);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_msg = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "celiagg/paint.pxi":45
 *     if arr.shape[0] != length:
 *         msg = 'points argument must be an iterable of length {}'.format(length)
 *         raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *     return arr
 */
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(7, 45, __pyx_L1_error)

    /* "celiagg/paint.pxi":43
 *                                          dtype=numpy.float64,
 *                                          order='c')
 *     if arr.shape[0] != length:             # <<<<<<<<<<<<<<
 *         msg = 'points argument must be an iterable of length {}'.format(length)
 *         raise ValueError(msg)
 */
  }

  /* "celiagg/paint.pxi":47
 *         raise ValueError(msg)
 * 
 *     return arr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_arr, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/paint.pxi":39
 * 
 * 
 * cdef _get_gradient_points(points, length):             # <<<<<<<<<<<<<<
 *     cdef double[::1] arr = numpy.asarray(points,
 *                                          dtype=numpy.float64,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("celiagg._celiagg._get_gradient_points", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_arr, 1);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":50
 * 
 * 
 * cdef _get_gradient_stops(stops):             # <<<<<<<<<<<<<<
 *     cdef double[:,::1] arr = numpy.asarray(stops,
 *                                            dtype=numpy.float64,
 */

static PyObject *__pyx_f_7celiagg_8_celiagg__get_gradient_stops(PyObject *__pyx_v_stops) {
  __Pyx_memviewslice __pyx_v_arr = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_get_gradient_stops", 0);

  /* "celiagg/paint.pxi":51
 * 
 * cdef _get_gradient_stops(stops):
 *     cdef double[:,::1] arr = numpy.asarray(stops,             # <<<<<<<<<<<<<<
 *                                            dtype=numpy.float64,
 *                                            order='c')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_stops);
  __Pyx_GIVEREF(__pyx_v_stops);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_stops);

  /* "celiagg/paint.pxi":52
 * cdef _get_gradient_stops(stops):
 *     cdef double[:,::1] arr = numpy.asarray(stops,
 *                                            dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                            order='c')
 *     if arr.shape[1] != 5:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(7, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(7, 52, __pyx_L1_error)

  /* "celiagg/paint.pxi":51
 * 
 * cdef _get_gradient_stops(stops):
 *     cdef double[:,::1] arr = numpy.asarray(stops,             # <<<<<<<<<<<<<<
 *                                            dtype=numpy.float64,
 *                                            order='c')
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(7, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_arr = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "celiagg/paint.pxi":54
 *                                            dtype=numpy.float64,
 *                                            order='c')
 *     if arr.shape[1] != 5:             # <<<<<<<<<<<<<<
 *         msg = ('stops argument must be an iterable of (off, r, g, b, a) '
 *                 'tuples.')
 */
  __pyx_t_7 = (((__pyx_v_arr.shape[1]) != 5) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "celiagg/paint.pxi":55
 *                                            order='c')
 *     if arr.shape[1] != 5:
 *         msg = ('stops argument must be an iterable of (off, r, g, b, a) '             # <<<<<<<<<<<<<<
 *                 'tuples.')
 *         raise ValueError(msg)
 */
    __Pyx_INCREF(__pyx_kp_s_stops_argument_must_be_an_iterab);
    __pyx_v_msg = __pyx_kp_s_stops_argument_must_be_an_iterab;

    /* "celiagg/paint.pxi":57
 *         msg = ('stops argument must be an iterable of (off, r, g, b, a) '
 *                 'tuples.')
 *         raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *     return arr
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(7, 57, __pyx_L1_error)

    /* "celiagg/paint.pxi":54
 *                                            dtype=numpy.float64,
 *                                            order='c')
 *     if arr.shape[1] != 5:             # <<<<<<<<<<<<<<
 *         msg = ('stops argument must be an iterable of (off, r, g, b, a) '
 *                 'tuples.')
 */
  }

  /* "celiagg/paint.pxi":59
 *         raise ValueError(msg)
 * 
 *     return arr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_arr, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "celiagg/paint.pxi":50
 * 
 * 
 * cdef _get_gradient_stops(stops):             # <<<<<<<<<<<<<<
 *     cdef double[:,::1] arr = numpy.asarray(stops,
 *                                            dtype=numpy.float64,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("celiagg._celiagg._get_gradient_stops", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_arr, 1);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":66
 *     cdef _paint.Paint* _this
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

/* Python wrapper */
static void __pyx_pw_7celiagg_8_celiagg_5Paint_1__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7celiagg_8_celiagg_5Paint_1__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7celiagg_8_celiagg_5Paint___dealloc__(((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7celiagg_8_celiagg_5Paint___dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "celiagg/paint.pxi":67
 * 
 *     def __dealloc__(self):
 *         del self._this             # <<<<<<<<<<<<<<
 * 
 *     property transform:
 */
  delete __pyx_v_self->_this;

  /* "celiagg/paint.pxi":66
 *     cdef _paint.Paint* _this
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "celiagg/paint.pxi":70
 * 
 *     property transform:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef Transform trans = Transform()
 *             trans._assign_obj(self._this.transform())
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Paint_9transform_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Paint_9transform_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Paint_9transform___get__(((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Paint_9transform___get__(struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self) {
  struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_trans = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/paint.pxi":71
 *     property transform:
 *         def __get__(self):
 *             cdef Transform trans = Transform()             # <<<<<<<<<<<<<<
 *             trans._assign_obj(self._this.transform())
 *             return trans
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_Transform)); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_trans = ((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":72
 *         def __get__(self):
 *             cdef Transform trans = Transform()
 *             trans._assign_obj(self._this.transform())             # <<<<<<<<<<<<<<
 *             return trans
 *         def __set__(self, trans):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_7celiagg_8_celiagg_Transform *)__pyx_v_trans->__pyx_vtab)->_assign_obj(__pyx_v_trans, __pyx_v_self->_this->transform()); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":73
 *             cdef Transform trans = Transform()
 *             trans._assign_obj(self._this.transform())
 *             return trans             # <<<<<<<<<<<<<<
 *         def __set__(self, trans):
 *             if not isinstance(trans, Transform):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_trans));
  __pyx_r = ((PyObject *)__pyx_v_trans);
  goto __pyx_L0;

  /* "celiagg/paint.pxi":70
 * 
 *     property transform:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef Transform trans = Transform()
 *             trans._assign_obj(self._this.transform())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Paint.transform.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_trans);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":74
 *             trans._assign_obj(self._this.transform())
 *             return trans
 *         def __set__(self, trans):             # <<<<<<<<<<<<<<
 *             if not isinstance(trans, Transform):
 *                 raise TypeError("transform must be a Transform instance")
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_5Paint_9transform_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_trans); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_5Paint_9transform_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_trans) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Paint_9transform_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_v_self), ((PyObject *)__pyx_v_trans));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_5Paint_9transform_2__set__(struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self, PyObject *__pyx_v_trans) {
  struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v__trans = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/paint.pxi":75
 *             return trans
 *         def __set__(self, trans):
 *             if not isinstance(trans, Transform):             # <<<<<<<<<<<<<<
 *                 raise TypeError("transform must be a Transform instance")
 *             cdef Transform _trans = <Transform>trans
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_trans, __pyx_ptype_7celiagg_8_celiagg_Transform); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/paint.pxi":76
 *         def __set__(self, trans):
 *             if not isinstance(trans, Transform):
 *                 raise TypeError("transform must be a Transform instance")             # <<<<<<<<<<<<<<
 *             cdef Transform _trans = <Transform>trans
 *             self._this.transform(dereference(_trans._this))
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(7, 76, __pyx_L1_error)

    /* "celiagg/paint.pxi":75
 *             return trans
 *         def __set__(self, trans):
 *             if not isinstance(trans, Transform):             # <<<<<<<<<<<<<<
 *                 raise TypeError("transform must be a Transform instance")
 *             cdef Transform _trans = <Transform>trans
 */
  }

  /* "celiagg/paint.pxi":77
 *             if not isinstance(trans, Transform):
 *                 raise TypeError("transform must be a Transform instance")
 *             cdef Transform _trans = <Transform>trans             # <<<<<<<<<<<<<<
 *             self._this.transform(dereference(_trans._this))
 * 
 */
  __pyx_t_3 = __pyx_v_trans;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v__trans = ((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/paint.pxi":78
 *                 raise TypeError("transform must be a Transform instance")
 *             cdef Transform _trans = <Transform>trans
 *             self._this.transform(dereference(_trans._this))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->_this->transform((*__pyx_v__trans->_this));

  /* "celiagg/paint.pxi":74
 *             trans._assign_obj(self._this.transform())
 *             return trans
 *         def __set__(self, trans):             # <<<<<<<<<<<<<<
 *             if not isinstance(trans, Transform):
 *                 raise TypeError("transform must be a Transform instance")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.Paint.transform.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v__trans);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Paint_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Paint_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Paint_2__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Paint_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Paint.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Paint_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_5Paint_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_5Paint_4__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Paint *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_5Paint_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Paint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Paint.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":96
 *     cdef object _stops
 * 
 *     def __cinit__(self, double x1, double y1, double x2, double y2,             # <<<<<<<<<<<<<<
 *                   stops, GradientSpread spread, GradientUnits units):
 *         cdef double[::1] _points = _get_gradient_points((x1, y1, x2, y2), 4)
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x1;
  double __pyx_v_y1;
  double __pyx_v_x2;
  double __pyx_v_y2;
  PyObject *__pyx_v_stops = 0;
  enum __pyx_t_7celiagg_8_celiagg_GradientSpread __pyx_v_spread;
  enum __pyx_t_7celiagg_8_celiagg_GradientUnits __pyx_v_units;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x1,&__pyx_n_s_y1,&__pyx_n_s_x2,&__pyx_n_s_y2,&__pyx_n_s_stops,&__pyx_n_s_spread,&__pyx_n_s_units,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 7, 7, 1); __PYX_ERR(7, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 7, 7, 2); __PYX_ERR(7, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 7, 7, 3); __PYX_ERR(7, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stops)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 7, 7, 4); __PYX_ERR(7, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_spread)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 7, 7, 5); __PYX_ERR(7, 96, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_units)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 7, 7, 6); __PYX_ERR(7, 96, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(7, 96, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_x1 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 96, __pyx_L3_error)
    __pyx_v_y1 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 96, __pyx_L3_error)
    __pyx_v_x2 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_x2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 96, __pyx_L3_error)
    __pyx_v_y2 = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_y2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 96, __pyx_L3_error)
    __pyx_v_stops = values[4];
    __pyx_v_spread = ((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(values[5])); if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 97, __pyx_L3_error)
    __pyx_v_units = ((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(values[6])); if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 97, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(7, 96, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.LinearGradientPaint.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *)__pyx_v_self), __pyx_v_x1, __pyx_v_y1, __pyx_v_x2, __pyx_v_y2, __pyx_v_stops, __pyx_v_spread, __pyx_v_units);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint___cinit__(struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *__pyx_v_self, double __pyx_v_x1, double __pyx_v_y1, double __pyx_v_x2, double __pyx_v_y2, PyObject *__pyx_v_stops, enum __pyx_t_7celiagg_8_celiagg_GradientSpread __pyx_v_spread, enum __pyx_t_7celiagg_8_celiagg_GradientUnits __pyx_v_units) {
  __Pyx_memviewslice __pyx_v__points = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v__stops = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/paint.pxi":98
 *     def __cinit__(self, double x1, double y1, double x2, double y2,
 *                   stops, GradientSpread spread, GradientUnits units):
 *         cdef double[::1] _points = _get_gradient_points((x1, y1, x2, y2), 4)             # <<<<<<<<<<<<<<
 *         cdef double[:,::1] _stops = _get_gradient_stops(stops)
 *         self._this = new _paint.Paint(_enums.k_PaintTypeLinearGradient,
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_x1); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_y1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_x2); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_y2); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __pyx_f_7celiagg_8_celiagg__get_gradient_points(__pyx_t_5, __pyx_int_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(7, 98, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v__points = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "celiagg/paint.pxi":99
 *                   stops, GradientSpread spread, GradientUnits units):
 *         cdef double[::1] _points = _get_gradient_points((x1, y1, x2, y2), 4)
 *         cdef double[:,::1] _stops = _get_gradient_stops(stops)             # <<<<<<<<<<<<<<
 *         self._this = new _paint.Paint(_enums.k_PaintTypeLinearGradient,
 *                                       &_points[0], _points.shape[0],
 */
  __pyx_t_4 = __pyx_f_7celiagg_8_celiagg__get_gradient_stops(__pyx_v_stops); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(7, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v__stops = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "celiagg/paint.pxi":101
 *         cdef double[:,::1] _stops = _get_gradient_stops(stops)
 *         self._this = new _paint.Paint(_enums.k_PaintTypeLinearGradient,
 *                                       &_points[0], _points.shape[0],             # <<<<<<<<<<<<<<
 *                                       &_stops[0][0], _stops.shape[0],
 *                                       spread, units)
 */
  __pyx_t_8 = 0;
  __pyx_t_9 = -1;
  if (__pyx_t_8 < 0) {
    __pyx_t_8 += __pyx_v__points.shape[0];
    if (unlikely(__pyx_t_8 < 0)) __pyx_t_9 = 0;
  } else if (unlikely(__pyx_t_8 >= __pyx_v__points.shape[0])) __pyx_t_9 = 0;
  if (unlikely(__pyx_t_9 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_9);
    __PYX_ERR(7, 101, __pyx_L1_error)
  }

  /* "celiagg/paint.pxi":102
 *         self._this = new _paint.Paint(_enums.k_PaintTypeLinearGradient,
 *                                       &_points[0], _points.shape[0],
 *                                       &_stops[0][0], _stops.shape[0],             # <<<<<<<<<<<<<<
 *                                       spread, units)
 * 
 */
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_9 = -1;
  if (__pyx_t_10 < 0) {
    __pyx_t_10 += __pyx_v__stops.shape[0];
    if (unlikely(__pyx_t_10 < 0)) __pyx_t_9 = 0;
  } else if (unlikely(__pyx_t_10 >= __pyx_v__stops.shape[0])) __pyx_t_9 = 0;
  if (__pyx_t_11 < 0) {
    __pyx_t_11 += __pyx_v__stops.shape[1];
    if (unlikely(__pyx_t_11 < 0)) __pyx_t_9 = 1;
  } else if (unlikely(__pyx_t_11 >= __pyx_v__stops.shape[1])) __pyx_t_9 = 1;
  if (unlikely(__pyx_t_9 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_9);
    __PYX_ERR(7, 102, __pyx_L1_error)
  }

  /* "celiagg/paint.pxi":100
 *         cdef double[::1] _points = _get_gradient_points((x1, y1, x2, y2), 4)
 *         cdef double[:,::1] _stops = _get_gradient_stops(stops)
 *         self._this = new _paint.Paint(_enums.k_PaintTypeLinearGradient,             # <<<<<<<<<<<<<<
 *                                       &_points[0], _points.shape[0],
 *                                       &_stops[0][0], _stops.shape[0],
 */
  __pyx_v_self->__pyx_base._this = new Paint(Paint::k_PaintTypeLinearGradient, (&(*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v__points.data) + __pyx_t_8)) )))), (__pyx_v__points.shape[0]), (&(*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__stops.data + __pyx_t_10 * __pyx_v__stops.strides[0]) )) + __pyx_t_11)) )))), (__pyx_v__stops.shape[0]), ((enum Paint::GradientSpread)__pyx_v_spread), ((enum Paint::GradientUnits)__pyx_v_units));

  /* "celiagg/paint.pxi":106
 * 
 *         # Hold on to the allocated arrays for safety
 *         self._points = _points             # <<<<<<<<<<<<<<
 *         self._stops = _stops
 * 
 */
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v__points, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_points);
  __Pyx_DECREF(__pyx_v_self->_points);
  __pyx_v_self->_points = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "celiagg/paint.pxi":107
 *         # Hold on to the allocated arrays for safety
 *         self._points = _points
 *         self._stops = _stops             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v__stops, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_stops);
  __Pyx_DECREF(__pyx_v_self->_stops);
  __pyx_v_self->_stops = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "celiagg/paint.pxi":96
 *     cdef object _stops
 * 
 *     def __cinit__(self, double x1, double y1, double x2, double y2,             # <<<<<<<<<<<<<<
 *                   stops, GradientSpread spread, GradientUnits units):
 *         cdef double[::1] _points = _get_gradient_points((x1, y1, x2, y2), 4)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_AddTraceback("celiagg._celiagg.LinearGradientPaint.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v__points, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v__stops, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":109
 *         self._stops = _stops
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         points = self._points  # Not a deep copy!
 *         stops = self._stops  # Not a deep copy!
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint_2copy(((struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint_2copy(struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *__pyx_v_self) {
  PyObject *__pyx_v_points = NULL;
  PyObject *__pyx_v_stops = NULL;
  struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *__pyx_v_cpy = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/paint.pxi":110
 * 
 *     def copy(self):
 *         points = self._points  # Not a deep copy!             # <<<<<<<<<<<<<<
 *         stops = self._stops  # Not a deep copy!
 *         cpy = LinearGradientPaint(
 */
  __pyx_t_1 = __pyx_v_self->_points;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_points = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":111
 *     def copy(self):
 *         points = self._points  # Not a deep copy!
 *         stops = self._stops  # Not a deep copy!             # <<<<<<<<<<<<<<
 *         cpy = LinearGradientPaint(
 *             points[0], points[1], points[2], points[3],
 */
  __pyx_t_1 = __pyx_v_self->_stops;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_stops = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":113
 *         stops = self._stops  # Not a deep copy!
 *         cpy = LinearGradientPaint(
 *             points[0], points[1], points[2], points[3],             # <<<<<<<<<<<<<<
 *             stops, self._this.spread(), self._this.units()
 *         )
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_points, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_points, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_points, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_points, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "celiagg/paint.pxi":114
 *         cpy = LinearGradientPaint(
 *             points[0], points[1], points[2], points[3],
 *             stops, self._this.spread(), self._this.units()             # <<<<<<<<<<<<<<
 *         )
 *         cpy.transform = self.transform
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum__Paint_3a__3a_GradientSpread(__pyx_v_self->__pyx_base._this->spread()); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_enum__Paint_3a__3a_GradientUnits(__pyx_v_self->__pyx_base._this->units()); if (unlikely(!__pyx_t_6)) __PYX_ERR(7, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "celiagg/paint.pxi":112
 *         points = self._points  # Not a deep copy!
 *         stops = self._stops  # Not a deep copy!
 *         cpy = LinearGradientPaint(             # <<<<<<<<<<<<<<
 *             points[0], points[1], points[2], points[3],
 *             stops, self._this.spread(), self._this.units()
 */
  __pyx_t_7 = PyTuple_New(7); if (unlikely(!__pyx_t_7)) __PYX_ERR(7, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_stops);
  __Pyx_GIVEREF(__pyx_v_stops);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_v_stops);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 6, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_LinearGradientPaint), __pyx_t_7, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(7, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_cpy = ((struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "celiagg/paint.pxi":116
 *             stops, self._this.spread(), self._this.units()
 *         )
 *         cpy.transform = self.transform             # <<<<<<<<<<<<<<
 *         return cpy
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_transform); if (unlikely(!__pyx_t_6)) __PYX_ERR(7, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cpy), __pyx_n_s_transform, __pyx_t_6) < 0) __PYX_ERR(7, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "celiagg/paint.pxi":117
 *         )
 *         cpy.transform = self.transform
 *         return cpy             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cpy));
  __pyx_r = ((PyObject *)__pyx_v_cpy);
  goto __pyx_L0;

  /* "celiagg/paint.pxi":109
 *         self._stops = _stops
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         points = self._points  # Not a deep copy!
 *         stops = self._stops  # Not a deep copy!
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("celiagg._celiagg.LinearGradientPaint.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_points);
  __Pyx_XDECREF(__pyx_v_stops);
  __Pyx_XDECREF((PyObject *)__pyx_v_cpy);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint_4__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.LinearGradientPaint.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint_6__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_19LinearGradientPaint_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.LinearGradientPaint.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":136
 *     cdef object _stops
 * 
 *     def __cinit__(self, double cx, double cy, double r, double fx, double fy,             # <<<<<<<<<<<<<<
 *                   stops, GradientSpread spread, GradientUnits units):
 *         cdef double[::1] _points = _get_gradient_points([cx, cy, r, fx, fy], 5)
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_cx;
  double __pyx_v_cy;
  double __pyx_v_r;
  double __pyx_v_fx;
  double __pyx_v_fy;
  PyObject *__pyx_v_stops = 0;
  enum __pyx_t_7celiagg_8_celiagg_GradientSpread __pyx_v_spread;
  enum __pyx_t_7celiagg_8_celiagg_GradientUnits __pyx_v_units;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cx,&__pyx_n_s_cy,&__pyx_n_s_r,&__pyx_n_s_fx,&__pyx_n_s_fy,&__pyx_n_s_stops,&__pyx_n_s_spread,&__pyx_n_s_units,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cx)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 1); __PYX_ERR(7, 136, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_r)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 2); __PYX_ERR(7, 136, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 3); __PYX_ERR(7, 136, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 4); __PYX_ERR(7, 136, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stops)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 5); __PYX_ERR(7, 136, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_spread)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 6); __PYX_ERR(7, 136, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_units)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, 7); __PYX_ERR(7, 136, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(7, 136, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_cx = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_cx == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 136, __pyx_L3_error)
    __pyx_v_cy = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_cy == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 136, __pyx_L3_error)
    __pyx_v_r = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_r == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 136, __pyx_L3_error)
    __pyx_v_fx = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_fx == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 136, __pyx_L3_error)
    __pyx_v_fy = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_fy == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 136, __pyx_L3_error)
    __pyx_v_stops = values[5];
    __pyx_v_spread = ((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(values[6])); if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 137, __pyx_L3_error)
    __pyx_v_units = ((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(values[7])); if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 137, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(7, 136, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.RadialGradientPaint.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *)__pyx_v_self), __pyx_v_cx, __pyx_v_cy, __pyx_v_r, __pyx_v_fx, __pyx_v_fy, __pyx_v_stops, __pyx_v_spread, __pyx_v_units);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint___cinit__(struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *__pyx_v_self, double __pyx_v_cx, double __pyx_v_cy, double __pyx_v_r, double __pyx_v_fx, double __pyx_v_fy, PyObject *__pyx_v_stops, enum __pyx_t_7celiagg_8_celiagg_GradientSpread __pyx_v_spread, enum __pyx_t_7celiagg_8_celiagg_GradientUnits __pyx_v_units) {
  __Pyx_memviewslice __pyx_v__points = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v__stops = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/paint.pxi":138
 *     def __cinit__(self, double cx, double cy, double r, double fx, double fy,
 *                   stops, GradientSpread spread, GradientUnits units):
 *         cdef double[::1] _points = _get_gradient_points([cx, cy, r, fx, fy], 5)             # <<<<<<<<<<<<<<
 *         cdef double[:,::1] _stops = _get_gradient_stops(stops)
 *         self._this = new _paint.Paint(_enums.k_PaintTypeRadialGradient,
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_cx); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_cy); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_r); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_fx); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_fy); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyList_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(7, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_6, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_6, 4, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_f_7celiagg_8_celiagg__get_gradient_points(__pyx_t_6, __pyx_int_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(7, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v__points = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "celiagg/paint.pxi":139
 *                   stops, GradientSpread spread, GradientUnits units):
 *         cdef double[::1] _points = _get_gradient_points([cx, cy, r, fx, fy], 5)
 *         cdef double[:,::1] _stops = _get_gradient_stops(stops)             # <<<<<<<<<<<<<<
 *         self._this = new _paint.Paint(_enums.k_PaintTypeRadialGradient,
 *                                       &_points[0], _points.shape[0],
 */
  __pyx_t_5 = __pyx_f_7celiagg_8_celiagg__get_gradient_stops(__pyx_v_stops); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(7, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v__stops = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "celiagg/paint.pxi":141
 *         cdef double[:,::1] _stops = _get_gradient_stops(stops)
 *         self._this = new _paint.Paint(_enums.k_PaintTypeRadialGradient,
 *                                       &_points[0], _points.shape[0],             # <<<<<<<<<<<<<<
 *                                       &_stops[0][0], _stops.shape[0],
 *                                       spread, units)
 */
  __pyx_t_9 = 0;
  __pyx_t_10 = -1;
  if (__pyx_t_9 < 0) {
    __pyx_t_9 += __pyx_v__points.shape[0];
    if (unlikely(__pyx_t_9 < 0)) __pyx_t_10 = 0;
  } else if (unlikely(__pyx_t_9 >= __pyx_v__points.shape[0])) __pyx_t_10 = 0;
  if (unlikely(__pyx_t_10 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_10);
    __PYX_ERR(7, 141, __pyx_L1_error)
  }

  /* "celiagg/paint.pxi":142
 *         self._this = new _paint.Paint(_enums.k_PaintTypeRadialGradient,
 *                                       &_points[0], _points.shape[0],
 *                                       &_stops[0][0], _stops.shape[0],             # <<<<<<<<<<<<<<
 *                                       spread, units)
 * 
 */
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_10 = -1;
  if (__pyx_t_11 < 0) {
    __pyx_t_11 += __pyx_v__stops.shape[0];
    if (unlikely(__pyx_t_11 < 0)) __pyx_t_10 = 0;
  } else if (unlikely(__pyx_t_11 >= __pyx_v__stops.shape[0])) __pyx_t_10 = 0;
  if (__pyx_t_12 < 0) {
    __pyx_t_12 += __pyx_v__stops.shape[1];
    if (unlikely(__pyx_t_12 < 0)) __pyx_t_10 = 1;
  } else if (unlikely(__pyx_t_12 >= __pyx_v__stops.shape[1])) __pyx_t_10 = 1;
  if (unlikely(__pyx_t_10 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_10);
    __PYX_ERR(7, 142, __pyx_L1_error)
  }

  /* "celiagg/paint.pxi":140
 *         cdef double[::1] _points = _get_gradient_points([cx, cy, r, fx, fy], 5)
 *         cdef double[:,::1] _stops = _get_gradient_stops(stops)
 *         self._this = new _paint.Paint(_enums.k_PaintTypeRadialGradient,             # <<<<<<<<<<<<<<
 *                                       &_points[0], _points.shape[0],
 *                                       &_stops[0][0], _stops.shape[0],
 */
  __pyx_v_self->__pyx_base._this = new Paint(Paint::k_PaintTypeRadialGradient, (&(*((double *) ( /* dim=0 */ ((char *) (((double *) __pyx_v__points.data) + __pyx_t_9)) )))), (__pyx_v__points.shape[0]), (&(*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__stops.data + __pyx_t_11 * __pyx_v__stops.strides[0]) )) + __pyx_t_12)) )))), (__pyx_v__stops.shape[0]), ((enum Paint::GradientSpread)__pyx_v_spread), ((enum Paint::GradientUnits)__pyx_v_units));

  /* "celiagg/paint.pxi":146
 * 
 *         # Hold on to the allocated arrays for safety
 *         self._points = _points             # <<<<<<<<<<<<<<
 *         self._stops = _stops
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v__points, 1, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->_points);
  __Pyx_DECREF(__pyx_v_self->_points);
  __pyx_v_self->_points = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "celiagg/paint.pxi":147
 *         # Hold on to the allocated arrays for safety
 *         self._points = _points
 *         self._stops = _stops             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v__stops, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->_stops);
  __Pyx_DECREF(__pyx_v_self->_stops);
  __pyx_v_self->_stops = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "celiagg/paint.pxi":136
 *     cdef object _stops
 * 
 *     def __cinit__(self, double cx, double cy, double r, double fx, double fy,             # <<<<<<<<<<<<<<
 *                   stops, GradientSpread spread, GradientUnits units):
 *         cdef double[::1] _points = _get_gradient_points([cx, cy, r, fx, fy], 5)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_AddTraceback("celiagg._celiagg.RadialGradientPaint.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v__points, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v__stops, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":149
 *         self._stops = _stops
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         points = self._points  # Not a deep copy!
 *         stops = self._stops  # Not a deep copy!
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint_2copy(((struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint_2copy(struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *__pyx_v_self) {
  PyObject *__pyx_v_points = NULL;
  PyObject *__pyx_v_stops = NULL;
  struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *__pyx_v_cpy = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/paint.pxi":150
 * 
 *     def copy(self):
 *         points = self._points  # Not a deep copy!             # <<<<<<<<<<<<<<
 *         stops = self._stops  # Not a deep copy!
 *         cpy = RadialGradientPaint(
 */
  __pyx_t_1 = __pyx_v_self->_points;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_points = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":151
 *     def copy(self):
 *         points = self._points  # Not a deep copy!
 *         stops = self._stops  # Not a deep copy!             # <<<<<<<<<<<<<<
 *         cpy = RadialGradientPaint(
 *             points[0], points[1], points[2], points[3], points[4],
 */
  __pyx_t_1 = __pyx_v_self->_stops;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_stops = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":153
 *         stops = self._stops  # Not a deep copy!
 *         cpy = RadialGradientPaint(
 *             points[0], points[1], points[2], points[3], points[4],             # <<<<<<<<<<<<<<
 *             stops, self._this.spread(), self._this.units()
 *         )
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_points, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_points, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_points, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_points, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_points, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "celiagg/paint.pxi":154
 *         cpy = RadialGradientPaint(
 *             points[0], points[1], points[2], points[3], points[4],
 *             stops, self._this.spread(), self._this.units()             # <<<<<<<<<<<<<<
 *         )
 *         cpy.transform = self.transform
 */
  __pyx_t_6 = __Pyx_PyInt_From_enum__Paint_3a__3a_GradientSpread(__pyx_v_self->__pyx_base._this->spread()); if (unlikely(!__pyx_t_6)) __PYX_ERR(7, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_enum__Paint_3a__3a_GradientUnits(__pyx_v_self->__pyx_base._this->units()); if (unlikely(!__pyx_t_7)) __PYX_ERR(7, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "celiagg/paint.pxi":152
 *         points = self._points  # Not a deep copy!
 *         stops = self._stops  # Not a deep copy!
 *         cpy = RadialGradientPaint(             # <<<<<<<<<<<<<<
 *             points[0], points[1], points[2], points[3], points[4],
 *             stops, self._this.spread(), self._this.units()
 */
  __pyx_t_8 = PyTuple_New(8); if (unlikely(!__pyx_t_8)) __PYX_ERR(7, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_stops);
  __Pyx_GIVEREF(__pyx_v_stops);
  PyTuple_SET_ITEM(__pyx_t_8, 5, __pyx_v_stops);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_8, 6, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 7, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_RadialGradientPaint), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(7, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_cpy = ((struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "celiagg/paint.pxi":156
 *             stops, self._this.spread(), self._this.units()
 *         )
 *         cpy.transform = self.transform             # <<<<<<<<<<<<<<
 *         return cpy
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_transform); if (unlikely(!__pyx_t_7)) __PYX_ERR(7, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cpy), __pyx_n_s_transform, __pyx_t_7) < 0) __PYX_ERR(7, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "celiagg/paint.pxi":157
 *         )
 *         cpy.transform = self.transform
 *         return cpy             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cpy));
  __pyx_r = ((PyObject *)__pyx_v_cpy);
  goto __pyx_L0;

  /* "celiagg/paint.pxi":149
 *         self._stops = _stops
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         points = self._points  # Not a deep copy!
 *         stops = self._stops  # Not a deep copy!
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("celiagg._celiagg.RadialGradientPaint.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_points);
  __Pyx_XDECREF(__pyx_v_stops);
  __Pyx_XDECREF((PyObject *)__pyx_v_cpy);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint_4__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.RadialGradientPaint.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint_6__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_19RadialGradientPaint_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.RadialGradientPaint.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":170
 *     cdef Image img_obj
 * 
 *     def __cinit__(self, PatternStyle style, image):             # <<<<<<<<<<<<<<
 *         if not isinstance(image, Image):
 *             raise TypeError("image must be an Image instance")
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_12PatternPaint_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_12PatternPaint_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  enum __pyx_t_7celiagg_8_celiagg_PatternStyle __pyx_v_style;
  PyObject *__pyx_v_image = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_style,&__pyx_n_s_image,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_style)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(7, 170, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(7, 170, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_style = ((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(values[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 170, __pyx_L3_error)
    __pyx_v_image = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(7, 170, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.PatternPaint.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12PatternPaint___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)__pyx_v_self), __pyx_v_style, __pyx_v_image);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_12PatternPaint___cinit__(struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_PatternStyle __pyx_v_style, PyObject *__pyx_v_image) {
  struct __pyx_obj_7celiagg_8_celiagg_Image *__pyx_v_img = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/paint.pxi":171
 * 
 *     def __cinit__(self, PatternStyle style, image):
 *         if not isinstance(image, Image):             # <<<<<<<<<<<<<<
 *             raise TypeError("image must be an Image instance")
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_image, __pyx_ptype_7celiagg_8_celiagg_Image); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/paint.pxi":172
 *     def __cinit__(self, PatternStyle style, image):
 *         if not isinstance(image, Image):
 *             raise TypeError("image must be an Image instance")             # <<<<<<<<<<<<<<
 * 
 *         cdef Image img = <Image>image
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(7, 172, __pyx_L1_error)

    /* "celiagg/paint.pxi":171
 * 
 *     def __cinit__(self, PatternStyle style, image):
 *         if not isinstance(image, Image):             # <<<<<<<<<<<<<<
 *             raise TypeError("image must be an Image instance")
 * 
 */
  }

  /* "celiagg/paint.pxi":174
 *             raise TypeError("image must be an Image instance")
 * 
 *         cdef Image img = <Image>image             # <<<<<<<<<<<<<<
 *         self._this = new _paint.Paint(style, img._this)
 * 
 */
  __pyx_t_3 = __pyx_v_image;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_img = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/paint.pxi":175
 * 
 *         cdef Image img = <Image>image
 *         self._this = new _paint.Paint(style, img._this)             # <<<<<<<<<<<<<<
 * 
 *         # Hold a reference to the pattern image
 */
  __pyx_v_self->__pyx_base._this = new Paint(((enum Paint::PatternStyle)__pyx_v_style), __pyx_v_img->_this);

  /* "celiagg/paint.pxi":178
 * 
 *         # Hold a reference to the pattern image
 *         self.img_obj = img             # <<<<<<<<<<<<<<
 *         self.style = style
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_img));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_img));
  __Pyx_GOTREF(__pyx_v_self->img_obj);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->img_obj));
  __pyx_v_self->img_obj = __pyx_v_img;

  /* "celiagg/paint.pxi":179
 *         # Hold a reference to the pattern image
 *         self.img_obj = img
 *         self.style = style             # <<<<<<<<<<<<<<
 * 
 *     def _with_format(self, PixelFormat fmt):
 */
  __pyx_v_self->style = __pyx_v_style;

  /* "celiagg/paint.pxi":170
 *     cdef Image img_obj
 * 
 *     def __cinit__(self, PatternStyle style, image):             # <<<<<<<<<<<<<<
 *         if not isinstance(image, Image):
 *             raise TypeError("image must be an Image instance")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.PatternPaint.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_img);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":181
 *         self.style = style
 * 
 *     def _with_format(self, PixelFormat fmt):             # <<<<<<<<<<<<<<
 *         if self.img_obj.format == fmt:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12PatternPaint_3_with_format(PyObject *__pyx_v_self, PyObject *__pyx_arg_fmt); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12PatternPaint_3_with_format(PyObject *__pyx_v_self, PyObject *__pyx_arg_fmt) {
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_fmt;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_with_format (wrapper)", 0);
  assert(__pyx_arg_fmt); {
    __pyx_v_fmt = ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_arg_fmt)); if (unlikely(PyErr_Occurred())) __PYX_ERR(7, 181, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.PatternPaint._with_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12PatternPaint_2_with_format(((struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)__pyx_v_self), ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)__pyx_v_fmt));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12PatternPaint_2_with_format(struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self, enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_v_fmt) {
  PyObject *__pyx_v_fmt_image = NULL;
  struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_fmt_pattern = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_with_format", 0);

  /* "celiagg/paint.pxi":182
 * 
 *     def _with_format(self, PixelFormat fmt):
 *         if self.img_obj.format == fmt:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->img_obj), __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_fmt); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(7, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "celiagg/paint.pxi":183
 *     def _with_format(self, PixelFormat fmt):
 *         if self.img_obj.format == fmt:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         fmt_image = convert_image(self.img_obj, fmt,
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "celiagg/paint.pxi":182
 * 
 *     def _with_format(self, PixelFormat fmt):
 *         if self.img_obj.format == fmt:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "celiagg/paint.pxi":185
 *             return self
 * 
 *         fmt_image = convert_image(self.img_obj, fmt,             # <<<<<<<<<<<<<<
 *                                   bottom_up=self.img_obj.bottom_up)
 *         fmt_pattern = PatternPaint(self.style, fmt_image)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_convert_image); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_fmt); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->img_obj));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->img_obj));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self->img_obj));
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "celiagg/paint.pxi":186
 * 
 *         fmt_image = convert_image(self.img_obj, fmt,
 *                                   bottom_up=self.img_obj.bottom_up)             # <<<<<<<<<<<<<<
 *         fmt_pattern = PatternPaint(self.style, fmt_image)
 *         fmt_pattern.transform = self.transform
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->img_obj->bottom_up); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_bottom_up, __pyx_t_5) < 0) __PYX_ERR(7, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "celiagg/paint.pxi":185
 *             return self
 * 
 *         fmt_image = convert_image(self.img_obj, fmt,             # <<<<<<<<<<<<<<
 *                                   bottom_up=self.img_obj.bottom_up)
 *         fmt_pattern = PatternPaint(self.style, fmt_image)
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_fmt_image = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "celiagg/paint.pxi":187
 *         fmt_image = convert_image(self.img_obj, fmt,
 *                                   bottom_up=self.img_obj.bottom_up)
 *         fmt_pattern = PatternPaint(self.style, fmt_image)             # <<<<<<<<<<<<<<
 *         fmt_pattern.transform = self.transform
 *         return fmt_pattern
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(__pyx_v_self->style); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_fmt_image);
  __Pyx_GIVEREF(__pyx_v_fmt_image);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_fmt_image);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_PatternPaint), __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_fmt_pattern = ((struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "celiagg/paint.pxi":188
 *                                   bottom_up=self.img_obj.bottom_up)
 *         fmt_pattern = PatternPaint(self.style, fmt_image)
 *         fmt_pattern.transform = self.transform             # <<<<<<<<<<<<<<
 *         return fmt_pattern
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_transform); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_fmt_pattern), __pyx_n_s_transform, __pyx_t_5) < 0) __PYX_ERR(7, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "celiagg/paint.pxi":189
 *         fmt_pattern = PatternPaint(self.style, fmt_image)
 *         fmt_pattern.transform = self.transform
 *         return fmt_pattern             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_fmt_pattern));
  __pyx_r = ((PyObject *)__pyx_v_fmt_pattern);
  goto __pyx_L0;

  /* "celiagg/paint.pxi":181
 *         self.style = style
 * 
 *     def _with_format(self, PixelFormat fmt):             # <<<<<<<<<<<<<<
 *         if self.img_obj.format == fmt:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg.PatternPaint._with_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fmt_image);
  __Pyx_XDECREF((PyObject *)__pyx_v_fmt_pattern);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":191
 *         return fmt_pattern
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         img = self.img_obj.copy()
 *         cpy = PatternPaint(self.style, img)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12PatternPaint_5copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12PatternPaint_5copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12PatternPaint_4copy(((struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12PatternPaint_4copy(struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self) {
  PyObject *__pyx_v_img = NULL;
  struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_cpy = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/paint.pxi":192
 * 
 *     def copy(self):
 *         img = self.img_obj.copy()             # <<<<<<<<<<<<<<
 *         cpy = PatternPaint(self.style, img)
 *         cpy.transform = self.transform
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->img_obj), __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_img = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":193
 *     def copy(self):
 *         img = self.img_obj.copy()
 *         cpy = PatternPaint(self.style, img)             # <<<<<<<<<<<<<<
 *         cpy.transform = self.transform
 *         return cpy
 */
  __pyx_t_1 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(__pyx_v_self->style); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_img);
  __Pyx_GIVEREF(__pyx_v_img);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_img);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_PatternPaint), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_cpy = ((struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":194
 *         img = self.img_obj.copy()
 *         cpy = PatternPaint(self.style, img)
 *         cpy.transform = self.transform             # <<<<<<<<<<<<<<
 *         return cpy
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_transform); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cpy), __pyx_n_s_transform, __pyx_t_1) < 0) __PYX_ERR(7, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":195
 *         cpy = PatternPaint(self.style, img)
 *         cpy.transform = self.transform
 *         return cpy             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cpy));
  __pyx_r = ((PyObject *)__pyx_v_cpy);
  goto __pyx_L0;

  /* "celiagg/paint.pxi":191
 *         return fmt_pattern
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         img = self.img_obj.copy()
 *         cpy = PatternPaint(self.style, img)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.PatternPaint.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_img);
  __Pyx_XDECREF((PyObject *)__pyx_v_cpy);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12PatternPaint_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12PatternPaint_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12PatternPaint_6__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12PatternPaint_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.PatternPaint.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12PatternPaint_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12PatternPaint_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12PatternPaint_8__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12PatternPaint_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__49, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.PatternPaint.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":207
 *     :param a: Alpha value in [0, 1] (defaults to 1.0)
 *     """
 *     def __cinit__(self, double r, double g, double b, double a=1.0):             # <<<<<<<<<<<<<<
 *         self._this = new _paint.Paint(r, g, b, a)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_r;
  double __pyx_v_g;
  double __pyx_v_b;
  double __pyx_v_a;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_r,&__pyx_n_s_g,&__pyx_n_s_b,&__pyx_n_s_a,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_r)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_g)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 4, 1); __PYX_ERR(7, 207, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 4, 2); __PYX_ERR(7, 207, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(7, 207, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_r = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_r == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 207, __pyx_L3_error)
    __pyx_v_g = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_g == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 207, __pyx_L3_error)
    __pyx_v_b = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_b == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 207, __pyx_L3_error)
    if (values[3]) {
      __pyx_v_a = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_a == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 207, __pyx_L3_error)
    } else {
      __pyx_v_a = ((double)1.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(7, 207, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self), __pyx_v_r, __pyx_v_g, __pyx_v_b, __pyx_v_a);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint___cinit__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_r, double __pyx_v_g, double __pyx_v_b, double __pyx_v_a) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/paint.pxi":208
 *     """
 *     def __cinit__(self, double r, double g, double b, double a=1.0):
 *         self._this = new _paint.Paint(r, g, b, a)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_v_self->__pyx_base._this = new Paint(__pyx_v_r, __pyx_v_g, __pyx_v_b, __pyx_v_a);

  /* "celiagg/paint.pxi":207
 *     :param a: Alpha value in [0, 1] (defaults to 1.0)
 *     """
 *     def __cinit__(self, double r, double g, double b, double a=1.0):             # <<<<<<<<<<<<<<
 *         self._this = new _paint.Paint(r, g, b, a)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":210
 *         self._this = new _paint.Paint(r, g, b, a)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         name = type(self).__name__
 *         return "{}({}, {}, {}, {})".format(
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_3__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_2__repr__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_2__repr__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "celiagg/paint.pxi":211
 * 
 *     def __repr__(self):
 *         name = type(self).__name__             # <<<<<<<<<<<<<<
 *         return "{}({}, {}, {}, {})".format(
 *             name, self._this.r(), self._this.g(), self._this.b(),
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/paint.pxi":212
 *     def __repr__(self):
 *         name = type(self).__name__
 *         return "{}({}, {}, {}, {})".format(             # <<<<<<<<<<<<<<
 *             name, self._this.r(), self._this.g(), self._this.b(),
 *             self._this.a()
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s__3, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "celiagg/paint.pxi":213
 *         name = type(self).__name__
 *         return "{}({}, {}, {}, {})".format(
 *             name, self._this.r(), self._this.g(), self._this.b(),             # <<<<<<<<<<<<<<
 *             self._this.a()
 *         )
 */
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->r()); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->g()); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->b()); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "celiagg/paint.pxi":214
 *         return "{}({}, {}, {}, {})".format(
 *             name, self._this.r(), self._this.g(), self._this.b(),
 *             self._this.a()             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->a()); if (unlikely(!__pyx_t_6)) __PYX_ERR(7, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[6] = {__pyx_t_7, __pyx_v_name, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 212, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[6] = {__pyx_t_7, __pyx_v_name, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 5+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 212, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(5+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(7, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_name);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_8, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 4+__pyx_t_8, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/paint.pxi":210
 *         self._this = new _paint.Paint(r, g, b, a)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         name = type(self).__name__
 *         return "{}({}, {}, {}, {})".format(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":217
 *         )
 * 
 *     def __richcmp__(SolidPaint self, SolidPaint other, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return (self._this.a() == other._this.a() and
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7celiagg_8_celiagg_SolidPaint, 1, "other", 0))) __PYX_ERR(7, 217, __pyx_L1_error)
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_4__richcmp__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self), ((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_4__richcmp__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_v_msg = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "celiagg/paint.pxi":218
 * 
 *     def __richcmp__(SolidPaint self, SolidPaint other, int op):
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return (self._this.a() == other._this.a() and
 *                     self._this.r() == other._this.r() and
 */
  __pyx_t_1 = ((__pyx_v_op == 2) != 0);
  if (likely(__pyx_t_1)) {

    /* "celiagg/paint.pxi":219
 *     def __richcmp__(SolidPaint self, SolidPaint other, int op):
 *         if op == 2:  # ==
 *             return (self._this.a() == other._this.a() and             # <<<<<<<<<<<<<<
 *                     self._this.r() == other._this.r() and
 *                     self._this.g() == other._this.g() and
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = (__pyx_v_self->__pyx_base._this->a() == __pyx_v_other->__pyx_base._this->a());
    if (__pyx_t_1) {
    } else {
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 219, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L4_bool_binop_done;
    }

    /* "celiagg/paint.pxi":220
 *         if op == 2:  # ==
 *             return (self._this.a() == other._this.a() and
 *                     self._this.r() == other._this.r() and             # <<<<<<<<<<<<<<
 *                     self._this.g() == other._this.g() and
 *                     self._this.b() == other._this.b())
 */
    __pyx_t_1 = (__pyx_v_self->__pyx_base._this->r() == __pyx_v_other->__pyx_base._this->r());
    if (__pyx_t_1) {
    } else {
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L4_bool_binop_done;
    }

    /* "celiagg/paint.pxi":221
 *             return (self._this.a() == other._this.a() and
 *                     self._this.r() == other._this.r() and
 *                     self._this.g() == other._this.g() and             # <<<<<<<<<<<<<<
 *                     self._this.b() == other._this.b())
 *         else:
 */
    __pyx_t_1 = (__pyx_v_self->__pyx_base._this->g() == __pyx_v_other->__pyx_base._this->g());
    if (__pyx_t_1) {
    } else {
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 221, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L4_bool_binop_done;
    }

    /* "celiagg/paint.pxi":222
 *                     self._this.r() == other._this.r() and
 *                     self._this.g() == other._this.g() and
 *                     self._this.b() == other._this.b())             # <<<<<<<<<<<<<<
 *         else:
 *             msg = "That type of comparison is not implemented for SolidPaint"
 */
    __pyx_t_1 = (__pyx_v_self->__pyx_base._this->b() == __pyx_v_other->__pyx_base._this->b());
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_L4_bool_binop_done:;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "celiagg/paint.pxi":218
 * 
 *     def __richcmp__(SolidPaint self, SolidPaint other, int op):
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return (self._this.a() == other._this.a() and
 *                     self._this.r() == other._this.r() and
 */
  }

  /* "celiagg/paint.pxi":224
 *                     self._this.b() == other._this.b())
 *         else:
 *             msg = "That type of comparison is not implemented for SolidPaint"             # <<<<<<<<<<<<<<
 *             raise NotImplementedError(msg)
 * 
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_kp_s_That_type_of_comparison_is_not_i_2);
    __pyx_v_msg = __pyx_kp_s_That_type_of_comparison_is_not_i_2;

    /* "celiagg/paint.pxi":225
 *         else:
 *             msg = "That type of comparison is not implemented for SolidPaint"
 *             raise NotImplementedError(msg)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_NotImplementedError, __pyx_v_msg); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(7, 225, __pyx_L1_error)
  }

  /* "celiagg/paint.pxi":217
 *         )
 * 
 *     def __richcmp__(SolidPaint self, SolidPaint other, int op):             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return (self._this.a() == other._this.a() and
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":227
 *             raise NotImplementedError(msg)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cpy = SolidPaint(self._this.r(), self._this.g(), self._this.b(),
 *                          self._this.a())
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_7copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_7copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_6copy(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_6copy(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self) {
  struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_cpy = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/paint.pxi":228
 * 
 *     def copy(self):
 *         cpy = SolidPaint(self._this.r(), self._this.g(), self._this.b(),             # <<<<<<<<<<<<<<
 *                          self._this.a())
 *         cpy.transform = self.transform
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->r()); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->g()); if (unlikely(!__pyx_t_2)) __PYX_ERR(7, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->b()); if (unlikely(!__pyx_t_3)) __PYX_ERR(7, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "celiagg/paint.pxi":229
 *     def copy(self):
 *         cpy = SolidPaint(self._this.r(), self._this.g(), self._this.b(),
 *                          self._this.a())             # <<<<<<<<<<<<<<
 *         cpy.transform = self.transform
 *         return cpy
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->a()); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "celiagg/paint.pxi":228
 * 
 *     def copy(self):
 *         cpy = SolidPaint(self._this.r(), self._this.g(), self._this.b(),             # <<<<<<<<<<<<<<
 *                          self._this.a())
 *         cpy.transform = self.transform
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(7, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_SolidPaint), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_cpy = ((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "celiagg/paint.pxi":230
 *         cpy = SolidPaint(self._this.r(), self._this.g(), self._this.b(),
 *                          self._this.a())
 *         cpy.transform = self.transform             # <<<<<<<<<<<<<<
 *         return cpy
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_transform); if (unlikely(!__pyx_t_4)) __PYX_ERR(7, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cpy), __pyx_n_s_transform, __pyx_t_4) < 0) __PYX_ERR(7, 230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "celiagg/paint.pxi":231
 *                          self._this.a())
 *         cpy.transform = self.transform
 *         return cpy             # <<<<<<<<<<<<<<
 * 
 *     property a:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cpy));
  __pyx_r = ((PyObject *)__pyx_v_cpy);
  goto __pyx_L0;

  /* "celiagg/paint.pxi":227
 *             raise NotImplementedError(msg)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cpy = SolidPaint(self._this.r(), self._this.g(), self._this.b(),
 *                          self._this.a())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cpy);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":234
 * 
 *     property a:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.a()
 *         def __set__(self, double value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_1a_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_1a_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1a___get__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_1a___get__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/paint.pxi":235
 *     property a:
 *         def __get__(self):
 *             return self._this.a()             # <<<<<<<<<<<<<<
 *         def __set__(self, double value):
 *             self._this.a(value)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->a()); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/paint.pxi":234
 * 
 *     property a:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.a()
 *         def __set__(self, double value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.a.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":236
 *         def __get__(self):
 *             return self._this.a()
 *         def __set__(self, double value):             # <<<<<<<<<<<<<<
 *             self._this.a(value)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1a_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1a_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 236, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.a.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1a_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1a_2__set__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/paint.pxi":237
 *             return self._this.a()
 *         def __set__(self, double value):
 *             self._this.a(value)             # <<<<<<<<<<<<<<
 * 
 *     property r:
 */
  __pyx_v_self->__pyx_base._this->a(__pyx_v_value);

  /* "celiagg/paint.pxi":236
 *         def __get__(self):
 *             return self._this.a()
 *         def __set__(self, double value):             # <<<<<<<<<<<<<<
 *             self._this.a(value)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":240
 * 
 *     property r:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.r()
 *         def __set__(self, double value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_1r_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_1r_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1r___get__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_1r___get__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/paint.pxi":241
 *     property r:
 *         def __get__(self):
 *             return self._this.r()             # <<<<<<<<<<<<<<
 *         def __set__(self, double value):
 *             self._this.r(value)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->r()); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/paint.pxi":240
 * 
 *     property r:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.r()
 *         def __set__(self, double value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.r.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":242
 *         def __get__(self):
 *             return self._this.r()
 *         def __set__(self, double value):             # <<<<<<<<<<<<<<
 *             self._this.r(value)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1r_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1r_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 242, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.r.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1r_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1r_2__set__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/paint.pxi":243
 *             return self._this.r()
 *         def __set__(self, double value):
 *             self._this.r(value)             # <<<<<<<<<<<<<<
 * 
 *     property g:
 */
  __pyx_v_self->__pyx_base._this->r(__pyx_v_value);

  /* "celiagg/paint.pxi":242
 *         def __get__(self):
 *             return self._this.r()
 *         def __set__(self, double value):             # <<<<<<<<<<<<<<
 *             self._this.r(value)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":246
 * 
 *     property g:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.g()
 *         def __set__(self, double value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_1g_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_1g_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1g___get__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_1g___get__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/paint.pxi":247
 *     property g:
 *         def __get__(self):
 *             return self._this.g()             # <<<<<<<<<<<<<<
 *         def __set__(self, double value):
 *             self._this.g(value)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->g()); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/paint.pxi":246
 * 
 *     property g:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.g()
 *         def __set__(self, double value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.g.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":248
 *         def __get__(self):
 *             return self._this.g()
 *         def __set__(self, double value):             # <<<<<<<<<<<<<<
 *             self._this.g(value)
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1g_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1g_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 248, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.g.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1g_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1g_2__set__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/paint.pxi":249
 *             return self._this.g()
 *         def __set__(self, double value):
 *             self._this.g(value)             # <<<<<<<<<<<<<<
 * 
 *     property b:
 */
  __pyx_v_self->__pyx_base._this->g(__pyx_v_value);

  /* "celiagg/paint.pxi":248
 *         def __get__(self):
 *             return self._this.g()
 *         def __set__(self, double value):             # <<<<<<<<<<<<<<
 *             self._this.g(value)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":252
 * 
 *     property b:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.b()
 *         def __set__(self, double value):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_1b_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_1b_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1b___get__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_1b___get__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/paint.pxi":253
 *     property b:
 *         def __get__(self):
 *             return self._this.b()             # <<<<<<<<<<<<<<
 *         def __set__(self, double value):
 *             self._this.b(value)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base._this->b()); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/paint.pxi":252
 * 
 *     property b:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.b()
 *         def __set__(self, double value):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.b.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/paint.pxi":254
 *         def __get__(self):
 *             return self._this.b()
 *         def __set__(self, double value):             # <<<<<<<<<<<<<<
 *             self._this.b(value)
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1b_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1b_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(7, 254, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.b.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1b_2__set__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_10SolidPaint_1b_2__set__(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "celiagg/paint.pxi":255
 *             return self._this.b()
 *         def __set__(self, double value):
 *             self._this.b(value)             # <<<<<<<<<<<<<<
 */
  __pyx_v_self->__pyx_base._this->b(__pyx_v_value);

  /* "celiagg/paint.pxi":254
 *         def __get__(self):
 *             return self._this.b()
 *         def __set__(self, double value):             # <<<<<<<<<<<<<<
 *             self._this.b(value)
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_8__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_10SolidPaint_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_10SolidPaint_10__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_10SolidPaint_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_SolidPaint *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__51, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.SolidPaint.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":46
 *     cdef _transform.trans_affine* _this
 * 
 *     def __init__(self, double sx=1.0, double shy=0.0, double shx=0.0,             # <<<<<<<<<<<<<<
 *                  double sy=1.0, double tx=0.0, double ty=0.0):
 *         self._this.sx = sx
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_9Transform_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_9Transform_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_sx;
  double __pyx_v_shy;
  double __pyx_v_shx;
  double __pyx_v_sy;
  double __pyx_v_tx;
  double __pyx_v_ty;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sx,&__pyx_n_s_shy,&__pyx_n_s_shx,&__pyx_n_s_sy,&__pyx_n_s_tx,&__pyx_n_s_ty,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sx);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shy);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shx);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sy);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tx);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ty);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(8, 46, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_sx = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_sx == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 46, __pyx_L3_error)
    } else {
      __pyx_v_sx = ((double)1.0);
    }
    if (values[1]) {
      __pyx_v_shy = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_shy == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 46, __pyx_L3_error)
    } else {
      __pyx_v_shy = ((double)0.0);
    }
    if (values[2]) {
      __pyx_v_shx = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_shx == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 46, __pyx_L3_error)
    } else {
      __pyx_v_shx = ((double)0.0);
    }
    if (values[3]) {
      __pyx_v_sy = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_sy == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 47, __pyx_L3_error)
    } else {
      __pyx_v_sy = ((double)1.0);
    }
    if (values[4]) {
      __pyx_v_tx = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_tx == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 47, __pyx_L3_error)
    } else {
      __pyx_v_tx = ((double)0.0);
    }
    if (values[5]) {
      __pyx_v_ty = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_ty == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 47, __pyx_L3_error)
    } else {
      __pyx_v_ty = ((double)0.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(8, 46, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Transform.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform___init__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), __pyx_v_sx, __pyx_v_shy, __pyx_v_shx, __pyx_v_sy, __pyx_v_tx, __pyx_v_ty);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_9Transform___init__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_sx, double __pyx_v_shy, double __pyx_v_shx, double __pyx_v_sy, double __pyx_v_tx, double __pyx_v_ty) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "celiagg/transform.pxi":48
 *     def __init__(self, double sx=1.0, double shy=0.0, double shx=0.0,
 *                  double sy=1.0, double tx=0.0, double ty=0.0):
 *         self._this.sx = sx             # <<<<<<<<<<<<<<
 *         self._this.sy = sy
 *         self._this.shx = shx
 */
  __pyx_v_self->_this->sx = __pyx_v_sx;

  /* "celiagg/transform.pxi":49
 *                  double sy=1.0, double tx=0.0, double ty=0.0):
 *         self._this.sx = sx
 *         self._this.sy = sy             # <<<<<<<<<<<<<<
 *         self._this.shx = shx
 *         self._this.shy = shy
 */
  __pyx_v_self->_this->sy = __pyx_v_sy;

  /* "celiagg/transform.pxi":50
 *         self._this.sx = sx
 *         self._this.sy = sy
 *         self._this.shx = shx             # <<<<<<<<<<<<<<
 *         self._this.shy = shy
 *         self._this.tx = tx
 */
  __pyx_v_self->_this->shx = __pyx_v_shx;

  /* "celiagg/transform.pxi":51
 *         self._this.sy = sy
 *         self._this.shx = shx
 *         self._this.shy = shy             # <<<<<<<<<<<<<<
 *         self._this.tx = tx
 *         self._this.ty = ty
 */
  __pyx_v_self->_this->shy = __pyx_v_shy;

  /* "celiagg/transform.pxi":52
 *         self._this.shx = shx
 *         self._this.shy = shy
 *         self._this.tx = tx             # <<<<<<<<<<<<<<
 *         self._this.ty = ty
 * 
 */
  __pyx_v_self->_this->tx = __pyx_v_tx;

  /* "celiagg/transform.pxi":53
 *         self._this.shy = shy
 *         self._this.tx = tx
 *         self._this.ty = ty             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
 */
  __pyx_v_self->_this->ty = __pyx_v_ty;

  /* "celiagg/transform.pxi":46
 *     cdef _transform.trans_affine* _this
 * 
 *     def __init__(self, double sx=1.0, double shy=0.0, double shx=0.0,             # <<<<<<<<<<<<<<
 *                  double sy=1.0, double tx=0.0, double ty=0.0):
 *         self._this.sx = sx
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":55
 *         self._this.ty = ty
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._this = new _transform.trans_affine()
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_9Transform_3__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_9Transform_3__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_2__cinit__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_9Transform_2__cinit__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/transform.pxi":56
 * 
 *     def __cinit__(self):
 *         self._this = new _transform.trans_affine()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->_this = new agg::trans_affine();

  /* "celiagg/transform.pxi":55
 *         self._this.ty = ty
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._this = new _transform.trans_affine()
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":58
 *         self._this = new _transform.trans_affine()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

/* Python wrapper */
static void __pyx_pw_7celiagg_8_celiagg_9Transform_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7celiagg_8_celiagg_9Transform_5__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7celiagg_8_celiagg_9Transform_4__dealloc__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7celiagg_8_celiagg_9Transform_4__dealloc__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "celiagg/transform.pxi":59
 * 
 *     def __dealloc__(self):
 *         del self._this             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  delete __pyx_v_self->_this;

  /* "celiagg/transform.pxi":58
 *         self._this = new _transform.trans_affine()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "celiagg/transform.pxi":61
 *         del self._this
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         name = type(self).__name__
 *         return "{}({}, {}, {}, {}, {}, {})".format(
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_7__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_7__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_6__repr__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_6__repr__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "celiagg/transform.pxi":62
 * 
 *     def __repr__(self):
 *         name = type(self).__name__             # <<<<<<<<<<<<<<
 *         return "{}({}, {}, {}, {}, {}, {})".format(
 *             name, self._this.sx, self._this.shy, self._this.shx, self._this.sy,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))), __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/transform.pxi":63
 *     def __repr__(self):
 *         name = type(self).__name__
 *         return "{}({}, {}, {}, {}, {}, {})".format(             # <<<<<<<<<<<<<<
 *             name, self._this.sx, self._this.shy, self._this.shx, self._this.sy,
 *             self._this.tx, self._this.ty
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s__52, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "celiagg/transform.pxi":64
 *         name = type(self).__name__
 *         return "{}({}, {}, {}, {}, {}, {})".format(
 *             name, self._this.sx, self._this.shy, self._this.shx, self._this.sy,             # <<<<<<<<<<<<<<
 *             self._this.tx, self._this.ty
 *         )
 */
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->_this->sx); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->_this->shy); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->_this->shx); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->_this->sy); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "celiagg/transform.pxi":65
 *         return "{}({}, {}, {}, {}, {}, {})".format(
 *             name, self._this.sx, self._this.shy, self._this.shx, self._this.sy,
 *             self._this.tx, self._this.ty             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_self->_this->tx); if (unlikely(!__pyx_t_7)) __PYX_ERR(8, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyFloat_FromDouble(__pyx_v_self->_this->ty); if (unlikely(!__pyx_t_8)) __PYX_ERR(8, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[8] = {__pyx_t_9, __pyx_v_name, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 7+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 63, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[8] = {__pyx_t_9, __pyx_v_name, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 7+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 63, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  {
    __pyx_t_11 = PyTuple_New(7+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(8, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__pyx_t_9) {
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_9); __pyx_t_9 = NULL;
    }
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_v_name);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_10, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_11, 3+__pyx_t_10, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_11, 4+__pyx_t_10, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_11, 5+__pyx_t_10, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_11, 6+__pyx_t_10, __pyx_t_8);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":61
 *         del self._this
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         name = type(self).__name__
 *         return "{}({}, {}, {}, {}, {}, {})".format(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":69
 * 
 *     property sx:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.sx
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_2sx_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_2sx_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_2sx___get__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_2sx___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/transform.pxi":70
 *     property sx:
 *         def __get__(self):
 *             return self._this.sx             # <<<<<<<<<<<<<<
 * 
 *     property sy:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->sx); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":69
 * 
 *     property sx:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.sx
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.sx.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":73
 * 
 *     property sy:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.sy
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_2sy_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_2sy_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_2sy___get__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_2sy___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/transform.pxi":74
 *     property sy:
 *         def __get__(self):
 *             return self._this.sy             # <<<<<<<<<<<<<<
 * 
 *     property shx:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->sy); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":73
 * 
 *     property sy:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.sy
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.sy.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":77
 * 
 *     property shx:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.shx
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_3shx_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_3shx_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_3shx___get__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_3shx___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/transform.pxi":78
 *     property shx:
 *         def __get__(self):
 *             return self._this.shx             # <<<<<<<<<<<<<<
 * 
 *     property shy:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->shx); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":77
 * 
 *     property shx:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.shx
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.shx.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":81
 * 
 *     property shy:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.shy
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_3shy_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_3shy_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_3shy___get__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_3shy___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/transform.pxi":82
 *     property shy:
 *         def __get__(self):
 *             return self._this.shy             # <<<<<<<<<<<<<<
 * 
 *     property tx:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->shy); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":81
 * 
 *     property shy:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.shy
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.shy.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":85
 * 
 *     property tx:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.tx
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_2tx_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_2tx_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_2tx___get__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_2tx___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/transform.pxi":86
 *     property tx:
 *         def __get__(self):
 *             return self._this.tx             # <<<<<<<<<<<<<<
 * 
 *     property ty:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->tx); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":85
 * 
 *     property tx:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.tx
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.tx.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":89
 * 
 *     property ty:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.ty
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_2ty_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_2ty_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_2ty___get__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_2ty___get__(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "celiagg/transform.pxi":90
 *     property ty:
 *         def __get__(self):
 *             return self._this.ty             # <<<<<<<<<<<<<<
 * 
 *     cdef _assign_obj(self, _transform.trans_affine trans):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->ty); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":89
 * 
 *     property ty:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._this.ty
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.ty.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":92
 *             return self._this.ty
 * 
 *     cdef _assign_obj(self, _transform.trans_affine trans):             # <<<<<<<<<<<<<<
 *         self._this.reset()
 *         self._this.multiply(trans)
 */

static PyObject *__pyx_f_7celiagg_8_celiagg_9Transform__assign_obj(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, agg::trans_affine __pyx_v_trans) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_assign_obj", 0);

  /* "celiagg/transform.pxi":93
 * 
 *     cdef _assign_obj(self, _transform.trans_affine trans):
 *         self._this.reset()             # <<<<<<<<<<<<<<
 *         self._this.multiply(trans)
 * 
 */
  (void)(__pyx_v_self->_this->reset());

  /* "celiagg/transform.pxi":94
 *     cdef _assign_obj(self, _transform.trans_affine trans):
 *         self._this.reset()
 *         self._this.multiply(trans)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  (void)(__pyx_v_self->_this->multiply(__pyx_v_trans));

  /* "celiagg/transform.pxi":92
 *             return self._this.ty
 * 
 *     cdef _assign_obj(self, _transform.trans_affine trans):             # <<<<<<<<<<<<<<
 *         self._this.reset()
 *         self._this.multiply(trans)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":96
 *         self._this.multiply(trans)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """copy()
 *         Returns a deep copy of the object.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_9copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_8copy[] = "copy()\n        Returns a deep copy of the object.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_9copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_8copy(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_8copy(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/transform.pxi":100
 *         Returns a deep copy of the object.
 *         """
 *         return Transform(self._this.sx, self._this.shy, self._this.shx,             # <<<<<<<<<<<<<<
 *                          self._this.sy, self._this.tx, self._this.ty)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_this->sx); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->_this->shy); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->_this->shx); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "celiagg/transform.pxi":101
 *         """
 *         return Transform(self._this.sx, self._this.shy, self._this.shx,
 *                          self._this.sy, self._this.tx, self._this.ty)             # <<<<<<<<<<<<<<
 * 
 *     def reset(self):
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->_this->sy); if (unlikely(!__pyx_t_4)) __PYX_ERR(8, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->_this->tx); if (unlikely(!__pyx_t_5)) __PYX_ERR(8, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->_this->ty); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "celiagg/transform.pxi":100
 *         Returns a deep copy of the object.
 *         """
 *         return Transform(self._this.sx, self._this.shy, self._this.shx,             # <<<<<<<<<<<<<<
 *                          self._this.sy, self._this.tx, self._this.ty)
 * 
 */
  __pyx_t_7 = PyTuple_New(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(8, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_Transform), __pyx_t_7, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(8, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":96
 *         self._this.multiply(trans)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """copy()
 *         Returns a deep copy of the object.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":103
 *                          self._this.sy, self._this.tx, self._this.ty)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         """reset()
 *         Resets to the identity transform.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_11reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_10reset[] = "reset()\n        Resets to the identity transform.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_11reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_10reset(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_10reset(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset", 0);

  /* "celiagg/transform.pxi":107
 *         Resets to the identity transform.
 *         """
 *         self._this.reset()             # <<<<<<<<<<<<<<
 * 
 *     def invert(self):
 */
  (void)(__pyx_v_self->_this->reset());

  /* "celiagg/transform.pxi":103
 *                          self._this.sy, self._this.tx, self._this.ty)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         """reset()
 *         Resets to the identity transform.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":109
 *         self._this.reset()
 * 
 *     def invert(self):             # <<<<<<<<<<<<<<
 *         """invert()
 *         Inverts the matrix.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_13invert(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_12invert[] = "invert()\n        Inverts the matrix.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_13invert(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("invert (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_12invert(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_12invert(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("invert", 0);

  /* "celiagg/transform.pxi":113
 *         Inverts the matrix.
 *         """
 *         self._this.invert()             # <<<<<<<<<<<<<<
 * 
 *     def scale(self, double sx, double sy):
 */
  (void)(__pyx_v_self->_this->invert());

  /* "celiagg/transform.pxi":109
 *         self._this.reset()
 * 
 *     def invert(self):             # <<<<<<<<<<<<<<
 *         """invert()
 *         Inverts the matrix.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":115
 *         self._this.invert()
 * 
 *     def scale(self, double sx, double sy):             # <<<<<<<<<<<<<<
 *         """scale(x, y)
 *         Apply a scaling to the transform
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_15scale(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_14scale[] = "scale(x, y)\n        Apply a scaling to the transform\n\n        :param x: The scale factor in X\n        :param y: The scale factor in Y\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_15scale(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_sx;
  double __pyx_v_sy;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scale (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sx,&__pyx_n_s_sy,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sx)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scale", 1, 2, 2, 1); __PYX_ERR(8, 115, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scale") < 0)) __PYX_ERR(8, 115, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_sx = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_sx == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 115, __pyx_L3_error)
    __pyx_v_sy = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_sy == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 115, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scale", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(8, 115, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Transform.scale", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_14scale(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), __pyx_v_sx, __pyx_v_sy);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_14scale(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_sx, double __pyx_v_sy) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scale", 0);

  /* "celiagg/transform.pxi":122
 *         :param y: The scale factor in Y
 *         """
 *         self._this.premultiply(_transform.trans_affine_scaling(sx, sy))             # <<<<<<<<<<<<<<
 * 
 *     def rotate(self, double angle):
 */
  (void)(__pyx_v_self->_this->premultiply(agg::trans_affine_scaling(__pyx_v_sx, __pyx_v_sy)));

  /* "celiagg/transform.pxi":115
 *         self._this.invert()
 * 
 *     def scale(self, double sx, double sy):             # <<<<<<<<<<<<<<
 *         """scale(x, y)
 *         Apply a scaling to the transform
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":124
 *         self._this.premultiply(_transform.trans_affine_scaling(sx, sy))
 * 
 *     def rotate(self, double angle):             # <<<<<<<<<<<<<<
 *         """rotate(angle)
 *         Apply a rotation to the transform.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_17rotate(PyObject *__pyx_v_self, PyObject *__pyx_arg_angle); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_16rotate[] = "rotate(angle)\n        Apply a rotation to the transform.\n\n        :param angle: The desired rotation in radians.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_17rotate(PyObject *__pyx_v_self, PyObject *__pyx_arg_angle) {
  double __pyx_v_angle;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rotate (wrapper)", 0);
  assert(__pyx_arg_angle); {
    __pyx_v_angle = __pyx_PyFloat_AsDouble(__pyx_arg_angle); if (unlikely((__pyx_v_angle == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 124, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Transform.rotate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_16rotate(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), ((double)__pyx_v_angle));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_16rotate(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_angle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rotate", 0);

  /* "celiagg/transform.pxi":130
 *         :param angle: The desired rotation in radians.
 *         """
 *         self._this.premultiply(_transform.trans_affine_rotation(angle))             # <<<<<<<<<<<<<<
 * 
 *     def skew(self, double sx, double sy):
 */
  (void)(__pyx_v_self->_this->premultiply(agg::trans_affine_rotation(__pyx_v_angle)));

  /* "celiagg/transform.pxi":124
 *         self._this.premultiply(_transform.trans_affine_scaling(sx, sy))
 * 
 *     def rotate(self, double angle):             # <<<<<<<<<<<<<<
 *         """rotate(angle)
 *         Apply a rotation to the transform.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":132
 *         self._this.premultiply(_transform.trans_affine_rotation(angle))
 * 
 *     def skew(self, double sx, double sy):             # <<<<<<<<<<<<<<
 *         """skew(x, y)
 *         Apply a skew to the transform.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_19skew(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_18skew[] = "skew(x, y)\n        Apply a skew to the transform.\n\n        :param x: The skew factor in X\n        :param y: The skew factor in Y\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_19skew(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_sx;
  double __pyx_v_sy;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("skew (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sx,&__pyx_n_s_sy,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sx)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("skew", 1, 2, 2, 1); __PYX_ERR(8, 132, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "skew") < 0)) __PYX_ERR(8, 132, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_sx = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_sx == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 132, __pyx_L3_error)
    __pyx_v_sy = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_sy == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 132, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("skew", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(8, 132, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Transform.skew", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_18skew(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), __pyx_v_sx, __pyx_v_sy);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_18skew(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_sx, double __pyx_v_sy) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("skew", 0);

  /* "celiagg/transform.pxi":139
 *         :param y: The skew factor in Y
 *         """
 *         self._this.premultiply(_transform.trans_affine_skewing(sx, sy))             # <<<<<<<<<<<<<<
 * 
 *     def translate(self, double x, double y):
 */
  (void)(__pyx_v_self->_this->premultiply(agg::trans_affine_skewing(__pyx_v_sx, __pyx_v_sy)));

  /* "celiagg/transform.pxi":132
 *         self._this.premultiply(_transform.trans_affine_rotation(angle))
 * 
 *     def skew(self, double sx, double sy):             # <<<<<<<<<<<<<<
 *         """skew(x, y)
 *         Apply a skew to the transform.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":141
 *         self._this.premultiply(_transform.trans_affine_skewing(sx, sy))
 * 
 *     def translate(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """translate(x, y)
 *         Apply a translation to the transform.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_21translate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_20translate[] = "translate(x, y)\n        Apply a translation to the transform.\n\n        :param x: The translation in X\n        :param y: The translation in Y\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_21translate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x;
  double __pyx_v_y;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("translate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate", 1, 2, 2, 1); __PYX_ERR(8, 141, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "translate") < 0)) __PYX_ERR(8, 141, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 141, __pyx_L3_error)
    __pyx_v_y = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 141, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("translate", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(8, 141, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Transform.translate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_20translate(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), __pyx_v_x, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_20translate(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_x, double __pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("translate", 0);

  /* "celiagg/transform.pxi":148
 *         :param y: The translation in Y
 *         """
 *         self._this.premultiply(_transform.trans_affine_translation(x, y))             # <<<<<<<<<<<<<<
 * 
 *     def premultiply(self, Transform other):
 */
  (void)(__pyx_v_self->_this->premultiply(agg::trans_affine_translation(__pyx_v_x, __pyx_v_y)));

  /* "celiagg/transform.pxi":141
 *         self._this.premultiply(_transform.trans_affine_skewing(sx, sy))
 * 
 *     def translate(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """translate(x, y)
 *         Apply a translation to the transform.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":150
 *         self._this.premultiply(_transform.trans_affine_translation(x, y))
 * 
 *     def premultiply(self, Transform other):             # <<<<<<<<<<<<<<
 *         """premultiply(other)
 *         Pre-multiply this transform with another transform.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_23premultiply(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_22premultiply[] = "premultiply(other)\n        Pre-multiply this transform with another transform.\n\n        :param other: A ``Transform`` instance to premultiply\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_23premultiply(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("premultiply (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7celiagg_8_celiagg_Transform, 1, "other", 0))) __PYX_ERR(8, 150, __pyx_L1_error)
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_22premultiply(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), ((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_22premultiply(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("premultiply", 0);

  /* "celiagg/transform.pxi":156
 *         :param other: A ``Transform`` instance to premultiply
 *         """
 *         self._this.premultiply(dereference(other._this))             # <<<<<<<<<<<<<<
 * 
 *     def multiply(self, Transform other):
 */
  (void)(__pyx_v_self->_this->premultiply((*__pyx_v_other->_this)));

  /* "celiagg/transform.pxi":150
 *         self._this.premultiply(_transform.trans_affine_translation(x, y))
 * 
 *     def premultiply(self, Transform other):             # <<<<<<<<<<<<<<
 *         """premultiply(other)
 *         Pre-multiply this transform with another transform.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":158
 *         self._this.premultiply(dereference(other._this))
 * 
 *     def multiply(self, Transform other):             # <<<<<<<<<<<<<<
 *         """multiply(other)
 *         Post-multiply this transform with another transform.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_25multiply(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_24multiply[] = "multiply(other)\n        Post-multiply this transform with another transform.\n\n        :param other: A ``Transform`` instance to multiply\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_25multiply(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("multiply (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7celiagg_8_celiagg_Transform, 1, "other", 0))) __PYX_ERR(8, 158, __pyx_L1_error)
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_24multiply(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), ((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_24multiply(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("multiply", 0);

  /* "celiagg/transform.pxi":164
 *         :param other: A ``Transform`` instance to multiply
 *         """
 *         self._this.multiply(dereference(other._this))             # <<<<<<<<<<<<<<
 * 
 *     def worldToScreen(self, double x, double y):
 */
  (void)(__pyx_v_self->_this->multiply((*__pyx_v_other->_this)));

  /* "celiagg/transform.pxi":158
 *         self._this.premultiply(dereference(other._this))
 * 
 *     def multiply(self, Transform other):             # <<<<<<<<<<<<<<
 *         """multiply(other)
 *         Post-multiply this transform with another transform.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":166
 *         self._this.multiply(dereference(other._this))
 * 
 *     def worldToScreen(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """worldToScreen(x, y)
 *         Transforms a point from world space to screen space.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_27worldToScreen(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_26worldToScreen[] = "worldToScreen(x, y)\n        Transforms a point from world space to screen space.\n\n        :param x: The X coordinate of the point to transform.\n        :param y: The Y coordinate of the point to transform.\n        :return: An (x, y) tuple containing the transformed point.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_27worldToScreen(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x;
  double __pyx_v_y;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("worldToScreen (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("worldToScreen", 1, 2, 2, 1); __PYX_ERR(8, 166, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "worldToScreen") < 0)) __PYX_ERR(8, 166, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 166, __pyx_L3_error)
    __pyx_v_y = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 166, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("worldToScreen", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(8, 166, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Transform.worldToScreen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_26worldToScreen(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), __pyx_v_x, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_26worldToScreen(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_x, double __pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("worldToScreen", 0);

  /* "celiagg/transform.pxi":174
 *         :return: An (x, y) tuple containing the transformed point.
 *         """
 *         self._this.transform(&x, &y)             # <<<<<<<<<<<<<<
 *         return (x, y)
 * 
 */
  __pyx_v_self->_this->transform((&__pyx_v_x), (&__pyx_v_y));

  /* "celiagg/transform.pxi":175
 *         """
 *         self._this.transform(&x, &y)
 *         return (x, y)             # <<<<<<<<<<<<<<
 * 
 *     def screenToWorld(self, double x, double y):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":166
 *         self._this.multiply(dereference(other._this))
 * 
 *     def worldToScreen(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """worldToScreen(x, y)
 *         Transforms a point from world space to screen space.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.worldToScreen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/transform.pxi":177
 *         return (x, y)
 * 
 *     def screenToWorld(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """screenToWorld(x, y)
 *         Transforms a point from screen space to world space.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_29screenToWorld(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_9Transform_28screenToWorld[] = "screenToWorld(x, y)\n        Transforms a point from screen space to world space.\n\n        :param x: The X coordinate of the point to transform.\n        :param y: The Y coordinate of the point to transform.\n        :return: An (x, y) tuple containing the transformed point.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_29screenToWorld(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x;
  double __pyx_v_y;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("screenToWorld (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("screenToWorld", 1, 2, 2, 1); __PYX_ERR(8, 177, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "screenToWorld") < 0)) __PYX_ERR(8, 177, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 177, __pyx_L3_error)
    __pyx_v_y = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y == (double)-1) && PyErr_Occurred())) __PYX_ERR(8, 177, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("screenToWorld", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(8, 177, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Transform.screenToWorld", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_28screenToWorld(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), __pyx_v_x, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_28screenToWorld(struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, double __pyx_v_x, double __pyx_v_y) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("screenToWorld", 0);

  /* "celiagg/transform.pxi":185
 *         :return: An (x, y) tuple containing the transformed point.
 *         """
 *         self._this.inverse_transform(&x, &y)             # <<<<<<<<<<<<<<
 *         return (x, y)
 */
  __pyx_v_self->_this->inverse_transform((&__pyx_v_x), (&__pyx_v_y));

  /* "celiagg/transform.pxi":186
 *         """
 *         self._this.inverse_transform(&x, &y)
 *         return (x, y)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(8, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(8, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/transform.pxi":177
 *         return (x, y)
 * 
 *     def screenToWorld(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """screenToWorld(x, y)
 *         Transforms a point from screen space to world space.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.screenToWorld", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_31__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_31__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_30__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_30__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__53, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_33__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_9Transform_33__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_9Transform_32__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Transform *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_9Transform_32__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Transform *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Transform.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":33
 *     cdef _vertex_source.VertexSource* _this
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

/* Python wrapper */
static void __pyx_pw_7celiagg_8_celiagg_12VertexSource_1__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7celiagg_8_celiagg_12VertexSource_1__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_7celiagg_8_celiagg_12VertexSource___dealloc__(((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7celiagg_8_celiagg_12VertexSource___dealloc__(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "celiagg/vertex_source.pxi":34
 * 
 *     def __dealloc__(self):
 *         del self._this             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  delete __pyx_v_self->_this;

  /* "celiagg/vertex_source.pxi":33
 *     cdef _vertex_source.VertexSource* _this
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._this
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}
static PyObject *__pyx_gb_7celiagg_8_celiagg_12VertexSource_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "celiagg/vertex_source.pxi":36
 *         del self._this
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         for pt in self.vertices():
 *             yield pt
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_3__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12VertexSource_2__iter__(((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_2__iter__(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self) {
  struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *)__pyx_tp_new_7celiagg_8_celiagg___pyx_scope_struct____iter__(__pyx_ptype_7celiagg_8_celiagg___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(2, 36, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_7celiagg_8_celiagg_12VertexSource_4generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_VertexSource___iter, __pyx_n_s_celiagg__celiagg); if (unlikely(!gen)) __PYX_ERR(2, 36, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("celiagg._celiagg.VertexSource.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_7celiagg_8_celiagg_12VertexSource_4generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(2, 36, __pyx_L1_error)

  /* "celiagg/vertex_source.pxi":37
 * 
 *     def __iter__(self):
 *         for pt in self.vertices():             # <<<<<<<<<<<<<<
 *             yield pt
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_vertices); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 37, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(2, 37, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 37, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(2, 37, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 37, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(2, 37, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_pt);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_pt, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "celiagg/vertex_source.pxi":38
 *     def __iter__(self):
 *         for pt in self.vertices():
 *             yield pt             # <<<<<<<<<<<<<<
 * 
 *     def bounding_rect(self):
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_pt);
    __pyx_r = __pyx_cur_scope->__pyx_v_pt;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(2, 38, __pyx_L1_error)

    /* "celiagg/vertex_source.pxi":37
 * 
 *     def __iter__(self):
 *         for pt in self.vertices():             # <<<<<<<<<<<<<<
 *             yield pt
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "celiagg/vertex_source.pxi":36
 *         del self._this
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         for pt in self.vertices():
 *             yield pt
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":40
 *             yield pt
 * 
 *     def bounding_rect(self):             # <<<<<<<<<<<<<<
 *         """Returns a bounding rectangle for the vertices of the source in the
 *         format: (x, y, w, h)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_6bounding_rect(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_12VertexSource_5bounding_rect[] = "Returns a bounding rectangle for the vertices of the source in the\n        format: (x, y, w, h)\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_6bounding_rect(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bounding_rect (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12VertexSource_5bounding_rect(((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_5bounding_rect(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self) {
  PyObject *__pyx_v_verts = NULL;
  PyObject *__pyx_v_xs = NULL;
  PyObject *__pyx_v_ys = NULL;
  PyObject *__pyx_v_xmin = NULL;
  PyObject *__pyx_v_xmax = NULL;
  PyObject *__pyx_v_ymin = NULL;
  PyObject *__pyx_v_ymax = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("bounding_rect", 0);

  /* "celiagg/vertex_source.pxi":44
 *         format: (x, y, w, h)
 *         """
 *         if self.length() == 0:             # <<<<<<<<<<<<<<
 *             return (0, 0, 0, 0)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_length); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(2, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "celiagg/vertex_source.pxi":45
 *         """
 *         if self.length() == 0:
 *             return (0, 0, 0, 0)             # <<<<<<<<<<<<<<
 * 
 *         verts = self.vertices()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_tuple__55);
    __pyx_r = __pyx_tuple__55;
    goto __pyx_L0;

    /* "celiagg/vertex_source.pxi":44
 *         format: (x, y, w, h)
 *         """
 *         if self.length() == 0:             # <<<<<<<<<<<<<<
 *             return (0, 0, 0, 0)
 * 
 */
  }

  /* "celiagg/vertex_source.pxi":47
 *             return (0, 0, 0, 0)
 * 
 *         verts = self.vertices()             # <<<<<<<<<<<<<<
 *         xs = verts[:, 0]
 *         ys = verts[:, 1]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_vertices); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_verts = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "celiagg/vertex_source.pxi":48
 * 
 *         verts = self.vertices()
 *         xs = verts[:, 0]             # <<<<<<<<<<<<<<
 *         ys = verts[:, 1]
 *         xmin, xmax = xs.min(), xs.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_verts, __pyx_tuple__57); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_xs = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "celiagg/vertex_source.pxi":49
 *         verts = self.vertices()
 *         xs = verts[:, 0]
 *         ys = verts[:, 1]             # <<<<<<<<<<<<<<
 *         xmin, xmax = xs.min(), xs.max()
 *         ymin, ymax = ys.min(), ys.max()
 */
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_verts, __pyx_tuple__58); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_ys = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "celiagg/vertex_source.pxi":50
 *         xs = verts[:, 0]
 *         ys = verts[:, 1]
 *         xmin, xmax = xs.min(), xs.max()             # <<<<<<<<<<<<<<
 *         ymin, ymax = ys.min(), ys.max()
 *         return (xmin, ymin, xmax - xmin, ymax - ymin)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_xs, __pyx_n_s_min); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_xs, __pyx_n_s_max); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_xmin = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_xmax = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/vertex_source.pxi":51
 *         ys = verts[:, 1]
 *         xmin, xmax = xs.min(), xs.max()
 *         ymin, ymax = ys.min(), ys.max()             # <<<<<<<<<<<<<<
 *         return (xmin, ymin, xmax - xmin, ymax - ymin)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_ys, __pyx_n_s_min); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_ys, __pyx_n_s_max); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ymin = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_ymax = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "celiagg/vertex_source.pxi":52
 *         xmin, xmax = xs.min(), xs.max()
 *         ymin, ymax = ys.min(), ys.max()
 *         return (xmin, ymin, xmax - xmin, ymax - ymin)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyNumber_Subtract(__pyx_v_xmax, __pyx_v_xmin); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_ymax, __pyx_v_ymin); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_xmin);
  __Pyx_GIVEREF(__pyx_v_xmin);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_xmin);
  __Pyx_INCREF(__pyx_v_ymin);
  __Pyx_GIVEREF(__pyx_v_ymin);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_ymin);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/vertex_source.pxi":40
 *             yield pt
 * 
 *     def bounding_rect(self):             # <<<<<<<<<<<<<<
 *         """Returns a bounding rectangle for the vertices of the source in the
 *         format: (x, y, w, h)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg.VertexSource.bounding_rect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_verts);
  __Pyx_XDECREF(__pyx_v_xs);
  __Pyx_XDECREF(__pyx_v_ys);
  __Pyx_XDECREF(__pyx_v_xmin);
  __Pyx_XDECREF(__pyx_v_xmax);
  __Pyx_XDECREF(__pyx_v_ymin);
  __Pyx_XDECREF(__pyx_v_ymax);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":54
 *         return (xmin, ymin, xmax - xmin, ymax - ymin)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the object.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_8copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_12VertexSource_7copy[] = "Returns a deep copy of the object.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_8copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12VertexSource_7copy(((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_7copy(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/vertex_source.pxi":57
 *         """Returns a deep copy of the object.
 *         """
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def final_point(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(2, 57, __pyx_L1_error)

  /* "celiagg/vertex_source.pxi":54
 *         return (xmin, ymin, xmax - xmin, ymax - ymin)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the object.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("celiagg._celiagg.VertexSource.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":59
 *         raise NotImplementedError
 * 
 *     def final_point(self):             # <<<<<<<<<<<<<<
 *         """Returns the last vertex that will be returned by the source.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_10final_point(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_12VertexSource_9final_point[] = "Returns the last vertex that will be returned by the source.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_10final_point(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("final_point (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12VertexSource_9final_point(((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_9final_point(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("final_point", 0);

  /* "celiagg/vertex_source.pxi":62
 *         """Returns the last vertex that will be returned by the source.
 *         """
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def length(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(2, 62, __pyx_L1_error)

  /* "celiagg/vertex_source.pxi":59
 *         raise NotImplementedError
 * 
 *     def final_point(self):             # <<<<<<<<<<<<<<
 *         """Returns the last vertex that will be returned by the source.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("celiagg._celiagg.VertexSource.final_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":64
 *         raise NotImplementedError
 * 
 *     def length(self):             # <<<<<<<<<<<<<<
 *         """Returns the number of vertices returned by ``vertices()``
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_12length(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_12VertexSource_11length[] = "Returns the number of vertices returned by ``vertices()``\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_12length(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("length (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12VertexSource_11length(((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_11length(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("length", 0);

  /* "celiagg/vertex_source.pxi":67
 *         """Returns the number of vertices returned by ``vertices()``
 *         """
 *         return self._this.total_vertices()             # <<<<<<<<<<<<<<
 * 
 *     def vertices(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_self->_this->total_vertices()); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/vertex_source.pxi":64
 *         raise NotImplementedError
 * 
 *     def length(self):             # <<<<<<<<<<<<<<
 *         """Returns the number of vertices returned by ``vertices()``
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.VertexSource.length", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":69
 *         return self._this.total_vertices()
 * 
 *     def vertices(self):             # <<<<<<<<<<<<<<
 *         """Returns all the vertices in the source as a numpy array.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_14vertices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_12VertexSource_13vertices[] = "Returns all the vertices in the source as a numpy array.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_14vertices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("vertices (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12VertexSource_13vertices(((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_13vertices(struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self) {
  unsigned int __pyx_v_count;
  __Pyx_memviewslice __pyx_v_points = { 0, 0, { 0 }, { 0 }, { 0 } };
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  unsigned int __pyx_t_7;
  unsigned int __pyx_t_8;
  long __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  __Pyx_RefNannySetupContext("vertices", 0);

  /* "celiagg/vertex_source.pxi":72
 *         """Returns all the vertices in the source as a numpy array.
 *         """
 *         cdef unsigned count = self._this.total_vertices()             # <<<<<<<<<<<<<<
 *         cdef double[:,::1] points = numpy.empty((count, 2), dtype=numpy.float64,
 *                                                 order='c')
 */
  __pyx_v_count = __pyx_v_self->_this->total_vertices();

  /* "celiagg/vertex_source.pxi":73
 *         """
 *         cdef unsigned count = self._this.total_vertices()
 *         cdef double[:,::1] points = numpy.empty((count, 2), dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                                 order='c')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_2);
  __Pyx_GIVEREF(__pyx_int_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_2);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(2, 73, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(2, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_points = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "celiagg/vertex_source.pxi":76
 *                                                 order='c')
 * 
 *         self._this.rewind(0)             # <<<<<<<<<<<<<<
 *         for i in range(0, count):
 *             self._this.vertex(&points[i][0], &points[i][1])
 */
  __pyx_v_self->_this->rewind(0);

  /* "celiagg/vertex_source.pxi":77
 * 
 *         self._this.rewind(0)
 *         for i in range(0, count):             # <<<<<<<<<<<<<<
 *             self._this.vertex(&points[i][0], &points[i][1])
 * 
 */
  __pyx_t_7 = __pyx_v_count;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "celiagg/vertex_source.pxi":78
 *         self._this.rewind(0)
 *         for i in range(0, count):
 *             self._this.vertex(&points[i][0], &points[i][1])             # <<<<<<<<<<<<<<
 * 
 *         return points.base
 */
    __pyx_t_10 = __pyx_v_i;
    __pyx_t_11 = 0;
    __pyx_t_12 = -1;
    if (__pyx_t_10 < 0) {
      __pyx_t_10 += __pyx_v_points.shape[0];
      if (unlikely(__pyx_t_10 < 0)) __pyx_t_12 = 0;
    } else if (unlikely(__pyx_t_10 >= __pyx_v_points.shape[0])) __pyx_t_12 = 0;
    if (__pyx_t_11 < 0) {
      __pyx_t_11 += __pyx_v_points.shape[1];
      if (unlikely(__pyx_t_11 < 0)) __pyx_t_12 = 1;
    } else if (unlikely(__pyx_t_11 >= __pyx_v_points.shape[1])) __pyx_t_12 = 1;
    if (unlikely(__pyx_t_12 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_12);
      __PYX_ERR(2, 78, __pyx_L1_error)
    }
    __pyx_t_13 = __pyx_v_i;
    __pyx_t_14 = 1;
    __pyx_t_12 = -1;
    if (__pyx_t_13 < 0) {
      __pyx_t_13 += __pyx_v_points.shape[0];
      if (unlikely(__pyx_t_13 < 0)) __pyx_t_12 = 0;
    } else if (unlikely(__pyx_t_13 >= __pyx_v_points.shape[0])) __pyx_t_12 = 0;
    if (__pyx_t_14 < 0) {
      __pyx_t_14 += __pyx_v_points.shape[1];
      if (unlikely(__pyx_t_14 < 0)) __pyx_t_12 = 1;
    } else if (unlikely(__pyx_t_14 >= __pyx_v_points.shape[1])) __pyx_t_12 = 1;
    if (unlikely(__pyx_t_12 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_12);
      __PYX_ERR(2, 78, __pyx_L1_error)
    }
    (void)(__pyx_v_self->_this->vertex((&(*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v_points.data + __pyx_t_10 * __pyx_v_points.strides[0]) )) + __pyx_t_11)) )))), (&(*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v_points.data + __pyx_t_13 * __pyx_v_points.strides[0]) )) + __pyx_t_14)) ))))));
  }

  /* "celiagg/vertex_source.pxi":80
 *             self._this.vertex(&points[i][0], &points[i][1])
 * 
 *         return points.base             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_points, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_base); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/vertex_source.pxi":69
 *         return self._this.total_vertices()
 * 
 *     def vertices(self):             # <<<<<<<<<<<<<<
 *         """Returns all the vertices in the source as a numpy array.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("celiagg._celiagg.VertexSource.vertices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_points, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_16__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_16__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12VertexSource_15__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_15__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.VertexSource.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_18__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_12VertexSource_18__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_12VertexSource_17__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_12VertexSource_17__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.VertexSource.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":91
 *     cdef object _points
 * 
 *     def __cinit__(self, points):             # <<<<<<<<<<<<<<
 *         cdef double[:,::1] _points = numpy.asarray(points, dtype=numpy.float64,
 *                                                    order='c')
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_7BSpline_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_7BSpline_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_points = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_points,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_points)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 91, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_points = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 91, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.BSpline.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_7BSpline___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_BSpline *)__pyx_v_self), __pyx_v_points);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_7BSpline___cinit__(struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self, PyObject *__pyx_v_points) {
  __Pyx_memviewslice __pyx_v__points = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_msg = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/vertex_source.pxi":92
 * 
 *     def __cinit__(self, points):
 *         cdef double[:,::1] _points = numpy.asarray(points, dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                                    order='c')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_points);
  __Pyx_GIVEREF(__pyx_v_points);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_points);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(2, 92, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(2, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v__points = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "celiagg/vertex_source.pxi":95
 *                                                    order='c')
 * 
 *         if _points.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = 'Points argument must be an iterable of (x, y) pairs.'
 *             raise ValueError(msg)
 */
  __pyx_t_7 = (((__pyx_v__points.shape[1]) != 2) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "celiagg/vertex_source.pxi":96
 * 
 *         if _points.shape[1] != 2:
 *             msg = 'Points argument must be an iterable of (x, y) pairs.'             # <<<<<<<<<<<<<<
 *             raise ValueError(msg)
 * 
 */
    __Pyx_INCREF(__pyx_kp_s_Points_argument_must_be_an_itera);
    __pyx_v_msg = __pyx_kp_s_Points_argument_must_be_an_itera;

    /* "celiagg/vertex_source.pxi":97
 *         if _points.shape[1] != 2:
 *             msg = 'Points argument must be an iterable of (x, y) pairs.'
 *             raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.BsplineSource(
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(2, 97, __pyx_L1_error)

    /* "celiagg/vertex_source.pxi":95
 *                                                    order='c')
 * 
 *         if _points.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = 'Points argument must be an iterable of (x, y) pairs.'
 *             raise ValueError(msg)
 */
  }

  /* "celiagg/vertex_source.pxi":100
 * 
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.BsplineSource(
 *             &_points[0][0], _points.shape[0]             # <<<<<<<<<<<<<<
 *         )
 *         # hold a reference to the backing ndarray
 */
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = -1;
  if (__pyx_t_8 < 0) {
    __pyx_t_8 += __pyx_v__points.shape[0];
    if (unlikely(__pyx_t_8 < 0)) __pyx_t_10 = 0;
  } else if (unlikely(__pyx_t_8 >= __pyx_v__points.shape[0])) __pyx_t_10 = 0;
  if (__pyx_t_9 < 0) {
    __pyx_t_9 += __pyx_v__points.shape[1];
    if (unlikely(__pyx_t_9 < 0)) __pyx_t_10 = 1;
  } else if (unlikely(__pyx_t_9 >= __pyx_v__points.shape[1])) __pyx_t_10 = 1;
  if (unlikely(__pyx_t_10 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_10);
    __PYX_ERR(2, 100, __pyx_L1_error)
  }

  /* "celiagg/vertex_source.pxi":99
 *             raise ValueError(msg)
 * 
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.BsplineSource(             # <<<<<<<<<<<<<<
 *             &_points[0][0], _points.shape[0]
 *         )
 */
  __pyx_v_self->__pyx_base._this = ((VertexSource *)new BsplineSource((&(*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__points.data + __pyx_t_8 * __pyx_v__points.strides[0]) )) + __pyx_t_9)) )))), (__pyx_v__points.shape[0])));

  /* "celiagg/vertex_source.pxi":103
 *         )
 *         # hold a reference to the backing ndarray
 *         self._points = _points             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v__points, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->_points);
  __Pyx_DECREF(__pyx_v_self->_points);
  __pyx_v_self->_points = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "celiagg/vertex_source.pxi":91
 *     cdef object _points
 * 
 *     def __cinit__(self, points):             # <<<<<<<<<<<<<<
 *         cdef double[:,::1] _points = numpy.asarray(points, dtype=numpy.float64,
 *                                                    order='c')
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("celiagg._celiagg.BSpline.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v__points, 1);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":105
 *         self._points = _points
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the object.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_7BSpline_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_7BSpline_2copy[] = "Returns a deep copy of the object.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_7BSpline_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_7BSpline_2copy(((struct __pyx_obj_7celiagg_8_celiagg_BSpline *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_7BSpline_2copy(struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self) {
  PyObject *__pyx_v_points = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/vertex_source.pxi":108
 *         """Returns a deep copy of the object.
 *         """
 *         points = self._points.copy()             # <<<<<<<<<<<<<<
 *         return BSpline(points)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_points, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_points = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/vertex_source.pxi":109
 *         """
 *         points = self._points.copy()
 *         return BSpline(points)             # <<<<<<<<<<<<<<
 * 
 *     def final_point(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_BSpline), __pyx_v_points); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/vertex_source.pxi":105
 *         self._points = _points
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the object.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.BSpline.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_points);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":111
 *         return BSpline(points)
 * 
 *     def final_point(self):             # <<<<<<<<<<<<<<
 *         """Returns the last vertex that will be returned by the source.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_7BSpline_5final_point(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_7BSpline_4final_point[] = "Returns the last vertex that will be returned by the source.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_7BSpline_5final_point(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("final_point (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_7BSpline_4final_point(((struct __pyx_obj_7celiagg_8_celiagg_BSpline *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_7BSpline_4final_point(struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("final_point", 0);

  /* "celiagg/vertex_source.pxi":114
 *         """Returns the last vertex that will be returned by the source.
 *         """
 *         return self._points[-1]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->_points, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/vertex_source.pxi":111
 *         return BSpline(points)
 * 
 *     def final_point(self):             # <<<<<<<<<<<<<<
 *         """Returns the last vertex that will be returned by the source.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.BSpline.final_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_7BSpline_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_7BSpline_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_7BSpline_6__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_BSpline *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_7BSpline_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__61, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.BSpline.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_7BSpline_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_7BSpline_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_7BSpline_8__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_BSpline *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_7BSpline_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_BSpline *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__62, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.BSpline.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":121
 *     """
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.PathSource()
 * 
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_4Path_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_4Path_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_4Path___cinit__(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/vertex_source.pxi":122
 * 
 *     def __cinit__(self):
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.PathSource()             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __pyx_v_self->__pyx_base._this = ((VertexSource *)new PathSource());

  /* "celiagg/vertex_source.pxi":121
 *     """
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.PathSource()
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":124
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.PathSource()
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the object.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_2copy[] = "Returns a deep copy of the object.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_2copy(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_2copy(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self) {
  struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_cpy = 0;
  PathSource *__pyx_v_ths;
  PathSource *__pyx_v_other;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/vertex_source.pxi":128
 *         """
 *         cdef:
 *             Path cpy = Path()             # <<<<<<<<<<<<<<
 *             _vertex_source.PathSource* ths = <_vertex_source.PathSource*>self._this
 *             _vertex_source.PathSource* other = <_vertex_source.PathSource*>cpy._this
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_Path)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cpy = ((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/vertex_source.pxi":129
 *         cdef:
 *             Path cpy = Path()
 *             _vertex_source.PathSource* ths = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *             _vertex_source.PathSource* other = <_vertex_source.PathSource*>cpy._this
 * 
 */
  __pyx_v_ths = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":130
 *             Path cpy = Path()
 *             _vertex_source.PathSource* ths = <_vertex_source.PathSource*>self._this
 *             _vertex_source.PathSource* other = <_vertex_source.PathSource*>cpy._this             # <<<<<<<<<<<<<<
 * 
 *         other.concat_path[_vertex_source.PathSource](dereference(ths), 0)
 */
  __pyx_v_other = ((PathSource *)__pyx_v_cpy->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":132
 *             _vertex_source.PathSource* other = <_vertex_source.PathSource*>cpy._this
 * 
 *         other.concat_path[_vertex_source.PathSource](dereference(ths), 0)             # <<<<<<<<<<<<<<
 *         return cpy
 * 
 */
  __pyx_v_other->concat_path<PathSource>((*__pyx_v_ths), 0);

  /* "celiagg/vertex_source.pxi":133
 * 
 *         other.concat_path[_vertex_source.PathSource](dereference(ths), 0)
 *         return cpy             # <<<<<<<<<<<<<<
 * 
 *     def final_point(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cpy));
  __pyx_r = ((PyObject *)__pyx_v_cpy);
  goto __pyx_L0;

  /* "celiagg/vertex_source.pxi":124
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.PathSource()
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the object.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Path.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cpy);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":135
 *         return cpy
 * 
 *     def final_point(self):             # <<<<<<<<<<<<<<
 *         """Returns the last vertex that will be returned by the source.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_5final_point(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_4final_point[] = "Returns the last vertex that will be returned by the source.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_5final_point(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("final_point (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_4final_point(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_4final_point(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self) {
  PathSource *__pyx_v_pth;
  double __pyx_v_x;
  double __pyx_v_y;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("final_point", 0);

  /* "celiagg/vertex_source.pxi":139
 *         """
 *         cdef:
 *             _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *             double x, y
 *         pth.last_vertex(&x, &y)
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":141
 *             _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *             double x, y
 *         pth.last_vertex(&x, &y)             # <<<<<<<<<<<<<<
 *         return (x, y)
 * 
 */
  (void)(__pyx_v_pth->last_vertex((&__pyx_v_x), (&__pyx_v_y)));

  /* "celiagg/vertex_source.pxi":142
 *             double x, y
 *         pth.last_vertex(&x, &y)
 *         return (x, y)             # <<<<<<<<<<<<<<
 * 
 *     def begin(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/vertex_source.pxi":135
 *         return cpy
 * 
 *     def final_point(self):             # <<<<<<<<<<<<<<
 *         """Returns the last vertex that will be returned by the source.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.Path.final_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":144
 *         return (x, y)
 * 
 *     def begin(self):             # <<<<<<<<<<<<<<
 *         """Begins a new subpath.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_7begin(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_6begin[] = "Begins a new subpath.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_7begin(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("begin (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_6begin(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_6begin(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("begin", 0);

  /* "celiagg/vertex_source.pxi":147
 *         """Begins a new subpath.
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.begin()
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":148
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.begin()             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __pyx_v_pth->begin();

  /* "celiagg/vertex_source.pxi":144
 *         return (x, y)
 * 
 *     def begin(self):             # <<<<<<<<<<<<<<
 *         """Begins a new subpath.
 *         """
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":150
 *         pth.begin()
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Closes the current subpath.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_9close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_8close[] = "Closes the current subpath.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_9close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_8close(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_8close(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close", 0);

  /* "celiagg/vertex_source.pxi":153
 *         """Closes the current subpath.
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.close()
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":154
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.close()             # <<<<<<<<<<<<<<
 * 
 *     def reset(self):
 */
  __pyx_v_pth->close();

  /* "celiagg/vertex_source.pxi":150
 *         pth.begin()
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Closes the current subpath.
 *         """
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":156
 *         pth.close()
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         """Clears the path.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_11reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_10reset[] = "Clears the path.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_11reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_10reset(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_10reset(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset", 0);

  /* "celiagg/vertex_source.pxi":159
 *         """Clears the path.
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.reset()
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":160
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.reset()             # <<<<<<<<<<<<<<
 * 
 *     def add_path(self, Path other):
 */
  __pyx_v_pth->reset();

  /* "celiagg/vertex_source.pxi":156
 *         pth.close()
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         """Clears the path.
 *         """
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":162
 *         pth.reset()
 * 
 *     def add_path(self, Path other):             # <<<<<<<<<<<<<<
 *         """add_path(other)
 *         Add another path to this path.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_13add_path(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_12add_path[] = "add_path(other)\n        Add another path to this path.\n\n        :param other: A ``Path`` object which will be concatenated onto the end\n                      of this path.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_13add_path(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_path (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_7celiagg_8_celiagg_Path, 1, "other", 0))) __PYX_ERR(2, 162, __pyx_L1_error)
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_12add_path(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), ((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_12add_path(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_other) {
  PathSource *__pyx_v_pth;
  PathSource *__pyx_v_othr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_path", 0);

  /* "celiagg/vertex_source.pxi":169
 *                       of this path.
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         cdef _vertex_source.PathSource* othr = <_vertex_source.PathSource*>self._this
 *         pth.concat_path[_vertex_source.PathSource](dereference(othr), 0)
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":170
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         cdef _vertex_source.PathSource* othr = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.concat_path[_vertex_source.PathSource](dereference(othr), 0)
 * 
 */
  __pyx_v_othr = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":171
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         cdef _vertex_source.PathSource* othr = <_vertex_source.PathSource*>self._this
 *         pth.concat_path[_vertex_source.PathSource](dereference(othr), 0)             # <<<<<<<<<<<<<<
 * 
 *     def move_to(self, double x, double y):
 */
  __pyx_v_pth->concat_path<PathSource>((*__pyx_v_othr), 0);

  /* "celiagg/vertex_source.pxi":162
 *         pth.reset()
 * 
 *     def add_path(self, Path other):             # <<<<<<<<<<<<<<
 *         """add_path(other)
 *         Add another path to this path.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":173
 *         pth.concat_path[_vertex_source.PathSource](dereference(othr), 0)
 * 
 *     def move_to(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """move_to(x, y)
 *         Moves the current position of the path.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_15move_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_14move_to[] = "move_to(x, y)\n        Moves the current position of the path.\n\n        :param x: X coordinate of the new current position\n        :param y: Y coordinate of the new current position\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_15move_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x;
  double __pyx_v_y;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("move_to (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("move_to", 1, 2, 2, 1); __PYX_ERR(2, 173, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "move_to") < 0)) __PYX_ERR(2, 173, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 173, __pyx_L3_error)
    __pyx_v_y = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 173, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("move_to", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 173, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Path.move_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_14move_to(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), __pyx_v_x, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_14move_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x, double __pyx_v_y) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("move_to", 0);

  /* "celiagg/vertex_source.pxi":180
 *         :param y: Y coordinate of the new current position
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.move_to(x, y)
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":181
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.move_to(x, y)             # <<<<<<<<<<<<<<
 * 
 *     def line_to(self, double x, double y):
 */
  __pyx_v_pth->move_to(__pyx_v_x, __pyx_v_y);

  /* "celiagg/vertex_source.pxi":173
 *         pth.concat_path[_vertex_source.PathSource](dereference(othr), 0)
 * 
 *     def move_to(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """move_to(x, y)
 *         Moves the current position of the path.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":183
 *         pth.move_to(x, y)
 * 
 *     def line_to(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """line_to(x, y)
 *         Adds a line segment to the path from the current position to the given
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_17line_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_16line_to[] = "line_to(x, y)\n        Adds a line segment to the path from the current position to the given\n        position.\n\n        :param x: X coordinate of the line end point\n        :param y: Y coordinate of the line end point\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_17line_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x;
  double __pyx_v_y;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("line_to (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("line_to", 1, 2, 2, 1); __PYX_ERR(2, 183, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "line_to") < 0)) __PYX_ERR(2, 183, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 183, __pyx_L3_error)
    __pyx_v_y = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 183, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("line_to", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 183, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Path.line_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_16line_to(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), __pyx_v_x, __pyx_v_y);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_16line_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x, double __pyx_v_y) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("line_to", 0);

  /* "celiagg/vertex_source.pxi":191
 *         :param y: Y coordinate of the line end point
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.line_to(x, y)
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":192
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.line_to(x, y)             # <<<<<<<<<<<<<<
 * 
 *     def arc_to(self, double x1, double y1, double x2, double y2, double radius):
 */
  __pyx_v_pth->line_to(__pyx_v_x, __pyx_v_y);

  /* "celiagg/vertex_source.pxi":183
 *         pth.move_to(x, y)
 * 
 *     def line_to(self, double x, double y):             # <<<<<<<<<<<<<<
 *         """line_to(x, y)
 *         Adds a line segment to the path from the current position to the given
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":194
 *         pth.line_to(x, y)
 * 
 *     def arc_to(self, double x1, double y1, double x2, double y2, double radius):             # <<<<<<<<<<<<<<
 *         """arc_to(x1, y1, x2, y2, radius)
 *         Adds an arc to the path connecting two points
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_19arc_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_18arc_to[] = "arc_to(x1, y1, x2, y2, radius)\n        Adds an arc to the path connecting two points\n\n        :param x1: The starting X point of the arc\n        :param y1: The starting Y point of the arc\n        :param x2: The end X point of the arc\n        :param y2: The end Y point of the arc\n        :param radius: The radius of the arc\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_19arc_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x1;
  double __pyx_v_y1;
  double __pyx_v_x2;
  double __pyx_v_y2;
  double __pyx_v_radius;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("arc_to (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x1,&__pyx_n_s_y1,&__pyx_n_s_x2,&__pyx_n_s_y2,&__pyx_n_s_radius,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("arc_to", 1, 5, 5, 1); __PYX_ERR(2, 194, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("arc_to", 1, 5, 5, 2); __PYX_ERR(2, 194, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("arc_to", 1, 5, 5, 3); __PYX_ERR(2, 194, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("arc_to", 1, 5, 5, 4); __PYX_ERR(2, 194, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "arc_to") < 0)) __PYX_ERR(2, 194, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_x1 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 194, __pyx_L3_error)
    __pyx_v_y1 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 194, __pyx_L3_error)
    __pyx_v_x2 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_x2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 194, __pyx_L3_error)
    __pyx_v_y2 = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_y2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 194, __pyx_L3_error)
    __pyx_v_radius = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_radius == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 194, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("arc_to", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 194, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Path.arc_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_18arc_to(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), __pyx_v_x1, __pyx_v_y1, __pyx_v_x2, __pyx_v_y2, __pyx_v_radius);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_18arc_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x1, double __pyx_v_y1, double __pyx_v_x2, double __pyx_v_y2, double __pyx_v_radius) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("arc_to", 0);

  /* "celiagg/vertex_source.pxi":204
 *         :param radius: The radius of the arc
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.arc_to(x1, y1, x2, y2, radius)
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":205
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.arc_to(x1, y1, x2, y2, radius)             # <<<<<<<<<<<<<<
 * 
 *     def quadric_to(self, double x_ctrl, double y_ctrl, double x_to, double y_to):
 */
  __pyx_v_pth->arc_to(__pyx_v_x1, __pyx_v_y1, __pyx_v_x2, __pyx_v_y2, __pyx_v_radius);

  /* "celiagg/vertex_source.pxi":194
 *         pth.line_to(x, y)
 * 
 *     def arc_to(self, double x1, double y1, double x2, double y2, double radius):             # <<<<<<<<<<<<<<
 *         """arc_to(x1, y1, x2, y2, radius)
 *         Adds an arc to the path connecting two points
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":207
 *         pth.arc_to(x1, y1, x2, y2, radius)
 * 
 *     def quadric_to(self, double x_ctrl, double y_ctrl, double x_to, double y_to):             # <<<<<<<<<<<<<<
 *         """Adds a quadratic Bezier curve to the path, starting from the
 *         current position.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_21quadric_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_20quadric_to[] = "Adds a quadratic Bezier curve to the path, starting from the\n        current position.\n\n        :param x_ctrl: Control point X\n        :param y_ctrl: Control point Y\n        :param x_to: End point X\n        :param y_to: End point Y\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_21quadric_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x_ctrl;
  double __pyx_v_y_ctrl;
  double __pyx_v_x_to;
  double __pyx_v_y_to;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("quadric_to (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x_ctrl,&__pyx_n_s_y_ctrl,&__pyx_n_s_x_to,&__pyx_n_s_y_to,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x_ctrl)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_ctrl)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("quadric_to", 1, 4, 4, 1); __PYX_ERR(2, 207, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x_to)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("quadric_to", 1, 4, 4, 2); __PYX_ERR(2, 207, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_to)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("quadric_to", 1, 4, 4, 3); __PYX_ERR(2, 207, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "quadric_to") < 0)) __PYX_ERR(2, 207, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_x_ctrl = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x_ctrl == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 207, __pyx_L3_error)
    __pyx_v_y_ctrl = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y_ctrl == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 207, __pyx_L3_error)
    __pyx_v_x_to = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_x_to == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 207, __pyx_L3_error)
    __pyx_v_y_to = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_y_to == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 207, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("quadric_to", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 207, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Path.quadric_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_20quadric_to(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), __pyx_v_x_ctrl, __pyx_v_y_ctrl, __pyx_v_x_to, __pyx_v_y_to);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_20quadric_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x_ctrl, double __pyx_v_y_ctrl, double __pyx_v_x_to, double __pyx_v_y_to) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("quadric_to", 0);

  /* "celiagg/vertex_source.pxi":216
 *         :param y_to: End point Y
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.quadric_to(x_ctrl, y_ctrl, x_to, y_to)
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":217
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.quadric_to(x_ctrl, y_ctrl, x_to, y_to)             # <<<<<<<<<<<<<<
 * 
 *     def cubic_to(self, double x_ctrl1, double y_ctrl1,
 */
  __pyx_v_pth->quadric_to(__pyx_v_x_ctrl, __pyx_v_y_ctrl, __pyx_v_x_to, __pyx_v_y_to);

  /* "celiagg/vertex_source.pxi":207
 *         pth.arc_to(x1, y1, x2, y2, radius)
 * 
 *     def quadric_to(self, double x_ctrl, double y_ctrl, double x_to, double y_to):             # <<<<<<<<<<<<<<
 *         """Adds a quadratic Bezier curve to the path, starting from the
 *         current position.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":219
 *         pth.quadric_to(x_ctrl, y_ctrl, x_to, y_to)
 * 
 *     def cubic_to(self, double x_ctrl1, double y_ctrl1,             # <<<<<<<<<<<<<<
 *                  double x_ctrl2, double y_ctrl2, double x_to, double y_to):
 *         """cubic_to(x_ctrl1, y_ctrl1, x_ctrl2, y_ctrl2, x_to, y_to)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_23cubic_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_22cubic_to[] = "cubic_to(x_ctrl1, y_ctrl1, x_ctrl2, y_ctrl2, x_to, y_to)\n        Adds a cubic Bezier curve to the path, starting from the current pen\n        position.\n\n        :param x_ctrl1: First control point X\n        :param y_ctrl1: First control point Y\n        :param x_ctrl2: Second control point X\n        :param y_ctrl2: Second control point Y\n        :param x_to: End point X\n        :param y_to: End point Y\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_23cubic_to(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x_ctrl1;
  double __pyx_v_y_ctrl1;
  double __pyx_v_x_ctrl2;
  double __pyx_v_y_ctrl2;
  double __pyx_v_x_to;
  double __pyx_v_y_to;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cubic_to (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x_ctrl1,&__pyx_n_s_y_ctrl1,&__pyx_n_s_x_ctrl2,&__pyx_n_s_y_ctrl2,&__pyx_n_s_x_to,&__pyx_n_s_y_to,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x_ctrl1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_ctrl1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cubic_to", 1, 6, 6, 1); __PYX_ERR(2, 219, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x_ctrl2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cubic_to", 1, 6, 6, 2); __PYX_ERR(2, 219, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_ctrl2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cubic_to", 1, 6, 6, 3); __PYX_ERR(2, 219, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x_to)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cubic_to", 1, 6, 6, 4); __PYX_ERR(2, 219, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y_to)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cubic_to", 1, 6, 6, 5); __PYX_ERR(2, 219, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cubic_to") < 0)) __PYX_ERR(2, 219, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_x_ctrl1 = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x_ctrl1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 219, __pyx_L3_error)
    __pyx_v_y_ctrl1 = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y_ctrl1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 219, __pyx_L3_error)
    __pyx_v_x_ctrl2 = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_x_ctrl2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 220, __pyx_L3_error)
    __pyx_v_y_ctrl2 = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_y_ctrl2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 220, __pyx_L3_error)
    __pyx_v_x_to = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_x_to == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 220, __pyx_L3_error)
    __pyx_v_y_to = __pyx_PyFloat_AsDouble(values[5]); if (unlikely((__pyx_v_y_to == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 220, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cubic_to", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 219, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Path.cubic_to", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_22cubic_to(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), __pyx_v_x_ctrl1, __pyx_v_y_ctrl1, __pyx_v_x_ctrl2, __pyx_v_y_ctrl2, __pyx_v_x_to, __pyx_v_y_to);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_22cubic_to(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x_ctrl1, double __pyx_v_y_ctrl1, double __pyx_v_x_ctrl2, double __pyx_v_y_ctrl2, double __pyx_v_x_to, double __pyx_v_y_to) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cubic_to", 0);

  /* "celiagg/vertex_source.pxi":232
 *         :param y_to: End point Y
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.cubic_to(x_ctrl1, y_ctrl1, x_ctrl2, y_ctrl2, x_to, y_to)
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":233
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.cubic_to(x_ctrl1, y_ctrl1, x_ctrl2, y_ctrl2, x_to, y_to)             # <<<<<<<<<<<<<<
 * 
 *     def arc(self, double x, double y, double radius, double start_angle,
 */
  __pyx_v_pth->cubic_to(__pyx_v_x_ctrl1, __pyx_v_y_ctrl1, __pyx_v_x_ctrl2, __pyx_v_y_ctrl2, __pyx_v_x_to, __pyx_v_y_to);

  /* "celiagg/vertex_source.pxi":219
 *         pth.quadric_to(x_ctrl, y_ctrl, x_to, y_to)
 * 
 *     def cubic_to(self, double x_ctrl1, double y_ctrl1,             # <<<<<<<<<<<<<<
 *                  double x_ctrl2, double y_ctrl2, double x_to, double y_to):
 *         """cubic_to(x_ctrl1, y_ctrl1, x_ctrl2, y_ctrl2, x_to, y_to)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":235
 *         pth.cubic_to(x_ctrl1, y_ctrl1, x_ctrl2, y_ctrl2, x_to, y_to)
 * 
 *     def arc(self, double x, double y, double radius, double start_angle,             # <<<<<<<<<<<<<<
 *             double end_angle, bool clockwise):
 *         """arc(x, y, radius, start_angle, end_angle, clockwise)
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_25arc(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_24arc[] = "arc(x, y, radius, start_angle, end_angle, clockwise)\n        Adds an arc to the path.\n\n        :param x: Center X coordinate of the arc\n        :param y: Center Y coordinate of the arc\n        :param radius: Radius of the arc\n        :param start_angle: Starting angle, in radians\n        :param end_angle: Ending angle, in radians\n        :param clockwise: If True, the arc is drawn in a clockwise direction\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_25arc(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x;
  double __pyx_v_y;
  double __pyx_v_radius;
  double __pyx_v_start_angle;
  double __pyx_v_end_angle;
  bool __pyx_v_clockwise;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("arc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_radius,&__pyx_n_s_start_angle,&__pyx_n_s_end_angle,&__pyx_n_s_clockwise,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("arc", 1, 6, 6, 1); __PYX_ERR(2, 235, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_radius)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("arc", 1, 6, 6, 2); __PYX_ERR(2, 235, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start_angle)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("arc", 1, 6, 6, 3); __PYX_ERR(2, 235, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end_angle)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("arc", 1, 6, 6, 4); __PYX_ERR(2, 235, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_clockwise)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("arc", 1, 6, 6, 5); __PYX_ERR(2, 235, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "arc") < 0)) __PYX_ERR(2, 235, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 6) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
    }
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 235, __pyx_L3_error)
    __pyx_v_y = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 235, __pyx_L3_error)
    __pyx_v_radius = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_radius == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 235, __pyx_L3_error)
    __pyx_v_start_angle = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_start_angle == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 235, __pyx_L3_error)
    __pyx_v_end_angle = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_end_angle == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 236, __pyx_L3_error)
    __pyx_v_clockwise = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_clockwise == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(2, 236, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("arc", 1, 6, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 235, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Path.arc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_24arc(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), __pyx_v_x, __pyx_v_y, __pyx_v_radius, __pyx_v_start_angle, __pyx_v_end_angle, __pyx_v_clockwise);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_24arc(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x, double __pyx_v_y, double __pyx_v_radius, double __pyx_v_start_angle, double __pyx_v_end_angle, bool __pyx_v_clockwise) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("arc", 0);

  /* "celiagg/vertex_source.pxi":247
 *         :param clockwise: If True, the arc is drawn in a clockwise direction
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.arc(x, y, radius, start_angle, end_angle, clockwise)
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":248
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.arc(x, y, radius, start_angle, end_angle, clockwise)             # <<<<<<<<<<<<<<
 * 
 *     def ellipse(self, double cx, double cy, double rx, double ry):
 */
  __pyx_v_pth->arc(__pyx_v_x, __pyx_v_y, __pyx_v_radius, __pyx_v_start_angle, __pyx_v_end_angle, __pyx_v_clockwise);

  /* "celiagg/vertex_source.pxi":235
 *         pth.cubic_to(x_ctrl1, y_ctrl1, x_ctrl2, y_ctrl2, x_to, y_to)
 * 
 *     def arc(self, double x, double y, double radius, double start_angle,             # <<<<<<<<<<<<<<
 *             double end_angle, bool clockwise):
 *         """arc(x, y, radius, start_angle, end_angle, clockwise)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":250
 *         pth.arc(x, y, radius, start_angle, end_angle, clockwise)
 * 
 *     def ellipse(self, double cx, double cy, double rx, double ry):             # <<<<<<<<<<<<<<
 *         """ellipse(cx, cy, rx, ry)
 *         Adds an ellipse to the path.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_27ellipse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_26ellipse[] = "ellipse(cx, cy, rx, ry)\n        Adds an ellipse to the path.\n\n        :param cx: Center X coordinate of the ellipse\n        :param cy: Center Y coordinate of the ellipse\n        :param rx: Ellipse radii x dimension.\n        :param ry: Ellipse radii y dimension.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_27ellipse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_cx;
  double __pyx_v_cy;
  double __pyx_v_rx;
  double __pyx_v_ry;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ellipse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cx,&__pyx_n_s_cy,&__pyx_n_s_rx,&__pyx_n_s_ry,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cx)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cy)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ellipse", 1, 4, 4, 1); __PYX_ERR(2, 250, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ellipse", 1, 4, 4, 2); __PYX_ERR(2, 250, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ry)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ellipse", 1, 4, 4, 3); __PYX_ERR(2, 250, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ellipse") < 0)) __PYX_ERR(2, 250, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_cx = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_cx == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 250, __pyx_L3_error)
    __pyx_v_cy = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_cy == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 250, __pyx_L3_error)
    __pyx_v_rx = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_rx == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 250, __pyx_L3_error)
    __pyx_v_ry = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_ry == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 250, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ellipse", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 250, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Path.ellipse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_26ellipse(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), __pyx_v_cx, __pyx_v_cy, __pyx_v_rx, __pyx_v_ry);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_26ellipse(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_cx, double __pyx_v_cy, double __pyx_v_rx, double __pyx_v_ry) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ellipse", 0);

  /* "celiagg/vertex_source.pxi":259
 *         :param ry: Ellipse radii y dimension.
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         pth.ellipse(cx, cy, rx, ry)
 * 
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":260
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         pth.ellipse(cx, cy, rx, ry)             # <<<<<<<<<<<<<<
 * 
 *     def rect(self, double x, double y, double width, double height):
 */
  __pyx_v_pth->ellipse(__pyx_v_cx, __pyx_v_cy, __pyx_v_rx, __pyx_v_ry);

  /* "celiagg/vertex_source.pxi":250
 *         pth.arc(x, y, radius, start_angle, end_angle, clockwise)
 * 
 *     def ellipse(self, double cx, double cy, double rx, double ry):             # <<<<<<<<<<<<<<
 *         """ellipse(cx, cy, rx, ry)
 *         Adds an ellipse to the path.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":262
 *         pth.ellipse(cx, cy, rx, ry)
 * 
 *     def rect(self, double x, double y, double width, double height):             # <<<<<<<<<<<<<<
 *         """rect(x, y, width, height)
 *         Adds a rectangle to the path.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_29rect(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_28rect[] = "rect(x, y, width, height)\n        Adds a rectangle to the path.\n\n        :param x: Rectangle X position\n        :param y: Rectangle Y position\n        :param width: Rectangle widtht\n        :param height: Rectangle height\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_29rect(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_x;
  double __pyx_v_y;
  double __pyx_v_width;
  double __pyx_v_height;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rect (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_width,&__pyx_n_s_height,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rect", 1, 4, 4, 1); __PYX_ERR(2, 262, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_width)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rect", 1, 4, 4, 2); __PYX_ERR(2, 262, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_height)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rect", 1, 4, 4, 3); __PYX_ERR(2, 262, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "rect") < 0)) __PYX_ERR(2, 262, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 262, __pyx_L3_error)
    __pyx_v_y = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_y == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 262, __pyx_L3_error)
    __pyx_v_width = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_width == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 262, __pyx_L3_error)
    __pyx_v_height = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_height == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 262, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("rect", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 262, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Path.rect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_28rect(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), __pyx_v_x, __pyx_v_y, __pyx_v_width, __pyx_v_height);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_28rect(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, double __pyx_v_x, double __pyx_v_y, double __pyx_v_width, double __pyx_v_height) {
  PathSource *__pyx_v_pth;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rect", 0);

  /* "celiagg/vertex_source.pxi":271
 *         :param height: Rectangle height
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 * 
 *         pth.move_to(x, y)
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":273
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 * 
 *         pth.move_to(x, y)             # <<<<<<<<<<<<<<
 *         pth.line_to(x + width, y)
 *         pth.line_to(x + width, y + height)
 */
  __pyx_v_pth->move_to(__pyx_v_x, __pyx_v_y);

  /* "celiagg/vertex_source.pxi":274
 * 
 *         pth.move_to(x, y)
 *         pth.line_to(x + width, y)             # <<<<<<<<<<<<<<
 *         pth.line_to(x + width, y + height)
 *         pth.line_to(x, y + height)
 */
  __pyx_v_pth->line_to((__pyx_v_x + __pyx_v_width), __pyx_v_y);

  /* "celiagg/vertex_source.pxi":275
 *         pth.move_to(x, y)
 *         pth.line_to(x + width, y)
 *         pth.line_to(x + width, y + height)             # <<<<<<<<<<<<<<
 *         pth.line_to(x, y + height)
 *         pth.close()
 */
  __pyx_v_pth->line_to((__pyx_v_x + __pyx_v_width), (__pyx_v_y + __pyx_v_height));

  /* "celiagg/vertex_source.pxi":276
 *         pth.line_to(x + width, y)
 *         pth.line_to(x + width, y + height)
 *         pth.line_to(x, y + height)             # <<<<<<<<<<<<<<
 *         pth.close()
 * 
 */
  __pyx_v_pth->line_to(__pyx_v_x, (__pyx_v_y + __pyx_v_height));

  /* "celiagg/vertex_source.pxi":277
 *         pth.line_to(x + width, y + height)
 *         pth.line_to(x, y + height)
 *         pth.close()             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_v_pth->close();

  /* "celiagg/vertex_source.pxi":262
 *         pth.ellipse(cx, cy, rx, ry)
 * 
 *     def rect(self, double x, double y, double width, double height):             # <<<<<<<<<<<<<<
 *         """rect(x, y, width, height)
 *         Adds a rectangle to the path.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":281
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     def rects(self, rects):             # <<<<<<<<<<<<<<
 *         """rects(rects)
 *         Adds multiple rectangles to the path.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_31rects(PyObject *__pyx_v_self, PyObject *__pyx_v_rects); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_30rects[] = "rects(rects)\n        Adds multiple rectangles to the path.\n\n        :param rects: 2D array of (x, y, w, h) sequences\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_31rects(PyObject *__pyx_v_self, PyObject *__pyx_v_rects) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rects (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_30rects(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), ((PyObject *)__pyx_v_rects));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_30rects(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, PyObject *__pyx_v_rects) {
  PathSource *__pyx_v_pth;
  __Pyx_memviewslice __pyx_v__rects = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_count;
  double __pyx_v_x;
  double __pyx_v_y;
  double __pyx_v_w;
  double __pyx_v_h;
  PyObject *__pyx_v_msg = NULL;
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  long __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  __Pyx_RefNannySetupContext("rects", 0);

  /* "celiagg/vertex_source.pxi":287
 *         :param rects: 2D array of (x, y, w, h) sequences
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         cdef double[:,::1] _rects = numpy.asarray(rects, dtype=numpy.float64,
 *                                                   order='c')
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":288
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         cdef double[:,::1] _rects = numpy.asarray(rects, dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                                   order='c')
 *         cdef int count = _rects.shape[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_rects);
  __Pyx_GIVEREF(__pyx_v_rects);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_rects);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(2, 288, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v__rects = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "celiagg/vertex_source.pxi":290
 *         cdef double[:,::1] _rects = numpy.asarray(rects, dtype=numpy.float64,
 *                                                   order='c')
 *         cdef int count = _rects.shape[0]             # <<<<<<<<<<<<<<
 *         cdef double x, y, w, h
 * 
 */
  __pyx_v_count = (__pyx_v__rects.shape[0]);

  /* "celiagg/vertex_source.pxi":293
 *         cdef double x, y, w, h
 * 
 *         if _rects.shape[1] != 4:             # <<<<<<<<<<<<<<
 *             msg = "rects argument must be an iterable of (x, y, w, h) tuples."
 *             raise ValueError(msg)
 */
  __pyx_t_7 = (((__pyx_v__rects.shape[1]) != 4) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "celiagg/vertex_source.pxi":294
 * 
 *         if _rects.shape[1] != 4:
 *             msg = "rects argument must be an iterable of (x, y, w, h) tuples."             # <<<<<<<<<<<<<<
 *             raise ValueError(msg)
 * 
 */
    __Pyx_INCREF(__pyx_kp_s_rects_argument_must_be_an_iterab);
    __pyx_v_msg = __pyx_kp_s_rects_argument_must_be_an_iterab;

    /* "celiagg/vertex_source.pxi":295
 *         if _rects.shape[1] != 4:
 *             msg = "rects argument must be an iterable of (x, y, w, h) tuples."
 *             raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(0, count):
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 295, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(2, 295, __pyx_L1_error)

    /* "celiagg/vertex_source.pxi":293
 *         cdef double x, y, w, h
 * 
 *         if _rects.shape[1] != 4:             # <<<<<<<<<<<<<<
 *             msg = "rects argument must be an iterable of (x, y, w, h) tuples."
 *             raise ValueError(msg)
 */
  }

  /* "celiagg/vertex_source.pxi":297
 *             raise ValueError(msg)
 * 
 *         for i in range(0, count):             # <<<<<<<<<<<<<<
 *             x = _rects[i, 0]
 *             y = _rects[i, 1]
 */
  __pyx_t_8 = __pyx_v_count;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "celiagg/vertex_source.pxi":298
 * 
 *         for i in range(0, count):
 *             x = _rects[i, 0]             # <<<<<<<<<<<<<<
 *             y = _rects[i, 1]
 *             w = _rects[i, 2]
 */
    __pyx_t_11 = __pyx_v_i;
    __pyx_t_12 = 0;
    __pyx_v_x = (*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__rects.data + __pyx_t_11 * __pyx_v__rects.strides[0]) )) + __pyx_t_12)) )));

    /* "celiagg/vertex_source.pxi":299
 *         for i in range(0, count):
 *             x = _rects[i, 0]
 *             y = _rects[i, 1]             # <<<<<<<<<<<<<<
 *             w = _rects[i, 2]
 *             h = _rects[i, 3]
 */
    __pyx_t_13 = __pyx_v_i;
    __pyx_t_14 = 1;
    __pyx_v_y = (*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__rects.data + __pyx_t_13 * __pyx_v__rects.strides[0]) )) + __pyx_t_14)) )));

    /* "celiagg/vertex_source.pxi":300
 *             x = _rects[i, 0]
 *             y = _rects[i, 1]
 *             w = _rects[i, 2]             # <<<<<<<<<<<<<<
 *             h = _rects[i, 3]
 * 
 */
    __pyx_t_15 = __pyx_v_i;
    __pyx_t_16 = 2;
    __pyx_v_w = (*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__rects.data + __pyx_t_15 * __pyx_v__rects.strides[0]) )) + __pyx_t_16)) )));

    /* "celiagg/vertex_source.pxi":301
 *             y = _rects[i, 1]
 *             w = _rects[i, 2]
 *             h = _rects[i, 3]             # <<<<<<<<<<<<<<
 * 
 *             pth.move_to(x, y)
 */
    __pyx_t_17 = __pyx_v_i;
    __pyx_t_18 = 3;
    __pyx_v_h = (*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__rects.data + __pyx_t_17 * __pyx_v__rects.strides[0]) )) + __pyx_t_18)) )));

    /* "celiagg/vertex_source.pxi":303
 *             h = _rects[i, 3]
 * 
 *             pth.move_to(x, y)             # <<<<<<<<<<<<<<
 *             pth.line_to(x + w, y)
 *             pth.line_to(x + w, y + h)
 */
    __pyx_v_pth->move_to(__pyx_v_x, __pyx_v_y);

    /* "celiagg/vertex_source.pxi":304
 * 
 *             pth.move_to(x, y)
 *             pth.line_to(x + w, y)             # <<<<<<<<<<<<<<
 *             pth.line_to(x + w, y + h)
 *             pth.line_to(x, y + h)
 */
    __pyx_v_pth->line_to((__pyx_v_x + __pyx_v_w), __pyx_v_y);

    /* "celiagg/vertex_source.pxi":305
 *             pth.move_to(x, y)
 *             pth.line_to(x + w, y)
 *             pth.line_to(x + w, y + h)             # <<<<<<<<<<<<<<
 *             pth.line_to(x, y + h)
 *             pth.close()
 */
    __pyx_v_pth->line_to((__pyx_v_x + __pyx_v_w), (__pyx_v_y + __pyx_v_h));

    /* "celiagg/vertex_source.pxi":306
 *             pth.line_to(x + w, y)
 *             pth.line_to(x + w, y + h)
 *             pth.line_to(x, y + h)             # <<<<<<<<<<<<<<
 *             pth.close()
 * 
 */
    __pyx_v_pth->line_to(__pyx_v_x, (__pyx_v_y + __pyx_v_h));

    /* "celiagg/vertex_source.pxi":307
 *             pth.line_to(x + w, y + h)
 *             pth.line_to(x, y + h)
 *             pth.close()             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_v_pth->close();
  }

  /* "celiagg/vertex_source.pxi":281
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     def rects(self, rects):             # <<<<<<<<<<<<<<
 *         """rects(rects)
 *         Adds multiple rectangles to the path.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("celiagg._celiagg.Path.rects", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v__rects, 1);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":311
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     def lines(self, points):             # <<<<<<<<<<<<<<
 *         """lines(points)
 *         Adds multiple connected line segmests to the path.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_33lines(PyObject *__pyx_v_self, PyObject *__pyx_v_points); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_32lines[] = "lines(points)\n        Adds multiple connected line segmests to the path.\n\n        :param points: 2D array of (x, y) pairs\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_33lines(PyObject *__pyx_v_self, PyObject *__pyx_v_points) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lines (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_32lines(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), ((PyObject *)__pyx_v_points));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_32lines(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, PyObject *__pyx_v_points) {
  PathSource *__pyx_v_pth;
  __Pyx_memviewslice __pyx_v__points = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_count;
  PyObject *__pyx_v_msg = NULL;
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  long __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  __Pyx_RefNannySetupContext("lines", 0);

  /* "celiagg/vertex_source.pxi":317
 *         :param points: 2D array of (x, y) pairs
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         cdef double[:,::1] _points = numpy.asarray(points,
 *                                                       dtype=numpy.float64,
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":318
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         cdef double[:,::1] _points = numpy.asarray(points,             # <<<<<<<<<<<<<<
 *                                                       dtype=numpy.float64,
 *                                                       order='c')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_points);
  __Pyx_GIVEREF(__pyx_v_points);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_points);

  /* "celiagg/vertex_source.pxi":319
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         cdef double[:,::1] _points = numpy.asarray(points,
 *                                                       dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                                       order='c')
 *         cdef int count = _points.shape[0]
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(2, 319, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(2, 319, __pyx_L1_error)

  /* "celiagg/vertex_source.pxi":318
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         cdef double[:,::1] _points = numpy.asarray(points,             # <<<<<<<<<<<<<<
 *                                                       dtype=numpy.float64,
 *                                                       order='c')
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(2, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v__points = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "celiagg/vertex_source.pxi":321
 *                                                       dtype=numpy.float64,
 *                                                       order='c')
 *         cdef int count = _points.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         if _points.shape[1] != 2:
 */
  __pyx_v_count = (__pyx_v__points.shape[0]);

  /* "celiagg/vertex_source.pxi":323
 *         cdef int count = _points.shape[0]
 * 
 *         if _points.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = "points argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 */
  __pyx_t_7 = (((__pyx_v__points.shape[1]) != 2) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "celiagg/vertex_source.pxi":324
 * 
 *         if _points.shape[1] != 2:
 *             msg = "points argument must be an iterable of (x, y) tuples."             # <<<<<<<<<<<<<<
 *             raise ValueError(msg)
 * 
 */
    __Pyx_INCREF(__pyx_kp_s_points_argument_must_be_an_itera_2);
    __pyx_v_msg = __pyx_kp_s_points_argument_must_be_an_itera_2;

    /* "celiagg/vertex_source.pxi":325
 *         if _points.shape[1] != 2:
 *             msg = "points argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *         pth.move_to(_points[0, 0], _points[0, 1])
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(2, 325, __pyx_L1_error)

    /* "celiagg/vertex_source.pxi":323
 *         cdef int count = _points.shape[0]
 * 
 *         if _points.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = "points argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 */
  }

  /* "celiagg/vertex_source.pxi":327
 *             raise ValueError(msg)
 * 
 *         pth.move_to(_points[0, 0], _points[0, 1])             # <<<<<<<<<<<<<<
 *         for i in range(1, count):
 *             pth.line_to(_points[i, 0], _points[i, 1])
 */
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 1;
  __pyx_v_pth->move_to((*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__points.data + __pyx_t_8 * __pyx_v__points.strides[0]) )) + __pyx_t_9)) ))), (*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__points.data + __pyx_t_10 * __pyx_v__points.strides[0]) )) + __pyx_t_11)) ))));

  /* "celiagg/vertex_source.pxi":328
 * 
 *         pth.move_to(_points[0, 0], _points[0, 1])
 *         for i in range(1, count):             # <<<<<<<<<<<<<<
 *             pth.line_to(_points[i, 0], _points[i, 1])
 * 
 */
  __pyx_t_12 = __pyx_v_count;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 1; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "celiagg/vertex_source.pxi":329
 *         pth.move_to(_points[0, 0], _points[0, 1])
 *         for i in range(1, count):
 *             pth.line_to(_points[i, 0], _points[i, 1])             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_t_15 = __pyx_v_i;
    __pyx_t_16 = 0;
    __pyx_t_17 = __pyx_v_i;
    __pyx_t_18 = 1;
    __pyx_v_pth->line_to((*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__points.data + __pyx_t_15 * __pyx_v__points.strides[0]) )) + __pyx_t_16)) ))), (*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__points.data + __pyx_t_17 * __pyx_v__points.strides[0]) )) + __pyx_t_18)) ))));
  }

  /* "celiagg/vertex_source.pxi":311
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     def lines(self, points):             # <<<<<<<<<<<<<<
 *         """lines(points)
 *         Adds multiple connected line segmests to the path.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("celiagg._celiagg.Path.lines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v__points, 1);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":333
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     def lines_set(self, starts, ends):             # <<<<<<<<<<<<<<
 *         """lines_set(starts, ends)
 *         Adds multiple disconnected line segments to the path.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_35lines_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_4Path_34lines_set[] = "lines_set(starts, ends)\n        Adds multiple disconnected line segments to the path.\n\n        :param starts: 2D array of (x, y) pairs\n        :param ends: 2D array of (x, y) pairs\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_35lines_set(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_starts = 0;
  PyObject *__pyx_v_ends = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lines_set (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_starts,&__pyx_n_s_ends,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_starts)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ends)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("lines_set", 1, 2, 2, 1); __PYX_ERR(2, 333, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "lines_set") < 0)) __PYX_ERR(2, 333, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_starts = values[0];
    __pyx_v_ends = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lines_set", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 333, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.Path.lines_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_34lines_set(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), __pyx_v_starts, __pyx_v_ends);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_34lines_set(struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, PyObject *__pyx_v_starts, PyObject *__pyx_v_ends) {
  PathSource *__pyx_v_pth;
  __Pyx_memviewslice __pyx_v__starts = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v__ends = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_v_starts_count;
  int __pyx_v_ends_count;
  PyObject *__pyx_v_msg = NULL;
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  long __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  __Pyx_RefNannySetupContext("lines_set", 0);

  /* "celiagg/vertex_source.pxi":340
 *         :param ends: 2D array of (x, y) pairs
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this             # <<<<<<<<<<<<<<
 *         cdef double[:,::1] _starts = numpy.asarray(starts, dtype=numpy.float64,
 *                                                    order='c')
 */
  __pyx_v_pth = ((PathSource *)__pyx_v_self->__pyx_base._this);

  /* "celiagg/vertex_source.pxi":341
 *         """
 *         cdef _vertex_source.PathSource* pth = <_vertex_source.PathSource*>self._this
 *         cdef double[:,::1] _starts = numpy.asarray(starts, dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                                    order='c')
 *         cdef double[:,::1] _ends = numpy.asarray(ends, dtype=numpy.float64,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_starts);
  __Pyx_GIVEREF(__pyx_v_starts);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_starts);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_numpy); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(2, 341, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(2, 341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v__starts = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "celiagg/vertex_source.pxi":343
 *         cdef double[:,::1] _starts = numpy.asarray(starts, dtype=numpy.float64,
 *                                                    order='c')
 *         cdef double[:,::1] _ends = numpy.asarray(ends, dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                                  order='c')
 *         cdef int starts_count = _starts.shape[0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_numpy); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_ends);
  __Pyx_GIVEREF(__pyx_v_ends);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_ends);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_numpy); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float64); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_4) < 0) __PYX_ERR(2, 343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(2, 343, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_4, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(2, 343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v__ends = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "celiagg/vertex_source.pxi":345
 *         cdef double[:,::1] _ends = numpy.asarray(ends, dtype=numpy.float64,
 *                                                  order='c')
 *         cdef int starts_count = _starts.shape[0]             # <<<<<<<<<<<<<<
 *         cdef int ends_count = _ends.shape[0]
 * 
 */
  __pyx_v_starts_count = (__pyx_v__starts.shape[0]);

  /* "celiagg/vertex_source.pxi":346
 *                                                  order='c')
 *         cdef int starts_count = _starts.shape[0]
 *         cdef int ends_count = _ends.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         if _starts.shape[1] != 2:
 */
  __pyx_v_ends_count = (__pyx_v__ends.shape[0]);

  /* "celiagg/vertex_source.pxi":348
 *         cdef int ends_count = _ends.shape[0]
 * 
 *         if _starts.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = "starts argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 */
  __pyx_t_7 = (((__pyx_v__starts.shape[1]) != 2) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "celiagg/vertex_source.pxi":349
 * 
 *         if _starts.shape[1] != 2:
 *             msg = "starts argument must be an iterable of (x, y) tuples."             # <<<<<<<<<<<<<<
 *             raise ValueError(msg)
 *         if _ends.shape[1] != 2:
 */
    __Pyx_INCREF(__pyx_kp_s_starts_argument_must_be_an_itera);
    __pyx_v_msg = __pyx_kp_s_starts_argument_must_be_an_itera;

    /* "celiagg/vertex_source.pxi":350
 *         if _starts.shape[1] != 2:
 *             msg = "starts argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)             # <<<<<<<<<<<<<<
 *         if _ends.shape[1] != 2:
 *             msg = "ends argument must be an iterable of (x, y) tuples."
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 350, __pyx_L1_error)

    /* "celiagg/vertex_source.pxi":348
 *         cdef int ends_count = _ends.shape[0]
 * 
 *         if _starts.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = "starts argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 */
  }

  /* "celiagg/vertex_source.pxi":351
 *             msg = "starts argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 *         if _ends.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = "ends argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 */
  __pyx_t_7 = (((__pyx_v__ends.shape[1]) != 2) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "celiagg/vertex_source.pxi":352
 *             raise ValueError(msg)
 *         if _ends.shape[1] != 2:
 *             msg = "ends argument must be an iterable of (x, y) tuples."             # <<<<<<<<<<<<<<
 *             raise ValueError(msg)
 *         if starts_count != ends_count:
 */
    __Pyx_INCREF(__pyx_kp_s_ends_argument_must_be_an_iterabl);
    __pyx_v_msg = __pyx_kp_s_ends_argument_must_be_an_iterabl;

    /* "celiagg/vertex_source.pxi":353
 *         if _ends.shape[1] != 2:
 *             msg = "ends argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)             # <<<<<<<<<<<<<<
 *         if starts_count != ends_count:
 *             msg = "starts argument must match length of ends argument."
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 353, __pyx_L1_error)

    /* "celiagg/vertex_source.pxi":351
 *             msg = "starts argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 *         if _ends.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = "ends argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 */
  }

  /* "celiagg/vertex_source.pxi":354
 *             msg = "ends argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 *         if starts_count != ends_count:             # <<<<<<<<<<<<<<
 *             msg = "starts argument must match length of ends argument."
 *             raise ValueError(msg)
 */
  __pyx_t_7 = ((__pyx_v_starts_count != __pyx_v_ends_count) != 0);
  if (unlikely(__pyx_t_7)) {

    /* "celiagg/vertex_source.pxi":355
 *             raise ValueError(msg)
 *         if starts_count != ends_count:
 *             msg = "starts argument must match length of ends argument."             # <<<<<<<<<<<<<<
 *             raise ValueError(msg)
 * 
 */
    __Pyx_INCREF(__pyx_kp_s_starts_argument_must_match_lengt);
    __pyx_v_msg = __pyx_kp_s_starts_argument_must_match_lengt;

    /* "celiagg/vertex_source.pxi":356
 *         if starts_count != ends_count:
 *             msg = "starts argument must match length of ends argument."
 *             raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(0, starts_count):
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 356, __pyx_L1_error)

    /* "celiagg/vertex_source.pxi":354
 *             msg = "ends argument must be an iterable of (x, y) tuples."
 *             raise ValueError(msg)
 *         if starts_count != ends_count:             # <<<<<<<<<<<<<<
 *             msg = "starts argument must match length of ends argument."
 *             raise ValueError(msg)
 */
  }

  /* "celiagg/vertex_source.pxi":358
 *             raise ValueError(msg)
 * 
 *         for i in range(0, starts_count):             # <<<<<<<<<<<<<<
 *             pth.move_to(_starts[i, 0], _starts[i, 1])
 *             pth.line_to(_ends[i, 0], _ends[i, 1])
 */
  __pyx_t_8 = __pyx_v_starts_count;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "celiagg/vertex_source.pxi":359
 * 
 *         for i in range(0, starts_count):
 *             pth.move_to(_starts[i, 0], _starts[i, 1])             # <<<<<<<<<<<<<<
 *             pth.line_to(_ends[i, 0], _ends[i, 1])
 * 
 */
    __pyx_t_11 = __pyx_v_i;
    __pyx_t_12 = 0;
    __pyx_t_13 = __pyx_v_i;
    __pyx_t_14 = 1;
    __pyx_v_pth->move_to((*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__starts.data + __pyx_t_11 * __pyx_v__starts.strides[0]) )) + __pyx_t_12)) ))), (*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__starts.data + __pyx_t_13 * __pyx_v__starts.strides[0]) )) + __pyx_t_14)) ))));

    /* "celiagg/vertex_source.pxi":360
 *         for i in range(0, starts_count):
 *             pth.move_to(_starts[i, 0], _starts[i, 1])
 *             pth.line_to(_ends[i, 0], _ends[i, 1])             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_15 = __pyx_v_i;
    __pyx_t_16 = 0;
    __pyx_t_17 = __pyx_v_i;
    __pyx_t_18 = 1;
    __pyx_v_pth->line_to((*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__ends.data + __pyx_t_15 * __pyx_v__ends.strides[0]) )) + __pyx_t_16)) ))), (*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__ends.data + __pyx_t_17 * __pyx_v__ends.strides[0]) )) + __pyx_t_18)) ))));
  }

  /* "celiagg/vertex_source.pxi":333
 *     @cython.boundscheck(False)
 *     @cython.wraparound(False)
 *     def lines_set(self, starts, ends):             # <<<<<<<<<<<<<<
 *         """lines_set(starts, ends)
 *         Adds multiple disconnected line segments to the path.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("celiagg._celiagg.Path.lines_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v__starts, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v__ends, 1);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_37__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_37__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_36__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_36__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__63, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Path.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_39__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_4Path_39__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_4Path_38__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_Path *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_4Path_38__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_Path *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__64, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.Path.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":374
 *     cdef object _points
 * 
 *     def __cinit__(self, source, points):             # <<<<<<<<<<<<<<
 *         if not isinstance(source, VertexSource):
 *             raise TypeError("source must be a VertexSource instance")
 */

/* Python wrapper */
static int __pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_source = 0;
  PyObject *__pyx_v_points = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_source,&__pyx_n_s_points,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_source)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_points)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(2, 374, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(2, 374, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_source = values[0];
    __pyx_v_points = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 374, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.ShapeAtPoints.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints___cinit__(((struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *)__pyx_v_self), __pyx_v_source, __pyx_v_points);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints___cinit__(struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self, PyObject *__pyx_v_source, PyObject *__pyx_v_points) {
  struct __pyx_obj_7celiagg_8_celiagg_VertexSource *__pyx_v_vs = 0;
  __Pyx_memviewslice __pyx_v__points = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_msg = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "celiagg/vertex_source.pxi":375
 * 
 *     def __cinit__(self, source, points):
 *         if not isinstance(source, VertexSource):             # <<<<<<<<<<<<<<
 *             raise TypeError("source must be a VertexSource instance")
 * 
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_source, __pyx_ptype_7celiagg_8_celiagg_VertexSource); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/vertex_source.pxi":376
 *     def __cinit__(self, source, points):
 *         if not isinstance(source, VertexSource):
 *             raise TypeError("source must be a VertexSource instance")             # <<<<<<<<<<<<<<
 * 
 *         cdef VertexSource vs = <VertexSource>source
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 376, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 376, __pyx_L1_error)

    /* "celiagg/vertex_source.pxi":375
 * 
 *     def __cinit__(self, source, points):
 *         if not isinstance(source, VertexSource):             # <<<<<<<<<<<<<<
 *             raise TypeError("source must be a VertexSource instance")
 * 
 */
  }

  /* "celiagg/vertex_source.pxi":378
 *             raise TypeError("source must be a VertexSource instance")
 * 
 *         cdef VertexSource vs = <VertexSource>source             # <<<<<<<<<<<<<<
 *         cdef double[:,::1] _points = numpy.asarray(points, dtype=numpy.float64,
 *                                                    order='c')
 */
  __pyx_t_3 = __pyx_v_source;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_vs = ((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/vertex_source.pxi":379
 * 
 *         cdef VertexSource vs = <VertexSource>source
 *         cdef double[:,::1] _points = numpy.asarray(points, dtype=numpy.float64,             # <<<<<<<<<<<<<<
 *                                                    order='c')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_points);
  __Pyx_GIVEREF(__pyx_v_points);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_points);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_numpy); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_float64); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_7) < 0) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_order, __pyx_n_s_c) < 0) __PYX_ERR(2, 379, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(__pyx_t_7, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(2, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v__points = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "celiagg/vertex_source.pxi":382
 *                                                    order='c')
 * 
 *         if _points.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = 'Points argument must be an iterable of (x, y) pairs.'
 *             raise ValueError(msg)
 */
  __pyx_t_2 = (((__pyx_v__points.shape[1]) != 2) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/vertex_source.pxi":383
 * 
 *         if _points.shape[1] != 2:
 *             msg = 'Points argument must be an iterable of (x, y) pairs.'             # <<<<<<<<<<<<<<
 *             raise ValueError(msg)
 * 
 */
    __Pyx_INCREF(__pyx_kp_s_Points_argument_must_be_an_itera);
    __pyx_v_msg = __pyx_kp_s_Points_argument_must_be_an_itera;

    /* "celiagg/vertex_source.pxi":384
 *         if _points.shape[1] != 2:
 *             msg = 'Points argument must be an iterable of (x, y) pairs.'
 *             raise ValueError(msg)             # <<<<<<<<<<<<<<
 * 
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.RepeatedSource(
 */
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_v_msg); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(2, 384, __pyx_L1_error)

    /* "celiagg/vertex_source.pxi":382
 *                                                    order='c')
 * 
 *         if _points.shape[1] != 2:             # <<<<<<<<<<<<<<
 *             msg = 'Points argument must be an iterable of (x, y) pairs.'
 *             raise ValueError(msg)
 */
  }

  /* "celiagg/vertex_source.pxi":387
 * 
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.RepeatedSource(
 *             dereference(vs._this), &_points[0][0], _points.shape[0]             # <<<<<<<<<<<<<<
 *         )
 *         # Keep references for safety
 */
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = -1;
  if (__pyx_t_9 < 0) {
    __pyx_t_9 += __pyx_v__points.shape[0];
    if (unlikely(__pyx_t_9 < 0)) __pyx_t_11 = 0;
  } else if (unlikely(__pyx_t_9 >= __pyx_v__points.shape[0])) __pyx_t_11 = 0;
  if (__pyx_t_10 < 0) {
    __pyx_t_10 += __pyx_v__points.shape[1];
    if (unlikely(__pyx_t_10 < 0)) __pyx_t_11 = 1;
  } else if (unlikely(__pyx_t_10 >= __pyx_v__points.shape[1])) __pyx_t_11 = 1;
  if (unlikely(__pyx_t_11 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_11);
    __PYX_ERR(2, 387, __pyx_L1_error)
  }

  /* "celiagg/vertex_source.pxi":386
 *             raise ValueError(msg)
 * 
 *         self._this = <_vertex_source.VertexSource*> new _vertex_source.RepeatedSource(             # <<<<<<<<<<<<<<
 *             dereference(vs._this), &_points[0][0], _points.shape[0]
 *         )
 */
  __pyx_v_self->__pyx_base._this = ((VertexSource *)new RepeatedSource((*__pyx_v_vs->_this), (&(*((double *) ( /* dim=1 */ ((char *) (((double *) ( /* dim=0 */ (__pyx_v__points.data + __pyx_t_9 * __pyx_v__points.strides[0]) )) + __pyx_t_10)) )))), (__pyx_v__points.shape[0])));

  /* "celiagg/vertex_source.pxi":390
 *         )
 *         # Keep references for safety
 *         self._sub_source = vs             # <<<<<<<<<<<<<<
 *         self._points = _points
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_vs));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_vs));
  __Pyx_GOTREF(__pyx_v_self->_sub_source);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_sub_source));
  __pyx_v_self->_sub_source = __pyx_v_vs;

  /* "celiagg/vertex_source.pxi":391
 *         # Keep references for safety
 *         self._sub_source = vs
 *         self._points = _points             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v__points, 2, (PyObject *(*)(char *)) __pyx_memview_get_double, (int (*)(char *, PyObject *)) __pyx_memview_set_double, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __Pyx_GOTREF(__pyx_v_self->_points);
  __Pyx_DECREF(__pyx_v_self->_points);
  __pyx_v_self->_points = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "celiagg/vertex_source.pxi":374
 *     cdef object _points
 * 
 *     def __cinit__(self, source, points):             # <<<<<<<<<<<<<<
 *         if not isinstance(source, VertexSource):
 *             raise TypeError("source must be a VertexSource instance")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_AddTraceback("celiagg._celiagg.ShapeAtPoints.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_vs);
  __PYX_XDEC_MEMVIEW(&__pyx_v__points, 1);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":393
 *         self._points = _points
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the object.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_13ShapeAtPoints_2copy[] = "Returns a deep copy of the object.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_3copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_2copy(((struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_2copy(struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self) {
  PyObject *__pyx_v_source = NULL;
  PyObject *__pyx_v_points = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "celiagg/vertex_source.pxi":396
 *         """Returns a deep copy of the object.
 *         """
 *         source = self._sub_source.copy()             # <<<<<<<<<<<<<<
 *         points = self._points.copy()
 *         return ShapeAtPoints(source, points)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_sub_source), __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_source = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/vertex_source.pxi":397
 *         """
 *         source = self._sub_source.copy()
 *         points = self._points.copy()             # <<<<<<<<<<<<<<
 *         return ShapeAtPoints(source, points)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_points, __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_points = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/vertex_source.pxi":398
 *         source = self._sub_source.copy()
 *         points = self._points.copy()
 *         return ShapeAtPoints(source, points)             # <<<<<<<<<<<<<<
 * 
 *     def final_point(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_source);
  __Pyx_GIVEREF(__pyx_v_source);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_source);
  __Pyx_INCREF(__pyx_v_points);
  __Pyx_GIVEREF(__pyx_v_points);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_points);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_ShapeAtPoints), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "celiagg/vertex_source.pxi":393
 *         self._points = _points
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         """Returns a deep copy of the object.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("celiagg._celiagg.ShapeAtPoints.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_source);
  __Pyx_XDECREF(__pyx_v_points);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/vertex_source.pxi":400
 *         return ShapeAtPoints(source, points)
 * 
 *     def final_point(self):             # <<<<<<<<<<<<<<
 *         """Returns the last vertex that will be returned by the source.
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_5final_point(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_13ShapeAtPoints_4final_point[] = "Returns the last vertex that will be returned by the source.\n        ";
static PyObject *__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_5final_point(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("final_point (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_4final_point(((struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_4final_point(struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_v_tx = NULL;
  PyObject *__pyx_v_ty = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  __Pyx_RefNannySetupContext("final_point", 0);

  /* "celiagg/vertex_source.pxi":403
 *         """Returns the last vertex that will be returned by the source.
 *         """
 *         x, y = self._sub_source.last_vertex()             # <<<<<<<<<<<<<<
 *         tx, ty = self._points[-1]
 *         return (x + tx, y + ty)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_sub_source), __pyx_n_s_last_vertex); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(2, 403, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(2, 403, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(2, 403, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_x = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_y = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "celiagg/vertex_source.pxi":404
 *         """
 *         x, y = self._sub_source.last_vertex()
 *         tx, ty = self._points[-1]             # <<<<<<<<<<<<<<
 *         return (x + tx, y + ty)
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->_points, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(2, 404, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L5_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(2, 404, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L6_unpacking_done;
    __pyx_L5_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(2, 404, __pyx_L1_error)
    __pyx_L6_unpacking_done:;
  }
  __pyx_v_tx = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_ty = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "celiagg/vertex_source.pxi":405
 *         x, y = self._sub_source.last_vertex()
 *         tx, ty = self._points[-1]
 *         return (x + tx, y + ty)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_v_x, __pyx_v_tx); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_v_y, __pyx_v_ty); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/vertex_source.pxi":400
 *         return ShapeAtPoints(source, points)
 * 
 *     def final_point(self):             # <<<<<<<<<<<<<<
 *         """Returns the last vertex that will be returned by the source.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.ShapeAtPoints.final_point", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XDECREF(__pyx_v_tx);
  __Pyx_XDECREF(__pyx_v_ty);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_6__reduce_cython__(((struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__66, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.ShapeAtPoints.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_8__setstate_cython__(((struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_13ShapeAtPoints_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__67, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("celiagg._celiagg.ShapeAtPoints.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":25
 * # Authors: John Wiggins
 * 
 * cdef _get_component_indices(fmt):             # <<<<<<<<<<<<<<
 *     order_map = {
 *         (0, 0, 0, 0): (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 */

static PyObject *__pyx_f_7celiagg_8_celiagg__get_component_indices(PyObject *__pyx_v_fmt) {
  PyObject *__pyx_v_order_map = NULL;
  PyObject *__pyx_v_rev_order_map = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_o = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_fs = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  __Pyx_RefNannySetupContext("_get_component_indices", 0);

  /* "celiagg/conversion.pxi":27
 * cdef _get_component_indices(fmt):
 *     order_map = {
 *         (0, 0, 0, 0): (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),             # <<<<<<<<<<<<<<
 *         (0, 1, 2, 0): (PixelFormat.RGB24, PixelFormat.RGB48, PixelFormat.RGB96),
 *         (2, 1, 0, 0): (PixelFormat.BGR24, PixelFormat.BGR48, PixelFormat.BGR96),
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray8); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray16); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray32); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_tuple__55, __pyx_t_5) < 0) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "celiagg/conversion.pxi":28
 *     order_map = {
 *         (0, 0, 0, 0): (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 *         (0, 1, 2, 0): (PixelFormat.RGB24, PixelFormat.RGB48, PixelFormat.RGB96),             # <<<<<<<<<<<<<<
 *         (2, 1, 0, 0): (PixelFormat.BGR24, PixelFormat.BGR48, PixelFormat.BGR96),
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB24); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB48); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB96); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_tuple__68, __pyx_t_2) < 0) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "celiagg/conversion.pxi":29
 *         (0, 0, 0, 0): (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 *         (0, 1, 2, 0): (PixelFormat.RGB24, PixelFormat.RGB48, PixelFormat.RGB96),
 *         (2, 1, 0, 0): (PixelFormat.BGR24, PixelFormat.BGR48, PixelFormat.BGR96),             # <<<<<<<<<<<<<<
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR24); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR48); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR96); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_tuple__69, __pyx_t_5) < 0) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "celiagg/conversion.pxi":30
 *         (0, 1, 2, 0): (PixelFormat.RGB24, PixelFormat.RGB48, PixelFormat.RGB96),
 *         (2, 1, 0, 0): (PixelFormat.BGR24, PixelFormat.BGR48, PixelFormat.BGR96),
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),             # <<<<<<<<<<<<<<
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),
 *         (3, 0, 1, 2): (PixelFormat.ARGB32, PixelFormat.ARGB64, PixelFormat.ARGB128),
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA32); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA64); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA128); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_tuple__70, __pyx_t_2) < 0) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "celiagg/conversion.pxi":31
 *         (2, 1, 0, 0): (PixelFormat.BGR24, PixelFormat.BGR48, PixelFormat.BGR96),
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),             # <<<<<<<<<<<<<<
 *         (3, 0, 1, 2): (PixelFormat.ARGB32, PixelFormat.ARGB64, PixelFormat.ARGB128),
 *         (3, 2, 1, 0): (PixelFormat.ABGR32, PixelFormat.ABGR64, PixelFormat.ABGR128),
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA32); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA64); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA128); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_tuple__71, __pyx_t_5) < 0) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "celiagg/conversion.pxi":32
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),
 *         (3, 0, 1, 2): (PixelFormat.ARGB32, PixelFormat.ARGB64, PixelFormat.ARGB128),             # <<<<<<<<<<<<<<
 *         (3, 2, 1, 0): (PixelFormat.ABGR32, PixelFormat.ABGR64, PixelFormat.ABGR128),
 *     }
 */
  __pyx_t_5 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB32); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB64); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB128); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_tuple__72, __pyx_t_2) < 0) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "celiagg/conversion.pxi":33
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),
 *         (3, 0, 1, 2): (PixelFormat.ARGB32, PixelFormat.ARGB64, PixelFormat.ARGB128),
 *         (3, 2, 1, 0): (PixelFormat.ABGR32, PixelFormat.ABGR64, PixelFormat.ABGR128),             # <<<<<<<<<<<<<<
 *     }
 *     rev_order_map = {f: o for o, fs in order_map.items() for f in fs}
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR32); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR64); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR128); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_tuple__73, __pyx_t_5) < 0) __PYX_ERR(9, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_order_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":35
 *         (3, 2, 1, 0): (PixelFormat.ABGR32, PixelFormat.ABGR64, PixelFormat.ABGR128),
 *     }
 *     rev_order_map = {f: o for o, fs in order_map.items() for f in fs}             # <<<<<<<<<<<<<<
 *     return rev_order_map[fmt]
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 35, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyDict_Items(__pyx_v_order_map); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 35, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_4 = __pyx_t_5; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 35, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(9, 35, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(9, 35, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 35, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(9, 35, __pyx_L5_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 35, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(9, 35, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(9, 35, __pyx_L5_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        #else
        __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 35, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 35, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(9, 35, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
        index = 0; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_2 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_2)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(9, 35, __pyx_L5_error)
        __pyx_t_9 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(9, 35, __pyx_L5_error)
        __pyx_L9_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_o, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_fs, __pyx_t_2);
      __pyx_t_2 = 0;
      if (likely(PyList_CheckExact(__pyx_8genexpr2__pyx_v_fs)) || PyTuple_CheckExact(__pyx_8genexpr2__pyx_v_fs)) {
        __pyx_t_5 = __pyx_8genexpr2__pyx_v_fs; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_8genexpr2__pyx_v_fs); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 35, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_11 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(9, 35, __pyx_L5_error)
      }
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(9, 35, __pyx_L5_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 35, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(9, 35, __pyx_L5_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 35, __pyx_L5_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_11(__pyx_t_5);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(9, 35, __pyx_L5_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_f, __pyx_t_2);
        __pyx_t_2 = 0;
        if (unlikely(PyDict_SetItem(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_f, (PyObject*)__pyx_8genexpr2__pyx_v_o))) __PYX_ERR(9, 35, __pyx_L5_error)
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_f); __pyx_8genexpr2__pyx_v_f = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_fs); __pyx_8genexpr2__pyx_v_fs = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_o); __pyx_8genexpr2__pyx_v_o = 0;
    goto __pyx_L12_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_f); __pyx_8genexpr2__pyx_v_f = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_fs); __pyx_8genexpr2__pyx_v_fs = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_o); __pyx_8genexpr2__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L12_exit_scope:;
  } /* exit inner scope */
  __pyx_v_rev_order_map = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":36
 *     }
 *     rev_order_map = {f: o for o, fs in order_map.items() for f in fs}
 *     return rev_order_map[fmt]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_rev_order_map, __pyx_v_fmt); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":25
 * # Authors: John Wiggins
 * 
 * cdef _get_component_indices(fmt):             # <<<<<<<<<<<<<<
 *     order_map = {
 *         (0, 0, 0, 0): (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("celiagg._celiagg._get_component_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_map);
  __Pyx_XDECREF(__pyx_v_rev_order_map);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_o);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_fs);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":39
 * 
 * 
 * cdef bool _has_alpha(fmt):             # <<<<<<<<<<<<<<
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 */

static bool __pyx_f_7celiagg_8_celiagg__has_alpha(PyObject *__pyx_v_fmt) {
  bool __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("_has_alpha", 0);

  /* "celiagg/conversion.pxi":40
 * 
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,             # <<<<<<<<<<<<<<
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 */
  __Pyx_INCREF(__pyx_v_fmt);
  __pyx_t_1 = __pyx_v_fmt;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray8); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray16); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray32); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }

  /* "celiagg/conversion.pxi":41
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,             # <<<<<<<<<<<<<<
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR24); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":40
 * 
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,             # <<<<<<<<<<<<<<
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }

  /* "celiagg/conversion.pxi":41
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,             # <<<<<<<<<<<<<<
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB24); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":40
 * 
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,             # <<<<<<<<<<<<<<
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }

  /* "celiagg/conversion.pxi":41
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,             # <<<<<<<<<<<<<<
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR48); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":40
 * 
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,             # <<<<<<<<<<<<<<
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }

  /* "celiagg/conversion.pxi":42
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB48); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":40
 * 
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,             # <<<<<<<<<<<<<<
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }

  /* "celiagg/conversion.pxi":42
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR96); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":40
 * 
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,             # <<<<<<<<<<<<<<
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }

  /* "celiagg/conversion.pxi":42
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB96); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":40
 * 
 * cdef bool _has_alpha(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,             # <<<<<<<<<<<<<<
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 *                        PixelFormat.RGB48,  PixelFormat.BGR96, PixelFormat.RGB96)
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_5;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":39
 * 
 * 
 * cdef bool _has_alpha(fmt):             # <<<<<<<<<<<<<<
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32,
 *                        PixelFormat.BGR24, PixelFormat.RGB24, PixelFormat.BGR48,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("celiagg._celiagg._has_alpha", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":45
 * 
 * 
 * cdef bool _is_gray(fmt):             # <<<<<<<<<<<<<<
 *     return fmt in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32)
 * 
 */

static bool __pyx_f_7celiagg_8_celiagg__is_gray(PyObject *__pyx_v_fmt) {
  bool __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("_is_gray", 0);

  /* "celiagg/conversion.pxi":46
 * 
 * cdef bool _is_gray(fmt):
 *     return fmt in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_fmt);
  __pyx_t_1 = __pyx_v_fmt;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray8); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray16); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray32); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_5;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":45
 * 
 * 
 * cdef bool _is_gray(fmt):             # <<<<<<<<<<<<<<
 *     return fmt in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("celiagg._celiagg._is_gray", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":49
 * 
 * 
 * cdef bool _is_rgb(fmt):             # <<<<<<<<<<<<<<
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32)
 * 
 */

static bool __pyx_f_7celiagg_8_celiagg__is_rgb(PyObject *__pyx_v_fmt) {
  bool __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("_is_rgb", 0);

  /* "celiagg/conversion.pxi":50
 * 
 * cdef bool _is_rgb(fmt):
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_fmt);
  __pyx_t_1 = __pyx_v_fmt;
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray8); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray16); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray32); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(9, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_5;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":49
 * 
 * 
 * cdef bool _is_rgb(fmt):             # <<<<<<<<<<<<<<
 *     return fmt not in (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("celiagg._celiagg._is_rgb", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":53
 * 
 * 
 * cdef float _get_max_value(dtype):             # <<<<<<<<<<<<<<
 *     if dtype.kind == 'f':
 *         return 1.0
 */

static float __pyx_f_7celiagg_8_celiagg__get_max_value(PyObject *__pyx_v_dtype) {
  float __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  __Pyx_RefNannySetupContext("_get_max_value", 0);

  /* "celiagg/conversion.pxi":54
 * 
 * cdef float _get_max_value(dtype):
 *     if dtype.kind == 'f':             # <<<<<<<<<<<<<<
 *         return 1.0
 *     return float(numpy.iinfo(dtype).max)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_t_1, __pyx_n_s_f, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(9, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "celiagg/conversion.pxi":55
 * cdef float _get_max_value(dtype):
 *     if dtype.kind == 'f':
 *         return 1.0             # <<<<<<<<<<<<<<
 *     return float(numpy.iinfo(dtype).max)
 * 
 */
    __pyx_r = 1.0;
    goto __pyx_L0;

    /* "celiagg/conversion.pxi":54
 * 
 * cdef float _get_max_value(dtype):
 *     if dtype.kind == 'f':             # <<<<<<<<<<<<<<
 *         return 1.0
 *     return float(numpy.iinfo(dtype).max)
 */
  }

  /* "celiagg/conversion.pxi":56
 *     if dtype.kind == 'f':
 *         return 1.0
 *     return float(numpy.iinfo(dtype).max)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_numpy); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_iinfo); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_dtype) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_dtype);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_max); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_AsDouble(__pyx_t_4); if (unlikely(__pyx_t_5 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(9, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":53
 * 
 * 
 * cdef float _get_max_value(dtype):             # <<<<<<<<<<<<<<
 *     if dtype.kind == 'f':
 *         return 1.0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("celiagg._celiagg._get_max_value", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":59
 * 
 * 
 * cdef gray2rgb(src, dst_order, dtype=None, alpha=False):             # <<<<<<<<<<<<<<
 *     """Convert gray pixels to RGB[A].
 *     """
 */

static PyObject *__pyx_f_7celiagg_8_celiagg_gray2rgb(PyObject *__pyx_v_src, PyObject *__pyx_v_dst_order, struct __pyx_opt_args_7celiagg_8_celiagg_gray2rgb *__pyx_optional_args) {
  PyObject *__pyx_v_dtype = ((PyObject *)Py_None);
  PyObject *__pyx_v_alpha = ((PyObject *)Py_False);
  PyObject *__pyx_v_last_dim = NULL;
  PyObject *__pyx_v_dst = NULL;
  PyObject *__pyx_v_max_from = NULL;
  PyObject *__pyx_v_max_to = NULL;
  PyObject *__pyx_v_scaled_src = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("gray2rgb", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_dtype = __pyx_optional_args->dtype;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_alpha = __pyx_optional_args->alpha;
      }
    }
  }

  /* "celiagg/conversion.pxi":62
 *     """Convert gray pixels to RGB[A].
 *     """
 *     last_dim = (4,) if alpha else (3,)             # <<<<<<<<<<<<<<
 *     dst = numpy.empty(src.shape + last_dim, dtype=dtype)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_alpha); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(9, 62, __pyx_L1_error)
  if (__pyx_t_2) {
    __Pyx_INCREF(__pyx_tuple__74);
    __pyx_t_1 = __pyx_tuple__74;
  } else {
    __Pyx_INCREF(__pyx_tuple__75);
    __pyx_t_1 = __pyx_tuple__75;
  }
  __pyx_v_last_dim = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":63
 *     """
 *     last_dim = (4,) if alpha else (3,)
 *     dst = numpy.empty(src.shape + last_dim, dtype=dtype)             # <<<<<<<<<<<<<<
 * 
 *     max_from = _get_max_value(src.dtype)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_v_last_dim); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_v_dtype) < 0) __PYX_ERR(9, 63, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_dst = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "celiagg/conversion.pxi":65
 *     dst = numpy.empty(src.shape + last_dim, dtype=dtype)
 * 
 *     max_from = _get_max_value(src.dtype)             # <<<<<<<<<<<<<<
 *     max_to = _get_max_value(dst.dtype)
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_f_7celiagg_8_celiagg__get_max_value(__pyx_t_5)); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_max_from = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":66
 * 
 *     max_from = _get_max_value(src.dtype)
 *     max_to = _get_max_value(dst.dtype)             # <<<<<<<<<<<<<<
 * 
 *     scaled_src = src[:] * (max_to / max_from)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_f_7celiagg_8_celiagg__get_max_value(__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_max_to = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "celiagg/conversion.pxi":68
 *     max_to = _get_max_value(dst.dtype)
 * 
 *     scaled_src = src[:] * (max_to / max_from)             # <<<<<<<<<<<<<<
 *     dst[..., dst_order[0]] = scaled_src
 *     dst[..., dst_order[1]] = scaled_src
 */
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_src, 0, 0, NULL, NULL, &__pyx_slice__56, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyNumber_Divide(__pyx_v_max_to, __pyx_v_max_from); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyNumber_Multiply(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_scaled_src = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":69
 * 
 *     scaled_src = src[:] * (max_to / max_from)
 *     dst[..., dst_order[0]] = scaled_src             # <<<<<<<<<<<<<<
 *     dst[..., dst_order[1]] = scaled_src
 *     dst[..., dst_order[2]] = scaled_src
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_dst_order, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_4, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(PyObject_SetItem(__pyx_v_dst, __pyx_t_4, __pyx_v_scaled_src) < 0)) __PYX_ERR(9, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":70
 *     scaled_src = src[:] * (max_to / max_from)
 *     dst[..., dst_order[0]] = scaled_src
 *     dst[..., dst_order[1]] = scaled_src             # <<<<<<<<<<<<<<
 *     dst[..., dst_order[2]] = scaled_src
 *     if alpha:
 */
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_dst_order, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_1, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (unlikely(PyObject_SetItem(__pyx_v_dst, __pyx_t_1, __pyx_v_scaled_src) < 0)) __PYX_ERR(9, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":71
 *     dst[..., dst_order[0]] = scaled_src
 *     dst[..., dst_order[1]] = scaled_src
 *     dst[..., dst_order[2]] = scaled_src             # <<<<<<<<<<<<<<
 *     if alpha:
 *         dst[..., dst_order[3]] = max_to
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_dst_order, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_4, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(PyObject_SetItem(__pyx_v_dst, __pyx_t_4, __pyx_v_scaled_src) < 0)) __PYX_ERR(9, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":72
 *     dst[..., dst_order[1]] = scaled_src
 *     dst[..., dst_order[2]] = scaled_src
 *     if alpha:             # <<<<<<<<<<<<<<
 *         dst[..., dst_order[3]] = max_to
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_alpha); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(9, 72, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "celiagg/conversion.pxi":73
 *     dst[..., dst_order[2]] = scaled_src
 *     if alpha:
 *         dst[..., dst_order[3]] = max_to             # <<<<<<<<<<<<<<
 * 
 *     return dst
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_dst_order, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(Py_Ellipsis);
    __Pyx_GIVEREF(Py_Ellipsis);
    PyTuple_SET_ITEM(__pyx_t_1, 0, Py_Ellipsis);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
    __pyx_t_4 = 0;
    if (unlikely(PyObject_SetItem(__pyx_v_dst, __pyx_t_1, __pyx_v_max_to) < 0)) __PYX_ERR(9, 73, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "celiagg/conversion.pxi":72
 *     dst[..., dst_order[1]] = scaled_src
 *     dst[..., dst_order[2]] = scaled_src
 *     if alpha:             # <<<<<<<<<<<<<<
 *         dst[..., dst_order[3]] = max_to
 * 
 */
  }

  /* "celiagg/conversion.pxi":75
 *         dst[..., dst_order[3]] = max_to
 * 
 *     return dst             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dst);
  __pyx_r = __pyx_v_dst;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":59
 * 
 * 
 * cdef gray2rgb(src, dst_order, dtype=None, alpha=False):             # <<<<<<<<<<<<<<
 *     """Convert gray pixels to RGB[A].
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg.gray2rgb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_last_dim);
  __Pyx_XDECREF(__pyx_v_dst);
  __Pyx_XDECREF(__pyx_v_max_from);
  __Pyx_XDECREF(__pyx_v_max_to);
  __Pyx_XDECREF(__pyx_v_scaled_src);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":78
 * 
 * 
 * cdef rgb2rgb(src, src_order, dst_order, dtype=None, alpha=False):             # <<<<<<<<<<<<<<
 *     """Convert RGB[A] pixels to RGB[A] pixels.
 *     """
 */

static PyObject *__pyx_f_7celiagg_8_celiagg_rgb2rgb(PyObject *__pyx_v_src, PyObject *__pyx_v_src_order, PyObject *__pyx_v_dst_order, struct __pyx_opt_args_7celiagg_8_celiagg_rgb2rgb *__pyx_optional_args) {
  PyObject *__pyx_v_dtype = ((PyObject *)Py_None);
  PyObject *__pyx_v_alpha = ((PyObject *)Py_False);
  PyObject *__pyx_v_last_dim = NULL;
  PyObject *__pyx_v_dst = NULL;
  PyObject *__pyx_v_max_from = NULL;
  PyObject *__pyx_v_max_to = NULL;
  PyObject *__pyx_v_value_scale = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("rgb2rgb", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_dtype = __pyx_optional_args->dtype;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_alpha = __pyx_optional_args->alpha;
      }
    }
  }

  /* "celiagg/conversion.pxi":81
 *     """Convert RGB[A] pixels to RGB[A] pixels.
 *     """
 *     last_dim = (4,) if alpha else (3,)             # <<<<<<<<<<<<<<
 *     dst = numpy.empty(src.shape[:2] + last_dim, dtype=dtype)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_alpha); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(9, 81, __pyx_L1_error)
  if (__pyx_t_2) {
    __Pyx_INCREF(__pyx_tuple__74);
    __pyx_t_1 = __pyx_tuple__74;
  } else {
    __Pyx_INCREF(__pyx_tuple__75);
    __pyx_t_1 = __pyx_tuple__75;
  }
  __pyx_v_last_dim = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":82
 *     """
 *     last_dim = (4,) if alpha else (3,)
 *     dst = numpy.empty(src.shape[:2] + last_dim, dtype=dtype)             # <<<<<<<<<<<<<<
 * 
 *     max_from = _get_max_value(src.dtype)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 2, NULL, NULL, &__pyx_slice__76, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_4, __pyx_v_last_dim); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_v_dtype) < 0) __PYX_ERR(9, 82, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dst = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "celiagg/conversion.pxi":84
 *     dst = numpy.empty(src.shape[:2] + last_dim, dtype=dtype)
 * 
 *     max_from = _get_max_value(src.dtype)             # <<<<<<<<<<<<<<
 *     max_to = _get_max_value(dst.dtype)
 *     value_scale = (max_to / max_from)
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_dtype); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_7celiagg_8_celiagg__get_max_value(__pyx_t_5)); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_max_from = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":85
 * 
 *     max_from = _get_max_value(src.dtype)
 *     max_to = _get_max_value(dst.dtype)             # <<<<<<<<<<<<<<
 *     value_scale = (max_to / max_from)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_f_7celiagg_8_celiagg__get_max_value(__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_max_to = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "celiagg/conversion.pxi":86
 *     max_from = _get_max_value(src.dtype)
 *     max_to = _get_max_value(dst.dtype)
 *     value_scale = (max_to / max_from)             # <<<<<<<<<<<<<<
 * 
 *     dst[..., dst_order[0]] = src[..., src_order[0]] * value_scale
 */
  __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_v_max_to, __pyx_v_max_from); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_value_scale = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "celiagg/conversion.pxi":88
 *     value_scale = (max_to / max_from)
 * 
 *     dst[..., dst_order[0]] = src[..., src_order[0]] * value_scale             # <<<<<<<<<<<<<<
 *     dst[..., dst_order[1]] = src[..., src_order[1]] * value_scale
 *     dst[..., dst_order[2]] = src[..., src_order[2]] * value_scale
 */
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_src_order, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_1, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_t_5, __pyx_v_value_scale); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_dst_order, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_4, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  if (unlikely(PyObject_SetItem(__pyx_v_dst, __pyx_t_4, __pyx_t_1) < 0)) __PYX_ERR(9, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":89
 * 
 *     dst[..., dst_order[0]] = src[..., src_order[0]] * value_scale
 *     dst[..., dst_order[1]] = src[..., src_order[1]] * value_scale             # <<<<<<<<<<<<<<
 *     dst[..., dst_order[2]] = src[..., src_order[2]] * value_scale
 *     if alpha:
 */
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_order, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_4, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_v_value_scale); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_dst_order, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_5, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(PyObject_SetItem(__pyx_v_dst, __pyx_t_5, __pyx_t_4) < 0)) __PYX_ERR(9, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":90
 *     dst[..., dst_order[0]] = src[..., src_order[0]] * value_scale
 *     dst[..., dst_order[1]] = src[..., src_order[1]] * value_scale
 *     dst[..., dst_order[2]] = src[..., src_order[2]] * value_scale             # <<<<<<<<<<<<<<
 *     if alpha:
 *         if src.shape[2] == 4:
 */
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_src_order, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_5, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyNumber_Multiply(__pyx_t_4, __pyx_v_value_scale); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_dst_order, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_1, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  if (unlikely(PyObject_SetItem(__pyx_v_dst, __pyx_t_1, __pyx_t_5) < 0)) __PYX_ERR(9, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "celiagg/conversion.pxi":91
 *     dst[..., dst_order[1]] = src[..., src_order[1]] * value_scale
 *     dst[..., dst_order[2]] = src[..., src_order[2]] * value_scale
 *     if alpha:             # <<<<<<<<<<<<<<
 *         if src.shape[2] == 4:
 *             dst[..., dst_order[3]] = src[..., src_order[3]] * value_scale
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_alpha); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(9, 91, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "celiagg/conversion.pxi":92
 *     dst[..., dst_order[2]] = src[..., src_order[2]] * value_scale
 *     if alpha:
 *         if src.shape[2] == 4:             # <<<<<<<<<<<<<<
 *             dst[..., dst_order[3]] = src[..., src_order[3]] * value_scale
 *         else:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_5, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(9, 92, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_2) {

      /* "celiagg/conversion.pxi":93
 *     if alpha:
 *         if src.shape[2] == 4:
 *             dst[..., dst_order[3]] = src[..., src_order[3]] * value_scale             # <<<<<<<<<<<<<<
 *         else:
 *             dst[..., dst_order[3]] = max_to
 */
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_src_order, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(Py_Ellipsis);
      __Pyx_GIVEREF(Py_Ellipsis);
      PyTuple_SET_ITEM(__pyx_t_1, 0, Py_Ellipsis);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Multiply(__pyx_t_5, __pyx_v_value_scale); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_dst_order, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(Py_Ellipsis);
      __Pyx_GIVEREF(Py_Ellipsis);
      PyTuple_SET_ITEM(__pyx_t_4, 0, Py_Ellipsis);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
      __pyx_t_5 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_dst, __pyx_t_4, __pyx_t_1) < 0)) __PYX_ERR(9, 93, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "celiagg/conversion.pxi":92
 *     dst[..., dst_order[2]] = src[..., src_order[2]] * value_scale
 *     if alpha:
 *         if src.shape[2] == 4:             # <<<<<<<<<<<<<<
 *             dst[..., dst_order[3]] = src[..., src_order[3]] * value_scale
 *         else:
 */
      goto __pyx_L4;
    }

    /* "celiagg/conversion.pxi":95
 *             dst[..., dst_order[3]] = src[..., src_order[3]] * value_scale
 *         else:
 *             dst[..., dst_order[3]] = max_to             # <<<<<<<<<<<<<<
 * 
 *     return dst
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_dst_order, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(Py_Ellipsis);
      __Pyx_GIVEREF(Py_Ellipsis);
      PyTuple_SET_ITEM(__pyx_t_4, 0, Py_Ellipsis);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
      __pyx_t_1 = 0;
      if (unlikely(PyObject_SetItem(__pyx_v_dst, __pyx_t_4, __pyx_v_max_to) < 0)) __PYX_ERR(9, 95, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_L4:;

    /* "celiagg/conversion.pxi":91
 *     dst[..., dst_order[1]] = src[..., src_order[1]] * value_scale
 *     dst[..., dst_order[2]] = src[..., src_order[2]] * value_scale
 *     if alpha:             # <<<<<<<<<<<<<<
 *         if src.shape[2] == 4:
 *             dst[..., dst_order[3]] = src[..., src_order[3]] * value_scale
 */
  }

  /* "celiagg/conversion.pxi":97
 *             dst[..., dst_order[3]] = max_to
 * 
 *     return dst             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dst);
  __pyx_r = __pyx_v_dst;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":78
 * 
 * 
 * cdef rgb2rgb(src, src_order, dst_order, dtype=None, alpha=False):             # <<<<<<<<<<<<<<
 *     """Convert RGB[A] pixels to RGB[A] pixels.
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("celiagg._celiagg.rgb2rgb", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_last_dim);
  __Pyx_XDECREF(__pyx_v_dst);
  __Pyx_XDECREF(__pyx_v_max_from);
  __Pyx_XDECREF(__pyx_v_max_to);
  __Pyx_XDECREF(__pyx_v_value_scale);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":100
 * 
 * 
 * cdef gray2gray(src, dtype=None):             # <<<<<<<<<<<<<<
 *     """Convert gray pixels to gray pixels.
 *     """
 */

static PyObject *__pyx_f_7celiagg_8_celiagg_gray2gray(PyObject *__pyx_v_src, struct __pyx_opt_args_7celiagg_8_celiagg_gray2gray *__pyx_optional_args) {
  PyObject *__pyx_v_dtype = ((PyObject *)Py_None);
  PyObject *__pyx_v_dst = NULL;
  PyObject *__pyx_v_max_from = NULL;
  PyObject *__pyx_v_max_to = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("gray2gray", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_dtype = __pyx_optional_args->dtype;
    }
  }

  /* "celiagg/conversion.pxi":103
 *     """Convert gray pixels to gray pixels.
 *     """
 *     dst = numpy.empty(src.shape, dtype=dtype)             # <<<<<<<<<<<<<<
 *     max_from = _get_max_value(src.dtype)
 *     max_to = _get_max_value(dst.dtype)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_v_dtype) < 0) __PYX_ERR(9, 103, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dst = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":104
 *     """
 *     dst = numpy.empty(src.shape, dtype=dtype)
 *     max_from = _get_max_value(src.dtype)             # <<<<<<<<<<<<<<
 *     max_to = _get_max_value(dst.dtype)
 *     dst[:] = src[:] * (max_to / max_from)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_7celiagg_8_celiagg__get_max_value(__pyx_t_4)); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_max_from = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":105
 *     dst = numpy.empty(src.shape, dtype=dtype)
 *     max_from = _get_max_value(src.dtype)
 *     max_to = _get_max_value(dst.dtype)             # <<<<<<<<<<<<<<
 *     dst[:] = src[:] * (max_to / max_from)
 *     return dst
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_f_7celiagg_8_celiagg__get_max_value(__pyx_t_1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_max_to = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":106
 *     max_from = _get_max_value(src.dtype)
 *     max_to = _get_max_value(dst.dtype)
 *     dst[:] = src[:] * (max_to / max_from)             # <<<<<<<<<<<<<<
 *     return dst
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_src, 0, 0, NULL, NULL, &__pyx_slice__56, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_max_to, __pyx_v_max_from); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetSlice(__pyx_v_dst, __pyx_t_3, 0, 0, NULL, NULL, &__pyx_slice__56, 0, 0, 1) < 0) __PYX_ERR(9, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":107
 *     max_to = _get_max_value(dst.dtype)
 *     dst[:] = src[:] * (max_to / max_from)
 *     return dst             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dst);
  __pyx_r = __pyx_v_dst;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":100
 * 
 * 
 * cdef gray2gray(src, dtype=None):             # <<<<<<<<<<<<<<
 *     """Convert gray pixels to gray pixels.
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.gray2gray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dst);
  __Pyx_XDECREF(__pyx_v_max_from);
  __Pyx_XDECREF(__pyx_v_max_to);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":110
 * 
 * 
 * cdef rgb2gray(src, src_order, dtype=None):             # <<<<<<<<<<<<<<
 *     """Convert RGB[A] pixels to gray.
 *     """
 */

static PyObject *__pyx_f_7celiagg_8_celiagg_rgb2gray(PyObject *__pyx_v_src, PyObject *__pyx_v_src_order, struct __pyx_opt_args_7celiagg_8_celiagg_rgb2gray *__pyx_optional_args) {
  PyObject *__pyx_v_dtype = ((PyObject *)Py_None);
  PyObject *__pyx_v_dst = NULL;
  PyObject *__pyx_v_max_from = NULL;
  PyObject *__pyx_v_max_to = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("rgb2gray", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_dtype = __pyx_optional_args->dtype;
    }
  }

  /* "celiagg/conversion.pxi":113
 *     """Convert RGB[A] pixels to gray.
 *     """
 *     dst = numpy.empty(src.shape[:2], dtype=dtype)             # <<<<<<<<<<<<<<
 * 
 *     max_from = _get_max_value(dst.dtype)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_numpy); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_t_1, 0, 2, NULL, NULL, &__pyx_slice__76, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_v_dtype) < 0) __PYX_ERR(9, 113, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_dst = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":115
 *     dst = numpy.empty(src.shape[:2], dtype=dtype)
 * 
 *     max_from = _get_max_value(dst.dtype)             # <<<<<<<<<<<<<<
 *     max_to = _get_max_value(src.dtype)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_dtype); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_f_7celiagg_8_celiagg__get_max_value(__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_max_from = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":116
 * 
 *     max_from = _get_max_value(dst.dtype)
 *     max_to = _get_max_value(src.dtype)             # <<<<<<<<<<<<<<
 * 
 *     dst[:] = (src[..., src_order[0]] + src[..., src_order[1]] +
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_f_7celiagg_8_celiagg__get_max_value(__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_max_to = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":118
 *     max_to = _get_max_value(src.dtype)
 * 
 *     dst[:] = (src[..., src_order[0]] + src[..., src_order[1]] +             # <<<<<<<<<<<<<<
 *               src[..., src_order[2]]) * (max_to / (3.0 * max_from))
 *     return dst
 */
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_src_order, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_src_order, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_1, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":119
 * 
 *     dst[:] = (src[..., src_order[0]] + src[..., src_order[1]] +
 *               src[..., src_order[2]]) * (max_to / (3.0 * max_from))             # <<<<<<<<<<<<<<
 *     return dst
 * 
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_src_order, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(Py_Ellipsis);
  __Pyx_GIVEREF(Py_Ellipsis);
  PyTuple_SET_ITEM(__pyx_t_4, 0, Py_Ellipsis);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_src, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "celiagg/conversion.pxi":118
 *     max_to = _get_max_value(src.dtype)
 * 
 *     dst[:] = (src[..., src_order[0]] + src[..., src_order[1]] +             # <<<<<<<<<<<<<<
 *               src[..., src_order[2]]) * (max_to / (3.0 * max_from))
 *     return dst
 */
  __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":119
 * 
 *     dst[:] = (src[..., src_order[0]] + src[..., src_order[1]] +
 *               src[..., src_order[2]]) * (max_to / (3.0 * max_from))             # <<<<<<<<<<<<<<
 *     return dst
 * 
 */
  __pyx_t_3 = PyNumber_Multiply(__pyx_float_3_0, __pyx_v_max_from); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_max_to, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/conversion.pxi":118
 *     max_to = _get_max_value(src.dtype)
 * 
 *     dst[:] = (src[..., src_order[0]] + src[..., src_order[1]] +             # <<<<<<<<<<<<<<
 *               src[..., src_order[2]]) * (max_to / (3.0 * max_from))
 *     return dst
 */
  if (__Pyx_PyObject_SetSlice(__pyx_v_dst, __pyx_t_3, 0, 0, NULL, NULL, &__pyx_slice__56, 0, 0, 1) < 0) __PYX_ERR(9, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":120
 *     dst[:] = (src[..., src_order[0]] + src[..., src_order[1]] +
 *               src[..., src_order[2]]) * (max_to / (3.0 * max_from))
 *     return dst             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_dst);
  __pyx_r = __pyx_v_dst;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":110
 * 
 * 
 * cdef rgb2gray(src, src_order, dtype=None):             # <<<<<<<<<<<<<<
 *     """Convert RGB[A] pixels to gray.
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("celiagg._celiagg.rgb2gray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dst);
  __Pyx_XDECREF(__pyx_v_max_from);
  __Pyx_XDECREF(__pyx_v_max_to);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "celiagg/conversion.pxi":123
 * 
 * 
 * def convert_image(src, to_format, bottom_up=False):             # <<<<<<<<<<<<<<
 *     """convert_image(src, to_format, bottom_up=False)
 *     Create a new image with a desired pixel format and orientation.
 */

/* Python wrapper */
static PyObject *__pyx_pw_7celiagg_8_celiagg_1convert_image(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_7celiagg_8_celiagg_convert_image[] = "convert_image(src, to_format, bottom_up=False)\n    Create a new image with a desired pixel format and orientation.\n\n    :param image: An Image instance\n    :param to_format: A PixelFormat describing the desired output format\n    :param bottom_up: If True, the image data is flipped in the y axis\n    ";
static PyMethodDef __pyx_mdef_7celiagg_8_celiagg_1convert_image = {"convert_image", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_1convert_image, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_convert_image};
static PyObject *__pyx_pw_7celiagg_8_celiagg_1convert_image(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_src = 0;
  PyObject *__pyx_v_to_format = 0;
  PyObject *__pyx_v_bottom_up = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert_image (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src,&__pyx_n_s_to_format,&__pyx_n_s_bottom_up,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_src)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_to_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convert_image", 0, 2, 3, 1); __PYX_ERR(9, 123, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bottom_up);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convert_image") < 0)) __PYX_ERR(9, 123, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_src = values[0];
    __pyx_v_to_format = values[1];
    __pyx_v_bottom_up = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert_image", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(9, 123, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("celiagg._celiagg.convert_image", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7celiagg_8_celiagg_convert_image(__pyx_self, __pyx_v_src, __pyx_v_to_format, __pyx_v_bottom_up);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7celiagg_8_celiagg_convert_image(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_src, PyObject *__pyx_v_to_format, PyObject *__pyx_v_bottom_up) {
  PyObject *__pyx_v_from_format = NULL;
  PyObject *__pyx_v_dst_dtype = NULL;
  bool __pyx_v_dst_alpha;
  bool __pyx_v_src_gray;
  bool __pyx_v_dst_gray;
  CYTHON_UNUSED bool __pyx_v_src_rgb;
  CYTHON_UNUSED bool __pyx_v_dst_rgb;
  PyObject *__pyx_v_src_order = NULL;
  PyObject *__pyx_v_dst_order = NULL;
  PyObject *__pyx_v_dst = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  enum __pyx_t_7celiagg_8_celiagg_PixelFormat __pyx_t_7;
  struct __pyx_opt_args_7celiagg_8_celiagg_gray2gray __pyx_t_8;
  struct __pyx_opt_args_7celiagg_8_celiagg_gray2rgb __pyx_t_9;
  struct __pyx_opt_args_7celiagg_8_celiagg_rgb2gray __pyx_t_10;
  struct __pyx_opt_args_7celiagg_8_celiagg_rgb2rgb __pyx_t_11;
  __Pyx_RefNannySetupContext("convert_image", 0);
  __Pyx_INCREF(__pyx_v_to_format);

  /* "celiagg/conversion.pxi":131
 *     :param bottom_up: If True, the image data is flipped in the y axis
 *     """
 *     if not isinstance(src, Image):             # <<<<<<<<<<<<<<
 *         raise TypeError("src must an Image instance")
 *     if not isinstance(to_format, (int, PixelFormat)):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_src, __pyx_ptype_7celiagg_8_celiagg_Image); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "celiagg/conversion.pxi":132
 *     """
 *     if not isinstance(src, Image):
 *         raise TypeError("src must an Image instance")             # <<<<<<<<<<<<<<
 *     if not isinstance(to_format, (int, PixelFormat)):
 *         raise TypeError("to_format must be a PixelFormat value")
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__77, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(9, 132, __pyx_L1_error)

    /* "celiagg/conversion.pxi":131
 *     :param bottom_up: If True, the image data is flipped in the y axis
 *     """
 *     if not isinstance(src, Image):             # <<<<<<<<<<<<<<
 *         raise TypeError("src must an Image instance")
 *     if not isinstance(to_format, (int, PixelFormat)):
 */
  }

  /* "celiagg/conversion.pxi":133
 *     if not isinstance(src, Image):
 *         raise TypeError("src must an Image instance")
 *     if not isinstance(to_format, (int, PixelFormat)):             # <<<<<<<<<<<<<<
 *         raise TypeError("to_format must be a PixelFormat value")
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyInt_Check(__pyx_v_to_format); 
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = PyObject_IsInstance(__pyx_v_to_format, __pyx_t_3); 
  __pyx_t_1 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_1;
  __pyx_L5_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "celiagg/conversion.pxi":134
 *         raise TypeError("src must an Image instance")
 *     if not isinstance(to_format, (int, PixelFormat)):
 *         raise TypeError("to_format must be a PixelFormat value")             # <<<<<<<<<<<<<<
 * 
 *     to_format = PixelFormat(to_format)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__78, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(9, 134, __pyx_L1_error)

    /* "celiagg/conversion.pxi":133
 *     if not isinstance(src, Image):
 *         raise TypeError("src must an Image instance")
 *     if not isinstance(to_format, (int, PixelFormat)):             # <<<<<<<<<<<<<<
 *         raise TypeError("to_format must be a PixelFormat value")
 * 
 */
  }

  /* "celiagg/conversion.pxi":136
 *         raise TypeError("to_format must be a PixelFormat value")
 * 
 *     to_format = PixelFormat(to_format)             # <<<<<<<<<<<<<<
 *     from_format = src.format
 *     dst_dtype = _get_format_dtype(to_format)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_v_to_format) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_to_format);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_to_format, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":137
 * 
 *     to_format = PixelFormat(to_format)
 *     from_format = src.format             # <<<<<<<<<<<<<<
 *     dst_dtype = _get_format_dtype(to_format)
 *     cdef bool dst_alpha = _has_alpha(to_format)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_from_format = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":138
 *     to_format = PixelFormat(to_format)
 *     from_format = src.format
 *     dst_dtype = _get_format_dtype(to_format)             # <<<<<<<<<<<<<<
 *     cdef bool dst_alpha = _has_alpha(to_format)
 *     cdef bool src_gray = _is_gray(from_format)
 */
  __pyx_t_7 = ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)__Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_v_to_format)); if (unlikely(PyErr_Occurred())) __PYX_ERR(9, 138, __pyx_L1_error)
  __pyx_t_3 = __pyx_f_7celiagg_8_celiagg__get_format_dtype(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_dst_dtype = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":139
 *     from_format = src.format
 *     dst_dtype = _get_format_dtype(to_format)
 *     cdef bool dst_alpha = _has_alpha(to_format)             # <<<<<<<<<<<<<<
 *     cdef bool src_gray = _is_gray(from_format)
 *     cdef bool dst_gray = _is_gray(to_format)
 */
  __pyx_v_dst_alpha = __pyx_f_7celiagg_8_celiagg__has_alpha(__pyx_v_to_format);

  /* "celiagg/conversion.pxi":140
 *     dst_dtype = _get_format_dtype(to_format)
 *     cdef bool dst_alpha = _has_alpha(to_format)
 *     cdef bool src_gray = _is_gray(from_format)             # <<<<<<<<<<<<<<
 *     cdef bool dst_gray = _is_gray(to_format)
 *     cdef bool src_rgb = _is_rgb(from_format)
 */
  __pyx_v_src_gray = __pyx_f_7celiagg_8_celiagg__is_gray(__pyx_v_from_format);

  /* "celiagg/conversion.pxi":141
 *     cdef bool dst_alpha = _has_alpha(to_format)
 *     cdef bool src_gray = _is_gray(from_format)
 *     cdef bool dst_gray = _is_gray(to_format)             # <<<<<<<<<<<<<<
 *     cdef bool src_rgb = _is_rgb(from_format)
 *     cdef bool dst_rgb = _is_rgb(to_format)
 */
  __pyx_v_dst_gray = __pyx_f_7celiagg_8_celiagg__is_gray(__pyx_v_to_format);

  /* "celiagg/conversion.pxi":142
 *     cdef bool src_gray = _is_gray(from_format)
 *     cdef bool dst_gray = _is_gray(to_format)
 *     cdef bool src_rgb = _is_rgb(from_format)             # <<<<<<<<<<<<<<
 *     cdef bool dst_rgb = _is_rgb(to_format)
 * 
 */
  __pyx_v_src_rgb = __pyx_f_7celiagg_8_celiagg__is_rgb(__pyx_v_from_format);

  /* "celiagg/conversion.pxi":143
 *     cdef bool dst_gray = _is_gray(to_format)
 *     cdef bool src_rgb = _is_rgb(from_format)
 *     cdef bool dst_rgb = _is_rgb(to_format)             # <<<<<<<<<<<<<<
 * 
 *     src_order = _get_component_indices(from_format)
 */
  __pyx_v_dst_rgb = __pyx_f_7celiagg_8_celiagg__is_rgb(__pyx_v_to_format);

  /* "celiagg/conversion.pxi":145
 *     cdef bool dst_rgb = _is_rgb(to_format)
 * 
 *     src_order = _get_component_indices(from_format)             # <<<<<<<<<<<<<<
 *     dst_order = _get_component_indices(to_format)
 * 
 */
  __pyx_t_3 = __pyx_f_7celiagg_8_celiagg__get_component_indices(__pyx_v_from_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_src_order = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":146
 * 
 *     src_order = _get_component_indices(from_format)
 *     dst_order = _get_component_indices(to_format)             # <<<<<<<<<<<<<<
 * 
 *     if src_gray:
 */
  __pyx_t_3 = __pyx_f_7celiagg_8_celiagg__get_component_indices(__pyx_v_to_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_dst_order = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "celiagg/conversion.pxi":148
 *     dst_order = _get_component_indices(to_format)
 * 
 *     if src_gray:             # <<<<<<<<<<<<<<
 *         if dst_gray:
 *             dst = gray2gray(src.pixels, dtype=dst_dtype)
 */
  __pyx_t_1 = (__pyx_v_src_gray != 0);
  if (__pyx_t_1) {

    /* "celiagg/conversion.pxi":149
 * 
 *     if src_gray:
 *         if dst_gray:             # <<<<<<<<<<<<<<
 *             dst = gray2gray(src.pixels, dtype=dst_dtype)
 *         else:
 */
    __pyx_t_1 = (__pyx_v_dst_gray != 0);
    if (__pyx_t_1) {

      /* "celiagg/conversion.pxi":150
 *     if src_gray:
 *         if dst_gray:
 *             dst = gray2gray(src.pixels, dtype=dst_dtype)             # <<<<<<<<<<<<<<
 *         else:
 *             dst = gray2rgb(src.pixels, dst_order, dtype=dst_dtype,
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_pixels); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8.__pyx_n = 1;
      __pyx_t_8.dtype = __pyx_v_dst_dtype;
      __pyx_t_5 = __pyx_f_7celiagg_8_celiagg_gray2gray(__pyx_t_3, &__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 150, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_dst = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "celiagg/conversion.pxi":149
 * 
 *     if src_gray:
 *         if dst_gray:             # <<<<<<<<<<<<<<
 *             dst = gray2gray(src.pixels, dtype=dst_dtype)
 *         else:
 */
      goto __pyx_L8;
    }

    /* "celiagg/conversion.pxi":152
 *             dst = gray2gray(src.pixels, dtype=dst_dtype)
 *         else:
 *             dst = gray2rgb(src.pixels, dst_order, dtype=dst_dtype,             # <<<<<<<<<<<<<<
 *                            alpha=dst_alpha)
 *     else:
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_pixels); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "celiagg/conversion.pxi":153
 *         else:
 *             dst = gray2rgb(src.pixels, dst_order, dtype=dst_dtype,
 *                            alpha=dst_alpha)             # <<<<<<<<<<<<<<
 *     else:
 *         if dst_gray:
 */
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_dst_alpha); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "celiagg/conversion.pxi":152
 *             dst = gray2gray(src.pixels, dtype=dst_dtype)
 *         else:
 *             dst = gray2rgb(src.pixels, dst_order, dtype=dst_dtype,             # <<<<<<<<<<<<<<
 *                            alpha=dst_alpha)
 *     else:
 */
      __pyx_t_9.__pyx_n = 2;
      __pyx_t_9.dtype = __pyx_v_dst_dtype;
      __pyx_t_9.alpha = __pyx_t_3;
      __pyx_t_6 = __pyx_f_7celiagg_8_celiagg_gray2rgb(__pyx_t_5, __pyx_v_dst_order, &__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(9, 152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_dst = __pyx_t_6;
      __pyx_t_6 = 0;
    }
    __pyx_L8:;

    /* "celiagg/conversion.pxi":148
 *     dst_order = _get_component_indices(to_format)
 * 
 *     if src_gray:             # <<<<<<<<<<<<<<
 *         if dst_gray:
 *             dst = gray2gray(src.pixels, dtype=dst_dtype)
 */
    goto __pyx_L7;
  }

  /* "celiagg/conversion.pxi":155
 *                            alpha=dst_alpha)
 *     else:
 *         if dst_gray:             # <<<<<<<<<<<<<<
 *             dst = rgb2gray(src.pixels, src_order, dtype=dst_dtype)
 *         else:
 */
  /*else*/ {
    __pyx_t_1 = (__pyx_v_dst_gray != 0);
    if (__pyx_t_1) {

      /* "celiagg/conversion.pxi":156
 *     else:
 *         if dst_gray:
 *             dst = rgb2gray(src.pixels, src_order, dtype=dst_dtype)             # <<<<<<<<<<<<<<
 *         else:
 *             dst = rgb2rgb(src.pixels, src_order, dst_order, dtype=dst_dtype,
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_pixels); if (unlikely(!__pyx_t_6)) __PYX_ERR(9, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10.__pyx_n = 1;
      __pyx_t_10.dtype = __pyx_v_dst_dtype;
      __pyx_t_3 = __pyx_f_7celiagg_8_celiagg_rgb2gray(__pyx_t_6, __pyx_v_src_order, &__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_dst = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "celiagg/conversion.pxi":155
 *                            alpha=dst_alpha)
 *     else:
 *         if dst_gray:             # <<<<<<<<<<<<<<
 *             dst = rgb2gray(src.pixels, src_order, dtype=dst_dtype)
 *         else:
 */
      goto __pyx_L9;
    }

    /* "celiagg/conversion.pxi":158
 *             dst = rgb2gray(src.pixels, src_order, dtype=dst_dtype)
 *         else:
 *             dst = rgb2rgb(src.pixels, src_order, dst_order, dtype=dst_dtype,             # <<<<<<<<<<<<<<
 *                           alpha=dst_alpha)
 * 
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_pixels); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);

      /* "celiagg/conversion.pxi":159
 *         else:
 *             dst = rgb2rgb(src.pixels, src_order, dst_order, dtype=dst_dtype,
 *                           alpha=dst_alpha)             # <<<<<<<<<<<<<<
 * 
 *     return Image(dst, to_format, bottom_up=bottom_up)
 */
      __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_dst_alpha); if (unlikely(!__pyx_t_6)) __PYX_ERR(9, 159, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "celiagg/conversion.pxi":158
 *             dst = rgb2gray(src.pixels, src_order, dtype=dst_dtype)
 *         else:
 *             dst = rgb2rgb(src.pixels, src_order, dst_order, dtype=dst_dtype,             # <<<<<<<<<<<<<<
 *                           alpha=dst_alpha)
 * 
 */
      __pyx_t_11.__pyx_n = 2;
      __pyx_t_11.dtype = __pyx_v_dst_dtype;
      __pyx_t_11.alpha = __pyx_t_6;
      __pyx_t_5 = __pyx_f_7celiagg_8_celiagg_rgb2rgb(__pyx_t_3, __pyx_v_src_order, __pyx_v_dst_order, &__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 158, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_dst = __pyx_t_5;
      __pyx_t_5 = 0;
    }
    __pyx_L9:;
  }
  __pyx_L7:;

  /* "celiagg/conversion.pxi":161
 *                           alpha=dst_alpha)
 * 
 *     return Image(dst, to_format, bottom_up=bottom_up)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(9, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_dst);
  __Pyx_GIVEREF(__pyx_v_dst);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_dst);
  __Pyx_INCREF(__pyx_v_to_format);
  __Pyx_GIVEREF(__pyx_v_to_format);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_to_format);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(9, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_bottom_up, __pyx_v_bottom_up) < 0) __PYX_ERR(9, 161, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_7celiagg_8_celiagg_Image), __pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(9, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "celiagg/conversion.pxi":123
 * 
 * 
 * def convert_image(src, to_format, bottom_up=False):             # <<<<<<<<<<<<<<
 *     """convert_image(src, to_format, bottom_up=False)
 *     Create a new image with a desired pixel format and orientation.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("celiagg._celiagg.convert_image", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_from_format);
  __Pyx_XDECREF(__pyx_v_dst_dtype);
  __Pyx_XDECREF(__pyx_v_src_order);
  __Pyx_XDECREF(__pyx_v_dst_order);
  __Pyx_XDECREF(__pyx_v_dst);
  __Pyx_XDECREF(__pyx_v_to_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__79, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(10, 272, __pyx_L1_error)

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__80, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(10, 276, __pyx_L1_error)

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__81, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(10, 306, __pyx_L1_error)

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(10, 325, __pyx_L1_error)
      break;
    }

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(10, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":822
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":825
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":824
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":828
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":831
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":834
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":838
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":847
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":848
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(10, 851, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(10, 851, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":852
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(10, 852, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 852, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(10, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":853
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(10, 853, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(10, 853, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(10, 853, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(10, 855, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__82, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 856, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(10, 856, __pyx_L1_error)

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":855
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":859
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":860
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__81, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 860, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(10, 860, __pyx_L1_error)

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":858
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":870
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 870, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 870, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":873
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":875
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":878
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 878, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(10, 880, __pyx_L1_error)

        /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":883
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":899
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 899, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(10, 899, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(10, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(10, 901, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(10, 901, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":902
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":877
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(10, 906, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":851
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":907
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1024
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1030
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(10, 1036, __pyx_L3_error)

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(10, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__84, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 1038, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(10, 1038, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(10, 1042, __pyx_L3_error)

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(10, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__85, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 1044, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(10, 1044, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(10, 1048, __pyx_L3_error)

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(10, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1050
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__85, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(10, 1050, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(10, 1050, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_double")
 * cdef object __pyx_convert_vector_to_py_double(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

static PyObject *__pyx_convert_vector_to_py_double(const std::vector<double>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_double", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_double")
 * cdef object __pyx_convert_vector_to_py_double(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = PyFloat_FromDouble((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_double")
 * cdef object __pyx_convert_vector_to_py_double(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":16
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

/* Python wrapper */
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__(PyObject *__pyx_v_cls, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_parents = 0;
  PyObject *__pyx_v_dct = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_parents,&__pyx_n_s_dct,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parents)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(4, 16, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dct)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(4, 16, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(4, 16, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_name = values[0];
    __pyx_v_parents = values[1];
    __pyx_v_dct = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8EnumBase_14__Pyx_EnumMeta___init__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name, PyObject *__pyx_v_parents, PyObject *__pyx_v_dct) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "EnumBase":17
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)             # <<<<<<<<<<<<<<
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyType_Type)), __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, ((PyObject *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 17, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, ((PyObject *)__pyx_v_cls), __pyx_v_name, __pyx_v_parents, __pyx_v_dct};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 17, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_cls));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_cls));
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, ((PyObject *)__pyx_v_cls));
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_parents);
    __Pyx_GIVEREF(__pyx_v_parents);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_parents);
    __Pyx_INCREF(__pyx_v_dct);
    __Pyx_GIVEREF(__pyx_v_dct);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_dct);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 17, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":18
 *     def __init__(cls, name, parents, dct):
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()             # <<<<<<<<<<<<<<
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 */
  __Pyx_INCREF(__Pyx_OrderedDict);
  __pyx_t_2 = __Pyx_OrderedDict; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members, __pyx_t_1) < 0) __PYX_ERR(4, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":16
 * @cython.internal
 * cdef class __Pyx_EnumMeta(type):
 *     def __init__(cls, name, parents, dct):             # <<<<<<<<<<<<<<
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":19
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__(PyObject *__pyx_v_cls) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_2__iter__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "EnumBase":20
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())             # <<<<<<<<<<<<<<
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "EnumBase":19
 *         type.__init__(cls, name, parents, dct)
 *         cls.__members__ = __Pyx_OrderedDict()
 *     def __iter__(cls):             # <<<<<<<<<<<<<<
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":21
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__(PyObject *__pyx_v_cls, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_cls), ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_4__getitem__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_cls, PyObject *__pyx_v_name) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "EnumBase":22
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):
 *         return cls.__members__[name]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_cls), __pyx_n_s_members); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "EnumBase":21
 *     def __iter__(cls):
 *         return iter(cls.__members__.values())
 *     def __getitem__(cls, name):             # <<<<<<<<<<<<<<
 *         return cls.__members__[name]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_6__reduce_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumMeta_8__setstate_cython__(struct __pyx_obj___Pyx_EnumMeta *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(4, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle___Pyx_EnumMeta__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle___Pyx_EnumMeta, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle___Pyx_EnumMeta__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumMeta.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__ = {"__new__", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_1__new__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_cls = 0;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__new__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cls,&__pyx_n_s_value,&__pyx_n_s_name_2,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cls)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, 1); __PYX_ERR(4, 28, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__new__") < 0)) __PYX_ERR(4, 28, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_cls = values[0];
    __pyx_v_value = values[1];
    __pyx_v_name = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__new__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 28, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(__pyx_self, __pyx_v_cls, __pyx_v_value, __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase___new__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_cls, PyObject *__pyx_v_value, PyObject *__pyx_v_name) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__new__", 0);

  /* "EnumBase":29
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */
  if (likely(PyList_CheckExact(__pyx_v_cls)) || PyTuple_CheckExact(__pyx_v_cls)) {
    __pyx_t_1 = __pyx_v_cls; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_cls); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 29, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(4, 29, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(4, 29, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 29, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(4, 29, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "EnumBase":30
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */
    __pyx_t_4 = PyObject_RichCompare(__pyx_v_v, __pyx_v_value, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 30, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(4, 30, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "EnumBase":31
 *         for v in cls:
 *             if v == value:
 *                 return v             # <<<<<<<<<<<<<<
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_v);
      __pyx_r = __pyx_v_v;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;

      /* "EnumBase":30
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 *             if v == value:             # <<<<<<<<<<<<<<
 *                 return v
 *         if name is None:
 */
    }

    /* "EnumBase":29
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 *         for v in cls:             # <<<<<<<<<<<<<<
 *             if v == value:
 *                 return v
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":32
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */
  __pyx_t_5 = (__pyx_v_name == Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (unlikely(__pyx_t_6)) {

    /* "EnumBase":33
 *                 return v
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)             # <<<<<<<<<<<<<<
 *         res = int.__new__(cls, value)
 *         res.name = name
 */
    __pyx_t_1 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Unknown_enum_value_s, __pyx_v_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 33, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(4, 33, __pyx_L1_error)

    /* "EnumBase":32
 *             if v == value:
 *                 return v
 *         if name is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 */
  }

  /* "EnumBase":34
 *         if name is None:
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)             # <<<<<<<<<<<<<<
 *         res.name = name
 *         setattr(cls, name, res)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyInt_Type)), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_cls, __pyx_v_value};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 34, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_v_cls, __pyx_v_value};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 34, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_v_cls);
    __Pyx_GIVEREF(__pyx_v_cls);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_cls);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_value);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_res = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "EnumBase":35
 *             raise ValueError("Unknown enum value: '%s'" % value)
 *         res = int.__new__(cls, value)
 *         res.name = name             # <<<<<<<<<<<<<<
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_res, __pyx_n_s_name_2, __pyx_v_name) < 0) __PYX_ERR(4, 35, __pyx_L1_error)

  /* "EnumBase":36
 *         res = int.__new__(cls, value)
 *         res.name = name
 *         setattr(cls, name, res)             # <<<<<<<<<<<<<<
 *         cls.__members__[name] = res
 *         return res
 */
  __pyx_t_10 = PyObject_SetAttr(__pyx_v_cls, __pyx_v_name, __pyx_v_res); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(4, 36, __pyx_L1_error)

  /* "EnumBase":37
 *         res.name = name
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res             # <<<<<<<<<<<<<<
 *         return res
 *     def __repr__(self):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_cls, __pyx_n_s_members); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(PyObject_SetItem(__pyx_t_4, __pyx_v_name, __pyx_v_res) < 0)) __PYX_ERR(4, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumBase":38
 *         setattr(cls, name, res)
 *         cls.__members__[name] = res
 *         return res             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_res);
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__new__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__ = {"__repr__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__, METH_O, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_3__repr__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_2__repr__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "EnumBase":40
 *         return res
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)             # <<<<<<<<<<<<<<
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s_d, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__ = {"__str__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__, METH_O, 0};
static PyObject *__pyx_pw_8EnumBase_14__Pyx_EnumBase_5__str__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase_14__Pyx_EnumBase_4__str__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "EnumBase":42
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 *         return "%s.%s" % (self.__class__.__name__, self.name)             # <<<<<<<<<<<<<<
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("EnumBase.__Pyx_EnumBase.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta = {"__pyx_unpickle___Pyx_EnumMeta", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 1); __PYX_ERR(4, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, 2); __PYX_ERR(4, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle___Pyx_EnumMeta") < 0)) __PYX_ERR(4, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(4, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle___Pyx_EnumMeta", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8EnumBase___pyx_unpickle___Pyx_EnumMeta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype___Pyx_EnumMeta), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(4, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle___Pyx_EnumMeta__set_state(((struct __pyx_obj___Pyx_EnumMeta *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = __Pyx_EnumMeta.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_unpickle___Pyx_EnumMeta__set_state(struct __pyx_obj___Pyx_EnumMeta *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle___Pyx_EnumMeta__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(4, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle___Pyx_EnumMeta__set_state(<__Pyx_EnumMeta> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle___Pyx_EnumMeta__set_state(__Pyx_EnumMeta __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("EnumBase.__pyx_unpickle___Pyx_EnumMeta__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(4, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(4, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(4, 122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 122, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 123, __pyx_L3_error)
    } else {

      /* "View.MemoryView":123
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(4, 122, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(4, 122, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_dim;
  PyObject **__pyx_v_p;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":129
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(4, 129, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(4, 129, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":130
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__86, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 133, __pyx_L1_error)

    /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  }

  /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__87, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 136, __pyx_L1_error)

    /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  }

  /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":139
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":140
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(4, 140, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":141
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(4, 141, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(4, 141, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":144
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":145
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__88, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 148, __pyx_L1_error)

    /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  }

  /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  __pyx_t_8 = 0;
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(4, 151, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":153
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(4, 153, __pyx_L1_error)

      /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":154
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(4, 157, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":158
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":159
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(4, 160, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":161
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":162
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":164
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(4, 164, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":166
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":169
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":170
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 170, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":174
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__89, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(4, 176, __pyx_L1_error)

      /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    }

    /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":179
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":180
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */
      if (unlikely(__pyx_v_itemsize == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        __PYX_ERR(4, 180, __pyx_L1_error)
      }
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
        __PYX_ERR(4, 180, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_itemsize);
      __pyx_t_9 = __pyx_t_1;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":181
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":182
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
        Py_INCREF(Py_None);
      }

      /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    }

    /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":186
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(4, 187, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":188
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(4, 189, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":190
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__90, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 192, __pyx_L1_error)

    /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":193
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __pyx_t_4 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":194
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __pyx_t_5 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":195
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __pyx_t_6 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":196
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __pyx_t_7 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":197
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __pyx_t_7 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":199
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __pyx_t_5 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":200
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":203
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":205
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":207
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":213
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  __pyx_t_1 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":216
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    }

    /* "View.MemoryView":218
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":219
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":223
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":227
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":228
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":231
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":234
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":237
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":240
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(4, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__91, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__92, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":249
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
  /*else*/ {
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;

    /* "View.MemoryView":252
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(4, 252, __pyx_L1_error)

    /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "View.MemoryView":253
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":255
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(4, 281, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":282
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":284
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(4, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":300
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":304
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":307
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":309
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(4, 345, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(4, 345, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 345, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 345, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":346
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":347
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":349
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(4, 349, __pyx_L1_error)

    /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":351
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":352
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         global __pyx_memoryview_thread_locks_used
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":356
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 */
    __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

    /* "View.MemoryView":357
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 */
    __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

    /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  }

  /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":359
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 raise MemoryError
 */
    __pyx_v_self->lock = PyThread_allocate_lock();

    /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":361
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
      PyErr_NoMemory(); __PYX_ERR(4, 361, __pyx_L1_error)

      /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  }

  /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":364
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":366
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L10:;

  /* "View.MemoryView":368
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":370
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":374
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":377
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":378
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":383
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":385
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":388
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":387
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":389
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":391
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":395
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 397, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(4, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(4, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(4, 397, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":398
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 398, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(4, 398, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":400
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":405
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":407
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(4, 407, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(4, 407, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(4, 410, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":411
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":413
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(4, 413, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":414
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__93, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 418, __pyx_L1_error)

    /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":420
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(4, 420, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(4, 420, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(4, 422, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":423
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(4, 424, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":425
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":427
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(4, 427, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":429
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":435
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 435, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":436
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(4, 436, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":439
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(4, 445, __pyx_L1_error)

  /* "View.MemoryView":446
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(4, 446, __pyx_L1_error)

  /* "View.MemoryView":447
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(4, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __pyx_t_4 = __pyx_memoryview_copy_contents((__pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice))[0]), (__pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice))[0]), __pyx_t_2, __pyx_t_3, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(4, 445, __pyx_L1_error)

  /* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  char const *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":451
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":456
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_v_dst_slice = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice));

  /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_1 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":459
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_1 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":461
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(4, 461, __pyx_L1_error)

      /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":462
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":464
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":466
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":468
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":470
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 470, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_1 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":475
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_2 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 475, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":476
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":479
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __pyx_t_3 = __pyx_lineno; __pyx_t_4 = __pyx_clineno; __pyx_t_5 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ErrRestore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0;
      __pyx_lineno = __pyx_t_3; __pyx_clineno = __pyx_t_4; __pyx_filename = __pyx_t_5;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":482
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(4, 482, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":483
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":488
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":491
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":493
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_bytesitem);
        __Pyx_GIVEREF(__pyx_v_bytesitem);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_11) {

        /* "View.MemoryView":498
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 498, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":499
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":494
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 494, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(4, 494, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__94, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 495, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(4, 495, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  char *__pyx_t_14;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":504
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":510
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(4, 510, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":512
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(4, 512, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_9 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(4, 514, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_10 = __pyx_v_bytesvalue;
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {
    __pyx_t_11 = __pyx_t_14;
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  char *__pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__95, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 520, __pyx_L1_error)

    /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":523
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":525
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":528
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":530
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":533
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":535
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":538
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_5 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":540
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":542
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_6 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":543
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_7 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":544
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_8 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":545
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_8 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":546
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":547
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":554
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(4, 554, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":555
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(4, 555, __pyx_L1_error)

  /* "View.MemoryView":556
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":560
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":564
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_length = (__pyx_t_2[0]);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(4, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__96, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 570, __pyx_L1_error)

    /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  }

  /* "View.MemoryView":572
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
    __pyx_t_3 = __pyx_t_5;
    __pyx_v_stride = (__pyx_t_3[0]);
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(4, 572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__97, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":579
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
    __pyx_t_4 = __pyx_t_6;
    __pyx_v_suboffset = (__pyx_t_4[0]);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(4, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":583
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":587
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":591
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":596
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":598
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":599
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 599, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":601
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":603
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":607
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":609
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":613
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":616
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":622
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_v_mslice = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp));

  /* "View.MemoryView":623
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":628
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_v_mslice = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp));

  /* "View.MemoryView":629
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":633
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":635
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":636
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 636, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":641
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":645
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":647
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":648
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(4, 648, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":653
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__98, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__99, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":658
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":659
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":660
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":664
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":672
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_v_tup = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":674
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_index);
    __pyx_v_tup = __pyx_v_index;
  }
  __pyx_L3:;

  /* "View.MemoryView":676
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":677
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":678
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 679, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(4, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(4, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(4, 679, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_1) {

        /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(4, 682, __pyx_L1_error)
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        { Py_ssize_t __pyx_temp;
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {
            __Pyx_INCREF(__pyx_slice__56);
            __Pyx_GIVEREF(__pyx_slice__56);
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__56);
          }
        }
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(4, 682, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":683
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":685
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__56); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(4, 685, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":686
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":689
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(4, 689, __pyx_L1_error)

        /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
      }

      /* "View.MemoryView":691
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */
      __pyx_t_10 = (__pyx_v_have_slices != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_10 != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":692
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(4, 692, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":694
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(4, 694, __pyx_L1_error)
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  __pyx_t_1 = (__pyx_v_nslices != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":696
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {
        __Pyx_INCREF(__pyx_slice__56);
        __Pyx_GIVEREF(__pyx_slice__56);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__56);
      }
    }
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(4, 696, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  }

  /* "View.MemoryView":698
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_L14_bool_binop_done:;
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":701
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__100, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(4, 703, __pyx_L1_error)

      /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    }
  }

  /* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":711
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":718
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":722
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(4, 722, __pyx_L1_error)
    }
  }
  #endif

  /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":725
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(4, 725, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":726
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":728
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":729
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":735
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":736
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":741
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":742
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 746, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(4, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(4, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(4, 746, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":751
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 751, __pyx_L1_error)

      /* "View.MemoryView":748
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(4, 748, __pyx_L1_error)

      /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":755
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":756
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":757
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":758
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":760
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(4, 760, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 760, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":761
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(4, 761, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 761, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":762
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(4, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 762, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":764
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":765
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":766
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(4, 768, __pyx_L1_error)

      /* "View.MemoryView":774
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":778
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(4, 778, __pyx_L1_error) }

    /* "View.MemoryView":779
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(4, 779, __pyx_L1_error) }

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(4, 777, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":783
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(4, 782, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":830
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":832
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(4, 832, __pyx_L1_error)

      /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":835
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    __pyx_t_1 = (__pyx_v_have_step != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "View.MemoryView":838
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(4, 838, __pyx_L1_error)

      /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    }

    /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":843
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":845
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":848
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "View.MemoryView":850
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L14:;

        /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L12:;

      /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L11;
    }

    /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":853
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L15;
      }

      /* "View.MemoryView":855
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L15:;
    }
    __pyx_L11:;

    /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":859
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":861
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L17;
      }

      /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":863
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L17:;

      /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L16;
    }

    /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":866
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L19;
      }

      /* "View.MemoryView":868
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L19:;
    }
    __pyx_L16:;

    /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":871
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;

      /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    }

    /* "View.MemoryView":875
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":878
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":881
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":884
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":885
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":886
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":890
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L23;
  }

  /* "View.MemoryView":892
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L23:;

  /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":897
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L26;
      }

      /* "View.MemoryView":899
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":900
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(4, 899, __pyx_L1_error)
      }
      __pyx_L26:;

      /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L25;
    }

    /* "View.MemoryView":902
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L25:;

    /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":904
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":912
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":913
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":917
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(4, 917, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(4, 917, __pyx_L1_error)
    }
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":918
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":920
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":921
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":923
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":926
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":928
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(4, 928, __pyx_L1_error)

      /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    }

    /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":931
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 931, __pyx_L1_error)

    /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  }

  /* "View.MemoryView":933
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":935
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":937
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;

  /* "View.MemoryView":944
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":946
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":947
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":951
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":952
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":953
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":954
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":957
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(4, 957, __pyx_L1_error)

      /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":959
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":977
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":981
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":983
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":987
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(4, 987, __pyx_L1_error)

    /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":989
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":993
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__101, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__102, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1008
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1013
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1015
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1016
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1018
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1019
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1021
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1022
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1023
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1024
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1025
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1028
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1030
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1033
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1036
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1037
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1039
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1040
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1042
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1043
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1044
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(4, 1044, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1046
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1047
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1049
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice):
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice):
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1056
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(4, 1056, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1057
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice):
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1059
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1060
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice):
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1067
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1068
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1069
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1071
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1072
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1074
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1075
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1076
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1077
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1083
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1084
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1095
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1096
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1098
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1099
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1101
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1103
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(4, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1111
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */
    __pyx_r = (-__pyx_v_arg);
    goto __pyx_L0;

    /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  }

  /* "View.MemoryView":1113
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  /*else*/ {
    __pyx_r = __pyx_v_arg;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1121
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1122
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1124
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1126
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1127
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1129
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1131
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1132
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1135
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1137
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1147
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1148
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1149
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1154
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1155
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1157
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1158
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1159
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1160
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1162
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1163
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1167
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1168
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1173
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef int i
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1180
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef int i
 *     cdef Py_ssize_t size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1182
 *     cdef Py_ssize_t size = src.memview.view.itemsize
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         size *= src.shape[i]
 * 
 */
  __pyx_t_2 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1183
 * 
 *     for i in range(ndim):
 *         size *= src.shape[i]             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * (__pyx_v_src->shape[__pyx_v_i]));
  }

  /* "View.MemoryView":1185
 *         size *= src.shape[i]
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef int i
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1188
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1197
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1198
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride = stride * shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1199
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride = stride * shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1200
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride = stride * shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1197
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1202
 *             stride = stride * shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride = stride * shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1203
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride = stride * shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1204
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride = stride * shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1206
 *             stride = stride * shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1188
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1209
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;

  /* "View.MemoryView":1220
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1221
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1223
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1224
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1225
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(4, 1225, __pyx_L1_error)

    /* "View.MemoryView":1224
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  }

  /* "View.MemoryView":1228
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1229
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1230
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1231
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1232
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1234
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1238
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1239
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1240
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1239
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1242
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1243
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1242
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1245
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1247
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1209
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1252
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1255
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "View.MemoryView":1254
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(4, 1254, __pyx_L1_error)

  /* "View.MemoryView":1252
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1258
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1259
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 1259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(4, 1259, __pyx_L1_error)

  /* "View.MemoryView":1258
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1262
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1263
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1264
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(4, 1264, __pyx_L1_error)

    /* "View.MemoryView":1263
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  }

  /* "View.MemoryView":1266
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */
  /*else*/ {
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(4, 1266, __pyx_L1_error)
  }

  /* "View.MemoryView":1262
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1269
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;

  /* "View.MemoryView":1277
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1278
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1280
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1281
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1282
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1285
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1286
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1285
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1287
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1288
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1287
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1290
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1292
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1293
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1294
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1295
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1296
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1294
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1298
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(4, 1298, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1293
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1300
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1301
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(4, 1301, __pyx_L1_error)

      /* "View.MemoryView":1300
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1303
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1305
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1306
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1305
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1308
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(4, 1308, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1309
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1303
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1311
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1314
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1315
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1314
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1316
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1317
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1316
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1319
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1321
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1322
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1323
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1324
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1325
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1319
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    }

    /* "View.MemoryView":1311
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1327
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1330
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(4, 1330, __pyx_L1_error)

    /* "View.MemoryView":1331
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(4, 1331, __pyx_L1_error)

    /* "View.MemoryView":1327
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1333
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1334
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1335
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1337
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1338
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1269
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1341
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1345
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1347
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1348
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1349
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1350
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1352
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1353
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1354
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1355
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1341
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1363
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  int __pyx_t_1;

  /* "View.MemoryView":1367
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1368
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1367
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  }

  /* "View.MemoryView":1363
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":1372
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1375
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1372
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1378
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1382
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1383
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1384
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1385
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1384
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1387
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1383
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1389
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */
    /*else*/ {

      /* "View.MemoryView":1390
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1392
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));
  }

  /* "View.MemoryView":1378
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1398
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1401
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1402
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1404
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1398
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1408
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1412
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1413
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1415
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1416
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1417
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1418
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1415
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1420
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1421
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1423
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1408
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(4, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(4, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(4, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(4, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xb068931) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(4, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(4, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(4, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(4, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(4, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(4, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(4, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Font(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_4Font_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_Font(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7celiagg_8_celiagg_4Font_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Font_cache_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Font_10cache_type_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Font_face_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Font_10face_index_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Font_filepath(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Font_8filepath_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Font_height(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Font_6height_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Font_hinting(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Font_7hinting_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_4Font_hinting(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_4Font_7hinting_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_Font[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Font_5copy, METH_NOARGS, 0},
  {"width", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Font_7width, METH_O, __pyx_doc_7celiagg_8_celiagg_4Font_6width},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Font_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Font_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7celiagg_8_celiagg_Font[] = {
  {(char *)"cache_type", __pyx_getprop_7celiagg_8_celiagg_4Font_cache_type, 0, (char *)0, 0},
  {(char *)"face_index", __pyx_getprop_7celiagg_8_celiagg_4Font_face_index, 0, (char *)0, 0},
  {(char *)"filepath", __pyx_getprop_7celiagg_8_celiagg_4Font_filepath, 0, (char *)0, 0},
  {(char *)"height", __pyx_getprop_7celiagg_8_celiagg_4Font_height, 0, (char *)0, 0},
  {(char *)"hinting", __pyx_getprop_7celiagg_8_celiagg_4Font_hinting, __pyx_setprop_7celiagg_8_celiagg_4Font_hinting, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_Font = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.Font", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_Font), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_Font, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Font(path, size, cache_type)\n\n        :param path: A Unicode string containing the path of a ``Font`` file\n        :param size: The size of the font\n        :param cache_type: A ``FontCacheType``\n        :param face_index: For .ttc fonts, the index of the desired font within\n                           the collection.\n        ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_Font, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7celiagg_8_celiagg_Font, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_Font, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Rect(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_4Rect_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_Rect(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7celiagg_8_celiagg_4Rect_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Rect_x(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Rect_1x_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_4Rect_x(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_4Rect_1x_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Rect_w(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Rect_1w_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_4Rect_w(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_4Rect_1w_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Rect_y(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Rect_1y_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_4Rect_y(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_4Rect_1y_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Rect_h(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Rect_1h_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_4Rect_h(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_4Rect_1h_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_4Rect_valid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_4Rect_5valid_1__get__(o);
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_Rect[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Rect_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Rect_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7celiagg_8_celiagg_Rect[] = {
  {(char *)"x", __pyx_getprop_7celiagg_8_celiagg_4Rect_x, __pyx_setprop_7celiagg_8_celiagg_4Rect_x, (char *)0, 0},
  {(char *)"w", __pyx_getprop_7celiagg_8_celiagg_4Rect_w, __pyx_setprop_7celiagg_8_celiagg_4Rect_w, (char *)0, 0},
  {(char *)"y", __pyx_getprop_7celiagg_8_celiagg_4Rect_y, __pyx_setprop_7celiagg_8_celiagg_4Rect_y, (char *)0, 0},
  {(char *)"h", __pyx_getprop_7celiagg_8_celiagg_4Rect_h, __pyx_setprop_7celiagg_8_celiagg_4Rect_h, (char *)0, 0},
  {(char *)"valid", __pyx_getprop_7celiagg_8_celiagg_4Rect_valid, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_Rect = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.Rect", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_Rect), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_Rect, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7celiagg_8_celiagg_4Rect_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Rect(x, y, w, h)\n    A rectangle.\n\n    :param x: The X coordinate of the top left corner\n    :param y: The Y coordinate of the top left corner\n    :param w: The width of the rectangle in pixels\n    :param h: The height of the rectangle in pixels\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_7celiagg_8_celiagg_4Rect_7__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_Rect, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7celiagg_8_celiagg_Rect, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_Rect, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_GraphicsState __pyx_vtable_7celiagg_8_celiagg_GraphicsState;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_GraphicsState(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)o);
  p->__pyx_vtab = __pyx_vtabptr_7celiagg_8_celiagg_GraphicsState;
  p->_stencil_img = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_13GraphicsState_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_GraphicsState(PyObject *o) {
  struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *p = (struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7celiagg_8_celiagg_13GraphicsState_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_stencil_img);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_7celiagg_8_celiagg_GraphicsState(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *p = (struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)o;
  if (p->_stencil_img) {
    e = (*v)(((PyObject *)p->_stencil_img), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7celiagg_8_celiagg_GraphicsState(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *p = (struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *)o;
  tmp = ((PyObject*)p->_stencil_img);
  p->_stencil_img = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_anti_aliased(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12anti_aliased_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_anti_aliased(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12anti_aliased_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_clip_box(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_8clip_box_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_clip_box(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_8clip_box_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_drawing_mode(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12drawing_mode_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_drawing_mode(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12drawing_mode_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_text_drawing_mode(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_text_drawing_mode(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17text_drawing_mode_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_blend_mode(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10blend_mode_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_blend_mode(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10blend_mode_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_image_blend_mode(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_image_blend_mode(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_16image_blend_mode_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_master_alpha(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12master_alpha_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_master_alpha(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_12master_alpha_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_width(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10line_width_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_width(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10line_width_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_cap(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_8line_cap_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_cap(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_8line_cap_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_join(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_9line_join_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_join(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_9line_join_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_inner_join(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10inner_join_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_inner_join(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_10inner_join_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_miter_limit(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_11miter_limit_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_miter_limit(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_11miter_limit_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_inner_miter_limit(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_inner_miter_limit(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17inner_miter_limit_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_dash_pattern(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_dash_pattern(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_17line_dash_pattern_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_dash_phase(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_dash_phase(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_15line_dash_phase_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_13GraphicsState_stencil(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_7stencil_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_stencil(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_13GraphicsState_7stencil_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_GraphicsState[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_13GraphicsState_7copy, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_13GraphicsState_6copy},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_13GraphicsState_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_13GraphicsState_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7celiagg_8_celiagg_GraphicsState[] = {
  {(char *)"anti_aliased", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_anti_aliased, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_anti_aliased, (char *)0, 0},
  {(char *)"clip_box", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_clip_box, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_clip_box, (char *)0, 0},
  {(char *)"drawing_mode", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_drawing_mode, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_drawing_mode, (char *)0, 0},
  {(char *)"text_drawing_mode", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_text_drawing_mode, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_text_drawing_mode, (char *)0, 0},
  {(char *)"blend_mode", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_blend_mode, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_blend_mode, (char *)0, 0},
  {(char *)"image_blend_mode", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_image_blend_mode, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_image_blend_mode, (char *)0, 0},
  {(char *)"master_alpha", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_master_alpha, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_master_alpha, (char *)0, 0},
  {(char *)"line_width", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_width, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_width, (char *)0, 0},
  {(char *)"line_cap", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_cap, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_cap, (char *)0, 0},
  {(char *)"line_join", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_join, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_join, (char *)0, 0},
  {(char *)"inner_join", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_inner_join, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_inner_join, (char *)0, 0},
  {(char *)"miter_limit", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_miter_limit, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_miter_limit, (char *)0, 0},
  {(char *)"inner_miter_limit", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_inner_miter_limit, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_inner_miter_limit, (char *)0, 0},
  {(char *)"line_dash_pattern", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_dash_pattern, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_dash_pattern, (char *)0, 0},
  {(char *)"line_dash_phase", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_line_dash_phase, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_line_dash_phase, (char *)0, 0},
  {(char *)"stencil", __pyx_getprop_7celiagg_8_celiagg_13GraphicsState_stencil, __pyx_setprop_7celiagg_8_celiagg_13GraphicsState_stencil, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_GraphicsState = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.GraphicsState", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_GraphicsState, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "GraphicsState(**properties)\n    A container for drawing state. Can be initialized using keyword\n    arguments and works as a namespace for the following properties:\n\n    * anti_aliased: A boolean denoting whether drawing is anti-aliased or not.\n    * drawing_mode: A ``DrawingMode`` value denoting the drawing mode.\n    * text_drawing_mode: A ``TextDrawingMode`` value denoting the text drawing\n                         mode.\n    * blend_mode: A ``BlendMode`` for non-image drawing. (ignored)\n    * image_blend_mode: A ``BlendMode`` for image drawing. (ignored)\n    * line_cap: A ``LineCap`` value denoting the style of line ends.\n    * line_join: A ``LineJoin`` value denoting the style of joins.\n    * inner_join: An ``InnerJoin`` value denoting the style of inner joins.\n    * miter_limit: The miter limit\n    * inner_miter_limit: The miter limit for inner joins\n    * master_alpha: A master opacity value.\n    * line_width: The width when stroking lines.\n    * clip_box: A ``Rect`` which defines a simple clipping area.\n    * line_dash_pattern: A sequence of (dash length, gap length) pairs.\n    * line_dash_phase: Where in ``line_dash_pattern`` to start, when drawing.\n    * stencil: An ``Image`` with format ``Gray8`` which will mask any drawing.\n               The dimensions must match the size of the canvas that is being\n               drawn to.\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_GraphicsState, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_GraphicsState, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_GraphicsState, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7celiagg_8_celiagg_GraphicsState, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_7celiagg_8_celiagg_13GraphicsState_5__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_GraphicsState, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Image(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_Image *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)o);
  p->pixel_array = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_5Image_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_Image(PyObject *o) {
  struct __pyx_obj_7celiagg_8_celiagg_Image *p = (struct __pyx_obj_7celiagg_8_celiagg_Image *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7celiagg_8_celiagg_5Image_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->pixel_array);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_7celiagg_8_celiagg_Image(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7celiagg_8_celiagg_Image *p = (struct __pyx_obj_7celiagg_8_celiagg_Image *)o;
  if (p->pixel_array) {
    e = (*v)(p->pixel_array, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7celiagg_8_celiagg_Image(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7celiagg_8_celiagg_Image *p = (struct __pyx_obj_7celiagg_8_celiagg_Image *)o;
  tmp = ((PyObject*)p->pixel_array);
  p->pixel_array = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_5Image_format(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_5Image_6format_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_5Image_pixels(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_5Image_6pixels_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_5Image_height(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_5Image_6height_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_5Image_width(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_5Image_5width_1__get__(o);
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_Image[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_5Image_5copy, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_5Image_4copy},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_5Image_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_5Image_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7celiagg_8_celiagg_Image[] = {
  {(char *)"format", __pyx_getprop_7celiagg_8_celiagg_5Image_format, 0, (char *)0, 0},
  {(char *)"pixels", __pyx_getprop_7celiagg_8_celiagg_5Image_pixels, 0, (char *)0, 0},
  {(char *)"height", __pyx_getprop_7celiagg_8_celiagg_5Image_height, 0, (char *)0, 0},
  {(char *)"width", __pyx_getprop_7celiagg_8_celiagg_5Image_width, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_Image = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.Image", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_Image), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_Image, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Image(array, pixel_format, bottom_up=False)\n\n    :param image: A 2D or 3D numpy array containing image data\n    :param pixel_format: A PixelFormat describing the image's pixel format\n    :param bottom_up: If True, the image data starts at the bottom of the image\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_Image, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_Image, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_Image, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7celiagg_8_celiagg_Image, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_Image, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase __pyx_vtable_7celiagg_8_celiagg_CanvasBase;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_7celiagg_8_celiagg_CanvasBase;
  p->py_array = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_CanvasBase(PyObject *o) {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *p = (struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7celiagg_8_celiagg_10CanvasBase_1__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->py_array);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_7celiagg_8_celiagg_CanvasBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *p = (struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)o;
  if (p->py_array) {
    e = (*v)(p->py_array, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7celiagg_8_celiagg_CanvasBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *p = (struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *)o;
  tmp = ((PyObject*)p->py_array);
  p->py_array = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_10CanvasBase_channel_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_10CanvasBase_13channel_count_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_10CanvasBase_width(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_10CanvasBase_5width_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_10CanvasBase_height(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_10CanvasBase_6height_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_10CanvasBase_array(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_10CanvasBase_5array_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_10CanvasBase_image(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_10CanvasBase_5image_1__get__(o);
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_CanvasBase[] = {
  {"clear", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_10CanvasBase_3clear, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_10CanvasBase_2clear},
  {"draw_image", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_10CanvasBase_5draw_image, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_10CanvasBase_4draw_image},
  {"draw_shape", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_10CanvasBase_7draw_shape, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_10CanvasBase_6draw_shape},
  {"draw_text", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_10CanvasBase_9draw_text, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_10CanvasBase_8draw_text},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_10CanvasBase_11__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_10CanvasBase_13__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7celiagg_8_celiagg_CanvasBase[] = {
  {(char *)"channel_count", __pyx_getprop_7celiagg_8_celiagg_10CanvasBase_channel_count, 0, (char *)0, 0},
  {(char *)"width", __pyx_getprop_7celiagg_8_celiagg_10CanvasBase_width, 0, (char *)0, 0},
  {(char *)"height", __pyx_getprop_7celiagg_8_celiagg_10CanvasBase_height, 0, (char *)0, 0},
  {(char *)"array", __pyx_getprop_7celiagg_8_celiagg_10CanvasBase_array, 0, (char *)0, 0},
  {(char *)"image", __pyx_getprop_7celiagg_8_celiagg_10CanvasBase_image, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_CanvasBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.CanvasBase", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_CanvasBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_CanvasBase, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_CanvasBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_CanvasBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7celiagg_8_celiagg_CanvasBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_CanvasBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGBA128 __pyx_vtable_7celiagg_8_celiagg_CanvasRGBA128;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasRGBA128(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_CanvasBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128 *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase*)__pyx_vtabptr_7celiagg_8_celiagg_CanvasRGBA128;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_13CanvasRGBA128_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_CanvasRGBA128[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_13CanvasRGBA128_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_13CanvasRGBA128_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_CanvasRGBA128 = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.CanvasRGBA128", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA128), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_CanvasBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "CanvasRGBA128(array, bottom_up=False)\n    Provides AGG (Anti-Grain Geometry) drawing routines that render to the\n    numpy array passed as the constructor argument. Because this array is\n    modified in place, it must be of type ``numpy.float32``, must be\n    C-contiguous, and must be MxNx4 (4 channels: red, green, blue, and alpha).\n\n    :param array: A ``numpy.float32`` array with shape (H, W, 4).\n    :param bottom_up: If True, the origin is the bottom left, instead of top-left\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_CanvasBase, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_CanvasBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_CanvasRGBA128, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_CanvasRGBA128, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBGRA32 __pyx_vtable_7celiagg_8_celiagg_CanvasBGRA32;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasBGRA32(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_CanvasBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32 *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase*)__pyx_vtabptr_7celiagg_8_celiagg_CanvasBGRA32;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_12CanvasBGRA32_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_CanvasBGRA32[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12CanvasBGRA32_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12CanvasBGRA32_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_CanvasBGRA32 = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.CanvasBGRA32", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_CanvasBGRA32), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_CanvasBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "CanvasBGRA32(array, bottom_up=False)\n    Provides AGG (Anti-Grain Geometry) drawing routines that render to the\n    numpy array passed as the constructor argument. Because this array is\n    modified in place, it must be of type ``numpy.uint8``, must be\n    C-contiguous, and must be MxNx4 (4 channels: blue, green, red, alpha).\n\n    :param array: A ``numpy.uint8`` array with shape (H, W, 4).\n    :param bottom_up: If True, the origin is the bottom left, instead of top-left\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_CanvasBase, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_CanvasBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_CanvasBGRA32, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_CanvasBGRA32, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGBA32 __pyx_vtable_7celiagg_8_celiagg_CanvasRGBA32;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasRGBA32(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_CanvasBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32 *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase*)__pyx_vtabptr_7celiagg_8_celiagg_CanvasRGBA32;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_12CanvasRGBA32_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_CanvasRGBA32[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12CanvasRGBA32_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12CanvasRGBA32_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_CanvasRGBA32 = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.CanvasRGBA32", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_CanvasRGBA32), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_CanvasBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "CanvasRGBA32(array, bottom_up=False)\n    Provides AGG (Anti-Grain Geometry) drawing routines that render to the\n    numpy array passed as the constructor argument. Because this array is\n    modified in place, it must be of type ``numpy.uint8``, must be\n    C-contiguous, and must be MxNx4 (4 channels: red, green, blue, and alpha).\n\n    :param array: A ``numpy.uint8`` array with shape (H, W, 4).\n    :param bottom_up: If True, the origin is the bottom left, instead of top-left\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_CanvasBase, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_CanvasBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_CanvasRGBA32, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_CanvasRGBA32, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasRGB24 __pyx_vtable_7celiagg_8_celiagg_CanvasRGB24;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasRGB24(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_CanvasBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24 *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase*)__pyx_vtabptr_7celiagg_8_celiagg_CanvasRGB24;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_11CanvasRGB24_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_CanvasRGB24[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_11CanvasRGB24_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_11CanvasRGB24_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_CanvasRGB24 = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.CanvasRGB24", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_CanvasRGB24), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_CanvasBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "CanvasRGB24(array, bottom_up=False)\n    Provides AGG (Anti-Grain Geometry) drawing routines that render to the\n    numpy array passed as the constructor argument. Because this array is\n    modified in place, it must be of type ``numpy.uint8``, must be\n    C-contiguous, and must be MxNx3 (3 channels: red, green, blue).\n\n    :param array: A ``numpy.uint8`` array with shape (H, W, 3).\n    :param bottom_up: If True, the origin is the bottom left, instead of top-left\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_CanvasBase, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_CanvasBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_CanvasRGB24, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_CanvasRGB24, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasGA16 __pyx_vtable_7celiagg_8_celiagg_CanvasGA16;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasGA16(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_CanvasBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16 *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase*)__pyx_vtabptr_7celiagg_8_celiagg_CanvasGA16;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_10CanvasGA16_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_CanvasGA16[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_10CanvasGA16_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_10CanvasGA16_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_CanvasGA16 = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.CanvasGA16", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_CanvasGA16), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_CanvasBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "CanvasGA16(array, bottom_up=False)\n    Provides AGG (Anti-Grain Geometry) drawing routines that render to the\n    numpy array passed as the constructor argument. Because this array is\n    modified in place, it must be of type ``numpy.uint8``, must be\n    C-contiguous, and must be MxNx2 (2 channels: intensity and alpha).\n\n    :param array: A ``numpy.uint8`` array with shape (H, W, 2).\n    :param bottom_up: If True, the origin is the bottom left, instead of top-left\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_CanvasBase, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_CanvasBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_CanvasGA16, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_CanvasGA16, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasG8 __pyx_vtable_7celiagg_8_celiagg_CanvasG8;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_CanvasG8(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_CanvasBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_CanvasG8 *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7celiagg_8_celiagg_CanvasBase*)__pyx_vtabptr_7celiagg_8_celiagg_CanvasG8;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_8CanvasG8_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_CanvasG8[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_8CanvasG8_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_8CanvasG8_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_CanvasG8 = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.CanvasG8", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_CanvasG8), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_CanvasBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "CanvasG8(array, bottom_up=False)\n    Provides AGG (Anti-Grain Geometry) drawing routines that render to the\n    numpy array passed as the constructor argument. Because this array is\n    modified in place, it must be of type ``numpy.uint8``, must be\n    C-contiguous, and must be MxN (1 channel: intensity).\n\n    :param array: A ``numpy.uint8`` array with shape (H, W).\n    :param bottom_up: If True, the origin is the bottom left, instead of top-left\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_CanvasBase, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_CanvasBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_CanvasG8, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_CanvasG8, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Paint(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_Paint(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7celiagg_8_celiagg_5Paint_1__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_5Paint_transform(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_5Paint_9transform_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_5Paint_transform(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_5Paint_9transform_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_Paint[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_5Paint_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_5Paint_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7celiagg_8_celiagg_Paint[] = {
  {(char *)"transform", __pyx_getprop_7celiagg_8_celiagg_5Paint_transform, __pyx_setprop_7celiagg_8_celiagg_5Paint_transform, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_Paint = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.Paint", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_Paint), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_Paint, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_Paint, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7celiagg_8_celiagg_Paint, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_Paint, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_LinearGradientPaint(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_Paint(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *)o);
  p->_points = Py_None; Py_INCREF(Py_None);
  p->_stops = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_LinearGradientPaint(PyObject *o) {
  struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *p = (struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_points);
  Py_CLEAR(p->_stops);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_7celiagg_8_celiagg_Paint(o);
}

static int __pyx_tp_traverse_7celiagg_8_celiagg_LinearGradientPaint(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *p = (struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *)o;
  e = ((likely(__pyx_ptype_7celiagg_8_celiagg_Paint)) ? ((__pyx_ptype_7celiagg_8_celiagg_Paint->tp_traverse) ? __pyx_ptype_7celiagg_8_celiagg_Paint->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7celiagg_8_celiagg_LinearGradientPaint)); if (e) return e;
  if (p->_points) {
    e = (*v)(p->_points, a); if (e) return e;
  }
  if (p->_stops) {
    e = (*v)(p->_stops, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7celiagg_8_celiagg_LinearGradientPaint(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *p = (struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint *)o;
  if (likely(__pyx_ptype_7celiagg_8_celiagg_Paint)) { if (__pyx_ptype_7celiagg_8_celiagg_Paint->tp_clear) __pyx_ptype_7celiagg_8_celiagg_Paint->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7celiagg_8_celiagg_LinearGradientPaint);
  tmp = ((PyObject*)p->_points);
  p->_points = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_stops);
  p->_stops = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_LinearGradientPaint[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_3copy, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_19LinearGradientPaint_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_LinearGradientPaint = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.LinearGradientPaint", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_LinearGradientPaint), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_LinearGradientPaint, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "LinearGradientPaint(x1, y1, x2, y2, stops, spread, units)\n    A ``Paint`` for drawing linear gradients\n\n    :param x1: Gradient start point X\n    :param y1: Gradient start point Y\n    :param x2: Gradient end point X\n    :param y2: Gradient end point Y\n    :param stops: An iterable of gradient stop tuples: (offset, r, g, b, a)\n    :param spread: The ``GradientSpread`` type for this gradient\n    :param units: The ``GradientUnits`` type for this gradient\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_LinearGradientPaint, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_LinearGradientPaint, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_LinearGradientPaint, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_LinearGradientPaint, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_RadialGradientPaint(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_Paint(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *)o);
  p->_points = Py_None; Py_INCREF(Py_None);
  p->_stops = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_RadialGradientPaint(PyObject *o) {
  struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *p = (struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_points);
  Py_CLEAR(p->_stops);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_7celiagg_8_celiagg_Paint(o);
}

static int __pyx_tp_traverse_7celiagg_8_celiagg_RadialGradientPaint(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *p = (struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *)o;
  e = ((likely(__pyx_ptype_7celiagg_8_celiagg_Paint)) ? ((__pyx_ptype_7celiagg_8_celiagg_Paint->tp_traverse) ? __pyx_ptype_7celiagg_8_celiagg_Paint->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7celiagg_8_celiagg_RadialGradientPaint)); if (e) return e;
  if (p->_points) {
    e = (*v)(p->_points, a); if (e) return e;
  }
  if (p->_stops) {
    e = (*v)(p->_stops, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7celiagg_8_celiagg_RadialGradientPaint(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *p = (struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint *)o;
  if (likely(__pyx_ptype_7celiagg_8_celiagg_Paint)) { if (__pyx_ptype_7celiagg_8_celiagg_Paint->tp_clear) __pyx_ptype_7celiagg_8_celiagg_Paint->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7celiagg_8_celiagg_RadialGradientPaint);
  tmp = ((PyObject*)p->_points);
  p->_points = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_stops);
  p->_stops = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_RadialGradientPaint[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_3copy, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_19RadialGradientPaint_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_RadialGradientPaint = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.RadialGradientPaint", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_RadialGradientPaint), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_RadialGradientPaint, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "RadialGradientPaint(cx, cy, r, fx, fy, stops, spread, units)\n    A ``Paint`` for drawing radial gradients\n\n    :param cx: Gradient center point X\n    :param cy: Gradient center point Y\n    :param r: Gradient radius\n    :param fx: Gradient focus point X\n    :param fy: Gradient focus point Y\n    :param stops: An iterable of gradient stop tuples: (offset, r, g, b, a)\n    :param spread: The ``GradientSpread`` type for this gradient\n    :param units: The ``GradientUnits`` type for this gradient\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_RadialGradientPaint, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_RadialGradientPaint, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_RadialGradientPaint, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_RadialGradientPaint, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_PatternPaint(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_Paint(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)o);
  p->img_obj = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_12PatternPaint_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_PatternPaint(PyObject *o) {
  struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *p = (struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->img_obj);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_7celiagg_8_celiagg_Paint(o);
}

static int __pyx_tp_traverse_7celiagg_8_celiagg_PatternPaint(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *p = (struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)o;
  e = ((likely(__pyx_ptype_7celiagg_8_celiagg_Paint)) ? ((__pyx_ptype_7celiagg_8_celiagg_Paint->tp_traverse) ? __pyx_ptype_7celiagg_8_celiagg_Paint->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7celiagg_8_celiagg_PatternPaint)); if (e) return e;
  if (p->img_obj) {
    e = (*v)(((PyObject *)p->img_obj), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7celiagg_8_celiagg_PatternPaint(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *p = (struct __pyx_obj_7celiagg_8_celiagg_PatternPaint *)o;
  if (likely(__pyx_ptype_7celiagg_8_celiagg_Paint)) { if (__pyx_ptype_7celiagg_8_celiagg_Paint->tp_clear) __pyx_ptype_7celiagg_8_celiagg_Paint->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7celiagg_8_celiagg_PatternPaint);
  tmp = ((PyObject*)p->img_obj);
  p->img_obj = ((struct __pyx_obj_7celiagg_8_celiagg_Image *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_PatternPaint[] = {
  {"_with_format", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12PatternPaint_3_with_format, METH_O, 0},
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12PatternPaint_5copy, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12PatternPaint_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12PatternPaint_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_PatternPaint = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.PatternPaint", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_PatternPaint), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_PatternPaint, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "PatternPaint(style, image)\n    A ``Paint`` for repeating patterns.\n\n    :param style: A ``PatternStyle``\n    :param image: An ``Image`` object\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_PatternPaint, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_PatternPaint, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_PatternPaint, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_PatternPaint, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_SolidPaint(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_Paint(t, a, k);
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_10SolidPaint_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_10SolidPaint_a(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1a_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_10SolidPaint_a(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1a_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_10SolidPaint_r(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1r_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_10SolidPaint_r(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1r_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_10SolidPaint_g(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1g_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_10SolidPaint_g(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1g_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_10SolidPaint_b(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1b_1__get__(o);
}

static int __pyx_setprop_7celiagg_8_celiagg_10SolidPaint_b(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_7celiagg_8_celiagg_10SolidPaint_1b_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_SolidPaint[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_10SolidPaint_7copy, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_10SolidPaint_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_10SolidPaint_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7celiagg_8_celiagg_SolidPaint[] = {
  {(char *)"a", __pyx_getprop_7celiagg_8_celiagg_10SolidPaint_a, __pyx_setprop_7celiagg_8_celiagg_10SolidPaint_a, (char *)0, 0},
  {(char *)"r", __pyx_getprop_7celiagg_8_celiagg_10SolidPaint_r, __pyx_setprop_7celiagg_8_celiagg_10SolidPaint_r, (char *)0, 0},
  {(char *)"g", __pyx_getprop_7celiagg_8_celiagg_10SolidPaint_g, __pyx_setprop_7celiagg_8_celiagg_10SolidPaint_g, (char *)0, 0},
  {(char *)"b", __pyx_getprop_7celiagg_8_celiagg_10SolidPaint_b, __pyx_setprop_7celiagg_8_celiagg_10SolidPaint_b, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_SolidPaint = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.SolidPaint", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_SolidPaint), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_Paint, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7celiagg_8_celiagg_10SolidPaint_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "SolidPaint(r, g, b, a)\n    A ``Paint`` for solid colors.\n\n    :param r: Red value in [0, 1]\n    :param g: Green value in [0, 1]\n    :param b: Blue value in [0, 1]\n    :param a: Alpha value in [0, 1] (defaults to 1.0)\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_7celiagg_8_celiagg_10SolidPaint_5__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_SolidPaint, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7celiagg_8_celiagg_SolidPaint, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_SolidPaint, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_7celiagg_8_celiagg_Transform __pyx_vtable_7celiagg_8_celiagg_Transform;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Transform(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_Transform *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_Transform *)o);
  p->__pyx_vtab = __pyx_vtabptr_7celiagg_8_celiagg_Transform;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_9Transform_3__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_Transform(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7celiagg_8_celiagg_9Transform_5__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_9Transform_sx(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_9Transform_2sx_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_9Transform_sy(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_9Transform_2sy_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_9Transform_shx(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_9Transform_3shx_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_9Transform_shy(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_9Transform_3shy_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_9Transform_tx(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_9Transform_2tx_1__get__(o);
}

static PyObject *__pyx_getprop_7celiagg_8_celiagg_9Transform_ty(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_7celiagg_8_celiagg_9Transform_2ty_1__get__(o);
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_Transform[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_9Transform_9copy, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_9Transform_8copy},
  {"reset", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_9Transform_11reset, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_9Transform_10reset},
  {"invert", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_9Transform_13invert, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_9Transform_12invert},
  {"scale", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_9Transform_15scale, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_9Transform_14scale},
  {"rotate", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_9Transform_17rotate, METH_O, __pyx_doc_7celiagg_8_celiagg_9Transform_16rotate},
  {"skew", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_9Transform_19skew, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_9Transform_18skew},
  {"translate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_9Transform_21translate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_9Transform_20translate},
  {"premultiply", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_9Transform_23premultiply, METH_O, __pyx_doc_7celiagg_8_celiagg_9Transform_22premultiply},
  {"multiply", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_9Transform_25multiply, METH_O, __pyx_doc_7celiagg_8_celiagg_9Transform_24multiply},
  {"worldToScreen", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_9Transform_27worldToScreen, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_9Transform_26worldToScreen},
  {"screenToWorld", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_9Transform_29screenToWorld, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_9Transform_28screenToWorld},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_9Transform_31__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_9Transform_33__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_7celiagg_8_celiagg_Transform[] = {
  {(char *)"sx", __pyx_getprop_7celiagg_8_celiagg_9Transform_sx, 0, (char *)0, 0},
  {(char *)"sy", __pyx_getprop_7celiagg_8_celiagg_9Transform_sy, 0, (char *)0, 0},
  {(char *)"shx", __pyx_getprop_7celiagg_8_celiagg_9Transform_shx, 0, (char *)0, 0},
  {(char *)"shy", __pyx_getprop_7celiagg_8_celiagg_9Transform_shy, 0, (char *)0, 0},
  {(char *)"tx", __pyx_getprop_7celiagg_8_celiagg_9Transform_tx, 0, (char *)0, 0},
  {(char *)"ty", __pyx_getprop_7celiagg_8_celiagg_9Transform_ty, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_Transform = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.Transform", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_Transform), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_Transform, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_7celiagg_8_celiagg_9Transform_7__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Transform(sx=1.0, shy=0.0, shx=0.0, sy=1.0, tx=0.0, ty=0.0)\n    A 2D affine transform.\n\n    After construction, the same names can be used as get/set properties for\n    individual components of the transform.\n\n    NOTE: If you're familar with SVG, you'll notice the format of this\n    constructor. The arguments are exactly the same as the SVG `matrix`\n    attribute.\n\n    :param sx: Scale X\n    :param shy: Shear/Skew Y\n    :param shx: Shear/Skew X\n    :param sy: Scale Y\n    :param tx: Translation X\n    :param ty: Translation Y\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_Transform, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_7celiagg_8_celiagg_Transform, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_7celiagg_8_celiagg_9Transform_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_Transform, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_VertexSource(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_VertexSource(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_7celiagg_8_celiagg_12VertexSource_1__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_VertexSource[] = {
  {"bounding_rect", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12VertexSource_6bounding_rect, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_12VertexSource_5bounding_rect},
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12VertexSource_8copy, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_12VertexSource_7copy},
  {"final_point", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12VertexSource_10final_point, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_12VertexSource_9final_point},
  {"length", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12VertexSource_12length, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_12VertexSource_11length},
  {"vertices", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12VertexSource_14vertices, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_12VertexSource_13vertices},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12VertexSource_16__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_12VertexSource_18__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_VertexSource = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.VertexSource", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_VertexSource), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_VertexSource, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "An object which supplies vertex pairs and other information to\n    low-level drawing routines.\n\n    NOTE: This class should not be explicitly instantiated.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_7celiagg_8_celiagg_12VertexSource_3__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_VertexSource, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_VertexSource, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_BSpline(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_BSpline *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_VertexSource(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_BSpline *)o);
  p->_points = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_7BSpline_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_BSpline(PyObject *o) {
  struct __pyx_obj_7celiagg_8_celiagg_BSpline *p = (struct __pyx_obj_7celiagg_8_celiagg_BSpline *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_points);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_7celiagg_8_celiagg_VertexSource(o);
}

static int __pyx_tp_traverse_7celiagg_8_celiagg_BSpline(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7celiagg_8_celiagg_BSpline *p = (struct __pyx_obj_7celiagg_8_celiagg_BSpline *)o;
  e = ((likely(__pyx_ptype_7celiagg_8_celiagg_VertexSource)) ? ((__pyx_ptype_7celiagg_8_celiagg_VertexSource->tp_traverse) ? __pyx_ptype_7celiagg_8_celiagg_VertexSource->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7celiagg_8_celiagg_BSpline)); if (e) return e;
  if (p->_points) {
    e = (*v)(p->_points, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7celiagg_8_celiagg_BSpline(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7celiagg_8_celiagg_BSpline *p = (struct __pyx_obj_7celiagg_8_celiagg_BSpline *)o;
  if (likely(__pyx_ptype_7celiagg_8_celiagg_VertexSource)) { if (__pyx_ptype_7celiagg_8_celiagg_VertexSource->tp_clear) __pyx_ptype_7celiagg_8_celiagg_VertexSource->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7celiagg_8_celiagg_BSpline);
  tmp = ((PyObject*)p->_points);
  p->_points = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_BSpline[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_7BSpline_3copy, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_7BSpline_2copy},
  {"final_point", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_7BSpline_5final_point, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_7BSpline_4final_point},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_7BSpline_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_7BSpline_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_BSpline = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.BSpline", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_BSpline), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_BSpline, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "BSpline(points)\n    A B-Spline curve.\n\n    :param points: 2D array of (x, y) pairs\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_BSpline, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_BSpline, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_7celiagg_8_celiagg_12VertexSource_3__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_BSpline, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_BSpline, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_Path(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_VertexSource(t, a, k);
  if (unlikely(!o)) return 0;
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_4Path_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_Path[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_3copy, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_4Path_2copy},
  {"final_point", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_5final_point, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_4Path_4final_point},
  {"begin", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_7begin, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_4Path_6begin},
  {"close", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_9close, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_4Path_8close},
  {"reset", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_11reset, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_4Path_10reset},
  {"add_path", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_13add_path, METH_O, __pyx_doc_7celiagg_8_celiagg_4Path_12add_path},
  {"move_to", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_4Path_15move_to, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_4Path_14move_to},
  {"line_to", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_4Path_17line_to, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_4Path_16line_to},
  {"arc_to", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_4Path_19arc_to, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_4Path_18arc_to},
  {"quadric_to", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_4Path_21quadric_to, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_4Path_20quadric_to},
  {"cubic_to", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_4Path_23cubic_to, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_4Path_22cubic_to},
  {"arc", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_4Path_25arc, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_4Path_24arc},
  {"ellipse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_4Path_27ellipse, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_4Path_26ellipse},
  {"rect", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_4Path_29rect, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_4Path_28rect},
  {"rects", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_31rects, METH_O, __pyx_doc_7celiagg_8_celiagg_4Path_30rects},
  {"lines", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_33lines, METH_O, __pyx_doc_7celiagg_8_celiagg_4Path_32lines},
  {"lines_set", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_7celiagg_8_celiagg_4Path_35lines_set, METH_VARARGS|METH_KEYWORDS, __pyx_doc_7celiagg_8_celiagg_4Path_34lines_set},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_37__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_4Path_39__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_Path = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.Path", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_Path), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_VertexSource, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "A path object.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_7celiagg_8_celiagg_12VertexSource_3__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_Path, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_Path, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_7celiagg_8_celiagg_ShapeAtPoints(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *p;
  PyObject *o = __pyx_tp_new_7celiagg_8_celiagg_VertexSource(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *)o);
  p->_sub_source = ((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)Py_None); Py_INCREF(Py_None);
  p->_points = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg_ShapeAtPoints(PyObject *o) {
  struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *p = (struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_sub_source);
  Py_CLEAR(p->_points);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_7celiagg_8_celiagg_VertexSource(o);
}

static int __pyx_tp_traverse_7celiagg_8_celiagg_ShapeAtPoints(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *p = (struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *)o;
  e = ((likely(__pyx_ptype_7celiagg_8_celiagg_VertexSource)) ? ((__pyx_ptype_7celiagg_8_celiagg_VertexSource->tp_traverse) ? __pyx_ptype_7celiagg_8_celiagg_VertexSource->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7celiagg_8_celiagg_ShapeAtPoints)); if (e) return e;
  if (p->_sub_source) {
    e = (*v)(((PyObject *)p->_sub_source), a); if (e) return e;
  }
  if (p->_points) {
    e = (*v)(p->_points, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7celiagg_8_celiagg_ShapeAtPoints(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *p = (struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints *)o;
  if (likely(__pyx_ptype_7celiagg_8_celiagg_VertexSource)) { if (__pyx_ptype_7celiagg_8_celiagg_VertexSource->tp_clear) __pyx_ptype_7celiagg_8_celiagg_VertexSource->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7celiagg_8_celiagg_ShapeAtPoints);
  tmp = ((PyObject*)p->_sub_source);
  p->_sub_source = ((struct __pyx_obj_7celiagg_8_celiagg_VertexSource *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_points);
  p->_points = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7celiagg_8_celiagg_ShapeAtPoints[] = {
  {"copy", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_3copy, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_13ShapeAtPoints_2copy},
  {"final_point", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_5final_point, METH_NOARGS, __pyx_doc_7celiagg_8_celiagg_13ShapeAtPoints_4final_point},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_7celiagg_8_celiagg_13ShapeAtPoints_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_7celiagg_8_celiagg_ShapeAtPoints = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.ShapeAtPoints", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg_ShapeAtPoints), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg_ShapeAtPoints, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "ShapeAtPoints(source, points)\n    Replicates a shape at mutiple points.\n\n    :param source: A ``VertexSource`` object (``BSpline``, ``Path``, etc.)\n    :param points: A sequence of (x, y) pairs where the shape defined by\n                   ``source`` should be drawn.\n    ", /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg_ShapeAtPoints, /*tp_traverse*/
  __pyx_tp_clear_7celiagg_8_celiagg_ShapeAtPoints, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_7celiagg_8_celiagg_12VertexSource_3__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_7celiagg_8_celiagg_ShapeAtPoints, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg_ShapeAtPoints, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *__pyx_freelist_7celiagg_8_celiagg___pyx_scope_struct____iter__[8];
static int __pyx_freecount_7celiagg_8_celiagg___pyx_scope_struct____iter__ = 0;

static PyObject *__pyx_tp_new_7celiagg_8_celiagg___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_7celiagg_8_celiagg___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_7celiagg_8_celiagg___pyx_scope_struct____iter__[--__pyx_freecount_7celiagg_8_celiagg___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_7celiagg_8_celiagg___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *p = (struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_pt);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_7celiagg_8_celiagg___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__)))) {
    __pyx_freelist_7celiagg_8_celiagg___pyx_scope_struct____iter__[__pyx_freecount_7celiagg_8_celiagg___pyx_scope_struct____iter__++] = ((struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_7celiagg_8_celiagg___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *p = (struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_pt) {
    e = (*v)(p->__pyx_v_pt, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_7celiagg_8_celiagg___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_7celiagg_8_celiagg___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7celiagg_8_celiagg___pyx_scope_struct____iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7celiagg_8_celiagg___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7celiagg_8_celiagg___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new___Pyx_EnumMeta(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = (&PyType_Type)->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc___Pyx_EnumMeta(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  (&PyType_Type)->tp_dealloc(o);
}

static int __pyx_tp_traverse___Pyx_EnumMeta(PyObject *o, visitproc v, void *a) {
  int e;
  if (!(&PyType_Type)->tp_traverse); else { e = (&PyType_Type)->tp_traverse(o,v,a); if (e) return e; }
  return 0;
}

static int __pyx_tp_clear___Pyx_EnumMeta(PyObject *o) {
  if (!(&PyType_Type)->tp_clear); else (&PyType_Type)->tp_clear(o);
  return 0;
}
static PyObject *__pyx_sq_item___Pyx_EnumMeta(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods___Pyx_EnumMeta[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8EnumBase_14__Pyx_EnumMeta_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence___Pyx_EnumMeta = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item___Pyx_EnumMeta, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping___Pyx_EnumMeta = {
  0, /*mp_length*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_5__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __Pyx_EnumMeta = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.__Pyx_EnumMeta", /*tp_name*/
  sizeof(struct __pyx_obj___Pyx_EnumMeta), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___Pyx_EnumMeta, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence___Pyx_EnumMeta, /*tp_as_sequence*/
  &__pyx_tp_as_mapping___Pyx_EnumMeta, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse___Pyx_EnumMeta, /*tp_traverse*/
  __pyx_tp_clear___Pyx_EnumMeta, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_3__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods___Pyx_EnumMeta, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_8EnumBase_14__Pyx_EnumMeta_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___Pyx_EnumMeta, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_array___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_memoryview___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg.memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_memoryviewslice___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "celiagg._celiagg._memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Internal class for passing memoryview slices to Python", /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets__memoryviewslice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__celiagg(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__celiagg},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_celiagg",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ABGR128, __pyx_k_ABGR128, sizeof(__pyx_k_ABGR128), 0, 0, 1, 1},
  {&__pyx_n_s_ABGR32, __pyx_k_ABGR32, sizeof(__pyx_k_ABGR32), 0, 0, 1, 1},
  {&__pyx_n_s_ABGR64, __pyx_k_ABGR64, sizeof(__pyx_k_ABGR64), 0, 0, 1, 1},
  {&__pyx_n_s_ARGB128, __pyx_k_ARGB128, sizeof(__pyx_k_ARGB128), 0, 0, 1, 1},
  {&__pyx_n_s_ARGB32, __pyx_k_ARGB32, sizeof(__pyx_k_ARGB32), 0, 0, 1, 1},
  {&__pyx_n_s_ARGB64, __pyx_k_ARGB64, sizeof(__pyx_k_ARGB64), 0, 0, 1, 1},
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_n_s_AggError, __pyx_k_AggError, sizeof(__pyx_k_AggError), 0, 0, 1, 1},
  {&__pyx_kp_s_An_image_array_must_be_2_or_3_di, __pyx_k_An_image_array_must_be_2_or_3_di, sizeof(__pyx_k_An_image_array_must_be_2_or_3_di), 0, 0, 1, 0},
  {&__pyx_kp_s_An_image_must_be_provided_in_a_n, __pyx_k_An_image_must_be_provided_in_a_n, sizeof(__pyx_k_An_image_must_be_provided_in_a_n), 0, 0, 1, 0},
  {&__pyx_kp_s_Argument_must_be_a_unicode_strin, __pyx_k_Argument_must_be_a_unicode_strin, sizeof(__pyx_k_Argument_must_be_a_unicode_strin), 0, 0, 1, 0},
  {&__pyx_n_s_BGR24, __pyx_k_BGR24, sizeof(__pyx_k_BGR24), 0, 0, 1, 1},
  {&__pyx_n_s_BGR48, __pyx_k_BGR48, sizeof(__pyx_k_BGR48), 0, 0, 1, 1},
  {&__pyx_n_s_BGR96, __pyx_k_BGR96, sizeof(__pyx_k_BGR96), 0, 0, 1, 1},
  {&__pyx_n_s_BGRA128, __pyx_k_BGRA128, sizeof(__pyx_k_BGRA128), 0, 0, 1, 1},
  {&__pyx_n_s_BGRA32, __pyx_k_BGRA32, sizeof(__pyx_k_BGRA32), 0, 0, 1, 1},
  {&__pyx_n_s_BGRA64, __pyx_k_BGRA64, sizeof(__pyx_k_BGRA64), 0, 0, 1, 1},
  {&__pyx_n_s_BSpline, __pyx_k_BSpline, sizeof(__pyx_k_BSpline), 0, 0, 1, 1},
  {&__pyx_n_s_BlendAdd, __pyx_k_BlendAdd, sizeof(__pyx_k_BlendAdd), 0, 0, 1, 1},
  {&__pyx_n_s_BlendAlpha, __pyx_k_BlendAlpha, sizeof(__pyx_k_BlendAlpha), 0, 0, 1, 1},
  {&__pyx_n_s_BlendClear, __pyx_k_BlendClear, sizeof(__pyx_k_BlendClear), 0, 0, 1, 1},
  {&__pyx_n_s_BlendColorBurn, __pyx_k_BlendColorBurn, sizeof(__pyx_k_BlendColorBurn), 0, 0, 1, 1},
  {&__pyx_n_s_BlendColorDodge, __pyx_k_BlendColorDodge, sizeof(__pyx_k_BlendColorDodge), 0, 0, 1, 1},
  {&__pyx_n_s_BlendDarken, __pyx_k_BlendDarken, sizeof(__pyx_k_BlendDarken), 0, 0, 1, 1},
  {&__pyx_n_s_BlendDifference, __pyx_k_BlendDifference, sizeof(__pyx_k_BlendDifference), 0, 0, 1, 1},
  {&__pyx_n_s_BlendDst, __pyx_k_BlendDst, sizeof(__pyx_k_BlendDst), 0, 0, 1, 1},
  {&__pyx_n_s_BlendDstAtop, __pyx_k_BlendDstAtop, sizeof(__pyx_k_BlendDstAtop), 0, 0, 1, 1},
  {&__pyx_n_s_BlendDstIn, __pyx_k_BlendDstIn, sizeof(__pyx_k_BlendDstIn), 0, 0, 1, 1},
  {&__pyx_n_s_BlendDstOut, __pyx_k_BlendDstOut, sizeof(__pyx_k_BlendDstOut), 0, 0, 1, 1},
  {&__pyx_n_s_BlendDstOver, __pyx_k_BlendDstOver, sizeof(__pyx_k_BlendDstOver), 0, 0, 1, 1},
  {&__pyx_n_s_BlendExclusion, __pyx_k_BlendExclusion, sizeof(__pyx_k_BlendExclusion), 0, 0, 1, 1},
  {&__pyx_n_s_BlendHardLight, __pyx_k_BlendHardLight, sizeof(__pyx_k_BlendHardLight), 0, 0, 1, 1},
  {&__pyx_n_s_BlendLighten, __pyx_k_BlendLighten, sizeof(__pyx_k_BlendLighten), 0, 0, 1, 1},
  {&__pyx_n_s_BlendMode, __pyx_k_BlendMode, sizeof(__pyx_k_BlendMode), 0, 0, 1, 1},
  {&__pyx_n_s_BlendMultiply, __pyx_k_BlendMultiply, sizeof(__pyx_k_BlendMultiply), 0, 0, 1, 1},
  {&__pyx_n_s_BlendOverlay, __pyx_k_BlendOverlay, sizeof(__pyx_k_BlendOverlay), 0, 0, 1, 1},
  {&__pyx_n_s_BlendScreen, __pyx_k_BlendScreen, sizeof(__pyx_k_BlendScreen), 0, 0, 1, 1},
  {&__pyx_n_s_BlendSoftLight, __pyx_k_BlendSoftLight, sizeof(__pyx_k_BlendSoftLight), 0, 0, 1, 1},
  {&__pyx_n_s_BlendSrc, __pyx_k_BlendSrc, sizeof(__pyx_k_BlendSrc), 0, 0, 1, 1},
  {&__pyx_n_s_BlendSrcAtop, __pyx_k_BlendSrcAtop, sizeof(__pyx_k_BlendSrcAtop), 0, 0, 1, 1},
  {&__pyx_n_s_BlendSrcIn, __pyx_k_BlendSrcIn, sizeof(__pyx_k_BlendSrcIn), 0, 0, 1, 1},
  {&__pyx_n_s_BlendSrcOut, __pyx_k_BlendSrcOut, sizeof(__pyx_k_BlendSrcOut), 0, 0, 1, 1},
  {&__pyx_n_s_BlendSrcOver, __pyx_k_BlendSrcOver, sizeof(__pyx_k_BlendSrcOver), 0, 0, 1, 1},
  {&__pyx_n_s_BlendXor, __pyx_k_BlendXor, sizeof(__pyx_k_BlendXor), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_CanvasBGRA32, __pyx_k_CanvasBGRA32, sizeof(__pyx_k_CanvasBGRA32), 0, 0, 1, 1},
  {&__pyx_n_s_CanvasG8, __pyx_k_CanvasG8, sizeof(__pyx_k_CanvasG8), 0, 0, 1, 1},
  {&__pyx_n_s_CanvasGA16, __pyx_k_CanvasGA16, sizeof(__pyx_k_CanvasGA16), 0, 0, 1, 1},
  {&__pyx_n_s_CanvasRGB24, __pyx_k_CanvasRGB24, sizeof(__pyx_k_CanvasRGB24), 0, 0, 1, 1},
  {&__pyx_n_s_CanvasRGBA128, __pyx_k_CanvasRGBA128, sizeof(__pyx_k_CanvasRGBA128), 0, 0, 1, 1},
  {&__pyx_n_s_CanvasRGBA32, __pyx_k_CanvasRGBA32, sizeof(__pyx_k_CanvasRGBA32), 0, 0, 1, 1},
  {&__pyx_n_s_CapButt, __pyx_k_CapButt, sizeof(__pyx_k_CapButt), 0, 0, 1, 1},
  {&__pyx_n_s_CapRound, __pyx_k_CapRound, sizeof(__pyx_k_CapRound), 0, 0, 1, 1},
  {&__pyx_n_s_CapSquare, __pyx_k_CapSquare, sizeof(__pyx_k_CapSquare), 0, 0, 1, 1},
  {&__pyx_kp_s_Dashes_must_be_an_iterable_of_da, __pyx_k_Dashes_must_be_an_iterable_of_da, sizeof(__pyx_k_Dashes_must_be_an_iterable_of_da), 0, 0, 1, 0},
  {&__pyx_n_s_DrawEofFill, __pyx_k_DrawEofFill, sizeof(__pyx_k_DrawEofFill), 0, 0, 1, 1},
  {&__pyx_n_s_DrawEofFillStroke, __pyx_k_DrawEofFillStroke, sizeof(__pyx_k_DrawEofFillStroke), 0, 0, 1, 1},
  {&__pyx_n_s_DrawFill, __pyx_k_DrawFill, sizeof(__pyx_k_DrawFill), 0, 0, 1, 1},
  {&__pyx_n_s_DrawFillStroke, __pyx_k_DrawFillStroke, sizeof(__pyx_k_DrawFillStroke), 0, 0, 1, 1},
  {&__pyx_n_s_DrawStroke, __pyx_k_DrawStroke, sizeof(__pyx_k_DrawStroke), 0, 0, 1, 1},
  {&__pyx_n_s_DrawingMode, __pyx_k_DrawingMode, sizeof(__pyx_k_DrawingMode), 0, 0, 1, 1},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_n_s_EnumBase, __pyx_k_EnumBase, sizeof(__pyx_k_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_EnumType, __pyx_k_EnumType, sizeof(__pyx_k_EnumType), 0, 0, 1, 1},
  {&__pyx_n_s_Font, __pyx_k_Font, sizeof(__pyx_k_Font), 0, 0, 1, 1},
  {&__pyx_n_s_FontCacheType, __pyx_k_FontCacheType, sizeof(__pyx_k_FontCacheType), 0, 0, 1, 1},
  {&__pyx_kp_s_Font_path_must_be_a_unicode_stri, __pyx_k_Font_path_must_be_a_unicode_stri, sizeof(__pyx_k_Font_path_must_be_a_unicode_stri), 0, 0, 1, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_GetSetDescriptorType, __pyx_k_GetSetDescriptorType, sizeof(__pyx_k_GetSetDescriptorType), 0, 0, 1, 1},
  {&__pyx_n_s_GradientSpread, __pyx_k_GradientSpread, sizeof(__pyx_k_GradientSpread), 0, 0, 1, 1},
  {&__pyx_n_s_GradientUnits, __pyx_k_GradientUnits, sizeof(__pyx_k_GradientUnits), 0, 0, 1, 1},
  {&__pyx_n_s_GraphicsState, __pyx_k_GraphicsState, sizeof(__pyx_k_GraphicsState), 0, 0, 1, 1},
  {&__pyx_n_s_Gray16, __pyx_k_Gray16, sizeof(__pyx_k_Gray16), 0, 0, 1, 1},
  {&__pyx_n_s_Gray32, __pyx_k_Gray32, sizeof(__pyx_k_Gray32), 0, 0, 1, 1},
  {&__pyx_n_s_Gray8, __pyx_k_Gray8, sizeof(__pyx_k_Gray8), 0, 0, 1, 1},
  {&__pyx_n_s_HAS_TEXT, __pyx_k_HAS_TEXT, sizeof(__pyx_k_HAS_TEXT), 0, 0, 1, 1},
  {&__pyx_n_s_Image, __pyx_k_Image, sizeof(__pyx_k_Image), 0, 0, 1, 1},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_k_Incompatible_checksums_s_vs_0xb0, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xb0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_k_Incompatible_checksums_s_vs_0xd4, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd4), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_n_s_InnerBevel, __pyx_k_InnerBevel, sizeof(__pyx_k_InnerBevel), 0, 0, 1, 1},
  {&__pyx_n_s_InnerJag, __pyx_k_InnerJag, sizeof(__pyx_k_InnerJag), 0, 0, 1, 1},
  {&__pyx_n_s_InnerJoin, __pyx_k_InnerJoin, sizeof(__pyx_k_InnerJoin), 0, 0, 1, 1},
  {&__pyx_n_s_InnerMiter, __pyx_k_InnerMiter, sizeof(__pyx_k_InnerMiter), 0, 0, 1, 1},
  {&__pyx_n_s_InnerRound, __pyx_k_InnerRound, sizeof(__pyx_k_InnerRound), 0, 0, 1, 1},
  {&__pyx_n_s_IntEnum, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_JoinBevel, __pyx_k_JoinBevel, sizeof(__pyx_k_JoinBevel), 0, 0, 1, 1},
  {&__pyx_n_s_JoinMiter, __pyx_k_JoinMiter, sizeof(__pyx_k_JoinMiter), 0, 0, 1, 1},
  {&__pyx_n_s_JoinRound, __pyx_k_JoinRound, sizeof(__pyx_k_JoinRound), 0, 0, 1, 1},
  {&__pyx_n_s_LineCap, __pyx_k_LineCap, sizeof(__pyx_k_LineCap), 0, 0, 1, 1},
  {&__pyx_n_s_LineJoin, __pyx_k_LineJoin, sizeof(__pyx_k_LineJoin), 0, 0, 1, 1},
  {&__pyx_n_s_LinearGradientPaint, __pyx_k_LinearGradientPaint, sizeof(__pyx_k_LinearGradientPaint), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_n_s_MxN, __pyx_k_MxN, sizeof(__pyx_k_MxN), 0, 0, 1, 1},
  {&__pyx_kp_s_MxNx, __pyx_k_MxNx, sizeof(__pyx_k_MxNx), 0, 0, 1, 0},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_n_s_ObjectBoundingBox, __pyx_k_ObjectBoundingBox, sizeof(__pyx_k_ObjectBoundingBox), 0, 0, 1, 1},
  {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_Path, __pyx_k_Path, sizeof(__pyx_k_Path), 0, 0, 1, 1},
  {&__pyx_n_s_PatternPaint, __pyx_k_PatternPaint, sizeof(__pyx_k_PatternPaint), 0, 0, 1, 1},
  {&__pyx_n_s_PatternStyle, __pyx_k_PatternStyle, sizeof(__pyx_k_PatternStyle), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_PixelFormat, __pyx_k_PixelFormat, sizeof(__pyx_k_PixelFormat), 0, 0, 1, 1},
  {&__pyx_kp_s_Points_argument_must_be_an_itera, __pyx_k_Points_argument_must_be_an_itera, sizeof(__pyx_k_Points_argument_must_be_an_itera), 0, 0, 1, 0},
  {&__pyx_n_s_Pyx_EnumBase, __pyx_k_Pyx_EnumBase, sizeof(__pyx_k_Pyx_EnumBase), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___new, __pyx_k_Pyx_EnumBase___new, sizeof(__pyx_k_Pyx_EnumBase___new), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___repr, __pyx_k_Pyx_EnumBase___repr, sizeof(__pyx_k_Pyx_EnumBase___repr), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_EnumBase___str, __pyx_k_Pyx_EnumBase___str, sizeof(__pyx_k_Pyx_EnumBase___str), 0, 0, 1, 1},
  {&__pyx_n_s_RGB24, __pyx_k_RGB24, sizeof(__pyx_k_RGB24), 0, 0, 1, 1},
  {&__pyx_n_s_RGB48, __pyx_k_RGB48, sizeof(__pyx_k_RGB48), 0, 0, 1, 1},
  {&__pyx_n_s_RGB96, __pyx_k_RGB96, sizeof(__pyx_k_RGB96), 0, 0, 1, 1},
  {&__pyx_n_s_RGBA128, __pyx_k_RGBA128, sizeof(__pyx_k_RGBA128), 0, 0, 1, 1},
  {&__pyx_n_s_RGBA32, __pyx_k_RGBA32, sizeof(__pyx_k_RGBA32), 0, 0, 1, 1},
  {&__pyx_n_s_RGBA64, __pyx_k_RGBA64, sizeof(__pyx_k_RGBA64), 0, 0, 1, 1},
  {&__pyx_n_s_RadialGradientPaint, __pyx_k_RadialGradientPaint, sizeof(__pyx_k_RadialGradientPaint), 0, 0, 1, 1},
  {&__pyx_n_s_RasterFontCache, __pyx_k_RasterFontCache, sizeof(__pyx_k_RasterFontCache), 0, 0, 1, 1},
  {&__pyx_n_s_Rect, __pyx_k_Rect, sizeof(__pyx_k_Rect), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_ShapeAtPoints, __pyx_k_ShapeAtPoints, sizeof(__pyx_k_ShapeAtPoints), 0, 0, 1, 1},
  {&__pyx_n_s_SolidPaint, __pyx_k_SolidPaint, sizeof(__pyx_k_SolidPaint), 0, 0, 1, 1},
  {&__pyx_n_s_SpreadPad, __pyx_k_SpreadPad, sizeof(__pyx_k_SpreadPad), 0, 0, 1, 1},
  {&__pyx_n_s_SpreadReflect, __pyx_k_SpreadReflect, sizeof(__pyx_k_SpreadReflect), 0, 0, 1, 1},
  {&__pyx_n_s_SpreadRepeat, __pyx_k_SpreadRepeat, sizeof(__pyx_k_SpreadRepeat), 0, 0, 1, 1},
  {&__pyx_kp_s_Stencil_Images_must_have_format, __pyx_k_Stencil_Images_must_have_format, sizeof(__pyx_k_Stencil_Images_must_have_format), 0, 0, 1, 0},
  {&__pyx_n_s_StyleReflect, __pyx_k_StyleReflect, sizeof(__pyx_k_StyleReflect), 0, 0, 1, 1},
  {&__pyx_n_s_StyleRepeat, __pyx_k_StyleRepeat, sizeof(__pyx_k_StyleRepeat), 0, 0, 1, 1},
  {&__pyx_n_s_TextDrawClip, __pyx_k_TextDrawClip, sizeof(__pyx_k_TextDrawClip), 0, 0, 1, 1},
  {&__pyx_n_s_TextDrawFill, __pyx_k_TextDrawFill, sizeof(__pyx_k_TextDrawFill), 0, 0, 1, 1},
  {&__pyx_n_s_TextDrawFillClip, __pyx_k_TextDrawFillClip, sizeof(__pyx_k_TextDrawFillClip), 0, 0, 1, 1},
  {&__pyx_n_s_TextDrawFillStroke, __pyx_k_TextDrawFillStroke, sizeof(__pyx_k_TextDrawFillStroke), 0, 0, 1, 1},
  {&__pyx_n_s_TextDrawFillStrokeClip, __pyx_k_TextDrawFillStrokeClip, sizeof(__pyx_k_TextDrawFillStrokeClip), 0, 0, 1, 1},
  {&__pyx_n_s_TextDrawInvisible, __pyx_k_TextDrawInvisible, sizeof(__pyx_k_TextDrawInvisible), 0, 0, 1, 1},
  {&__pyx_n_s_TextDrawStroke, __pyx_k_TextDrawStroke, sizeof(__pyx_k_TextDrawStroke), 0, 0, 1, 1},
  {&__pyx_n_s_TextDrawStrokeClip, __pyx_k_TextDrawStrokeClip, sizeof(__pyx_k_TextDrawStrokeClip), 0, 0, 1, 1},
  {&__pyx_n_s_TextDrawingMode, __pyx_k_TextDrawingMode, sizeof(__pyx_k_TextDrawingMode), 0, 0, 1, 1},
  {&__pyx_kp_s_That_type_of_comparison_is_not_i, __pyx_k_That_type_of_comparison_is_not_i, sizeof(__pyx_k_That_type_of_comparison_is_not_i), 0, 0, 1, 0},
  {&__pyx_kp_s_That_type_of_comparison_is_not_i_2, __pyx_k_That_type_of_comparison_is_not_i_2, sizeof(__pyx_k_That_type_of_comparison_is_not_i_2), 0, 0, 1, 0},
  {&__pyx_kp_s_The_clip_box_property_must_be_a, __pyx_k_The_clip_box_property_must_be_a, sizeof(__pyx_k_The_clip_box_property_must_be_a), 0, 0, 1, 0},
  {&__pyx_kp_s_The_format_requires_an_array_of, __pyx_k_The_format_requires_an_array_of, sizeof(__pyx_k_The_format_requires_an_array_of), 0, 0, 1, 0},
  {&__pyx_kp_s_The_format_requires_an_array_of_2, __pyx_k_The_format_requires_an_array_of_2, sizeof(__pyx_k_The_format_requires_an_array_of_2), 0, 0, 1, 0},
  {&__pyx_kp_s_The_stencil_property_must_be_an, __pyx_k_The_stencil_property_must_be_an, sizeof(__pyx_k_The_stencil_property_must_be_an), 0, 0, 1, 0},
  {&__pyx_kp_s_The_stencil_set_on_the_GraphicsS, __pyx_k_The_stencil_set_on_the_GraphicsS, sizeof(__pyx_k_The_stencil_set_on_the_GraphicsS), 0, 0, 1, 0},
  {&__pyx_kp_s_The_text_argument_must_be_unicod, __pyx_k_The_text_argument_must_be_unicod, sizeof(__pyx_k_The_text_argument_must_be_unicod), 0, 0, 1, 0},
  {&__pyx_n_s_Transform, __pyx_k_Transform, sizeof(__pyx_k_Transform), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_kp_s_Unknown_enum_value_s, __pyx_k_Unknown_enum_value_s, sizeof(__pyx_k_Unknown_enum_value_s), 0, 0, 1, 0},
  {&__pyx_n_s_UserSpace, __pyx_k_UserSpace, sizeof(__pyx_k_UserSpace), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_VectorFontCache, __pyx_k_VectorFontCache, sizeof(__pyx_k_VectorFontCache), 0, 0, 1, 1},
  {&__pyx_n_s_VertexSource, __pyx_k_VertexSource, sizeof(__pyx_k_VertexSource), 0, 0, 1, 1},
  {&__pyx_n_s_VertexSource___iter, __pyx_k_VertexSource___iter, sizeof(__pyx_k_VertexSource___iter), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_kp_s__52, __pyx_k__52, sizeof(__pyx_k__52), 0, 0, 1, 0},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_asarray, __pyx_k_asarray, sizeof(__pyx_k_asarray), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_bottom_up, __pyx_k_bottom_up, sizeof(__pyx_k_bottom_up), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_celiagg__celiagg, __pyx_k_celiagg__celiagg, sizeof(__pyx_k_celiagg__celiagg), 0, 0, 1, 1},
  {&__pyx_kp_s_celiagg_conversion_pxi, __pyx_k_celiagg_conversion_pxi, sizeof(__pyx_k_celiagg_conversion_pxi), 0, 0, 1, 0},
  {&__pyx_n_s_ch, __pyx_k_ch, sizeof(__pyx_k_ch), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clockwise, __pyx_k_clockwise, sizeof(__pyx_k_clockwise), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_convert_image, __pyx_k_convert_image, sizeof(__pyx_k_convert_image), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_cx, __pyx_k_cx, sizeof(__pyx_k_cx), 0, 0, 1, 1},
  {&__pyx_n_s_cy, __pyx_k_cy, sizeof(__pyx_k_cy), 0, 0, 1, 1},
  {&__pyx_n_s_dct, __pyx_k_dct, sizeof(__pyx_k_dct), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dst, __pyx_k_dst, sizeof(__pyx_k_dst), 0, 0, 1, 1},
  {&__pyx_n_s_dst_alpha, __pyx_k_dst_alpha, sizeof(__pyx_k_dst_alpha), 0, 0, 1, 1},
  {&__pyx_n_s_dst_dtype, __pyx_k_dst_dtype, sizeof(__pyx_k_dst_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dst_gray, __pyx_k_dst_gray, sizeof(__pyx_k_dst_gray), 0, 0, 1, 1},
  {&__pyx_n_s_dst_order, __pyx_k_dst_order, sizeof(__pyx_k_dst_order), 0, 0, 1, 1},
  {&__pyx_n_s_dst_rgb, __pyx_k_dst_rgb, sizeof(__pyx_k_dst_rgb), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_end_angle, __pyx_k_end_angle, sizeof(__pyx_k_end_angle), 0, 0, 1, 1},
  {&__pyx_n_s_ends, __pyx_k_ends, sizeof(__pyx_k_ends), 0, 0, 1, 1},
  {&__pyx_kp_s_ends_argument_must_be_an_iterabl, __pyx_k_ends_argument_must_be_an_iterabl, sizeof(__pyx_k_ends_argument_must_be_an_iterabl), 0, 0, 1, 0},
  {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_face_index, __pyx_k_face_index, sizeof(__pyx_k_face_index), 0, 0, 1, 1},
  {&__pyx_n_s_fill, __pyx_k_fill, sizeof(__pyx_k_fill), 0, 0, 1, 1},
  {&__pyx_kp_s_fill_must_be_a_Paint_instance, __pyx_k_fill_must_be_a_Paint_instance, sizeof(__pyx_k_fill_must_be_a_Paint_instance), 0, 0, 1, 0},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_fmt, __pyx_k_fmt, sizeof(__pyx_k_fmt), 0, 0, 1, 1},
  {&__pyx_n_s_font, __pyx_k_font, sizeof(__pyx_k_font), 0, 0, 1, 1},
  {&__pyx_n_s_fontName, __pyx_k_fontName, sizeof(__pyx_k_fontName), 0, 0, 1, 1},
  {&__pyx_kp_s_font_must_be_a_Font_instance, __pyx_k_font_must_be_a_Font_instance, sizeof(__pyx_k_font_must_be_a_Font_instance), 0, 0, 1, 0},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_kp_s_format_must_be_a_PixelFormat_val, __pyx_k_format_must_be_a_PixelFormat_val, sizeof(__pyx_k_format_must_be_a_PixelFormat_val), 0, 0, 1, 0},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_from_format, __pyx_k_from_format, sizeof(__pyx_k_from_format), 0, 0, 1, 1},
  {&__pyx_n_s_fx, __pyx_k_fx, sizeof(__pyx_k_fx), 0, 0, 1, 1},
  {&__pyx_n_s_fy, __pyx_k_fy, sizeof(__pyx_k_fy), 0, 0, 1, 1},
  {&__pyx_n_s_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_h, __pyx_k_h, sizeof(__pyx_k_h), 0, 0, 1, 1},
  {&__pyx_n_s_height, __pyx_k_height, sizeof(__pyx_k_height), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_iinfo, __pyx_k_iinfo, sizeof(__pyx_k_iinfo), 0, 0, 1, 1},
  {&__pyx_n_s_image, __pyx_k_image, sizeof(__pyx_k_image), 0, 0, 1, 1},
  {&__pyx_kp_s_image_argument_must_be_channel_i, __pyx_k_image_argument_must_be_channel_i, sizeof(__pyx_k_image_argument_must_be_channel_i), 0, 0, 1, 0},
  {&__pyx_kp_s_image_argument_must_not_be_None, __pyx_k_image_argument_must_not_be_None, sizeof(__pyx_k_image_argument_must_not_be_None), 0, 0, 1, 0},
  {&__pyx_kp_s_image_must_be_an_Image_instance, __pyx_k_image_must_be_an_Image_instance, sizeof(__pyx_k_image_must_be_an_Image_instance), 0, 0, 1, 0},
  {&__pyx_kp_s_image_must_be_an_ndarray_or_Imag, __pyx_k_image_must_be_an_ndarray_or_Imag, sizeof(__pyx_k_image_must_be_an_ndarray_or_Imag), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_last_vertex, __pyx_k_last_vertex, sizeof(__pyx_k_last_vertex), 0, 0, 1, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_members, __pyx_k_members, sizeof(__pyx_k_members), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_parents, __pyx_k_parents, sizeof(__pyx_k_parents), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pixel_format, __pyx_k_pixel_format, sizeof(__pyx_k_pixel_format), 0, 0, 1, 1},
  {&__pyx_n_s_pixels, __pyx_k_pixels, sizeof(__pyx_k_pixels), 0, 0, 1, 1},
  {&__pyx_n_s_points, __pyx_k_points, sizeof(__pyx_k_points), 0, 0, 1, 1},
  {&__pyx_kp_s_points_argument_must_be_an_itera, __pyx_k_points_argument_must_be_an_itera, sizeof(__pyx_k_points_argument_must_be_an_itera), 0, 0, 1, 0},
  {&__pyx_kp_s_points_argument_must_be_an_itera_2, __pyx_k_points_argument_must_be_an_itera_2, sizeof(__pyx_k_points_argument_must_be_an_itera_2), 0, 0, 1, 0},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_k_pyx_unpickle___Pyx_EnumMeta, sizeof(__pyx_k_pyx_unpickle___Pyx_EnumMeta), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
  {&__pyx_n_s_radius, __pyx_k_radius, sizeof(__pyx_k_radius), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_kp_s_rects_argument_must_be_an_iterab, __pyx_k_rects_argument_must_be_an_iterab, sizeof(__pyx_k_rects_argument_must_be_an_iterab), 0, 0, 1, 0},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_res, __pyx_k_res, sizeof(__pyx_k_res), 0, 0, 1, 1},
  {&__pyx_n_s_rx, __pyx_k_rx, sizeof(__pyx_k_rx), 0, 0, 1, 1},
  {&__pyx_n_s_ry, __pyx_k_ry, sizeof(__pyx_k_ry), 0, 0, 1, 1},
  {&__pyx_kp_s_s_s, __pyx_k_s_s, sizeof(__pyx_k_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_d, __pyx_k_s_s_d, sizeof(__pyx_k_s_s_d), 0, 0, 1, 0},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self__this_cannot_be_converted_t, __pyx_k_self__this_cannot_be_converted_t, sizeof(__pyx_k_self__this_cannot_be_converted_t), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_kp_s_shape_must_be_a_VertexSource_Pat, __pyx_k_shape_must_be_a_VertexSource_Pat, sizeof(__pyx_k_shape_must_be_a_VertexSource_Pat), 0, 0, 1, 0},
  {&__pyx_n_s_shx, __pyx_k_shx, sizeof(__pyx_k_shx), 0, 0, 1, 1},
  {&__pyx_n_s_shy, __pyx_k_shy, sizeof(__pyx_k_shy), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
  {&__pyx_kp_s_source_must_be_a_VertexSource_in, __pyx_k_source_must_be_a_VertexSource_in, sizeof(__pyx_k_source_must_be_a_VertexSource_in), 0, 0, 1, 0},
  {&__pyx_n_s_spread, __pyx_k_spread, sizeof(__pyx_k_spread), 0, 0, 1, 1},
  {&__pyx_n_s_src, __pyx_k_src, sizeof(__pyx_k_src), 0, 0, 1, 1},
  {&__pyx_n_s_src_gray, __pyx_k_src_gray, sizeof(__pyx_k_src_gray), 0, 0, 1, 1},
  {&__pyx_kp_s_src_must_an_Image_instance, __pyx_k_src_must_an_Image_instance, sizeof(__pyx_k_src_must_an_Image_instance), 0, 0, 1, 0},
  {&__pyx_n_s_src_order, __pyx_k_src_order, sizeof(__pyx_k_src_order), 0, 0, 1, 1},
  {&__pyx_n_s_src_rgb, __pyx_k_src_rgb, sizeof(__pyx_k_src_rgb), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_start_angle, __pyx_k_start_angle, sizeof(__pyx_k_start_angle), 0, 0, 1, 1},
  {&__pyx_n_s_starts, __pyx_k_starts, sizeof(__pyx_k_starts), 0, 0, 1, 1},
  {&__pyx_kp_s_starts_argument_must_be_an_itera, __pyx_k_starts_argument_must_be_an_itera, sizeof(__pyx_k_starts_argument_must_be_an_itera), 0, 0, 1, 0},
  {&__pyx_kp_s_starts_argument_must_match_lengt, __pyx_k_starts_argument_must_match_lengt, sizeof(__pyx_k_starts_argument_must_match_lengt), 0, 0, 1, 0},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_kp_s_state_must_be_a_GraphicsState_in, __pyx_k_state_must_be_a_GraphicsState_in, sizeof(__pyx_k_state_must_be_a_GraphicsState_in), 0, 0, 1, 0},
  {&__pyx_n_s_stencil, __pyx_k_stencil, sizeof(__pyx_k_stencil), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_n_s_stops, __pyx_k_stops, sizeof(__pyx_k_stops), 0, 0, 1, 1},
  {&__pyx_kp_s_stops_argument_must_be_an_iterab, __pyx_k_stops_argument_must_be_an_iterab, sizeof(__pyx_k_stops_argument_must_be_an_iterab), 0, 0, 1, 0},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_stroke, __pyx_k_stroke, sizeof(__pyx_k_stroke), 0, 0, 1, 1},
  {&__pyx_kp_s_stroke_must_be_a_Paint_instance, __pyx_k_stroke_must_be_a_Paint_instance, sizeof(__pyx_k_stroke_must_be_a_Paint_instance), 0, 0, 1, 0},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_style, __pyx_k_style, sizeof(__pyx_k_style), 0, 0, 1, 1},
  {&__pyx_n_s_sx, __pyx_k_sx, sizeof(__pyx_k_sx), 0, 0, 1, 1},
  {&__pyx_n_s_sy, __pyx_k_sy, sizeof(__pyx_k_sy), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_text, __pyx_k_text, sizeof(__pyx_k_text), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_to_format, __pyx_k_to_format, sizeof(__pyx_k_to_format), 0, 0, 1, 1},
  {&__pyx_kp_s_to_format_must_be_a_PixelFormat, __pyx_k_to_format_must_be_a_PixelFormat, sizeof(__pyx_k_to_format_must_be_a_PixelFormat), 0, 0, 1, 0},
  {&__pyx_n_s_transform, __pyx_k_transform, sizeof(__pyx_k_transform), 0, 0, 1, 1},
  {&__pyx_kp_s_transform_must_be_a_Transform_in, __pyx_k_transform_must_be_a_Transform_in, sizeof(__pyx_k_transform_must_be_a_Transform_in), 0, 0, 1, 0},
  {&__pyx_n_s_tx, __pyx_k_tx, sizeof(__pyx_k_tx), 0, 0, 1, 1},
  {&__pyx_n_s_ty, __pyx_k_ty, sizeof(__pyx_k_ty), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_types, __pyx_k_types, sizeof(__pyx_k_types), 0, 0, 1, 1},
  {&__pyx_n_s_uint16, __pyx_k_uint16, sizeof(__pyx_k_uint16), 0, 0, 1, 1},
  {&__pyx_n_s_uint8, __pyx_k_uint8, sizeof(__pyx_k_uint8), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_units, __pyx_k_units, sizeof(__pyx_k_units), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_utf8, __pyx_k_utf8, sizeof(__pyx_k_utf8), 0, 0, 1, 1},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_vertices, __pyx_k_vertices, sizeof(__pyx_k_vertices), 0, 0, 1, 1},
  {&__pyx_n_s_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 0, 1, 1},
  {&__pyx_n_s_width, __pyx_k_width, sizeof(__pyx_k_width), 0, 0, 1, 1},
  {&__pyx_n_s_with_format, __pyx_k_with_format, sizeof(__pyx_k_with_format), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_x1, __pyx_k_x1, sizeof(__pyx_k_x1), 0, 0, 1, 1},
  {&__pyx_n_s_x2, __pyx_k_x2, sizeof(__pyx_k_x2), 0, 0, 1, 1},
  {&__pyx_n_s_x_ctrl, __pyx_k_x_ctrl, sizeof(__pyx_k_x_ctrl), 0, 0, 1, 1},
  {&__pyx_n_s_x_ctrl1, __pyx_k_x_ctrl1, sizeof(__pyx_k_x_ctrl1), 0, 0, 1, 1},
  {&__pyx_n_s_x_ctrl2, __pyx_k_x_ctrl2, sizeof(__pyx_k_x_ctrl2), 0, 0, 1, 1},
  {&__pyx_n_s_x_to, __pyx_k_x_to, sizeof(__pyx_k_x_to), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_y1, __pyx_k_y1, sizeof(__pyx_k_y1), 0, 0, 1, 1},
  {&__pyx_n_s_y2, __pyx_k_y2, sizeof(__pyx_k_y2), 0, 0, 1, 1},
  {&__pyx_n_s_y_ctrl, __pyx_k_y_ctrl, sizeof(__pyx_k_y_ctrl), 0, 0, 1, 1},
  {&__pyx_n_s_y_ctrl1, __pyx_k_y_ctrl1, sizeof(__pyx_k_y_ctrl1), 0, 0, 1, 1},
  {&__pyx_n_s_y_ctrl2, __pyx_k_y_ctrl2, sizeof(__pyx_k_y_ctrl2), 0, 0, 1, 1},
  {&__pyx_n_s_y_to, __pyx_k_y_to, sizeof(__pyx_k_y_to), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 59, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(1, 61, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(1, 253, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(1, 265, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(2, 77, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(10, 856, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(10, 1038, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(4, 148, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(4, 151, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(4, 404, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(4, 613, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(4, 832, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "celiagg/graphics_state.pxi":168
 *         def __set__(self, Rect box):
 *             if not isinstance(box, Rect):
 *                 raise TypeError("The clip_box property must be a Rect")             # <<<<<<<<<<<<<<
 * 
 *             cdef Rect rect = <Rect>box
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_The_clip_box_property_must_be_a); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "celiagg/graphics_state.pxi":253
 *         def __get__(self):
 *             cdef vector[double] dashes = self._this.line_dash_pattern()
 *             return list(zip(dashes[:-1:2], dashes[1::2]))             # <<<<<<<<<<<<<<
 * 
 *         def __set__(self, dashes):
 */
  __pyx_slice__7 = PySlice_New(Py_None, __pyx_int_neg_1, __pyx_int_2); if (unlikely(!__pyx_slice__7)) __PYX_ERR(1, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__7);
  __Pyx_GIVEREF(__pyx_slice__7);
  __pyx_slice__8 = PySlice_New(__pyx_int_1, Py_None, __pyx_int_2); if (unlikely(!__pyx_slice__8)) __PYX_ERR(1, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__8);
  __Pyx_GIVEREF(__pyx_slice__8);

  /* "celiagg/graphics_state.pxi":281
 *         def __set__(self, img):
 *             if img is not None and not isinstance(img, Image):
 *                 raise TypeError("The stencil property must be an Image")             # <<<<<<<<<<<<<<
 * 
 *             cdef Image image = <Image>img
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_The_stencil_property_must_be_an); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "celiagg/graphics_state.pxi":287
 *             if self._stencil_img is not None:
 *                 if image.format != PixelFormat.Gray8:
 *                     raise ValueError("Stencil Images must have format Gray8!")             # <<<<<<<<<<<<<<
 *                 self._this.stencil(image._this)
 *             else:
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_Stencil_Images_must_have_format); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "celiagg/image.pxi":130
 *         # Do several sanity checks...
 *         if not isinstance(array, numpy.ndarray):
 *             raise TypeError("An image must be provided in a numpy array")             # <<<<<<<<<<<<<<
 * 
 *         if array.dtype.type is not expected_dtype:
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_An_image_must_be_provided_in_a_n); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(5, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "celiagg/image.pxi":140
 * 
 *         if array.ndim not in (2, 3):
 *             raise ValueError("An image array must be 2 or 3 dimensions")             # <<<<<<<<<<<<<<
 * 
 *         last_dim = array.shape[-1] if array.ndim == 3 else 1
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_An_image_array_must_be_2_or_3_di); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(5, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "celiagg/ndarray_canvas.pxi":46
 *     cdef int base_init(self, image, int channel_count, bool has_alpha) except -1:
 *         if image is None:
 *             raise ValueError('image argument must not be None.')             # <<<<<<<<<<<<<<
 * 
 *         cdef numpy.npy_int32[:] image_shape = numpy.asarray(image.shape,
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_image_argument_must_not_be_None); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(6, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "celiagg/ndarray_canvas.pxi":126
 *         """
 *         if not isinstance(image, (numpy.ndarray, Image)):
 *             raise TypeError("image must be an ndarray or Image instance")             # <<<<<<<<<<<<<<
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):
 *             raise TypeError("format must be a PixelFormat value")
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_image_must_be_an_ndarray_or_Imag); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(6, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "celiagg/ndarray_canvas.pxi":128
 *             raise TypeError("image must be an ndarray or Image instance")
 *         if not isinstance(fmt, PixelFormat) and isinstance(image, numpy.ndarray):
 *             raise TypeError("format must be a PixelFormat value")             # <<<<<<<<<<<<<<
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_format_must_be_a_PixelFormat_val); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(6, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "celiagg/ndarray_canvas.pxi":130
 *             raise TypeError("format must be a PixelFormat value")
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")             # <<<<<<<<<<<<<<
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_transform_must_be_a_Transform_in); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(6, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "celiagg/ndarray_canvas.pxi":132
 *             raise TypeError("transform must be a Transform instance")
 *         if not isinstance(state, GraphicsState):
 *             raise TypeError("state must be a GraphicsState instance")             # <<<<<<<<<<<<<<
 * 
 *         cdef GraphicsState gs = <GraphicsState>state
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_state_must_be_a_GraphicsState_in); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(6, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "celiagg/ndarray_canvas.pxi":167
 *         """
 *         if not isinstance(shape, VertexSource):
 *             raise TypeError("shape must be a VertexSource (Path, BSpline, etc)")             # <<<<<<<<<<<<<<
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_shape_must_be_a_VertexSource_Pat); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(6, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "celiagg/ndarray_canvas.pxi":173
 *             raise TypeError("state must be a GraphicsState instance")
 *         if stroke is not None and not isinstance(stroke, Paint):
 *             raise TypeError("stroke must be a Paint instance")             # <<<<<<<<<<<<<<
 *         if fill is not None and not isinstance(fill, Paint):
 *             raise TypeError("fill must be a Paint instance")
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_stroke_must_be_a_Paint_instance); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(6, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "celiagg/ndarray_canvas.pxi":175
 *             raise TypeError("stroke must be a Paint instance")
 *         if fill is not None and not isinstance(fill, Paint):
 *             raise TypeError("fill must be a Paint instance")             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_fill_must_be_a_Paint_instance); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(6, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "celiagg/ndarray_canvas.pxi":218
 * 
 *         if not isinstance(font, Font):
 *             raise TypeError("font must be a Font instance")             # <<<<<<<<<<<<<<
 *         if not isinstance(transform, Transform):
 *             raise TypeError("transform must be a Transform instance")
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_font_must_be_a_Font_instance); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(6, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "celiagg/ndarray_canvas.pxi":281
 *         """
 *         if paint is None:
 *             return SolidPaint(0.0, 0.0, 0.0)             # <<<<<<<<<<<<<<
 * 
 *         if not hasattr(paint, '_with_format'):
 */
  __pyx_tuple__26 = PyTuple_Pack(3, __pyx_float_0_0, __pyx_float_0_0, __pyx_float_0_0); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(6, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_self__this_cannot_be_converted_t); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "(tree fragment)":4
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_self__this_cannot_be_converted_t); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_kp_s_self__this_cannot_be_converted_t); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "(tree fragment)":4
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_s_self__this_cannot_be_converted_t); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "celiagg/paint.pxi":172
 *     def __cinit__(self, PatternStyle style, image):
 *         if not isinstance(image, Image):
 *             raise TypeError("image must be an Image instance")             # <<<<<<<<<<<<<<
 * 
 *         cdef Image img = <Image>image
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_kp_s_image_must_be_an_Image_instance); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(7, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__49 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__51 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "celiagg/vertex_source.pxi":45
 *         """
 *         if self.length() == 0:
 *             return (0, 0, 0, 0)             # <<<<<<<<<<<<<<
 * 
 *         verts = self.vertices()
 */
  __pyx_tuple__55 = PyTuple_Pack(4, __pyx_int_0, __pyx_int_0, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(2, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);

  /* "celiagg/vertex_source.pxi":48
 * 
 *         verts = self.vertices()
 *         xs = verts[:, 0]             # <<<<<<<<<<<<<<
 *         ys = verts[:, 1]
 *         xmin, xmax = xs.min(), xs.max()
 */
  __pyx_slice__56 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__56)) __PYX_ERR(2, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__56);
  __Pyx_GIVEREF(__pyx_slice__56);
  __pyx_tuple__57 = PyTuple_Pack(2, __pyx_slice__56, __pyx_int_0); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(2, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "celiagg/vertex_source.pxi":49
 *         verts = self.vertices()
 *         xs = verts[:, 0]
 *         ys = verts[:, 1]             # <<<<<<<<<<<<<<
 *         xmin, xmax = xs.min(), xs.max()
 *         ymin, ymax = ys.min(), ys.max()
 */
  __pyx_tuple__58 = PyTuple_Pack(2, __pyx_slice__56, __pyx_int_1); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(2, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_kp_s_self__this_cannot_be_converted_t); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "(tree fragment)":4
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._this cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_s_self__this_cannot_be_converted_t); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__61 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__62 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__63 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__64 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "celiagg/vertex_source.pxi":376
 *     def __cinit__(self, source, points):
 *         if not isinstance(source, VertexSource):
 *             raise TypeError("source must be a VertexSource instance")             # <<<<<<<<<<<<<<
 * 
 *         cdef VertexSource vs = <VertexSource>source
 */
  __pyx_tuple__65 = PyTuple_Pack(1, __pyx_kp_s_source_must_be_a_VertexSource_in); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(2, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__66 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__67 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "celiagg/conversion.pxi":28
 *     order_map = {
 *         (0, 0, 0, 0): (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 *         (0, 1, 2, 0): (PixelFormat.RGB24, PixelFormat.RGB48, PixelFormat.RGB96),             # <<<<<<<<<<<<<<
 *         (2, 1, 0, 0): (PixelFormat.BGR24, PixelFormat.BGR48, PixelFormat.BGR96),
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),
 */
  __pyx_tuple__68 = PyTuple_Pack(4, __pyx_int_0, __pyx_int_1, __pyx_int_2, __pyx_int_0); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(9, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "celiagg/conversion.pxi":29
 *         (0, 0, 0, 0): (PixelFormat.Gray8, PixelFormat.Gray16, PixelFormat.Gray32),
 *         (0, 1, 2, 0): (PixelFormat.RGB24, PixelFormat.RGB48, PixelFormat.RGB96),
 *         (2, 1, 0, 0): (PixelFormat.BGR24, PixelFormat.BGR48, PixelFormat.BGR96),             # <<<<<<<<<<<<<<
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),
 */
  __pyx_tuple__69 = PyTuple_Pack(4, __pyx_int_2, __pyx_int_1, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(9, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);

  /* "celiagg/conversion.pxi":30
 *         (0, 1, 2, 0): (PixelFormat.RGB24, PixelFormat.RGB48, PixelFormat.RGB96),
 *         (2, 1, 0, 0): (PixelFormat.BGR24, PixelFormat.BGR48, PixelFormat.BGR96),
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),             # <<<<<<<<<<<<<<
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),
 *         (3, 0, 1, 2): (PixelFormat.ARGB32, PixelFormat.ARGB64, PixelFormat.ARGB128),
 */
  __pyx_tuple__70 = PyTuple_Pack(4, __pyx_int_0, __pyx_int_1, __pyx_int_2, __pyx_int_3); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(9, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);

  /* "celiagg/conversion.pxi":31
 *         (2, 1, 0, 0): (PixelFormat.BGR24, PixelFormat.BGR48, PixelFormat.BGR96),
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),             # <<<<<<<<<<<<<<
 *         (3, 0, 1, 2): (PixelFormat.ARGB32, PixelFormat.ARGB64, PixelFormat.ARGB128),
 *         (3, 2, 1, 0): (PixelFormat.ABGR32, PixelFormat.ABGR64, PixelFormat.ABGR128),
 */
  __pyx_tuple__71 = PyTuple_Pack(4, __pyx_int_2, __pyx_int_1, __pyx_int_0, __pyx_int_3); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(9, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "celiagg/conversion.pxi":32
 *         (0, 1, 2, 3): (PixelFormat.RGBA32, PixelFormat.RGBA64, PixelFormat.RGBA128),
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),
 *         (3, 0, 1, 2): (PixelFormat.ARGB32, PixelFormat.ARGB64, PixelFormat.ARGB128),             # <<<<<<<<<<<<<<
 *         (3, 2, 1, 0): (PixelFormat.ABGR32, PixelFormat.ABGR64, PixelFormat.ABGR128),
 *     }
 */
  __pyx_tuple__72 = PyTuple_Pack(4, __pyx_int_3, __pyx_int_0, __pyx_int_1, __pyx_int_2); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(9, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);

  /* "celiagg/conversion.pxi":33
 *         (2, 1, 0, 3): (PixelFormat.BGRA32, PixelFormat.BGRA64, PixelFormat.BGRA128),
 *         (3, 0, 1, 2): (PixelFormat.ARGB32, PixelFormat.ARGB64, PixelFormat.ARGB128),
 *         (3, 2, 1, 0): (PixelFormat.ABGR32, PixelFormat.ABGR64, PixelFormat.ABGR128),             # <<<<<<<<<<<<<<
 *     }
 *     rev_order_map = {f: o for o, fs in order_map.items() for f in fs}
 */
  __pyx_tuple__73 = PyTuple_Pack(4, __pyx_int_3, __pyx_int_2, __pyx_int_1, __pyx_int_0); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(9, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "celiagg/conversion.pxi":62
 *     """Convert gray pixels to RGB[A].
 *     """
 *     last_dim = (4,) if alpha else (3,)             # <<<<<<<<<<<<<<
 *     dst = numpy.empty(src.shape + last_dim, dtype=dtype)
 * 
 */
  __pyx_tuple__74 = PyTuple_Pack(1, __pyx_int_4); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(9, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);
  __pyx_tuple__75 = PyTuple_Pack(1, __pyx_int_3); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(9, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);

  /* "celiagg/conversion.pxi":82
 *     """
 *     last_dim = (4,) if alpha else (3,)
 *     dst = numpy.empty(src.shape[:2] + last_dim, dtype=dtype)             # <<<<<<<<<<<<<<
 * 
 *     max_from = _get_max_value(src.dtype)
 */
  __pyx_slice__76 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__76)) __PYX_ERR(9, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__76);
  __Pyx_GIVEREF(__pyx_slice__76);

  /* "celiagg/conversion.pxi":132
 *     """
 *     if not isinstance(src, Image):
 *         raise TypeError("src must an Image instance")             # <<<<<<<<<<<<<<
 *     if not isinstance(to_format, (int, PixelFormat)):
 *         raise TypeError("to_format must be a PixelFormat value")
 */
  __pyx_tuple__77 = PyTuple_Pack(1, __pyx_kp_s_src_must_an_Image_instance); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(9, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);

  /* "celiagg/conversion.pxi":134
 *         raise TypeError("src must an Image instance")
 *     if not isinstance(to_format, (int, PixelFormat)):
 *         raise TypeError("to_format must be a PixelFormat value")             # <<<<<<<<<<<<<<
 * 
 *     to_format = PixelFormat(to_format)
 */
  __pyx_tuple__78 = PyTuple_Pack(1, __pyx_kp_s_to_format_must_be_a_PixelFormat); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(9, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__79 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(10, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__80 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(10, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__81 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(10, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":856
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__82 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(10, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":880
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__83 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(10, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1038
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__84 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(10, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);

  /* "../../../.edm/envs/edm/lib/python3.6/site-packages/Cython/Includes/numpy/__init__.pxd":1044
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__85 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(10, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
  __pyx_tuple__86 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(4, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
  __pyx_tuple__87 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(4, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);

  /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__88 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(4, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);

  /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
  __pyx_tuple__89 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(4, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);

  /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
  __pyx_tuple__90 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(4, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__91 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__92 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);

  /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
  __pyx_tuple__93 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(4, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);

  /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
  __pyx_tuple__94 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(4, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);

  /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
  __pyx_tuple__95 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(4, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);

  /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
  __pyx_tuple__96 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(4, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);

  /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__97 = PyTuple_New(1); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(4, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__97, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__97);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__98 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__99 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);

  /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__100 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(4, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__101 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(4, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__102 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(4, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);

  /* "celiagg/conversion.pxi":123
 * 
 * 
 * def convert_image(src, to_format, bottom_up=False):             # <<<<<<<<<<<<<<
 *     """convert_image(src, to_format, bottom_up=False)
 *     Create a new image with a desired pixel format and orientation.
 */
  __pyx_tuple__103 = PyTuple_Pack(13, __pyx_n_s_src, __pyx_n_s_to_format, __pyx_n_s_bottom_up, __pyx_n_s_from_format, __pyx_n_s_dst_dtype, __pyx_n_s_dst_alpha, __pyx_n_s_src_gray, __pyx_n_s_dst_gray, __pyx_n_s_src_rgb, __pyx_n_s_dst_rgb, __pyx_n_s_src_order, __pyx_n_s_dst_order, __pyx_n_s_dst); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(9, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(3, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__103, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_celiagg_conversion_pxi, __pyx_n_s_convert_image, 123, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(9, 123, __pyx_L1_error)

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */
  __pyx_tuple__105 = PyTuple_Pack(5, __pyx_n_s_cls, __pyx_n_s_value, __pyx_n_s_name_2, __pyx_n_s_v, __pyx_n_s_res); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(4, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__105, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_new, 28, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(4, 28, __pyx_L1_error)
  __pyx_tuple__107 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(4, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */
  __pyx_tuple__108 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__108)) __PYX_ERR(4, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__108);
  __Pyx_GIVEREF(__pyx_tuple__108);
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__108, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_repr, 39, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(4, 39, __pyx_L1_error)

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */
  __pyx_tuple__110 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(4, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__110, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_str, 41, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(4, 41, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__112 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__112, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(4, 1, __pyx_L1_error)

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__114 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(4, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__115 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__115)) __PYX_ERR(4, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__116 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__116)) __PYX_ERR(4, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__116);
  __Pyx_GIVEREF(__pyx_tuple__116);

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__117 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(4, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__118 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(4, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__119 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);
  __pyx_codeobj__120 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__120)) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyList_Type_remove.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_3_0 = PyFloat_FromDouble(3.0); if (unlikely(!__pyx_float_3_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_OrderedDict = Py_None; Py_INCREF(Py_None);
  __Pyx_EnumBase = Py_None; Py_INCREF(Py_None);
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_globals = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_Font) < 0) __PYX_ERR(3, 32, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_Font.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_Font.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_Font.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_Font.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Font, (PyObject *)&__pyx_type_7celiagg_8_celiagg_Font) < 0) __PYX_ERR(3, 32, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_Font) < 0) __PYX_ERR(3, 32, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_Font = &__pyx_type_7celiagg_8_celiagg_Font;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_Rect) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_Rect.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_Rect.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_Rect.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_Rect.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Rect, (PyObject *)&__pyx_type_7celiagg_8_celiagg_Rect) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_Rect) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_Rect = &__pyx_type_7celiagg_8_celiagg_Rect;
  __pyx_vtabptr_7celiagg_8_celiagg_GraphicsState = &__pyx_vtable_7celiagg_8_celiagg_GraphicsState;
  __pyx_vtable_7celiagg_8_celiagg_GraphicsState._propnames = (PyObject *(*)(struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *))__pyx_f_7celiagg_8_celiagg_13GraphicsState__propnames;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_GraphicsState) < 0) __PYX_ERR(1, 92, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_GraphicsState.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_GraphicsState.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_GraphicsState.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_GraphicsState.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7celiagg_8_celiagg_GraphicsState.tp_dict, __pyx_vtabptr_7celiagg_8_celiagg_GraphicsState) < 0) __PYX_ERR(1, 92, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GraphicsState, (PyObject *)&__pyx_type_7celiagg_8_celiagg_GraphicsState) < 0) __PYX_ERR(1, 92, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_GraphicsState) < 0) __PYX_ERR(1, 92, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_GraphicsState = &__pyx_type_7celiagg_8_celiagg_GraphicsState;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_Image) < 0) __PYX_ERR(5, 112, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_Image.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_Image.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_Image.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_Image.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Image, (PyObject *)&__pyx_type_7celiagg_8_celiagg_Image) < 0) __PYX_ERR(5, 112, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_Image) < 0) __PYX_ERR(5, 112, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_Image = &__pyx_type_7celiagg_8_celiagg_Image;
  __pyx_vtabptr_7celiagg_8_celiagg_CanvasBase = &__pyx_vtable_7celiagg_8_celiagg_CanvasBase;
  __pyx_vtable_7celiagg_8_celiagg_CanvasBase.base_init = (int (*)(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *, PyObject *, int, bool))__pyx_f_7celiagg_8_celiagg_10CanvasBase_base_init;
  __pyx_vtable_7celiagg_8_celiagg_CanvasBase._check_stencil = (PyObject *(*)(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *, struct __pyx_obj_7celiagg_8_celiagg_GraphicsState *))__pyx_f_7celiagg_8_celiagg_10CanvasBase__check_stencil;
  __pyx_vtable_7celiagg_8_celiagg_CanvasBase._get_native_image = (struct __pyx_obj_7celiagg_8_celiagg_Image *(*)(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *, struct __pyx_obj_7celiagg_8_celiagg_Image *, enum __pyx_t_7celiagg_8_celiagg_PixelFormat))__pyx_f_7celiagg_8_celiagg_10CanvasBase__get_native_image;
  __pyx_vtable_7celiagg_8_celiagg_CanvasBase._get_native_paint = (struct __pyx_obj_7celiagg_8_celiagg_Paint *(*)(struct __pyx_obj_7celiagg_8_celiagg_CanvasBase *, PyObject *, enum __pyx_t_7celiagg_8_celiagg_PixelFormat))__pyx_f_7celiagg_8_celiagg_10CanvasBase__get_native_paint;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_CanvasBase) < 0) __PYX_ERR(6, 38, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_CanvasBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_CanvasBase.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_CanvasBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_CanvasBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7celiagg_8_celiagg_CanvasBase.tp_dict, __pyx_vtabptr_7celiagg_8_celiagg_CanvasBase) < 0) __PYX_ERR(6, 38, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_CanvasBase) < 0) __PYX_ERR(6, 38, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_CanvasBase = &__pyx_type_7celiagg_8_celiagg_CanvasBase;
  __pyx_vtabptr_7celiagg_8_celiagg_CanvasRGBA128 = &__pyx_vtable_7celiagg_8_celiagg_CanvasRGBA128;
  __pyx_vtable_7celiagg_8_celiagg_CanvasRGBA128.__pyx_base = *__pyx_vtabptr_7celiagg_8_celiagg_CanvasBase;
  __pyx_type_7celiagg_8_celiagg_CanvasRGBA128.tp_base = __pyx_ptype_7celiagg_8_celiagg_CanvasBase;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_CanvasRGBA128) < 0) __PYX_ERR(6, 289, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_CanvasRGBA128.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_CanvasRGBA128.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_CanvasRGBA128.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_CanvasRGBA128.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7celiagg_8_celiagg_CanvasRGBA128.tp_dict, __pyx_vtabptr_7celiagg_8_celiagg_CanvasRGBA128) < 0) __PYX_ERR(6, 289, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CanvasRGBA128, (PyObject *)&__pyx_type_7celiagg_8_celiagg_CanvasRGBA128) < 0) __PYX_ERR(6, 289, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_CanvasRGBA128) < 0) __PYX_ERR(6, 289, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_CanvasRGBA128 = &__pyx_type_7celiagg_8_celiagg_CanvasRGBA128;
  __pyx_vtabptr_7celiagg_8_celiagg_CanvasBGRA32 = &__pyx_vtable_7celiagg_8_celiagg_CanvasBGRA32;
  __pyx_vtable_7celiagg_8_celiagg_CanvasBGRA32.__pyx_base = *__pyx_vtabptr_7celiagg_8_celiagg_CanvasBase;
  __pyx_type_7celiagg_8_celiagg_CanvasBGRA32.tp_base = __pyx_ptype_7celiagg_8_celiagg_CanvasBase;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_CanvasBGRA32) < 0) __PYX_ERR(6, 310, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_CanvasBGRA32.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_CanvasBGRA32.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_CanvasBGRA32.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_CanvasBGRA32.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7celiagg_8_celiagg_CanvasBGRA32.tp_dict, __pyx_vtabptr_7celiagg_8_celiagg_CanvasBGRA32) < 0) __PYX_ERR(6, 310, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CanvasBGRA32, (PyObject *)&__pyx_type_7celiagg_8_celiagg_CanvasBGRA32) < 0) __PYX_ERR(6, 310, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_CanvasBGRA32) < 0) __PYX_ERR(6, 310, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_CanvasBGRA32 = &__pyx_type_7celiagg_8_celiagg_CanvasBGRA32;
  __pyx_vtabptr_7celiagg_8_celiagg_CanvasRGBA32 = &__pyx_vtable_7celiagg_8_celiagg_CanvasRGBA32;
  __pyx_vtable_7celiagg_8_celiagg_CanvasRGBA32.__pyx_base = *__pyx_vtabptr_7celiagg_8_celiagg_CanvasBase;
  __pyx_type_7celiagg_8_celiagg_CanvasRGBA32.tp_base = __pyx_ptype_7celiagg_8_celiagg_CanvasBase;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_CanvasRGBA32) < 0) __PYX_ERR(6, 331, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_CanvasRGBA32.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_CanvasRGBA32.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_CanvasRGBA32.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_CanvasRGBA32.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7celiagg_8_celiagg_CanvasRGBA32.tp_dict, __pyx_vtabptr_7celiagg_8_celiagg_CanvasRGBA32) < 0) __PYX_ERR(6, 331, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CanvasRGBA32, (PyObject *)&__pyx_type_7celiagg_8_celiagg_CanvasRGBA32) < 0) __PYX_ERR(6, 331, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_CanvasRGBA32) < 0) __PYX_ERR(6, 331, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_CanvasRGBA32 = &__pyx_type_7celiagg_8_celiagg_CanvasRGBA32;
  __pyx_vtabptr_7celiagg_8_celiagg_CanvasRGB24 = &__pyx_vtable_7celiagg_8_celiagg_CanvasRGB24;
  __pyx_vtable_7celiagg_8_celiagg_CanvasRGB24.__pyx_base = *__pyx_vtabptr_7celiagg_8_celiagg_CanvasBase;
  __pyx_type_7celiagg_8_celiagg_CanvasRGB24.tp_base = __pyx_ptype_7celiagg_8_celiagg_CanvasBase;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_CanvasRGB24) < 0) __PYX_ERR(6, 352, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_CanvasRGB24.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_CanvasRGB24.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_CanvasRGB24.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_CanvasRGB24.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7celiagg_8_celiagg_CanvasRGB24.tp_dict, __pyx_vtabptr_7celiagg_8_celiagg_CanvasRGB24) < 0) __PYX_ERR(6, 352, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CanvasRGB24, (PyObject *)&__pyx_type_7celiagg_8_celiagg_CanvasRGB24) < 0) __PYX_ERR(6, 352, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_CanvasRGB24) < 0) __PYX_ERR(6, 352, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_CanvasRGB24 = &__pyx_type_7celiagg_8_celiagg_CanvasRGB24;
  __pyx_vtabptr_7celiagg_8_celiagg_CanvasGA16 = &__pyx_vtable_7celiagg_8_celiagg_CanvasGA16;
  __pyx_vtable_7celiagg_8_celiagg_CanvasGA16.__pyx_base = *__pyx_vtabptr_7celiagg_8_celiagg_CanvasBase;
  __pyx_type_7celiagg_8_celiagg_CanvasGA16.tp_base = __pyx_ptype_7celiagg_8_celiagg_CanvasBase;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_CanvasGA16) < 0) __PYX_ERR(6, 373, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_CanvasGA16.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_CanvasGA16.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_CanvasGA16.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_CanvasGA16.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7celiagg_8_celiagg_CanvasGA16.tp_dict, __pyx_vtabptr_7celiagg_8_celiagg_CanvasGA16) < 0) __PYX_ERR(6, 373, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CanvasGA16, (PyObject *)&__pyx_type_7celiagg_8_celiagg_CanvasGA16) < 0) __PYX_ERR(6, 373, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_CanvasGA16) < 0) __PYX_ERR(6, 373, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_CanvasGA16 = &__pyx_type_7celiagg_8_celiagg_CanvasGA16;
  __pyx_vtabptr_7celiagg_8_celiagg_CanvasG8 = &__pyx_vtable_7celiagg_8_celiagg_CanvasG8;
  __pyx_vtable_7celiagg_8_celiagg_CanvasG8.__pyx_base = *__pyx_vtabptr_7celiagg_8_celiagg_CanvasBase;
  __pyx_type_7celiagg_8_celiagg_CanvasG8.tp_base = __pyx_ptype_7celiagg_8_celiagg_CanvasBase;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_CanvasG8) < 0) __PYX_ERR(6, 394, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_CanvasG8.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_CanvasG8.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_CanvasG8.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_CanvasG8.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7celiagg_8_celiagg_CanvasG8.tp_dict, __pyx_vtabptr_7celiagg_8_celiagg_CanvasG8) < 0) __PYX_ERR(6, 394, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CanvasG8, (PyObject *)&__pyx_type_7celiagg_8_celiagg_CanvasG8) < 0) __PYX_ERR(6, 394, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_CanvasG8) < 0) __PYX_ERR(6, 394, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_CanvasG8 = &__pyx_type_7celiagg_8_celiagg_CanvasG8;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_Paint) < 0) __PYX_ERR(7, 63, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_Paint.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_Paint.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_Paint.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_Paint.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_Paint) < 0) __PYX_ERR(7, 63, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_Paint = &__pyx_type_7celiagg_8_celiagg_Paint;
  __pyx_type_7celiagg_8_celiagg_LinearGradientPaint.tp_base = __pyx_ptype_7celiagg_8_celiagg_Paint;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_LinearGradientPaint) < 0) __PYX_ERR(7, 81, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_LinearGradientPaint.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_LinearGradientPaint.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_LinearGradientPaint.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_LinearGradientPaint.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_LinearGradientPaint, (PyObject *)&__pyx_type_7celiagg_8_celiagg_LinearGradientPaint) < 0) __PYX_ERR(7, 81, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_LinearGradientPaint) < 0) __PYX_ERR(7, 81, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_LinearGradientPaint = &__pyx_type_7celiagg_8_celiagg_LinearGradientPaint;
  __pyx_type_7celiagg_8_celiagg_RadialGradientPaint.tp_base = __pyx_ptype_7celiagg_8_celiagg_Paint;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_RadialGradientPaint) < 0) __PYX_ERR(7, 120, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_RadialGradientPaint.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_RadialGradientPaint.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_RadialGradientPaint.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_RadialGradientPaint.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_RadialGradientPaint, (PyObject *)&__pyx_type_7celiagg_8_celiagg_RadialGradientPaint) < 0) __PYX_ERR(7, 120, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_RadialGradientPaint) < 0) __PYX_ERR(7, 120, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_RadialGradientPaint = &__pyx_type_7celiagg_8_celiagg_RadialGradientPaint;
  __pyx_type_7celiagg_8_celiagg_PatternPaint.tp_base = __pyx_ptype_7celiagg_8_celiagg_Paint;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_PatternPaint) < 0) __PYX_ERR(7, 160, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_PatternPaint.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_PatternPaint.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_PatternPaint.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_PatternPaint.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PatternPaint, (PyObject *)&__pyx_type_7celiagg_8_celiagg_PatternPaint) < 0) __PYX_ERR(7, 160, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_PatternPaint) < 0) __PYX_ERR(7, 160, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_PatternPaint = &__pyx_type_7celiagg_8_celiagg_PatternPaint;
  __pyx_type_7celiagg_8_celiagg_SolidPaint.tp_base = __pyx_ptype_7celiagg_8_celiagg_Paint;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_SolidPaint) < 0) __PYX_ERR(7, 198, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_SolidPaint.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_SolidPaint.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_SolidPaint.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_SolidPaint.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SolidPaint, (PyObject *)&__pyx_type_7celiagg_8_celiagg_SolidPaint) < 0) __PYX_ERR(7, 198, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_SolidPaint) < 0) __PYX_ERR(7, 198, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_SolidPaint = &__pyx_type_7celiagg_8_celiagg_SolidPaint;
  __pyx_vtabptr_7celiagg_8_celiagg_Transform = &__pyx_vtable_7celiagg_8_celiagg_Transform;
  __pyx_vtable_7celiagg_8_celiagg_Transform._assign_obj = (PyObject *(*)(struct __pyx_obj_7celiagg_8_celiagg_Transform *, agg::trans_affine))__pyx_f_7celiagg_8_celiagg_9Transform__assign_obj;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_Transform) < 0) __PYX_ERR(8, 26, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_Transform.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_Transform.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_Transform.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_Transform.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_7celiagg_8_celiagg_Transform.tp_dict, __pyx_vtabptr_7celiagg_8_celiagg_Transform) < 0) __PYX_ERR(8, 26, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Transform, (PyObject *)&__pyx_type_7celiagg_8_celiagg_Transform) < 0) __PYX_ERR(8, 26, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_Transform) < 0) __PYX_ERR(8, 26, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_Transform = &__pyx_type_7celiagg_8_celiagg_Transform;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_VertexSource) < 0) __PYX_ERR(2, 25, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_VertexSource.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_VertexSource.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_VertexSource.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_VertexSource.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_VertexSource, (PyObject *)&__pyx_type_7celiagg_8_celiagg_VertexSource) < 0) __PYX_ERR(2, 25, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_VertexSource) < 0) __PYX_ERR(2, 25, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_VertexSource = &__pyx_type_7celiagg_8_celiagg_VertexSource;
  __pyx_type_7celiagg_8_celiagg_BSpline.tp_base = __pyx_ptype_7celiagg_8_celiagg_VertexSource;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_BSpline) < 0) __PYX_ERR(2, 83, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_BSpline.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_BSpline.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_BSpline.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_BSpline.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_BSpline, (PyObject *)&__pyx_type_7celiagg_8_celiagg_BSpline) < 0) __PYX_ERR(2, 83, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_BSpline) < 0) __PYX_ERR(2, 83, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_BSpline = &__pyx_type_7celiagg_8_celiagg_BSpline;
  __pyx_type_7celiagg_8_celiagg_Path.tp_base = __pyx_ptype_7celiagg_8_celiagg_VertexSource;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_Path) < 0) __PYX_ERR(2, 117, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_Path.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_Path.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_Path.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_Path.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Path, (PyObject *)&__pyx_type_7celiagg_8_celiagg_Path) < 0) __PYX_ERR(2, 117, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_Path) < 0) __PYX_ERR(2, 117, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_Path = &__pyx_type_7celiagg_8_celiagg_Path;
  __pyx_type_7celiagg_8_celiagg_ShapeAtPoints.tp_base = __pyx_ptype_7celiagg_8_celiagg_VertexSource;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg_ShapeAtPoints) < 0) __PYX_ERR(2, 363, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg_ShapeAtPoints.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg_ShapeAtPoints.tp_dictoffset && __pyx_type_7celiagg_8_celiagg_ShapeAtPoints.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg_ShapeAtPoints.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ShapeAtPoints, (PyObject *)&__pyx_type_7celiagg_8_celiagg_ShapeAtPoints) < 0) __PYX_ERR(2, 363, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_7celiagg_8_celiagg_ShapeAtPoints) < 0) __PYX_ERR(2, 363, __pyx_L1_error)
  __pyx_ptype_7celiagg_8_celiagg_ShapeAtPoints = &__pyx_type_7celiagg_8_celiagg_ShapeAtPoints;
  if (PyType_Ready(&__pyx_type_7celiagg_8_celiagg___pyx_scope_struct____iter__) < 0) __PYX_ERR(2, 36, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_7celiagg_8_celiagg___pyx_scope_struct____iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_7celiagg_8_celiagg___pyx_scope_struct____iter__.tp_dictoffset && __pyx_type_7celiagg_8_celiagg___pyx_scope_struct____iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_7celiagg_8_celiagg___pyx_scope_struct____iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_7celiagg_8_celiagg___pyx_scope_struct____iter__ = &__pyx_type_7celiagg_8_celiagg___pyx_scope_struct____iter__;
  __Pyx_EnumMeta.tp_base = (&PyType_Type);
  if (PyType_Ready(&__Pyx_EnumMeta) < 0) __PYX_ERR(4, 15, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __Pyx_EnumMeta.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__Pyx_EnumMeta.tp_dictoffset && __Pyx_EnumMeta.tp_getattro == PyObject_GenericGetAttr)) {
    __Pyx_EnumMeta.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__Pyx_EnumMeta) < 0) __PYX_ERR(4, 15, __pyx_L1_error)
  __pyx_ptype___Pyx_EnumMeta = &__Pyx_EnumMeta;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(4, 105, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(4, 105, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(4, 105, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(4, 279, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(4, 279, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(4, 330, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(4, 330, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(4, 330, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(4, 965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(4, 965, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(4, 965, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(11, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(10, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(10, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(10, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(10, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(10, 918, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_celiagg(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_celiagg(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__celiagg(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__celiagg(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__celiagg(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  static PyThread_type_lock __pyx_t_6[8];
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_celiagg' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__celiagg(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_celiagg", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_celiagg___celiagg) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "celiagg._celiagg")) {
      if (unlikely(PyDict_SetItemString(modules, "celiagg._celiagg", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "celiagg/_celiagg.pyx":30
 * from cython.operator cimport dereference
 * cimport numpy
 * import numpy             # <<<<<<<<<<<<<<
 * 
 * cimport _enums
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_numpy, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/_celiagg.pyx":43
 * # This is a global flag which tells you if text rendering support was compiled
 * IF _ENABLE_TEXT_RENDERING:
 *     HAS_TEXT = True             # <<<<<<<<<<<<<<
 * ELSE:
 *     HAS_TEXT = False
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAS_TEXT, Py_True) < 0) __PYX_ERR(0, 43, __pyx_L1_error)

  /* "celiagg/_celiagg.pyx":48
 * 
 * 
 * class AggError(Exception):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __Pyx_GIVEREF(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_AggError, __pyx_n_s_AggError, (PyObject *) NULL, __pyx_n_s_celiagg__celiagg, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_AggError, __pyx_t_1, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AggError, __pyx_t_4) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "celiagg/graphics_state.pxi":25
 * # Authors: John Wiggins
 * 
 * from types import GetSetDescriptorType             # <<<<<<<<<<<<<<
 * from libcpp.vector cimport vector
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_GetSetDescriptorType);
  __Pyx_GIVEREF(__pyx_n_s_GetSetDescriptorType);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_GetSetDescriptorType);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_types, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_GetSetDescriptorType); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GetSetDescriptorType, __pyx_t_1) < 0) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "celiagg/conversion.pxi":123
 * 
 * 
 * def convert_image(src, to_format, bottom_up=False):             # <<<<<<<<<<<<<<
 *     """convert_image(src, to_format, bottom_up=False)
 *     Create a new image with a desired pixel format and orientation.
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_7celiagg_8_celiagg_1convert_image, NULL, __pyx_n_s_celiagg__celiagg); if (unlikely(!__pyx_t_2)) __PYX_ERR(9, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_convert_image, __pyx_t_2) < 0) __PYX_ERR(9, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "celiagg/_celiagg.pyx":1
 * # The MIT License (MIT)             # <<<<<<<<<<<<<<
 * #
 * # Copyright (c) 2014 WUSTL ZPLAB
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumBase":9
 * 
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x02070000) != 0);
  if (__pyx_t_5) {

    /* "EnumBase":10
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:
 *     from collections import OrderedDict as __Pyx_OrderedDict             # <<<<<<<<<<<<<<
 * else:
 *     __Pyx_OrderedDict = dict
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_OrderedDict);
    __Pyx_GIVEREF(__pyx_n_s_OrderedDict);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_OrderedDict);
    __pyx_t_1 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 10, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XGOTREF(__Pyx_OrderedDict);
    __Pyx_DECREF_SET(__Pyx_OrderedDict, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "EnumBase":9
 * 
 * cdef object __Pyx_OrderedDict
 * if PY_VERSION_HEX >= 0x02070000:             # <<<<<<<<<<<<<<
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 */
    goto __pyx_L2;
  }

  /* "EnumBase":12
 *     from collections import OrderedDict as __Pyx_OrderedDict
 * else:
 *     __Pyx_OrderedDict = dict             # <<<<<<<<<<<<<<
 * 
 * @cython.internal
 */
  /*else*/ {
    __Pyx_INCREF(((PyObject *)(&PyDict_Type)));
    __Pyx_XGOTREF(__Pyx_OrderedDict);
    __Pyx_DECREF_SET(__Pyx_OrderedDict, ((PyObject *)(&PyDict_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyDict_Type)));
  }
  __pyx_L2:;

  /* "EnumBase":26
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):             # <<<<<<<<<<<<<<
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)(&PyInt_Type)));
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_Pyx_EnumBase, __pyx_n_s_Pyx_EnumBase, (PyObject *) NULL, __pyx_n_s_EnumBase, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "EnumBase":27
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta             # <<<<<<<<<<<<<<
 *     def __new__(cls, value, name=None):
 *         for v in cls:
 */
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_metaclass, ((PyObject *)__pyx_ptype___Pyx_EnumMeta)) < 0) __PYX_ERR(4, 27, __pyx_L1_error)

  /* "EnumBase":28
 * class __Pyx_EnumBase(int):
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):             # <<<<<<<<<<<<<<
 *         for v in cls:
 *             if v == value:
 */
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_1__new__, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_Pyx_EnumBase___new, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__107);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_new, __pyx_t_4) < 0) __PYX_ERR(4, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumBase":39
 *         cls.__members__[name] = res
 *         return res
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):
 */
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_3__repr__, 0, __pyx_n_s_Pyx_EnumBase___repr, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_repr, __pyx_t_4) < 0) __PYX_ERR(4, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumBase":41
 *     def __repr__(self):
 *         return "<%s.%s: %d>" % (self.__class__.__name__, self.name, self)
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_8EnumBase_14__Pyx_EnumBase_5__str__, 0, __pyx_n_s_Pyx_EnumBase___str, NULL, __pyx_n_s_EnumBase, __pyx_d, ((PyObject *)__pyx_codeobj__111)); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_str, __pyx_t_4) < 0) __PYX_ERR(4, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "EnumBase":26
 * 
 * cdef object __Pyx_EnumBase
 * class __Pyx_EnumBase(int):             # <<<<<<<<<<<<<<
 *     __metaclass__ = __Pyx_EnumMeta
 *     def __new__(cls, value, name=None):
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_Pyx_EnumBase, __pyx_t_1, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_XGOTREF(__Pyx_EnumBase);
  __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "EnumBase":44
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumBase":45
 * 
 * if PY_VERSION_HEX >= 0x03040000:
 *     from enum import IntEnum as __Pyx_EnumBase             # <<<<<<<<<<<<<<
 * 
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_IntEnum);
    __Pyx_GIVEREF(__pyx_n_s_IntEnum);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_IntEnum);
    __pyx_t_2 = __Pyx_Import(__pyx_n_s_enum, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XGOTREF(__Pyx_EnumBase);
    __Pyx_DECREF_SET(__Pyx_EnumBase, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumBase":44
 *         return "%s.%s" % (self.__class__.__name__, self.name)
 * 
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 *     from enum import IntEnum as __Pyx_EnumBase
 * 
 */
  }

  /* "(tree fragment)":1
 * def __pyx_unpickle___Pyx_EnumMeta(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_8EnumBase_1__pyx_unpickle___Pyx_EnumMeta, NULL, __pyx_n_s_EnumBase); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle___Pyx_EnumMeta, __pyx_t_2) < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":209
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(4, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__114, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__115, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__116, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__117, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__118, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":316
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":317
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_6[0] = PyThread_allocate_lock();
  __pyx_t_6[1] = PyThread_allocate_lock();
  __pyx_t_6[2] = PyThread_allocate_lock();
  __pyx_t_6[3] = PyThread_allocate_lock();
  __pyx_t_6[4] = PyThread_allocate_lock();
  __pyx_t_6[5] = PyThread_allocate_lock();
  __pyx_t_6[6] = PyThread_allocate_lock();
  __pyx_t_6[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_6, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":549
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(4, 549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":995
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(4, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_2) < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyDict_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     BlendMode = __Pyx_EnumBase('BlendMode', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     BlendMode = __Pyx_EnumBase('BlendMode', __Pyx_OrderedDict([
 *         ('BlendAlpha', BlendAlpha),             # <<<<<<<<<<<<<<
 *         ('BlendClear', BlendClear),
 *         ('BlendSrc', BlendSrc),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendAlpha); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_BlendAlpha);
    __Pyx_GIVEREF(__pyx_n_s_BlendAlpha);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_BlendAlpha);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":55
 *     BlendMode = __Pyx_EnumBase('BlendMode', __Pyx_OrderedDict([
 *         ('BlendAlpha', BlendAlpha),
 *         ('BlendClear', BlendClear),             # <<<<<<<<<<<<<<
 *         ('BlendSrc', BlendSrc),
 *         ('BlendDst', BlendDst),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendClear); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_n_s_BlendClear);
    __Pyx_GIVEREF(__pyx_n_s_BlendClear);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_BlendClear);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":56
 *         ('BlendAlpha', BlendAlpha),
 *         ('BlendClear', BlendClear),
 *         ('BlendSrc', BlendSrc),             # <<<<<<<<<<<<<<
 *         ('BlendDst', BlendDst),
 *         ('BlendSrcOver', BlendSrcOver),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrc); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_BlendSrc);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrc);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_n_s_BlendSrc);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":57
 *         ('BlendClear', BlendClear),
 *         ('BlendSrc', BlendSrc),
 *         ('BlendDst', BlendDst),             # <<<<<<<<<<<<<<
 *         ('BlendSrcOver', BlendSrcOver),
 *         ('BlendDstOver', BlendDstOver),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDst); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_n_s_BlendDst);
    __Pyx_GIVEREF(__pyx_n_s_BlendDst);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_BlendDst);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":58
 *         ('BlendSrc', BlendSrc),
 *         ('BlendDst', BlendDst),
 *         ('BlendSrcOver', BlendSrcOver),             # <<<<<<<<<<<<<<
 *         ('BlendDstOver', BlendDstOver),
 *         ('BlendSrcIn', BlendSrcIn),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrcOver); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_BlendSrcOver);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrcOver);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_BlendSrcOver);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":59
 *         ('BlendDst', BlendDst),
 *         ('BlendSrcOver', BlendSrcOver),
 *         ('BlendDstOver', BlendDstOver),             # <<<<<<<<<<<<<<
 *         ('BlendSrcIn', BlendSrcIn),
 *         ('BlendDstIn', BlendDstIn),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDstOver); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_BlendDstOver);
    __Pyx_GIVEREF(__pyx_n_s_BlendDstOver);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_BlendDstOver);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":60
 *         ('BlendSrcOver', BlendSrcOver),
 *         ('BlendDstOver', BlendDstOver),
 *         ('BlendSrcIn', BlendSrcIn),             # <<<<<<<<<<<<<<
 *         ('BlendDstIn', BlendDstIn),
 *         ('BlendSrcOut', BlendSrcOut),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrcIn); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_n_s_BlendSrcIn);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrcIn);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_BlendSrcIn);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":61
 *         ('BlendDstOver', BlendDstOver),
 *         ('BlendSrcIn', BlendSrcIn),
 *         ('BlendDstIn', BlendDstIn),             # <<<<<<<<<<<<<<
 *         ('BlendSrcOut', BlendSrcOut),
 *         ('BlendDstOut', BlendDstOut),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDstIn); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_n_s_BlendDstIn);
    __Pyx_GIVEREF(__pyx_n_s_BlendDstIn);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_BlendDstIn);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":62
 *         ('BlendSrcIn', BlendSrcIn),
 *         ('BlendDstIn', BlendDstIn),
 *         ('BlendSrcOut', BlendSrcOut),             # <<<<<<<<<<<<<<
 *         ('BlendDstOut', BlendDstOut),
 *         ('BlendSrcAtop', BlendSrcAtop),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrcOut); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_BlendSrcOut);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrcOut);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_BlendSrcOut);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":63
 *         ('BlendDstIn', BlendDstIn),
 *         ('BlendSrcOut', BlendSrcOut),
 *         ('BlendDstOut', BlendDstOut),             # <<<<<<<<<<<<<<
 *         ('BlendSrcAtop', BlendSrcAtop),
 *         ('BlendDstAtop', BlendDstAtop),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDstOut); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_BlendDstOut);
    __Pyx_GIVEREF(__pyx_n_s_BlendDstOut);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_BlendDstOut);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":64
 *         ('BlendSrcOut', BlendSrcOut),
 *         ('BlendDstOut', BlendDstOut),
 *         ('BlendSrcAtop', BlendSrcAtop),             # <<<<<<<<<<<<<<
 *         ('BlendDstAtop', BlendDstAtop),
 *         ('BlendXor', BlendXor),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrcAtop); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_BlendSrcAtop);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrcAtop);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_s_BlendSrcAtop);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":65
 *         ('BlendDstOut', BlendDstOut),
 *         ('BlendSrcAtop', BlendSrcAtop),
 *         ('BlendDstAtop', BlendDstAtop),             # <<<<<<<<<<<<<<
 *         ('BlendXor', BlendXor),
 *         ('BlendAdd', BlendAdd),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDstAtop); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_BlendDstAtop);
    __Pyx_GIVEREF(__pyx_n_s_BlendDstAtop);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_s_BlendDstAtop);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":66
 *         ('BlendSrcAtop', BlendSrcAtop),
 *         ('BlendDstAtop', BlendDstAtop),
 *         ('BlendXor', BlendXor),             # <<<<<<<<<<<<<<
 *         ('BlendAdd', BlendAdd),
 *         ('BlendMultiply', BlendMultiply),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendXor); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_n_s_BlendXor);
    __Pyx_GIVEREF(__pyx_n_s_BlendXor);
    PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_n_s_BlendXor);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":67
 *         ('BlendDstAtop', BlendDstAtop),
 *         ('BlendXor', BlendXor),
 *         ('BlendAdd', BlendAdd),             # <<<<<<<<<<<<<<
 *         ('BlendMultiply', BlendMultiply),
 *         ('BlendScreen', BlendScreen),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendAdd); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_INCREF(__pyx_n_s_BlendAdd);
    __Pyx_GIVEREF(__pyx_n_s_BlendAdd);
    PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_n_s_BlendAdd);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":68
 *         ('BlendXor', BlendXor),
 *         ('BlendAdd', BlendAdd),
 *         ('BlendMultiply', BlendMultiply),             # <<<<<<<<<<<<<<
 *         ('BlendScreen', BlendScreen),
 *         ('BlendOverlay', BlendOverlay),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendMultiply); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_INCREF(__pyx_n_s_BlendMultiply);
    __Pyx_GIVEREF(__pyx_n_s_BlendMultiply);
    PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_n_s_BlendMultiply);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":69
 *         ('BlendAdd', BlendAdd),
 *         ('BlendMultiply', BlendMultiply),
 *         ('BlendScreen', BlendScreen),             # <<<<<<<<<<<<<<
 *         ('BlendOverlay', BlendOverlay),
 *         ('BlendDarken', BlendDarken),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendScreen); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_INCREF(__pyx_n_s_BlendScreen);
    __Pyx_GIVEREF(__pyx_n_s_BlendScreen);
    PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_n_s_BlendScreen);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":70
 *         ('BlendMultiply', BlendMultiply),
 *         ('BlendScreen', BlendScreen),
 *         ('BlendOverlay', BlendOverlay),             # <<<<<<<<<<<<<<
 *         ('BlendDarken', BlendDarken),
 *         ('BlendLighten', BlendLighten),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendOverlay); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_INCREF(__pyx_n_s_BlendOverlay);
    __Pyx_GIVEREF(__pyx_n_s_BlendOverlay);
    PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_s_BlendOverlay);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":71
 *         ('BlendScreen', BlendScreen),
 *         ('BlendOverlay', BlendOverlay),
 *         ('BlendDarken', BlendDarken),             # <<<<<<<<<<<<<<
 *         ('BlendLighten', BlendLighten),
 *         ('BlendColorDodge', BlendColorDodge),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDarken); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_INCREF(__pyx_n_s_BlendDarken);
    __Pyx_GIVEREF(__pyx_n_s_BlendDarken);
    PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_s_BlendDarken);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":72
 *         ('BlendOverlay', BlendOverlay),
 *         ('BlendDarken', BlendDarken),
 *         ('BlendLighten', BlendLighten),             # <<<<<<<<<<<<<<
 *         ('BlendColorDodge', BlendColorDodge),
 *         ('BlendColorBurn', BlendColorBurn),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendLighten); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_BlendLighten);
    __Pyx_GIVEREF(__pyx_n_s_BlendLighten);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_BlendLighten);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":73
 *         ('BlendDarken', BlendDarken),
 *         ('BlendLighten', BlendLighten),
 *         ('BlendColorDodge', BlendColorDodge),             # <<<<<<<<<<<<<<
 *         ('BlendColorBurn', BlendColorBurn),
 *         ('BlendHardLight', BlendHardLight),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendColorDodge); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(4, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_BlendColorDodge);
    __Pyx_GIVEREF(__pyx_n_s_BlendColorDodge);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_BlendColorDodge);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":74
 *         ('BlendLighten', BlendLighten),
 *         ('BlendColorDodge', BlendColorDodge),
 *         ('BlendColorBurn', BlendColorBurn),             # <<<<<<<<<<<<<<
 *         ('BlendHardLight', BlendHardLight),
 *         ('BlendSoftLight', BlendSoftLight),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendColorBurn); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_BlendColorBurn);
    __Pyx_GIVEREF(__pyx_n_s_BlendColorBurn);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_BlendColorBurn);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":75
 *         ('BlendColorDodge', BlendColorDodge),
 *         ('BlendColorBurn', BlendColorBurn),
 *         ('BlendHardLight', BlendHardLight),             # <<<<<<<<<<<<<<
 *         ('BlendSoftLight', BlendSoftLight),
 *         ('BlendDifference', BlendDifference),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendHardLight); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_BlendHardLight);
    __Pyx_GIVEREF(__pyx_n_s_BlendHardLight);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_BlendHardLight);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":76
 *         ('BlendColorBurn', BlendColorBurn),
 *         ('BlendHardLight', BlendHardLight),
 *         ('BlendSoftLight', BlendSoftLight),             # <<<<<<<<<<<<<<
 *         ('BlendDifference', BlendDifference),
 *         ('BlendExclusion', BlendExclusion),
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSoftLight); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_BlendSoftLight);
    __Pyx_GIVEREF(__pyx_n_s_BlendSoftLight);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_BlendSoftLight);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":77
 *         ('BlendHardLight', BlendHardLight),
 *         ('BlendSoftLight', BlendSoftLight),
 *         ('BlendDifference', BlendDifference),             # <<<<<<<<<<<<<<
 *         ('BlendExclusion', BlendExclusion),
 *     ]))
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDifference); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_BlendDifference);
    __Pyx_GIVEREF(__pyx_n_s_BlendDifference);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_BlendDifference);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":78
 *         ('BlendSoftLight', BlendSoftLight),
 *         ('BlendDifference', BlendDifference),
 *         ('BlendExclusion', BlendExclusion),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['BlendAlpha'] = BlendMode.BlendAlpha
 */
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendExclusion); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendExclusion);
    __Pyx_GIVEREF(__pyx_n_s_BlendExclusion);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_BlendExclusion);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     BlendMode = __Pyx_EnumBase('BlendMode', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('BlendAlpha', BlendAlpha),
 *         ('BlendClear', BlendClear),
 */
    __pyx_t_2 = PyList_New(25); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_7);
    PyList_SET_ITEM(__pyx_t_2, 3, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyList_SET_ITEM(__pyx_t_2, 4, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_2, 5, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_10);
    PyList_SET_ITEM(__pyx_t_2, 6, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    PyList_SET_ITEM(__pyx_t_2, 7, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyList_SET_ITEM(__pyx_t_2, 8, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyList_SET_ITEM(__pyx_t_2, 9, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyList_SET_ITEM(__pyx_t_2, 10, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyList_SET_ITEM(__pyx_t_2, 11, __pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_16);
    PyList_SET_ITEM(__pyx_t_2, 12, __pyx_t_16);
    __Pyx_GIVEREF(__pyx_t_17);
    PyList_SET_ITEM(__pyx_t_2, 13, __pyx_t_17);
    __Pyx_GIVEREF(__pyx_t_18);
    PyList_SET_ITEM(__pyx_t_2, 14, __pyx_t_18);
    __Pyx_GIVEREF(__pyx_t_19);
    PyList_SET_ITEM(__pyx_t_2, 15, __pyx_t_19);
    __Pyx_GIVEREF(__pyx_t_20);
    PyList_SET_ITEM(__pyx_t_2, 16, __pyx_t_20);
    __Pyx_GIVEREF(__pyx_t_21);
    PyList_SET_ITEM(__pyx_t_2, 17, __pyx_t_21);
    __Pyx_GIVEREF(__pyx_t_22);
    PyList_SET_ITEM(__pyx_t_2, 18, __pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_23);
    PyList_SET_ITEM(__pyx_t_2, 19, __pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_2, 20, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_2, 21, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_26);
    PyList_SET_ITEM(__pyx_t_2, 22, __pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_27);
    PyList_SET_ITEM(__pyx_t_2, 23, __pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyList_SET_ITEM(__pyx_t_2, 24, __pyx_t_28);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_t_13 = 0;
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_t_16 = 0;
    __pyx_t_17 = 0;
    __pyx_t_18 = 0;
    __pyx_t_19 = 0;
    __pyx_t_20 = 0;
    __pyx_t_21 = 0;
    __pyx_t_22 = 0;
    __pyx_t_23 = 0;
    __pyx_t_24 = 0;
    __pyx_t_25 = 0;
    __pyx_t_26 = 0;
    __pyx_t_27 = 0;
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_BlendMode);
    __Pyx_GIVEREF(__pyx_n_s_BlendMode);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_BlendMode);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_28);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_2, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_BlendMode, __pyx_t_28) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":80
 *         ('BlendExclusion', BlendExclusion),
 *     ]))
 *     __Pyx_globals['BlendAlpha'] = BlendMode.BlendAlpha             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendClear'] = BlendMode.BlendClear
 *     __Pyx_globals['BlendSrc'] = BlendMode.BlendSrc
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendAlpha); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 80, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendAlpha, __pyx_t_2) < 0)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":81
 *     ]))
 *     __Pyx_globals['BlendAlpha'] = BlendMode.BlendAlpha
 *     __Pyx_globals['BlendClear'] = BlendMode.BlendClear             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrc'] = BlendMode.BlendSrc
 *     __Pyx_globals['BlendDst'] = BlendMode.BlendDst
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendClear); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 81, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendClear, __pyx_t_28) < 0)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":82
 *     __Pyx_globals['BlendAlpha'] = BlendMode.BlendAlpha
 *     __Pyx_globals['BlendClear'] = BlendMode.BlendClear
 *     __Pyx_globals['BlendSrc'] = BlendMode.BlendSrc             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDst'] = BlendMode.BlendDst
 *     __Pyx_globals['BlendSrcOver'] = BlendMode.BlendSrcOver
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendSrc); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 82, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrc, __pyx_t_2) < 0)) __PYX_ERR(4, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":83
 *     __Pyx_globals['BlendClear'] = BlendMode.BlendClear
 *     __Pyx_globals['BlendSrc'] = BlendMode.BlendSrc
 *     __Pyx_globals['BlendDst'] = BlendMode.BlendDst             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrcOver'] = BlendMode.BlendSrcOver
 *     __Pyx_globals['BlendDstOver'] = BlendMode.BlendDstOver
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendDst); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 83, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDst, __pyx_t_28) < 0)) __PYX_ERR(4, 83, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":84
 *     __Pyx_globals['BlendSrc'] = BlendMode.BlendSrc
 *     __Pyx_globals['BlendDst'] = BlendMode.BlendDst
 *     __Pyx_globals['BlendSrcOver'] = BlendMode.BlendSrcOver             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDstOver'] = BlendMode.BlendDstOver
 *     __Pyx_globals['BlendSrcIn'] = BlendMode.BlendSrcIn
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendSrcOver); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 84, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrcOver, __pyx_t_2) < 0)) __PYX_ERR(4, 84, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":85
 *     __Pyx_globals['BlendDst'] = BlendMode.BlendDst
 *     __Pyx_globals['BlendSrcOver'] = BlendMode.BlendSrcOver
 *     __Pyx_globals['BlendDstOver'] = BlendMode.BlendDstOver             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrcIn'] = BlendMode.BlendSrcIn
 *     __Pyx_globals['BlendDstIn'] = BlendMode.BlendDstIn
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendDstOver); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 85, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDstOver, __pyx_t_28) < 0)) __PYX_ERR(4, 85, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":86
 *     __Pyx_globals['BlendSrcOver'] = BlendMode.BlendSrcOver
 *     __Pyx_globals['BlendDstOver'] = BlendMode.BlendDstOver
 *     __Pyx_globals['BlendSrcIn'] = BlendMode.BlendSrcIn             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDstIn'] = BlendMode.BlendDstIn
 *     __Pyx_globals['BlendSrcOut'] = BlendMode.BlendSrcOut
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendSrcIn); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 86, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrcIn, __pyx_t_2) < 0)) __PYX_ERR(4, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":87
 *     __Pyx_globals['BlendDstOver'] = BlendMode.BlendDstOver
 *     __Pyx_globals['BlendSrcIn'] = BlendMode.BlendSrcIn
 *     __Pyx_globals['BlendDstIn'] = BlendMode.BlendDstIn             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrcOut'] = BlendMode.BlendSrcOut
 *     __Pyx_globals['BlendDstOut'] = BlendMode.BlendDstOut
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendDstIn); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 87, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDstIn, __pyx_t_28) < 0)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":88
 *     __Pyx_globals['BlendSrcIn'] = BlendMode.BlendSrcIn
 *     __Pyx_globals['BlendDstIn'] = BlendMode.BlendDstIn
 *     __Pyx_globals['BlendSrcOut'] = BlendMode.BlendSrcOut             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDstOut'] = BlendMode.BlendDstOut
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode.BlendSrcAtop
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendSrcOut); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 88, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrcOut, __pyx_t_2) < 0)) __PYX_ERR(4, 88, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":89
 *     __Pyx_globals['BlendDstIn'] = BlendMode.BlendDstIn
 *     __Pyx_globals['BlendSrcOut'] = BlendMode.BlendSrcOut
 *     __Pyx_globals['BlendDstOut'] = BlendMode.BlendDstOut             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode.BlendSrcAtop
 *     __Pyx_globals['BlendDstAtop'] = BlendMode.BlendDstAtop
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendDstOut); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 89, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDstOut, __pyx_t_28) < 0)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":90
 *     __Pyx_globals['BlendSrcOut'] = BlendMode.BlendSrcOut
 *     __Pyx_globals['BlendDstOut'] = BlendMode.BlendDstOut
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode.BlendSrcAtop             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDstAtop'] = BlendMode.BlendDstAtop
 *     __Pyx_globals['BlendXor'] = BlendMode.BlendXor
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendSrcAtop); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 90, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrcAtop, __pyx_t_2) < 0)) __PYX_ERR(4, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":91
 *     __Pyx_globals['BlendDstOut'] = BlendMode.BlendDstOut
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode.BlendSrcAtop
 *     __Pyx_globals['BlendDstAtop'] = BlendMode.BlendDstAtop             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendXor'] = BlendMode.BlendXor
 *     __Pyx_globals['BlendAdd'] = BlendMode.BlendAdd
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendDstAtop); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 91, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDstAtop, __pyx_t_28) < 0)) __PYX_ERR(4, 91, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":92
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode.BlendSrcAtop
 *     __Pyx_globals['BlendDstAtop'] = BlendMode.BlendDstAtop
 *     __Pyx_globals['BlendXor'] = BlendMode.BlendXor             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendAdd'] = BlendMode.BlendAdd
 *     __Pyx_globals['BlendMultiply'] = BlendMode.BlendMultiply
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendXor); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 92, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendXor, __pyx_t_2) < 0)) __PYX_ERR(4, 92, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":93
 *     __Pyx_globals['BlendDstAtop'] = BlendMode.BlendDstAtop
 *     __Pyx_globals['BlendXor'] = BlendMode.BlendXor
 *     __Pyx_globals['BlendAdd'] = BlendMode.BlendAdd             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendMultiply'] = BlendMode.BlendMultiply
 *     __Pyx_globals['BlendScreen'] = BlendMode.BlendScreen
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendAdd); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 93, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendAdd, __pyx_t_28) < 0)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":94
 *     __Pyx_globals['BlendXor'] = BlendMode.BlendXor
 *     __Pyx_globals['BlendAdd'] = BlendMode.BlendAdd
 *     __Pyx_globals['BlendMultiply'] = BlendMode.BlendMultiply             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendScreen'] = BlendMode.BlendScreen
 *     __Pyx_globals['BlendOverlay'] = BlendMode.BlendOverlay
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendMultiply); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 94, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendMultiply, __pyx_t_2) < 0)) __PYX_ERR(4, 94, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":95
 *     __Pyx_globals['BlendAdd'] = BlendMode.BlendAdd
 *     __Pyx_globals['BlendMultiply'] = BlendMode.BlendMultiply
 *     __Pyx_globals['BlendScreen'] = BlendMode.BlendScreen             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendOverlay'] = BlendMode.BlendOverlay
 *     __Pyx_globals['BlendDarken'] = BlendMode.BlendDarken
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendScreen); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 95, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendScreen, __pyx_t_28) < 0)) __PYX_ERR(4, 95, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":96
 *     __Pyx_globals['BlendMultiply'] = BlendMode.BlendMultiply
 *     __Pyx_globals['BlendScreen'] = BlendMode.BlendScreen
 *     __Pyx_globals['BlendOverlay'] = BlendMode.BlendOverlay             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDarken'] = BlendMode.BlendDarken
 *     __Pyx_globals['BlendLighten'] = BlendMode.BlendLighten
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendOverlay); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 96, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendOverlay, __pyx_t_2) < 0)) __PYX_ERR(4, 96, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":97
 *     __Pyx_globals['BlendScreen'] = BlendMode.BlendScreen
 *     __Pyx_globals['BlendOverlay'] = BlendMode.BlendOverlay
 *     __Pyx_globals['BlendDarken'] = BlendMode.BlendDarken             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendLighten'] = BlendMode.BlendLighten
 *     __Pyx_globals['BlendColorDodge'] = BlendMode.BlendColorDodge
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendDarken); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 97, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 97, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDarken, __pyx_t_28) < 0)) __PYX_ERR(4, 97, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":98
 *     __Pyx_globals['BlendOverlay'] = BlendMode.BlendOverlay
 *     __Pyx_globals['BlendDarken'] = BlendMode.BlendDarken
 *     __Pyx_globals['BlendLighten'] = BlendMode.BlendLighten             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendColorDodge'] = BlendMode.BlendColorDodge
 *     __Pyx_globals['BlendColorBurn'] = BlendMode.BlendColorBurn
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendLighten); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 98, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendLighten, __pyx_t_2) < 0)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":99
 *     __Pyx_globals['BlendDarken'] = BlendMode.BlendDarken
 *     __Pyx_globals['BlendLighten'] = BlendMode.BlendLighten
 *     __Pyx_globals['BlendColorDodge'] = BlendMode.BlendColorDodge             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendColorBurn'] = BlendMode.BlendColorBurn
 *     __Pyx_globals['BlendHardLight'] = BlendMode.BlendHardLight
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendColorDodge); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 99, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendColorDodge, __pyx_t_28) < 0)) __PYX_ERR(4, 99, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":100
 *     __Pyx_globals['BlendLighten'] = BlendMode.BlendLighten
 *     __Pyx_globals['BlendColorDodge'] = BlendMode.BlendColorDodge
 *     __Pyx_globals['BlendColorBurn'] = BlendMode.BlendColorBurn             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendHardLight'] = BlendMode.BlendHardLight
 *     __Pyx_globals['BlendSoftLight'] = BlendMode.BlendSoftLight
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendColorBurn); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 100, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendColorBurn, __pyx_t_2) < 0)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":101
 *     __Pyx_globals['BlendColorDodge'] = BlendMode.BlendColorDodge
 *     __Pyx_globals['BlendColorBurn'] = BlendMode.BlendColorBurn
 *     __Pyx_globals['BlendHardLight'] = BlendMode.BlendHardLight             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSoftLight'] = BlendMode.BlendSoftLight
 *     __Pyx_globals['BlendDifference'] = BlendMode.BlendDifference
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendHardLight); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 101, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendHardLight, __pyx_t_28) < 0)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":102
 *     __Pyx_globals['BlendColorBurn'] = BlendMode.BlendColorBurn
 *     __Pyx_globals['BlendHardLight'] = BlendMode.BlendHardLight
 *     __Pyx_globals['BlendSoftLight'] = BlendMode.BlendSoftLight             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDifference'] = BlendMode.BlendDifference
 *     __Pyx_globals['BlendExclusion'] = BlendMode.BlendExclusion
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendSoftLight); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 102, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSoftLight, __pyx_t_2) < 0)) __PYX_ERR(4, 102, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":103
 *     __Pyx_globals['BlendHardLight'] = BlendMode.BlendHardLight
 *     __Pyx_globals['BlendSoftLight'] = BlendMode.BlendSoftLight
 *     __Pyx_globals['BlendDifference'] = BlendMode.BlendDifference             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendExclusion'] = BlendMode.BlendExclusion
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BlendDifference); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 103, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDifference, __pyx_t_28) < 0)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":104
 *     __Pyx_globals['BlendSoftLight'] = BlendMode.BlendSoftLight
 *     __Pyx_globals['BlendDifference'] = BlendMode.BlendDifference
 *     __Pyx_globals['BlendExclusion'] = BlendMode.BlendExclusion             # <<<<<<<<<<<<<<
 * else:
 *     class BlendMode(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_BlendExclusion); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 104, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendExclusion, __pyx_t_2) < 0)) __PYX_ERR(4, 104, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     BlendMode = __Pyx_EnumBase('BlendMode', __Pyx_OrderedDict([
 */
    goto __pyx_L4;
  }

  /* "EnumType":106
 *     __Pyx_globals['BlendExclusion'] = BlendMode.BlendExclusion
 * else:
 *     class BlendMode(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['BlendAlpha'] = BlendMode(BlendAlpha, 'BlendAlpha')
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __Pyx_EnumBase);
    __pyx_t_28 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_Py3MetaclassPrepare(__pyx_t_28, __pyx_t_2, __pyx_n_s_BlendMode, __pyx_n_s_BlendMode, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_26 = __Pyx_Py3ClassCreate(__pyx_t_28, __pyx_n_s_BlendMode, __pyx_t_2, __pyx_t_27, NULL, 0, 1); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_BlendMode, __pyx_t_26) < 0) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":108
 *     class BlendMode(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['BlendAlpha'] = BlendMode(BlendAlpha, 'BlendAlpha')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendClear'] = BlendMode(BlendClear, 'BlendClear')
 *     __Pyx_globals['BlendSrc'] = BlendMode(BlendSrc, 'BlendSrc')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendAlpha); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendAlpha);
    __Pyx_GIVEREF(__pyx_n_s_BlendAlpha);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_n_s_BlendAlpha);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_27, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 108, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendAlpha, __pyx_t_28) < 0)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":109
 *         pass
 *     __Pyx_globals['BlendAlpha'] = BlendMode(BlendAlpha, 'BlendAlpha')
 *     __Pyx_globals['BlendClear'] = BlendMode(BlendClear, 'BlendClear')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrc'] = BlendMode(BlendSrc, 'BlendSrc')
 *     __Pyx_globals['BlendDst'] = BlendMode(BlendDst, 'BlendDst')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendClear); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_BlendClear);
    __Pyx_GIVEREF(__pyx_n_s_BlendClear);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_BlendClear);
    __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_2, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 109, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendClear, __pyx_t_27) < 0)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

    /* "EnumType":110
 *     __Pyx_globals['BlendAlpha'] = BlendMode(BlendAlpha, 'BlendAlpha')
 *     __Pyx_globals['BlendClear'] = BlendMode(BlendClear, 'BlendClear')
 *     __Pyx_globals['BlendSrc'] = BlendMode(BlendSrc, 'BlendSrc')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDst'] = BlendMode(BlendDst, 'BlendDst')
 *     __Pyx_globals['BlendSrcOver'] = BlendMode(BlendSrcOver, 'BlendSrcOver')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrc); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_BlendSrc);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrc);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_BlendSrc);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_t_28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 110, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrc, __pyx_t_2) < 0)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":111
 *     __Pyx_globals['BlendClear'] = BlendMode(BlendClear, 'BlendClear')
 *     __Pyx_globals['BlendSrc'] = BlendMode(BlendSrc, 'BlendSrc')
 *     __Pyx_globals['BlendDst'] = BlendMode(BlendDst, 'BlendDst')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrcOver'] = BlendMode(BlendSrcOver, 'BlendSrcOver')
 *     __Pyx_globals['BlendDstOver'] = BlendMode(BlendDstOver, 'BlendDstOver')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDst); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendDst);
    __Pyx_GIVEREF(__pyx_n_s_BlendDst);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_n_s_BlendDst);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_27, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 111, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDst, __pyx_t_28) < 0)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":112
 *     __Pyx_globals['BlendSrc'] = BlendMode(BlendSrc, 'BlendSrc')
 *     __Pyx_globals['BlendDst'] = BlendMode(BlendDst, 'BlendDst')
 *     __Pyx_globals['BlendSrcOver'] = BlendMode(BlendSrcOver, 'BlendSrcOver')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDstOver'] = BlendMode(BlendDstOver, 'BlendDstOver')
 *     __Pyx_globals['BlendSrcIn'] = BlendMode(BlendSrcIn, 'BlendSrcIn')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrcOver); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_BlendSrcOver);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrcOver);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_BlendSrcOver);
    __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_2, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 112, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrcOver, __pyx_t_27) < 0)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

    /* "EnumType":113
 *     __Pyx_globals['BlendDst'] = BlendMode(BlendDst, 'BlendDst')
 *     __Pyx_globals['BlendSrcOver'] = BlendMode(BlendSrcOver, 'BlendSrcOver')
 *     __Pyx_globals['BlendDstOver'] = BlendMode(BlendDstOver, 'BlendDstOver')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrcIn'] = BlendMode(BlendSrcIn, 'BlendSrcIn')
 *     __Pyx_globals['BlendDstIn'] = BlendMode(BlendDstIn, 'BlendDstIn')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDstOver); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_BlendDstOver);
    __Pyx_GIVEREF(__pyx_n_s_BlendDstOver);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_BlendDstOver);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_t_28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 113, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDstOver, __pyx_t_2) < 0)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":114
 *     __Pyx_globals['BlendSrcOver'] = BlendMode(BlendSrcOver, 'BlendSrcOver')
 *     __Pyx_globals['BlendDstOver'] = BlendMode(BlendDstOver, 'BlendDstOver')
 *     __Pyx_globals['BlendSrcIn'] = BlendMode(BlendSrcIn, 'BlendSrcIn')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDstIn'] = BlendMode(BlendDstIn, 'BlendDstIn')
 *     __Pyx_globals['BlendSrcOut'] = BlendMode(BlendSrcOut, 'BlendSrcOut')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrcIn); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendSrcIn);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrcIn);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_n_s_BlendSrcIn);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_27, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 114, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrcIn, __pyx_t_28) < 0)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":115
 *     __Pyx_globals['BlendDstOver'] = BlendMode(BlendDstOver, 'BlendDstOver')
 *     __Pyx_globals['BlendSrcIn'] = BlendMode(BlendSrcIn, 'BlendSrcIn')
 *     __Pyx_globals['BlendDstIn'] = BlendMode(BlendDstIn, 'BlendDstIn')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrcOut'] = BlendMode(BlendSrcOut, 'BlendSrcOut')
 *     __Pyx_globals['BlendDstOut'] = BlendMode(BlendDstOut, 'BlendDstOut')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDstIn); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_BlendDstIn);
    __Pyx_GIVEREF(__pyx_n_s_BlendDstIn);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_BlendDstIn);
    __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_2, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 115, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDstIn, __pyx_t_27) < 0)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

    /* "EnumType":116
 *     __Pyx_globals['BlendSrcIn'] = BlendMode(BlendSrcIn, 'BlendSrcIn')
 *     __Pyx_globals['BlendDstIn'] = BlendMode(BlendDstIn, 'BlendDstIn')
 *     __Pyx_globals['BlendSrcOut'] = BlendMode(BlendSrcOut, 'BlendSrcOut')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDstOut'] = BlendMode(BlendDstOut, 'BlendDstOut')
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode(BlendSrcAtop, 'BlendSrcAtop')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrcOut); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_BlendSrcOut);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrcOut);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_BlendSrcOut);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_t_28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 116, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrcOut, __pyx_t_2) < 0)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":117
 *     __Pyx_globals['BlendDstIn'] = BlendMode(BlendDstIn, 'BlendDstIn')
 *     __Pyx_globals['BlendSrcOut'] = BlendMode(BlendSrcOut, 'BlendSrcOut')
 *     __Pyx_globals['BlendDstOut'] = BlendMode(BlendDstOut, 'BlendDstOut')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode(BlendSrcAtop, 'BlendSrcAtop')
 *     __Pyx_globals['BlendDstAtop'] = BlendMode(BlendDstAtop, 'BlendDstAtop')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDstOut); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendDstOut);
    __Pyx_GIVEREF(__pyx_n_s_BlendDstOut);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_n_s_BlendDstOut);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_27, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 117, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDstOut, __pyx_t_28) < 0)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":118
 *     __Pyx_globals['BlendSrcOut'] = BlendMode(BlendSrcOut, 'BlendSrcOut')
 *     __Pyx_globals['BlendDstOut'] = BlendMode(BlendDstOut, 'BlendDstOut')
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode(BlendSrcAtop, 'BlendSrcAtop')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDstAtop'] = BlendMode(BlendDstAtop, 'BlendDstAtop')
 *     __Pyx_globals['BlendXor'] = BlendMode(BlendXor, 'BlendXor')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSrcAtop); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_BlendSrcAtop);
    __Pyx_GIVEREF(__pyx_n_s_BlendSrcAtop);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_BlendSrcAtop);
    __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_2, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 118, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSrcAtop, __pyx_t_27) < 0)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

    /* "EnumType":119
 *     __Pyx_globals['BlendDstOut'] = BlendMode(BlendDstOut, 'BlendDstOut')
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode(BlendSrcAtop, 'BlendSrcAtop')
 *     __Pyx_globals['BlendDstAtop'] = BlendMode(BlendDstAtop, 'BlendDstAtop')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendXor'] = BlendMode(BlendXor, 'BlendXor')
 *     __Pyx_globals['BlendAdd'] = BlendMode(BlendAdd, 'BlendAdd')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDstAtop); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_BlendDstAtop);
    __Pyx_GIVEREF(__pyx_n_s_BlendDstAtop);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_BlendDstAtop);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_t_28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 119, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDstAtop, __pyx_t_2) < 0)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":120
 *     __Pyx_globals['BlendSrcAtop'] = BlendMode(BlendSrcAtop, 'BlendSrcAtop')
 *     __Pyx_globals['BlendDstAtop'] = BlendMode(BlendDstAtop, 'BlendDstAtop')
 *     __Pyx_globals['BlendXor'] = BlendMode(BlendXor, 'BlendXor')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendAdd'] = BlendMode(BlendAdd, 'BlendAdd')
 *     __Pyx_globals['BlendMultiply'] = BlendMode(BlendMultiply, 'BlendMultiply')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendXor); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendXor);
    __Pyx_GIVEREF(__pyx_n_s_BlendXor);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_n_s_BlendXor);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_27, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 120, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendXor, __pyx_t_28) < 0)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":121
 *     __Pyx_globals['BlendDstAtop'] = BlendMode(BlendDstAtop, 'BlendDstAtop')
 *     __Pyx_globals['BlendXor'] = BlendMode(BlendXor, 'BlendXor')
 *     __Pyx_globals['BlendAdd'] = BlendMode(BlendAdd, 'BlendAdd')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendMultiply'] = BlendMode(BlendMultiply, 'BlendMultiply')
 *     __Pyx_globals['BlendScreen'] = BlendMode(BlendScreen, 'BlendScreen')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendAdd); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_BlendAdd);
    __Pyx_GIVEREF(__pyx_n_s_BlendAdd);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_BlendAdd);
    __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_2, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 121, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendAdd, __pyx_t_27) < 0)) __PYX_ERR(4, 121, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

    /* "EnumType":122
 *     __Pyx_globals['BlendXor'] = BlendMode(BlendXor, 'BlendXor')
 *     __Pyx_globals['BlendAdd'] = BlendMode(BlendAdd, 'BlendAdd')
 *     __Pyx_globals['BlendMultiply'] = BlendMode(BlendMultiply, 'BlendMultiply')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendScreen'] = BlendMode(BlendScreen, 'BlendScreen')
 *     __Pyx_globals['BlendOverlay'] = BlendMode(BlendOverlay, 'BlendOverlay')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendMultiply); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_BlendMultiply);
    __Pyx_GIVEREF(__pyx_n_s_BlendMultiply);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_BlendMultiply);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_t_28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 122, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendMultiply, __pyx_t_2) < 0)) __PYX_ERR(4, 122, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":123
 *     __Pyx_globals['BlendAdd'] = BlendMode(BlendAdd, 'BlendAdd')
 *     __Pyx_globals['BlendMultiply'] = BlendMode(BlendMultiply, 'BlendMultiply')
 *     __Pyx_globals['BlendScreen'] = BlendMode(BlendScreen, 'BlendScreen')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendOverlay'] = BlendMode(BlendOverlay, 'BlendOverlay')
 *     __Pyx_globals['BlendDarken'] = BlendMode(BlendDarken, 'BlendDarken')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendScreen); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendScreen);
    __Pyx_GIVEREF(__pyx_n_s_BlendScreen);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_n_s_BlendScreen);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_27, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 123, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendScreen, __pyx_t_28) < 0)) __PYX_ERR(4, 123, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":124
 *     __Pyx_globals['BlendMultiply'] = BlendMode(BlendMultiply, 'BlendMultiply')
 *     __Pyx_globals['BlendScreen'] = BlendMode(BlendScreen, 'BlendScreen')
 *     __Pyx_globals['BlendOverlay'] = BlendMode(BlendOverlay, 'BlendOverlay')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDarken'] = BlendMode(BlendDarken, 'BlendDarken')
 *     __Pyx_globals['BlendLighten'] = BlendMode(BlendLighten, 'BlendLighten')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendOverlay); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_BlendOverlay);
    __Pyx_GIVEREF(__pyx_n_s_BlendOverlay);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_BlendOverlay);
    __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_2, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 124, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendOverlay, __pyx_t_27) < 0)) __PYX_ERR(4, 124, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

    /* "EnumType":125
 *     __Pyx_globals['BlendScreen'] = BlendMode(BlendScreen, 'BlendScreen')
 *     __Pyx_globals['BlendOverlay'] = BlendMode(BlendOverlay, 'BlendOverlay')
 *     __Pyx_globals['BlendDarken'] = BlendMode(BlendDarken, 'BlendDarken')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendLighten'] = BlendMode(BlendLighten, 'BlendLighten')
 *     __Pyx_globals['BlendColorDodge'] = BlendMode(BlendColorDodge, 'BlendColorDodge')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDarken); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_BlendDarken);
    __Pyx_GIVEREF(__pyx_n_s_BlendDarken);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_BlendDarken);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_t_28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 125, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDarken, __pyx_t_2) < 0)) __PYX_ERR(4, 125, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":126
 *     __Pyx_globals['BlendOverlay'] = BlendMode(BlendOverlay, 'BlendOverlay')
 *     __Pyx_globals['BlendDarken'] = BlendMode(BlendDarken, 'BlendDarken')
 *     __Pyx_globals['BlendLighten'] = BlendMode(BlendLighten, 'BlendLighten')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendColorDodge'] = BlendMode(BlendColorDodge, 'BlendColorDodge')
 *     __Pyx_globals['BlendColorBurn'] = BlendMode(BlendColorBurn, 'BlendColorBurn')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendLighten); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendLighten);
    __Pyx_GIVEREF(__pyx_n_s_BlendLighten);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_n_s_BlendLighten);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_27, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 126, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendLighten, __pyx_t_28) < 0)) __PYX_ERR(4, 126, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":127
 *     __Pyx_globals['BlendDarken'] = BlendMode(BlendDarken, 'BlendDarken')
 *     __Pyx_globals['BlendLighten'] = BlendMode(BlendLighten, 'BlendLighten')
 *     __Pyx_globals['BlendColorDodge'] = BlendMode(BlendColorDodge, 'BlendColorDodge')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendColorBurn'] = BlendMode(BlendColorBurn, 'BlendColorBurn')
 *     __Pyx_globals['BlendHardLight'] = BlendMode(BlendHardLight, 'BlendHardLight')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendColorDodge); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_BlendColorDodge);
    __Pyx_GIVEREF(__pyx_n_s_BlendColorDodge);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_BlendColorDodge);
    __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_2, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 127, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendColorDodge, __pyx_t_27) < 0)) __PYX_ERR(4, 127, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

    /* "EnumType":128
 *     __Pyx_globals['BlendLighten'] = BlendMode(BlendLighten, 'BlendLighten')
 *     __Pyx_globals['BlendColorDodge'] = BlendMode(BlendColorDodge, 'BlendColorDodge')
 *     __Pyx_globals['BlendColorBurn'] = BlendMode(BlendColorBurn, 'BlendColorBurn')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendHardLight'] = BlendMode(BlendHardLight, 'BlendHardLight')
 *     __Pyx_globals['BlendSoftLight'] = BlendMode(BlendSoftLight, 'BlendSoftLight')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendColorBurn); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_BlendColorBurn);
    __Pyx_GIVEREF(__pyx_n_s_BlendColorBurn);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_BlendColorBurn);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_t_28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 128, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendColorBurn, __pyx_t_2) < 0)) __PYX_ERR(4, 128, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":129
 *     __Pyx_globals['BlendColorDodge'] = BlendMode(BlendColorDodge, 'BlendColorDodge')
 *     __Pyx_globals['BlendColorBurn'] = BlendMode(BlendColorBurn, 'BlendColorBurn')
 *     __Pyx_globals['BlendHardLight'] = BlendMode(BlendHardLight, 'BlendHardLight')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendSoftLight'] = BlendMode(BlendSoftLight, 'BlendSoftLight')
 *     __Pyx_globals['BlendDifference'] = BlendMode(BlendDifference, 'BlendDifference')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendHardLight); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendHardLight);
    __Pyx_GIVEREF(__pyx_n_s_BlendHardLight);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_n_s_BlendHardLight);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_27, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 129, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendHardLight, __pyx_t_28) < 0)) __PYX_ERR(4, 129, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":130
 *     __Pyx_globals['BlendColorBurn'] = BlendMode(BlendColorBurn, 'BlendColorBurn')
 *     __Pyx_globals['BlendHardLight'] = BlendMode(BlendHardLight, 'BlendHardLight')
 *     __Pyx_globals['BlendSoftLight'] = BlendMode(BlendSoftLight, 'BlendSoftLight')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendDifference'] = BlendMode(BlendDifference, 'BlendDifference')
 *     __Pyx_globals['BlendExclusion'] = BlendMode(BlendExclusion, 'BlendExclusion')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendSoftLight); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_27);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_BlendSoftLight);
    __Pyx_GIVEREF(__pyx_n_s_BlendSoftLight);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_BlendSoftLight);
    __pyx_t_27 = 0;
    __pyx_t_27 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_2, NULL); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 130, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendSoftLight, __pyx_t_27) < 0)) __PYX_ERR(4, 130, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;

    /* "EnumType":131
 *     __Pyx_globals['BlendHardLight'] = BlendMode(BlendHardLight, 'BlendHardLight')
 *     __Pyx_globals['BlendSoftLight'] = BlendMode(BlendSoftLight, 'BlendSoftLight')
 *     __Pyx_globals['BlendDifference'] = BlendMode(BlendDifference, 'BlendDifference')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BlendExclusion'] = BlendMode(BlendExclusion, 'BlendExclusion')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_27, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendDifference); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_BlendDifference);
    __Pyx_GIVEREF(__pyx_n_s_BlendDifference);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_BlendDifference);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_27, __pyx_t_28, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 131, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendDifference, __pyx_t_2) < 0)) __PYX_ERR(4, 131, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":132
 *     __Pyx_globals['BlendSoftLight'] = BlendMode(BlendSoftLight, 'BlendSoftLight')
 *     __Pyx_globals['BlendDifference'] = BlendMode(BlendDifference, 'BlendDifference')
 *     __Pyx_globals['BlendExclusion'] = BlendMode(BlendExclusion, 'BlendExclusion')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BlendMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(__pyx_e_7celiagg_8_celiagg_BlendExclusion); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_BlendExclusion);
    __Pyx_GIVEREF(__pyx_n_s_BlendExclusion);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_n_s_BlendExclusion);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_27, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_27); __pyx_t_27 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 132, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BlendExclusion, __pyx_t_28) < 0)) __PYX_ERR(4, 132, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  }
  __pyx_L4:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_28 = __Pyx_Globals(); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (!(likely(PyDict_CheckExact(__pyx_t_28))||((__pyx_t_28) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_28)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_28));
  __Pyx_GIVEREF(__pyx_t_28);
  __pyx_t_28 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     DrawingMode = __Pyx_EnumBase('DrawingMode', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     DrawingMode = __Pyx_EnumBase('DrawingMode', __Pyx_OrderedDict([
 *         ('DrawFill', DrawFill),             # <<<<<<<<<<<<<<
 *         ('DrawEofFill', DrawEofFill),
 *         ('DrawStroke', DrawStroke),
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawFill); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_DrawFill);
    __Pyx_GIVEREF(__pyx_n_s_DrawFill);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_DrawFill);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":55
 *     DrawingMode = __Pyx_EnumBase('DrawingMode', __Pyx_OrderedDict([
 *         ('DrawFill', DrawFill),
 *         ('DrawEofFill', DrawEofFill),             # <<<<<<<<<<<<<<
 *         ('DrawStroke', DrawStroke),
 *         ('DrawFillStroke', DrawFillStroke),
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawEofFill); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_DrawEofFill);
    __Pyx_GIVEREF(__pyx_n_s_DrawEofFill);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_DrawEofFill);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":56
 *         ('DrawFill', DrawFill),
 *         ('DrawEofFill', DrawEofFill),
 *         ('DrawStroke', DrawStroke),             # <<<<<<<<<<<<<<
 *         ('DrawFillStroke', DrawFillStroke),
 *         ('DrawEofFillStroke', DrawEofFillStroke),
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawStroke); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_DrawStroke);
    __Pyx_GIVEREF(__pyx_n_s_DrawStroke);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_DrawStroke);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":57
 *         ('DrawEofFill', DrawEofFill),
 *         ('DrawStroke', DrawStroke),
 *         ('DrawFillStroke', DrawFillStroke),             # <<<<<<<<<<<<<<
 *         ('DrawEofFillStroke', DrawEofFillStroke),
 *     ]))
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawFillStroke); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_DrawFillStroke);
    __Pyx_GIVEREF(__pyx_n_s_DrawFillStroke);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_DrawFillStroke);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":58
 *         ('DrawStroke', DrawStroke),
 *         ('DrawFillStroke', DrawFillStroke),
 *         ('DrawEofFillStroke', DrawEofFillStroke),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['DrawFill'] = DrawingMode.DrawFill
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawEofFillStroke); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_DrawEofFillStroke);
    __Pyx_GIVEREF(__pyx_n_s_DrawEofFillStroke);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_DrawEofFillStroke);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     DrawingMode = __Pyx_EnumBase('DrawingMode', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('DrawFill', DrawFill),
 *         ('DrawEofFill', DrawEofFill),
 */
    __pyx_t_28 = PyList_New(5); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_27);
    PyList_SET_ITEM(__pyx_t_28, 0, __pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_28, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_26);
    PyList_SET_ITEM(__pyx_t_28, 2, __pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_28, 3, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_28, 4, __pyx_t_24);
    __pyx_t_27 = 0;
    __pyx_t_2 = 0;
    __pyx_t_26 = 0;
    __pyx_t_25 = 0;
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_28); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_DrawingMode);
    __Pyx_GIVEREF(__pyx_n_s_DrawingMode);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_DrawingMode);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_24);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_28, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_DrawingMode, __pyx_t_24) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":60
 *         ('DrawEofFillStroke', DrawEofFillStroke),
 *     ]))
 *     __Pyx_globals['DrawFill'] = DrawingMode.DrawFill             # <<<<<<<<<<<<<<
 *     __Pyx_globals['DrawEofFill'] = DrawingMode.DrawEofFill
 *     __Pyx_globals['DrawStroke'] = DrawingMode.DrawStroke
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_DrawFill); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 60, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawFill, __pyx_t_28) < 0)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":61
 *     ]))
 *     __Pyx_globals['DrawFill'] = DrawingMode.DrawFill
 *     __Pyx_globals['DrawEofFill'] = DrawingMode.DrawEofFill             # <<<<<<<<<<<<<<
 *     __Pyx_globals['DrawStroke'] = DrawingMode.DrawStroke
 *     __Pyx_globals['DrawFillStroke'] = DrawingMode.DrawFillStroke
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_DrawEofFill); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 61, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawEofFill, __pyx_t_24) < 0)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":62
 *     __Pyx_globals['DrawFill'] = DrawingMode.DrawFill
 *     __Pyx_globals['DrawEofFill'] = DrawingMode.DrawEofFill
 *     __Pyx_globals['DrawStroke'] = DrawingMode.DrawStroke             # <<<<<<<<<<<<<<
 *     __Pyx_globals['DrawFillStroke'] = DrawingMode.DrawFillStroke
 *     __Pyx_globals['DrawEofFillStroke'] = DrawingMode.DrawEofFillStroke
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_DrawStroke); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 62, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawStroke, __pyx_t_28) < 0)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":63
 *     __Pyx_globals['DrawEofFill'] = DrawingMode.DrawEofFill
 *     __Pyx_globals['DrawStroke'] = DrawingMode.DrawStroke
 *     __Pyx_globals['DrawFillStroke'] = DrawingMode.DrawFillStroke             # <<<<<<<<<<<<<<
 *     __Pyx_globals['DrawEofFillStroke'] = DrawingMode.DrawEofFillStroke
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_DrawFillStroke); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 63, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawFillStroke, __pyx_t_24) < 0)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":64
 *     __Pyx_globals['DrawStroke'] = DrawingMode.DrawStroke
 *     __Pyx_globals['DrawFillStroke'] = DrawingMode.DrawFillStroke
 *     __Pyx_globals['DrawEofFillStroke'] = DrawingMode.DrawEofFillStroke             # <<<<<<<<<<<<<<
 * else:
 *     class DrawingMode(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_DrawEofFillStroke); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 64, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawEofFillStroke, __pyx_t_28) < 0)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     DrawingMode = __Pyx_EnumBase('DrawingMode', __Pyx_OrderedDict([
 */
    goto __pyx_L5;
  }

  /* "EnumType":66
 *     __Pyx_globals['DrawEofFillStroke'] = DrawingMode.DrawEofFillStroke
 * else:
 *     class DrawingMode(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['DrawFill'] = DrawingMode(DrawFill, 'DrawFill')
 */
  /*else*/ {
    __pyx_t_28 = PyTuple_New(1); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __Pyx_EnumBase);
    __pyx_t_24 = __Pyx_CalculateMetaclass(NULL, __pyx_t_28); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = __Pyx_Py3MetaclassPrepare(__pyx_t_24, __pyx_t_28, __pyx_n_s_DrawingMode, __pyx_n_s_DrawingMode, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_26 = __Pyx_Py3ClassCreate(__pyx_t_24, __pyx_n_s_DrawingMode, __pyx_t_28, __pyx_t_25, NULL, 0, 1); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_DrawingMode, __pyx_t_26) < 0) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":68
 *     class DrawingMode(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['DrawFill'] = DrawingMode(DrawFill, 'DrawFill')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['DrawEofFill'] = DrawingMode(DrawEofFill, 'DrawEofFill')
 *     __Pyx_globals['DrawStroke'] = DrawingMode(DrawStroke, 'DrawStroke')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawFill); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_DrawFill);
    __Pyx_GIVEREF(__pyx_n_s_DrawFill);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_n_s_DrawFill);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_25, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 68, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawFill, __pyx_t_24) < 0)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":69
 *         pass
 *     __Pyx_globals['DrawFill'] = DrawingMode(DrawFill, 'DrawFill')
 *     __Pyx_globals['DrawEofFill'] = DrawingMode(DrawEofFill, 'DrawEofFill')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['DrawStroke'] = DrawingMode(DrawStroke, 'DrawStroke')
 *     __Pyx_globals['DrawFillStroke'] = DrawingMode(DrawFillStroke, 'DrawFillStroke')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawEofFill); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_DrawEofFill);
    __Pyx_GIVEREF(__pyx_n_s_DrawEofFill);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_DrawEofFill);
    __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_28, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 69, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawEofFill, __pyx_t_25) < 0)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "EnumType":70
 *     __Pyx_globals['DrawFill'] = DrawingMode(DrawFill, 'DrawFill')
 *     __Pyx_globals['DrawEofFill'] = DrawingMode(DrawEofFill, 'DrawEofFill')
 *     __Pyx_globals['DrawStroke'] = DrawingMode(DrawStroke, 'DrawStroke')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['DrawFillStroke'] = DrawingMode(DrawFillStroke, 'DrawFillStroke')
 *     __Pyx_globals['DrawEofFillStroke'] = DrawingMode(DrawEofFillStroke, 'DrawEofFillStroke')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawStroke); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_DrawStroke);
    __Pyx_GIVEREF(__pyx_n_s_DrawStroke);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_DrawStroke);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_25, __pyx_t_24, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 70, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawStroke, __pyx_t_28) < 0)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":71
 *     __Pyx_globals['DrawEofFill'] = DrawingMode(DrawEofFill, 'DrawEofFill')
 *     __Pyx_globals['DrawStroke'] = DrawingMode(DrawStroke, 'DrawStroke')
 *     __Pyx_globals['DrawFillStroke'] = DrawingMode(DrawFillStroke, 'DrawFillStroke')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['DrawEofFillStroke'] = DrawingMode(DrawEofFillStroke, 'DrawEofFillStroke')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawFillStroke); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_DrawFillStroke);
    __Pyx_GIVEREF(__pyx_n_s_DrawFillStroke);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_n_s_DrawFillStroke);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_25, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 71, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawFillStroke, __pyx_t_24) < 0)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":72
 *     __Pyx_globals['DrawStroke'] = DrawingMode(DrawStroke, 'DrawStroke')
 *     __Pyx_globals['DrawFillStroke'] = DrawingMode(DrawFillStroke, 'DrawFillStroke')
 *     __Pyx_globals['DrawEofFillStroke'] = DrawingMode(DrawEofFillStroke, 'DrawEofFillStroke')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_DrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(__pyx_e_7celiagg_8_celiagg_DrawEofFillStroke); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_DrawEofFillStroke);
    __Pyx_GIVEREF(__pyx_n_s_DrawEofFillStroke);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_DrawEofFillStroke);
    __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_28, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 72, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_DrawEofFillStroke, __pyx_t_25) < 0)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
  }
  __pyx_L5:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_25 = __Pyx_Globals(); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_25);
  if (!(likely(PyDict_CheckExact(__pyx_t_25))||((__pyx_t_25) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_25)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_25));
  __Pyx_GIVEREF(__pyx_t_25);
  __pyx_t_25 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     FontCacheType = __Pyx_EnumBase('FontCacheType', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     FontCacheType = __Pyx_EnumBase('FontCacheType', __Pyx_OrderedDict([
 *         ('RasterFontCache', RasterFontCache),             # <<<<<<<<<<<<<<
 *         ('VectorFontCache', VectorFontCache),
 *     ]))
 */
    __pyx_t_25 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(__pyx_e_7celiagg_8_celiagg_RasterFontCache); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_RasterFontCache);
    __Pyx_GIVEREF(__pyx_n_s_RasterFontCache);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_RasterFontCache);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_25);
    __pyx_t_25 = 0;

    /* "EnumType":55
 *     FontCacheType = __Pyx_EnumBase('FontCacheType', __Pyx_OrderedDict([
 *         ('RasterFontCache', RasterFontCache),
 *         ('VectorFontCache', VectorFontCache),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['RasterFontCache'] = FontCacheType.RasterFontCache
 */
    __pyx_t_25 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(__pyx_e_7celiagg_8_celiagg_VectorFontCache); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_VectorFontCache);
    __Pyx_GIVEREF(__pyx_n_s_VectorFontCache);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_VectorFontCache);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_25);
    __pyx_t_25 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     FontCacheType = __Pyx_EnumBase('FontCacheType', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('RasterFontCache', RasterFontCache),
 *         ('VectorFontCache', VectorFontCache),
 */
    __pyx_t_25 = PyList_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_28);
    PyList_SET_ITEM(__pyx_t_25, 0, __pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_25, 1, __pyx_t_24);
    __pyx_t_28 = 0;
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_25); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_FontCacheType);
    __Pyx_GIVEREF(__pyx_n_s_FontCacheType);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_FontCacheType);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_24);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_25, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_FontCacheType, __pyx_t_24) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":57
 *         ('VectorFontCache', VectorFontCache),
 *     ]))
 *     __Pyx_globals['RasterFontCache'] = FontCacheType.RasterFontCache             # <<<<<<<<<<<<<<
 *     __Pyx_globals['VectorFontCache'] = FontCacheType.VectorFontCache
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_FontCacheType); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_RasterFontCache); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 57, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RasterFontCache, __pyx_t_25) < 0)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "EnumType":58
 *     ]))
 *     __Pyx_globals['RasterFontCache'] = FontCacheType.RasterFontCache
 *     __Pyx_globals['VectorFontCache'] = FontCacheType.VectorFontCache             # <<<<<<<<<<<<<<
 * else:
 *     class FontCacheType(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_n_s_FontCacheType); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_n_s_VectorFontCache); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 58, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_VectorFontCache, __pyx_t_24) < 0)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     FontCacheType = __Pyx_EnumBase('FontCacheType', __Pyx_OrderedDict([
 */
    goto __pyx_L6;
  }

  /* "EnumType":60
 *     __Pyx_globals['VectorFontCache'] = FontCacheType.VectorFontCache
 * else:
 *     class FontCacheType(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['RasterFontCache'] = FontCacheType(RasterFontCache, 'RasterFontCache')
 */
  /*else*/ {
    __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __Pyx_EnumBase);
    __pyx_t_25 = __Pyx_CalculateMetaclass(NULL, __pyx_t_24); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = __Pyx_Py3MetaclassPrepare(__pyx_t_25, __pyx_t_24, __pyx_n_s_FontCacheType, __pyx_n_s_FontCacheType, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = __Pyx_Py3ClassCreate(__pyx_t_25, __pyx_n_s_FontCacheType, __pyx_t_24, __pyx_t_28, NULL, 0, 1); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_FontCacheType, __pyx_t_26) < 0) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":62
 *     class FontCacheType(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['RasterFontCache'] = FontCacheType(RasterFontCache, 'RasterFontCache')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['VectorFontCache'] = FontCacheType(VectorFontCache, 'VectorFontCache')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_FontCacheType); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(__pyx_e_7celiagg_8_celiagg_RasterFontCache); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_RasterFontCache);
    __Pyx_GIVEREF(__pyx_n_s_RasterFontCache);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_RasterFontCache);
    __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_28, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 62, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RasterFontCache, __pyx_t_25) < 0)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "EnumType":63
 *         pass
 *     __Pyx_globals['RasterFontCache'] = FontCacheType(RasterFontCache, 'RasterFontCache')
 *     __Pyx_globals['VectorFontCache'] = FontCacheType(VectorFontCache, 'VectorFontCache')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_n_s_FontCacheType); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(__pyx_e_7celiagg_8_celiagg_VectorFontCache); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_VectorFontCache);
    __Pyx_GIVEREF(__pyx_n_s_VectorFontCache);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_VectorFontCache);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_25, __pyx_t_24, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 63, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_VectorFontCache, __pyx_t_28) < 0)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  }
  __pyx_L6:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_28 = __Pyx_Globals(); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (!(likely(PyDict_CheckExact(__pyx_t_28))||((__pyx_t_28) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_28)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_28));
  __Pyx_GIVEREF(__pyx_t_28);
  __pyx_t_28 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     GradientSpread = __Pyx_EnumBase('GradientSpread', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     GradientSpread = __Pyx_EnumBase('GradientSpread', __Pyx_OrderedDict([
 *         ('SpreadPad', SpreadPad),             # <<<<<<<<<<<<<<
 *         ('SpreadReflect', SpreadReflect),
 *         ('SpreadRepeat', SpreadRepeat),
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(__pyx_e_7celiagg_8_celiagg_SpreadPad); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_SpreadPad);
    __Pyx_GIVEREF(__pyx_n_s_SpreadPad);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_SpreadPad);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":55
 *     GradientSpread = __Pyx_EnumBase('GradientSpread', __Pyx_OrderedDict([
 *         ('SpreadPad', SpreadPad),
 *         ('SpreadReflect', SpreadReflect),             # <<<<<<<<<<<<<<
 *         ('SpreadRepeat', SpreadRepeat),
 *     ]))
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(__pyx_e_7celiagg_8_celiagg_SpreadReflect); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_SpreadReflect);
    __Pyx_GIVEREF(__pyx_n_s_SpreadReflect);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_SpreadReflect);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":56
 *         ('SpreadPad', SpreadPad),
 *         ('SpreadReflect', SpreadReflect),
 *         ('SpreadRepeat', SpreadRepeat),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['SpreadPad'] = GradientSpread.SpreadPad
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(__pyx_e_7celiagg_8_celiagg_SpreadRepeat); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_SpreadRepeat);
    __Pyx_GIVEREF(__pyx_n_s_SpreadRepeat);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_SpreadRepeat);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     GradientSpread = __Pyx_EnumBase('GradientSpread', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('SpreadPad', SpreadPad),
 *         ('SpreadReflect', SpreadReflect),
 */
    __pyx_t_28 = PyList_New(3); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_28, 0, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_28, 1, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_26);
    PyList_SET_ITEM(__pyx_t_28, 2, __pyx_t_26);
    __pyx_t_24 = 0;
    __pyx_t_25 = 0;
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_28); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_GradientSpread);
    __Pyx_GIVEREF(__pyx_n_s_GradientSpread);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_GradientSpread);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_26);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_28, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_GradientSpread, __pyx_t_26) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":58
 *         ('SpreadRepeat', SpreadRepeat),
 *     ]))
 *     __Pyx_globals['SpreadPad'] = GradientSpread.SpreadPad             # <<<<<<<<<<<<<<
 *     __Pyx_globals['SpreadReflect'] = GradientSpread.SpreadReflect
 *     __Pyx_globals['SpreadRepeat'] = GradientSpread.SpreadRepeat
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_GradientSpread); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_n_s_SpreadPad); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 58, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_SpreadPad, __pyx_t_28) < 0)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":59
 *     ]))
 *     __Pyx_globals['SpreadPad'] = GradientSpread.SpreadPad
 *     __Pyx_globals['SpreadReflect'] = GradientSpread.SpreadReflect             # <<<<<<<<<<<<<<
 *     __Pyx_globals['SpreadRepeat'] = GradientSpread.SpreadRepeat
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_GradientSpread); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_SpreadReflect); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 59, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_SpreadReflect, __pyx_t_26) < 0)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":60
 *     __Pyx_globals['SpreadPad'] = GradientSpread.SpreadPad
 *     __Pyx_globals['SpreadReflect'] = GradientSpread.SpreadReflect
 *     __Pyx_globals['SpreadRepeat'] = GradientSpread.SpreadRepeat             # <<<<<<<<<<<<<<
 * else:
 *     class GradientSpread(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_GradientSpread); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_n_s_SpreadRepeat); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 60, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_SpreadRepeat, __pyx_t_28) < 0)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     GradientSpread = __Pyx_EnumBase('GradientSpread', __Pyx_OrderedDict([
 */
    goto __pyx_L7;
  }

  /* "EnumType":62
 *     __Pyx_globals['SpreadRepeat'] = GradientSpread.SpreadRepeat
 * else:
 *     class GradientSpread(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['SpreadPad'] = GradientSpread(SpreadPad, 'SpreadPad')
 */
  /*else*/ {
    __pyx_t_28 = PyTuple_New(1); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __Pyx_EnumBase);
    __pyx_t_26 = __Pyx_CalculateMetaclass(NULL, __pyx_t_28); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_25 = __Pyx_Py3MetaclassPrepare(__pyx_t_26, __pyx_t_28, __pyx_n_s_GradientSpread, __pyx_n_s_GradientSpread, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_24 = __Pyx_Py3ClassCreate(__pyx_t_26, __pyx_n_s_GradientSpread, __pyx_t_28, __pyx_t_25, NULL, 0, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_GradientSpread, __pyx_t_24) < 0) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":64
 *     class GradientSpread(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['SpreadPad'] = GradientSpread(SpreadPad, 'SpreadPad')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['SpreadReflect'] = GradientSpread(SpreadReflect, 'SpreadReflect')
 *     __Pyx_globals['SpreadRepeat'] = GradientSpread(SpreadRepeat, 'SpreadRepeat')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_GradientSpread); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(__pyx_e_7celiagg_8_celiagg_SpreadPad); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_SpreadPad);
    __Pyx_GIVEREF(__pyx_n_s_SpreadPad);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_n_s_SpreadPad);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_25, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 64, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_SpreadPad, __pyx_t_26) < 0)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":65
 *         pass
 *     __Pyx_globals['SpreadPad'] = GradientSpread(SpreadPad, 'SpreadPad')
 *     __Pyx_globals['SpreadReflect'] = GradientSpread(SpreadReflect, 'SpreadReflect')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['SpreadRepeat'] = GradientSpread(SpreadRepeat, 'SpreadRepeat')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_GradientSpread); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_25 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(__pyx_e_7celiagg_8_celiagg_SpreadReflect); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_SpreadReflect);
    __Pyx_GIVEREF(__pyx_n_s_SpreadReflect);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_SpreadReflect);
    __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_t_28, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 65, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_SpreadReflect, __pyx_t_25) < 0)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "EnumType":66
 *     __Pyx_globals['SpreadPad'] = GradientSpread(SpreadPad, 'SpreadPad')
 *     __Pyx_globals['SpreadReflect'] = GradientSpread(SpreadReflect, 'SpreadReflect')
 *     __Pyx_globals['SpreadRepeat'] = GradientSpread(SpreadRepeat, 'SpreadRepeat')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_n_s_GradientSpread); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(__pyx_e_7celiagg_8_celiagg_SpreadRepeat); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_SpreadRepeat);
    __Pyx_GIVEREF(__pyx_n_s_SpreadRepeat);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_n_s_SpreadRepeat);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_25, __pyx_t_26, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 66, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_SpreadRepeat, __pyx_t_28) < 0)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
  }
  __pyx_L7:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_28 = __Pyx_Globals(); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_28);
  if (!(likely(PyDict_CheckExact(__pyx_t_28))||((__pyx_t_28) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_28)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_28));
  __Pyx_GIVEREF(__pyx_t_28);
  __pyx_t_28 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     GradientUnits = __Pyx_EnumBase('GradientUnits', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     GradientUnits = __Pyx_EnumBase('GradientUnits', __Pyx_OrderedDict([
 *         ('UserSpace', UserSpace),             # <<<<<<<<<<<<<<
 *         ('ObjectBoundingBox', ObjectBoundingBox),
 *     ]))
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(__pyx_e_7celiagg_8_celiagg_UserSpace); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_UserSpace);
    __Pyx_GIVEREF(__pyx_n_s_UserSpace);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_UserSpace);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":55
 *     GradientUnits = __Pyx_EnumBase('GradientUnits', __Pyx_OrderedDict([
 *         ('UserSpace', UserSpace),
 *         ('ObjectBoundingBox', ObjectBoundingBox),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['UserSpace'] = GradientUnits.UserSpace
 */
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(__pyx_e_7celiagg_8_celiagg_ObjectBoundingBox); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_ObjectBoundingBox);
    __Pyx_GIVEREF(__pyx_n_s_ObjectBoundingBox);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_ObjectBoundingBox);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_28);
    __pyx_t_28 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     GradientUnits = __Pyx_EnumBase('GradientUnits', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('UserSpace', UserSpace),
 *         ('ObjectBoundingBox', ObjectBoundingBox),
 */
    __pyx_t_28 = PyList_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_26);
    PyList_SET_ITEM(__pyx_t_28, 0, __pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_28, 1, __pyx_t_25);
    __pyx_t_26 = 0;
    __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_28); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_GradientUnits);
    __Pyx_GIVEREF(__pyx_n_s_GradientUnits);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_GradientUnits);
    __Pyx_GIVEREF(__pyx_t_25);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_25);
    __pyx_t_25 = 0;
    __pyx_t_25 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_28, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_GradientUnits, __pyx_t_25) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "EnumType":57
 *         ('ObjectBoundingBox', ObjectBoundingBox),
 *     ]))
 *     __Pyx_globals['UserSpace'] = GradientUnits.UserSpace             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ObjectBoundingBox'] = GradientUnits.ObjectBoundingBox
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_n_s_GradientUnits); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_n_s_UserSpace); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 57, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_UserSpace, __pyx_t_28) < 0)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":58
 *     ]))
 *     __Pyx_globals['UserSpace'] = GradientUnits.UserSpace
 *     __Pyx_globals['ObjectBoundingBox'] = GradientUnits.ObjectBoundingBox             # <<<<<<<<<<<<<<
 * else:
 *     class GradientUnits(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_GradientUnits); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_ObjectBoundingBox); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 58, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ObjectBoundingBox, __pyx_t_25) < 0)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     GradientUnits = __Pyx_EnumBase('GradientUnits', __Pyx_OrderedDict([
 */
    goto __pyx_L8;
  }

  /* "EnumType":60
 *     __Pyx_globals['ObjectBoundingBox'] = GradientUnits.ObjectBoundingBox
 * else:
 *     class GradientUnits(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['UserSpace'] = GradientUnits(UserSpace, 'UserSpace')
 */
  /*else*/ {
    __pyx_t_25 = PyTuple_New(1); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __Pyx_EnumBase);
    __pyx_t_28 = __Pyx_CalculateMetaclass(NULL, __pyx_t_25); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = __Pyx_Py3MetaclassPrepare(__pyx_t_28, __pyx_t_25, __pyx_n_s_GradientUnits, __pyx_n_s_GradientUnits, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_Py3ClassCreate(__pyx_t_28, __pyx_n_s_GradientUnits, __pyx_t_25, __pyx_t_26, NULL, 0, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_GradientUnits, __pyx_t_24) < 0) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

    /* "EnumType":62
 *     class GradientUnits(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['UserSpace'] = GradientUnits(UserSpace, 'UserSpace')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ObjectBoundingBox'] = GradientUnits(ObjectBoundingBox, 'ObjectBoundingBox')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_25, __pyx_n_s_GradientUnits); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(__pyx_e_7celiagg_8_celiagg_UserSpace); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_UserSpace);
    __Pyx_GIVEREF(__pyx_n_s_UserSpace);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_n_s_UserSpace);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_25, __pyx_t_26, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 62, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_UserSpace, __pyx_t_28) < 0)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":63
 *         pass
 *     __Pyx_globals['UserSpace'] = GradientUnits(UserSpace, 'UserSpace')
 *     __Pyx_globals['ObjectBoundingBox'] = GradientUnits(ObjectBoundingBox, 'ObjectBoundingBox')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_GradientUnits); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(__pyx_e_7celiagg_8_celiagg_ObjectBoundingBox); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_ObjectBoundingBox);
    __Pyx_GIVEREF(__pyx_n_s_ObjectBoundingBox);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_n_s_ObjectBoundingBox);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_25, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 63, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ObjectBoundingBox, __pyx_t_26) < 0)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  }
  __pyx_L8:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_26 = __Pyx_Globals(); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (!(likely(PyDict_CheckExact(__pyx_t_26))||((__pyx_t_26) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_26)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_26));
  __Pyx_GIVEREF(__pyx_t_26);
  __pyx_t_26 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     InnerJoin = __Pyx_EnumBase('InnerJoin', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     InnerJoin = __Pyx_EnumBase('InnerJoin', __Pyx_OrderedDict([
 *         ('InnerBevel', InnerBevel),             # <<<<<<<<<<<<<<
 *         ('InnerMiter', InnerMiter),
 *         ('InnerJag', InnerJag),
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(__pyx_e_7celiagg_8_celiagg_InnerBevel); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_InnerBevel);
    __Pyx_GIVEREF(__pyx_n_s_InnerBevel);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_InnerBevel);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":55
 *     InnerJoin = __Pyx_EnumBase('InnerJoin', __Pyx_OrderedDict([
 *         ('InnerBevel', InnerBevel),
 *         ('InnerMiter', InnerMiter),             # <<<<<<<<<<<<<<
 *         ('InnerJag', InnerJag),
 *         ('InnerRound', InnerRound),
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(__pyx_e_7celiagg_8_celiagg_InnerMiter); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_InnerMiter);
    __Pyx_GIVEREF(__pyx_n_s_InnerMiter);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_InnerMiter);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":56
 *         ('InnerBevel', InnerBevel),
 *         ('InnerMiter', InnerMiter),
 *         ('InnerJag', InnerJag),             # <<<<<<<<<<<<<<
 *         ('InnerRound', InnerRound),
 *     ]))
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(__pyx_e_7celiagg_8_celiagg_InnerJag); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_InnerJag);
    __Pyx_GIVEREF(__pyx_n_s_InnerJag);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_InnerJag);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":57
 *         ('InnerMiter', InnerMiter),
 *         ('InnerJag', InnerJag),
 *         ('InnerRound', InnerRound),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['InnerBevel'] = InnerJoin.InnerBevel
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(__pyx_e_7celiagg_8_celiagg_InnerRound); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_InnerRound);
    __Pyx_GIVEREF(__pyx_n_s_InnerRound);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_InnerRound);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     InnerJoin = __Pyx_EnumBase('InnerJoin', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('InnerBevel', InnerBevel),
 *         ('InnerMiter', InnerMiter),
 */
    __pyx_t_26 = PyList_New(4); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_26, 0, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_28);
    PyList_SET_ITEM(__pyx_t_26, 1, __pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_26, 2, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_26, 3, __pyx_t_2);
    __pyx_t_25 = 0;
    __pyx_t_28 = 0;
    __pyx_t_24 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_26); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_InnerJoin);
    __Pyx_GIVEREF(__pyx_n_s_InnerJoin);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_InnerJoin);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_26, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_InnerJoin, __pyx_t_2) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":59
 *         ('InnerRound', InnerRound),
 *     ]))
 *     __Pyx_globals['InnerBevel'] = InnerJoin.InnerBevel             # <<<<<<<<<<<<<<
 *     __Pyx_globals['InnerMiter'] = InnerJoin.InnerMiter
 *     __Pyx_globals['InnerJag'] = InnerJoin.InnerJag
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_InnerJoin); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_InnerBevel); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 59, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_InnerBevel, __pyx_t_26) < 0)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":60
 *     ]))
 *     __Pyx_globals['InnerBevel'] = InnerJoin.InnerBevel
 *     __Pyx_globals['InnerMiter'] = InnerJoin.InnerMiter             # <<<<<<<<<<<<<<
 *     __Pyx_globals['InnerJag'] = InnerJoin.InnerJag
 *     __Pyx_globals['InnerRound'] = InnerJoin.InnerRound
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_InnerJoin); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_n_s_InnerMiter); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 60, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_InnerMiter, __pyx_t_2) < 0)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":61
 *     __Pyx_globals['InnerBevel'] = InnerJoin.InnerBevel
 *     __Pyx_globals['InnerMiter'] = InnerJoin.InnerMiter
 *     __Pyx_globals['InnerJag'] = InnerJoin.InnerJag             # <<<<<<<<<<<<<<
 *     __Pyx_globals['InnerRound'] = InnerJoin.InnerRound
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_InnerJoin); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_InnerJag); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 61, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_InnerJag, __pyx_t_26) < 0)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":62
 *     __Pyx_globals['InnerMiter'] = InnerJoin.InnerMiter
 *     __Pyx_globals['InnerJag'] = InnerJoin.InnerJag
 *     __Pyx_globals['InnerRound'] = InnerJoin.InnerRound             # <<<<<<<<<<<<<<
 * else:
 *     class InnerJoin(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_InnerJoin); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_n_s_InnerRound); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 62, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_InnerRound, __pyx_t_2) < 0)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     InnerJoin = __Pyx_EnumBase('InnerJoin', __Pyx_OrderedDict([
 */
    goto __pyx_L9;
  }

  /* "EnumType":64
 *     __Pyx_globals['InnerRound'] = InnerJoin.InnerRound
 * else:
 *     class InnerJoin(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['InnerBevel'] = InnerJoin(InnerBevel, 'InnerBevel')
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __Pyx_EnumBase);
    __pyx_t_26 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_Py3MetaclassPrepare(__pyx_t_26, __pyx_t_2, __pyx_n_s_InnerJoin, __pyx_n_s_InnerJoin, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_28 = __Pyx_Py3ClassCreate(__pyx_t_26, __pyx_n_s_InnerJoin, __pyx_t_2, __pyx_t_24, NULL, 0, 1); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_InnerJoin, __pyx_t_28) < 0) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":66
 *     class InnerJoin(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['InnerBevel'] = InnerJoin(InnerBevel, 'InnerBevel')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['InnerMiter'] = InnerJoin(InnerMiter, 'InnerMiter')
 *     __Pyx_globals['InnerJag'] = InnerJoin(InnerJag, 'InnerJag')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_InnerJoin); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(__pyx_e_7celiagg_8_celiagg_InnerBevel); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_InnerBevel);
    __Pyx_GIVEREF(__pyx_n_s_InnerBevel);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_InnerBevel);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_24, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 66, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_InnerBevel, __pyx_t_26) < 0)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":67
 *         pass
 *     __Pyx_globals['InnerBevel'] = InnerJoin(InnerBevel, 'InnerBevel')
 *     __Pyx_globals['InnerMiter'] = InnerJoin(InnerMiter, 'InnerMiter')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['InnerJag'] = InnerJoin(InnerJag, 'InnerJag')
 *     __Pyx_globals['InnerRound'] = InnerJoin(InnerRound, 'InnerRound')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_InnerJoin); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(__pyx_e_7celiagg_8_celiagg_InnerMiter); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_InnerMiter);
    __Pyx_GIVEREF(__pyx_n_s_InnerMiter);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_InnerMiter);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_t_2, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 67, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_InnerMiter, __pyx_t_24) < 0)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":68
 *     __Pyx_globals['InnerBevel'] = InnerJoin(InnerBevel, 'InnerBevel')
 *     __Pyx_globals['InnerMiter'] = InnerJoin(InnerMiter, 'InnerMiter')
 *     __Pyx_globals['InnerJag'] = InnerJoin(InnerJag, 'InnerJag')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['InnerRound'] = InnerJoin(InnerRound, 'InnerRound')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_InnerJoin); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(__pyx_e_7celiagg_8_celiagg_InnerJag); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_InnerJag);
    __Pyx_GIVEREF(__pyx_n_s_InnerJag);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_n_s_InnerJag);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_26, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 68, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_InnerJag, __pyx_t_2) < 0)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":69
 *     __Pyx_globals['InnerMiter'] = InnerJoin(InnerMiter, 'InnerMiter')
 *     __Pyx_globals['InnerJag'] = InnerJoin(InnerJag, 'InnerJag')
 *     __Pyx_globals['InnerRound'] = InnerJoin(InnerRound, 'InnerRound')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_InnerJoin); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(__pyx_e_7celiagg_8_celiagg_InnerRound); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_InnerRound);
    __Pyx_GIVEREF(__pyx_n_s_InnerRound);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_InnerRound);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_24, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 69, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_InnerRound, __pyx_t_26) < 0)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  }
  __pyx_L9:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_26 = __Pyx_Globals(); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (!(likely(PyDict_CheckExact(__pyx_t_26))||((__pyx_t_26) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_26)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_26));
  __Pyx_GIVEREF(__pyx_t_26);
  __pyx_t_26 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     LineCap = __Pyx_EnumBase('LineCap', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     LineCap = __Pyx_EnumBase('LineCap', __Pyx_OrderedDict([
 *         ('CapButt', CapButt),             # <<<<<<<<<<<<<<
 *         ('CapSquare', CapSquare),
 *         ('CapRound', CapRound),
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineCap(__pyx_e_7celiagg_8_celiagg_CapButt); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_CapButt);
    __Pyx_GIVEREF(__pyx_n_s_CapButt);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_CapButt);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":55
 *     LineCap = __Pyx_EnumBase('LineCap', __Pyx_OrderedDict([
 *         ('CapButt', CapButt),
 *         ('CapSquare', CapSquare),             # <<<<<<<<<<<<<<
 *         ('CapRound', CapRound),
 *     ]))
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineCap(__pyx_e_7celiagg_8_celiagg_CapSquare); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_CapSquare);
    __Pyx_GIVEREF(__pyx_n_s_CapSquare);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_CapSquare);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":56
 *         ('CapButt', CapButt),
 *         ('CapSquare', CapSquare),
 *         ('CapRound', CapRound),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['CapButt'] = LineCap.CapButt
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineCap(__pyx_e_7celiagg_8_celiagg_CapRound); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_CapRound);
    __Pyx_GIVEREF(__pyx_n_s_CapRound);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_CapRound);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     LineCap = __Pyx_EnumBase('LineCap', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('CapButt', CapButt),
 *         ('CapSquare', CapSquare),
 */
    __pyx_t_26 = PyList_New(3); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_26, 0, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_26, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_28);
    PyList_SET_ITEM(__pyx_t_26, 2, __pyx_t_28);
    __pyx_t_24 = 0;
    __pyx_t_2 = 0;
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_26); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_LineCap);
    __Pyx_GIVEREF(__pyx_n_s_LineCap);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_LineCap);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_28);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_26, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_LineCap, __pyx_t_28) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":58
 *         ('CapRound', CapRound),
 *     ]))
 *     __Pyx_globals['CapButt'] = LineCap.CapButt             # <<<<<<<<<<<<<<
 *     __Pyx_globals['CapSquare'] = LineCap.CapSquare
 *     __Pyx_globals['CapRound'] = LineCap.CapRound
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_LineCap); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_CapButt); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 58, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_CapButt, __pyx_t_26) < 0)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":59
 *     ]))
 *     __Pyx_globals['CapButt'] = LineCap.CapButt
 *     __Pyx_globals['CapSquare'] = LineCap.CapSquare             # <<<<<<<<<<<<<<
 *     __Pyx_globals['CapRound'] = LineCap.CapRound
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_LineCap); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_28 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_n_s_CapSquare); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 59, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_CapSquare, __pyx_t_28) < 0)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":60
 *     __Pyx_globals['CapButt'] = LineCap.CapButt
 *     __Pyx_globals['CapSquare'] = LineCap.CapSquare
 *     __Pyx_globals['CapRound'] = LineCap.CapRound             # <<<<<<<<<<<<<<
 * else:
 *     class LineCap(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_LineCap); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_28, __pyx_n_s_CapRound); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 60, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_CapRound, __pyx_t_26) < 0)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     LineCap = __Pyx_EnumBase('LineCap', __Pyx_OrderedDict([
 */
    goto __pyx_L10;
  }

  /* "EnumType":62
 *     __Pyx_globals['CapRound'] = LineCap.CapRound
 * else:
 *     class LineCap(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['CapButt'] = LineCap(CapButt, 'CapButt')
 */
  /*else*/ {
    __pyx_t_26 = PyTuple_New(1); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __Pyx_EnumBase);
    __pyx_t_28 = __Pyx_CalculateMetaclass(NULL, __pyx_t_26); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_28, __pyx_t_26, __pyx_n_s_LineCap, __pyx_n_s_LineCap, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_24 = __Pyx_Py3ClassCreate(__pyx_t_28, __pyx_n_s_LineCap, __pyx_t_26, __pyx_t_2, NULL, 0, 1); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_LineCap, __pyx_t_24) < 0) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":64
 *     class LineCap(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['CapButt'] = LineCap(CapButt, 'CapButt')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['CapSquare'] = LineCap(CapSquare, 'CapSquare')
 *     __Pyx_globals['CapRound'] = LineCap(CapRound, 'CapRound')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_LineCap); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_28 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineCap(__pyx_e_7celiagg_8_celiagg_CapButt); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_28);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_CapButt);
    __Pyx_GIVEREF(__pyx_n_s_CapButt);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_CapButt);
    __pyx_t_28 = 0;
    __pyx_t_28 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_t_2, NULL); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 64, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_CapButt, __pyx_t_28) < 0)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

    /* "EnumType":65
 *         pass
 *     __Pyx_globals['CapButt'] = LineCap(CapButt, 'CapButt')
 *     __Pyx_globals['CapSquare'] = LineCap(CapSquare, 'CapSquare')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['CapRound'] = LineCap(CapRound, 'CapRound')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_LineCap); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineCap(__pyx_e_7celiagg_8_celiagg_CapSquare); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_CapSquare);
    __Pyx_GIVEREF(__pyx_n_s_CapSquare);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_n_s_CapSquare);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_28, __pyx_t_26, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 65, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_CapSquare, __pyx_t_2) < 0)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":66
 *     __Pyx_globals['CapButt'] = LineCap(CapButt, 'CapButt')
 *     __Pyx_globals['CapSquare'] = LineCap(CapSquare, 'CapSquare')
 *     __Pyx_globals['CapRound'] = LineCap(CapRound, 'CapRound')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LineCap); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineCap(__pyx_e_7celiagg_8_celiagg_CapRound); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_CapRound);
    __Pyx_GIVEREF(__pyx_n_s_CapRound);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_n_s_CapRound);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_28, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 66, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_CapRound, __pyx_t_26) < 0)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  }
  __pyx_L10:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_26 = __Pyx_Globals(); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (!(likely(PyDict_CheckExact(__pyx_t_26))||((__pyx_t_26) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_26)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_26));
  __Pyx_GIVEREF(__pyx_t_26);
  __pyx_t_26 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     LineJoin = __Pyx_EnumBase('LineJoin', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     LineJoin = __Pyx_EnumBase('LineJoin', __Pyx_OrderedDict([
 *         ('JoinMiter', JoinMiter),             # <<<<<<<<<<<<<<
 *         ('JoinRound', JoinRound),
 *         ('JoinBevel', JoinBevel),
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineJoin(__pyx_e_7celiagg_8_celiagg_JoinMiter); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_JoinMiter);
    __Pyx_GIVEREF(__pyx_n_s_JoinMiter);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_JoinMiter);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":55
 *     LineJoin = __Pyx_EnumBase('LineJoin', __Pyx_OrderedDict([
 *         ('JoinMiter', JoinMiter),
 *         ('JoinRound', JoinRound),             # <<<<<<<<<<<<<<
 *         ('JoinBevel', JoinBevel),
 *     ]))
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineJoin(__pyx_e_7celiagg_8_celiagg_JoinRound); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_JoinRound);
    __Pyx_GIVEREF(__pyx_n_s_JoinRound);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_JoinRound);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":56
 *         ('JoinMiter', JoinMiter),
 *         ('JoinRound', JoinRound),
 *         ('JoinBevel', JoinBevel),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['JoinMiter'] = LineJoin.JoinMiter
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineJoin(__pyx_e_7celiagg_8_celiagg_JoinBevel); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_JoinBevel);
    __Pyx_GIVEREF(__pyx_n_s_JoinBevel);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_JoinBevel);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     LineJoin = __Pyx_EnumBase('LineJoin', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('JoinMiter', JoinMiter),
 *         ('JoinRound', JoinRound),
 */
    __pyx_t_26 = PyList_New(3); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_28);
    PyList_SET_ITEM(__pyx_t_26, 0, __pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_26, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_26, 2, __pyx_t_24);
    __pyx_t_28 = 0;
    __pyx_t_2 = 0;
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_26); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_LineJoin);
    __Pyx_GIVEREF(__pyx_n_s_LineJoin);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_LineJoin);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_24);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_26, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_LineJoin, __pyx_t_24) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":58
 *         ('JoinBevel', JoinBevel),
 *     ]))
 *     __Pyx_globals['JoinMiter'] = LineJoin.JoinMiter             # <<<<<<<<<<<<<<
 *     __Pyx_globals['JoinRound'] = LineJoin.JoinRound
 *     __Pyx_globals['JoinBevel'] = LineJoin.JoinBevel
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_LineJoin); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_JoinMiter); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 58, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_JoinMiter, __pyx_t_26) < 0)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":59
 *     ]))
 *     __Pyx_globals['JoinMiter'] = LineJoin.JoinMiter
 *     __Pyx_globals['JoinRound'] = LineJoin.JoinRound             # <<<<<<<<<<<<<<
 *     __Pyx_globals['JoinBevel'] = LineJoin.JoinBevel
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_LineJoin); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_n_s_JoinRound); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 59, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_JoinRound, __pyx_t_24) < 0)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":60
 *     __Pyx_globals['JoinMiter'] = LineJoin.JoinMiter
 *     __Pyx_globals['JoinRound'] = LineJoin.JoinRound
 *     __Pyx_globals['JoinBevel'] = LineJoin.JoinBevel             # <<<<<<<<<<<<<<
 * else:
 *     class LineJoin(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_LineJoin); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_JoinBevel); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 60, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_JoinBevel, __pyx_t_26) < 0)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     LineJoin = __Pyx_EnumBase('LineJoin', __Pyx_OrderedDict([
 */
    goto __pyx_L11;
  }

  /* "EnumType":62
 *     __Pyx_globals['JoinBevel'] = LineJoin.JoinBevel
 * else:
 *     class LineJoin(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['JoinMiter'] = LineJoin(JoinMiter, 'JoinMiter')
 */
  /*else*/ {
    __pyx_t_26 = PyTuple_New(1); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __Pyx_EnumBase);
    __pyx_t_24 = __Pyx_CalculateMetaclass(NULL, __pyx_t_26); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_24, __pyx_t_26, __pyx_n_s_LineJoin, __pyx_n_s_LineJoin, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_28 = __Pyx_Py3ClassCreate(__pyx_t_24, __pyx_n_s_LineJoin, __pyx_t_26, __pyx_t_2, NULL, 0, 1); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_LineJoin, __pyx_t_28) < 0) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":64
 *     class LineJoin(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['JoinMiter'] = LineJoin(JoinMiter, 'JoinMiter')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['JoinRound'] = LineJoin(JoinRound, 'JoinRound')
 *     __Pyx_globals['JoinBevel'] = LineJoin(JoinBevel, 'JoinBevel')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_LineJoin); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineJoin(__pyx_e_7celiagg_8_celiagg_JoinMiter); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_JoinMiter);
    __Pyx_GIVEREF(__pyx_n_s_JoinMiter);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_JoinMiter);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_t_2, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 64, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_JoinMiter, __pyx_t_24) < 0)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":65
 *         pass
 *     __Pyx_globals['JoinMiter'] = LineJoin(JoinMiter, 'JoinMiter')
 *     __Pyx_globals['JoinRound'] = LineJoin(JoinRound, 'JoinRound')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['JoinBevel'] = LineJoin(JoinBevel, 'JoinBevel')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_LineJoin); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineJoin(__pyx_e_7celiagg_8_celiagg_JoinRound); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_JoinRound);
    __Pyx_GIVEREF(__pyx_n_s_JoinRound);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_n_s_JoinRound);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_26, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 65, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_JoinRound, __pyx_t_2) < 0)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":66
 *     __Pyx_globals['JoinMiter'] = LineJoin(JoinMiter, 'JoinMiter')
 *     __Pyx_globals['JoinRound'] = LineJoin(JoinRound, 'JoinRound')
 *     __Pyx_globals['JoinBevel'] = LineJoin(JoinBevel, 'JoinBevel')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_LineJoin); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineJoin(__pyx_e_7celiagg_8_celiagg_JoinBevel); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_JoinBevel);
    __Pyx_GIVEREF(__pyx_n_s_JoinBevel);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_JoinBevel);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_24, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 66, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_JoinBevel, __pyx_t_26) < 0)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
  }
  __pyx_L11:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_26 = __Pyx_Globals(); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_26);
  if (!(likely(PyDict_CheckExact(__pyx_t_26))||((__pyx_t_26) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_26)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_26));
  __Pyx_GIVEREF(__pyx_t_26);
  __pyx_t_26 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     PatternStyle = __Pyx_EnumBase('PatternStyle', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     PatternStyle = __Pyx_EnumBase('PatternStyle', __Pyx_OrderedDict([
 *         ('StyleRepeat', StyleRepeat),             # <<<<<<<<<<<<<<
 *         ('StyleReflect', StyleReflect),
 *     ]))
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(__pyx_e_7celiagg_8_celiagg_StyleRepeat); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_StyleRepeat);
    __Pyx_GIVEREF(__pyx_n_s_StyleRepeat);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_StyleRepeat);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":55
 *     PatternStyle = __Pyx_EnumBase('PatternStyle', __Pyx_OrderedDict([
 *         ('StyleRepeat', StyleRepeat),
 *         ('StyleReflect', StyleReflect),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['StyleRepeat'] = PatternStyle.StyleRepeat
 */
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(__pyx_e_7celiagg_8_celiagg_StyleReflect); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_StyleReflect);
    __Pyx_GIVEREF(__pyx_n_s_StyleReflect);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_StyleReflect);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_26);
    __pyx_t_26 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     PatternStyle = __Pyx_EnumBase('PatternStyle', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('StyleRepeat', StyleRepeat),
 *         ('StyleReflect', StyleReflect),
 */
    __pyx_t_26 = PyList_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_24);
    PyList_SET_ITEM(__pyx_t_26, 0, __pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_26, 1, __pyx_t_2);
    __pyx_t_24 = 0;
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_26); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_PatternStyle);
    __Pyx_GIVEREF(__pyx_n_s_PatternStyle);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_PatternStyle);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_26, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_PatternStyle, __pyx_t_2) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":57
 *         ('StyleReflect', StyleReflect),
 *     ]))
 *     __Pyx_globals['StyleRepeat'] = PatternStyle.StyleRepeat             # <<<<<<<<<<<<<<
 *     __Pyx_globals['StyleReflect'] = PatternStyle.StyleReflect
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PatternStyle); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_StyleRepeat); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 57, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_StyleRepeat, __pyx_t_26) < 0)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":58
 *     ]))
 *     __Pyx_globals['StyleRepeat'] = PatternStyle.StyleRepeat
 *     __Pyx_globals['StyleReflect'] = PatternStyle.StyleReflect             # <<<<<<<<<<<<<<
 * else:
 *     class PatternStyle(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_PatternStyle); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_n_s_StyleReflect); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 58, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_StyleReflect, __pyx_t_2) < 0)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     PatternStyle = __Pyx_EnumBase('PatternStyle', __Pyx_OrderedDict([
 */
    goto __pyx_L12;
  }

  /* "EnumType":60
 *     __Pyx_globals['StyleReflect'] = PatternStyle.StyleReflect
 * else:
 *     class PatternStyle(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['StyleRepeat'] = PatternStyle(StyleRepeat, 'StyleRepeat')
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __Pyx_EnumBase);
    __pyx_t_26 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_Py3MetaclassPrepare(__pyx_t_26, __pyx_t_2, __pyx_n_s_PatternStyle, __pyx_n_s_PatternStyle, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_28 = __Pyx_Py3ClassCreate(__pyx_t_26, __pyx_n_s_PatternStyle, __pyx_t_2, __pyx_t_24, NULL, 0, 1); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_PatternStyle, __pyx_t_28) < 0) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "EnumType":62
 *     class PatternStyle(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['StyleRepeat'] = PatternStyle(StyleRepeat, 'StyleRepeat')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['StyleReflect'] = PatternStyle(StyleReflect, 'StyleReflect')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PatternStyle); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_26 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(__pyx_e_7celiagg_8_celiagg_StyleRepeat); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_26);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_StyleRepeat);
    __Pyx_GIVEREF(__pyx_n_s_StyleRepeat);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_StyleRepeat);
    __pyx_t_26 = 0;
    __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_24, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 62, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_StyleRepeat, __pyx_t_26) < 0)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

    /* "EnumType":63
 *         pass
 *     __Pyx_globals['StyleRepeat'] = PatternStyle(StyleRepeat, 'StyleRepeat')
 *     __Pyx_globals['StyleReflect'] = PatternStyle(StyleReflect, 'StyleReflect')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_PatternStyle); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(__pyx_e_7celiagg_8_celiagg_StyleReflect); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_StyleReflect);
    __Pyx_GIVEREF(__pyx_n_s_StyleReflect);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_StyleReflect);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_26, __pyx_t_2, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 63, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_StyleReflect, __pyx_t_24) < 0)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  }
  __pyx_L12:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_24 = __Pyx_Globals(); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (!(likely(PyDict_CheckExact(__pyx_t_24))||((__pyx_t_24) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_24)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_24));
  __Pyx_GIVEREF(__pyx_t_24);
  __pyx_t_24 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     PixelFormat = __Pyx_EnumBase('PixelFormat', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     PixelFormat = __Pyx_EnumBase('PixelFormat', __Pyx_OrderedDict([
 *         ('Gray8', Gray8),             # <<<<<<<<<<<<<<
 *         ('Gray16', Gray16),
 *         ('Gray32', Gray32),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray8); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_Gray8);
    __Pyx_GIVEREF(__pyx_n_s_Gray8);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Gray8);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":55
 *     PixelFormat = __Pyx_EnumBase('PixelFormat', __Pyx_OrderedDict([
 *         ('Gray8', Gray8),
 *         ('Gray16', Gray16),             # <<<<<<<<<<<<<<
 *         ('Gray32', Gray32),
 *         ('BGR24', BGR24),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray16); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_26 = PyTuple_New(2); if (unlikely(!__pyx_t_26)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_26);
    __Pyx_INCREF(__pyx_n_s_Gray16);
    __Pyx_GIVEREF(__pyx_n_s_Gray16);
    PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_n_s_Gray16);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_26, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":56
 *         ('Gray8', Gray8),
 *         ('Gray16', Gray16),
 *         ('Gray32', Gray32),             # <<<<<<<<<<<<<<
 *         ('BGR24', BGR24),
 *         ('RGB24', RGB24),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_28 = PyTuple_New(2); if (unlikely(!__pyx_t_28)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_28);
    __Pyx_INCREF(__pyx_n_s_Gray32);
    __Pyx_GIVEREF(__pyx_n_s_Gray32);
    PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_n_s_Gray32);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_28, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":57
 *         ('Gray16', Gray16),
 *         ('Gray32', Gray32),
 *         ('BGR24', BGR24),             # <<<<<<<<<<<<<<
 *         ('RGB24', RGB24),
 *         ('BGR48', BGR48),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR24); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_25 = PyTuple_New(2); if (unlikely(!__pyx_t_25)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_25);
    __Pyx_INCREF(__pyx_n_s_BGR24);
    __Pyx_GIVEREF(__pyx_n_s_BGR24);
    PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_n_s_BGR24);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_25, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":58
 *         ('Gray32', Gray32),
 *         ('BGR24', BGR24),
 *         ('RGB24', RGB24),             # <<<<<<<<<<<<<<
 *         ('BGR48', BGR48),
 *         ('RGB48', RGB48),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB24); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_27 = PyTuple_New(2); if (unlikely(!__pyx_t_27)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_27);
    __Pyx_INCREF(__pyx_n_s_RGB24);
    __Pyx_GIVEREF(__pyx_n_s_RGB24);
    PyTuple_SET_ITEM(__pyx_t_27, 0, __pyx_n_s_RGB24);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_27, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":59
 *         ('BGR24', BGR24),
 *         ('RGB24', RGB24),
 *         ('BGR48', BGR48),             # <<<<<<<<<<<<<<
 *         ('RGB48', RGB48),
 *         ('BGR96', BGR96),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR48); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_23 = PyTuple_New(2); if (unlikely(!__pyx_t_23)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_n_s_BGR48);
    __Pyx_GIVEREF(__pyx_n_s_BGR48);
    PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_n_s_BGR48);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_23, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":60
 *         ('RGB24', RGB24),
 *         ('BGR48', BGR48),
 *         ('RGB48', RGB48),             # <<<<<<<<<<<<<<
 *         ('BGR96', BGR96),
 *         ('RGB96', RGB96),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB48); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_22 = PyTuple_New(2); if (unlikely(!__pyx_t_22)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __Pyx_INCREF(__pyx_n_s_RGB48);
    __Pyx_GIVEREF(__pyx_n_s_RGB48);
    PyTuple_SET_ITEM(__pyx_t_22, 0, __pyx_n_s_RGB48);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_22, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":61
 *         ('BGR48', BGR48),
 *         ('RGB48', RGB48),
 *         ('BGR96', BGR96),             # <<<<<<<<<<<<<<
 *         ('RGB96', RGB96),
 *         ('BGRA32', BGRA32),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR96); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_21 = PyTuple_New(2); if (unlikely(!__pyx_t_21)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_INCREF(__pyx_n_s_BGR96);
    __Pyx_GIVEREF(__pyx_n_s_BGR96);
    PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_n_s_BGR96);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_21, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":62
 *         ('RGB48', RGB48),
 *         ('BGR96', BGR96),
 *         ('RGB96', RGB96),             # <<<<<<<<<<<<<<
 *         ('BGRA32', BGRA32),
 *         ('RGBA32', RGBA32),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB96); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_20 = PyTuple_New(2); if (unlikely(!__pyx_t_20)) __PYX_ERR(4, 62, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    __Pyx_INCREF(__pyx_n_s_RGB96);
    __Pyx_GIVEREF(__pyx_n_s_RGB96);
    PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_n_s_RGB96);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_20, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":63
 *         ('BGR96', BGR96),
 *         ('RGB96', RGB96),
 *         ('BGRA32', BGRA32),             # <<<<<<<<<<<<<<
 *         ('RGBA32', RGBA32),
 *         ('ARGB32', ARGB32),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_19 = PyTuple_New(2); if (unlikely(!__pyx_t_19)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    __Pyx_INCREF(__pyx_n_s_BGRA32);
    __Pyx_GIVEREF(__pyx_n_s_BGRA32);
    PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_n_s_BGRA32);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_19, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":64
 *         ('RGB96', RGB96),
 *         ('BGRA32', BGRA32),
 *         ('RGBA32', RGBA32),             # <<<<<<<<<<<<<<
 *         ('ARGB32', ARGB32),
 *         ('ABGR32', ABGR32),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_18 = PyTuple_New(2); if (unlikely(!__pyx_t_18)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_INCREF(__pyx_n_s_RGBA32);
    __Pyx_GIVEREF(__pyx_n_s_RGBA32);
    PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_n_s_RGBA32);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_18, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":65
 *         ('BGRA32', BGRA32),
 *         ('RGBA32', RGBA32),
 *         ('ARGB32', ARGB32),             # <<<<<<<<<<<<<<
 *         ('ABGR32', ABGR32),
 *         ('BGRA64', BGRA64),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_17 = PyTuple_New(2); if (unlikely(!__pyx_t_17)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_INCREF(__pyx_n_s_ARGB32);
    __Pyx_GIVEREF(__pyx_n_s_ARGB32);
    PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_n_s_ARGB32);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":66
 *         ('RGBA32', RGBA32),
 *         ('ARGB32', ARGB32),
 *         ('ABGR32', ABGR32),             # <<<<<<<<<<<<<<
 *         ('BGRA64', BGRA64),
 *         ('RGBA64', RGBA64),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_n_s_ABGR32);
    __Pyx_GIVEREF(__pyx_n_s_ABGR32);
    PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_n_s_ABGR32);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":67
 *         ('ARGB32', ARGB32),
 *         ('ABGR32', ABGR32),
 *         ('BGRA64', BGRA64),             # <<<<<<<<<<<<<<
 *         ('RGBA64', RGBA64),
 *         ('ARGB64', ARGB64),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA64); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_BGRA64);
    __Pyx_GIVEREF(__pyx_n_s_BGRA64);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_s_BGRA64);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":68
 *         ('ABGR32', ABGR32),
 *         ('BGRA64', BGRA64),
 *         ('RGBA64', RGBA64),             # <<<<<<<<<<<<<<
 *         ('ARGB64', ARGB64),
 *         ('ABGR64', ABGR64),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA64); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_RGBA64);
    __Pyx_GIVEREF(__pyx_n_s_RGBA64);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_s_RGBA64);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":69
 *         ('BGRA64', BGRA64),
 *         ('RGBA64', RGBA64),
 *         ('ARGB64', ARGB64),             # <<<<<<<<<<<<<<
 *         ('ABGR64', ABGR64),
 *         ('BGRA128', BGRA128),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB64); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_ARGB64);
    __Pyx_GIVEREF(__pyx_n_s_ARGB64);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_ARGB64);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":70
 *         ('RGBA64', RGBA64),
 *         ('ARGB64', ARGB64),
 *         ('ABGR64', ABGR64),             # <<<<<<<<<<<<<<
 *         ('BGRA128', BGRA128),
 *         ('RGBA128', RGBA128),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR64); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_ABGR64);
    __Pyx_GIVEREF(__pyx_n_s_ABGR64);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_ABGR64);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":71
 *         ('ARGB64', ARGB64),
 *         ('ABGR64', ABGR64),
 *         ('BGRA128', BGRA128),             # <<<<<<<<<<<<<<
 *         ('RGBA128', RGBA128),
 *         ('ARGB128', ARGB128),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA128); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_n_s_BGRA128);
    __Pyx_GIVEREF(__pyx_n_s_BGRA128);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_BGRA128);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":72
 *         ('ABGR64', ABGR64),
 *         ('BGRA128', BGRA128),
 *         ('RGBA128', RGBA128),             # <<<<<<<<<<<<<<
 *         ('ARGB128', ARGB128),
 *         ('ABGR128', ABGR128),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA128); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_n_s_RGBA128);
    __Pyx_GIVEREF(__pyx_n_s_RGBA128);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_RGBA128);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":73
 *         ('BGRA128', BGRA128),
 *         ('RGBA128', RGBA128),
 *         ('ARGB128', ARGB128),             # <<<<<<<<<<<<<<
 *         ('ABGR128', ABGR128),
 *     ]))
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB128); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 73, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_ARGB128);
    __Pyx_GIVEREF(__pyx_n_s_ARGB128);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_ARGB128);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":74
 *         ('RGBA128', RGBA128),
 *         ('ARGB128', ARGB128),
 *         ('ABGR128', ABGR128),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['Gray8'] = PixelFormat.Gray8
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR128); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_ABGR128);
    __Pyx_GIVEREF(__pyx_n_s_ABGR128);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_ABGR128);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     PixelFormat = __Pyx_EnumBase('PixelFormat', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('Gray8', Gray8),
 *         ('Gray16', Gray16),
 */
    __pyx_t_24 = PyList_New(21); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_24, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_26);
    PyList_SET_ITEM(__pyx_t_24, 1, __pyx_t_26);
    __Pyx_GIVEREF(__pyx_t_28);
    PyList_SET_ITEM(__pyx_t_24, 2, __pyx_t_28);
    __Pyx_GIVEREF(__pyx_t_25);
    PyList_SET_ITEM(__pyx_t_24, 3, __pyx_t_25);
    __Pyx_GIVEREF(__pyx_t_27);
    PyList_SET_ITEM(__pyx_t_24, 4, __pyx_t_27);
    __Pyx_GIVEREF(__pyx_t_23);
    PyList_SET_ITEM(__pyx_t_24, 5, __pyx_t_23);
    __Pyx_GIVEREF(__pyx_t_22);
    PyList_SET_ITEM(__pyx_t_24, 6, __pyx_t_22);
    __Pyx_GIVEREF(__pyx_t_21);
    PyList_SET_ITEM(__pyx_t_24, 7, __pyx_t_21);
    __Pyx_GIVEREF(__pyx_t_20);
    PyList_SET_ITEM(__pyx_t_24, 8, __pyx_t_20);
    __Pyx_GIVEREF(__pyx_t_19);
    PyList_SET_ITEM(__pyx_t_24, 9, __pyx_t_19);
    __Pyx_GIVEREF(__pyx_t_18);
    PyList_SET_ITEM(__pyx_t_24, 10, __pyx_t_18);
    __Pyx_GIVEREF(__pyx_t_17);
    PyList_SET_ITEM(__pyx_t_24, 11, __pyx_t_17);
    __Pyx_GIVEREF(__pyx_t_16);
    PyList_SET_ITEM(__pyx_t_24, 12, __pyx_t_16);
    __Pyx_GIVEREF(__pyx_t_15);
    PyList_SET_ITEM(__pyx_t_24, 13, __pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_14);
    PyList_SET_ITEM(__pyx_t_24, 14, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_13);
    PyList_SET_ITEM(__pyx_t_24, 15, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_12);
    PyList_SET_ITEM(__pyx_t_24, 16, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_11);
    PyList_SET_ITEM(__pyx_t_24, 17, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_10);
    PyList_SET_ITEM(__pyx_t_24, 18, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_24, 19, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyList_SET_ITEM(__pyx_t_24, 20, __pyx_t_8);
    __pyx_t_2 = 0;
    __pyx_t_26 = 0;
    __pyx_t_28 = 0;
    __pyx_t_25 = 0;
    __pyx_t_27 = 0;
    __pyx_t_23 = 0;
    __pyx_t_22 = 0;
    __pyx_t_21 = 0;
    __pyx_t_20 = 0;
    __pyx_t_19 = 0;
    __pyx_t_18 = 0;
    __pyx_t_17 = 0;
    __pyx_t_16 = 0;
    __pyx_t_15 = 0;
    __pyx_t_14 = 0;
    __pyx_t_13 = 0;
    __pyx_t_12 = 0;
    __pyx_t_11 = 0;
    __pyx_t_10 = 0;
    __pyx_t_9 = 0;
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_24); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_PixelFormat);
    __Pyx_GIVEREF(__pyx_n_s_PixelFormat);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_PixelFormat);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_24, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_PixelFormat, __pyx_t_8) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":76
 *         ('ABGR128', ABGR128),
 *     ]))
 *     __Pyx_globals['Gray8'] = PixelFormat.Gray8             # <<<<<<<<<<<<<<
 *     __Pyx_globals['Gray16'] = PixelFormat.Gray16
 *     __Pyx_globals['Gray32'] = PixelFormat.Gray32
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Gray8); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 76, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Gray8, __pyx_t_24) < 0)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":77
 *     ]))
 *     __Pyx_globals['Gray8'] = PixelFormat.Gray8
 *     __Pyx_globals['Gray16'] = PixelFormat.Gray16             # <<<<<<<<<<<<<<
 *     __Pyx_globals['Gray32'] = PixelFormat.Gray32
 *     __Pyx_globals['BGR24'] = PixelFormat.BGR24
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_Gray16); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 77, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Gray16, __pyx_t_8) < 0)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":78
 *     __Pyx_globals['Gray8'] = PixelFormat.Gray8
 *     __Pyx_globals['Gray16'] = PixelFormat.Gray16
 *     __Pyx_globals['Gray32'] = PixelFormat.Gray32             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGR24'] = PixelFormat.BGR24
 *     __Pyx_globals['RGB24'] = PixelFormat.RGB24
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Gray32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 78, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Gray32, __pyx_t_24) < 0)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":79
 *     __Pyx_globals['Gray16'] = PixelFormat.Gray16
 *     __Pyx_globals['Gray32'] = PixelFormat.Gray32
 *     __Pyx_globals['BGR24'] = PixelFormat.BGR24             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGB24'] = PixelFormat.RGB24
 *     __Pyx_globals['BGR48'] = PixelFormat.BGR48
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_BGR24); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 79, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGR24, __pyx_t_8) < 0)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":80
 *     __Pyx_globals['Gray32'] = PixelFormat.Gray32
 *     __Pyx_globals['BGR24'] = PixelFormat.BGR24
 *     __Pyx_globals['RGB24'] = PixelFormat.RGB24             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGR48'] = PixelFormat.BGR48
 *     __Pyx_globals['RGB48'] = PixelFormat.RGB48
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_RGB24); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 80, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGB24, __pyx_t_24) < 0)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":81
 *     __Pyx_globals['BGR24'] = PixelFormat.BGR24
 *     __Pyx_globals['RGB24'] = PixelFormat.RGB24
 *     __Pyx_globals['BGR48'] = PixelFormat.BGR48             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGB48'] = PixelFormat.RGB48
 *     __Pyx_globals['BGR96'] = PixelFormat.BGR96
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_BGR48); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 81, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGR48, __pyx_t_8) < 0)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":82
 *     __Pyx_globals['RGB24'] = PixelFormat.RGB24
 *     __Pyx_globals['BGR48'] = PixelFormat.BGR48
 *     __Pyx_globals['RGB48'] = PixelFormat.RGB48             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGR96'] = PixelFormat.BGR96
 *     __Pyx_globals['RGB96'] = PixelFormat.RGB96
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_RGB48); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 82, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGB48, __pyx_t_24) < 0)) __PYX_ERR(4, 82, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":83
 *     __Pyx_globals['BGR48'] = PixelFormat.BGR48
 *     __Pyx_globals['RGB48'] = PixelFormat.RGB48
 *     __Pyx_globals['BGR96'] = PixelFormat.BGR96             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGB96'] = PixelFormat.RGB96
 *     __Pyx_globals['BGRA32'] = PixelFormat.BGRA32
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_BGR96); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 83, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGR96, __pyx_t_8) < 0)) __PYX_ERR(4, 83, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":84
 *     __Pyx_globals['RGB48'] = PixelFormat.RGB48
 *     __Pyx_globals['BGR96'] = PixelFormat.BGR96
 *     __Pyx_globals['RGB96'] = PixelFormat.RGB96             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGRA32'] = PixelFormat.BGRA32
 *     __Pyx_globals['RGBA32'] = PixelFormat.RGBA32
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_RGB96); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 84, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGB96, __pyx_t_24) < 0)) __PYX_ERR(4, 84, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":85
 *     __Pyx_globals['BGR96'] = PixelFormat.BGR96
 *     __Pyx_globals['RGB96'] = PixelFormat.RGB96
 *     __Pyx_globals['BGRA32'] = PixelFormat.BGRA32             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGBA32'] = PixelFormat.RGBA32
 *     __Pyx_globals['ARGB32'] = PixelFormat.ARGB32
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_BGRA32); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 85, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGRA32, __pyx_t_8) < 0)) __PYX_ERR(4, 85, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":86
 *     __Pyx_globals['RGB96'] = PixelFormat.RGB96
 *     __Pyx_globals['BGRA32'] = PixelFormat.BGRA32
 *     __Pyx_globals['RGBA32'] = PixelFormat.RGBA32             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ARGB32'] = PixelFormat.ARGB32
 *     __Pyx_globals['ABGR32'] = PixelFormat.ABGR32
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_RGBA32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 86, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGBA32, __pyx_t_24) < 0)) __PYX_ERR(4, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":87
 *     __Pyx_globals['BGRA32'] = PixelFormat.BGRA32
 *     __Pyx_globals['RGBA32'] = PixelFormat.RGBA32
 *     __Pyx_globals['ARGB32'] = PixelFormat.ARGB32             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ABGR32'] = PixelFormat.ABGR32
 *     __Pyx_globals['BGRA64'] = PixelFormat.BGRA64
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_ARGB32); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 87, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ARGB32, __pyx_t_8) < 0)) __PYX_ERR(4, 87, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":88
 *     __Pyx_globals['RGBA32'] = PixelFormat.RGBA32
 *     __Pyx_globals['ARGB32'] = PixelFormat.ARGB32
 *     __Pyx_globals['ABGR32'] = PixelFormat.ABGR32             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGRA64'] = PixelFormat.BGRA64
 *     __Pyx_globals['RGBA64'] = PixelFormat.RGBA64
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ABGR32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 88, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ABGR32, __pyx_t_24) < 0)) __PYX_ERR(4, 88, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":89
 *     __Pyx_globals['ARGB32'] = PixelFormat.ARGB32
 *     __Pyx_globals['ABGR32'] = PixelFormat.ABGR32
 *     __Pyx_globals['BGRA64'] = PixelFormat.BGRA64             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGBA64'] = PixelFormat.RGBA64
 *     __Pyx_globals['ARGB64'] = PixelFormat.ARGB64
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_BGRA64); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 89, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGRA64, __pyx_t_8) < 0)) __PYX_ERR(4, 89, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":90
 *     __Pyx_globals['ABGR32'] = PixelFormat.ABGR32
 *     __Pyx_globals['BGRA64'] = PixelFormat.BGRA64
 *     __Pyx_globals['RGBA64'] = PixelFormat.RGBA64             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ARGB64'] = PixelFormat.ARGB64
 *     __Pyx_globals['ABGR64'] = PixelFormat.ABGR64
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_RGBA64); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 90, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGBA64, __pyx_t_24) < 0)) __PYX_ERR(4, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":91
 *     __Pyx_globals['BGRA64'] = PixelFormat.BGRA64
 *     __Pyx_globals['RGBA64'] = PixelFormat.RGBA64
 *     __Pyx_globals['ARGB64'] = PixelFormat.ARGB64             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ABGR64'] = PixelFormat.ABGR64
 *     __Pyx_globals['BGRA128'] = PixelFormat.BGRA128
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_ARGB64); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 91, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ARGB64, __pyx_t_8) < 0)) __PYX_ERR(4, 91, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":92
 *     __Pyx_globals['RGBA64'] = PixelFormat.RGBA64
 *     __Pyx_globals['ARGB64'] = PixelFormat.ARGB64
 *     __Pyx_globals['ABGR64'] = PixelFormat.ABGR64             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGRA128'] = PixelFormat.BGRA128
 *     __Pyx_globals['RGBA128'] = PixelFormat.RGBA128
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ABGR64); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 92, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ABGR64, __pyx_t_24) < 0)) __PYX_ERR(4, 92, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":93
 *     __Pyx_globals['ARGB64'] = PixelFormat.ARGB64
 *     __Pyx_globals['ABGR64'] = PixelFormat.ABGR64
 *     __Pyx_globals['BGRA128'] = PixelFormat.BGRA128             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGBA128'] = PixelFormat.RGBA128
 *     __Pyx_globals['ARGB128'] = PixelFormat.ARGB128
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_BGRA128); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 93, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGRA128, __pyx_t_8) < 0)) __PYX_ERR(4, 93, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":94
 *     __Pyx_globals['ABGR64'] = PixelFormat.ABGR64
 *     __Pyx_globals['BGRA128'] = PixelFormat.BGRA128
 *     __Pyx_globals['RGBA128'] = PixelFormat.RGBA128             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ARGB128'] = PixelFormat.ARGB128
 *     __Pyx_globals['ABGR128'] = PixelFormat.ABGR128
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_RGBA128); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 94, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGBA128, __pyx_t_24) < 0)) __PYX_ERR(4, 94, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":95
 *     __Pyx_globals['BGRA128'] = PixelFormat.BGRA128
 *     __Pyx_globals['RGBA128'] = PixelFormat.RGBA128
 *     __Pyx_globals['ARGB128'] = PixelFormat.ARGB128             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ABGR128'] = PixelFormat.ABGR128
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_ARGB128); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 95, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 95, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ARGB128, __pyx_t_8) < 0)) __PYX_ERR(4, 95, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":96
 *     __Pyx_globals['RGBA128'] = PixelFormat.RGBA128
 *     __Pyx_globals['ARGB128'] = PixelFormat.ARGB128
 *     __Pyx_globals['ABGR128'] = PixelFormat.ABGR128             # <<<<<<<<<<<<<<
 * else:
 *     class PixelFormat(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ABGR128); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 96, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ABGR128, __pyx_t_24) < 0)) __PYX_ERR(4, 96, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     PixelFormat = __Pyx_EnumBase('PixelFormat', __Pyx_OrderedDict([
 */
    goto __pyx_L13;
  }

  /* "EnumType":98
 *     __Pyx_globals['ABGR128'] = PixelFormat.ABGR128
 * else:
 *     class PixelFormat(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['Gray8'] = PixelFormat(Gray8, 'Gray8')
 */
  /*else*/ {
    __pyx_t_24 = PyTuple_New(1); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __Pyx_EnumBase);
    __pyx_t_8 = __Pyx_CalculateMetaclass(NULL, __pyx_t_24); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_Py3MetaclassPrepare(__pyx_t_8, __pyx_t_24, __pyx_n_s_PixelFormat, __pyx_n_s_PixelFormat, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_Py3ClassCreate(__pyx_t_8, __pyx_n_s_PixelFormat, __pyx_t_24, __pyx_t_9, NULL, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_PixelFormat, __pyx_t_10) < 0) __PYX_ERR(4, 98, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":100
 *     class PixelFormat(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['Gray8'] = PixelFormat(Gray8, 'Gray8')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['Gray16'] = PixelFormat(Gray16, 'Gray16')
 *     __Pyx_globals['Gray32'] = PixelFormat(Gray32, 'Gray32')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray8); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_Gray8);
    __Pyx_GIVEREF(__pyx_n_s_Gray8);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_Gray8);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 100, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Gray8, __pyx_t_8) < 0)) __PYX_ERR(4, 100, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":101
 *         pass
 *     __Pyx_globals['Gray8'] = PixelFormat(Gray8, 'Gray8')
 *     __Pyx_globals['Gray16'] = PixelFormat(Gray16, 'Gray16')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['Gray32'] = PixelFormat(Gray32, 'Gray32')
 *     __Pyx_globals['BGR24'] = PixelFormat(BGR24, 'BGR24')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray16); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_Gray16);
    __Pyx_GIVEREF(__pyx_n_s_Gray16);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_Gray16);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_24, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 101, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Gray16, __pyx_t_9) < 0)) __PYX_ERR(4, 101, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "EnumType":102
 *     __Pyx_globals['Gray8'] = PixelFormat(Gray8, 'Gray8')
 *     __Pyx_globals['Gray16'] = PixelFormat(Gray16, 'Gray16')
 *     __Pyx_globals['Gray32'] = PixelFormat(Gray32, 'Gray32')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGR24'] = PixelFormat(BGR24, 'BGR24')
 *     __Pyx_globals['RGB24'] = PixelFormat(RGB24, 'RGB24')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_Gray32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_Gray32);
    __Pyx_GIVEREF(__pyx_n_s_Gray32);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_n_s_Gray32);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 102, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_Gray32, __pyx_t_24) < 0)) __PYX_ERR(4, 102, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":103
 *     __Pyx_globals['Gray16'] = PixelFormat(Gray16, 'Gray16')
 *     __Pyx_globals['Gray32'] = PixelFormat(Gray32, 'Gray32')
 *     __Pyx_globals['BGR24'] = PixelFormat(BGR24, 'BGR24')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGB24'] = PixelFormat(RGB24, 'RGB24')
 *     __Pyx_globals['BGR48'] = PixelFormat(BGR48, 'BGR48')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR24); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_BGR24);
    __Pyx_GIVEREF(__pyx_n_s_BGR24);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_BGR24);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 103, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGR24, __pyx_t_8) < 0)) __PYX_ERR(4, 103, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":104
 *     __Pyx_globals['Gray32'] = PixelFormat(Gray32, 'Gray32')
 *     __Pyx_globals['BGR24'] = PixelFormat(BGR24, 'BGR24')
 *     __Pyx_globals['RGB24'] = PixelFormat(RGB24, 'RGB24')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGR48'] = PixelFormat(BGR48, 'BGR48')
 *     __Pyx_globals['RGB48'] = PixelFormat(RGB48, 'RGB48')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB24); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_RGB24);
    __Pyx_GIVEREF(__pyx_n_s_RGB24);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_RGB24);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_24, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 104, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGB24, __pyx_t_9) < 0)) __PYX_ERR(4, 104, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "EnumType":105
 *     __Pyx_globals['BGR24'] = PixelFormat(BGR24, 'BGR24')
 *     __Pyx_globals['RGB24'] = PixelFormat(RGB24, 'RGB24')
 *     __Pyx_globals['BGR48'] = PixelFormat(BGR48, 'BGR48')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGB48'] = PixelFormat(RGB48, 'RGB48')
 *     __Pyx_globals['BGR96'] = PixelFormat(BGR96, 'BGR96')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR48); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_BGR48);
    __Pyx_GIVEREF(__pyx_n_s_BGR48);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_n_s_BGR48);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 105, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGR48, __pyx_t_24) < 0)) __PYX_ERR(4, 105, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":106
 *     __Pyx_globals['RGB24'] = PixelFormat(RGB24, 'RGB24')
 *     __Pyx_globals['BGR48'] = PixelFormat(BGR48, 'BGR48')
 *     __Pyx_globals['RGB48'] = PixelFormat(RGB48, 'RGB48')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGR96'] = PixelFormat(BGR96, 'BGR96')
 *     __Pyx_globals['RGB96'] = PixelFormat(RGB96, 'RGB96')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB48); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_RGB48);
    __Pyx_GIVEREF(__pyx_n_s_RGB48);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_RGB48);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 106, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGB48, __pyx_t_8) < 0)) __PYX_ERR(4, 106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":107
 *     __Pyx_globals['BGR48'] = PixelFormat(BGR48, 'BGR48')
 *     __Pyx_globals['RGB48'] = PixelFormat(RGB48, 'RGB48')
 *     __Pyx_globals['BGR96'] = PixelFormat(BGR96, 'BGR96')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGB96'] = PixelFormat(RGB96, 'RGB96')
 *     __Pyx_globals['BGRA32'] = PixelFormat(BGRA32, 'BGRA32')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGR96); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_BGR96);
    __Pyx_GIVEREF(__pyx_n_s_BGR96);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_BGR96);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_24, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 107, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGR96, __pyx_t_9) < 0)) __PYX_ERR(4, 107, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "EnumType":108
 *     __Pyx_globals['RGB48'] = PixelFormat(RGB48, 'RGB48')
 *     __Pyx_globals['BGR96'] = PixelFormat(BGR96, 'BGR96')
 *     __Pyx_globals['RGB96'] = PixelFormat(RGB96, 'RGB96')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGRA32'] = PixelFormat(BGRA32, 'BGRA32')
 *     __Pyx_globals['RGBA32'] = PixelFormat(RGBA32, 'RGBA32')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGB96); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_RGB96);
    __Pyx_GIVEREF(__pyx_n_s_RGB96);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_n_s_RGB96);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 108, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGB96, __pyx_t_24) < 0)) __PYX_ERR(4, 108, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":109
 *     __Pyx_globals['BGR96'] = PixelFormat(BGR96, 'BGR96')
 *     __Pyx_globals['RGB96'] = PixelFormat(RGB96, 'RGB96')
 *     __Pyx_globals['BGRA32'] = PixelFormat(BGRA32, 'BGRA32')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGBA32'] = PixelFormat(RGBA32, 'RGBA32')
 *     __Pyx_globals['ARGB32'] = PixelFormat(ARGB32, 'ARGB32')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA32); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_BGRA32);
    __Pyx_GIVEREF(__pyx_n_s_BGRA32);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_BGRA32);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 109, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGRA32, __pyx_t_8) < 0)) __PYX_ERR(4, 109, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":110
 *     __Pyx_globals['RGB96'] = PixelFormat(RGB96, 'RGB96')
 *     __Pyx_globals['BGRA32'] = PixelFormat(BGRA32, 'BGRA32')
 *     __Pyx_globals['RGBA32'] = PixelFormat(RGBA32, 'RGBA32')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ARGB32'] = PixelFormat(ARGB32, 'ARGB32')
 *     __Pyx_globals['ABGR32'] = PixelFormat(ABGR32, 'ABGR32')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA32); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_RGBA32);
    __Pyx_GIVEREF(__pyx_n_s_RGBA32);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_RGBA32);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_24, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 110, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGBA32, __pyx_t_9) < 0)) __PYX_ERR(4, 110, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "EnumType":111
 *     __Pyx_globals['BGRA32'] = PixelFormat(BGRA32, 'BGRA32')
 *     __Pyx_globals['RGBA32'] = PixelFormat(RGBA32, 'RGBA32')
 *     __Pyx_globals['ARGB32'] = PixelFormat(ARGB32, 'ARGB32')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ABGR32'] = PixelFormat(ABGR32, 'ABGR32')
 *     __Pyx_globals['BGRA64'] = PixelFormat(BGRA64, 'BGRA64')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB32); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_ARGB32);
    __Pyx_GIVEREF(__pyx_n_s_ARGB32);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_n_s_ARGB32);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 111, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ARGB32, __pyx_t_24) < 0)) __PYX_ERR(4, 111, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":112
 *     __Pyx_globals['RGBA32'] = PixelFormat(RGBA32, 'RGBA32')
 *     __Pyx_globals['ARGB32'] = PixelFormat(ARGB32, 'ARGB32')
 *     __Pyx_globals['ABGR32'] = PixelFormat(ABGR32, 'ABGR32')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGRA64'] = PixelFormat(BGRA64, 'BGRA64')
 *     __Pyx_globals['RGBA64'] = PixelFormat(RGBA64, 'RGBA64')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR32); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_ABGR32);
    __Pyx_GIVEREF(__pyx_n_s_ABGR32);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_ABGR32);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 112, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ABGR32, __pyx_t_8) < 0)) __PYX_ERR(4, 112, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":113
 *     __Pyx_globals['ARGB32'] = PixelFormat(ARGB32, 'ARGB32')
 *     __Pyx_globals['ABGR32'] = PixelFormat(ABGR32, 'ABGR32')
 *     __Pyx_globals['BGRA64'] = PixelFormat(BGRA64, 'BGRA64')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGBA64'] = PixelFormat(RGBA64, 'RGBA64')
 *     __Pyx_globals['ARGB64'] = PixelFormat(ARGB64, 'ARGB64')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA64); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_BGRA64);
    __Pyx_GIVEREF(__pyx_n_s_BGRA64);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_BGRA64);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_24, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 113, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGRA64, __pyx_t_9) < 0)) __PYX_ERR(4, 113, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "EnumType":114
 *     __Pyx_globals['ABGR32'] = PixelFormat(ABGR32, 'ABGR32')
 *     __Pyx_globals['BGRA64'] = PixelFormat(BGRA64, 'BGRA64')
 *     __Pyx_globals['RGBA64'] = PixelFormat(RGBA64, 'RGBA64')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ARGB64'] = PixelFormat(ARGB64, 'ARGB64')
 *     __Pyx_globals['ABGR64'] = PixelFormat(ABGR64, 'ABGR64')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA64); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_RGBA64);
    __Pyx_GIVEREF(__pyx_n_s_RGBA64);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_n_s_RGBA64);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 114, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGBA64, __pyx_t_24) < 0)) __PYX_ERR(4, 114, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":115
 *     __Pyx_globals['BGRA64'] = PixelFormat(BGRA64, 'BGRA64')
 *     __Pyx_globals['RGBA64'] = PixelFormat(RGBA64, 'RGBA64')
 *     __Pyx_globals['ARGB64'] = PixelFormat(ARGB64, 'ARGB64')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ABGR64'] = PixelFormat(ABGR64, 'ABGR64')
 *     __Pyx_globals['BGRA128'] = PixelFormat(BGRA128, 'BGRA128')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB64); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_ARGB64);
    __Pyx_GIVEREF(__pyx_n_s_ARGB64);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_ARGB64);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 115, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ARGB64, __pyx_t_8) < 0)) __PYX_ERR(4, 115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":116
 *     __Pyx_globals['RGBA64'] = PixelFormat(RGBA64, 'RGBA64')
 *     __Pyx_globals['ARGB64'] = PixelFormat(ARGB64, 'ARGB64')
 *     __Pyx_globals['ABGR64'] = PixelFormat(ABGR64, 'ABGR64')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['BGRA128'] = PixelFormat(BGRA128, 'BGRA128')
 *     __Pyx_globals['RGBA128'] = PixelFormat(RGBA128, 'RGBA128')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR64); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_ABGR64);
    __Pyx_GIVEREF(__pyx_n_s_ABGR64);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_ABGR64);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_24, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 116, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ABGR64, __pyx_t_9) < 0)) __PYX_ERR(4, 116, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "EnumType":117
 *     __Pyx_globals['ARGB64'] = PixelFormat(ARGB64, 'ARGB64')
 *     __Pyx_globals['ABGR64'] = PixelFormat(ABGR64, 'ABGR64')
 *     __Pyx_globals['BGRA128'] = PixelFormat(BGRA128, 'BGRA128')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['RGBA128'] = PixelFormat(RGBA128, 'RGBA128')
 *     __Pyx_globals['ARGB128'] = PixelFormat(ARGB128, 'ARGB128')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_BGRA128); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_BGRA128);
    __Pyx_GIVEREF(__pyx_n_s_BGRA128);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_n_s_BGRA128);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 117, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_BGRA128, __pyx_t_24) < 0)) __PYX_ERR(4, 117, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":118
 *     __Pyx_globals['ABGR64'] = PixelFormat(ABGR64, 'ABGR64')
 *     __Pyx_globals['BGRA128'] = PixelFormat(BGRA128, 'BGRA128')
 *     __Pyx_globals['RGBA128'] = PixelFormat(RGBA128, 'RGBA128')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ARGB128'] = PixelFormat(ARGB128, 'ARGB128')
 *     __Pyx_globals['ABGR128'] = PixelFormat(ABGR128, 'ABGR128')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_RGBA128); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_RGBA128);
    __Pyx_GIVEREF(__pyx_n_s_RGBA128);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_n_s_RGBA128);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 118, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_RGBA128, __pyx_t_8) < 0)) __PYX_ERR(4, 118, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "EnumType":119
 *     __Pyx_globals['BGRA128'] = PixelFormat(BGRA128, 'BGRA128')
 *     __Pyx_globals['RGBA128'] = PixelFormat(RGBA128, 'RGBA128')
 *     __Pyx_globals['ARGB128'] = PixelFormat(ARGB128, 'ARGB128')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['ABGR128'] = PixelFormat(ABGR128, 'ABGR128')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ARGB128); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_ARGB128);
    __Pyx_GIVEREF(__pyx_n_s_ARGB128);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_ARGB128);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_24, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 119, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ARGB128, __pyx_t_9) < 0)) __PYX_ERR(4, 119, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "EnumType":120
 *     __Pyx_globals['RGBA128'] = PixelFormat(RGBA128, 'RGBA128')
 *     __Pyx_globals['ARGB128'] = PixelFormat(ARGB128, 'ARGB128')
 *     __Pyx_globals['ABGR128'] = PixelFormat(ABGR128, 'ABGR128')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_PixelFormat); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(__pyx_e_7celiagg_8_celiagg_ABGR128); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_ABGR128);
    __Pyx_GIVEREF(__pyx_n_s_ABGR128);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_n_s_ABGR128);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 120, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_ABGR128, __pyx_t_24) < 0)) __PYX_ERR(4, 120, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
  }
  __pyx_L13:;

  /* "EnumType":50
 * 
 * 
 * cdef dict __Pyx_globals = globals()             # <<<<<<<<<<<<<<
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 */
  __pyx_t_24 = __Pyx_Globals(); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_24);
  if (!(likely(PyDict_CheckExact(__pyx_t_24))||((__pyx_t_24) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_24)->tp_name), 0))) __PYX_ERR(4, 50, __pyx_L1_error)
  __Pyx_XGOTREF(__Pyx_globals);
  __Pyx_DECREF_SET(__Pyx_globals, ((PyObject*)__pyx_t_24));
  __Pyx_GIVEREF(__pyx_t_24);
  __pyx_t_24 = 0;

  /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     TextDrawingMode = __Pyx_EnumBase('TextDrawingMode', __Pyx_OrderedDict([
 */
  __pyx_t_5 = ((PY_VERSION_HEX >= 0x03040000) != 0);
  if (__pyx_t_5) {

    /* "EnumType":54
 * 
 *     TextDrawingMode = __Pyx_EnumBase('TextDrawingMode', __Pyx_OrderedDict([
 *         ('TextDrawInvisible', TextDrawInvisible),             # <<<<<<<<<<<<<<
 *         ('TextDrawFill', TextDrawFill),
 *         ('TextDrawStroke', TextDrawStroke),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawInvisible); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(4, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_n_s_TextDrawInvisible);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawInvisible);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_TextDrawInvisible);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":55
 *     TextDrawingMode = __Pyx_EnumBase('TextDrawingMode', __Pyx_OrderedDict([
 *         ('TextDrawInvisible', TextDrawInvisible),
 *         ('TextDrawFill', TextDrawFill),             # <<<<<<<<<<<<<<
 *         ('TextDrawStroke', TextDrawStroke),
 *         ('TextDrawClip', TextDrawClip),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawFill); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(4, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_n_s_TextDrawFill);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawFill);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_n_s_TextDrawFill);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":56
 *         ('TextDrawInvisible', TextDrawInvisible),
 *         ('TextDrawFill', TextDrawFill),
 *         ('TextDrawStroke', TextDrawStroke),             # <<<<<<<<<<<<<<
 *         ('TextDrawClip', TextDrawClip),
 *         ('TextDrawFillStroke', TextDrawFillStroke),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawStroke); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(4, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_n_s_TextDrawStroke);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawStroke);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_TextDrawStroke);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":57
 *         ('TextDrawFill', TextDrawFill),
 *         ('TextDrawStroke', TextDrawStroke),
 *         ('TextDrawClip', TextDrawClip),             # <<<<<<<<<<<<<<
 *         ('TextDrawFillStroke', TextDrawFillStroke),
 *         ('TextDrawFillClip', TextDrawFillClip),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawClip); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(4, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_n_s_TextDrawClip);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawClip);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_TextDrawClip);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":58
 *         ('TextDrawStroke', TextDrawStroke),
 *         ('TextDrawClip', TextDrawClip),
 *         ('TextDrawFillStroke', TextDrawFillStroke),             # <<<<<<<<<<<<<<
 *         ('TextDrawFillClip', TextDrawFillClip),
 *         ('TextDrawStrokeClip', TextDrawStrokeClip),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawFillStroke); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(4, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_n_s_TextDrawFillStroke);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawFillStroke);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_n_s_TextDrawFillStroke);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":59
 *         ('TextDrawClip', TextDrawClip),
 *         ('TextDrawFillStroke', TextDrawFillStroke),
 *         ('TextDrawFillClip', TextDrawFillClip),             # <<<<<<<<<<<<<<
 *         ('TextDrawStrokeClip', TextDrawStrokeClip),
 *         ('TextDrawFillStrokeClip', TextDrawFillStrokeClip),
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawFillClip); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_n_s_TextDrawFillClip);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawFillClip);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_n_s_TextDrawFillClip);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":60
 *         ('TextDrawFillStroke', TextDrawFillStroke),
 *         ('TextDrawFillClip', TextDrawFillClip),
 *         ('TextDrawStrokeClip', TextDrawStrokeClip),             # <<<<<<<<<<<<<<
 *         ('TextDrawFillStrokeClip', TextDrawFillStrokeClip),
 *     ]))
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawStrokeClip); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_TextDrawStrokeClip);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawStrokeClip);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_n_s_TextDrawStrokeClip);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":61
 *         ('TextDrawFillClip', TextDrawFillClip),
 *         ('TextDrawStrokeClip', TextDrawStrokeClip),
 *         ('TextDrawFillStrokeClip', TextDrawFillStrokeClip),             # <<<<<<<<<<<<<<
 *     ]))
 *     __Pyx_globals['TextDrawInvisible'] = TextDrawingMode.TextDrawInvisible
 */
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawFillStrokeClip); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_TextDrawFillStrokeClip);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawFillStrokeClip);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_n_s_TextDrawFillStrokeClip);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_24);
    __pyx_t_24 = 0;

    /* "EnumType":53
 * if PY_VERSION_HEX >= 0x03040000:
 * 
 *     TextDrawingMode = __Pyx_EnumBase('TextDrawingMode', __Pyx_OrderedDict([             # <<<<<<<<<<<<<<
 *         ('TextDrawInvisible', TextDrawInvisible),
 *         ('TextDrawFill', TextDrawFill),
 */
    __pyx_t_24 = PyList_New(8); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_8);
    PyList_SET_ITEM(__pyx_t_24, 0, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyList_SET_ITEM(__pyx_t_24, 1, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_10);
    PyList_SET_ITEM(__pyx_t_24, 2, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    PyList_SET_ITEM(__pyx_t_24, 3, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyList_SET_ITEM(__pyx_t_24, 4, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyList_SET_ITEM(__pyx_t_24, 5, __pyx_t_13);
    __Pyx_GIVEREF(__pyx_t_14);
    PyList_SET_ITEM(__pyx_t_24, 6, __pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_15);
    PyList_SET_ITEM(__pyx_t_24, 7, __pyx_t_15);
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_10 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_t_13 = 0;
    __pyx_t_14 = 0;
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_CallOneArg(__Pyx_OrderedDict, __pyx_t_24); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_TextDrawingMode);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawingMode);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_n_s_TextDrawingMode);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_t_15);
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_Call(__Pyx_EnumBase, __pyx_t_24, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_TextDrawingMode, __pyx_t_15) < 0) __PYX_ERR(4, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":63
 *         ('TextDrawFillStrokeClip', TextDrawFillStrokeClip),
 *     ]))
 *     __Pyx_globals['TextDrawInvisible'] = TextDrawingMode.TextDrawInvisible             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawFill'] = TextDrawingMode.TextDrawFill
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode.TextDrawStroke
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_TextDrawInvisible); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 63, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawInvisible, __pyx_t_24) < 0)) __PYX_ERR(4, 63, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":64
 *     ]))
 *     __Pyx_globals['TextDrawInvisible'] = TextDrawingMode.TextDrawInvisible
 *     __Pyx_globals['TextDrawFill'] = TextDrawingMode.TextDrawFill             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode.TextDrawStroke
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode.TextDrawClip
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_TextDrawFill); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 64, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawFill, __pyx_t_15) < 0)) __PYX_ERR(4, 64, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":65
 *     __Pyx_globals['TextDrawInvisible'] = TextDrawingMode.TextDrawInvisible
 *     __Pyx_globals['TextDrawFill'] = TextDrawingMode.TextDrawFill
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode.TextDrawStroke             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode.TextDrawClip
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode.TextDrawFillStroke
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_TextDrawStroke); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 65, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawStroke, __pyx_t_24) < 0)) __PYX_ERR(4, 65, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":66
 *     __Pyx_globals['TextDrawFill'] = TextDrawingMode.TextDrawFill
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode.TextDrawStroke
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode.TextDrawClip             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode.TextDrawFillStroke
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode.TextDrawFillClip
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_TextDrawClip); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 66, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawClip, __pyx_t_15) < 0)) __PYX_ERR(4, 66, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":67
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode.TextDrawStroke
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode.TextDrawClip
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode.TextDrawFillStroke             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode.TextDrawFillClip
 *     __Pyx_globals['TextDrawStrokeClip'] = TextDrawingMode.TextDrawStrokeClip
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_TextDrawFillStroke); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 67, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawFillStroke, __pyx_t_24) < 0)) __PYX_ERR(4, 67, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":68
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode.TextDrawClip
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode.TextDrawFillStroke
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode.TextDrawFillClip             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawStrokeClip'] = TextDrawingMode.TextDrawStrokeClip
 *     __Pyx_globals['TextDrawFillStrokeClip'] = TextDrawingMode.TextDrawFillStrokeClip
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_TextDrawFillClip); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 68, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawFillClip, __pyx_t_15) < 0)) __PYX_ERR(4, 68, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":69
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode.TextDrawFillStroke
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode.TextDrawFillClip
 *     __Pyx_globals['TextDrawStrokeClip'] = TextDrawingMode.TextDrawStrokeClip             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawFillStrokeClip'] = TextDrawingMode.TextDrawFillStrokeClip
 * else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_24 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_TextDrawStrokeClip); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 69, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawStrokeClip, __pyx_t_24) < 0)) __PYX_ERR(4, 69, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":70
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode.TextDrawFillClip
 *     __Pyx_globals['TextDrawStrokeClip'] = TextDrawingMode.TextDrawStrokeClip
 *     __Pyx_globals['TextDrawFillStrokeClip'] = TextDrawingMode.TextDrawFillStrokeClip             # <<<<<<<<<<<<<<
 * else:
 *     class TextDrawingMode(__Pyx_EnumBase):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_24, __pyx_n_s_TextDrawFillStrokeClip); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 70, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawFillStrokeClip, __pyx_t_15) < 0)) __PYX_ERR(4, 70, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":51
 * 
 * cdef dict __Pyx_globals = globals()
 * if PY_VERSION_HEX >= 0x03040000:             # <<<<<<<<<<<<<<
 * 
 *     TextDrawingMode = __Pyx_EnumBase('TextDrawingMode', __Pyx_OrderedDict([
 */
    goto __pyx_L14;
  }

  /* "EnumType":72
 *     __Pyx_globals['TextDrawFillStrokeClip'] = TextDrawingMode.TextDrawFillStrokeClip
 * else:
 *     class TextDrawingMode(__Pyx_EnumBase):             # <<<<<<<<<<<<<<
 *         pass
 *     __Pyx_globals['TextDrawInvisible'] = TextDrawingMode(TextDrawInvisible, 'TextDrawInvisible')
 */
  /*else*/ {
    __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_INCREF(__Pyx_EnumBase);
    __Pyx_GIVEREF(__Pyx_EnumBase);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __Pyx_EnumBase);
    __pyx_t_24 = __Pyx_CalculateMetaclass(NULL, __pyx_t_15); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_14 = __Pyx_Py3MetaclassPrepare(__pyx_t_24, __pyx_t_15, __pyx_n_s_TextDrawingMode, __pyx_n_s_TextDrawingMode, (PyObject *) NULL, __pyx_n_s_EnumType, (PyObject *) NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_24, __pyx_n_s_TextDrawingMode, __pyx_t_15, __pyx_t_14, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_TextDrawingMode, __pyx_t_13) < 0) __PYX_ERR(4, 72, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":74
 *     class TextDrawingMode(__Pyx_EnumBase):
 *         pass
 *     __Pyx_globals['TextDrawInvisible'] = TextDrawingMode(TextDrawInvisible, 'TextDrawInvisible')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawFill'] = TextDrawingMode(TextDrawFill, 'TextDrawFill')
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode(TextDrawStroke, 'TextDrawStroke')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawInvisible); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_TextDrawInvisible);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawInvisible);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_n_s_TextDrawInvisible);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_14, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 74, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawInvisible, __pyx_t_24) < 0)) __PYX_ERR(4, 74, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":75
 *         pass
 *     __Pyx_globals['TextDrawInvisible'] = TextDrawingMode(TextDrawInvisible, 'TextDrawInvisible')
 *     __Pyx_globals['TextDrawFill'] = TextDrawingMode(TextDrawFill, 'TextDrawFill')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode(TextDrawStroke, 'TextDrawStroke')
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode(TextDrawClip, 'TextDrawClip')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_14 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawFill); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_TextDrawFill);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawFill);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_n_s_TextDrawFill);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_15, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 75, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawFill, __pyx_t_14) < 0)) __PYX_ERR(4, 75, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "EnumType":76
 *     __Pyx_globals['TextDrawInvisible'] = TextDrawingMode(TextDrawInvisible, 'TextDrawInvisible')
 *     __Pyx_globals['TextDrawFill'] = TextDrawingMode(TextDrawFill, 'TextDrawFill')
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode(TextDrawStroke, 'TextDrawStroke')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode(TextDrawClip, 'TextDrawClip')
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode(TextDrawFillStroke, 'TextDrawFillStroke')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawStroke); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_TextDrawStroke);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawStroke);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_TextDrawStroke);
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_24, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 76, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawStroke, __pyx_t_15) < 0)) __PYX_ERR(4, 76, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":77
 *     __Pyx_globals['TextDrawFill'] = TextDrawingMode(TextDrawFill, 'TextDrawFill')
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode(TextDrawStroke, 'TextDrawStroke')
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode(TextDrawClip, 'TextDrawClip')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode(TextDrawFillStroke, 'TextDrawFillStroke')
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode(TextDrawFillClip, 'TextDrawFillClip')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawClip); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_TextDrawClip);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawClip);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_n_s_TextDrawClip);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_14, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 77, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawClip, __pyx_t_24) < 0)) __PYX_ERR(4, 77, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":78
 *     __Pyx_globals['TextDrawStroke'] = TextDrawingMode(TextDrawStroke, 'TextDrawStroke')
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode(TextDrawClip, 'TextDrawClip')
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode(TextDrawFillStroke, 'TextDrawFillStroke')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode(TextDrawFillClip, 'TextDrawFillClip')
 *     __Pyx_globals['TextDrawStrokeClip'] = TextDrawingMode(TextDrawStrokeClip, 'TextDrawStrokeClip')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_14 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawFillStroke); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_TextDrawFillStroke);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawFillStroke);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_n_s_TextDrawFillStroke);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_15, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 78, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawFillStroke, __pyx_t_14) < 0)) __PYX_ERR(4, 78, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "EnumType":79
 *     __Pyx_globals['TextDrawClip'] = TextDrawingMode(TextDrawClip, 'TextDrawClip')
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode(TextDrawFillStroke, 'TextDrawFillStroke')
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode(TextDrawFillClip, 'TextDrawFillClip')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawStrokeClip'] = TextDrawingMode(TextDrawStrokeClip, 'TextDrawStrokeClip')
 *     __Pyx_globals['TextDrawFillStrokeClip'] = TextDrawingMode(TextDrawFillStrokeClip, 'TextDrawFillStrokeClip')
 */
    __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawFillClip); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_24 = PyTuple_New(2); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_24, 0, __pyx_t_15);
    __Pyx_INCREF(__pyx_n_s_TextDrawFillClip);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawFillClip);
    PyTuple_SET_ITEM(__pyx_t_24, 1, __pyx_n_s_TextDrawFillClip);
    __pyx_t_15 = 0;
    __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_24, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 79, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawFillClip, __pyx_t_15) < 0)) __PYX_ERR(4, 79, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "EnumType":80
 *     __Pyx_globals['TextDrawFillStroke'] = TextDrawingMode(TextDrawFillStroke, 'TextDrawFillStroke')
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode(TextDrawFillClip, 'TextDrawFillClip')
 *     __Pyx_globals['TextDrawStrokeClip'] = TextDrawingMode(TextDrawStrokeClip, 'TextDrawStrokeClip')             # <<<<<<<<<<<<<<
 *     __Pyx_globals['TextDrawFillStrokeClip'] = TextDrawingMode(TextDrawFillStrokeClip, 'TextDrawFillStrokeClip')
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_24 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawStrokeClip); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_24);
    PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_24);
    __Pyx_INCREF(__pyx_n_s_TextDrawStrokeClip);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawStrokeClip);
    PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_n_s_TextDrawStrokeClip);
    __pyx_t_24 = 0;
    __pyx_t_24 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_14, NULL); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 80, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawStrokeClip, __pyx_t_24) < 0)) __PYX_ERR(4, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;

    /* "EnumType":81
 *     __Pyx_globals['TextDrawFillClip'] = TextDrawingMode(TextDrawFillClip, 'TextDrawFillClip')
 *     __Pyx_globals['TextDrawStrokeClip'] = TextDrawingMode(TextDrawStrokeClip, 'TextDrawStrokeClip')
 *     __Pyx_globals['TextDrawFillStrokeClip'] = TextDrawingMode(TextDrawFillStrokeClip, 'TextDrawFillStrokeClip')             # <<<<<<<<<<<<<<
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_24, __pyx_n_s_TextDrawingMode); if (unlikely(!__pyx_t_24)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_24);
    __pyx_t_14 = __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(__pyx_e_7celiagg_8_celiagg_TextDrawFillStrokeClip); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = PyTuple_New(2); if (unlikely(!__pyx_t_15)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14);
    __Pyx_INCREF(__pyx_n_s_TextDrawFillStrokeClip);
    __Pyx_GIVEREF(__pyx_n_s_TextDrawFillStrokeClip);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_n_s_TextDrawFillStrokeClip);
    __pyx_t_14 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_24, __pyx_t_15, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (unlikely(__Pyx_globals == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(4, 81, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__Pyx_globals, __pyx_n_s_TextDrawFillStrokeClip, __pyx_t_14) < 0)) __PYX_ERR(4, 81, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  }
  __pyx_L14:;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init celiagg._celiagg", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init celiagg._celiagg");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_items */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_items, d);
    else
        return PyDict_Items(d);
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* BufferIndexError */
static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* MemviewSliceInit */
static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (memviewslice->memview || memviewslice->data) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#ifdef HAVE_STDARG_PROTOTYPES
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (!memview || (PyObject *) memview == Py_None)
        return;
    if (__pyx_get_slice_count(memview) < 0)
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (first_time) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (!memview ) {
        return;
    } else if ((PyObject *) memview == Py_None) {
        memslice->memview = NULL;
        return;
    }
    if (__pyx_get_slice_count(memview) <= 0)
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (last_time) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* pyobject_as_double */
static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if ((0)) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* SliceObject */
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* None */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    length = stop - start;
    if (unlikely(length <= 0))
        return PyUnicode_FromUnicode(NULL, 0);
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* None */
static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    Py_ssize_t i;
    PyObject *names;
    PyObject *globals = __pyx_d;
    Py_INCREF(globals);
    names = PyObject_Dir(__pyx_m);
    if (!names)
        goto bad;
    for (i = PyList_GET_SIZE(names)-1; i >= 0; i--) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject* name = PySequence_ITEM(names, i);
        if (!name)
            goto bad;
#else
        PyObject* name = PyList_GET_ITEM(names, i);
#endif
        if (!PyDict_Contains(globals, name)) {
            PyObject* value = __Pyx_GetAttr(__pyx_m, name);
            if (!value) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                goto bad;
            }
            if (PyDict_SetItem(globals, name, value) < 0) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                Py_DECREF(value);
                goto bad;
            }
        }
#if CYTHON_COMPILING_IN_PYPY
        Py_DECREF(name);
#endif
    }
    Py_DECREF(names);
    return globals;
bad:
    Py_XDECREF(names);
    Py_XDECREF(globals);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


/* MemviewSliceIsContig */
static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* TypeInfoCompare */
  static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
  static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (buf->strides[dim] != sizeof(void *)) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (buf->strides[dim] != buf->itemsize) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (stride < buf->itemsize) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (spec & (__Pyx_MEMVIEW_PTR)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (buf->suboffsets) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (buf->suboffsets && buf->suboffsets[dim] >= 0) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (!buf->suboffsets || (buf->suboffsets[dim] < 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (stride * buf->itemsize != buf->strides[i] &&
                    buf->shape[i] > 1)
            {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (stride * buf->itemsize != buf->strides[i] &&
                    buf->shape[i] > 1) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (buf->ndim != ndim) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
    }
    if ((unsigned) buf->itemsize != dtype->size) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    for (i = 0; i < ndim; i++) {
        spec = axes_specs[i];
        if (!__pyx_check_strides(buf, i, ndim, spec))
            goto fail;
        if (!__pyx_check_suboffsets(buf, i, ndim, spec))
            goto fail;
    }
    if (buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_float(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 3,
                                                 &__Pyx_TypeInfo_float, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn___pyx_t_7celiagg_8_celiagg__bytes_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 3,
                                                 &__Pyx_TypeInfo_nn___pyx_t_7celiagg_8_celiagg__bytes_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn___pyx_t_7celiagg_8_celiagg__bytes_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn___pyx_t_7celiagg_8_celiagg__bytes_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__Font_3a__3a_FontCacheType(enum Font::FontCacheType value) {
    const enum Font::FontCacheType neg_one = (enum Font::FontCacheType) ((enum Font::FontCacheType) 0 - (enum Font::FontCacheType) 1), const_zero = (enum Font::FontCacheType) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum Font::FontCacheType) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum Font::FontCacheType) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum Font::FontCacheType) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum Font::FontCacheType) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum Font::FontCacheType) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum Font::FontCacheType),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_DrawingMode(enum GraphicsState::DrawingMode value) {
    const enum GraphicsState::DrawingMode neg_one = (enum GraphicsState::DrawingMode) ((enum GraphicsState::DrawingMode) 0 - (enum GraphicsState::DrawingMode) 1), const_zero = (enum GraphicsState::DrawingMode) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum GraphicsState::DrawingMode) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum GraphicsState::DrawingMode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::DrawingMode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum GraphicsState::DrawingMode) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::DrawingMode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum GraphicsState::DrawingMode),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_TextDrawingMode(enum GraphicsState::TextDrawingMode value) {
    const enum GraphicsState::TextDrawingMode neg_one = (enum GraphicsState::TextDrawingMode) ((enum GraphicsState::TextDrawingMode) 0 - (enum GraphicsState::TextDrawingMode) 1), const_zero = (enum GraphicsState::TextDrawingMode) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum GraphicsState::TextDrawingMode) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum GraphicsState::TextDrawingMode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::TextDrawingMode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum GraphicsState::TextDrawingMode) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::TextDrawingMode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum GraphicsState::TextDrawingMode),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_BlendMode(enum GraphicsState::BlendMode value) {
    const enum GraphicsState::BlendMode neg_one = (enum GraphicsState::BlendMode) ((enum GraphicsState::BlendMode) 0 - (enum GraphicsState::BlendMode) 1), const_zero = (enum GraphicsState::BlendMode) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum GraphicsState::BlendMode) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum GraphicsState::BlendMode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::BlendMode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum GraphicsState::BlendMode) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::BlendMode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum GraphicsState::BlendMode),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_LineCap(enum GraphicsState::LineCap value) {
    const enum GraphicsState::LineCap neg_one = (enum GraphicsState::LineCap) ((enum GraphicsState::LineCap) 0 - (enum GraphicsState::LineCap) 1), const_zero = (enum GraphicsState::LineCap) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum GraphicsState::LineCap) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum GraphicsState::LineCap) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::LineCap) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum GraphicsState::LineCap) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::LineCap) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum GraphicsState::LineCap),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_LineJoin(enum GraphicsState::LineJoin value) {
    const enum GraphicsState::LineJoin neg_one = (enum GraphicsState::LineJoin) ((enum GraphicsState::LineJoin) 0 - (enum GraphicsState::LineJoin) 1), const_zero = (enum GraphicsState::LineJoin) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum GraphicsState::LineJoin) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum GraphicsState::LineJoin) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::LineJoin) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum GraphicsState::LineJoin) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::LineJoin) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum GraphicsState::LineJoin),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__GraphicsState_3a__3a_InnerJoin(enum GraphicsState::InnerJoin value) {
    const enum GraphicsState::InnerJoin neg_one = (enum GraphicsState::InnerJoin) ((enum GraphicsState::InnerJoin) 0 - (enum GraphicsState::InnerJoin) 1), const_zero = (enum GraphicsState::InnerJoin) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum GraphicsState::InnerJoin) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum GraphicsState::InnerJoin) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::InnerJoin) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum GraphicsState::InnerJoin) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum GraphicsState::InnerJoin) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum GraphicsState::InnerJoin),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(enum __pyx_t_7celiagg_8_celiagg_PixelFormat value) {
    const enum __pyx_t_7celiagg_8_celiagg_PixelFormat neg_one = (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat) 0 - (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_float(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(float *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_float(const char *itemp, PyObject *obj) {
    float value = __pyx_PyFloat_AsFloat(obj);
    if ((value == (float)-1) && PyErr_Occurred())
        return 0;
    *(float *) itemp = value;
    return 1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_char(unsigned char value) {
    const unsigned char neg_one = (unsigned char) ((unsigned char) 0 - (unsigned char) 1), const_zero = (unsigned char) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned char),
                                     little, !is_unsigned);
    }
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_7celiagg_8_celiagg__bytes_t(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_unsigned_char(*(__pyx_t_7celiagg_8_celiagg__bytes_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_7celiagg_8_celiagg__bytes_t(const char *itemp, PyObject *obj) {
    __pyx_t_7celiagg_8_celiagg__bytes_t value = __Pyx_PyInt_As_unsigned_char(obj);
    if ((value == (unsigned char)-1) && PyErr_Occurred())
        return 0;
    *(__pyx_t_7celiagg_8_celiagg__bytes_t *) itemp = value;
    return 1;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_double(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(double *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_double(const char *itemp, PyObject *obj) {
    double value = __pyx_PyFloat_AsDouble(obj);
    if ((value == (double)-1) && PyErr_Occurred())
        return 0;
    *(double *) itemp = value;
    return 1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__Paint_3a__3a_GradientSpread(enum Paint::GradientSpread value) {
    const enum Paint::GradientSpread neg_one = (enum Paint::GradientSpread) ((enum Paint::GradientSpread) 0 - (enum Paint::GradientSpread) 1), const_zero = (enum Paint::GradientSpread) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum Paint::GradientSpread) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum Paint::GradientSpread) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum Paint::GradientSpread) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum Paint::GradientSpread) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum Paint::GradientSpread) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum Paint::GradientSpread),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__Paint_3a__3a_GradientUnits(enum Paint::GradientUnits value) {
    const enum Paint::GradientUnits neg_one = (enum Paint::GradientUnits) ((enum Paint::GradientUnits) 0 - (enum Paint::GradientUnits) 1), const_zero = (enum Paint::GradientUnits) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum Paint::GradientUnits) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum Paint::GradientUnits) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum Paint::GradientUnits) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum Paint::GradientUnits) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum Paint::GradientUnits) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum Paint::GradientUnits),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(enum __pyx_t_7celiagg_8_celiagg_PatternStyle value) {
    const enum __pyx_t_7celiagg_8_celiagg_PatternStyle neg_one = (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) ((enum __pyx_t_7celiagg_8_celiagg_PatternStyle) 0 - (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle),
                                     little, !is_unsigned);
    }
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) ((enum NPY_TYPES) 0 - (enum NPY_TYPES) 1), const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* MemviewSliceCopyTemplate */
  static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (from_mvs->suboffsets[i] >= 0) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_FontCacheType __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_FontCacheType neg_one = (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) ((enum __pyx_t_7celiagg_8_celiagg_FontCacheType) 0 - (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) (((((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) (((((((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) (((((((((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) (((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) ((((((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) (((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) ((((((((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) (((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) ((((((((((enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_FontCacheType)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_FontCacheType, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_FontCacheType val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_FontCacheType val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_FontCacheType");
    return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_FontCacheType");
    return (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
    const unsigned int neg_one = (unsigned int) ((unsigned int) 0 - (unsigned int) 1), const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned int) -1;
        }
    } else {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_DrawingMode __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_DrawingMode neg_one = (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) ((enum __pyx_t_7celiagg_8_celiagg_DrawingMode) 0 - (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) (((((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) (((((((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) (((((((((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) (((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) ((((((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) (((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) ((((((((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) (((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) ((((((((((enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_DrawingMode)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_DrawingMode, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_DrawingMode val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_DrawingMode val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_DrawingMode");
    return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_DrawingMode");
    return (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode neg_one = (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) ((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) 0 - (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) (((((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) (((((((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) (((((((((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) (((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) ((((((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) (((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) ((((((((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) (((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) ((((((((((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode");
    return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode");
    return (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_BlendMode __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_BlendMode(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_BlendMode neg_one = (enum __pyx_t_7celiagg_8_celiagg_BlendMode) ((enum __pyx_t_7celiagg_8_celiagg_BlendMode) 0 - (enum __pyx_t_7celiagg_8_celiagg_BlendMode) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_BlendMode) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) (((((enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) (((((((enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) (((((((((enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_BlendMode, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_BlendMode, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) (((enum __pyx_t_7celiagg_8_celiagg_BlendMode)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) ((((((enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) (((enum __pyx_t_7celiagg_8_celiagg_BlendMode)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) ((((((((enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) (((enum __pyx_t_7celiagg_8_celiagg_BlendMode)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_BlendMode, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) ((((((((((enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_BlendMode)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_BlendMode, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_BlendMode, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_BlendMode val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_BlendMode val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_BlendMode(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_BlendMode");
    return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_BlendMode");
    return (enum __pyx_t_7celiagg_8_celiagg_BlendMode) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_LineCap __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_LineCap(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_LineCap neg_one = (enum __pyx_t_7celiagg_8_celiagg_LineCap) ((enum __pyx_t_7celiagg_8_celiagg_LineCap) 0 - (enum __pyx_t_7celiagg_8_celiagg_LineCap) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_LineCap) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_LineCap) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) (((((enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) (((((((enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) (((((((((enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_LineCap) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_LineCap, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_LineCap, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_LineCap) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) (((enum __pyx_t_7celiagg_8_celiagg_LineCap)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) ((((((enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) (((enum __pyx_t_7celiagg_8_celiagg_LineCap)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) ((((((((enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) (((enum __pyx_t_7celiagg_8_celiagg_LineCap)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineCap, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) ((((((((((enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineCap)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_LineCap, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_LineCap, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_LineCap val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_LineCap) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_LineCap val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_LineCap) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_LineCap(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_LineCap");
    return (enum __pyx_t_7celiagg_8_celiagg_LineCap) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_LineCap");
    return (enum __pyx_t_7celiagg_8_celiagg_LineCap) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_LineJoin __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_LineJoin(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_LineJoin neg_one = (enum __pyx_t_7celiagg_8_celiagg_LineJoin) ((enum __pyx_t_7celiagg_8_celiagg_LineJoin) 0 - (enum __pyx_t_7celiagg_8_celiagg_LineJoin) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_LineJoin) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) (((((enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) (((((((enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) (((((((((enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_LineJoin, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_LineJoin, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) (((enum __pyx_t_7celiagg_8_celiagg_LineJoin)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) ((((((enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) (((enum __pyx_t_7celiagg_8_celiagg_LineJoin)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) ((((((((enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) (((enum __pyx_t_7celiagg_8_celiagg_LineJoin)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_LineJoin, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) ((((((((((enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_LineJoin)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_LineJoin, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_LineJoin, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_LineJoin val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_LineJoin val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_LineJoin(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_LineJoin");
    return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_LineJoin");
    return (enum __pyx_t_7celiagg_8_celiagg_LineJoin) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_InnerJoin __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_InnerJoin neg_one = (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) ((enum __pyx_t_7celiagg_8_celiagg_InnerJoin) 0 - (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) (((((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) (((((((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) (((((((((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) (((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) ((((((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) (((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) ((((((((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) (((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) ((((((((((enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_InnerJoin)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_InnerJoin, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_InnerJoin val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_InnerJoin val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_InnerJoin");
    return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_InnerJoin");
    return (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_PixelFormat __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_PixelFormat neg_one = (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) ((enum __pyx_t_7celiagg_8_celiagg_PixelFormat) 0 - (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) (((((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) (((((((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) (((((((((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) (((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) ((((((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) (((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) ((((((((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) (((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) ((((((((((enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PixelFormat)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PixelFormat) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_PixelFormat, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_PixelFormat val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_PixelFormat val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PixelFormat(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_PixelFormat");
    return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_PixelFormat");
    return (enum __pyx_t_7celiagg_8_celiagg_PixelFormat) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_GradientSpread __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_GradientSpread neg_one = (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) ((enum __pyx_t_7celiagg_8_celiagg_GradientSpread) 0 - (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) (((((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) (((((((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) (((((((((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) (((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) ((((((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) (((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) ((((((((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) (((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) ((((((((((enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientSpread)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_GradientSpread, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_GradientSpread val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_GradientSpread val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_GradientSpread");
    return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_GradientSpread");
    return (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_GradientUnits __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_GradientUnits neg_one = (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) ((enum __pyx_t_7celiagg_8_celiagg_GradientUnits) 0 - (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) (((((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) (((((((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) (((((((((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) (((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) ((((((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) (((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) ((((((((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) (((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) ((((((((((enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_GradientUnits)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_GradientUnits, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_GradientUnits val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_GradientUnits val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_GradientUnits");
    return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_GradientUnits");
    return (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE enum __pyx_t_7celiagg_8_celiagg_PatternStyle __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(PyObject *x) {
    const enum __pyx_t_7celiagg_8_celiagg_PatternStyle neg_one = (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) ((enum __pyx_t_7celiagg_8_celiagg_PatternStyle) 0 - (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) 0;
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, digit, digits[0])
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) >= 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) (((((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) >= 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) (((((((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) >= 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) (((((((((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) 0;
                case -1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) (((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)-1)*(((((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) - 1 > 2 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) ((((((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) (((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)-1)*(((((((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) - 1 > 3 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) ((((((((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) (((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)-1)*(((((((((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) - 1 > 4 * PyLong_SHIFT) {
                            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) ((((((((((enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[3]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[2]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[1]) << PyLong_SHIFT) | (enum __pyx_t_7celiagg_8_celiagg_PatternStyle)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_PatternStyle) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(enum __pyx_t_7celiagg_8_celiagg_PatternStyle, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            enum __pyx_t_7celiagg_8_celiagg_PatternStyle val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) -1;
        }
    } else {
        enum __pyx_t_7celiagg_8_celiagg_PatternStyle val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) -1;
        val = __Pyx_PyInt_As_enum____pyx_t_7celiagg_8_celiagg_PatternStyle(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to enum __pyx_t_7celiagg_8_celiagg_PatternStyle");
    return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to enum __pyx_t_7celiagg_8_celiagg_PatternStyle");
    return (enum __pyx_t_7celiagg_8_celiagg_PatternStyle) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE unsigned char __Pyx_PyInt_As_unsigned_char(PyObject *x) {
    const unsigned char neg_one = (unsigned char) ((unsigned char) 0 - (unsigned char) 1), const_zero = (unsigned char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned char, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 2 * PyLong_SHIFT) {
                            return (unsigned char) (((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 3 * PyLong_SHIFT) {
                            return (unsigned char) (((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 4 * PyLong_SHIFT) {
                            return (unsigned char) (((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned char) ((((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned char) ((((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned char) ((((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned char) -1;
        }
    } else {
        unsigned char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned char) -1;
        val = __Pyx_PyInt_As_unsigned_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned char");
    return (unsigned char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned char");
    return (unsigned char) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
    const char neg_one = (char) ((char) 0 - (char) 1), const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_BlendMode(enum __pyx_t_7celiagg_8_celiagg_BlendMode value) {
    const enum __pyx_t_7celiagg_8_celiagg_BlendMode neg_one = (enum __pyx_t_7celiagg_8_celiagg_BlendMode) ((enum __pyx_t_7celiagg_8_celiagg_BlendMode) 0 - (enum __pyx_t_7celiagg_8_celiagg_BlendMode) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_BlendMode) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_BlendMode),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_DrawingMode(enum __pyx_t_7celiagg_8_celiagg_DrawingMode value) {
    const enum __pyx_t_7celiagg_8_celiagg_DrawingMode neg_one = (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) ((enum __pyx_t_7celiagg_8_celiagg_DrawingMode) 0 - (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_DrawingMode) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_DrawingMode),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_FontCacheType(enum __pyx_t_7celiagg_8_celiagg_FontCacheType value) {
    const enum __pyx_t_7celiagg_8_celiagg_FontCacheType neg_one = (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) ((enum __pyx_t_7celiagg_8_celiagg_FontCacheType) 0 - (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_FontCacheType) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_FontCacheType),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientSpread(enum __pyx_t_7celiagg_8_celiagg_GradientSpread value) {
    const enum __pyx_t_7celiagg_8_celiagg_GradientSpread neg_one = (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) ((enum __pyx_t_7celiagg_8_celiagg_GradientSpread) 0 - (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_GradientSpread) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientSpread),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_GradientUnits(enum __pyx_t_7celiagg_8_celiagg_GradientUnits value) {
    const enum __pyx_t_7celiagg_8_celiagg_GradientUnits neg_one = (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) ((enum __pyx_t_7celiagg_8_celiagg_GradientUnits) 0 - (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_GradientUnits) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_GradientUnits),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_InnerJoin(enum __pyx_t_7celiagg_8_celiagg_InnerJoin value) {
    const enum __pyx_t_7celiagg_8_celiagg_InnerJoin neg_one = (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) ((enum __pyx_t_7celiagg_8_celiagg_InnerJoin) 0 - (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_InnerJoin) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_InnerJoin),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineCap(enum __pyx_t_7celiagg_8_celiagg_LineCap value) {
    const enum __pyx_t_7celiagg_8_celiagg_LineCap neg_one = (enum __pyx_t_7celiagg_8_celiagg_LineCap) ((enum __pyx_t_7celiagg_8_celiagg_LineCap) 0 - (enum __pyx_t_7celiagg_8_celiagg_LineCap) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_LineCap) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_LineCap),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_LineJoin(enum __pyx_t_7celiagg_8_celiagg_LineJoin value) {
    const enum __pyx_t_7celiagg_8_celiagg_LineJoin neg_one = (enum __pyx_t_7celiagg_8_celiagg_LineJoin) ((enum __pyx_t_7celiagg_8_celiagg_LineJoin) 0 - (enum __pyx_t_7celiagg_8_celiagg_LineJoin) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_LineJoin) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_LineJoin),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_7celiagg_8_celiagg_TextDrawingMode(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode value) {
    const enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode neg_one = (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) ((enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) 0 - (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) 1), const_zero = (enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7celiagg_8_celiagg_TextDrawingMode),
                                     little, !is_unsigned);
    }
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_npy_uint8(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn_npy_uint8, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_npy_uint16(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn_npy_uint16, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_float(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_float, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn_npy_uint8(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 3,
                                                 &__Pyx_TypeInfo_nn_npy_uint8, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_d_dc_nn_npy_uint16(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 3,
                                                 &__Pyx_TypeInfo_nn_npy_uint16, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_npy_int32(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn_npy_int32, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dc_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 1,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* PyObjectGetMethod */
  static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
  static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
  #include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
  static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
  #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
  static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
